/* Example of plotting packets using the default data conversions.  This
 * example uses plplot, though any C lib plotting package would work.
 * 
 * PLplot can be installed via:
 *
 *   yum install plplot
 *
 * If that fails vist http://plplot.sourceforge.net to get the library.
 *
 * Build using the commands:
 *
 *
 * gcc -Wall -fexceptions -rdynamic -fPIC -ggdb -DGSE_POSIX_THREADS \
 *      -I /usr/local/juno/include -I /usr/local/plplot \
 *      -o waves_read.o -c waves_read.c
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb waves_read.o \
 *      /usr/local/juno/lib/libjwaves_t.a  /usr/local/juno/lib/librpwgse_t.a \
 *      -o waves_read -lm -ldl -lpthread
 *
 * And make sure the WAVES_DATA environment variable is set before running.
 */
#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

/* Source URL for librpwgse:
 * https://saturn.physics.uiowa.edu/svn/util/C/trunk/librpwgse
 */
#include <rpwgse/Gse.h>

/* Source URL for libjwaves:
 * https://ec.physics.uiowa.edu/control/juno_hw/GSE/trunk/libjwaves
 */
#include <jwaves/JWaves.h>

#include <plplot.h>

/*****************************************************************************/
void init_plplot(GseXYdata* pData)
{
	size_t nPts = 0;
	const float4* pXvals = NULL;
	int nAxisType = 0;
	uint4 uHints = 0U;
	char* sXlab;
	char* sYlab;
	char* sTopLab;
	
	/* Setup plplot */
	plsdev("Xwin");
	plinit();
	
	pXvals = pData->Xdata(pData);
	nPts   = pData->length(pData);
	uHints = pData->hints(pData);
	
	/* if(GSE_HNT_X_LOG(uHints)) nAxisType += 10; */
	/* if(GSE_HNT_Y_LOG(uHints)) nAxisType += 20; */
	
	plenv(pData->pXvals[0], pData->pXvals[nPts - 1],
			pData->yLoLimit(pData), pData->yHiLimit(pData), 0, 0);
	
	sXlab = gse_string("%s (%s)", pData->Xtype(pData), pData->Xunit(pData));
	sYlab = gse_string("%s (%s)", pData->Xtype(pData), pData->Xunit(pData));
	sTopLab = gse_copystr(pData->channelName(pData));
	
	pllab(sXlab, sYLab, sTopLab);
}

/*****************************************************************************/
int main(int argc, char** argv)
{
	char*          sFile    = NULL;
	GsePktSrc*     pSrc     = NULL;
	GsePkt*        pPkt     = NULL;
	JWavesSDSblk*  pSci     = NULL;
	GseXYdataAry*  pDataSets = NULL;
	GseXYdata*     pData    = NULL;
	int            nPkt     = 0;
	byte           uPktType = 0;
	float4*        pXvals   = NULL;
	float4*        pYvals   = NULL;
	size_t         nPts     = 0;
	bool           bInitPlplot = true;
	
	gse_thread_init(GSE_NOTICE, NULL);              /* Always do this first */
	
	if(argc > 1)                                    /* Get the file name */
		sFile = argv[1];
	else
		sFile = jwaves_latest_file(".useg", GSE_CRITICAL);
	
	gse_notice_v("Reading from %s\n", sFile);       /* Log the file used */
	
	
	pSrc = (GsePktSrc*) new_GseFilePktSrc(sFile, jwaves_blk_fact, at_eof_stop);
	
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
		nPkt += 1;
		
		/* Get the Science Data Status header block */
		pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		/* Save the id of the 1st packet type encountered, always plot that */
		if(uPktType == 0){
			uPktType = pSci->uPSID;
		}
		else{
			if(pSci->uPSID != uPktType)
				continue;
		}
		
		/* Get the packet data as an array of XY data structures.  Not that
		 * some packets contain more than one dataset (e.g. noise reduction
		 * diagnostic packets) so an array return type is needed.  However this
		 * example only uses the first dataset returned. */
		pDataSets = fact_JWavesXYdataAry(pPkt, 0);
		pData = pDataSets->pAry[0];
		
		if(bInitPlPlot) init_plplot(pData);
		
		pXvals = pData->Xdata(pData);
		nPts   = pData->length(pData);
				
		plline(
		
		
		/* We no longer need this packet, free it */
		pPkt->del(pPkt);
	}
	

	return 0;
}
