/* 
Example of plotting packets using the default data conversions.  This example
uses plplot, though any C lib plotting package would work.
  
PLplot can be installed via:

   yum install plplot

If that fails vist http://plplot.sourceforge.net to get the library.

Build using the commands:

gcc -Wall -fexceptions -rdynamic -fPIC -ggdb -DGSE_POSIX_THREADS \
    -I /usr/local/juno/include -I /usr/include/plplot \
    -o waves_plot.o -c waves_plot.c
 
gcc -Wall -fexceptions -rdynamic -fPIC -ggdb waves_plot.o \
    /usr/local/juno/lib/libjwaves_t.a  /usr/local/juno/lib/librpwgse_t.a \
    -o waves_plot -lplplotd -lm -ldl -lpthread
 
 
The program supports all plplot command line arguments.  So for example you
can specify the geometry of the top level window and a specific file to read
using:

./waves_plot -geometry 800x500 FILE_NAME

 
Make sure the WAVES_DATA environment variable is set before running.

*/
#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

/* Source URL for librpwgse:
 * https://saturn.physics.uiowa.edu/svn/util/C/trunk/librpwgse
 */
#include <rpwgse/Gse.h>

/* Source URL for libjwaves:
 * https://ec.physics.uiowa.edu/control/juno_hw/GSE/trunk/libjwaves
 */
#include <jwaves/JWaves.h>

#include <plplot.h>

#define MAX_PTS 65536

/*****************************************************************************/
void plot_data(GseXYdata* pData)
{
	size_t i, nPts;
	const float4* pXvals;
	const float4* pYvals;
	char* sTopLab;
	double X[MAX_PTS], Y[MAX_PTS];                 
	
	pXvals = pData->Xdata(pData);
	pYvals = pData->Ydata(pData);
	nPts   = pData->length(pData);
	
	plenv(pXvals[0], pXvals[nPts-1], pData->yLoLimit(pData), 
			pData->yHiLimit(pData), 0, 0);
	
	sTopLab = gse_string("%s @ %s", pData->channelName(pData),
			                          pData->captureSCLK(pData));
	
	pllab(pData->Xunit(pData), pData->Yunit(pData), sTopLab);
	
	if(nPts > MAX_PTS){
		gse_error_v("Datasets greater than %d points are not supported", MAX_PTS);
		return;
	}
	
	for(i = 0; i < nPts; i++){           /* PlPlot expects doubles */
		X[i] = (double)pXvals[i];
		Y[i] = (double)pYvals[i];
	}
	
	plline(nPts, (double*)X, (double*)Y);
	plflush();
	gse_notice_v("Plotted %s", sTopLab);
}

/*****************************************************************************/
int main(int argc, char** argv)
{
	char*          sFile    = NULL;
	GsePktSrc*     pSrc     = NULL;
	GsePkt*        pPkt     = NULL;
	JWavesSDSblk*  pSci     = NULL;
	GseXYdataAry*  pDataSets = NULL;
	GseXYdata*     pData    = NULL;
	int            nPkt     = 0;
	byte           uPktType = 0;
	
	gse_thread_init(GSE_NOTICE, NULL);              /* Always do this first */
	
	plparseopts(&argc, argv, 0);                    /* Let plplot eat it's args*/
	
	if(argc > 1)                                    /* Get the file name */
		sFile = argv[1];
	else
		sFile = jwaves_latest_file(".useg", GSE_CRITICAL);
	
	gse_notice_v("Reading from %s", sFile);         /* Log the file used */
	
	plsdev("xwin");                                 /* Setup plplot */
	plspause(0);
	plinit();
	
	pSrc = (GsePktSrc*) new_GseFilePktSrc(sFile, jwaves_blocks, at_eof_stop);
	
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
		nPkt += 1;
		gse_notice_v("Reading pkt: %04d", nPkt);
		
		/* Skip data that is not science, or is segmented, compressed, etc. 
		 * Content type codes are defined per mission, see:
		 *
		 *   GSE/trunk/libjwaves/doc/waves_gs_fmt_stor.txt
		 *
		 * for Juno Waves content codes.
		 */
		if(pPkt->pHdr->uContent != 0x84)
			continue;
		
		/* Get the Science Data Status header block */
		pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		/* Save the id of the 1st packet type encountered */
		if(uPktType == 0)
			uPktType = pSci->uPSID;
		
		/* To make for a smoother display, only display one type of packet in
		   the plot */
		if(pSci->uPSID != uPktType)                  
			continue;
		
		/* Get the packet data as an array of XY data structures.  Note that
		 * some packets contain more than one dataset (e.g. noise reduction
		 * diagnostic packets) so an array return type is needed.  However this
		 * example only uses the first dataset returned. */
		pDataSets = fact_JWavesXYdataAry(pPkt, 0);
		pData = pDataSets->pAry[0];
		
		plot_data(pData);
		
		pDataSets->del(pDataSets);                   /* free the datasets */
		pPkt->del(pPkt);                             /* free the packet */
		
		gse_millisleep(1000,false);                  /* Cross platform sleep */
	}
	
	gse_notice("Done");
	plend();
	return 0;
}
