/* Example of iterating over juno waves packets in a file, and printing
 * some of them.
 *
 * On Linux x86_64 compile with the commands:
 *
 *  source /opt/project/juno/Linux.x86_64/setup_env.sh
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb -DGSE_POSIX_THREADS \
 *      -I $PREFIX/include -o waves_read.o -c waves_read.c
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb waves_read.o \
 *      $PREFIX/lib64/libjwaves_t.a  -L $PREFIX/lib64 
 *      -o waves_read -lrpwgse_t -lm -ldl -lpthread
 *
 *
 * On the Waves GSE computers compile with the commands:
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb -DGSE_POSIX_THREADS \
 *      -I $PREFIX/include -o waves_read.o -c waves_read.c
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb waves_read.o \
 *      $PREFIX/lib/libjwaves_t.a  -L $PREFIX/lib 
 *      -o waves_read -lrpwgse_t -lm -ldl -lpthread
 *
 *
 * For compiling in other locations, note that $PREFIX is the top level of
 * the juno architecture specific directory.
 *
 * And make sure the WAVES_DATA environment variable is set before running.
 */
#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

/* Source URL for librpwgse:
 * https://saturn.physics.uiowa.edu/svn/util/C/trunk/librpwgse
 */
#include <rpwgse/Gse.h>

/* Source URL for libjwaves:
 * https://ec.physics.uiowa.edu/control/juno_hw/GSE/trunk/libjwaves
 */
#include <jwaves/JWaves.h>


int main(int argc, char** argv)
{
	char*          sFile = NULL;
	GsePktSrc*     pSrc = NULL;
	GsePkt*        pPkt = NULL;
	JWavesSDSblk*  pSci = NULL;
	cbytebuf_t     rawData = CBBUF_NULL;
	int            nPkt = 0;
	
	/* Always do this first */
	gse_thread_init(GSE_NOTICE, NULL);
	
	/* Get the file name from the command line or autoselect */
	if(argc > 1)
		sFile = argv[1];
	else
		sFile = jwaves_latest_file(".ucal", GSE_CRITICAL);
	
	gse_notice_v("Reading from %s", sFile);
	
	pSrc = (GsePktSrc*) new_GseFilePktSrc(sFile, jwaves_blocks, at_eof_stop);
	
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
		nPkt += 1;
		
		/* Get the Science Data Status header block */
		pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		/* Only print LFR Lo E Spectrum data, ignore the rest */
		if(pSci->uPSID == 0x92){
		
			/* Get the data as raw bytes */
			rawData = pPkt->getData(pPkt);
		
			/* Print them using a simple Hex printer */
			fprintf(stdout, "Data in packet no: %d\n", nPkt);
			gse_fmt_hexprn(stdout, rawData, 4, 2);
		}
		else
			fprintf(stdout, "Packet no: %d, Pkt ID & Src 0x%02hhX is not LFR Lo "
					  "E.\n", nPkt, pSci->uPSID);
		
		/* We no longer need this packet, free it */
		pPkt->del(pPkt);
	}
	

	return 0;
}
