/* Example of looking up packets by time and category and printing some 
 * information from each one.  Also serves as an examples of using GNU argp for
 * arugment processing.
 *
 *
 * On Linux x86_64 compile with the commands:
 *
 *  source /opt/project/juno/Linux.x86_64/setup_env.sh
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb -DGSE_POSIX_THREADS \
 *      -I $PREFIX/include -o waves_read_multi.o -c waves_read_multi.c
 *
 *  gcc -Wall -fexceptions -rdynamic -fPIC -ggdb waves_read_multi.o \
 *      $PREFIX/lib64/libjwaves_t.a  -L $PREFIX/lib64 
 *      -o waves_read_multi -lrpwgse_t -lm -ldl -lpthread
 *
 *
 *
 */

#define _POSIX_C_SOURCE 200112

#include <pthread.h>
#include <argp.h>
#include <strings.h>

#include <rpwgse/Gse.h>
#include <jwaves/JWaves.h>


/*****************************************************************************/
/* Argument Parsing */

typedef struct opts{
	const char* sGroup;
	char* sBeg;
	char* sEnd;
	const char* sChannels;
	gse_eof_disp eof_disp;
	int nLevel;
}Opts;

const char* argp_program_version = "waves_multi_read $Rev: 3424 $ $Date: 2009-07-08 17:43:01 -0500 (Wed, 08 Jul 2009) $";
const char* argp_program_bug_address = "<YOU@YOUR.EMAIL.ADDRESS.ORG>";
static char g_doc[] = 
"\n"
"your_program - Summary of program's purpose\n"
"Options:\v"
"Environment:\n"
"  WAVES_DATA is taken to be the root directory for the Juno Waves data\n"
"   storage tree.\n"
"\n"
"Example:\n"
"\n"
"Processing all flight model hrs ucal packets from antenna deployment:\n"
"\n"
"   your_program -T 2011-221,2011-223 -G flight -C hrs_rec.raw,hrs_bin.raw\n"
"\n"
"Continuously processing LRS QL data as it arrives in real time:\n"
"   your_program -T now,forever -G flight/ql -C lrs.raw\n"
"\n"
"Doing Something Else:\n"
"\n"
"   Your Command line example \n"
"\n";

static struct argp_option g_options[] = {
	
	{"log", 'l', "LEVEL", 0,
	 "Set the log verbosity, on of 'critical', 'error', 'warning', 'notice',"
	 " 'debug', or 'trace'.\n"},
	
	{"group", 'G', "GROUP", 0, 
	 "The data group, for example 'flight' or 'naked6/svt/2011_08_15_lvco_d5'. "
	 "The default is 'flight'.\n"
	},
			
	{"time", 'T', "TIME1,TIME2", 0,
	 "The time period of data to read.  TIME1 and TIME2 may be specified as a "
	 "Hexidecimal SCLK (starting with 0x), a signed decimal time offset from "
	 "now in minutes, or the special values 'now' or 'forever'.  The default "
	 "is -20,forever or \"Twenty minutes ago until forever\".\n"
	},
	
	{"channel", 'C', "CHAN1[,CHAN2,...]", 0,
	 "A comma separated list of data channels to read (no spaces).  The "
	 "supported data channels are: 'hsk.raw','lrs.raw','hrs_rec.raw', "
	 "'hrs_bin.raw','lrs.ucal','hrs_rec.ucal', and 'hrs_bin.ucal'.  The default "
	 "is all high rate data, 'hrs_rec.ucal,hrs_bin.ucal'\n"
	},

	{0}
};

static error_t parse_opt(int nKey, char* sArg, struct argp_state *state)
{
	Opts* pOpts = state->input;
	
	switch(nKey){
		case 'G':  pOpts->sGroup = sArg;            break;
		
		/* Pull out the begin an end times */
		case 'T':  
			pOpts->sBeg = CS(sArg);
			pOpts->sEnd = strchr(pOpts->sBeg, ',');

			if(pOpts->sEnd != NULL){
				*(pOpts->sEnd) = '\0';
				pOpts->sEnd += 1;
			}
			if(!pOpts->sEnd || (pOpts->sEnd == '\0') || (pOpts->sBeg == '\0') ){
				fprintf(stderr, "ERROR: Bad time range for -T argument: '%s'\n", sArg);
				return EINVAL;
			}
			
			/* Forever for an end value implies keep reading when hitting EOF */
			if(strcasecmp(pOpts->sEnd, "forever") == 0)
				pOpts->eof_disp = at_eof_block;
			else
				pOpts->eof_disp = at_eof_stop;
			break;
		
		case 'C':  pOpts->sChannels = sArg;         break;
		
		case 'l':
			switch(sArg[0]){
				case 'c': pOpts->nLevel = GSE_CRITICAL; break;
				case 'e': pOpts->nLevel = GSE_ERROR;    break;
				case 'w': pOpts->nLevel = GSE_WARNING;  break;
				case 'n': case 'i': pOpts->nLevel = GSE_NOTICE; break;
				case 'd': pOpts->nLevel = GSE_DEBUG;    break;
				case 't': pOpts->nLevel = GSE_TRACE;    break;
				default:
					return EINVAL;
			}
			break;
		case ARGP_KEY_ARG:    /* This program has no positional parameters */
			argp_usage(state); 
			break;
			
		default:
			return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

/* Structure tieing together all the parts of our arugment parser */
static struct argp g_parser = {g_options, parse_opt, NULL, g_doc}; 

/*****************************************************************************/
/* Main Program */
int main(int argc, char** argv)
{
	/* Set up default options */
	Opts opts;
	
	opts.sGroup = "flight";
	opts.sBeg = CS("-20");
	opts.sEnd = CS("forever");
	opts.sChannels = "hrs_rec.ucal,hrs_bin.ucal";
	opts.nLevel = GSE_NOTICE;
	opts.eof_disp = at_eof_block;
	
	/* Parse the command line */
	argp_parse(&g_parser, argc, argv, 0, 0, &opts);
	
	/* Init the library */
	gse_thread_init(opts.nLevel, NULL);	
	
	/* Setup the data source */
	GseFileChanMgr* pMgr = NULL;
	GsePktSrc*      pSrc = NULL;
	
	int nTimeOut = 10; /* How long to wait for re-try if a file we are parsing 
	                      is deleted while we are looking at it */
	
	
	pMgr = (GseFileChanMgr*)new_JWavesChanMgr(opts.sBeg, opts.sEnd,
	                                          opts.sChannels, opts.sGroup);
	pSrc = (GsePktSrc*)new_GseMultiFileSrc(pMgr, jwaves_blocks, opts.eof_disp, nTimeOut);
	
	
	/* The reading loop */
	GsePkt* pPkt = NULL;
	
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
		
		/* Do something... */
		fprintf(stdout, "Read packet %d from %s\n", pSrc->pktIdx(pSrc),
				  pSrc->name(pSrc));
		
		pPkt->del(pPkt);
	}
	
	if(pSrc != NULL) pSrc->del(pSrc);
	
	gse_thread_deinit();
	return 0;
}
		
	
			
	
