/** @file JWavesCCSDSblk.h */
#ifndef _JWavesCCSDSblk_H_
#define _JWavesCCSDSblk_H_

#include <sys/time.h>
#include <time.h>

#include <rpwgse/GseBlock.h>
#include <jwaves/JWavesDefs.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_jwaves_pkt */
/*@{*/

/** Interface for Juno CCSDS packet headers.
 * These are usually only available for ATLO housekeeping data.
 *
 * @class JWavesCCSDSblk
 * @extends GseBlock
 */
typedef struct jwaves_ccsds_blk{
	
	/** Base structure */
	GseBlock base;
	
	/** FIELD: The CCSDS Version number, defaults to 0x00 */
	byte uVer;

	/** FIELD: Packet type, defaults to 0 (telemetry) */
	byte uPktType;
	
	/** FIELD: Secondary Header Flag, defaults to 1 (always present) */
	byte uSecHdr;
	
	/** FIELD: Application Process Identifer (APID).
	 * For Juno Waves these are:
	 * 96 - Housekeeping data
	 * 97 - HRS data
	 * 98 - LRS data
	 * 99 - Unknown to Juno FDS, but was read from Waves.
	 */
	uint2 uAPID;
	
	/** FIELD: Grouping flags:
	 *
	 *  - 0 - Continuation of a segmented series
	 *  - 1 - First Packet of a segmented series
	 *  - 2 - Final Packet of a segmented series
	 *  - 3 - Non-segmented packet
	 *
	 * This is probably ignored by the waves instrument, as segmentation is
	 * handled in the minipacket header.
	 */
	byte uGrouping;
	
	
	/** FIELD: Data product part number.
	 * From the Flight Ground ICD:
	 *   Data products are partitioned into parts, with one part per
	 *   packet.  The packet with a count of zero contains the data product's
	 *   Metadata Protocol Data Unit.  The packets with count's one through
	 *   N contain the data product's individual File Data Units.  The packet
	 *   with a count of N also contains the data product's End-of-File
	 *   Protocol Data Unit.
	 *
	 * This is encoded as a 14-bit field so the maximum value is 16383.
	 * @note This is usually called the Source Sequence Count in general
	 *       CCSDS documentation.
	 */
	uint2 uPartNo;
	
	
	/** FIELD: Packet Length, (length of waves packet + 4)
	 * 
	 * Length of the packet, not including the primary header (first 6 bytes),
	 * expressed as a number of bytes minus one.
	 *
	 * On Juno the maximum length of a packet will be 8192, including headers.
	 */
	uint2 uPktLen;
	
	/** FIELD: Packet Emitt SCLK seconds.
	 * We think this time is assigned when reading the packet via the Juno FDS.
	 */
	uint4 uSCLK;
			
	/** FIELD: Packet Emitt SCLK fractional seconds. 
	 * We think this time is assigned when reading the packet via the Juno FDS.
	 */
	byte uSubSec;
	
	/** Destructor */
	void (*del)(void* vpThis);
	 
} JWavesCCSDSblk;

/** Generate a new Juno Waves CCSDS block. 
 * @public
 * @memberof JWavesCCSDSblk
 */
JWavesCCSDSblk* new_JWavesCCSDSblk(cbytebuf_t buf);


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesCCSDSblk_H_ */

