#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <errno.h>
#include <strings.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseTime.h>
#include <rpwgse/GseFile.h>

#include "JWavesStore.h"
#include "JWavesConv.h"
#include "JWavesSDSblk.h"
#include "JWavesMroBlk.h"
#include "JWavesUpkBlk.h"

#include "JWavesChanMgr.h"

/*****************************************************************************/
/* Globals */

const char* jwaves_file_types[] = {
	"hsk.raw",
	"lrs.raw",
	"hrs_rec.raw",
	"hrs_bin.raw",
	"lrs.ucal",
	"hrs_rec.ucal",
	"hrs_bin.ucal",
	"hsk.dat",
	"srv.pkt",
	"rec.pkt",
	"bst.pkt",
	NULL
};


/*****************************************************************************/
/* Making, removing file channels */

void delete_JWavesFileChan(void* vpThis)
{
	JWavesFileChan* pThis = (JWavesFileChan*)vpThis;
	
	pThis->pFiles->del(pThis->pFiles);
	
	free(pThis->sType);
	free(vpThis);
}

/* WARNING: I take control of the sType argument ! */
JWavesFileChan* new_JWavesFileChan(char* sType)
{
	JWavesFileChan* pThis = NULL;
	
	pThis = (JWavesFileChan*)gse_calloc(1, sizeof(JWavesFileChan));
	pThis->sType = sType;
	pThis->pFiles = new_JWavesDirList();
	pThis->bValid = true;
	pThis->bNeedsSort = false;
	pThis->del = delete_JWavesFileChan;
	
	return pThis;
}

/*****************************************************************************/
/*****************************************************************************/

size_t JWavesChanMgr_channels(const GseFileChanMgr* pBase)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*)pBase;
	return pThis->pChAry->num;
}

const char* JWavesChanMgr_begIdx(const GseFileChanMgr* pBase)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*)pBase;
	return pThis->sBegIdx;
}


const char* JWavesChanMgr_endIdx(const GseFileChanMgr* pBase)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*)pBase;
	return pThis->sEndIdx;	
}

char* JWavesChanMgr_getPktIdx(const GseFileChanMgr* pBase, const GsePkt* pPkt)
{
	float8 dSclk;
	uint2 nSeq = 0;
	byte iMP = 0;
	JWavesUpkBlk* pUpk = NULL;
	cbytebuf_t cbuf;
	uint4 uSclk = 0;
	
	/* nTs   = jwaves_pkt2ts(pPkt); */
	nSeq = jwaves_pkt2seq(pPkt);
	
	if(pPkt->hasUniqueBlkByType(pPkt, 0x70, JW_UPACK_PROC_STEP_ID)){
		pUpk = (JWavesUpkBlk*) pPkt->getBlkByType(pPkt, 0x70, 
		                                          JW_UPACK_PROC_STEP_ID);
		iMP = pUpk->iMiniPkt;
		dSclk = jwaves_pkt2sclk(pPkt);
	}
	else{
		/* Just use the SCLK from the waves header, don't try to be smart */
		cbuf = pPkt->getData(pPkt);
		uSclk = gse_u4FromMSBbytes(cbuf.ptr + 4);
		dSclk = (float8)uSclk; 
	}
	
	return NS("%014.3f:%05d:%03d", dSclk, nSeq, iMP);
}

/* Reuse channel type names as much as possible */
char* JWavesChanMgr_getPktCat(const GseFileChanMgr* pBase, int4 iChan,
		                        const GsePkt* pPkt)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*)pBase;
	
	cbytebuf_t cbuf;
	byte uPSID;
	const JWavesSDSblk* pSds = NULL;
	
	if((pPkt->pHdr->uContent & 0xF0) == 0x60)
		return CS("hsk.raw");
	
	if((pPkt->pHdr->uContent & 0xF0) == 0x70)
		return CS("bst.raw");
		
	if( ((pPkt->pHdr->uContent & 0xF0) != 0x80 ) &&
	    ((pPkt->pHdr->uContent & 0xF0) != 0xE0 ) )
		
		gse_throw(GSE_XCEPT, GSE_XASSERT);
	
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	if((pPkt->pHdr->uContent & 0x0F) == 0x1 ){
		
		if(strcmp(pThis->pChAry->pAry[iChan]->sType, "lrs.raw") == 0){
			
			if((pPkt->pHdr->uContent & 0xF0) == 0xE0 )
				return CS("mro.raw");
			else
				return CS("lrs.raw");
		}
		
		/* The packet ID for HRS data is in bytes 8 and 13 of the packet data */	
		cbuf = pPkt->getData(pPkt);		
		uPSID = ((cbuf.ptr[8]) & 0xF0) | ((cbuf.ptr[13]) & 0x0F);
		return NS("hrs.raw:0x%02X", uPSID);
	}
	
	/* I'm dealing with unpacked science data */
	
	if((pPkt->pHdr->uContent & 0xF0) == 0xE0)
		return CS("sci:mro");
	
	pSds = (const JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	return NS("sci:0x%02X", pSds->uPSID);
}
	

int4 JWavesChanMgr_compare(const GseFileChanMgr* pBase, const char* sIdx1,
		                     const char* sIdx2)
{
	int4 nRet;
	const char* pIdx1;
	const char* pIdx2;
	
	/* fprintf(stderr, "sIdx1 %s,  sIdx2 %s\n", sIdx1, sIdx2); */
	
	if((strlen(sIdx1) != 24) || (strlen(sIdx2) != 24)){
		
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "JWaves files channel manager indici"
		             "es are supposed to be 24 character strings (sIdx1 '%s', sIdx2 "
		             "'%s')", sIdx1, sIdx2);
	}
	
	nRet = strncmp(sIdx1, sIdx2, 14);
	if(nRet != 0) return nRet;
	
	pIdx1 = sIdx1 + 15;
	pIdx2 = sIdx2 + 15;
	nRet = strncmp(pIdx1, pIdx2, 5);
	
	/* Check for wrap around, give a 400 count boundary centered on the wrap */
	if(nRet > 0){
		
		if( (strncmp(pIdx1, "16184", 5) > 0) && (strncmp(pIdx2, "00200", 5) < 0) )
			return -1;
	}
	
	if(nRet < 0){
		if( (strncmp(pIdx2, "16184", 5) > 0) && (strncmp(pIdx1, "00200", 5) < 0) )
			return 1;	
	}
	
	if(nRet != 0) return nRet;
	
	pIdx1 = sIdx1 + 21;
	pIdx2 = sIdx2 + 21;
	
	return strcmp(pIdx1, pIdx2);
}

/*****************************************************************************/
/* Check to see if an RPWGSE file is in range based on it's filename, 
   dumb helper, returns the channel type name if in range. */

const char* JWavesChanMgr_gseFileInRng(JWavesChanMgr* pThis, const char* sPath)
{
	const char* sType = NULL;
	const char** pType = NULL;
	const char* sFile = NULL;
	gse_caltime_t beg = {0,0,0,0,0,0.0};
	gse_caltime_t end = {0,0,0,0,0,0.0};
	struct timeval tv = {0,0};
	float8 dFileBeg = 0.0, dFileEnd = 0.0;
	
	char sFmt[40];
	
	/* basename is not MT-safe, so find the directory character on your own */
	if( (sFile = rindex(sPath, '/') + 1) == (void*)0x1){
		sFile = sPath;
	}
	else{
		if(sFile == sPath + strlen(sPath))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Trailing slash in '%s'", sPath);
	}
	
	pType = jwaves_file_types;
	while(*pType != NULL){
		if(gse_endswith(sFile, *pType)){
			sType = *pType;	
			break;
		}
		pType++;
	}
	
	if(sType == NULL) return NULL;
	
	memset(sFmt, 0, 40);
	
	sprintf(sFmt, "%%4d_%%2d_%%2dT%%2d.%s", sType);
	
	if(sscanf(sFile, sFmt, &beg.year, &beg.month, &beg.dom, &beg.hour) < 4)
		return NULL;
	
	gse_ct_norm_dom(&beg);
	
	end = beg;
	end.hour++;
	gse_ct_norm_dom(&end);
	
	if( ! gse_ct2tv_gm(&beg, &tv, false) ) return NULL;
	dFileBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
	
	if( ! gse_ct2tv_gm(&end, &tv, false) ) return NULL;
	dFileEnd = jwaves_tv2sclk_d(&tv) + pThis->dSclkSlop;
	
	/* Range overlap check */
	if( (dFileBeg <= pThis->dBegSclk) && (pThis->dBegSclk < dFileEnd) ) return sType;
	if( (dFileBeg < pThis->dEndSclk) && (pThis->dEndSclk <= dFileEnd)) return sType;
	if( (pThis->dBegSclk <= dFileBeg) && (dFileEnd < pThis->dEndSclk)) return sType;

	return NULL;
}

/*****************************************************************************/
/* Check to see if a PDS file is in range based on it's filename, 
   dumb helper, returns the channel type name if in range. */

const char* JWavesChanMgr_pdsFileInRng(JWavesChanMgr* pThis, const char* sPath)
{
	const char* sFile = NULL;
	const char* sType = NULL;
	char sDate[18];
	struct timeval tv;
	float8 dFileBeg = -1.0, dFileEnd = -1.0;
	
	memset(sDate, 0, 18);
	
	/* basename is not MT-safe, so find the directory character on your own */
	if( (sFile = rindex(sPath, '/') + 1) == (void*)0x1){
		sFile = sPath;
	}
	else{
		if(sFile == sPath + strlen(sPath))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Trailing slash in '%s'", sPath);
	}
	
	if(strstr(sFile, "HSK") && gse_endswith(sFile, ".DAT")){
		if( strlen(sFile) != 23) return NULL;
		sType = "hsk.dat";
	}
	
	if(!sType && strstr(sFile, "SRV") && gse_endswith(sFile, ".PKT")){
		if( strlen(sFile) != 23) return NULL;
		sType = "srv.pkt";
	}
	
	/* All filetypes have the first section */
	memcpy(sDate, sFile + 4, 4);
	sDate[4] = '-';
	memcpy(sDate + 5, sFile + 8, 3);
	
	if(sType != NULL){
		if( ! gse_str2tv_gm(sDate, &tv, false)) return NULL;
		dFileBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
		dFileEnd = dFileBeg + 24.0*60.0*60.0 + pThis->dSclkSlop;			
	}
	else{
	
		if(!sType && strstr(sFile, "REC") && gse_endswith(sFile, ".PKT")){
			if( strlen(sFile) != 30) return NULL;
			sType = "rec.pkt";
		}
		
		if(!sType && strstr(sFile, "BST") && gse_endswith(sFile, ".PKT")){
			if( strlen(sFile) != 30) return NULL;
			sType = "bst.pkt";
		}
	
		if(!sType) return NULL;
		
		sDate[8] = 'T';
		memcpy(sDate + 9, sFile + 12, 2);
		sDate[11] = ':';
		memcpy(sDate + 12, sFile + 14, 2);
		sDate[14] = ':';
		memcpy(sDate + 15, sFile + 16, 2);
	
		if( ! gse_str2tv_gm(sDate, &tv, false)) return NULL;
		dFileBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
		
		/* Assume 1 hour max for high rate files */
		dFileEnd = dFileBeg + 60.0*60.0 + pThis->dSclkSlop;
		
	}
	
	/* Range overlap check */
	if( (dFileBeg <= pThis->dBegSclk) && (pThis->dBegSclk < dFileEnd) ) return sType;
	if( (dFileBeg < pThis->dEndSclk) && (pThis->dEndSclk <= dFileEnd)) return sType;
	if( (pThis->dBegSclk <= dFileBeg) && (dFileEnd < pThis->dEndSclk)) return sType;

	return NULL;

}
	
/*****************************************************************************/
/* Helper for path invalidation */
bool jwaves_is_in_path(void* sMovedDir, const char* sPath)
{
	if(strstr(sPath, sMovedDir)) return true;
	else return false;
}

/*****************************************************************************/
/* Returns true if all channels are still valid, false if one or more have 
   been invalided */

bool JWavesChanMgr_update(GseFileChanMgr* pBase)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	
	bool bValid = true;
	int4 nChanges, i;
	JWavesFileChan* pChannel = NULL;
	GseTreeChange* pChange = NULL;
	JWavesDirList* pFiles;
	
	const char* sType = NULL;
	
	nChanges = pThis->pWatch->numChanges(pThis->pWatch);
	
	if(nChanges == 0) return true;
	
	while((pChange = pThis->pWatch->getChange(pThis->pWatch)) != NULL){
		
		
		switch(pChange->cChange){
		
			case 'I':
			case 'D':
			for(i = 0; i < pThis->pChAry->num; i++){
				pChannel = pThis->pChAry->pAry[i];
				pFiles = pChannel->pFiles;
				if( pFiles->rmWhere(pFiles, jwaves_is_in_path, pChange->sPath) > 0)
					gse_info_v("File %s removed", pChange->sPath);
				else
					gse_debug_v("File %s removed", pChange->sPath);
				
				if(pChannel->sCurFile && strstr(pChange->sPath, pChannel->sCurFile)){
					pChannel->bValid = false;
					bValid = false;
				}
			}
			break;
		
			
			case 'C':
			
			/* A new file has been created, do I care about it? */
			if(strstr(pThis->sTop, "rpwgse"))
				sType = JWavesChanMgr_gseFileInRng(pThis, pChange->sPath);			
			else
				sType = JWavesChanMgr_pdsFileInRng(pThis, pChange->sPath);
		
			if(sType != NULL){
				gse_info_v("File %s added to filesystem.", pChange->sPath);
				pChannel = NULL;
				for(i = 0; i<pThis->pChAry->num; i++){
					if(strcmp(sType, pThis->pChAry->pAry[i]->sType) == 0){					
						pChannel = pThis->pChAry->pAry[i];
						pChannel->pFiles->append(pChannel->pFiles, CS(pChange->sPath));
						pChannel->bNeedsSort = true;	
					}
				}
			}
			else{
				gse_debug_v("File %s added to filesystem.", pChange->sPath);
			}
			break;
				
					
			default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
			break;		
		}
		pChange->del(pChange);
		
	}
	
	/* Re-sort any channels that need it... */
	for(i = 0; i < pThis->pChAry->num; i++){
		pChannel = pThis->pChAry->pAry[i];
		if(pChannel->bNeedsSort){
			pFiles = pChannel->pFiles;
			pFiles->sort(pFiles, strcmp);
			pChannel->bNeedsSort = false;
		}
	}
	
	return bValid;
}

/*****************************************************************************/
const char* JWavesChanMgr_chanName(const GseFileChanMgr* pBase, int4 iChan)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	return pThis->pChAry->pAry[iChan]->sType;
}

/*****************************************************************************/
bool JWavesChanMgr_isValid(GseFileChanMgr* pBase, int4 iChan)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	return pThis->pChAry->pAry[iChan]->bValid;
}

/*****************************************************************************/
void JWavesChanMgr_clearInvalid(GseFileChanMgr* pBase, int4 iChan)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	pThis->pChAry->pAry[iChan]->bValid = true;
}

/*****************************************************************************/
const char* JWavesChanMgr_getFile(GseFileChanMgr* pBase, int4 iChan, 
                                  const char* sIdx)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	JWavesFileChan* pChannel = NULL;
	JWavesDirList* pList = NULL;
	float8 dIdx, dFileBeg;
	struct timeval tv;
	const char* sFile;
	const char* sPath;
	char sFmt[40];
	char sDate[18];
	gse_caltime_t beg = {0,0,0,0,0,0.0};
	
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);

	
	/* Start at the bottom of the channel, compute lowest 1st packet times
	   in the file.  If the time is < sIdx, save it and go to the next one.
		the 1st file >= sIdx stops the loop.  If no files are less than the
		index, return the first file in the channel, or NULL if there are
		no files in the channel.
	*/
							 
	pChannel = pThis->pChAry->pAry[iChan];
	pList = pChannel->pFiles;
	
	free(pChannel->sCurFile);
	pChannel->sCurFile = NULL;
	
	pList = pChannel->pFiles;
	dFileBeg = -1.0;
	
	
	if( sIdx == NULL) {
		pChannel->sCurFile = pList->get(pList, 0);
		return pChannel->sCurFile;
	}
	else{
		if( ! gse_str2float8(sIdx, &dIdx) )
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Can't convert '%s' to a floating"
			             " point value.", sIdx);
	}

	
	for(sPath = pList->get(pList, 0); sPath != NULL; sPath = pList->next(pList)){
		
		/* Get the filename part from the stored path */
		if( (sFile = rindex(sPath, '/') + 1) == (void*)0x1)
			sFile = sPath;
		
		/* Get the earliest possible file start date */
		if(strstr(sPath, "rpwgse")){
			memset(sFmt, 0, 40);
			sprintf(sFmt, "%%4d_%%2d_%%2dT%%2d.%s", pChannel->sType);
	
			if(sscanf(sFile, sFmt, &beg.year, &beg.month, &beg.dom, &beg.hour) < 4)
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			gse_ct_norm_dom(&beg);
			if( ! gse_ct2tv_gm(&beg, &tv, false) ) 
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			dFileBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
		}
		else{
			
			memset(sDate, 0, 18);
			memcpy(sDate, sFile + 4, 4);
			sDate[4] = '-';
			memcpy(sDate + 5, sFile + 8, 3);
	
			if((strcmp(pChannel->sType, "hsk.dat") == 0)||
			   (strcmp(pChannel->sType, "srv.pkt") == 0)){
				
				sDate[8] = 'T';
				memcpy(sDate + 9, sFile + 12, 2);
				sDate[11] = ':';
				memcpy(sDate + 12, sFile + 14, 2);
				sDate[14] = ':';
				memcpy(sDate + 15, sFile + 16, 2);
			}
			
			if( ! gse_str2tv_gm(sDate, &tv, false))
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			dFileBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
		}
		
		if(dFileBeg < dIdx){
			free(pChannel->sCurFile);
			pChannel->sCurFile = CS(sPath);
		}
		else{
			break;
		}
	}
	
	/* Maybe nothing was less than dIdx, if so just return the 1st file */
	if(pChannel->sCurFile == NULL)
		pChannel->sCurFile = pList->get(pList, 0);
				
	return pChannel->sCurFile;
}


/*****************************************************************************/
/* Don't set pChan->sCurFile to NULL, ever! */

const char* JWavesChanMgr_nextFile(GseFileChanMgr* pBase, int4 iChan,
		                             const char* sCurFile)
{	
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	JWavesFileChan* pChan = NULL;
	JWavesDirList* pFiles = NULL;
	const char* sPath = NULL;
	bool bGetNext = false;
	
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	pChan = pThis->pChAry->pAry[iChan];
		
	if(sCurFile == NULL){
		free(pChan->sCurFile);
		pChan->sCurFile = CS( pChan->pFiles->get(pChan->pFiles, 0) );
		return pChan->sCurFile;
	}
	
	pFiles = pChan->pFiles;
	bGetNext = false;
	for(sPath=pFiles->get(pFiles,0); sPath!=NULL; sPath=pFiles->next(pFiles)){
		
		if(bGetNext){ 
			free(pChan->sCurFile);
			pChan->sCurFile = CS(sPath);
			return pChan->sCurFile;
			break;
		}
		
		if(strcmp(sPath, sCurFile) == 0)
			bGetNext = true;
	}
	
	if(! bGetNext)
		gse_error_v("nextFile: File %s doesn't exist in channel %s",
		            sCurFile, pChan->sType);
		
	return NULL;
}


/*****************************************************************************/
void JWavesChanMgr_clearCurFile(GseFileChanMgr* pBase, int4 iChan)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*) pBase;
	
	if(iChan < 0 || iChan >= pThis->pChAry->num)
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "Expected a channel number from "
		             "0 to %d, received %d", 0, pThis->pChAry->num - 1, iChan);
	
	free(pThis->pChAry->pAry[iChan]->sCurFile);
	pThis->pChAry->pAry[iChan]->sCurFile = NULL;
}
	
/*****************************************************************************/
/* Check to see if a month directory is in range, dumb loadGse helper */
/* dSclkSlop -- Sometimes packets don't get put in the proper places, usually */
/*           allow for a bit of seconds of sloppiness in the sclk values */

bool JWavesChanMgr_dirInRng(JWavesChanMgr* pThis, const char* sMonthDir,
		                      const char* sDayDir)
{
	gse_caltime_t beg = {0,0,0,0,0,0.0};
	gse_caltime_t end = {0,0,0,0,0,0.0};
	
	float8 dDirBeg = 0, dDirEnd = 0;
	struct timeval tv = {0,0};
	
	if(strcmp(".", sMonthDir) == 0) return false;
	if(strcmp("..", sMonthDir) == 0) return false;
	
	if(sDayDir){
		if(strcmp(".", sDayDir) == 0) return false;
		if(strcmp("..", sDayDir) == 0) return false;		
	}
	
	/* Does the month entry match the YYYY_MM pattern ? */
	if( sscanf(sMonthDir, "%4d_%2d", &beg.year, &beg.month) < 2)
		return false;
	
	/* If we have a day entry, does the entry match the DD pattern? */
	if(sDayDir){
		if( sscanf(sDayDir, "%2d", &beg.dom) < 1 )
			return false;
	}
	else{
		beg.dom = 1;
	}
	gse_ct_setdoy(&beg);
	
	/* Is this even a legal juno date? */
	if(!jwaves_legal_ct(&beg)) return false;
	
	/* Get the start and stop times for the directory in question. */
	end = beg;
	if(sDayDir == NULL)
		end.month += 1;
	else
		end.dom += 1;
	
	gse_ct_norm_dom(&end);
	
	if( ! gse_ct2tv_gm(&beg, &tv, false) ) return false;
	dDirBeg = jwaves_tv2sclk_d(&tv) - pThis->dSclkSlop;
	
	if( ! gse_ct2tv_gm(&end, &tv, false) ) return false;
	dDirEnd = jwaves_tv2sclk_d(&tv) + pThis->dSclkSlop;
	
		
	/* Range overlap check */
	if( (dDirBeg <= pThis->dBegSclk) && (pThis->dBegSclk < dDirEnd) ) return true;
	if( (dDirBeg < pThis->dEndSclk) && (pThis->dEndSclk <= dDirEnd)) return true;
	if( (pThis->dBegSclk <= dDirBeg) && (dDirEnd < pThis->dEndSclk)) return true;

	return false;
}


/*****************************************************************************/
/* Loading GSE files of interest */

/* dStart - Starting SCLK
 * dStop - up to, but excluding SCLK 
 */
void JWavesChanMgr_loadGse(JWavesChanMgr* pThis, const char* sRpwDir)
{
	char sErrBuf[128];
	
	DIR* pTopDir = NULL;
	struct dirent* pTopEnt = NULL;
	
	char* sMonthDir = NULL;
	DIR* pMonthDir = NULL;	
	struct dirent* pMonthEnt = NULL;
	
	char* sDayDir = NULL;
	DIR* pDayDir = NULL;
	struct dirent* pDayEnt = NULL;
	
	const char* sType = NULL;
	
	char* sHourFile = NULL;
	bool bAdded = false;
	int4 i;
	
	memset(sErrBuf, 0, 128);
	JWavesFileChan* pChannel = NULL;
	
	/* Look through the month directories for day directories */		
	if( (pTopDir = opendir(sRpwDir)) == NULL){
		gse_warning_v("Could not open %s: %s", sRpwDir,
				        strerror_r(errno, sErrBuf, 127));
		return;
	}
	
	/* Top level reading ... */
	while( (pTopEnt = readdir(pTopDir)) != NULL){
		
		/* Am I going to care about this month tree ? */
		if(! JWavesChanMgr_dirInRng(pThis, pTopEnt->d_name, NULL) )
			continue;
		
		sMonthDir = NS("%s/%s", sRpwDir, pTopEnt->d_name);
		if(!gse_isdir(sMonthDir)){
			free(sMonthDir);
			continue;
		}
		
		if( (pMonthDir = opendir(sMonthDir)) == NULL){
			gse_warning_v("Could not open %s: %s", sMonthDir,
				        strerror_r(errno, sErrBuf, 127));
			free(sMonthDir);
			continue;
		}
		
		/* Month level reading .... */
		while( (pMonthEnt = gse_readdir(pMonthDir)) != NULL){
				
			/* Am I going to care about this daily dir ? */
			if(!JWavesChanMgr_dirInRng(pThis, pTopEnt->d_name, pMonthEnt->d_name))
				continue;
					                            
			sDayDir = NS("%s/%s", sMonthDir, pMonthEnt->d_name);
			if(!gse_isdir(sDayDir)){
				free(sDayDir);
				continue;
			}
			
			if( (pDayDir = opendir(sDayDir)) == NULL){
				gse_warning_v("Could not open %s: %s", sDayDir,
				        strerror_r(errno, sErrBuf, 127));
				free(sDayDir);
				continue;	
			}
			
			/* Day level reading ... */
			while( (pDayEnt = gse_readdir(pDayDir)) != NULL){
				
				sType = JWavesChanMgr_gseFileInRng(pThis, pDayEnt->d_name);
				if(!sType)
					continue;
				
				sHourFile = NS("%s/%s", sDayDir, pDayEnt->d_name);
				if(!gse_isfile(sHourFile)){
					free(sHourFile);
					continue;
				}
				
				bAdded = false;	
				for(i = 0; i<pThis->pChAry->num; i++){
					pChannel = pThis->pChAry->pAry[i];
					if(strcmp(sType, pChannel->sType) == 0){
						gse_debug_v("Adding %s to file channel %s", sHourFile,
						            pChannel->sType);
						pChannel->pFiles->append(pChannel->pFiles, sHourFile);
						bAdded = true;
					}
				}
				
				if(!bAdded) free(sHourFile);
			}
			
			free(sDayDir);
		}
		
		free(sMonthDir);
	}
	
	/* No need to free sRpwDir, it's passed in */
	
	/* Sort the file channels */
	for(i = 0; i < pThis->pChAry->num; i++){
		pChannel = pThis->pChAry->pAry[i];
		pChannel->pFiles->sort(pChannel->pFiles, strcmp);
		pChannel->bNeedsSort = false;
	}	
}

/*****************************************************************************/
/* Dumb loadPds helper check to see if a 100-day dir is in range */

bool JWavesChanMgr_100dirInRng(JWavesChanMgr* pThis, const char* s100Dir,
                               const char* sDayDir)
{
	int4 n100bin;
	
	gse_caltime_t beg = {0,0,0,0,0,0.0};
	gse_caltime_t end = {0,0,0,0,0,0.0};
	
	float8 dDirBeg = 0, dDirEnd = 0;
	struct timeval tv = {0,0};
	
	if(strcmp(".", s100Dir) == 0) return false;
	if(strcmp("..", s100Dir) == 0) return false;
	
	if(sDayDir){
		if(strcmp(".", sDayDir) == 0) return false;
		if(strcmp("..", sDayDir) == 0) return false;		
	}
	
	
	/* Does the 100-day entry match the YYYY_NXX pattern? */
	if( sscanf(s100Dir, "%4d_%1dXX", &beg.year, &n100bin) < 2)
		return false;
	
	/* If we have a day entry, does the entry match the DDD pattern? */
	if(sDayDir){
		if( sscanf(sDayDir, "%3d", &beg.doy) < 1 )
			return false;
	}
	else{
		beg.doy =  (n100bin == 0 ? 1 : n100bin * 100);
	}
	gse_ct_setdom(&beg);
	
	/* Is this even a legal juno date? */
	if(!jwaves_legal_ct(&beg)) return false;
	
	
	/* Get the start and stop times for the directory in question. */
	end = beg;
	if(sDayDir == NULL){
		
		/* Do roll-over here, since it's a simple case */
		if(n100bin < 3){
			end.doy += 100;
		}
		else{
			end.doy = 1;
			end.year += 1;
		}
	}
	else
		end.doy += 1;
	
	gse_ct_norm_doy(&end);
	
	if( ! gse_ct2tv_gm(&beg, &tv, false) ) return false;
	dDirBeg = jwaves_tv2sclk_d(&tv);
	
	if( ! gse_ct2tv_gm(&end, &tv, false) ) return false;
	dDirEnd = jwaves_tv2sclk_d(&tv);
	
		
	/* Range overlap check */
	if( (dDirBeg <= pThis->dBegSclk) && (pThis->dBegSclk < dDirEnd) ) return true;
	if( (dDirBeg < pThis->dEndSclk) && (pThis->dEndSclk <= dDirEnd)) return true;
	if( (pThis->dBegSclk <= dDirBeg) && (dDirEnd < pThis->dEndSclk)) return true;

	return false;
}

/*****************************************************************************/
/* I know this is sufficiently similar to loadGse that the two could be a    */
/* single function, but since the PDS layout can change, I'm going to keep   */
/* them separate for now.                                                    */
	
void JWavesChanMgr_loadPds(JWavesChanMgr* pThis, const char* sVolData)
{
	char sErrBuf[128];
	
	DIR* pTopDir = NULL;
	struct dirent* pTopEnt = NULL;
	
	char* s100Dir = NULL;
	DIR* p100Dir = NULL;
	struct dirent* p100Ent = NULL;
	
	char* sDoyDir = NULL;
	DIR* pDoyDir = NULL;
	struct dirent* pDoyEnt = NULL;
	
	const char* sType = NULL;
	
	char* sDataFile = NULL;
	bool bAdded = false;
	int4 i;
	
	memset(sErrBuf, 0, 128);
	JWavesFileChan* pChannel = NULL;
	
	/* Look through 100-day directories for day directories */
	if( (pTopDir = opendir(sVolData)) == NULL){
		gse_warning_v("Could not open %s: %s", sVolData,
				        strerror_r(errno, sErrBuf, 127));
		return;
	}
	
	/* Top level reading ... */
	while( (pTopEnt = readdir(pTopDir)) != NULL){
		
		/* Am I going to care about this 100-day tree? */	
		if(! JWavesChanMgr_100dirInRng(pThis, pTopEnt->d_name, NULL) )
			continue;
		
		s100Dir = NS("%s/%s", sVolData, pTopEnt->d_name);
		if(!gse_isdir(s100Dir)){
			free(s100Dir);
			continue;
		}
		
		if( (p100Dir = opendir(s100Dir)) == NULL){
			gse_warning_v("Could not open %s: %s", s100Dir,
				        strerror_r(errno, sErrBuf, 127));
			free(s100Dir);
			continue;
		}
		
		
		/* 100-Day Dir level reading */
		while( (p100Ent = gse_readdir(p100Dir)) != NULL){
		
			/* Am I going to care about this daily dir ? */
			if(!JWavesChanMgr_100dirInRng(pThis, pTopEnt->d_name, p100Ent->d_name))
				continue;
			
			sDoyDir = NS("%s/%s", s100Dir, p100Ent->d_name);
			if(!gse_isdir(sDoyDir)){
				free(sDoyDir);
				continue;
			}
			
			if( (pDoyDir = opendir(sDoyDir)) == NULL){
				gse_warning_v("Could not open %s: %s", sDoyDir,
				        strerror_r(errno, sErrBuf, 127));
				free(sDoyDir);
				continue;	
			}
			
			/* Doy level reading ... */
			while( (pDoyEnt = gse_readdir(pDoyDir)) != NULL){
				
				sType = JWavesChanMgr_pdsFileInRng(pThis, pDoyEnt->d_name);
				if(!sType)
					continue;
				
				sDataFile = NS("%s/%s", sDoyDir, pDoyEnt->d_name);
				if(!gse_isfile(sDataFile)){
					free(sDataFile);
					continue;
				}
				
				bAdded = false;	
				for(i = 0; i<pThis->pChAry->num; i++){
					pChannel = pThis->pChAry->pAry[i];
					if(strcmp(sType, pChannel->sType) == 0){
						gse_debug_v("Adding %s to file channel %s", sDataFile,
						            pChannel->sType);
						pChannel->pFiles->append(pChannel->pFiles, sDataFile);
						bAdded = true;
					}
				}
				
				if(!bAdded) free(sDataFile);
			}
			free(sDoyDir);
			
		}
		free(s100Dir);
		
	}
	
	/* No need to free sVolData, it's passed in */
	
	/* Sort the file channels */
	for(i = 0; i < pThis->pChAry->num; i++){
		pChannel = pThis->pChAry->pAry[i];
		pChannel->pFiles->sort(pChannel->pFiles, strcmp);
		pChannel->bNeedsSort = false;
	}	

}


/*****************************************************************************/
void delete_JWavesChanMgr(void* vpThis)
{
	JWavesChanMgr* pThis = (JWavesChanMgr*)vpThis;
	
	if(pThis->pChAry != NULL) pThis->pChAry->del(pThis->pChAry);
	if(pThis->pWatch != NULL) pThis->pWatch->del(pThis->pWatch);
	free(pThis->sTop);
	free(pThis->sBegIdx);
	free(pThis->sEndIdx);
	free(vpThis);
}

/*****************************************************************************/

bool JWavesChanMgr_isPds(const char* sType){
	return ((strcmp(sType, "hsk.dat") == 0)||(strcmp(sType, "srv.pkt") == 0)||
			  (strcmp(sType, "rec.pkt") == 0)||(strcmp(sType, "bst.pkt") == 0)  );
}

JWavesChanMgr* new_JWavesChanMgr(const char* sBegTime, const char* sEndTime,
		                           const char* sTypes, const char* sGroup)
{
	JWavesChanMgr* pThis = NULL;
	GseFileChanMgr* pBase = NULL;
	bool bPds = false;
	bool bGse = false;
	bool bLegal = false;
	GseStrAry* pTypes = NULL;
	int4 j, i = 0;
	JWavesFileChan* pChannel = NULL;
	uint4 uFlags = 0;
	
	if((sTypes == NULL)||(sBegTime == NULL)||(sGroup == NULL))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	jwaves_ckGroup(sGroup, true); /* true = throw on problem */
	
	if(getenv("WAVES_DATA") == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XMISSINGVAR, "The WAVES_DATA Environmet "
		            "variable is not defined.");
	
	
	/* Check the types argument */
	pTypes = gse_strsplit(sTypes, ",");
	except_cleanup_push(pTypes->del, pTypes);
	
	for(i=0;i<pTypes->num;i++){
		bLegal = false;
		j = 0;
		while(jwaves_file_types[j] != NULL){
			if(strcmp(pTypes->pAry[i], jwaves_file_types[j]) == 0){
				bLegal = true;
				break;
			}
			j++;
		}
		if(!bLegal)
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Unknown file type '%s'", 
			             pTypes->pAry[i]);
		
		if(JWavesChanMgr_isPds(pTypes->pAry[i]))
			bPds = true;
		else
			bGse = true;
	}
	
	/* Get the top directory */
	if(bPds && bGse)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Both RPWGSE and PDS data cannot be"
		            "read from a single instance of the JWavesChanMgr class.");
		
	pThis = (JWavesChanMgr*)gse_calloc(1, sizeof(JWavesChanMgr));
	pBase = (GseFileChanMgr*)pThis;
	
	pBase->channels     = JWavesChanMgr_channels;
	pBase->chanName     = JWavesChanMgr_chanName;
	pBase->begIdx       = JWavesChanMgr_begIdx;
	pBase->endIdx       = JWavesChanMgr_endIdx;
	pBase->getPktIdx    = JWavesChanMgr_getPktIdx;
	pBase->getPktCat    = JWavesChanMgr_getPktCat;
	pBase->compare      = JWavesChanMgr_compare;
	
	pBase->update       = JWavesChanMgr_update;
	pBase->isValid      = JWavesChanMgr_isValid;
	pBase->clearInvalid = JWavesChanMgr_clearInvalid;
	
	pBase->getFile      = JWavesChanMgr_getFile;
	pBase->nextFile     = JWavesChanMgr_nextFile;
	pBase->clearCurFile = JWavesChanMgr_clearCurFile;
			
	pBase->del          = delete_JWavesChanMgr;
	
	pThis->del          = delete_JWavesChanMgr;
	except_cleanup_push(pThis->del, pThis);
	
	if(bGse)
		pThis->sTop = NS("%s/%s/rpwgse", getenv("WAVES_DATA"), sGroup);
	else
		pThis->sTop = NS("%s/%s/JNOWAV_0000/DATA", getenv("WAVES_DATA"), sGroup);

	
	/* Convert the start, stop times */
	pThis->dBegSclk = jwaves_str2sclk(sBegTime);
	
	if(sEndTime == NULL){
		pThis->dEndSclk = 0x7FFFFFFF;  /* Max Juno SCLK value */
	}
	else{
		pThis->dEndSclk  = jwaves_str2sclk(sEndTime);
		if(pThis->dBegSclk >= pThis->dEndSclk)
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Start and stop times are "
			             "the same '%s', '%s'", sBegTime, sEndTime);
	}

	pThis->dSclkSlop = 30*60.0;
	
	pThis->sBegIdx = NS("%014.3f:%05d:%03d", pThis->dBegSclk, 0, 0);
	pThis->sEndIdx = NS("%014.3f:%05d:%03d", pThis->dEndSclk, 0, 0);
	
	pThis->pChAry      = new_JWavesFileChAry(0);
	
	/* Make one file channel for each file type */
	for(i = 0; i < pTypes->num; i++){
		
		pChannel = new_JWavesFileChan(pTypes->pAry[i]); /* Takes the memory */
		pTypes->pAry[i] = NULL;
		pThis->pChAry->add(pThis->pChAry, pChannel);
	}

	
	/* Make a new file watcher 1st, then get the list of files, this may
	   cause us to get create messages about file we already know about
		but that's okay */
	uFlags = GSE_TW_CREATE|GSE_TW_DELETE|GSE_TW_FILES_ONLY;
	pThis->pWatch = new_GseTreeWatch(pThis->sTop, uFlags, NULL, NULL);
	
	/* Load the files of interest */
	if(bPds)
		JWavesChanMgr_loadPds(pThis, pThis->sTop);
	else
		JWavesChanMgr_loadGse(pThis, pThis->sTop);
	
	except_cleanup_pop(except_no_call); /* pThis */
		
	except_cleanup_pop(except_call); /* pTypes */
	
	return pThis;
}


/*****************************************************************************/
/* Template Instantiation */

#define C JWavesDirList
#define I char
#define CASCADE_FREE

#include <rpwgse/listTplt.c>

#undef CASCADE_FREE
#undef I
#undef C


#define C JWavesFileChAry
#define T JWavesFileChan
#define CASCADE_DEL

#include <rpwgse/ptrAryTplt.c>

#undef CASCADE_DEL
#undef T
#undef C




