#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#define _JWavesConv_C_

#include <math.h>
#include <strings.h>
#include <ctype.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseConv.h>
#include <rpwgse/GseTypes.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseLog.h>

#include "JWavesExcept.h"
#include "JWavesConv.h"
#include "JWavesDefs.h"
#include "JWavesSDSblk.h"
#include "JWavesMroBlk.h"

const int4 g_nJunoSclkOffset = 946728000;


/* Unfortunatly, the time has come to add leap second support.  

   How this is done: 
	
	When deriving a SCLK from a unix time, unix times above the values
	below cause an extra second to be added to the SCLK in addition to
	the straight subtraction of the offset.  For each value that is 
	surpassed, another second is added to the sclk.
	
	When deriving a unix time from a SCLK things are more complicated.
	
	if sclk + offset > repeat value + index, loose one from the unix time,
	where index is the number of the entry in the table below.
	
	Since there is not a one to one correspondance between the unix
	clock and the SCLK, certian SCLK values cannot be generated from
	unix time values.
*/
	
/* Unix SCLK repeats in the Juno Era */
static const int4 g_unixRepeats[] = {
	
	1136073600, /*Represents both: 2006-12-31T23:59:60 and 2007-01-01T00:00:00 */
	1230768000, /*Represents both: 2008-12-31T23:59:60 and 2009-01-01T00:00:00 */
	1325376000, /*Represents both: 2011-12-31T23:59:60 and 2012-01-01T00:00:00 */
	1341100800, /*Represents both: 2012-06-30T23:59:60 and 2012-07-01T00:00:00 */
	0,          /*End of table */
};


/*****************************************************************************/
/* Lookup tables to answer questions about packet data */

/****************************************************************************/

static const byte g_validPSID[] = {
	0x21, 0x26, 0x27, 0x44, 0x45, 0x47, 0x46, 0x5F, 0x5B, 0x7B, /* 10 */
	0x3F, 0x7F, 0x88, 0x89, 0x8C, 0x8D, 0x91, 0x92, 0x93, 0xA1, /* 20 */
   0x02, 0xA2, 0xA3, 0xA0, 0xA4, 0xA5, 0xF1, 0xF2, 0xF0, 0xF4, /* 30 */
	0xF5, 0x3B                                                  /* 32 */
};
	
static const int g_validPSID_sz = 32;

const byte g_hrs_data[] = { 0x7B, 0x7F, 0x3B, 0x3F, 0x88, 0x89, 0x8C, 0x8D,
                            0xA1, 0x02, 0xA2, 0xA3, 0xA0, 0xA4, 0xA5, 0xF1,
                            0xF2, 0x00 };


bool jwaves_valid_idsrc(byte uPSID, bool bThrow)
{
	int i;
	bool bRet = false;
	
	for(i = 0; i< g_validPSID_sz; i++){
		if(uPSID == g_validPSID[i]){
			bRet = true;
			break;
		}
	}
	if(!bRet){
		if(bThrow)
			gse_throw_fm(JW_XCEPT, JW_XPKTUNKNOWN, "Unknown Packet ID and Src "
			             "field combination: 0x%02X", uPSID);
		else
			gse_error_v("Unknown Packet ID and Src field combination: 0x%02X",
			             uPSID);
	}
	return bRet;
}

static const byte g_sectionedPSID[] = {0xA0, 0xA4, 0xA5, 0xF0, 0xF4, 0xF5};
static const int g_sectionedPSID_sz = 6;

bool jwaves_valid_section(byte uPSID, int nSec, bool bThrow)
{
	bool bRet = false;
	int i;
	if( jwaves_valid_idsrc(uPSID, bThrow) == false) return false;
	
	if(nSec == 0) return true;
	
	for(i = 0; i<g_sectionedPSID_sz; i++){
		if(uPSID == g_sectionedPSID[i]){
			if(nSec < 3){
				bRet = true;
				break;
			}
		}
	}
	
	if(!bRet){
		if(bThrow)
			gse_throw_fm(JW_XCEPT, JW_XPKTUNKNOWN, "Section %d is not valid for"
			             "packets of type: 0x%02X", nSec, uPSID);
		else
			gse_error_v("Section %d is not valid for packets of type: 0x%02X",
					      nSec, uPSID);
	}
	
	return bRet;
}

static const byte g_lfr_sci_data[] = {0x47, 0x46, 0x5F, 0x5B, 0x3B, 
                                      0x7B, 0x7F, 0x3F, 0x88, 0x89, 
                                      0x8C, 0x8D, 0xF1, 0xF2, 0x0};

static const byte g_hfr_sci_data[] = {0x91, 0x92, 0x93, 0xA1, 0x02, 0xA2, 
                                      0xA3, 0xA0, 0xA4, 0xA5, 0xF0, 0xF4, 
                                      0xF5, 0x0};
bool jwaves_islfrsci(byte uPSID){
	int i = 0;
	while(g_lfr_sci_data[i] != 0){
		if(uPSID == g_lfr_sci_data[i])
			return true;
		i++;
	}
	return false;
}

bool jwaves_ishfrsci(byte uPSID){
	int i = 0;
	while(g_hfr_sci_data[i] != 0){
		if(uPSID == g_hfr_sci_data[i])
			return true;
		i++;
	}
	return false;
}

static const byte g_b_sci_data[] = {0x91, 0xF0, 0xA1, 0x02, 0xA0, 0x0};
 
bool jwaves_isBsci(byte uPSID){
	int i = 0;
	while(g_b_sci_data[i] != 0){
		if(uPSID == g_b_sci_data[i])
			return true;
		i++;
	}
	return false;
}


static const byte g_logamp_data[] = {0x47, 0x46, 0x0};
static const byte g_wfrm_data[]   = {0x7B, 0x7F, 0x3F, 0x3B, 0x88, 0x89, 0x8C,
                                     0x8D, 0xA1, 0x02, 0xA2, 0xA3, 0xA0, 0xA4, 
                                     0xA5, 0x0 };
static const byte g_fft_data[]    = {0x5F, 0x5B, 0x91, 0x92, 0x93, 0xF1, 0xF2,
                                     0xF0, 0xF4, 0xF5, 0x0};
static const byte* g_data_kinds[] = {
  0x0,
  (const byte*)g_logamp_data, /* JWAVES_KIND_LOGAMP = 1*/
  (const byte*)g_wfrm_data,   /* JWAVES_KIND_WFRM = 2 */
  (const byte*)g_fft_data,    /* JWAVES_KIND_FFT = 3 */
};

int jwaves_dataKind(byte uPSID)
{
	int i, j;
	
	for(i=1; i<4; i++){
		j = 0;
		while(g_data_kinds[i][j] != 0){
			if(g_data_kinds[i][j] == uPSID)
				return i;
			j++;
		}
	}
	return 0;
}  

GseCStrAry* jwaves_channels(){
	int i, j;
	byte uPSID;
	GseCStrAry* pAry = new_GseCStrAry(0);
	
	
	for(i=0; i < g_validPSID_sz; i++){
		uPSID = g_validPSID[i];
		
		if(uPSID == 0xA0 || uPSID == 0xA4 || uPSID == 0xA5 ||
			uPSID == 0xF0 || uPSID == 0xF4 || uPSID == 0xF5){
			for(j=0; j<3; j++)
				pAry->add(pAry, jwaves_psid_sname(uPSID, j) );
		}
		else{
			pAry->add(pAry, jwaves_psid_sname(uPSID, 0) );
		}
	}
	return pAry;
}

const char* jwaves_psid_sname(byte uPSID, int section){
	switch(uPSID){
		case 0x21: return "LFR Q-Factor";
		case 0x27: return "HFR AGC";
		case 0x26: return "HFWBR AGC";
		case 0x44: return "HFWBR Lock";
		case 0x45: return "HFR Lock";
		case 0x47: return "HFR";
		case 0x46: return "HFR fallback";
		case 0x5F: return "HFR base";
		case 0x5B: return "HFR base fallback";
		case 0x7B: return "HFWBR base";
		case 0x3B: return "HFWBR base dual";
		case 0x3F: return "HFR base dual";
		case 0x7F: return "HFWBR base fallback";
		case 0x88: return "HFWBR I";
		case 0x89: return "HFWBR Q";
		case 0x8C: return "HFR I";
		case 0x8D: return "HFR Q";
		case 0x91: return "LFR B spec";
		case 0x92: return "LFR Lo E spec";
		case 0x93: return "LFR Hi E spec";
		case 0xA1: return "LFR B";
		case 0x02: return "LFR B";
		case 0xA2: return "LFR Lo E";
		case 0xA3: return "LFR Hi E";
		
		case 0xA0:
			switch(section){
				case 0: return "LFR B NRD S+N";
				case 1: return "LFR B NRD N";
				case 2: return "LFR B NRD S";
				case 3: return "LFR B NRD C";
				default: return NULL;
			}
			
		case 0xA4:
				switch(section){
				case 0: return "LFR Lo E NRD S+N";
				case 1: return "LFR Lo E NRD N";
				case 2: return "LFR Lo E NRD S";
				case 3: return "LFR Lo E NRD C";
				default: return NULL;
			}
			
		case 0xA5:
			switch(section){
				case 0: return "LFR Hi E NRD S+N";
				case 1: return "LFR Hi E NRD N";
				case 2: return "LFR Hi E NRD S";
				case 3: return "LFR Hi E NRD C";
				default: return NULL;
			}
			
		case 0xF1: return "HFWBR band spec";  /* Generated on the ground */
		case 0xF2: return "HFR band spec";    /* Generated on the ground */
		
		case 0xF0:                            /* Generated on the ground */
			switch(section){
				case 0: return "LFR B spec NRD S";
				case 1: return "LFR B spec NRD S+N";
				case 2: return "LFR B spec NRD N";
				default: return NULL;
			}
			
		case 0xF4:                            /* Generated on the ground */
				switch(section){
				case 0: return "LFR Lo E spec NRD S";
				case 1: return "LFR Lo E spec NRD S+N";
				case 2: return "LFR Lo E spec NRD N";
				default: return NULL;
			}
			
		case 0xF5:                            /* Generated on the ground */
			switch(section){
				case 0: return "LFR Hi E spec NRD S";
				case 1: return "LFR Hi E spec NRD S+N";
				case 2: return "LFR Hi E spec NRD N";
				default: return NULL;
			}
	}
	
	return NULL;
}

const char* jwaves_psid_lname(byte uPSID, int section, int nSCLK){
	
	bool bNew = nSCLK > 332859600 ? true : false;
	
	
	switch(uPSID){
		case 0x21: return "LFR Q-Factor";
		
		case 0x27: 
			if(bNew) return "Automatic Gain Control from 44";
			else     return "HFR Automatic Gain Control";
		
		case 0x26: 
			if(bNew) return "Automatic Gain Control from 45";
			else     return "HFWBR Automatic Gain Control";
		
		case 0x45: 
			if(bNew) return "PLL Lock from 44";
			else     return "HFR PLL Lock";
		
		case 0x44:
			if(bNew) return "PLL Lock on 45";
			else     return "HFWBR PLL Lock";
		
		case 0x47: 
			if(bNew) return "Sweep Spectrum from 44";
			else     return "HFR Sweep Spectrum";
		
		case 0x46: 
			if(bNew) return "Sweep Spectrum from 45";
			else     return "HFWBR Sweep Spectrum";
		
		case 0x5F: 
			if(bNew) return "Baseband Spectrum from 44";
			else     return "HFR Baseband Spectrum";
		
		case 0x5B: 
			if(bNew) return "Baseband Spectrum from 45";
			else     return "HFWBR Baseband Spectrum";
		
		case 0x7B: 
			if(bNew) return "Baseband Waveform from 45";
			else     return "HFWBR Baseband Waveform";
		
		case 0x3B: 
			if(bNew) return "Baseband Waveform from 45 (dual-route)";
			else     return "HFWBR Baseband Waveform";
			
		case 0x3F: 
			if(bNew) return "Baseband Waveform from 44 (dual-route)";
			else     return "HFR Baseband Waveform";
		
		case 0x7F: 
			if(bNew) return "Baseband Waveform from 44";
			else     return "HFWBR Baseband Waveform from HFR";
		
		case 0x88: 
			if(bNew) return "Downshifted I Waveform from 45";
			else     return "HFWBR I Mixed Waveform";
		
		case 0x89:
			if(bNew) return "Downshifted Q Waveform from 45";
			else return "HFWBR Q Mixed Waveform";
		
		case 0x8C: 
			if(bNew) return "Downshifted I Waveform from 44";
			else     return "HFR I Mixed Waveform";
		
		case 0x8D:  
			if(bNew) return "Downshifted Q Waveform from 44";
			else     return "HFR Q Mixed Waveform";
		
		case 0x91: return "LFR B Spectrum";
		case 0x92: return "LFR Lo E Spectrum";
		case 0x93: return "LFR Hi E Spectrum";
		case 0xA1: return "LFR B Waveform";
		case 0x02: return "LFR B Waveform";
		case 0xA2: return "LFR Lo E Waveform";
		case 0xA3: return "LFR Hi E Waveform";
		
		case 0xA0: 
			switch(section){
				case 0: return "LFR B NRD Signal+Noise";
				case 1: return "LFR B NRD Noise";
				case 2: return "LFR B NRD Signal";
				case 3: return "LFR B NRD Coefficients";
				default: return NULL;
			}
		
		case 0xA4: 
			switch(section){
				case 0: return "LFR Lo E NRD Signal+Noise";
				case 1: return "LFR Lo E NRD Noise";
				case 2: return "LFR Lo E NRD Signal";
				case 3: return "LFR Lo E NRD Coefficients";
				default: return NULL;
			}
			
		case 0xA5:
			switch(section){
				case 0: return "LFR Hi E NRD Signal+Noise";
				case 1: return "LFR Hi E NRD Noise";
				case 2: return "LFR Hi E NRD Signal";
				case 3: return "LFR Hi E NRD Coefficients";
				default: return NULL;
			}
			
		case 0xF1:                                /* Generated on the ground */
			if(bNew) return "Band Spectrum from 45";  
			else     return "HFWBR Band Spectrum";  
			
		case 0xF2:                                /* Generated on the ground */
			if(bNew) return "Band Spectrum from 44";
			else     return "HFR Band Spectrum";    
		
		case 0xF0:                                /* Generated on the ground */
			switch(section){
				case 0: return "LFR B NRD Spec. Signal";
				case 1: return "LFR B NRD Spec. Signal+Noise";
				case 2: return "LFR B NRD Spec. Noise";
				default: return NULL;
			}
		
		case 0xF4:                                /* Generated on the ground */
			switch(section){
				case 0: return "LFR Lo E NRD Spec. Signal";
				case 1: return "LFR Lo E NRD Spec. Signal+Noise";
				case 2: return "LFR Lo E NRD Spec. Noise";
				default: return NULL;
			}
			
		case 0xF5:                                /* Generated on the ground */
			switch(section){
				case 0: return "LFR Hi E NRD Spec. Signal";
				case 1: return "LFR Hi E NRD Spec. Signal+Noise";
				case 2: return "LFR Hi E NRD Spec. Noise";
				default: return NULL;
			}
	
	}
	
	return NULL;
}

/*****************************************************************************/
/* Compression questions */

static const byte jwaves_compressed_formats[] = {
	0x14, 0x13, 0x09, 0x0E, 0x0F, 0x00
};
	
bool jwaves_isCompressed(byte uFmt)
{
	byte uComp;
	int i = 0;
	while( (uComp = jwaves_compressed_formats[i]) != 0x00)
	{
		if(uFmt == uComp) return true;
		i++;
	}
	return false;
}

byte jwaves_decompFmt(byte uCompFmt)
{
	if(!jwaves_isCompressed(uCompFmt))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Format type 0x%02hhX is not "
				       "compressed", uCompFmt);
	
	switch(uCompFmt){
		case JW_8BIT_RICE:  return JW_8BIT_IN_16BIT_lo;
		case JW_12BIT_RICE: return JW_12BIT_IN_16BIT_lo;
		case JW_16BIT_RICE: return JW_16BIT_IN_16BIT;
		case JW_9BIT_IN_12BIT_lo_RICE:  return JW_9BIT_IN_12BIT_lo;
		case JW_9BIT_IN_16BIT_lo_RICE:  return JW_9BIT_IN_16BIT_lo;
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Unknown compressed format"
			             "0x%02hhX", uCompFmt);
	}
	return 0; /* Never gets here */
}


int jwaves_dataBlockBits(byte uFmt)
{
	switch(uFmt){
	case 0x4:
	case 0x10:
	case 0xA:
	case 0x14:
		return 8;
		
	case 0x5:
		return 9;
	
	case 0x13:
	case 0x15:
	case 0x6:
	case 0xB:
	case 0xE:
		return 12;
		
	case 0x11:
	case 0x09:
	case 0x7:
	case 0xC:
	case 0x1C:
	case 0x3:
	case 0xF:
		return 16;
		
	case 0x1B:
		return 24;
			
	case 0x8:
	case 0xD:
	case 0x12:
	case 0x2:
		return 32;
	
	default:
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Juno Waves data format 0x%02X "
		             "is not recognized", uFmt);
	}
	return 0; 
}


/*****************************************************************************/
/* 9-bit data */
float4 jwaves_conv_flt9(uint4 f9_lo)
{
	byte exp, frac;
	
	float4 ret = 0;
	byte* pRet;
	
	exp = (f9_lo >> 3)&0x3F;
	frac = f9_lo & 0x07;
	
	exp -= 20;
	exp += 127;
	
	pRet = (byte*)(&ret);
	if( g_bIsLSBfirst){		
		/* Set the high byte to contain most of the exponent, mask sign bit */
		*(pRet + 3) = (byte)( (exp >> 1) & 0x7F );
		
		/* set the next to highest byte to contian last of exponent + frac */
		*(pRet + 2) = (byte)((((exp & 0x01) << 7) & 0x80) | ((frac << 4)&0x70));
	}
	else{
		*(pRet) = (byte)((exp >> 1) & 0x7F);
		*(pRet + 1) = (byte)( (((exp & 0x01) << 7) & 0x80) | ((frac << 4)&0x70));
	}
	
	return ret; 
}

/*****************************************************************************/
/* Converting to floating point */

void init_jwaves_float4_conv(struct jwaves_float4_conv* pConv)
{
	pConv->ptr = NULL;
	pConv->len = 0;
	pConv->min = GSE_FLOAT4_MAX;
	pConv->max = -GSE_FLOAT4_MAX;
}

void fin_jwaves_float4_conv(void* vp)
{
	struct jwaves_float4_conv* pConv = (struct jwaves_float4_conv*)vp;
	if(pConv->ptr != NULL) free(pConv->ptr);
}

/* A helper for jwaves_bytes2float4, Assumes pConv points to enough memory to
 * hold all the samples.  Here's a little chart to keep the shifts straight:
 *
 *   Byte 0   Byte 1   Byte 2   Byte 3   Byte 4   Byte 5   Byte 6   Byte 7   Byte 8  
 * +--------+--------+--------+--------+--------+--------+--------+--------+--------+
 * |76543210|76543210|76543210|76543210|76543210|76543210|76543210|76543210|76543210|
 * +--------+--------+--------+--------+--------+--------+--------+--------+--------+
 *  87654321 08765432 10876543 21087654 32108765 43210876 54321087 65432108 76543210
 *           ^         ^         ^         ^         ^         ^         ^         ^
 *  Sample 0  Sample 1  Sample 2  Sample 3  Sample 4  Sample 5  Sample 6  Sample 7
 */
void jwaves_9in9_upk(struct jwaves_float4_conv* pConv, cbytebuf_t cbuf)
{
	int i, j;
	uint4 u4;
	
	j = -1;
	for(i = 0; i<cbuf.len; i += 9){
				
		if(i+1 < cbuf.len){                        /* Sample 0 */
			u4 = ((uint4)cbuf.ptr[i] << 1) & 0x1FE;
			u4 |= (cbuf.ptr[i+1] >> 7) & 0x001;
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
				
		if(i+2 < cbuf.len){                        /* Sample 1 */
			u4 = ((uint4)cbuf.ptr[i+1] << 2) & 0x1FC;
			u4 |= (cbuf.ptr[i+2] >> 6) & 0x003;
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
				
		if(i+3 < cbuf.len){                         /* Sample 2 */
			u4 = ((uint4)cbuf.ptr[i+2] << 3) & 0x1F8;
			u4 |= (cbuf.ptr[i+3] >> 5) & 0x007;
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
				
		if(i+4 < cbuf.len){                         /* Sample 3 */
			u4 = ((uint4)cbuf.ptr[i+3] << 4) & 0x1F0;
			u4 |= (cbuf.ptr[i+4] >> 4) & 0x00F;	
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
				
		if(i+5 < cbuf.len){                         /* Sample 4 */
			u4 = ((uint4)cbuf.ptr[i+4] << 5) & 0x1E0;
			u4 |= (cbuf.ptr[i+5] >> 3) & 0x01F;	
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
		
		if(i+6 < cbuf.len){                         /* Sample 5 */
			u4 = ((uint4)cbuf.ptr[i+5] << 6) & 0x1C0;
			u4 |= (cbuf.ptr[i+6] >> 2) & 0x03F;	
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
		
		if(i+7 < cbuf.len){                         /* Sample 6 */
			u4 = ((uint4)cbuf.ptr[i+6] << 7) & 0x180;
			u4 |= (cbuf.ptr[i+7] >> 1) & 0x07F;	
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
		
		if(i+8 < cbuf.len){                         /* Sample 7 */
			u4 = ((uint4)cbuf.ptr[i+7] << 8) & 0x100;
			u4 |= cbuf.ptr[i+8];
			
			pConv->ptr[++j] = jwaves_conv_flt9( u4 );
		}
	}
		
	for(j=0; j<pConv->len; j++){
		if(pConv->ptr[j] > pConv->max) pConv->max = pConv->ptr[j];
		if(pConv->ptr[j] < pConv->min) pConv->min = pConv->ptr[j];
	}
			
}

/****************************************************************************/
cbytebuf_t jwaves_loc_nrd_spec_data(const GsePkt* pPkt, int iPart)
{
	const JWavesSDSblk* pSci;
	int nBlkBits;
	size_t iOffset = 0;
	size_t nLen = 0;
	cbytebuf_t cbuf = CBBUF_NULL;
	const char* sType;
	
	if((iPart < 0)||(iPart > 2)) gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	pSci = (const JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	nBlkBits = jwaves_dataBlockBits(pSci->uFmt);
	
	/* The offset and lengths are based on:
	 * 1. The data block size.
	 * 2. LFR Hi or Lo band.
	 * 3. The section desired.
	 * 4. The fact that the sub blocks are on 4-byte boundaries.
	 */
	if((pSci->uPSID == 0xF0)||(pSci->uPSID == 0xF4)){
		switch(nBlkBits){
		case 8:
			iOffset = iPart * (43 + 1);
			nLen = 43;
			break;
			
		case 9:
			iOffset = iPart * 52; /* ceil(43 * (9.0/8.0) ) + 3 */
			nLen = 49;            /* ceil(43 * (9.0/8.0) ) */
			break;
		
		case 16:
			iOffset = iPart * (43*2 + 2);
			nLen = 86;
			break;
			
		case 32:
			iOffset = iPart * (43*4);
			nLen = 172;
			break;
		default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		}
	}
	else{
		switch(nBlkBits){
		case 8:
			iOffset = iPart * (18 + 2);
			nLen = 18;
			break;
			
		case 9:
			iOffset = iPart * 24; /* ceil(18 * (9.0/8.0) ) + 3 */
			nLen = 21;            /* ceil(18 * (9.0/8.0) ) */
			break;
		
		case 16:
			iOffset = iPart * (18*2);
			nLen = 36;
			break;
			
		case 32:
			iOffset = iPart * (18*4);
			nLen = 72;
			break;
		default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		}
		
	}
	
	cbuf = pPkt->getData(pPkt);
	
	if( iOffset + nLen > cbuf.len){
		switch(pSci->uPSID){
			case 0xF0: sType = "LFR B NRD spectrum"; break;
			case 0xF4: sType = "LFR Lo E NRD spectrum"; break;
			case 0xF5: sType = "LFR Hi E NRD spectrum"; break;
			default: gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		}
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Packet is too short to contain "
		             "section %d of 3 for the %s data @ SCLK %X", iPart + 1, 
		             sType, pSci->uColSCLK);
	}
	
	cbuf.len = nLen;
	cbuf.ptr = cbuf.ptr + iOffset;
	
	return cbuf;
}

/*****************************************************************************/
size_t jwaves_bytes2float4(byte uFmt, cbytebuf_t bbuf, uint4 flags,
                           struct jwaves_float4_conv* pResult)
{
	size_t i;
	int nBlkBits = 0;
	uint2 u2;
	uint4 u4;
	
	init_jwaves_float4_conv(pResult);
	
	if(bbuf.len == 0) return 0;
	
	if(flags & JW_CONV_INT_SIGNED)
		gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Converting signed integers is "
		            "not yet implimented");
	
	if(jwaves_isCompressed(uFmt))
		gse_throw_m(JW_XCEPT, JW_XPKTNOTREADY, "Packet data requires "
		             "decompression before it can be read as float4");
	
	nBlkBits = jwaves_dataBlockBits(uFmt);
	
	/* Allocate memory */
	switch(nBlkBits){
		case 8:   pResult->len = bbuf.len; break;
		case 16:  pResult->len = (bbuf.len)/2; break;
		case 32:  pResult->len = (bbuf.len)/4; break;
		
		case 9:
			pResult->len = bbuf.len * 8 / 9;   /* truncate num samples */
			break;
		
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "Handling %d bit items is "
			             "not yet implimented.", nBlkBits);
	}
	
	pResult->ptr = (float4*)gse_calloc(pResult->len, sizeof(float4));
	except_cleanup_push(fin_jwaves_float4_conv, pResult);
		
	
	switch(uFmt){
		case 0x2:  /* The trivial case, we are already 32bit floats */
			for(i = 0; i<pResult->len; i++){
				pResult->ptr[i] = gse_f4FromLSBbytes(bbuf.ptr + i*4);
				if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
				if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
			}
			break;
			
		case 0x4:  /* 8bit int in 8bits */
			for(i=0; i<pResult->len; i++){
				pResult->ptr[i] = bbuf.ptr[i];
				if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
				if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
			}
			break;
			
		case 0x11: /* 8bit int in lower portion of 16 bits */
		case 0x7:  /* 12bit int in lower portion of 16 bits */
		case 0xC:  /* 16bit int in 16 bits */
		case 0x1C: /* Willy's waisting bits again, this is 0xC */
			
			if(flags & JW_HACK_12_IN_16_HI_SHIFT){
				for(i=0; i<pResult->len; i++){
					pResult->ptr[i]=(gse_u2FromLSBbytes(bbuf.ptr + i*2) >> 4)&0x0FFF;
					if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
					if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
				}
			}
			else{
				for(i=0; i<pResult->len; i++){
					pResult->ptr[i] = gse_u2FromLSBbytes(bbuf.ptr + i*2);
					if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
					if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
				}
			}
			break;
		
		case 0x8:  /* 12bit int in lower portion of 32 bits */
		case 0xD:  /* 16bit int in lower portion of 32 bits */
			for(i = 0; i<pResult->len; i++){
				pResult->ptr[i] = gse_u4FromLSBbytes(bbuf.ptr + i*4);
				if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
				if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
			}
			break;
			
		case 0x3:  /* 9 bit in 16 bit lo */
			for(i = 0; i<pResult->len; i++){
				u2 = gse_u2FromLSBbytes(bbuf.ptr + i*2);
				pResult->ptr[i] = jwaves_conv_flt9( u2 );
				if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
				if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
			}
			break;
		
		case 0x5:  /* 9 bit in 9 bits */
			jwaves_9in9_upk(pResult, bbuf);
			break;
			
		case 0x12: /* 9 bit in 32 bit lo */
			for(i = 0; i<pResult->len; i++){
				u4 = gse_u2FromLSBbytes(bbuf.ptr + i*4);
				pResult->ptr[i] = jwaves_conv_flt9( u4 );
				if(pResult->ptr[i] > pResult->max) pResult->max = pResult->ptr[i];
				if(pResult->ptr[i] < pResult->min) pResult->min = pResult->ptr[i];
			}
			break;
		
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "Handling data format type "
			             "0x%02X is not yet implimented", uFmt); 
		
	}
	
	except_cleanup_pop(except_no_call); /* pResult->ptr */
	return pResult->len;
}

/*****************************************************************************/
/* Recovering SCLK from SCLK bits in MP header */


/* Copied from /opt/project/juno/devel/src/util/UTIL_event_time.c
 * which is not in SVN */


/* Although LRS data time-tag applied after the WAVES record is formatted (glopping can 
 *	extend the delta between the mp time-tag and the Waves time-tag), mini-packet
 * blocking can cause the Waves timetag to be placed before data is collected).
 * -wtr
 */

/* The MAX_DELTA value determines how long we can expect the data formatting routines to
 * stall with data prior to delivery */
#define MAX_DELTA   64 

 /*Extract instrument event time
  * Data from the spacecraft is delivered with enough information to
  * recover an event time for the attached dataset.  This time indicates
  * the time at which data acquisition began.  Accuracy is limited by timing
  * resulution within the instrument (usually good to better than 2.5mSec)
  * and by the accuracy achieved when generating a SCLK/SCET relationship 
  * that is (hopefully) delivered with the data record.
  *
  * arguments:
  *	waves_time_in is the seconds field from the telemetry
  *		transport packet.
  *
  * returns:
  *	an 4-byte unsigned int containing the decoded instrument event time.
  */

uint4 jwaves_recover_sclk(uint2 uRTI, uint4 uRepSCLK)
{
	uint4 pkt_time;
	uint4 mp_time;
	int4 delta;

	pkt_time =  uRepSCLK & 0xFFFFFFFE; /* strip Time quality bit */
	
	mp_time = uRTI / 40;	/* 10 bit second in mP */
	delta = (mp_time  & 0x000003FF) -
              	    (pkt_time & 0x000003FF);	/* 32 bit SCLK */
            	    
   /* OK, nominally 'mp - SCLK' is negative.  If it's positive, then there is a roll
	 * involved and we need to adjust the delta we using by 1024 (10 bit roll value)
	 * UNLESS it's tiny..
	 *
	 * Currently (2008) the telemetry model doesn't let data hang around (*), so
	 * there's no need to worry about stale data...
	 */
     if (delta > MAX_DELTA)
              delta = delta - 0x0400;
     return pkt_time + delta;
} 


/*****************************************************************************/
/* Getting a SCLK form SCET */


/* Leap second: 2008-12-31T23:59:60 */
/* Leap second: 2005-12-31T23:59:60 */

char* jwaves_tv2sclk(struct timeval* pTv)
{
	struct timeval tv;
	
	if(pTv == NULL){
		gettimeofday(&tv, NULL);
		pTv = &tv;	
	}
	
	int nSclk = pTv->tv_sec - g_nJunoSclkOffset;
	const int4* pRepeat = g_unixRepeats;
	while( *pRepeat != 0.0 ){
		if( pTv->tv_sec > *pRepeat ) nSclk++;
		pRepeat++;
	}
	
	return gse_string("%010d.%03d", nSclk, pTv->tv_usec / 1000); 
}

char* jwaves_tv2sclk_s(struct timeval* pTv, char* buf, size_t len)
{
	char* s;
	
	if(buf == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL char pointer in "
		            "jwaves_tv2sclk_s");
	if(len < 15)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "result buffer len < 15 in "
		            "jwaves_tv2sclk_s");
	
	s = jwaves_tv2sclk(pTv);
	strncpy(buf, s, 14);
	buf[14] = '\0';
	return buf;
}

float8 jwaves_tv2sclk_d(struct timeval* pTv)
{
	struct timeval tv;
	double dRet;
	
	if(pTv == NULL){
		gettimeofday(&tv, NULL);
		pTv = &tv;	
	}
	
	int nSclk = pTv->tv_sec - g_nJunoSclkOffset;
	const int4* pRepeat = g_unixRepeats;
	while( *pRepeat != 0.0 ){
		if( pTv->tv_sec > *pRepeat ) nSclk++;
		pRepeat++;
	}
	
	dRet = nSclk + pTv->tv_usec / 1000000.0;
	return dRet;
}

void jwaves_sclk2tv(float8 dSclk, struct timeval* pTv){
	double dTrunc;
	
	if(dSclk < 0)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "SCLK value, %f is negative.",dSclk);
	
	if(pTv == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "struct timeval pointer is NULL");
	
	if(dSclk > 0x7FFFFFFF)
		gse_throw_fm(GSE_XCEPT, GSE_XOVERFLOW, "SCLK value, %f, is too big to "
		             " fit in a struct timeval.", dSclk);
		
	dTrunc = trunc(dSclk);
	pTv->tv_sec = (time_t)dTrunc + g_nJunoSclkOffset;
	
	const int4* pRepeat = g_unixRepeats;
	int i = 0;
	while( *pRepeat != 0.0 ){
		if((int4)dTrunc + g_nJunoSclkOffset > *pRepeat + i) (pTv->tv_sec)--;
		pRepeat++;
		i++;
	}
	
	pTv->tv_usec = (suseconds_t)( (dSclk - dTrunc) * 1000000);
}

float8 jwaves_str2sclk(const char* sTime)
{
	float8 dSclk;
	int4 nSclk, i;
	float8 dOffset = 0.0;
	struct timeval tv;
	bool bRelative = false;

	const int4* pRepeat = g_unixRepeats;
	
	if((sTime == NULL)||(sTime[0] == '\0'))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null or empty time value in "
		            "jwaves_str2sclk");
	
	/* Relative times */
	
	/* If this starts with '+', '-', or = to 'NOW' the the sclk that corresponds
	   to GMT, but with an offset */
	
	if((sTime[0] == '+') || (sTime[0] == '-')){
		bRelative = true;
		if(! gse_str2float8(sTime, &dOffset))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert time offset"
			             "'%s' in seconds to a Juno SCLK value.", sTime);
		else
			dOffset *= 60.0;
	}
	
	if(strcasecmp(sTime, "FOREVER") == 0){
		return (float8) 0x7FFFFFFF;
	}
	
	if(strcasecmp(sTime, "NOW") == 0){
		bRelative = true;
		dOffset = 0.0;
	}
	
	if(bRelative){
		gettimeofday(&tv, NULL);  /* <-- This is GMT */
	
		dSclk = tv.tv_sec + (tv.tv_usec / 1000000.0) - g_nJunoSclkOffset;
		dSclk += dOffset;
		 	
		while( *pRepeat != 0.0 ){
			if( tv.tv_sec > *pRepeat ) dSclk += 1.0;
			pRepeat++;
		}
		
		return dSclk;
	}
	
	/* SCLKs */
	
	/* See if this is a HEX integer */
	if((strlen(sTime) > 2)&&(sTime[0] == '0')&&(sTime[1] == 'x')){
		if(!gse_str2int4(sTime, &nSclk))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert Hex SCLK "
			             "str '%s' into an integer.", sTime);	
		else
			return (float8)nSclk;
	}
	
	/* Decimal sclks */
	bool bIsDigit = true;
	for(i=0;i<strlen(sTime);i++) 
		if( !isdigit(sTime[i]) && sTime[i] != '.') 
			bIsDigit = false;
	
	if(bIsDigit){
		if(!gse_str2float8(sTime, &dSclk))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert decimal SCLK "
			             "str '%s' into a double-precision value.", sTime);	
		else
			return dSclk;
	}
	
	/* Finnally SCETs */
	gse_str2tv_gm(sTime, &tv, true);
	if( (dSclk = jwaves_tv2sclk_d(&tv)) < 0.0){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "UTC time '%s' converts to a "
		             "negative Juno SCLK.", sTime);
	}
	
	return dSclk;
}

/*****************************************************************************/
float8 jwaves_pkt2sclk(const GsePkt* pPkt)
{
	byte uCont = 0;
	uint4 uSclk = 0;
	uint2 uRTI  = 0;
	JWavesSDSblk* pSci = NULL;
	JWavesMroBlk* pMro = NULL;
	cbytebuf_t cbuf;
	float8 dSclk;
	
	if(pPkt == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "pPkt = NULL in jwaves_pkt2sclk");
	
	if(pPkt->pHdr->nSpaceCraft != -61){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Packet is for some spacecraft "
		             "other than Juno, Spice ID = %d, expected -61",
				       pPkt->pHdr->nSpaceCraft);
	}
	
	/* If this is a raw packet, then get bytes 4-7 of the data section */
	uCont = (pPkt->pHdr->uContent & 0xF);
	if(uCont == 0x1){
		cbuf = pPkt->getData(pPkt);
		
		uSclk = gse_u4FromMSBbytes(cbuf.ptr + 4);
		dSclk = (float8)uSclk; 
		
		/* Can only do RTI recovery on science and MRO data */
		if( ((pPkt->pHdr->uContent & 0xF0) != 0x80) &&
		    ((pPkt->pHdr->uContent & 0xF0) != 0xE0) ) 
			return dSclk;
		
		uRTI = gse_u2FromMSBbytes(cbuf.ptr + 10);
		
		uSclk = jwaves_recover_sclk(uRTI, uSclk);
		
		dSclk = (float8)uSclk  + (float8) ( (uRTI % 40) / 40.0);
		
		return dSclk;
	}
	
	/* Okay, this is processed data.   Must be either an MRO or a SCI packet */
	if(pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0)){
		
		pSci = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		dSclk = ((float8)pSci->uColSCLK) + (((float8)pSci->uColRTI)/40.0);
		
		return dSclk;
	}
	
	if(pPkt->hasUniqueBlkByType(pPkt, JW_MRO_BLK_TYPE_ID, 0)){
		
		pMro = (JWavesMroBlk*)pPkt->getBlkByType(pPkt, JW_MRO_BLK_TYPE_ID, 0);
		
		dSclk = ((float8)pMro->uColSCLK) + (((float8)pMro->uColRTI)/40.0);
		
		return dSclk;
	}
	
	gse_throw_m(GSE_XCEPT, GSE_XASSERT, "Packet is not raw data, an MRO, or a SCI"
			    " packet, what the heck is it?");
	
	return -1.0;
}
		

/*****************************************************************************/

uint2 jwaves_pkt2seq(const GsePkt* pPkt)
{
	byte uCont = 0;
	cbytebuf_t cbuf = CBBUF_NULL;
	uint2 uSeq = 0;
	JWavesSDSblk* pSci = NULL;
	JWavesMroBlk* pMro = NULL;
	
	/* If this is a raw packet, then get bytes 2-3 of the data section */
	uCont = (pPkt->pHdr->uContent & 0xF);
	if(uCont == 0x1){
		cbuf = pPkt->getData(pPkt);
		
		uSeq = gse_u2FromMSBbytes(cbuf.ptr + 2);
		uSeq &= 0x3FFF;
				
		return uSeq;
	}
	
	/* Okay, this is processed data.   Must be either an MRO or a SCI packet */
	if(pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0)){
		
		pSci = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		return pSci->uSeqNoLo;
	}
	
	if(pPkt->hasUniqueBlkByType(pPkt, JW_MRO_BLK_TYPE_ID, 0)){
		
		pMro = (JWavesMroBlk*)pPkt->getBlkByType(pPkt, JW_MRO_BLK_TYPE_ID, 0);
		return pMro->uSeqNo;
	}
	
	gse_throw_m(GSE_XCEPT, GSE_XASSERT, "Packet is not raw data, an MRO, or a SCI"
			    " packet, what the heck is it?");
	
	return 0xFFFF;
}

/*****************************************************************************/
/* Is this even a possible juno cal or sclk value? */

const gse_caltime_t jwaves_ct_beg = {2000, 01, 01, 01, 12, 0, 0.0};
const gse_caltime_t jwaves_ct_end = {2068, 01, 19, 19, 15, 14, 7.0};

bool jwaves_legal_ct(const gse_caltime_t* pCal)
{
	if(!gse_ct_isnorm(pCal)) return false;
	
	if(gse_ctcmp(pCal, &jwaves_ct_beg) < 0) return false;
	if(gse_ctcmp(pCal, &jwaves_ct_end) >= 0) return false;
	
	return true;
}





