/** @file JWavesConv.h */
#ifndef _JWavesConv_H_
#define _JWavesConv_H_

#include <sys/time.h>
#include <rpwgse/GseTypes.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseTime.h>
#include <rpwgse/GsePkt.h>

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_junogse */
/*@{*/
/** @defgroup mod_jwaves_conv Data Conversions 
 * Helpers for converting Juno Waves data into floats and for getting SCET
 * from SCLK.
 */
/*@}*/

/** @addtogroup mod_jwaves_conv */
/*@{*/

/** Check to see if a particular combination of packet id and packet source
 *  are understood by this library.
 *
 * @param uPSID A combination of the mini packet ID (high-bits) and Src
 *        (low-bits) fields.
 *
 * @param bThrow One of JWAVES_THROW or JWAVES_LOG.
 *
 * @returns @c true if the packet ID and Source are understood by this
 *          library, @c false otherwise.  If @a bThrow is JWAVES_THROW and
 *          the packet and source combination are not understood by this
 *          library then an exception is throw and there is no return value.
 *
 * @throws (JW_XCEPT, JW_XPKTUNKNOWN) if the packet id and source combination
 *         appears invalid @b and bThrow = JWAVES_THROW.  If bThrow = JWAVES_LOG
 *         errors are logged instead of thrown as exceptions.
 */
bool jwaves_valid_idsrc(byte uPSID, bool bThrow);


/** Check to see if a particular combination of packet id and packet source
 *  and data section are understood by this library.
 *
 * @param uPSID A combination of the mini packet ID (high-bits) and Src
 *        (low-bits) fields.
 *
 * @param nSec The data section index, for most packets the only valid
 *        value is 0, for Noise reduction packets this can be 0 to 2.
 *
 * @param bThrow One of JWAVES_THROW or JWAVES_LOG.
 *
 * @returns @c true if the packet ID and Source are understood by this
 *          library, @c false otherwise.  If @a bThrow is JWAVES_THROW and
 *          the packet and source combination are not understood by this
 *          library then an exception is throw and there is no return value.
 *
 * @throws (JW_XCEPT, JW_XPKTUNKNOWN) if the packet id, source and section
 *         combination appears invalid @b and bThrow = JWAVES_THROW.  
 *         If bThrow = JWAVES_LOG errors are logged instead of thrown as
 *         exceptions.
 */

bool jwaves_valid_section(byte uPSID, int nSec, bool bThrow);



/** Contians the results of calling jwaves_bytes2float4() */
struct jwaves_float4_conv{

	/** A pointer to a float4 array allocated on the heap, or NULL */
	float* ptr;
	
	/** The length of the array pointed to by @a ptr */
	size_t len;
	
	/** The maximum value encountered during the conversion */
	float min;
	
	/** The minimum value encontered during the conversion */
	float max;
};

void init_jwaves_float4_conv(struct jwaves_float4_conv* pConv);

#define JW_CONV_INT_SIGNED           0x00000001
#define JW_HACK_12_IN_16_HI_SHIFT    0x00010000


/** Convert a chunk of bytes to an array of 4-byte floats
 *
 * @param uFmt the format of the data bytes see @ref JWavesSDSblk::uFmt
 *
 * @param bbuf a byte buffer of data to convert.
 *
 * @param override a set of override flags for the conversion process
 * these are:
 * - JW_CONV_INT_SIGNED: Assume integers are signed, overrides uFmt
 *
 * @param pResult a jwaves_float4_conv structure that will contian the 
 *        results of the conversion.
 * @return the number of floats generated.
 *
 * @throws TBD
 */
size_t jwaves_bytes2float4(byte uFmt, cbytebuf_t bbuf, uint4 override,
                           struct jwaves_float4_conv* pResult);


/** Convert a 9 bit float to an IEEE float
 *
 * @param f9_lo a 2 byte unsigned integer were the lower 9 bits contain the
 *        9-bit floating point value from the Waves instrument.
 */
float4 jwaves_conv_flt9(uint4 f9_lo);


/** Does this format represent compressed data */
bool jwaves_isCompressed(byte uFmt); 

/** Find corresponding uncompressed format.
 *
 * @param uCompFmt A value of JWavesSDSblk::uFmt that represents compressed
 *        data.
 * @return The uncompressed format corresponding to @a uCompFmt.
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) if @a uCompFmt does not represent a 
 *         compressed format.
 */
byte jwaves_decompFmt(byte uCompFmt);

/** How many bits are used for each sample in the data section */
int jwaves_dataBlockBits(byte uFmt);

/** Get one section of an LFR Nrd Spectrum that is 9-bit compressed */
cbytebuf_t jwaves_loc_nrd_spec_data(const GsePkt* pPkt, int iPart);


/** Find out if the data is from the logamp or from the fft engine
 *  or just plain waveforms.  The kind of data is independent of
 *  the actual data format, though data of a particular kind is often
 *  encountered with just a small range of formats.
 *
 * @param uPSID The packet and source combinded id.
 *
 * @return One of:
 *    - JWAVES_KIND_LOGAMP if the data is from the logamp
 *    - JWAVES_KIND_WFRM   if the data is just a waveform
 *    - JWAVES_KIND_FFT    if the data is fft of a waveform(s)
 *    - JWAVES_KIND_NOTSCI if the data is not a science data product
 */
int jwaves_dataKind(byte uPSID);

/** Is this science data from the LFR reciever ? */
bool jwaves_islfrsci(byte uPSID);

/** Is this science data from the HFR receiver ? */
bool jwaves_ishfrsci(byte uPSID);

/** Is this science data from the magnetic field sensor? */
bool jwaves_isBsci(byte uPSID);

/* Unlike most defines, these are not random, if they are changed,
   g_data_kinds, and jwaves_dataKind() must be updated */
#define JWAVES_KIND_NOTSCI  0
#define JWAVES_KIND_LOGAMP  1
#define JWAVES_KIND_WFRM    2
#define JWAVES_KIND_FFT     3

/** Get a human readable short name for packet and source ID's
 *
 *@param uPSID the conbination of the packet source and ID bits,
 *       @see JWavesSDSblk for the uPSID
 *
 *@param section The noise reduction diagnostic packets have three
 *      data sections, use the value 0, 1, or 2 to get the
 *      channel name for the desired section.
 */
const char* jwaves_psid_sname(byte uPSID, int section);

/** Get a human readable long name for packet and source ID's.
 *
 *@param uPSID the conbination of the packet source and ID bits,
 *       @see JWavesSDSblk for the uPSID
 *
 *@param section The noise reduction diagnostic packets have three
 *      data sections, use the value 0, 1, or 2 to get the
 *         channel name for the desired section.
 *
 *@param nSCLK Long channel names were changed during development,
 *       after 332859600.0 (2010-07-20T01:00 UTC) the board model
 *       numbers are used in the titles for HFR and HFWBR data to 
 *       reduce confusion
 *       
 */
const char* jwaves_psid_lname(byte uPSID, int section, int nSCLK);

/** The all important SCLK voodo.
 *
 * This is Willy's algorithm, I'm not even about to touch it. --cwp
 *
 * @param uRTI the 2 byte RTI field from the MP header.
 *
 * @param uRepSCLK the 2 byte spacecraft SCLK from the Waves Header
 *
 * @returns the SCLK at the time the data was collected. To get the collection
 *          RTI take uRTI % 40.
 *          
 */
uint4 jwaves_recover_sclk(uint2 uRTI, uint4 uRepSCLK);

/** Convert a UXIX time value to a Juno SCLK string.
 * The Juno SCLK a 32bit counter set so that 2000-01-01T00:00 is zero seconds.
 *
 * This conversion accounts for the leap seconds that have occured
 * since the start of the Juno Epoch (2000-01-01T12:00).
 *
 * @param pTv The time value to convert.  If NULL, the current UTC time 
 *        is converted to a Juno SCLK value.
 *
 * @return A string allocated on the heap in the format: @n@n
 *         <b> SSSSSSSSS.sss </b> @n@n
 *         where @b SSSSSSSSS is the SCLK seconds and @b sss is the SCLK
 *         micro seconds.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) If pTvUTC contains a time before the
 *         start of the Juno Epoch.
 */
char* jwaves_tv2sclk(struct timeval* pTv);


/** Convert a UNIX time value to a Juno SCLK string, static memory version.
 *
 * This conversion accounts for the leap seconds that have occured
 * since the start of the Juno Epoch (2000-01-01T12:00).
 *
 * @see jwaves_tv2sclk for a dynamic memory version of the same function.
 *
 * @param pTv the time value to convert, if NULL the current UTC time is
 *        converted.
 *
 * @param buf a memory location at least 14 bytes long, to hold the null
 *        terminated string.
 *
 * @param len The length of the memory location to hold the null terminated
 *        string.
 *
 * @returns A pointer to @a buf, which contains a string in the format: @n@n
 *         <b> SSSSSSSSS.sss </b> @n@n
 *         where @b SSSSSSSSS is the SCLK seconds and @b sss is the SCLK
 *         micro seconds.
 */ 
char* jwaves_tv2sclk_s(struct timeval* pTv, char* buf, size_t len);


/** Convert a UNIX time value to a Juno SCLK floating point seconds value.
 *
 * This conversion accounts for the leap seconds that have occured
 * since the start of the Juno Epoch (2000-01-01T12:00).
 *
 * @param pTv the time value to convert.
 *
 * @return The UNIX time value converted to a single double precision 
 *         floating point value and then down shifted to the Juno epoch.
 */
float8 jwaves_tv2sclk_d(struct timeval* pTv);


/** Convert a waves sclk to a struct timeval in the UNIX epoch.
 * 
 * This conversion accounts for the leap seconds that have occured
 * since the start of the Juno Epoch (2000-01-01T12:00).
 *
 * The floating point spacecraft clock is split into seconds and microseconds
 * fields.  The seconds field is then shifted to the UNIX epoch.
 *
 * @param dSclk A juno sclk value than may have finer than 1 second
 *        resolution.
 *
 * @param pTv The floating point spacecraft clock is converted cWhere to store the result.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if @a pTv is NULL or if @a dSclk is negative.
 *         (GSE_XCEPT, GSE_XOVERFLOW) if @a bSclk is too big to fit in @a pTv.
 */
void jwaves_sclk2tv(float8 dSclk, struct timeval* pTv);


/** Convert a string to an SCLK.
 *
 * This conversion accounts for the leap seconds that have occured
 * since the start of the Juno Epoch (2000-01-01T12:00).
 *
 * @param sTime in a veriety of formats.
 *        - If this looks like a datetime it is converted to a SCLK.
 *        - If this looks like a decimal or hexidecimal number it is taken as
 *          a SCLK.
 *        - If the string 'NOW' (not case sensitive) is encontered the current
 *          UTC time is determined and converted into the equivalent Juno SCLK.
 *        - If the string 'FOREVER' (not case sensitive) is encountered the
 *          maximum possible juno SCLK volue is returned (0x7FFFFFFF).
 *        - If the value starts with '+' or '-' it is taken as a time offset
 *          from NOW in minutes and converted to a SCLK.
 *
 * @return A double precision sclk value.
 * @throws (GSE_XCEPT,GSE_XBADARG) if the time string is not understood
 */
float8 jwaves_str2sclk(const char* sTime);

/** Get the SCLK from any juno GSE packet.
 *
 * @param pPkt a pointer to a Juno GSE packet structure
 *
 * @return a floating point sclk value, does not take into account fine
 *         times from the housekeeping data.
 */
float8 jwaves_pkt2sclk(const GsePkt* pPkt);


/** Get the sequence number from any juno GSE packet.
 *
 * @param pPkt a pointer to a Juno GSE packet structure
 *
 * @return the 14 bit sequence number from the Waves header in a unsigned 
 *         16bit integer.
 */
uint2 jwaves_pkt2seq(const GsePkt* pPkt);

/** Inquire wether a calendar time could be a valid Juno SCLK time.
 *
 *
 * @param pCal The calender time to inspect, if this time is not normalized
 *        then the function returns false automatically.
 *
 * @returns true if the calender time is normalize and represents a
 *       date-time between 2000-01-01T12:00:00.000 and
 *       2068-01-19T15:14:07.000.
 */
bool jwaves_legal_ct(const gse_caltime_t* pCal);

/* * Get a list of the output data channel names.
 *
 * @return An array of channel name strings allocated on the heap.
 */
GseCStrAry* jwaves_channels();


#ifndef _JWavesConv_C_
extern const byte g_hrs_data[];
#endif


/*@}*/
#ifdef __cplusplus
}
#endif

#endif /* _JWavesConv_H_ */
