/** @file JWavesData.h */
#ifndef _JWavesData_H_
#define _JWavesData_H_

#include <rpwgse/GsePkt.h>
#include <rpwgse/GseData.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_jwaves_proc */
/*@{*/

#define JW_DATA_HFWBR_SIGNED     0x00000001
#define JW_DATA_LFR_WFRM_SIGNED  0x00000002
#define JW_DATA_HFRLOCK_OSC_FREQ 0x00000004
#define JW_DATA_DIV_BANDWIDTH    0x00000008
#define JW_DATA_MARK_DCMP        0x00000010
#define JW_DATA_RAW_FLOATS       0x00000020

#ifndef _JWavesData_C_

extern const size_t jw_LFR_lo_spec_len;
extern const size_t jw_LFR_hi_spec_len;
extern const size_t jw_HFR_base_spec_len;
extern const size_t jw_HFR_spec_len;

#endif

/** Generate XY data structures from a Juno Waves Packet
 *
 * @param pPkt a GSE packet with content type JWAVES_CNT_CAPTURE. 
 *        See @ref page_spec_4_2_1 for more information on the 
 *        GSE header content byte as it applies to Juno Waves.
 *
 * @param flags An or'ed set of 0 or more flags:
 *    - JW_DATA_HFWBR_SIGNED - Assume HFR and HFWBR waveform data is signed.
 *
 *    - JW_DATA_HFRLOCK_OSC_FREQ - Max HFR Lock datasets use the 
 *        ocillator frequency instead of the Center frequency for the X
 *        axis.
 *
 *    - JW_DATA_DIV_BANDWIDTH - Divide onboard spectra values by the
 *        number of FFT bins each value is made up of.  This gives a
 *        value proportional to spectral density.
 *
 *    - JW_DATA_MARK_DCMP - Add the text ' dcmp' to the channel name
 *        for decompressed packets.  Allows for testing of decompressed
 *        data.
 *
 *    - JW_DATA_RAW_FLOATS -- Do nothing to the data values except convert
 *        them to floats.  No scaling, no logs, no / square roots.
 *
 * @param iPktIdx The index of this packet in the source location, used
 *        for error reporting that does not throw.
 *
 * @param nOffset The offset in the data source to the start of the packet,
 *        used for error reporting that does not throw.
 *
 * @throws (JW_XCEPT, JW_XNUMPTS) If an improper number of data samples is 
 *         detected
 * @throws (JW_XCEPT, JW_XPKTNOTSCI) The @a pPkt is not a science data packet
 * @throws (JW_XCEPT, JW_XPKTNOTREADY) The packet has not been processed
 *         i.e is still segmented, comperssed etc.
 * @throws (JW_XCEPT, JW_XPKTUNKNOWN) In invalid 
 *
 * @returns one or more GseXYdata objects for display.  
 *   see jwaves_channels(), to get a runtime discription of the possible
 *   channel names for the output data.
 */
GseXYdataAry* fact_JWavesXYdataAry(const GsePkt* pPkt, uint4 flags,
                                   int4 nPktIdx, size_t nOffset);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesData_H_ */
