#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseLog.h>
#include <rpwgse/GseRiceDecomp.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseConv.h>

#include "JWavesDefs.h"
#include "JWavesExcept.h"
#include "JWavesConv.h"
#include "JWavesSDSblk.h"
#include "JWavesDcmpFilter.h"
#include "JWavesDcmpBlk.h"

/*****************************************************************************/
/* A helper for readNext, unpacks 9-bit data */

GsePkt* JWavesDcmpFilter_9bitTo32bit(JWavesDcmpFilter* pThis, GsePkt* pPkt)
{
	cbytebuf_t old;
	bytebuf_t  new;
	JWavesSDSblk* pSci = NULL;
	JWavesDcmpBlk* pDcmpBlk = NULL;
	struct jwaves_float4_conv conv;
	size_t nSamples = 0;
	size_t nLastSamples = 0;
	int iSec = 0;
	
	gse_debug("Decompressing 9-bit data to little endian IEEE floats");
	
	if( ! pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0) )
		gse_throw_m(GSE_XCEPT, GSE_XINCONCEIVABLE, "Are you calling "
		            "JWavesDcmpFilter_9bitTo32bit directly instead of "
		            "using JWavesDcmpFilter::readNext?");
	
	pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	init_jwaves_float4_conv(&conv);
	
	/* For data that does not come in a multi-section minipacket we can
	   do this is one shot, otherwise individual subsections must be
		converted */
		 
	switch(pSci->uPSID){
	case 0xF0:
	case 0xF4:
	case 0xF5:
		
		new.len = 0; new.ptr = NULL;
		
		nLastSamples = -1;
		for(iSec = 0; iSec < 3; iSec++){
			old = jwaves_loc_nrd_spec_data(pPkt, iSec);
			nSamples = jwaves_bytes2float4(pSci->uFmt, old, 0, &conv);
			except_cleanup_push(free, conv.ptr);
			
			if(nLastSamples == -1) nLastSamples = nSamples;
			
			if(nSamples != nLastSamples)
				gse_throw_m(GSE_XCEPT, GSE_XDECODE, "Unequal length minipacket "
				        "sections, probably an error in jwaves_loc_nrd_spec_data");
			
			if(new.len == 0){
				new.len = nSamples * 3 * 4;
				new.ptr = (byte*)gse_calloc(new.len, sizeof(byte));
			}
			
			if(!g_bIsLSBfirst){
				gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Byte swapping needed for "
			   	         "big endian machines in JWavesDcmpFilter_9bitTo32bit");
			}
			memcpy(new.ptr + iSec * nSamples * 4, conv.ptr, nSamples * 4);
			except_cleanup_pop(except_call);  /* conv.ptr */
			
			nLastSamples = nSamples;
		}
		
		pPkt->setData(pPkt, new);				
		break;
		
	case 0xA0:
	case 0xA4:
	case 0xA5:
		gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Handling 9-bit encoded NRD"
		            " waveforms is not yet implemented");
		break;
		
	default:
		old = pPkt->getData(pPkt);
		nSamples = jwaves_bytes2float4(pSci->uFmt, old, 0, &conv);
		except_cleanup_push(free, conv.ptr);
	
		/* Swap out the data in the packet */
		new.len = conv.len * 4;
		new.ptr = (byte*)conv.ptr;
		if(!g_bIsLSBfirst){
			gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Byte swapping needed for big "
			            "endian machines in JWavesDcmpFilter_9bitTo32bit");
		}
		pPkt->setData(pPkt, new);
	
		except_cleanup_pop(except_no_call);  /* conv.ptr */
		break;
	}
	
	/* Update the science data status block */
	pSci->uFmt = 0x2; /* 32bit in 32bit */
	pSci->setMPlen(pSci, new.len);
	
	/* Make a decompression processing record */
	pDcmpBlk = new_JWavesDcmpBlk(CBBUF_NULL, 0);
	((GseProcBlk*)pDcmpBlk)->setNow((GseProcBlk*)pDcmpBlk);
	((GseProcBlk*)pDcmpBlk)->uProgID = pThis->base.uProgID;
	pDcmpBlk->fCmpRatio = ((float4)new.len) / ((float4)old.len);
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pDcmpBlk);
	
	/* Set the header type */
	if(!pSci->isGlopped(pSci)) /* Only glopping can remain */
		pPkt->pHdr->uContent = 0x84;
	
	return pPkt;	
}


/*****************************************************************************/
GsePkt* JWavesDcmpFilter_readNext(GsePktSrc* pBase)
{
	GsePkt*       pPkt    = NULL;
	JWavesSDSblk* pSci    = NULL;
	JWavesSDSblk* pNewSci = NULL;
	JWavesDcmpBlk* pDcmpBlk = NULL;
	cbytebuf_t cbuf       = CBBUF_NULL;
	int nBitWidth;
	struct rice_decomp_t decomp;
	GsePkt* pNewPkt;
	bytebuf_t newData;
	size_t i;
	JWavesDcmpFilter* pThis = (JWavesDcmpFilter*)pBase;
	
	
	/* Decompressing is a one packet in to one packet out processes */
	
	if( (pPkt = pThis->base.pUpSrc->readNext(pThis->base.pUpSrc)) == NULL)
		return NULL;	
	
	/* Don't operate on raw data, or data that is not science */
	if( ((pPkt->pHdr->uContent >> 4) & 0xF) != 0x8)
		return pPkt;
	
	/* Don't operate on raw data */
	if( (pPkt->pHdr->uContent & 0xF) < 0x3)
		return pPkt;
		
	/* If no SDS block can't do anything */
	if( ! pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0) ){
		gse_error_v("Missing Science Data Status Block in packet at "
				      "offset 0x%X, source: %s", 
				      pThis->base.pUpSrc->offset(pThis->base.pUpSrc),
				      pThis->base.pUpSrc->name(pThis->base.pUpSrc)  );
		return pPkt;
	}
	
	pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	/* If segmented can't operate on it */
	if( pSci->isSegmented(pSci)){
		gse_warning_v("Decompression routine encountered a segmented packet at "
				        "offset 0x%X, source %s", 
		               pThis->base.pUpSrc->offset(pThis->base.pUpSrc),
					      pThis->base.pUpSrc->name(pThis->base.pUpSrc)
		            );
		return pPkt;
	}
	
	/* Check for data that is not compressed, but is in an 9-bit float format */
	if( ! pSci->isCompressed(pSci) && (pSci->uFmt == 0x5 || pSci->uFmt == 0x3)){
		return JWavesDcmpFilter_9bitTo32bit(pThis, pPkt);
	}
	
	
	/* If not compressed, then nothing to do */
	if( ! pSci->isCompressed(pSci))
		return pPkt;
	
	/* Okay, data requires decompression, determine the bit mode */
	cbuf = pPkt->getData(pPkt);
	nBitWidth = jwaves_dataBlockBits(pSci->uFmt);
	if(! gse_rice_decomp(cbuf.ptr, cbuf.len, nBitWidth, &decomp) )
	{
		if(decomp.pMsg != NULL)
			gse_error_v("Decompression failed with error code %d, msg was: %s",
			            decomp.errorNum, decomp.pMsg);
		else
			gse_error_v("Decompression failed with error code %d, (no message)",
			            decomp.errorNum);
		
		return pPkt;
	}
	
	except_cleanup_push(free, decomp.pDecomp);
	
	/* Make a new packet as a copy of the old one */
	pNewPkt = pPkt->copy(pPkt, 0);
	except_cleanup_push(pNewPkt->del, pNewPkt);
	
	/* Set the format */
	pNewSci = (JWavesSDSblk*)pNewPkt->getBlkByType(pNewPkt,JW_SDS_BLK_TYPE_ID,0);
	pNewSci->uFmt = jwaves_decompFmt(pSci->uFmt);
	pNewSci->setMPlen(pNewSci, decomp.uSamples * 2);

	/* Save the data */
	newData.len = decomp.uSamples * 2;
	newData.ptr = (byte*) gse_calloc(newData.len, sizeof(byte));
	except_cleanup_push(free, newData.ptr);
	
	/* Let's high shift 12 bit data, it's a Willyism, don't ask me, of course
	   this would break glopped data !!! */
	if( pNewSci->uFmt == JW_12BIT_IN_16BIT_lo ){
		pNewSci->uFmt = JW_16BIT_IN_16BIT;
		
		if(pNewSci->isGlopped(pNewSci))
			gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Decompressing Glopped data is"
					      "not yet supported !!!");
		
		for(i = 0; i<decomp.uSamples; i++)
			gse_u2ToLSBbytes(newData.ptr + i*2, (decomp.pDecomp[i] << 4)&0xFFF0 );
	}
	else{
		for(i = 0; i<decomp.uSamples; i++)
			gse_u2ToLSBbytes(newData.ptr + i*2, decomp.pDecomp[i]);
	}
		
	pNewPkt->setData(pNewPkt, newData);
	except_cleanup_pop(except_no_call);  /* newData */
	
	if(!pNewSci->isGlopped(pNewSci)) /* Only glopping can remain */
		pNewPkt->pHdr->uContent = 0x84;
		
	pDcmpBlk = new_JWavesDcmpBlk(CBBUF_NULL, 0);
	((GseProcBlk*)pDcmpBlk)->setNow((GseProcBlk*)pDcmpBlk);
	((GseProcBlk*)pDcmpBlk)->uProgID = pThis->base.uProgID;
	pDcmpBlk->fCmpRatio = ((float4)newData.len) / ((float4)cbuf.len);
	
	pNewPkt->addHdrBlk(pNewPkt, (GseBlock*)pDcmpBlk);
	
	except_cleanup_pop(except_no_call);  /* pNewPkt */
	except_cleanup_pop(except_call);     /* always delete decomp.pDecomp */
	
	return pNewPkt;
}

/*****************************************************************************/

/* No destructor needed, just use free */

/*****************************************************************************/

JWavesDcmpFilter* new_JWavesDcmpFilter(GsePktSrc* pSrc, byte uProgID)
{
	JWavesDcmpFilter* pThis;
	
	pThis = (JWavesDcmpFilter*)new_GsePktFilter(
			    pSrc, 
	          uProgID,
	          sizeof(JWavesDcmpFilter),
	          JWavesDcmpFilter_readNext,
	          free
	        );
	
	
	pThis->del = free;
	
	return pThis;	
}
