#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseConv.h>

#include "JWavesDefs.h"
#include "JWavesHRSblk.h"


bytebuf_t JWavesHRSblk_encode(GseBlock* pBlk, byte uExtraLen, byte* pOffset)
{
	size_t i;
	byte offset;
	bytebuf_t buf;
	
	JWavesHRSblk* pThis = (JWavesHRSblk*)pBlk;
	GseProcBlk* pProc   = (GseProcBlk*)pBlk;
	
	/* Check prior to encode */
	if(pProc->uStepID != JW_SCEHRS_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Processing step ID was %hhu"
				       "expected %hhu.", pProc->uStepID, JW_SCEHRS_PROC_STEP_ID);
	
	gse_overflowUinU(uExtraLen + 15, sizeof(byte));
	
	buf = JWavesRdrBlk_encode(pBlk, uExtraLen + 15, &offset);
	gse_assert(buf.len >= offset + 15);
	
	buf.ptr[offset]     = (byte) ((pThis->uUID >> 16) & 0xFF);
	buf.ptr[offset + 1] = (byte) ((pThis->uUID >> 8) & 0xFF);
	buf.ptr[offset + 2] = (byte) (pThis->uUID & 0xFF);
	
	gse_u4ToMSBbytes(buf.ptr + offset + 3, pThis->uRemoteIP);
	gse_u4ToMSBbytes(buf.ptr + offset + 7, pThis->uLocalIP);
	
	for(i = 0; i<4; i++)
		buf.ptr[offset + 11 + i] = pThis->sync[i];
	
	
	if(pOffset != NULL) *pOffset = offset + 15;
	return buf;
}

byte JWavesHRSblk_decode(GseBlock* pBlk, cbytebuf_t buf)
{
	GseProcBlk*   pProc = (GseProcBlk*)pBlk;
	JWavesHRSblk* pThis = (JWavesHRSblk*)pBlk;
	
	size_t i;
	byte uOffset;
	
	uOffset = JWavesRdrBlk_decode(pBlk, buf);
	
	if(pProc->uStepID != JW_SCEHRS_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x%hhX for the block step "
		             "ID, but recieved 0x%hhX", JW_SCEHRS_PROC_STEP_ID,
				       pProc->uStepID);
	
	/* The uid is tricky because it is a 3 byte field */
	pThis->uUID = gse_u4FromMSBbytes(buf.ptr + uOffset - 1);
	pThis->uUID &= 0x00FFFFFF;  
	
	pThis->uRemoteIP = gse_u4FromMSBbytes(buf.ptr + uOffset + 3);
	pThis->uLocalIP  = gse_u4FromMSBbytes(buf.ptr + uOffset + 7);
	
	for(i = 0; i<4; i++)
		pThis->sync[i] = buf.ptr[uOffset + 11 + i];
		
	return uOffset + 15;
}

bool JWavesHRSblk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{
	JWavesHRSblk* pThis = (JWavesHRSblk*)pBase;
	JWavesHRSblk* pHRSoth = (JWavesHRSblk*)pOther;
	
	if(! JWavesRdrBlk_canMerge(pBase, pOther, uFlags) ) return false;
	
	if( (pThis->uUID      != pHRSoth->uUID) || 
	    (pThis->uRemoteIP != pHRSoth->uRemoteIP) ||
	    (pThis->uLocalIP  != pHRSoth->uLocalIP)  )  return false;
	
	return memcmp(pThis->sync, pHRSoth->sync, 4) == 0 ? true: false;
}

/* Copy constructor */
GseBlock* JWavesHRSblk_copy(GseBlock* pBase, size_t uOver)
{
	size_t i;
	JWavesHRSblk* pNew = NULL;
	JWavesHRSblk* pThis = (JWavesHRSblk*)pBase;
	
	if(uOver < sizeof(JWavesHRSblk))
		pNew = (JWavesHRSblk*) JWavesRdrBlk_copy(pBase, sizeof(JWavesHRSblk));
	else
		pNew = (JWavesHRSblk*) JWavesRdrBlk_copy(pBase, uOver); 
	
	pNew->uUID      = pThis->uUID;
	pNew->uRemoteIP = pThis->uRemoteIP;
	pNew->uLocalIP  = pThis->uLocalIP;
	pNew->del       = pThis->del;
	
	for(i = 0; i < 4; i++) pNew->sync[i] = pThis->sync[i];
	
	return (GseBlock*)pNew;
}

/* GseBlock destructor is fine for deleting this */

JWavesHRSblk* new_JWavesHRSblk(cbytebuf_t buf, size_t uOver)
{
	GseBlock*     pBase = NULL;
	GseProcBlk*   pProc = NULL;
	JWavesRdrBlk* pRdr  = NULL;
	JWavesHRSblk* pThis = NULL;
	
	if(uOver < sizeof(JWavesHRSblk))
		pRdr = new_empty_JWavesRdrBlk(sizeof(JWavesHRSblk));
	else
		pRdr = new_empty_JWavesRdrBlk(uOver); 
	
	pBase = (GseBlock*)pRdr;
	pBase->encode   = JWavesHRSblk_encode;
	pBase->decode   = JWavesHRSblk_decode;
	pBase->copy     = JWavesHRSblk_copy;
	pBase->canMerge = JWavesHRSblk_canMerge;
	
	/* Since merge can only happen if the fields are identical, no reason to 
	   have a merge override, it would only call the base clase anyway       */
	
	pBase->_sName  = "Waves SCE HRS Reader";
	pBase->_uMinLen += 15;
	
	pProc = (GseProcBlk*)pRdr;
	pProc->uStepID = JW_SCEHRS_PROC_STEP_ID;
	
	pThis = (JWavesHRSblk*)pRdr;
	pThis->del     = pRdr->del;
	
	
	/* Calloc in base makes the rest of this 0, as we want */
	if(buf.ptr == NULL) return pThis;
	
	except_cleanup_push(pThis->del, pThis);
	
	JWavesHRSblk_decode(pBase, buf);

	except_cleanup_pop(except_no_call);
	
	return pThis;
}
