#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseConv.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseCrc.h>

#include "JWavesIDP.h"
#include "JWavesSciMP.h"
#include "JWavesMroMP.h"

size_t JWavesSciIDP_origLen(JWavesSciIDP* pThis)
{ 
	return pThis->_len; 
}

uint2 JWavesSciIDP_calcCRC(struct jwaves_sci_idp* pThis)
{ 
	return pThis->_calculatedCRC;
}

bool JWavesSciIDP_CRCmatch(JWavesSciIDP* pThis)
{ 
	return (pThis->uCRC == pThis->_calculatedCRC); 
}

uint2 JWavesSciIDP_origSeq(JWavesSciIDP* pThis)
{
	return  ( pThis->uSeq | ((pThis->ts << 14)&0xC000)  );
}

/*****************************************************************************/
/* MP factory */
JWavesMP* make_minipacket(cbytebuf_t cbuf, size_t* pOffset){
	byte uID;
	/* Read ahead for the 1st nibble, if 0xE make an MRO packet, else 
	   make a Sci packet */
	
	if( *pOffset >= cbuf.len) return NULL;
	
	if( cbuf.ptr[*pOffset] == 0x00) return NULL;
	
	uID = (cbuf.ptr[*pOffset] >> 4) & 0x0F;
	
	if(uID == 0x0E) 
		return (JWavesMP*) new_JWavesMroMP(cbuf, pOffset);
	else 
		return (JWavesMP*) new_JWavesSciMP(cbuf, pOffset);
}

/*****************************************************************************/
void delete_JWavesSciIDP(void* vpThis)
{
	JWavesSciIDP* pThis = (JWavesSciIDP*)vpThis;
	if(pThis->pMPs != NULL) pThis->pMPs->del(pThis->pMPs);
	free(pThis);
}

JWavesSciIDP* new_JWavesSciIDP(cbytebuf_t cbuf, size_t uPktOff)
{
	uint2 uReadCRC, uCalcCRC;
	size_t offset;
	JWavesSciIDP* pThis;
	JWavesMP* pMiniPkt;
	
	/* Check the input */
	if(cbuf.len < 20)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "At least 20 bytes are needed for "
		             "a minimum Waves IDP packet, but the buffer was %d bytes.",
		             cbuf.len);
	
	/* Waves CRC is little endian */
	uReadCRC = gse_u2FromLSBbytes(cbuf.ptr);
	
	/* Check the CRC */
	uCalcCRC = gse_crc_ccitt(cbuf.ptr + 2, cbuf.len - 2);
	
	if(uReadCRC != uCalcCRC){
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, 
				      "CRC mismatch for input packet at offset 0x%08X: CRC from file:"
				      "0x%04hX, CRC calculated: 0x%04hX.  CRC is over %d bytes "
                  "starting at %02hhX %02hhX %02hhX %02hhX %02hhX %02hhX.", 
	               uPktOff, uReadCRC, uCalcCRC, cbuf.len - 2, cbuf.ptr[2],
				      cbuf.ptr[3], cbuf.ptr[4], cbuf.ptr[5], cbuf.ptr[6],
				      cbuf.ptr[7]);
		return NULL;
	}
	
	/* Okay, good CRC, go ahead and read the packet */
	pThis = (JWavesSciIDP*)gse_calloc(1, sizeof(JWavesSciIDP));
	pThis->del = delete_JWavesSciIDP;
	pThis->origLen = JWavesSciIDP_origLen;
	pThis->CRCmatch = JWavesSciIDP_CRCmatch;
	pThis->origSeq = JWavesSciIDP_origSeq;
	pThis->calcCRC = JWavesSciIDP_calcCRC;
	
	except_cleanup_push(delete_JWavesSciIDP, pThis);	
	
	pThis->uCRC = uReadCRC;
	
	/* Pull out the seq and telemetry source */
	pThis->uSeq = gse_u2FromMSBbytes(cbuf.ptr + 2);
	pThis->ts   = (pThis->uSeq >> 14) & 0x3;
	pThis->uSeq &= 0x3FFF;
	
	/* Get the SCLK */
	pThis->uSCLK = gse_u4FromMSBbytes(cbuf.ptr + 4);
	
	/* Now get all the mini-packets */
	offset = 8;
	
	pThis->pMPs = new_JWavesMPptrAry(0);	
	while( (pMiniPkt = make_minipacket(cbuf, &offset)) != NULL){
		/* Add error and warning recording */
		pThis->pMPs->add(pThis->pMPs, pMiniPkt);
	}
	
	if(pThis->pMPs->num < 1)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "IDP packet contained no minipackets"
		             ", at offset 0x%08X", uPktOff);
	
	except_cleanup_pop(except_no_call);
	return pThis;	
}
