/** @file JWavesIDP.h Reading Raw IDP data */
#ifndef _JWavesIDP_H_
#define _JWavesIDP_H_

#include <jwaves/JWavesMP.h>
#include <jwaves/JWavesSDSblk.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_jwaves_raw */
/*@{*/
	
/** Contians the waves header plus 0-N minipkt structures.
 * @class JWavesSciIDP
 */ 
typedef struct jwaves_sci_idp{
	
	/** FIELD: The CRC field, a 16-bit @b LSB first integer. */
	uint2 uCRC;
	
	/** FIELD: The sequence field */
	uint2 uSeq;
	
	/** FIELD: The telemetry source bits */
	jwaves_tlm_src ts;
	
	/** FIELD: The Spacecraft Clock and time quality bit.
	 * @warning this clock only has half hertz resolution, the least significant
	 *          bit is the time quality filed. 
	 */
	uint4 uSCLK; 
	
	/** minipkts array pointer */
	JWavesMPptrAry* pMPs;
	
	/** Get the original data length */
	size_t (*origLen)(struct jwaves_sci_idp* pThis);
	
	/** Get the origial sequence value with the telemetry Source embedded in
	 * bytes 15 and 14 */
	uint2 (*origSeq)(struct jwaves_sci_idp* pThis);
	
	/** Does the CRC checkout? */
	bool (*CRCmatch)(struct jwaves_sci_idp* pThis);
	
	/** Get the calculated CRC.
	 * The CRC is calculated in new_JWavesSciIDP(), this just returns the
	 * previously calculated value.
	 */
	uint2 (*calcCRC)(struct jwaves_sci_idp* pThis);

	/** Destructor */
	void (*del)(void* vpThis);
	
	/* private length field, don't use */
	size_t _len; 
	
	/* private, was calculated CRC instead */
	uint2 _calculatedCRC; 
	
}JWavesSciIDP;


/** Make a new IDP structure from raw bytes. 
 * 
 * Empty IDP messages are not parsed.  If the raw bytes contain only
 * an IDP header without an enclosed Minipacket, an exception is thrown.
 
 * The assumed format for the IDP block is:
 * <pre>
 *
 *         +-----+-----+-----+-----+-----+-----+-----+-----+
 *   0 - 7 |    CRC    |    SEQ    |         SCLK          |
 *         +-----+-----+-----+-----+-----+-----+-----+-----+
 *  8 - 16 |           Version 1 Minipkt Header            |
 *         +-----+-----+-----+-----+-----+-----+-----+-----+
 *         |  Mini Packet Data ...
 *	  
 *  
 *         +-----+-----+-----+-----+-----+-----+-----+-----+
 * N - N+7 |           Version 1 Minipkt Header            |
 *         +-----+-----+-----+-----+-----+-----+-----+-----+
 *         |  Mini Packet Data ...
 * </pre>
 *
 * Were the Minipkt header is defined in: @n
 * @n
 *	@c DPU_FSW/trunk/kernel/minipkt.inc
 *	
 * @note This function cannot be used to parse any packets other than SCI data,
 * so Hsk Packet parsing has to be handled by something else.
 *
 * @public
 * @memberof JWavesSciIDP
 */
JWavesSciIDP* new_JWavesSciIDP(cbytebuf_t cbuf, size_t uPktOff);

/** destructor
 * @protected 
 * @memberof JWavesSciIDP
 */
void delete_JWavesSciIDP(void* vpThis);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesIDP_H_ */
