#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseConv.h>

#include "JWavesDefs.h"
#include "JWavesLRSblk.h"


bytebuf_t JWavesLRSblk_encode(GseBlock* pBlk, byte uExtraLen, byte* pOffset)
{
	size_t i;
	byte offset;
	bytebuf_t buf;
	
	JWavesLRSblk* pThis = (JWavesLRSblk*)pBlk;
	GseProcBlk* pProc   = (GseProcBlk*)pBlk;
	
	/* Check prior to encode */
	if(pProc->uStepID != JW_SCELRS_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Processing step ID was %hhu"
				       "expected %hhu.", pProc->uStepID, JW_SCELRS_PROC_STEP_ID);
	
	gse_overflowUinU(uExtraLen + 47, sizeof(byte));
	
	buf = JWavesRdrBlk_encode(pBlk, uExtraLen + 47, &offset);
	gse_assert(buf.len >= offset + 47);
	
	buf.ptr[offset]     = (byte) ((pThis->uUID >> 16) & 0xFF);
	buf.ptr[offset + 1] = (byte) ((pThis->uUID >> 8) & 0xFF);
	buf.ptr[offset + 2] = (byte) (pThis->uUID & 0xFF);
	
	gse_u4ToMSBbytes(buf.ptr + offset + 3, pThis->uRemoteIP);
	gse_u4ToMSBbytes(buf.ptr + offset + 7, pThis->uLocalIP);
	
	for(i = 0; i<20; i++)
		buf.ptr[offset + 11 + i] = pThis->IPhdr[i];
	
	for(i = 0; i<8; i++)
		buf.ptr[offset + 31 + i] = pThis->UDPhdr[i];
	
	for(i = 0; i<8; i++)
		buf.ptr[offset + 39 + i] = pThis->CIPhdr[i];
	
	if(pOffset != NULL) *pOffset = offset + 47;
	return buf;
}

byte JWavesLRSblk_decode(GseBlock* pBlk, cbytebuf_t buf)
{
	GseProcBlk*   pProc = (GseProcBlk*)pBlk;
	JWavesLRSblk* pThis = (JWavesLRSblk*)pBlk;
	
	size_t i;
	byte uOffset;
	
	uOffset = JWavesRdrBlk_decode(pBlk, buf);
	
	if(pProc->uStepID != JW_SCELRS_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x%hhX for the block step "
		             "ID, but recieved 0x%hhX", JW_SCELRS_PROC_STEP_ID,
				       pProc->uStepID);
	
	/* The uid is tricky because it is a 3 byte field */
	pThis->uUID = gse_u4FromMSBbytes(buf.ptr + uOffset - 1);
	pThis->uUID &= 0x00FFFFFF;  
	
	pThis->uRemoteIP = gse_u4FromMSBbytes(buf.ptr + uOffset + 3);
	pThis->uLocalIP  = gse_u4FromMSBbytes(buf.ptr + uOffset + 7);
	
	for(i = 0; i<20; i++)
		pThis->IPhdr[i] = buf.ptr[uOffset + 11 + i];
	
	for(i = 0; i<8; i++)
		pThis->UDPhdr[i] = buf.ptr[uOffset + 31 + i];
	
	for(i = 0; i<8; i++)
		pThis->CIPhdr[i] = buf.ptr[uOffset + 39 + i];
	
	return uOffset + 47;
}

bool JWavesLRSblk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{
	JWavesLRSblk* pThis = (JWavesLRSblk*)pBase;	
	JWavesLRSblk* pLRSoth = (JWavesLRSblk*)pOther;
	int i;
	
	if(! JWavesRdrBlk_canMerge(pBase, pOther, uFlags) ) return false;
	
	if( (pThis->uUID      != pLRSoth->uUID) || 
	    (pThis->uRemoteIP != pLRSoth->uRemoteIP) ||
	    (pThis->uLocalIP  != pLRSoth->uLocalIP)  )  return false;
	
	for(i=0;i<20;i++){
		if( (i<2) || ((i>3)&&(i<10)) || (i>11))
			if(pThis->IPhdr[i] != pLRSoth->IPhdr[i])
				return false;
	}
	
	if(pThis->CIPhdr[7] != pLRSoth->IPhdr[7]) return false;
	
	return true;    
}
				
bool JWavesLRSblk_merge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{
	GseProcBlk* pProc = (GseProcBlk*)pBase;
	JWavesLRSblk* pThis = (JWavesLRSblk*)pBase;	
	
	GseProcBlk* pProcOth = (GseProcBlk*)pOther;
	JWavesLRSblk* pLRSoth = (JWavesLRSblk*)pOther;
	
	uint8 uThis = pProc->microEpoch(pProc);
	uint8 uOther = pProcOth->microEpoch(pProcOth);
	
	if(! JWavesRdrBlk_merge(pBase, pOther, uFlags) ) return false;
	
	if( ((uFlags&GSE_PROC_MERGE_LATE)&&(uOther > uThis)) || (uOther < uThis)){
		memcpy(pThis->IPhdr, pLRSoth->IPhdr, 20);
		memcpy(pThis->UDPhdr, pLRSoth->UDPhdr, 8);
		memcpy(pThis->CIPhdr, pLRSoth->CIPhdr, 8);
	}
	
	return true;
}

/* Copy constructor */
GseBlock* JWavesLRSblk_copy(GseBlock* pBase, size_t uOver)
{
	size_t i;
	JWavesLRSblk* pNew = NULL;
	JWavesLRSblk* pThis = (JWavesLRSblk*)pBase;
	
	if(uOver < sizeof(JWavesLRSblk))
		pNew = (JWavesLRSblk*) JWavesRdrBlk_copy(pBase, sizeof(JWavesLRSblk));
	else
		pNew = (JWavesLRSblk*) JWavesRdrBlk_copy(pBase, uOver); 
	
	pNew->uUID      = pThis->uUID;
	pNew->uRemoteIP = pThis->uRemoteIP;
	pNew->uLocalIP  = pThis->uLocalIP;
	pNew->del       = pThis->del;
	
	for(i = 0; i < 20; i++) pNew->IPhdr[i] = pThis->IPhdr[i];
	for(i = 0; i < 8; i++) pNew->UDPhdr[i] = pThis->UDPhdr[i];
	for(i = 0; i < 8; i++) pNew->CIPhdr[i] = pThis->CIPhdr[i];
	
	return (GseBlock*)pNew;
}

/* GseBlock destructor is fine for deleting this */

JWavesLRSblk* new_JWavesLRSblk(cbytebuf_t buf, size_t uOver)
{
	GseBlock*     pBase = NULL;
	GseProcBlk*   pProc = NULL;
	JWavesRdrBlk* pRdr  = NULL;
	JWavesLRSblk* pThis = NULL;
	
	if(uOver < sizeof(JWavesLRSblk))
		pRdr = new_empty_JWavesRdrBlk(sizeof(JWavesLRSblk));
	else
		pRdr = new_empty_JWavesRdrBlk(uOver); 
	
	pBase = (GseBlock*)pRdr;
	pBase->encode   = JWavesLRSblk_encode;
	pBase->decode   = JWavesLRSblk_decode;
	pBase->copy     = JWavesLRSblk_copy;
	pBase->canMerge = JWavesLRSblk_canMerge;
	pBase->merge    = JWavesLRSblk_merge;
	
	pBase->_sName   = "Waves SCE LRS Reader";
	pBase->_uMinLen += 47;
	
	pProc = (GseProcBlk*)pRdr;
	pProc->uStepID = JW_SCELRS_PROC_STEP_ID;
	
	pThis = (JWavesLRSblk*)pRdr;
	pThis->del     = pRdr->del;
	
	
	/* Calloc in base makes the rest of this 0, as we want */
	if(buf.ptr == NULL) return pThis;
	
	except_cleanup_push(pThis->del, pThis);
	
	JWavesLRSblk_decode(pBase, buf);

	except_cleanup_pop(except_no_call);
	
	return pThis;
}
