/** @file JWavesMP.h Reading Raw Minipkt data */
#ifndef _JWavesMP_H_
#define _JWavesMP_H_

#include <rpwgse/GseTypes.h> 
 
#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_junogse */
/*@{*/
/** @defgroup mod_jwaves_raw Raw Data 
 *
 * Decoding Raw Waves IDP Packets.
 */
/*@}*/
 
/** @addtogroup mod_jwaves_raw */
/*@{*/

#define JW_MRO_MINIPACKET 0x20
#define JW_SCI_MINIPACKET 0xE0

/** Contains the common header and data buffers for one minipkt.@n
 * @n
 * @class JWavesMP
 */
typedef struct jwaves_mp{

	/** The type of minipacket, either JW_SCI_MINIPACKET or JW_MRO_MINIPACKET */
	byte uType;
	
	/** FIELD: Packet ID
	 *
	 * Valid values are:
	 * - 0x4 HFR spectrum
	 * - 0x5 HFR baseband spectrum
	 * - 0x7 HFWBR baseband waveform
	 * - 0x8 HFWBR I or Q waveform
	 * - 0x9 LFR spectrum
	 * - 0xA LBR waveform
	 */
	byte uID;
	
	/** FIELD: Data + Glopped Attn bytes length - 1.
	 * Valid range is 0 to 4095.
	 * @note Use @a getData or @a getDataLen to get the data length regardless
	 * of wether glopping is in use.
	 */
	uint2 uLen;
	
	/** FIELD: Real time interrupt number.
	 * This is a combination of the 40Hz interrput counter plus the SCLK as
	 * follows:
	 * @code
	 *  RTI field = (0x3FF & SCLK)*40 + rti
	 * @endcode
	 * So to get the fraction of a second at which data collection was 
	 * initiated:
	 * @code
	 * frac = (pMiniPkt->uRTI % 40)/40.0;
	 * @endcode
	 * This field can be used in conjunction JWavesMP::uSCLK to determine when
	 * data collection was initiated for this minipacket.
	 * See @ref JWavesMP::collectSCLK.  
	 */ 
	uint2 uRTI;
	
	/** FIELD: MSF Bytes can contain anything must be 0, 2, 4, or 8 bytes long.*/
	cbytebuf_t bufMSF;
	
	/** Get the time of data collection.
	 *
	 * One could practically write a thesis about this function.  
	 * @todo Add diagram about collection time recovery and what assumptions
	 * are in play.
	 *
	 * @param uRepSclk The Space craft clock at time data is reported out of
	 *        the instrument in an IDP packet.  By definition this has to
	 *        be in the same or a later second as the collection initiation 
	 *        time.  This is usually obtained from JWavesSciIDP::uSCLK.
	 *
	 * @return The seconds field of the time of collection.
	 */
	uint4 (*collectSCLK)(struct jwaves_mp* pThis, uint4 uRepSCLK);
	
	/** Get the bytes from which the header was decoded.
	 * @return a constant buffer setup to point to the header bytes.
	 */
	cbytebuf_t (*getHdr)(struct jwaves_mp* pThis);
	
	/** Get the number of header bytes.
	 * This returns the proper header length even if miniptk glopping is
	 * in use.  
	 * @return the actual length, @b not length - 1.
	 */
	byte (*getHdrLen)(struct jwaves_mp* pThis);
	
	/** Get the data bytes.
	 * @return a constant buffer setup to point to the data bytes.
	 */
	cbytebuf_t (*getData)(struct jwaves_mp* pThis);
	
	/** Take the data bytes away.
	 * This function removes ownership of the data byte buffer from the
	 * JWavesMP structure.  Future calls to JWaves::getData will return
	 * an empty buffer, and @a del will not delete the data
	 */
	bytebuf_t (*takeData)(struct jwaves_mp* pThis);
	
	/** Get the number of data bytes.
	 * This returns the proper data length even if minipkt glopping is
	 * in use.
	 * @return the actual length, @b not length - 1.
	 */
	uint2 (*getDataLen)(struct jwaves_mp* pThis);
	
	/* Contains only the header bytes, including any MSF & GLOP bytes */
	bytebuf_t _hdr;
	
	/* Contains only the data bytes */
	bytebuf_t _data;
	
	/* area to store MSF bytes */
	byte _msf[8];
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} JWavesMP;

/** Generate a new empty minipacket base class.
 *
 * @protected
 * @memberof JWavesMP
 */
JWavesMP* new_JWavesMP(size_t uMalloc);

/** Extract the common minipacket fields
 * This is a helper for derived class decoders
 *
 * @throws (GSE_XCEPT, GSE_XDECODE) if @a cbuf is too short.
 * @memberof JWavesMP
 * @protected
 */
void JWavesMP_readCommon(JWavesMP* pThis, cbytebuf_t cbuf);

/** Destructor for Minipacket objects 
 *
 * @protected 
 * @memberof JWavesMP
 */
void delete_JWavesMP(void* vpThis);

/* dynamic array of pointers to jwaves_mp structures */
#define S jwaves_mp_ptr_ary
#define C JWavesMPptrAry
#define T JWavesMP

/* Note, the included file has doxygen documentation that is automatically
   extracted via doc_filter.py at the top level */
#include <rpwgse/ptrAryTplt.h>

#undef S
#undef C
#undef T

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesMP_H_ */
