#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseConv.h>
#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>

#include "JWavesMroBlk.h"
#include "JWavesConv.h"

/****************************************************************************/
/* Helper functions */

void JWavesMroBlk_collectUTC(const JWavesMroBlk* pThis, struct timeval* pTV){
	
	jwaves_sclk2tv(pThis->uColSCLK + (pThis->uColRTI / 40.0), pTV);
}

void JWavesMroBlk_setMPlen(JWavesMroBlk* pThis, uint8 uLen)
{
	if(uLen > 4096)
		pThis->uMPlen = 0;
	else
		pThis->uMPlen = (uint2)(uLen - 1);
}


/****************************************************************************/
/* Encode */

bytebuf_t JWavesMroBlk_encode(GseBlock* pBase, byte uExtraLen, byte* pOffset)
{
	byte offset;
	uint2 u2tmp;
	int i;
	bytebuf_t buf;
	JWavesMroBlk* pThis = (JWavesMroBlk*)pBase;
	
	if(uExtraLen != 0)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "The JWavesMroBlk class is final "
		            "(i.e. should not be derived from), thus uExtraLen should "
		            "always be 0.");
	
	/* Check values before writing crap to the disk */
	if(pThis->uModStep == 0x0)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uModStep cannot be 0");

	if(pThis->uColRTI > 39)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uColRTI > 39");
	
	if(pThis->uSeqNo > 16383)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uSeqNo > 16383");

	if(pThis->uTs > 3)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uTS > 3");

	if(pThis->uMPlen > 4095)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uMPlen > 4095");

	if(pThis->uFrame > 15)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uFrame > 15");

	if(pThis->uReg > 3)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uFrame > 3");
	
	if(pThis->uAddress > 0xFFFFF)
		gse_throw_m(GSE_XCEPT, GSE_XENCODE, "JWavesMroBlk::uAddress > 0xFFFFF");
		
	
	/* Write to the disk */
	buf = GseBlock_encode(pBase, 30, &offset);
	buf.ptr[offset] = pThis->uModStep;
	
	buf.ptr[offset + 1] = pThis->uColRTI;
	
	gse_u4ToMSBbytes(buf.ptr + offset + 2, pThis->uColSCLK);
	gse_u2ToLSBbytes(buf.ptr + offset + 6, pThis->uIDP_CRC);
	
	u2tmp = pThis->uSeqNo + (pThis->uTs << 14);
	gse_u2ToMSBbytes(buf.ptr + offset + 8, u2tmp);
	
	gse_u4ToMSBbytes(buf.ptr + offset + 10, pThis->uRepSclk);
	
	/* 0th MPH byte */
	buf.ptr[offset + 14] = 0xE0 | ((pThis->uMPlen >> 8) & 0x0F);
	
	/* 1st MPH byte */
	buf.ptr[offset + 15] = (pThis->uMPlen & 0xFF);
	
	/* 2nd and 3rd MPH bytes */
	gse_u2ToMSBbytes(buf.ptr + offset + 16, pThis->uMPrti);
	
	/* 4th MPH byte */
	buf.ptr[offset + 18] = pThis->uFrame;
	
	/* 5th MPH byte */
	buf.ptr[offset + 19] = ((pThis->uAddress >> 12) & 0xFF);
	
	/* 6th MPH byte */
	buf.ptr[offset + 20] = ((pThis->uAddress >> 4) & 0xFF);
	
	/* 7th MPH byte, contains three sub fields */
	switch(pThis->uMSFlen){
		case 0: buf.ptr[offset + 21] = 0;    break;
		case 2: buf.ptr[offset + 21] = 0x40; break;
		case 4: buf.ptr[offset + 21] = 0x80; break;
		case 8: buf.ptr[offset + 21] = 0xC0; break;
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "uMSFlen has an invalid "
			             "value: %hhu bytes", pThis->uMSFlen);
	}	
	
	buf.ptr[offset + 21] |= ((pThis->uReg << 4) & 0x30);
	
	buf.ptr[offset + 21] |= (pThis->uAddress & 0xF);
	
	/* The MSF bytes */
	for(i=0; i < pThis->uMSFlen; i++)
		buf.ptr[offset + 22 + i] = pThis->MSFary[i];
	
	
	if(pOffset != NULL) *pOffset = offset + 30;
	return buf;
}

/****************************************************************************/
/* Decode */

byte JWavesMroBlk_decode(struct gse_block* pBase, cbytebuf_t cbuf)
{
	int i;
	byte offset;
	byte uMSFbits;
	uint2 u2Tmp;
	
	JWavesMroBlk* pThis = (JWavesMroBlk*)pBase;
	
	if(pBase->uType != JW_MRO_BLK_TYPE_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x%hhX for the block type "
		             "ID, but recieved 0x%hhX", JW_MRO_BLK_TYPE_ID, pBase->uType);
	
	offset = GseBlock_decode(pBase, cbuf);
	if(pBase->uLen != offset + 30)
	
	pThis->uModStep       = cbuf.ptr[offset];
	pThis->uColRTI        = cbuf.ptr[offset + 1];
	
	pThis->uColSCLK       = gse_u4FromMSBbytes(cbuf.ptr + offset + 2);
	pThis->uIDP_CRC       = gse_u2FromLSBbytes(cbuf.ptr + offset + 6);
	
	u2Tmp                 = gse_u2FromMSBbytes(cbuf.ptr + offset + 8);
	pThis->uTs            = (u2Tmp >> 14) & 0x3;
	pThis->uSeqNo         = u2Tmp & 0x3FFF;
	
	pThis->uRepSclk       = gse_u4FromMSBbytes(cbuf.ptr + offset + 10);
	
	/* Check ID field in the 0th byte */
	if( ((cbuf.ptr[offset + 14] >> 4)&0xF) != 0xE)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected a minipacket ID of 0xE"
		             "instead I got 0x%hhX.", 
		             (byte)((cbuf.ptr[offset + 14] >> 4)&0xF));
	
	
	/* 0th and 1st MPH bytes have the Length Field */
	u2Tmp                 = gse_u2FromMSBbytes(cbuf.ptr + offset + 14);
	pThis->uMPlen         = u2Tmp & 0x0FFF;
	
	/* 2nd and 3rd MPH bytes have the RTI Field */
	pThis->uMPrti         = gse_u2FromMSBbytes(cbuf.ptr + offset + 16);
	
	/* 4th byte contains the frame */
	pThis->uFrame         = cbuf.ptr[offset + 18];
	
	/* 5th and 6th bytes have the upper 16 bits of the Address,
	   and the 7th bytes has the lower 4 bits */
	pThis->uAddress       = gse_u2FromMSBbytes(cbuf.ptr + offset + 19);
	pThis->uAddress       = (pThis->uAddress << 4) & 0xFFFF0;
	pThis->uAddress      |= (cbuf.ptr[offset + 21] & 0xF);
	
	/* The region is also in the 7th byte */
	pThis->uReg           = (cbuf.ptr[offset + 21] >> 4) & 0x03;
	
	/* The MSF size is encoded in the top 2bits of the 7th byte */
	uMSFbits = (cbuf.ptr[offset + 21] >> 6) & 0x03;
	
	switch(uMSFbits){
		case 0:  pThis->uMSFlen = 0; break;
		case 1:  pThis->uMSFlen = 2; break;
		case 2:  pThis->uMSFlen = 4; break;
		case 3:  pThis->uMSFlen = 8; break;
	}
	
	/* Now read out the MSF bytes, if any are meaningful */
	for(i = 0; i< pThis->uMSFlen; i++)
		pThis->MSFary[i] = cbuf.ptr[offset + 22 + i];
	
	return offset + 30;
}

/****************************************************************************/
/* Constructor, and Copy Constructor, (uses base Destructor)                */

GseBlock* JWavesMroBlk_copy(GseBlock* pBase, size_t uMalloc)
{
	JWavesMroBlk* pNew;
	JWavesMroBlk* pThis = (JWavesMroBlk*)pBase;
	size_t uSize = 0;
	int i;
	
	if(uMalloc > sizeof(JWavesMroBlk))
		uSize = uMalloc;
	else
		uSize = sizeof(JWavesMroBlk);
	
	pNew = (JWavesMroBlk*)GseBlock_copy(pBase, uSize);
	
	pNew->uModStep = pThis->uModStep;
	pNew->uTs      = pThis->uTs;
	pNew->uColSCLK = pThis->uColSCLK;
	pNew->uIDP_CRC = pThis->uIDP_CRC;
	pNew->uSeqNo   = pThis->uSeqNo;
	pNew->uRepSclk = pThis->uRepSclk;
	pNew->uMPlen   = pThis->uMPlen;
	pNew->uMPrti   = pThis->uMPrti;
	pNew->uFrame   = pThis->uFrame;
	pNew->uReg     = pThis->uReg;
	pNew->uAddress = pThis->uAddress;

	for(i=0; i<8; i++) pNew->MSFary[i] = pThis->MSFary[i];
	
	pNew->collectUTC = pThis->collectUTC;
	pNew->setMPlen   = pThis->setMPlen;
	pNew->del        = pThis->del;
	
	return (GseBlock*)pNew;
}


JWavesMroBlk* new_JWavesMroBlk(cbytebuf_t buf){

	GseBlock* pBase = NULL;
	JWavesMroBlk* pThis = NULL;
	
	pBase = new_GseBlock(CBBUF_NULL, sizeof(JWavesMroBlk));
	pThis = (JWavesMroBlk*)pBase;/* Override base class items */
	
	/* No new dynalloc stuff, so just use base destructor */
	pBase->encode = JWavesMroBlk_encode;
	pBase->decode = JWavesMroBlk_decode;
	pBase->_sName = "Juno Waves MRO Description";
	pBase->uType  = JW_MRO_BLK_TYPE_ID;
	pBase->copy   = JWavesMroBlk_copy;
	
	pBase->_uMinLen += 30;
	
	/* New items */
	pThis->del          = delete_GseBlock;
	pThis->collectUTC   = JWavesMroBlk_collectUTC;
	pThis->setMPlen     = JWavesMroBlk_setMPlen;
	
	if(buf.ptr == NULL){
		/* Rest of the defaults are all zero, so calloc gets this */ 
		return pThis;
	}
		
	except_cleanup_push(delete_GseBlock, pThis);
	
	JWavesMroBlk_decode(pBase, buf);

	except_cleanup_pop(except_no_call);
	return pThis;	  
}
