/** @file JWavesMroBlk.h Memeory Readout Data Definition Block */
#ifndef _JWavesMroBlk_H_
#define _JWavesMroBlk_H_

#include <sys/time.h>
#include <time.h>

#include <rpwgse/GseBlock.h>
#include <jwaves/JWavesDefs.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_jwaves_pkt */
/*@{*/


/** Interface for Juno Waves Memory Readout Blocks.
 *
 * @class JWavesMroBlk
 * @extends GseBlock
 *
 * The data for this block is encoded in the following manner,
 * (from waves_gs_fmts_stor.txt):
 *
 * <pre>
 *  4.3.1.2 Memory Readout Block
 * ----------------------------
 * This is only present for memory readout blocks.  It is not present if the
 * scince data status block is present.  The data body consists of a memory 
 * dump at the address given it this header.
 * 
 * 
 * Block Size:  56 bytes
 * Block Start: Offset 16 from the start of the GSE header.
 * 
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * 016-023|  0x11  |  0x37  |MODIFIER| COL_RTI|           COLLECT_SCLK            |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * 024-031| 	 IDP_CRC 	 |     SEQ_TS      |            REPORT_SCLK            |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * 032-039|  MPH0  |  MPH1  |  MPH2  |  MPH3  |  MPH4  |  MPH5  |  MPH6  |  MPH7  |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * 040-047|  MSF0  |  MSF1  |  MSF2  |  MSF3  |  MSF4  |  MSF5  |  MSF6  |  MSF7  |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * 
 * MODIFIER:
 *    The ID of the last processing step to modify this block.
 * 
 * COLLECT_SCLK:
 *    The SCLK at the time of the memory dump, gathered from the IDP header SCLK
 *    and the mini-packet RTI field, takes into account the fact that the last bit
 *    of the SCLK is a time quality field.
 * 
 * IDP_CRC:
 *    The original CRC generated by the instrument for the IDP packet that 
 *    contained this memory read out.  This is close to useless.
 * 
 * SEQ_TS:
 *    This is a 16-bit MSB first unsigned integer. 
 * 
 *    * The upper 2 bits of this field contain the telemetry source.  This is
 *    retained so that we always know where the packets originated from.  This
 *    sub field may have the following values:
 *    
 *       0x0 - Housekeeping
 *       0x1 - Low Rate Science
 *       0x2 - High Rate Science
 *     
 *    For MRO data This should always have the value 0x1 for Low Rate Science.
 *    
 *    * The lower 14 bits of this field contain the sequence number.
 * 
 * REPORT_SCLK:
 *    MSB first 32-bit unsigned integer containing the spacecraft sclk at the
 *    time the IDP packet containing this MRO was reported out of the instrument.
 * 
 * MPH0-7:
 *   The minipacket header.  After unpacking, these are the original bits
 *   describing the data and extended headers.  These bits are broken out as
 *   follows:
 *   
 *             bit7    bit6    bit5    bit4    bit3    bit2    bit1    bit0
 *          +---------------------------------------------------------------+
 *    byte0 |   1   |   1   |   1   |   0   | Len11 | Len10 | Len9  | Len8  |
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte1 | Len7  | Len6  | Len5  | Len4  | Len3  | Len3  | Len1  | Len0  |
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte2 | RTI15 | RTI14 | RTI13 | RTI12 | RTI11 | RTI10 | RTI9  | RTI8  |
 *          +---------------------------------------------------------------+
 * 
 *          +---------------------------------------------------------------+
 *    byte3 | RTI7  | RTI6  | RTI5  | RTI4  | RTI3  | RTI3  | RTI1  | RTI0  |
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte4 |                               | FRAME3| FRAME2| FRAME1| FRAME0|
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte5 |  A19  |  A18  |  A17  |  A16  |  A15  |  A14  |  A13  |  A12  |
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte6 |  A11  |  A10  |   A9  |   A8  |   A7  |   A6  |   A5  |   A4  |
 *          +---------------------------------------------------------------+
 *   
 *          +---------------------------------------------------------------+
 *    byte7 | MSF1  | MSF0  | REG1  | REG0  |   A3  |   A2  |   A1  |   A0  |
 *          +---------------------------------------------------------------+
 * 			
 * 
 *    Len
 *    ---
 *    The data length.  Initially this in the amount of minipkt data - 1.
 *    This field is zero'ed by the unsegmentor.
 *  
 *    RTI
 *    ---
 *    This is a combination of the 40Hz interrput counter plus the SCLK as
 *    follows:
 *    
 *      RTI field = (0x3FF & SCLK)*40 + rti
 *     
 *    So to get the fraction of a second at which MRO was read:
 *     
 *       frac = (pMiniPkt->uRTI % 40)/40.0;
 *     
 *    This field can be used in conjunction with the IDP_SCLK to determine
 *    when the memory read was initiated for this minipacket.
 * 
 *    FRAME
 *    -----
 *    If bit A19 is set to 1 then the address is within one of the 16 FFT
 *    512 MB memory frames.  This indicates which WvFE frame was currently
 *    mapped into the upper Y180 memory region.
 * 
 *    A19-A0
 *    ------
 *    The lowest memory address for the dump.  The first byte in the data 
 *    section was taken from this address.  Subsequent bytes are from 
 *    incrementing addresses.
 *   
 *    MSF1-0
 *    ------
 *    The number of extra header bytes.  These contain extra info that is TBD.
 *    This is an enumeration as follows:
 *    
 *       MSF = 0, No MSF bytes
 *       MSF = 1, 2 MSF bytes
 *       MSF = 2, 4 MSF bytes
 *       MSF = 3, 8 MSF bytes
 *      
 *    REG
 *    ---
 *    Most of the lower half of the Y180 memory space may be swapped between
 *    one of four memory regions.  Each of which is 512 MB in length.  This
 *    field indicates from which region the memory was dumped:
 *      
 *      REG = 0, Main Memory
 *      REG = 1, Program Memory
 *      REG = 2, IEB Bulk Memory
 *      REG = 3, Scratch Memory
 *          
 *    There is a suspicion that REG must = 0 if the address is less than 0x5000.
 * 
 * MSF0-7:
 *   Extra header bytes.  Though the "slot" for all 8 MSF bytes is always present
 *   the number that are valid is defined by the MSF subfield inside the MPH
 *   bytes.
 * 
 *</pre>
 */
typedef struct jwaves_mro_blk{

	/** Base structure */
	GseBlock base;

	/** FIELD: MODIFIER */
	byte uModStep;
	
	/** FIELD: COL_RTI */
	byte uColRTI;
	
	/** FIELD: COLLECT_SCLK */
	uint4 uColSCLK;
	
	/** FIELD: IDP_CRC */
	uint2 uIDP_CRC;
	
	/** FIELD: SEQ_NO */
	uint2 uSeqNo;
	
	/** FIELD: TLM_SRC */
	jwaves_tlm_src uTs;
	
	/** FIELD: REPORT_SCLK */
	uint4 uRepSclk;
	
	/** SUBFIELD:  MPH.LEN */
	uint2 uMPlen;
	
	/** SUBFIELD: MPH.RTI */
	uint2 uMPrti;
		
	/** SUBFIELD: MPH.FRAME */
	byte uFrame;
	
	/** SUBFIELD: MPH.REG (region) */
	byte uReg;
	
	/** SUBFIELD: MPH.A (address) */
	uint4 uAddress;
	
	/** FIELD: Number of valid MSF bytes, may be 0, 2, 4, or 8. */
	byte uMSFlen;
	
	/** FIELD: MSF */
	byte MSFary[8];
	
	/** Helper function, get the UTC time value at which this memory was dumped.*/
	void (*collectUTC)(const struct jwaves_mro_blk* pThis, struct timeval* pTV);
	
	/** Helper function, set the @a uMPlen field given the data length.
	 *
	 * @param pThis pointer to a JWavesMroBlk object 
	 * @param uLen The size of the data section of the enclosing GSE Packet 
	 */
	void (*setMPlen)(struct jwaves_mro_blk* pThis, uint8 uLen);
	
	
	/** Destructor */
	void (*del)(void* vpThis);
	
}JWavesMroBlk;


/** Generates a new Juno Waves Memory Dump Definition Block
 * @public
 * @memberof JWavesMroBlk
 */
JWavesMroBlk* new_JWavesMroBlk(cbytebuf_t cbuf);


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesMroBlk_H_ */
