#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseConv.h>

#include "JWavesMroMP.h"

/*****************************************************************************/
/* Constructor */

JWavesMroMP* new_JWavesMroMP(cbytebuf_t bulkBuf, size_t* pOffset)
{
	JWavesMroMP* pThis;
	JWavesMP*    pBase;
	size_t offset;
	cbytebuf_t cbuf;
	int i;
	
	offset = *pOffset;
	
	if(offset >= bulkBuf.len)
		return NULL;
	
	/* Set the start of my area */
	cbuf.ptr = bulkBuf.ptr + offset;
	cbuf.len = bulkBuf.len - offset;
	
	/* Need at least 8 bytes remaining for a minipkt header */
	if(cbuf.len < 8){
		for(i = 0; i<cbuf.len; i++){
			if(cbuf.ptr[i] != 0){
				gse_warning("Possible data loss: Remaining non-zero bytes after the "
				            "previous minipacket are too few to be a minipkt header"); 
				break;
			}
		}				
		return NULL;
	}
	
	pBase = new_JWavesMP(sizeof(JWavesMroMP));
	pBase->uType = JW_MRO_MINIPACKET;
	
	pThis = (JWavesMroMP*)pBase;
	pThis->del = delete_JWavesMP;
	
	/* Now lets setup my vars */
	except_cleanup_push(pThis->del, pThis);
	
	/* Decode the basic stuff: ID, Len, RTI, MSF */
	JWavesMP_readCommon(pBase, cbuf);
	
	if(pBase->uID != 0xE) gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	
	/* NOTE: There is a bug in the ROM MRO generator, it sets the length
	   too high by one.  If we *assume* that MRO packets will never be an
		odd length, we can know the length down by one for odd length packets
		and not affect processing it the future when this bug is corrected.
		*/
	if((pBase->uLen + 1) % 2 != 0){
		gse_notice_v("Correcting improper MRO packet length %d -to-> %d", 
		              pBase->uLen + 1, pBase->uLen);
		
		gse_notice_v("    MRO MP Hdr: %02X %02X %02X %02X %02X %02X %02X %02X ",
		             cbuf.ptr[0], cbuf.ptr[1], cbuf.ptr[2], cbuf.ptr[3], 
				       cbuf.ptr[4], cbuf.ptr[5], cbuf.ptr[6], cbuf.ptr[7]);
		
		pBase->uLen -= 1;
	}
	
	pThis->uFrame = (cbuf.ptr[4] & 0xF);
	
	pThis->uAddress = (((uint4) gse_u2FromMSBbytes(cbuf.ptr+5) ) << 4) & 0xFFFF0;
	pThis->uAddress |= (uint4)(cbuf.ptr[7] & 0x0F);
	
	pThis->uReg = (cbuf.ptr[7] >> 4) & 0x03;
	
	/* Store the raw header bytes */
	pBase->_hdr.len = 8 + pBase->bufMSF.len;
	pBase->_hdr.ptr = (byte*)gse_calloc(pBase->_hdr.len, sizeof(byte));
	memcpy(pBase->_hdr.ptr, cbuf.ptr, pBase->_hdr.len);
	
	/* Store the data bytes */
	pBase->_data.len = pBase->uLen + 1;
	
	if(cbuf.len < (pBase->_hdr.len + pBase->_data.len))
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Remaining buffer is length %d, but at "
		             "least %d bytes are needed for the header (%d bytes) "
		             "and data (%d bytes)",  cbuf.len,
		             pBase->_hdr.len + pBase->_data.len,
		             pBase->_hdr.len, pBase->_data.len);
	
	pBase->_data.ptr = (byte*)gse_calloc(pBase->_data.len, sizeof(byte));
	memcpy(pBase->_data.ptr, cbuf.ptr + pBase->_hdr.len, pBase->_data.len);
	
	/* Increment the offset counter */
	*pOffset = offset + pBase->_hdr.len + pBase->_data.len;
	
	except_cleanup_pop(except_no_call);
	return pThis;
}












