#define _POSIX_C_SOURCE 200112
#include <pthread.h>

#include <errno.h>
#include <limits.h>
#include <strings.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseFilePktSrc.h>
#include <rpwgse/GseMultiFileSrc.h>
#include <rpwgse/GseFileStr.h>

#include "JWavesChanMgr.h"
#include "JWavesStore.h"
#include "JWavesProc.h"
#include "JWavesPkt.h"
#include "JWavesDefs.h"
#include "JWavesConv.h"
#include "JWavesData.h"

#include "JWavesPPD.h"

#define AUTO_LRS  0
#define AUTO_HRS  1

/****************************************************************************
 STUFF THAT HAPPENS IN RUNNING THREAD WITHOUT CONSUMER MUTEX LOCK...
****************************************************************************/

const char* jwreader_sSufBank[] = {
	".lrs.upk|.lrs.useg|.lrs.dcmp|.lrs.ucal",
	".hrs.upk|.hrs.useg|.hrs.dcmp|.hrs.ucal"
};
	

GsePktSrc* JWReader_mkSingleFileSrc(JWReader* pThis, char** psErr)
{
	
	GsePktSrc* pSrc = NULL;
	const char* sSufficies = 0;
	char* sFile = NULL;
	char _sFile[PATH_MAX + 1];
	_sFile[0] = _sFile[PATH_MAX] = '\0';
	
	if(pThis->sFile == NULL){
				
		sSufficies = jwreader_sSufBank[pThis->iWhichAuto];
				
		if( pThis->sGroup == NULL ){
			if(pThis->iWhichAuto == AUTO_LRS)
				*psErr = CS("Error, data source set to \"autolrs\", but no "
			               "data group has been specified.\n");
			else
				*psErr = CS("Error, data source set to \"autohrs\", but no "
	   	               "data group has been specified.\n");

			return NULL;
		}
				
		if(! jwaves_latest_file_s(pThis->sGroup, sSufficies, _sFile, PATH_MAX, GSE_DEBUG)){
			*psErr = gse_string("Auto file select could not locate an input "
				         "under $WAVES_DATA/%s.", pThis->sGroup);
			return NULL;
		}
		sFile = _sFile;
	}
	else{
		sFile = pThis->sFile;
	}
	
	gse_info_v("Reading from file %s\n", sFile);
	pSrc = (GsePktSrc*)new_GseFilePktSrc(sFile, jwaves_blocks, at_eof_retry);
			
	if(pThis->bSkipToEnd){
		pSrc->skipToEnd(pSrc);
		gse_info_v("Waiting at the end of %s.\n", sFile);
		pThis->bSkipToEnd = false;
	}
	
	return pSrc;
}
	
GsePktSrc* JWReader_mkMultiFileSrc(JWReader* pThis, char** psErr)
{
	GsePktSrc* pSrc = NULL;
	GseFileChanMgr* pMgr = NULL;
	const char* sTypes = NULL;
	
	if(pThis->sStop == NULL){
		*psErr = CS("Stop time is missing.\n");
		return NULL;
	}
	if(pThis->sGroup == NULL){
		*psErr = CS("To select data by time index requires that you specify a "
				      "data group.\n");
		return NULL;
	}
	
	if(pThis->iWhichAuto == AUTO_LRS)
		sTypes = "lrs.ucal";
	else if(pThis->iWhichAuto == AUTO_HRS)
		sTypes = "hrs_rec.ucal,hrs_bin.ucal";
	else
		gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		
		
	/* else if(strcasecmp(pThis->sFile, "autosrv") == 0)
		sTypes = "srv.pkt";
	else if(strcasecmp(pThis->sFile, "autobst") == 0)
		sTypes = "rec.pkt,bst.pkt"; */
	
	if(sTypes == NULL){
		*psErr = NS("Unrecognized configuration 'file = %s'\n", pThis->sFile);
		return NULL;
	}
		
	
	pMgr = (GseFileChanMgr*) new_JWavesChanMgr(pThis->sStart, pThis->sStop,
	                                           sTypes, pThis->sGroup);
	
	pSrc = (GsePktSrc*)new_GseMultiFileSrc(pMgr, jwaves_blocks, at_eof_retry, 10);
	
	return pSrc;
}


int JWReader_captureSetup(Producer* pBase, char** psErr)
{
	JWReader* pThis = (JWReader*)pBase;
	
	static const except_id_t catch_spec[] = {{XCEPT_GROUP_ANY, XCEPT_CODE_ANY}};
	except_t* pExcept = NULL;
	int nRet = 0;
	
	except_try_push(catch_spec, 1, &pExcept);
	/* Will return to here if an exception is thrown */
	
	if(pExcept == NULL){
		
		if(pThis->pSrc == NULL){
		
			if(pThis->sStart == NULL)
				pThis->pSrc = JWReader_mkSingleFileSrc(pThis, psErr);
			else	
				pThis->pSrc = JWReader_mkMultiFileSrc(pThis, psErr);
			
			if(pThis->pSrc == NULL) nRet = -1;
		}
	}
	else{
		gse_log_except(GSE_ERROR, pExcept);
		nRet = -1;
	}
	except_try_pop();
	
	return nRet;
}

/****************************************************************************/
/* Read the next packet, if is of a type that is ready to display, then save*/
/* it, only return the halt flag if something is wrong with the packet      */
/* source                                                                   */

int JWReader_capture(Producer* pBase, char** psErr)
{
	static const except_id_t catch_spec[] = {{XCEPT_GROUP_ANY, XCEPT_CODE_ANY}};
	GsePkt* pPkt = NULL;
	GsePktSrc* pSrc;
	JWReader* pThis;
	except_t* pExcept;
	size_t nReadOffset = 0;
	
	pThis = (JWReader*)pBase;
	if(pThis->pSrc == NULL) return -1;
	if(pThis->pDataAry != NULL){ 
		pThis->pDataAry->del(pThis->pDataAry);
		pThis->pDataAry = NULL;
	}
	
	pSrc = pThis->pSrc; /* shorten the code */
	
	if(pThis->bForward){
		if( (pPkt = pSrc->readNext(pSrc)) == NULL){
			*psErr = CS("Null return from readNext\n");
			return 0; /* Try again */
		}
	}
	else{
		if(! pThis->pSrc->supportsPrev(pThis->pSrc) ){
			*psErr = CS("Packet Source does not support reading backwards.");
			pThis->bForward = true;
			return -1;
		}
		if( (pPkt = pSrc->readPrev(pSrc)) == NULL ){
			
			if( pSrc->offset(pSrc) > 0){
				return 0; /* Try again */
			}
			else{
				*psErr = NS("Reached the beginning of %s\n", pSrc->name(pSrc));
				return -1;
			}	
		}
	}
	
	/* Save offest to the top of packet */
	nReadOffset = pSrc->offset(pSrc) - pPkt->pHdr->uTotalLen;
	
	if(pThis->pDataAry != NULL) gse_throw(GSE_XCEPT, GSE_XASSERT);
	
	/* If it ain't science data, silently skip it */
	if( (pPkt->pHdr->uContent & 0xF0) != 0x80 ){
		pPkt->del(pPkt);
		return 0;
	}
	
	if( (pPkt->pHdr->uContent & 0x0F) >= JWAVES_CNT_CAPTURE){

		except_try_push(catch_spec, 1, &pExcept);
		/* Will return to here if an exception is thrown */
		if(pExcept == NULL){
			pThis->pDataAry = fact_JWavesXYdataAry(pPkt, pThis->uXYFlags,
			                                 pSrc->pktIdx(pSrc), nReadOffset);
		}
		else{
			pThis->pDataAry = NULL;
			gse_error_v("Exception caught for packet %d at offset 0x%X, "
			            "report follows:", pSrc->pktIdx(pSrc), nReadOffset);
			gse_log_except(GSE_ERROR, pExcept);
		}
		except_try_pop();
	}
	else
		*psErr = CS("Packet requires further processing before display\n");
	
	pPkt->del(pPkt);
	
	/* If we are trying to get to a certian sclk */
	
	return 0;
}

int JWReader_gotoCapture(Producer* pBase, char** psErr)
{
	
	return -1;
}

/* ************************************************************************ */
/* Free up the most recient XY data structure */

int JWReader_postCapture(Producer* pBase, char** psErr){
	JWReader* pThis = (JWReader*)pBase;
	if(pThis->pDataAry != NULL){
		pThis->pDataAry->del(pThis->pDataAry);
		pThis->pDataAry = NULL;
	}
	return 0;
}

/****************************************************************************
     STUFF THAT HAPPENS WITH CONSUMER MUTEX LOCK...
****************************************************************************/

int JWReader_makeWork(Producer* pBase, Consumer* pCons, int nCons, char** psErr){
	JWReader* pThis = (JWReader*)pBase;
	int i;
	
	/* Make a work item message for the indicated consumer and add link it
	 * to the consumers worklist.  Return non-zero if processing should
	 * stop.  This is the ONLY time we can directly manipulate consumers,
	 * as Producer.main has locked the consumer before this call.  Be quick
	 * here as the consumer may need to get back to work */
	GseXYdata* pCopy = NULL;
	
	/* Take the most recent pData and send it to the consumer */
	if(pThis->pDataAry == NULL){
		/* fprintf(stderr, "Don't have data to send.\n"); */
		return 0;
	}
	
	/* If there is only one consumer, let it have the structures without a 
	 * copy */
	if(nCons < 2){
		
		for(i=0; i<pThis->pDataAry->num; i++){
			pCons->pItem = WI_newHead(pCons->pItem, WT_GSE_XY_DATA, 
					                    (void*)(pThis->pDataAry->pAry[i]) );
			pThis->pDataAry->pAry[i] = NULL;
		}
		pThis->pDataAry->del(pThis->pDataAry);
		pThis->pDataAry = NULL;
		
	}
	else{
		for(i=0; i<pThis->pDataAry->num; i++){
			pCopy = pThis->pDataAry->pAry[i]->copy(pThis->pDataAry->pAry[i]);
			pCons->pItem = WI_newHead(pCons->pItem, WT_GSE_XY_DATA, (void*)pCopy);
		}
	}
	
	return 0;
}


/****************************************************************************
     STUFF THAT HAPPENS BEFORE OR AFTER THREAD MAIN ...
****************************************************************************/

char* JWReader_describe(ConfInter* pConf)
{
	char* pRet;
	int i;
	GseCStrAry* pChan;
	
	GseFileStr* pFS = new_GseFileStr(200);
	
	pFS->putstr(pFS, 
		"\nThis producer reads Juno Waves data.  For each packet read it\n"
		"produces one or more XY data sets.  All data sets in Jowags are\n"
		"identified by a channel name.  The possible channel names\n"
		"associated with data from this producer are:\n\n"
	);
	
	pChan = jwaves_channels();
	for(i = 0; i < pChan->num; i++)
		pFS->printf(pFS, "   %2d. %s\n", i+1, pChan->pAry[i]);
	
	pFS->putstr(pFS, "\nSome of these channels may not appear in flight data.\n");
	
	pRet = pFS->takestr(pFS);
	pFS->del(pFS);
	return pRet;
}

/* Setting the group to read from **************************************** */
int jwreader_setGroup(ConfInter* pConf, const char* sKey, const char* sExt, 
                     void* pVal, char** psRes)
{
	char* sVal = (char*)pVal;
	JWReader* pThis = (JWReader*)pConf;
	
	if(! jwaves_ckGroup(sVal, JWAVES_LOG) ){
		*psRes = NS("Invalid Group name: '%s'\n", sVal);
		return -1;
	}
	else{
		*psRes = NS("Auto select will look for data in group '%s'\n", sVal);
		if(pThis->sGroup != NULL) free(pThis->sGroup);
		pThis->sGroup = CS(sVal);
		return 0;
	}
}

char* jwreader_getGroup(ConfInter* pConf, const char* sKey){
	JWReader* pThis = (JWReader*)pConf;
	
	if( pThis->sGroup != NULL)
		return gse_string("group      = %s\n", pThis->sGroup);	
	else
		return NULL;
}



/* Setting the file to read ********************************************** */
int jwreader_setFile(ConfInter* pConf, const char* sKey, const char* sExt, 
                     void* pVal, char** psRes)
{
	char* sVal = (char*)pVal;
	JWReader* pThis = (JWReader*)pConf;
	
	if(pThis->pSrc != NULL){
		pThis->pSrc->del(pThis->pSrc);
		pThis->pSrc = NULL;
	}
	
	if(pThis->sFile != NULL) free(pThis->sFile);
	
	/* To make people's lives easier, treat 'automatic' as 'autolrs' */
	
	if((strncasecmp(sVal, "autoselect", 9) == 0) ||
		(strncasecmp(sVal, "autolrs", 7) == 0)||
		(strncasecmp(sVal, "autohrs", 7) == 0)  )
	{
		
		if( getenv("WAVES_DATA") == NULL){
			*psRes = gse_string("Error, data source set to \"%s\" but the environ"
		                       "ment variable WAVES_DATA is not defined.\n",sVal);
			return -1;
		}
		
		if(strncasecmp(sVal, "autohrs", 7) == 0){
			*psRes = gse_string("Automatically selecting High Rate Science Data"
					              " from: %s.\n", getenv("WAVES_DATA"));
			pThis->iWhichAuto = AUTO_HRS;
		}
		else{
			*psRes = gse_string("Automatically selecting Low Rate Science Data"
					              " from: %s.\n", getenv("WAVES_DATA"));
			pThis->iWhichAuto = AUTO_LRS;
		}
	}
	else{
		*psRes = NS("Data will be read from: %s.\n", sVal);
		pThis->sFile = CS(sVal);
	}
	return 0;
}

char* jwreader_getFile(ConfInter* pConf, const char* sKey){
	JWReader* pThis = (JWReader*)pConf;
	if( pThis->sFile != NULL){
		return gse_string("file       = %s\n", pThis->sFile);	
	}
	else{
		if(pThis->iWhichAuto == AUTO_LRS)
			return gse_strcopy("file       = autolrs\n");
		if(pThis->iWhichAuto == AUTO_HRS)
			return gse_strcopy("file       = autohrs\n");
	}
	gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	return NULL;
}

/* Reading data by time indicies ******************************************** */
int JWReader_setTime(ConfInter* pConf, const char* sKey, const char* sExt, 
                      void* pVal, char** psRes)
{
	JWReader* pThis = (JWReader*)pConf;	
	char* sVal = (char*)pVal;
	char* pComma = NULL;
	char* pStart = NULL;
	char* pStop  = NULL;

	
	if( (pComma = strchr(pVal, ',') ) == NULL){
		*psRes = NS("Comma missing in time range: '%s'\n", sVal);
		return -1;
	}
	
	pStart = sVal;
	pStop = pComma + 1;
	*pComma = '\0';      /* keyValue seems to allocate this on the heap, so okay to mod */
	
	if(strlen(pStart) < 1){
		*psRes = CS("Empty start time\n");
		return -1;
	}
	if(strlen(pStop) < 1){
		*psRes = CS("Empty stop time\n");
		return -1;
	}
	
	if(pThis->sStop != NULL) free(pThis->sStop);
	if(pThis->sStart != NULL) free(pThis->sStart);
	
	pThis->sStart = CS(pStart);
	pThis->sStop = CS(pStop);
	
	*psRes = NS("Data will be read from times %s to %s\n", pThis->sStart, pThis->sStop);
	return 0;
}

char* JWReader_getTime(ConfInter* pConf, const char* sKey){
	JWReader* pThis = (JWReader*)pConf;
	
	if(pThis->sStart != NULL)
		return NS("time       = %s,%s\n", pThis->sStart, pThis->sStop);
	else	
		return NULL;
}



/* Interactive command, reset to begininng of the file ******************** */
int jwreader_setReset(ConfInter* pConf, const char* sKey, const char* sExt, 
                      void* pVal, char** psRes){
	JWReader* pThis = (JWReader*)pConf;
	
	if(pThis->pSrc != NULL){
		*psRes = gse_string("Reset reading to the beginning of %s\n", 
		                    pThis->pSrc->name(pThis->pSrc));
		if(pThis->pSrc->supportsRewind(pThis->pSrc)){
			pThis->pSrc->rewind(pThis->pSrc);
		}
		else{
			/* okay do this the hard way */
			pThis->pSrc->del(pThis->pSrc);
			pThis->pSrc = NULL;
		}
		return 0;
	}		
	
	if(pThis->sFile != NULL)
		*psRes = gse_string("Command ignored (already set to beginning of %s)\n",
		                    pThis->sFile);
	else
		*psRes = gse_strcopy("Command ignored (no-current file).");

	return 0;
}

/* Set the reading direction */
int JWReader_setReadDir(ConfInter* pConf, const char* sKey, const char* sExt, 
                      void* pVal, char** psRes)
{
	int nBack;
	JWReader* pThis = (JWReader*)pConf;
	nBack = *((int*)pVal);
	
	if(nBack == 0){
		pThis->bForward = false;
		*psRes = gse_strcopy("The data file will be read in reverse\n");
	}
	else{
		pThis->bForward = true;
		*psRes = gse_strcopy("The data file will be read normally\n");
	}
	return 0;
}

char* JWReader_getReadDir(ConfInter* pConf, const char* sKey){
	JWReader* pThis = (JWReader*)pConf;
	if(pThis->bForward)
		return NULL;
	else
		return gse_strcopy("back =      0\n");
}

/* PLL Lock Packets X Axis */
int JWReader_setPllOsc(ConfInter* pConf, const char* sKey, const char* sExt, 
                       void* pVal, char** psRes)
{
	int nPllOsc;
	JWReader* pThis = (JWReader*)pConf;
	nPllOsc = *((int*)pVal);
	
	if(nPllOsc == 0){
		pThis->uXYFlags &= ~JW_DATA_HFRLOCK_OSC_FREQ;
		*psRes = gse_strcopy("PLL Lock data will have the Center Freq. as X axis\n");
	}
	else{
		pThis->uXYFlags |= JW_DATA_HFRLOCK_OSC_FREQ;
		*psRes = gse_strcopy("PLL Lock data will have the Oscillator Freq. as X axis\n");
	}
	
	return 0;
}
		
char* JWReader_getPllOsc(ConfInter* pConf, const char* sKey){
	JWReader* pThis = (JWReader*)pConf;
	if(pThis->uXYFlags & JW_DATA_HFRLOCK_OSC_FREQ)
		return gse_strcopy("pll_osc =   1\n");
	else
		return NULL;
}

/* SPECTRAL DENSITY MODE */
int JWReader_setDivBand(ConfInter* pConf, const char* sKey, const char* sExt, 
                       void* pVal, char** psRes)
{
	int nDivBand;
	JWReader* pThis = (JWReader*)pConf;
	nDivBand = *((int*)pVal);
	
	if(nDivBand == 0){
		pThis->uXYFlags &= ~JW_DATA_DIV_BANDWIDTH;
		*psRes = gse_strcopy("Set to display pure tones from onboard specta data \n");
	}
	else{
		pThis->uXYFlags |= JW_DATA_DIV_BANDWIDTH;
		*psRes = gse_strcopy("Dividing onboard specta data by the num of bins.\n");
	}
	
	return 0;
}
		
char* JWReader_getDivBand(ConfInter* pConf, const char* sKey)
{
	JWReader* pThis = (JWReader*)pConf;
	if(pThis->uXYFlags & JW_DATA_DIV_BANDWIDTH)
		return gse_strcopy("div_band =    1\n");
	else
		return NULL;
}

/* Marking Decompressed Packets */
int JWReader_setMarkDcmp(ConfInter* pConf, const char* sKey, const char* sExt, 
                         void* pVal, char** psRes)
{
	int nMarkDcmp;
	JWReader* pThis = (JWReader*)pConf;
	nMarkDcmp = *((int*)pVal);
	
	if(nMarkDcmp == 0){
		pThis->uXYFlags &= ~JW_DATA_MARK_DCMP;
		*psRes = gse_strcopy("Stardard channel names will be used for decomp. data\n");
	}
	else{
		pThis->uXYFlags |= JW_DATA_MARK_DCMP;
		*psRes = gse_strcopy("Decomp. data will have ' dcmp' appended to the channel name\n");
	}
	
	return 0;
}
		
char* JWReader_getMarkDcmp(ConfInter* pConf, const char* sKey)
{
	JWReader* pThis = (JWReader*)pConf;
	if(pThis->uXYFlags & JW_DATA_MARK_DCMP)
		return gse_strcopy("mark_dcmp =   1\n");
	else
		return NULL;
}	

/* Skipping to the end of the input file */

int JWReader_setSkipToEnd(ConfInter* pConf, const char* sKey, const char* sExt, 
                         void* pVal, char** psRes)
{
	JWReader* pThis = (JWReader*)pConf;
	int nSkipToEnd = *((int*)pVal);
	
	if(nSkipToEnd != 0){
		
		if(pThis->sStart != NULL){
			*psRes = CS("skip_to_end is incompatible with time based data access\n");
			return -1;
		}
		
		*psRes = gse_strcopy("Skipping to the end of the input file before the 1st read\n");
		pThis->bSkipToEnd = true;
	}
	else{
		*psRes = gse_strcopy("Reading will start at the beginning of the file\n");
		pThis->bSkipToEnd = false;
	}
	return 0;
}

char* JWReader_getSkipToEnd(ConfInter* pConf, const char* sKey)
{
	JWReader* pThis = (JWReader*)pConf;
	if(pThis->bSkipToEnd)
		return gse_strcopy("skip_to_end = 1\n");
	else
		return NULL;
}


/****************************************************************************/
/* Allocators, Deallocators */

void delete_JWReader(void* vpThis)
{
	JWReader* pThis = (JWReader*)vpThis;
	
	if(pThis->pDataAry != NULL) pThis->pDataAry->del(pThis->pDataAry);
	if(pThis->pSrc != NULL) pThis->pSrc->del(pThis->pSrc);
	if(pThis->sGroup != NULL) free(pThis->sGroup);
	if(pThis->sFile != NULL) free(pThis->sFile);
	
	/* BUG HERE?  sStart and sStop are not freed */
	
	delProducer(vpThis);
}


JWReader* new_JWReader(int* pFlag, int4 nLevel, size_t uOverMalloc)
{
	size_t uSize;
	ConfInter* pConf = NULL;
	Producer* pBase = NULL;
	JWReader* pThis = NULL;
	
	if(uOverMalloc > sizeof(JWReader))
		uSize = uOverMalloc;
	else
		uSize = sizeof(JWReader);
	
	if( (pBase = newProducer(pFlag, nLevel, uSize)) == NULL) return NULL;
	
	pConf = (ConfInter*)pBase;
	pThis = (JWReader*)pBase;
	
	free(pConf->sName);
	pConf->sName = gse_string("JW-Reader");
	pConf->describe = JWReader_describe;
	
	
	pConf->addItemNB(pConf, "group", DT_STR, "Juno Waves data group",
	"The data group containing the data to display.  Only needed if the \n"
	"'file' paramenter is on one of the automatic settings.  At present \n"
	"there is no default value for this parameter, it must be provided\n"
	"if 'file' is 'autolrs' or 'autohrs'\n",
	&jwreader_setGroup, &jwreader_getGroup);
	
	pConf->addItemNB(pConf, "file", DT_STR, "Juno Waves gse packet file",
	"A relative or absolute path to waves data that is in gse-packet format\n"
	"Packets that can be read from this file are converted to X-Y data and\n"
	"passed on to data consumers.\n"
	"Special values:\n"
	"\n"
	"    autolrs - Automatically lookup the latest Low Rate Science Data\n"
	"    autohrs - Automatically lookup the latest High Rate Science Data\n"
	"\n"
	"Using automatic lookup requires that a data 'group' be specified\n",
/*	"The packets do not need to be processed, thus packed, segmented, \n"
	"compressed and glopped data are readable, but with a prefromance hit for\n"
	"on the fly packet processing." */
	&jwreader_setFile, &jwreader_getFile);
	
	pConf->addItemNB(pConf, "reset", DT_NONE, 
	"Reset packet reading to the beginning of the file.",
	"When stopped the JW-Reader doesn't start back at the beginning of the\n"
	"file.  This command, with no equal sign will reset it.\n",
	&jwreader_setReset, NULL);
	
	
	pConf->addItemNB(pConf, "time", DT_STR, "Time based data selection",
	"This option sets the reader up to gather data by SCLK or SCET ranges\n"
	"instead of by filenames.  Either SCLKs or SCETs may be used for the time\n"
	"values.  Mixing of SCLK and SCET values is no problem.  Some examples:\n"
	"\n"
	"   time=0x15708120,0x15709700   # SCLKs for STL selftest on 2011-05-26\n"
	"\n"
	"   time=2011-05-26,2011-05-27   # all data for the 26th\n"
	"\n"
	"   time=2011-05-26T16:06:22.014,2011-05-26T16:10:38 # small slice of data\n"
	"\n"
	"   time=now,forever             # from now till forever\n"
	"\n"
	"   time=-20,now                 # from 20 minutes ago to now\n"
	"\n"
	"Using the 'time' option also requires the 'group' option\n",
	&JWReader_setTime, &JWReader_getTime);
	
	
	pConf->addItem(pConf, "back", DT_INT, 0, 1, "Set the read direction",
	"To read backwards:\n"
	"\n"
	"    back = 1\n"
	"\n"
	"To read forwards:\n"
	"\n"
	"    back = 0\n", 
	&JWReader_setReadDir, &JWReader_getReadDir);
	
	pConf->addItem(pConf, "pll_osc", DT_INT, 0, 1, 
	"Output PLL Lock packets with the Oscillator freq. as the X axis",
	"By default, PLL Lock packets use the mixer center frequency as \n"
	"the independent variable.  Setting this to 1 alters the X axis\n"
	"to be the oscillator frequency in MHz and reorders the Y data\n"
	"accordingly.\n", &JWReader_setPllOsc, &JWReader_getPllOsc);
	
	pConf->addItem(pConf, "div_band", DT_INT, 0, 1, 
	"Spectral Density Mode: Divide onboard FFT products by the num. FFT bins",
	"Waves flight software often uses multiple FFT bins to produce a \n"
	"single output spectral value.  In order to faithfully reproduce the\n"
	"amplitude of of pure tones, this reader does not divide by the number\n"
	"FFT bins by default.  This has the effect of raising the noise floor \n"
	"at higher frequencies.  Setting this option to '1' will divide the \n"
	"output data by the number of onboard FFT bins that were averaged \n"
	"together.\n",
	&JWReader_setDivBand, &JWReader_getDivBand);
	
	pConf->addItem(pConf, "mark_dcmp", DT_INT, 0, 1,
	"Add the text ' dcmp' to channel names for decompressed data",
	"This is useful for dual route data, makes the channel names for \n"
	"data that has been decompressed different from equivalent packets\n"
	"that have not been run through the onboard Rice compressor.\n",
	&JWReader_setMarkDcmp, &JWReader_getMarkDcmp);
	
	pConf->addItem(pConf, "skip_to_end", DT_INT, 0, 1,
	"Skip to the end of the input file",
	"Skip to the end of the input file and wait for more data, useful \n"
	"for live feeds on long files.\n",
	&JWReader_setSkipToEnd, &JWReader_getSkipToEnd);
	
	/* The following three functions are called from the Producer.prodMain.
	 * Since thread syncronization is tricky, instead of creating a whole
	 * new thread main, just redirect certian function calls.  To understand
	 * when these will be called, read the source for prodMain in prodThread.c
	 */
	 
	/* When called to setup for capturing data, use this function */
	pBase->captureSetup = JWReader_captureSetup;
	
	/* When called to capture data, use this capture function */
	pBase->capture = JWReader_capture;
	
	/* When called to pass off data to a consumer use this function */
	pBase->makeWork = JWReader_makeWork;
	
	/* When called to clean up after a capture, use this function */
	pBase->postCapture = JWReader_postCapture;
	
	/* Derived class setup only involves setting pointers to null but this
	   was handled by calloc in newConfIntr */
	
	pBase->del = delete_JWReader;
	pThis->del = delete_JWReader;
	
	pThis->bSkipToEnd = false;
	pThis->bForward = true;
	pThis->uXYFlags = 0U;
	pThis->iWhichAuto = AUTO_LRS;
	pThis->sGroup = NULL;
	pThis->sStart = NULL;
	pThis->sStop =  NULL;
	
	return pThis;
}

/****************************************************************************/
/* The PPD plugin interface */

ConfInter* newGVThreadObj(int* pFlag, int4 nLevel)
{
	return (ConfInter*) new_JWReader(pFlag, nLevel, 0);
}

/** Required plugin function, returns a pointer to the Producer name */
const char* GVThreadObjName(){
	return (const char*)"JW-Reader";
}

/** Required plugin funciton.
 * @returns 1, since this plugin is a producer.
 */
int GVThreadObjType(){
	return 1;
}

/** Required plugin function. 
 * @returns the version as a new string allocated on the heap.
 */
char* GVThreadObjVer(){
	return gse_strcopy("0.9, " __DATE__);
}
