/** @file JWavesPkt.h */
#ifndef _JWavesPkt_H_
#define _JWavesPkt_H_

#include <rpwgse/GseBlock.h>
#include <rpwgse/GseConv.h>

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_junogse */
/*@{*/

/** @example waves_read.c
 * Reading packets from a file and displaying LFR Lo E data bytes in hex.
 */

/** @example waves_plot.c
 * Plotting data using plplot.
 */
 
/** @example waves_orig.c Example of iterating over juno waves packets in
 * a file and printing bytes in a similar manner as they are generated in the
 * instrument.
 */

	

/** @defgroup mod_jwaves_pkt Juno Waves Packets 
 *
 * Defines the GSE Header blocks specific to Juno Waves packets.
 * The top level data structure containing a Waves GSE packet is still
 * the @ref GsePkt. Only the @ref GseBlock derived structures are specific.
 */
/*@}*/


/** @addtogroup mod_jwaves_pkt */
/*@{*/

/* These are useful for "blob of bytes" reading mentality, these are not
 * generically useful.  They only apply to unpacked Juno Waves GSE packets
 *
 * The object oriented factory function below works in all cases.
 */

#ifndef _JWavesPkt_C_
/* Location and definition of last modifing process in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MODIFIER;

/* Location and definition of lowest numeric telemetry source in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_LO_TLM_SRC;

/* Location and definition of highest numeric in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_HI_TLM_SRC;

/* Location and definition of collection time SCLK in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_COL_SCLK;

/* Location and definition of collection time RTI in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_COL_RTI;

/* Location and definition of number of collection averaged in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_AVG;

/* Location and definition of highest sequence number in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_HI_SEQ_NO;

/* Location and definition of latest report SCLK in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_HI_REP_SCLK;

/* Location and definition of instrument generated CRCs in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_IDP_CRC;

/* Location and definition of lowest sequence number in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_LO_SEQ_NO;

/* Alais for JW_FD_LO_SEQ_NO */
extern const gse_field_def JW_FD_SEQ_NO;

/* Location and definition of earliest report out SCLK in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_LO_REPORT_SCLK;

/* Alias for JW_FD_LO_REPORT_SCLK */
extern const gse_field_def JW_FD_REPORT_SCLK;

/* Location and definition of minipacket ID in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_ID;

/* Location and definition of minipacket length in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_LEN;

/* Location and definition of minipacket RTI in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_RTI;

/* Location and definition of glop bit in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_GLOP;

/* Location and definition of preamp attenuation bit in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_PATN;

/* Location and definition of LFR gain in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_BGAIN;

/* Location and definition of End of Segments bit in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_EOS;

/* Location and definition of the Cycle or Segment field in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_SEG_0R_CYCLES;

/* Location and definition of HFR/HFWBR step attenuator in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_ATTN;

/* Location and definition of the source in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_SRC;

/* Location and definition of mixer band or cycle seconds in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_BND_OR_CYCSEC;

/* Location and definition of more status follows bits in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_MSF;

/* Location and definition of the format bits in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_MP_FMT;

/* Location and definition of the initial read time in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_WKS_TIME;

/* Location and definition of the reader account user ID in unpacked Waves GSE Packets */
extern const gse_field_def JW_FD_WKS_UID;

/* Location and definition of the SCE IP address in unpacked Waves GSE 
 * Packets that were originally collected from the LRS interface.
 */
extern const gse_field_def JW_FD_LRS_SCE_IP;

/* Location and definition of the Workstation IP address in unpacked Waves GSE 
 * Packets that were originally collected from the LRS interface.
 */
extern const gse_field_def JW_FD_LRS_WKS_IP;

#endif /* _JWavesPkt_C_ */

/** Bytes in standard Waves GSE header, can be longer */
#define  WAVES_GSE_HDR_SZ 256

/** The offset to the IP header in Waves GSE packets collected from the SCE's
 * LRS interface.
 */ 
#define JW_FL_LRS_IP_HDR 96

/** The offset to the UDP header in Waves GSE packets collected from the SCE's
 * LRS interface.
 */ 
#define JW_FL_LRS_UDP_HDR 116

/** The offset to the CIP header in Waves GSE packets collected from the SCE's
 * LRS interface.
 */ 
#define JW_FL_LRS_CIP_HDR 124


/* This is useful for Object based handling of Juno Waves Packets */

/** Juno Waves Header block generator.
 * 
 * This is usually supplied as a callback to new_GseFilePktSrc().  It handles
 * parsing header block bytes into Juno Waves header block structures.  Any
 * unrecognized block types are passed on to gse_generic_blocks().
 * 
 * @param nSpaceCraft The NAIF spacecraft code, supplied by the packet reader.
 *
 * @param uType The block type, supplied by the packet reader.
 *
 * @param uProcess For processing blocks, the process type id.  Supplied by the
 *        packet reader.
 *
 * @param buf The byte blob, supplied by the packet reader.
 * 
 * @return A pointer to a GseBlock, GseProcBlk, JWavesSDSblk, JWavesRdrBlk,
 *         JWavesHRSblk, JWavesLRSblk, JWavesUpkBlk, or JWavesUsegBlk structure
 *         depending on the values of @a Type and @a uProcess.  All these 
 *         structures are <i>cast compatible</i> with GseBlock.  To free the
 *         structure call its @a del member.
 *
 * @see GseFilePktSrc and gse_generic_blocks.
 */
GseBlock* jwaves_blocks(int2 nSpaceCraft, byte uType, byte uProcess,
                        cbytebuf_t buf);


/** Reproduce "original" header out of the Waves instrument.
 *
 * Convenience function intended for flight software implimentors that wish to
 * view the data in a manner similar to how it is generated in the instrument.
 * Not that data can be be unpacked, unsegmented, decompressed. and deglopped
 * by the ground software.  To get a faithful reproduction, supply this funtion
 * with data from ".raw" or ".upk" files.
 *
 * @see example @link waves_orig.c @endlink for an example of using this
 * function.
 *
 * @warning If more than one minipacket has been placed within a single 
 *          waves packet then only the 1st header is returned.
 *
 * @param pPkt A GSE packet structure produced from Juno Waves data.  Note
 *        that the generic block parsing function may be used to produce the
 *        packet.
 *
 * @return A constant bytebuffer structure that points to memory within
 *         @a pPkt. Thus when @a pPkt is deleted, the return value for
 *         this function is invalidated.
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) If @a pPkt is not from Juno Waves.
 * @throws (GSE_XCEPT,GSE_XDECODE) If @a pPkt is too short to contian a minipacket header.
 *       
 */
cbytebuf_t jwaves_inst_hdr(GsePkt* pPkt); 

/** Reproduce "original" data out of the Waves instrument.
 *
 * Convenience function intended for flight software implimentors that wish to
 * view the data in a manner similar to how it is generated in the instrument.
 * Not that data can be be unpacked, unsegmented, decompressed. and deglopped
 * by the ground software.  To get a faithful reproduction, supply this funtion
 * with data from ".raw" or ".upk" files.
 *
 * @see example @link waves_orig.c @endlink for an example of using this
 * function.
 *
 * @warning If more than one minipacket has been placed within a single
 *          waves packet then only the data section will contain headers
 *          for packets after the lst.
 *
 * @param pPkt A GSE packet structure produced from Juno Waves data.  Note
 *        that the generic block parsing function may be used to produce the
 *        packet.
 *
 * @return A constant bytebuffer structure that points to memory within
 *         @a pPkt.  Thus when @a pPkt is deleted, the return value for
 *         this function is invalidated.
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) If @a pPkt is not from Juno Waves.
 */
cbytebuf_t jwaves_inst_data(GsePkt* pPkt);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesPkt_H_ */
