#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseConv.h>

#include "JWavesDefs.h"
#include "JWavesRdrBlk.h"


/*****************************************************************************/
/* Utility functions, get any one of a number of reader blocks */


JWavesRdrBlk* jwaves_get_rdr_blk(GsePkt* pPkt)
{
	int i;
	
	/* known compatible reader blocks */
	byte types[] = {JW_RDR_PROC_STEP_ID,   JW_SCELRS_PROC_STEP_ID,
			           JW_SCEHRS_PROC_STEP_ID, JW_FDM_PROC_STEP_ID};
	
	for(i = 0; i < 4; i++){
		if( pPkt->hasUniqueBlkByType(pPkt, GSE_PROC_BLK_TYPE, types[i]))
			return (JWavesRdrBlk*)pPkt->getBlkByType(pPkt, GSE_PROC_BLK_TYPE, types[i]);
	}
				
	return NULL;		
}


/*****************************************************************************/

bytebuf_t JWavesRdrBlk_encode(GseBlock* pBase, byte uExtraLen, byte* pOffset)
{
	byte offset;
	bytebuf_t buf;
	JWavesRdrBlk* pThis = (JWavesRdrBlk*)pBase;
	
	gse_overflowUinU(uExtraLen + 1, sizeof(byte)); 
	
	buf = GseProcBlk_encode(pBase, uExtraLen + 1, &offset);		
	
	buf.ptr[offset] = (pThis->uSource & 0x0F);
	
	if(pOffset != NULL) *pOffset = offset + 1;
	return buf;
}

byte JWavesRdrBlk_decode(GseBlock* pBase, cbytebuf_t buf)
{
	JWavesRdrBlk* pThis = (JWavesRdrBlk*)pBase;
	
	if(GseProcBlk_decode(pBase, buf) != 12)
		gse_throw(GSE_XCEPT, GSE_XLIBINCOMPAT);
	 	
	if(pBase->uLen < 12)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected at least 12 (i.e. 13 - 1) "
		             "for the Reader block  length, but recieved %hhu",
		             pBase->uLen);
	
	/* Decoding */
	pThis->uSource = (buf.ptr[12]) & 0x0F;

	return 13;		
}

bool JWavesRdrBlk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{
	if(!GseProcBlk_canMerge(pBase, pOther, uFlags)) return false;
	
	if( ((JWavesRdrBlk*)pBase)->uSource == ((JWavesRdrBlk*)pOther)->uSource )
		return true;
	else
		return false;
}

GseBlock* JWavesRdrBlk_copy(GseBlock* pBase, size_t uOverMalloc)
{
	JWavesRdrBlk* pNew = NULL;
	JWavesRdrBlk* pThis = (JWavesRdrBlk*)pBase;
	size_t uSize = 0;
	 
 	if(uOverMalloc > sizeof(JWavesRdrBlk))
		uSize = uOverMalloc;
	else
		uSize = sizeof(JWavesRdrBlk);
	
	pNew = (JWavesRdrBlk*)GseProcBlk_copy(pBase, uSize);
	pNew->uSource = pThis->uSource;
	pNew->del     = pThis->del;
	
	return (GseBlock*)pNew;
}

JWavesRdrBlk* new_JWavesRdrBlk(cbytebuf_t buf, size_t uOver)
{
	JWavesRdrBlk* pThis = new_empty_JWavesRdrBlk(uOver);
	
	if(buf.ptr == NULL) return pThis;
	
	except_cleanup_push(delete_GseBlock, pThis);
	
	JWavesRdrBlk_decode((GseBlock*)pThis, buf);

	except_cleanup_pop(except_no_call);
	return pThis;
}

JWavesRdrBlk* new_empty_JWavesRdrBlk(size_t uOver)
{
	GseBlock* pBase = NULL;
	GseProcBlk* pProc = NULL;
	JWavesRdrBlk* pThis = NULL;
	
	if(uOver < sizeof(JWavesRdrBlk))
		pProc = new_GseProcBlk(CBBUF_NULL, sizeof(JWavesRdrBlk));
	else
		pProc = new_GseProcBlk(CBBUF_NULL, uOver);
	
	pBase = (GseBlock*)pProc;
	pThis = (JWavesRdrBlk*)pProc;
	
	/* Override a couple base class functions */
	pBase->encode = JWavesRdrBlk_encode;
	pBase->decode = JWavesRdrBlk_decode;
	pBase->_sName = "Waves Reader";
	pBase->copy   = JWavesRdrBlk_copy;
	pBase->canMerge  = JWavesRdrBlk_canMerge;
	
	/* Since we can't merge unless the uSource is the same, no special 
	   merging functionality is needed */
	
	pProc->uStepID = JW_RDR_PROC_STEP_ID;
	
	pBase->_uMinLen += 1;
	
	/* Set new items */
	pThis->del = pProc->del;
	
	/* Rest of the defaults are all zero, so calloc gets this */
	return pThis;
}
