#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseConv.h>
#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>

#include "JWavesSDSblk.h"
#include "JWavesConv.h"


/****************************************************************************/

bool JWavesSDSblk_isSegmented(const JWavesSDSblk* pThis){
	/* If SEOF is set and the segment no. is zero this is
	   not segmented, otherwise it is */
	
	return ! (pThis->bEOC && (pThis->uSegNo == 0));
}

bool JWavesSDSblk_isGlopped(const JWavesSDSblk* pThis){
	return pThis->bGlop;
}

bool JWavesSDSblk_isCompressed(const JWavesSDSblk* pThis){
	return jwaves_isCompressed(pThis->uFmt);
}

void JWavesSDSblk_collectUTC(const JWavesSDSblk* pThis, struct timeval* pTV){
	
	jwaves_sclk2tv(pThis->uColSCLK + (pThis->uColRTI / 40.0), pTV);
}

void JWavesSDSblk_setMPlen(JWavesSDSblk* pThis, uint8 uLen)
{
	if(uLen > 4096)
		pThis->uMPlen = 0;
	else
		pThis->uMPlen = (uint2)(uLen - 1);
}

/****************************************************************************/
bytebuf_t JWavesSDSblk_encode(GseBlock* pBase, byte uExtraLen, byte* pOffset)
{
	int i, iPos;
	bytebuf_t buf;
	byte offset;
	byte uTmp;
	
	JWavesSDSblk* pThis = (JWavesSDSblk*)pBase;
	
	if(uExtraLen != 0)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "The JWavesSDSBlk class is final "
		            "(i.e. should not be derived from), thus uExtraLen should "
		            "always be 0.");
	
	jwaves_valid_idsrc(pThis->uPSID, JWAVES_LOG);
	
	/* TODO: add a throw if uExtraLen or pOffset if they aren't zero */
	buf = GseBlock_encode(pBase, 54, &offset);
	gse_assert( offset == 2 ); 
	
	buf.ptr[2] = pThis->uModStep;
	buf.ptr[3] = ((pThis->uTsLo << 4)&0x30) | (pThis->uTsHi & 0x03);
	gse_u4ToMSBbytes(buf.ptr + 4, pThis->uColSCLK);
	buf.ptr[8] = pThis->uColRTI;
	buf.ptr[9] = pThis->uAvg;
	gse_u2ToMSBbytes(buf.ptr + 10, pThis->uSeqNoHi);
	gse_u4ToMSBbytes(buf.ptr + 12, pThis->uRepSclkHi);
	gse_u2ToLSBbytes(buf.ptr + 16, pThis->uIDP_CRC);
	gse_u2ToMSBbytes(buf.ptr + 18, pThis->uSeqNoLo);
	gse_u4ToMSBbytes(buf.ptr + 20, pThis->uRepSclkLo);
	
	/* 0th Minipkt header byte */
	buf.ptr[24] = (pThis->uPSID & 0xF0) | ((pThis->uMPlen >> 8) & 0x0F);
	
	/* 1st Minipkt header byte */
	buf.ptr[25] = (pThis->uMPlen & 0xFF);
	
	/* 2nd and 3rd Minipkt Header byte */ 
	gse_u2ToMSBbytes(buf.ptr + 26, pThis->uMPrti); 
	
	/* 4th byte of the MPH is radically different depending on glopping */
	if(pThis->bGlop) 
		uTmp = 0x80;
	else 
		uTmp = 0x00;
	
	if(pThis->uAttn & 0x20) uTmp |= 0x40; /* Preamp Attn */
	if(pThis->uAttn & 0x10) uTmp |= 0x20; /* BGAIN */
	
	/* Glop Cycle or SegNo */
	if(pThis->bGlop){ 
		uTmp |= (pThis->uCycles & 0x0F) + 0x10;
	}
	else{
		uTmp |= (pThis->uSegNo & 0x0F);
		if(pThis->bEOC) 
			uTmp |= 0x10;
	}
	buf.ptr[28] = uTmp;
	
	/* 5th Minipkt byte */
	buf.ptr[29]  = ((pThis->uAttn << 4) & 0xF0);
	buf.ptr[29] |= pThis->uPSID & 0x0F;
	
	/* 6th Minipkt byte depends on glopping */ 
	if(pThis->bGlop)
		buf.ptr[30] = pThis->uCycleSec;
	else
		buf.ptr[30] = pThis->uMixer;
	
	/* 7th Minipkt byte */
	
	buf.ptr[31] = pThis->uFmt & 0x3F;
	
	switch(pThis->uMSFlen)
	{
		case 0: break;
		case 2: buf.ptr[31] |= 0x40; break;
		case 4: buf.ptr[31] |= 0x80; break;
		case 8: buf.ptr[31] |= 0xC0; break;
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "uMSFlen has an invalid "
			             "value: %hhu bytes", pThis->uMSFlen);
	}
	
	/* Pack MSF and Glop Attn bytes */
	iPos = 32;
	
	for(i=0; i < 8; i++){
		if(i < pThis->uMSFlen){
			buf.ptr[iPos] = pThis->MSFary[i];
			iPos++;
		}
		else
			break;
	}
	
	for(i=0; i < 15; i++){
		if(i < pThis->uCycles){
			buf.ptr[iPos] = pThis->GlopAttnAry[i];
			iPos++;
		}
		else
			break;
	}
	
	buf.ptr[55] = pThis->uFlags;
	
	if(pOffset != NULL) *pOffset = 56;
	return buf;
}

/****************************************************************************/
byte JWavesSDSblk_decode(struct gse_block* pBase, cbytebuf_t buf)
{
	int i, iPos;
	JWavesSDSblk* pThis = (JWavesSDSblk*)pBase;
	
	if(GseBlock_decode(pBase, buf) != 2)
		gse_throw(GSE_XCEPT, GSE_XLIBINCOMPAT);
	
	if(pBase->uType != JW_SDS_BLK_TYPE_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x%hhX for the block type "
		             "ID, but recieved 0x%hhX", JW_SDS_BLK_TYPE_ID, pBase->uType);
	
	if(pBase->uLen != 55)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 55 (i.e. 56 - 1) for the "
		             "SDS block length, but recieved %hhu", pBase->uLen);
	
	/* Decoding, idx 0 and 1 handled by base class */
	
	/* TODO: Alot of range checking could be added here */
	pThis->uModStep   = buf.ptr[2];
	pThis->uTsLo      = (buf.ptr[3] >> 4)&0x03;
	pThis->uTsHi      = buf.ptr[3] & 0x03;
	pThis->uColSCLK   = gse_u4FromMSBbytes(buf.ptr + 4);
	pThis->uColRTI    = buf.ptr[8];
	pThis->uAvg       = buf.ptr[9];
	pThis->uSeqNoHi   = gse_u2FromMSBbytes(buf.ptr + 10);
	pThis->uRepSclkHi = gse_u4FromMSBbytes(buf.ptr + 12);
	pThis->uIDP_CRC   = gse_u2FromLSBbytes(buf.ptr + 16);
	pThis->uSeqNoLo   = gse_u2FromMSBbytes(buf.ptr + 18);
	pThis->uRepSclkLo = gse_u4FromMSBbytes(buf.ptr + 20);
	
	/* Minipkt header bytes */
	pThis->uPSID      = (buf.ptr[24] & 0xF0) | (buf.ptr[29] & 0x0F);
	pThis->uMPlen     = gse_u2FromMSBbytes(buf.ptr + 24) & 0x0FFF;
	pThis->uMPrti     = gse_u2FromMSBbytes(buf.ptr + 26);
	
	if(buf.ptr[28] & 0x80)
		pThis->bGlop = true;
	else
		pThis->bGlop = false;
	
	pThis->uAttn = ((buf.ptr[28] >> 1)&0x30) | ((buf.ptr[29] >> 4)&0x0F);
	
	if(buf.ptr[28] & 0x10)
		pThis->bEOC = true;
	else
		pThis->bEOC = false;
	
	if(pThis->bGlop){
		pThis->uCycles   = buf.ptr[28] & 0x0F;
		pThis->uCycleSec = buf.ptr[30];
	}
	else{
		pThis->uSegNo  = buf.ptr[28] & 0x0F;
		pThis->uMixer  = buf.ptr[30];
	}
	
	switch((buf.ptr[31] >> 6) & 0x03){
		case 0: pThis->uMSFlen = 0; break;
		case 1: pThis->uMSFlen = 2; break;
		case 2: pThis->uMSFlen = 4; break;
		case 3: pThis->uMSFlen = 8; break;
	}
	
	pThis->uFmt = buf.ptr[31] & 0x3F;
	
	iPos = 32;
	for(i = 0; i<pThis->uMSFlen; i++){
		pThis->MSFary[i] = buf.ptr[iPos];
		iPos++;
	}
	
	for(i = 0; i<pThis->uCycles; i++){
		pThis->GlopAttnAry[i] = buf.ptr[iPos];
		iPos++;
	}
	
	pThis->uFlags = buf.ptr[55];
	
	jwaves_valid_idsrc(pThis->uPSID, JWAVES_LOG);
	return 56;
}

GseBlock* JWavesSDSblk_copy(GseBlock* pBase, size_t uOverMalloc)
{
	JWavesSDSblk* pNew;
	JWavesSDSblk* pThis = (JWavesSDSblk*)pBase;	
	size_t uSize = 0;
	int i;
	 
 	if(uOverMalloc > sizeof(JWavesSDSblk))
		uSize = uOverMalloc;
	else
		uSize = sizeof(JWavesSDSblk);
	
	pNew  = (JWavesSDSblk*)GseBlock_copy(pBase, uSize);
	
	pNew->uModStep   = pThis->uModStep;
	pNew->uTsLo      = pThis->uTsLo;
	pNew->uTsHi      = pThis->uTsHi;
	pNew->uColSCLK   = pThis->uColSCLK;
	pNew->uColRTI    = pThis->uColRTI;
	pNew->uAvg       = pThis->uAvg;
	pNew->uSeqNoHi   = pThis->uSeqNoHi;
	pNew->uRepSclkHi = pThis->uRepSclkHi;
	pNew->uIDP_CRC   = pThis->uIDP_CRC;
	pNew->uSeqNoLo   = pThis->uSeqNoLo;
	pNew->uRepSclkLo = pThis->uRepSclkLo;
	
	pNew->uPSID     = pThis->uPSID;
	pNew->uMPlen    = pThis->uMPlen;
	pNew->uMPrti    = pThis->uMPrti;
	pNew->bGlop     = pThis->bGlop;
	pNew->uAttn     = pThis->uAttn;
	pNew->bEOC      = pThis->bEOC;
	pNew->uSegNo    = pThis->uSegNo;
	pNew->uCycles   = pThis->uCycles;
	pNew->uMixer    = pThis->uMixer;
	pNew->uCycleSec = pThis->uCycleSec;
	pNew->uMSFlen   = pThis->uMSFlen;
	pNew->uFmt      = pThis->uFmt;
	
	for(i=0; i<8; i++)  pNew->MSFary[i]      = pThis->MSFary[i];
	for(i=0; i<15; i++) pNew->GlopAttnAry[i] = pThis->GlopAttnAry[i];
	
	pNew->uFlags = pThis->uFlags;
	
	pNew->isSegmented  = pThis->isSegmented;
	pNew->isGlopped    = pThis->isGlopped;
	pNew->isCompressed = pThis->isCompressed;
	pNew->setMPlen     = pThis->setMPlen;
	pNew->del = pThis->del;
	
	return (GseBlock*)pNew;
}

JWavesSDSblk* new_JWavesSDSblk(cbytebuf_t buf){
	GseBlock* pBase = NULL;
	JWavesSDSblk* pThis = NULL;
	
	pBase = new_GseBlock(CBBUF_NULL, sizeof(JWavesSDSblk));
	pThis = (JWavesSDSblk*)pBase;
	
	/* Override base class items */
	/* No new dynalloc stuff, so just use base destructor */
	pBase->encode = JWavesSDSblk_encode;
	pBase->decode = JWavesSDSblk_decode;
	pBase->_sName = "Juno Waves Science Data Status";
	pBase->uType  = JW_SDS_BLK_TYPE_ID;
	pBase->copy   = JWavesSDSblk_copy;
	
	pBase->_uMinLen = 56;
	
	/* New items */
	pThis->del          = delete_GseBlock;
	pThis->isSegmented  = JWavesSDSblk_isSegmented;
	pThis->isGlopped    = JWavesSDSblk_isGlopped;
	pThis->isCompressed = JWavesSDSblk_isCompressed;
	pThis->collectUTC   = JWavesSDSblk_collectUTC;
	pThis->setMPlen     = JWavesSDSblk_setMPlen;
	
	if(buf.ptr == NULL){
		pThis->uAvg = 1;
		/* Rest of the defaults are all zero, so calloc gets this */ 
		return pThis;
	}
		
	except_cleanup_push(delete_GseBlock, pThis);
	
	JWavesSDSblk_decode(pBase, buf);

	except_cleanup_pop(except_no_call);
	return pThis;	  
}

