/** @file JWavesSDSblk.h */
#ifndef _JWavesSDSblk_H_
#define _JWavesSDSblk_H_

#include <sys/time.h>
#include <time.h>

#include <rpwgse/GseBlock.h>
#include <jwaves/JWavesDefs.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_jwaves_pkt */
/*@{*/

/** Interface for Juno Waves Science Data Blocks.
 * See @ref page_spec_4_3_3_1 for more information on these fields.
 *
 * @class JWavesSDSblk
 * @extends GseBlock
 */
typedef struct jwaves_sds_blk{
	
	/** Base structure */
	GseBlock base;
	
	/** FIELD: The ID of the last procssing step to modify this packet */
	byte uModStep;

	/** FIELD: The lowest numeric Telemetry source of the waves IDP packets that
	 * originally contained this data */
	jwaves_tlm_src uTsLo;
	
	/** FIELD: The highest numeric Telemetry source of the waves IDP packets that
	 * originally contained this data */
	jwaves_tlm_src uTsHi;
	
	/** FIELD: The sample collection start time seconds field */
	uint4 uColSCLK;
	
	/** FIELD: The sample collection start time RTI field */
	byte uColRTI;
	
	/** FIELD: The number of original data captures averaged to arrive at this
	  * capture.*/
	byte uAvg;
	
	/** FIELD: The greatest sequence number of the waves IDP packets that originally
	 * contained this data. */
	uint2 uSeqNoHi;
	
	/** FIELD: The highest SCLK from the Waves IDP packets that originally
	 * contained this data.  Note that this SCLK represents the 
	 * report-out-of-the-instrument time of the data NOT the collection time. */
	uint4 uRepSclkHi;
	
	/** FIELD: The CRC for the original waves IDP packet containing this data */
	uint2 uIDP_CRC;

	/** FIELD: The least sequence number of the waves IDP packets that
	 * originally contained this data. */
	uint2 uSeqNoLo;
	
	/** FIELD: The lowest SCLK from the Waves IDP packets that originally contained 
	 * this data.  Note that this SCLK represents the 
	 * report-out-of-the-instrument time of the data NOT the collection time. */
	uint4 uRepSclkLo;

	/** FIELD: The packet and source ID.
	 * 
	 * This determines the type of data, either waveform or spectrum, as well
	 * as the time between samples for waveforms and the frequency table for
	 * spectrums.@n
	 * The upper for bits are taken from JWavesMP::uID and the lower bits are
	 * from JWavesMP::uSrc.
	 * @n
	 * Valid Values are:
	 * - 0x47 HFR Spectrum, from the HFR Log Detector.
	 * - 0x46 HFR Spectrum, from the HFWBR Log Detector.
	 * - 0x5F HFR Baseband Spectrum, from the HFR I channel, through FFT engine. 
	 * - 0x5B HFR Baseband Spectrum, from the HFWBR I channel, through FFT engine.
	 * - 0x7B HFWBR Baseband Waveform, from the HFWBR I channel.
	 * - 0x7F HFWBR Baseband Waveform, from the HFR I channel.
	 * - 0x88 HFWBR I or Q Waveform, from the HFWBR I channel.
	 * - 0x89 HFWBR I or Q Waveform, from the HFWBR Q channel.
	 * - 0x8C HFWBR I or Q Waveform, from the HFR I channel.
	 * - 0x8D HFWBR I or Q Waveform, from the HFR Q channel.
	 * - 0x91 LFR Spectrum, from the B channel, through the FFT engine.
	 * - 0x92 LFR Spectrum, from the Lo E channel, through the FFT engine.
	 * - 0x93 LFR Spectrum, from the HI E channel, through the FFT engine.
	 * - 0xA1 LFR Waveform, from the B channel
	 * - 0xA2 LFR Waveform, from the Lo E channel
	 * - 0xA3 LFR Waveform, from the HI E channel
	 *
	 * See $(SVNSERVER)/GSE/trunk/ground_proc/doc/packet_types.ods for more info.
	 */
	byte uPSID;
	
	/** FIELD: Original Minipkt Length (use of this field is discouraged)
	 * Use GseHdr::uHeadTailLen and GseHdr::TotalLen instead.
	 */
	uint2 uMPlen;
	
	/** FIELD: Original Minipkt Real time interrupt number. (use discouraged)
	 * This is a combination of the RTI and SCLK, use JWavesSDSblk::uColRTI 
	 * instead.
	 */ 
	uint2 uMPrti;
	
	/** TEMPORARY FIELD: Glopping enabled
	 * If this is @c true, @a uSegNo and uBand are not relevent.
	 */
	bool bGlop;
	
	/** FIELD: Attenuator settings byte.
	 * Here's how it's broken out:
    * <pre>
    *     7       6       5       4       3       2       1       0
    * +-------+-------+-------+-------+-------+-------+-------+-------+
    * |               | PATN  | BGAIN | StpAt | StpAt | StpAt | StpAt |
    * +-------+-------+-------+-------+-------+-------+-------+-------+
    * </pre>
    * PATN - If 1 the preamp attenuator is on.  The attenuation is 25dB for
    *        the Lo E and B LFR channels, and 20dB for LFR Hi E, HFR and 
    *        HFWBR.
    *    
    * BGAIN - If 1 the back end LFR gain amp is on for the Hi and Lo E 
    *         channels which adds 20dB to the signal.
    *    
    * StpAt Number between 0 and 15 which is determines the attenuation
    *        settings for (which modes?) via the following simple relation:
    *  
    *           STPAT * 2 = dB attenuationm
	 */
	byte uAttn;
	
	/** TEMPORARY FIELD: End of collection marker.
	 *
	 * If @c true, this is the last minipacket in a sequence of
	 * minipackets containing the data for a single data collection event.
	 * If @c false expect subsequent minipackets to contain more data for
	 * the collection event.
	 * For unsegmented data, this is always true. 
	 */
	bool bEOC;
	
	/** TEMPORARY FIELD: Segment number.
	 *
	 * This fields increments by one, starting at zero, for each minipacket
	 * that contains extra data for a data collection event.
	 * @note This field in irrelevant if @a bGlop is set, because glopping
	 * is the combinding of multiple small collection events into a single
	 * minipacket where as segmenting is the spliting of a single collection
	 * event over multiple minipackets.
	 */
	byte uSegNo;
	
	/** TEMPORARY FIELD: Cycle Count, the number of collections in this
	 * minipacket - 1.
	 *
	 * @warning Glopped packets have larger headers! @n
	 * There is one extra byte of header information for each cycle after the
	 * first.  Note the extra bytes at the bottom of the minipacket header
	 * diagram in new_JWavesMP()
	 *
	 * This field is 0 if @a bGlop is false.
	 */
	byte uCycles; 
	
	/** FIELD: Mixer selection byte.
	 * The mixer clock runs at 4 times the center frequency thus:
    *  
    *            Mixer / 4.0 = Band Center MHz
    *  
    * As you can see the center frequency is adjustable in .25 MHz steps.
    * A couple of values have special meaning:
	 *
	 * BND = 0x01
	 *      Indicates the HFR is sweeping.
    *  
    * BND = 0xF0:
    *      This indicates base band, the mixer is not affecting the signal
    *      and that noise processing is not being preformed.
    *         
    * BND = 0xF1:
    *      This indicates base band, the mixer is not affecting the signal
    *      and that noise mitigation is being preformed.
	 *
	 * BND = 0xF3:
	 *      This indicates baseband, the mixer is not affecting the signal
	 *      and that all three LFR channels are being transmitted, i.e.
	 *      (Signal + N), N, (Signal + N) - N
	 *         
    * The maximum legal value for BND is 180, which corresponds to 45 MHz. */
	byte uMixer;
	
	
	/** TEMPORARY FIELD: for glopped captures.  Contains the time between each
	 * glopped-on capture.  Set to 0 by a successful deglopping process. */
	byte uCycleSec;
	
	
	/** FIELD: Number of MSF bytes, may be 0, 2, 4, or 8. */
	byte uMSFlen;
	
	/** FIELD: The data format.
	 * The Data section of this packet contains pure samples of this type
	 * @b if the lower 4-bits of the uCont field of the packet header is 
	 * either @c 0x3 or @c 0x4.  When the content field is not one
	 * of these values, instrument and/or ground headers are mixed in with the
	 * samples in the Data section.
	 * @see mod_jwaves_conv for a description of the values. */
	byte uFmt;
	
	/** FIELD: Container for extra bytes that may have been added to the minipacket
	 * headers if the MSF bytes were set. */
	byte MSFary[8];
	
	/** FIELD: Temporary buffer for glopped captures.  
	 * Contains the Attn settings for each glopped on capture. Though the array
	 * is fixed length, useful data is of size @a uGlopCyc.  It is nulled by a
	 * successful deglopping process. 
	 */
	byte GlopAttnAry[15];
	
	/** FIELD: TBD */
	byte uFlags;
	
	/** Is the data only part of a segmented capture? */
	bool (*isSegmented)(const struct jwaves_sds_blk* pThis);
	
	/** Have multiple captures' been glopped together? */
	bool (*isGlopped)(const struct jwaves_sds_blk* pThis);
	
	/** Does this data require decompression? */
	bool (*isCompressed)(const struct jwaves_sds_blk* pThis);
	
	/** Get the data collection time in UTC instead of the spacecraft clock */
	void (*collectUTC)(const struct jwaves_sds_blk* pThis, struct timeval* pTV);
	
	/** Set the MPlen field if possible
	 * If the indicated length is longer than 4096 then the MPlen field is
	 * set to zero */
	void (*setMPlen)(struct jwaves_sds_blk* pThis, uint8 uNewLen);
	
	/** Destructor */
	void (*del)(void* vpThis);
	 
} JWavesSDSblk;

/** Generate a new Juno Waves Science Data Status block. 
 * @public
 * @memberof JWavesSDSblk
 */
JWavesSDSblk* new_JWavesSDSblk(cbytebuf_t buf);


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesSDSblk_H_ */

