#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseWrap.h>

#include "JWavesSciMP.h"

/*****************************************************************************/
/* Constructor, destructor is from the base class */

JWavesSciMP* new_JWavesSciMP(cbytebuf_t bulkBuf, size_t* pOffset)
{	
	size_t offset;
	JWavesSciMP* pThis;
	JWavesMP*    pBase;
	cbytebuf_t cbuf;
	
	size_t nMSF;
	int i;
	
	offset = *pOffset;
	
	if(offset >= bulkBuf.len)
		return NULL;
	
	/* Set the start of my area */
	cbuf.ptr = bulkBuf.ptr + offset;
	cbuf.len = bulkBuf.len - offset;

	
	/* Need at least 8 bytes remaining for a minipkt header */
	if(cbuf.len < 8){
		for(i = 0; i<cbuf.len; i++){
			if(cbuf.ptr[i] != 0){
				gse_warning("Possible data loss: Remaining non-zero bytes after the "
				            "previous minipacket are too few to be a minipkt header"); 
				break;
			}
		}				
		return NULL;
	}
	
	pBase = new_JWavesMP(sizeof(JWavesSciMP));
	pBase->uType = JW_SCI_MINIPACKET;
	
	pThis = (JWavesSciMP*)pBase;
	pThis->del = delete_JWavesMP;
	
	/* Now lets setup my vars */
	except_cleanup_push(pThis->del, pThis);
	
	
	/* Decode the basic stuff: ID, Len, RTI, MSF */
	JWavesMP_readCommon(pBase, cbuf);
	
	/* pStart = cbuf.ptr + offset; */
	
	
	pThis->bEOC  = ( cbuf.ptr[4] & 0x10) ? true: false;
			
	if( cbuf.ptr[4] & 0x80 ){
		pThis->bGlop = true;
		pThis->uCycles = cbuf.ptr[4] & 0x0F;
		pThis->uCycleSec = cbuf.ptr[6];
		if( ! pThis->bEOC )
			gse_warning("Inconsistant minipkt header, Glop bit is set, but end "
			            "of segment sequence bit is not set");
	}
	else{
		pThis->bGlop = false;		
		pThis->uSegNo = cbuf.ptr[4] & 0x0F;
		pThis->uMixer = cbuf.ptr[6];
	}
	
	pThis->uSrc = cbuf.ptr[5] & 0x0F;
	pThis->uFmt = cbuf.ptr[7] & 0x3F;
	
	/* Handle attn, always have at least one of these */
	pThis->bufAttn.ptr = pThis->_attn;
	pThis->bufAttn.len = 1 + pThis->uCycles;
	
	pThis->_attn[0] |= (cbuf.ptr[4] >> 1) & 0x20;  /* Preamp bit */
	pThis->_attn[0] |= (cbuf.ptr[4] >> 1) & 0x10;  /* Back Gain bit */
	pThis->_attn[0] |= (cbuf.ptr[5] >> 4) & 0x0F;  /* Step attn bits */
	
	
	nMSF = pBase->bufMSF.len;
	
	/* Now for all the rest of the attn bytes if glopping */
	if(cbuf.len < 8 + nMSF + pThis->uCycles)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Read buffer is length %d, but at "
		             "least %hhu bytes are needed for the standard header plus "
		             "%hhu MSF bytes and %hhu glopped attenuator setting bytes",
		             cbuf.len - offset, 8 + nMSF + pThis->uCycles, nMSF, 
		             pThis->uCycles);
	
	for(i = 0; i< pThis->uCycles; i++){
		pThis->_attn[1+i] |= (cbuf.ptr[8 + nMSF + i] << 4) & 0x20; /* Preamp */ 
		pThis->_attn[1+i] |= (cbuf.ptr[8 + nMSF + i] << 4) & 0x10; /* Back G */
		pThis->_attn[1+i] |= (cbuf.ptr[8 + nMSF + i] >> 4) & 0x0F; /* Step A */
	}
	
	/* Store all the header bytes, un-interperated */
	pBase->_hdr.len = 8 + nMSF + pThis->uCycles;
	pBase->_hdr.ptr = (byte*)gse_calloc(pBase->_hdr.len, sizeof(byte));
	memcpy(pBase->_hdr.ptr, cbuf.ptr, pBase->_hdr.len);
	
	/* Now for the data */
	pBase->_data.len = pBase->uLen + 1;
	
	if(cbuf.len < (pBase->_hdr.len + pBase->_data.len))
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Read buffer is length %d, but at "
		             "least %d bytes are needed for the header (%d bytes) "
		             "and data (%d bytes)",  cbuf.len,
		             pBase->_hdr.len + pBase->_data.len,
		             pBase->_hdr.len, pBase->_data.len);
	
	pBase->_data.ptr = (byte*)gse_calloc(pBase->_data.len, sizeof(byte));
	memcpy(pBase->_data.ptr, cbuf.ptr + pBase->_hdr.len, pBase->_data.len);
	
	*pOffset = offset + pBase->_hdr.len + pBase->_data.len;
	except_cleanup_pop(except_no_call);
	return pThis;
}
