/** @file JWavesSciMP.h */
#ifndef _JWavesSciMP_H_
#define _JWavesSciMP_H_

#include <jwaves/JWavesMP.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_jwaves_raw */
/*@{*/

/** Contains the header and data buffers for a science data minipacket
 * @n
 * @class JWavesSciMP
 * @extends JWavesMP
 */

typedef struct jwaves_sci_mp{

	/** Base structure containing the common minipacket fields */
	JWavesMP base;
	
	/** FIELD: Packet Source.
	 *
	 * Valid values are: 
	 * - 0x1 LFR B channel @ 50kHz
	 * - 0x2 LFR Lo E channel @ 50kHz
	 * - 0x3 LBR Hi E channel @ 375kHz
	 * - 0x6 HFWBR Log Detector
	 * - 0x7 HFR Log Detector
	 * - 0x8 HFWBR I channel @ 1.3125MHz
	 * - 0x9 HFWBR Q channel @ 1.3125MHz
	 * - 0xB HFWBR I channel @ 7MHz
	 * - 0xC HFR I channel @ 1.3125MHz
	 * - 0xD HFR Q channel @ 1.3125MHz
	 * - 0xF HBR I channel @ 7MHz
	 */
	byte uSrc;
	
		/** FIELD: Glopping enabled
	 * If this is @c true, @a uSegNo and uBand are not relevent.
	 */
	bool bGlop;
	
	/** FIELD: End of collection marker.
	 * If @c true, this is the last minipacket in a sequence of
	 * minipackets containing the data for a single data collection event.
	 * If @c false expect subsequent minipackets to contain more data for
	 * the collection event.
	 * For unsegmented data, this is always true. 
	 */
	bool bEOC;
	
	/** FIELD: Segment number.
	 * This fields increments by one, starting at zero, for each minipacket
	 * that contains extra data for a data collection event.
	 * @note This field in irrelevant if @a bGlop is set, because glopping
	 * is the combinding of multiple small collection events into a single
	 * minipacket where as segmenting is the spliting of a single collection
	 * event over multiple minipackets.
	 */
	byte uSegNo;
	
	/** FIELD: Cycle Count, the number of collections in this minipacket - 1.
	 *
	 * @warning Glopped packets have larger headers! @n
	 * There is one extra byte of header information for each cycle after the
	 * first.  Note the extra bytes at the bottom of the minipacket header
	 * diagram in @ref new_JWavesMP()
	 *
	 * This field is 0 if @a bGlop is false.
	 */
	byte uCycles;
	
	/** FIELD: Time between cycles in seconds - 1.
	 * Time between cycles, expressed as seconds-1 when GLOPping multiple data
	 * sets into a single minipacket.
	 *
	 * This is 0 if @a bGlop is false. */
	byte uCycleSec;
	
	/** FIELD: Mixer setting in MHz for HFR/HFWBR I and Q Waveform Packets.
	 * Value ranges:
	 * - 0 Reciever is in baseband, A/D converter runs at 7MHz
	 * - 1 Reciever is sweeping for log detection (Making HFR spectrum packets)
	 * - 12 to 192: Mixer setting in MHz, must be divisible by 4.
	 *
	 * This is 0 if @a bGlop is false, as waveforms are too big to glob.
	 * @todo: Check on the meaning of the JWavesMP::uBand field with Willy.
	 */
	byte uMixer;
	
	/** FIELD: The data format.
	 * Determines the contents of the the buffer returned by @a getData.
	 @see JWavesFmts.h for a description of values. */
	byte uFmt;
	
	/** FIELD: Gain and Attenuation settings.
	 * This buffer contains the same information as @ref JWavesSDSblk::uAttn.
	 * there is one byte in this buffer for each collection in the minipacket.
	 */
	cbytebuf_t bufAttn;
	
	/* area to store glopped attenuation bytes since there can never be more
	   than 16 */
	byte _attn[16];

	/** Destructor */
	void (*del)(void* vpThis);
} JWavesSciMP;

/** Generate Science Data Minipacket structures from raw bytes.
 * 
 * Definitive documentation for the mini-packet header is in the
 * assembly include file: @n
 * http://ec.physics.uiowa.edu/control/juno_hw/FSW/trunk/kernel/minipkt.inc
 * @n@n
 *
 * @param bulkBuf a buffer containing the minipacket header in the 
 *        minipacket data.  The expected science data minipacket header format
 *        is defined in waves_gs_fmts_stor.txt
 *
 * @param pOffset is used for both input and output.  Byte decoding starts
 *        @c *pOffset bytes into @c cbuf.ptr.  After decoding, 
 *        *pOffset is incremented by the number of bytes read.
 * 
 * @memberof JWavesSciMP
 */
JWavesSciMP* new_JWavesSciMP(cbytebuf_t bulkBuf, size_t* pOffset);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesSciMP_H_ */
