#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>

#include "JWavesDefs.h"
#include "JWavesSfduBlk.h"

/*****************************************************************************/
/* JPL Telecom Time to Unix Timevals */
bool JWavesSfduBlk_ert2tv(const struct jwaves_sfdu_blk* pThis, struct timeval* pTv)
{
	gse_throw(GSE_XCEPT, GSE_XNOTYETIMP);
	return false;
}

bool JWavesSfduBlk_rectm2tv(const struct jwaves_sfdu_blk* pThis, struct timeval* pTv)
{
	gse_throw(GSE_XCEPT, GSE_XNOTYETIMP);
	return false;	
}

/*****************************************************************************/
/* Encode */
bytebuf_t JWavesSfduBlk_encode(GseBlock* pBlk, byte uExtraLen, byte* pOffset)
{
	byte offset;
	JWavesSfduBlk* pThis = (JWavesSfduBlk*)pBlk;
	GseProcBlk*    pProc = (GseProcBlk*)pBlk;
	bytebuf_t buf;
	
	/* Check that our processing step id is correct */
	if(pProc->uStepID != JW_SFDU_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Processing step ID was %hhu"
				       "expected %hhu.", pProc->uStepID, JW_SFDU_PROC_STEP_ID);
	
	gse_overflowUinU(uExtraLen + 99, sizeof(byte));
	
	buf = JWavesRdrBlk_encode(pBlk, uExtraLen + 99, &offset);
	gse_assert(buf.len >= offset + 99);
	
	buf.ptr[offset]      = pThis->uVer;
	buf.ptr[offset + 1]  = pThis->bTrunc ? 1 : 0;
	memcpy(buf.ptr + offset + 2, pThis->sSrc, 64);
	buf.ptr[offset + 66] = 0;  /* Insure Null termination */
	gse_u4ToMSBbytes(buf.ptr + offset + 67, pThis->uOffset);
	buf.ptr[offset + 71] = pThis->uFlags;
	gse_u2ToMSBbytes(buf.ptr + offset + 72, pThis->uCraft);
	buf.ptr[offset + 74] = pThis->uDsnID;
	memcpy(buf.ptr + offset + 75, pThis->uERT, 6);
	gse_f4ToMSBbytes(buf.ptr + offset + 81, pThis->fBitRate);
	memcpy(buf.ptr + offset + 85, pThis->uRecTm, 6);
	memcpy(buf.ptr + offset + 91, pThis->sPub, 6);
	gse_u2ToMSBbytes(buf.ptr + offset + 97, pThis->uNonFillLen);
	
	if(pOffset != NULL) *pOffset = offset + 99;
	return buf;
}

/*****************************************************************************/
/* Decode */
byte JWavesSfduBlk_decode(GseBlock* pBlk, cbytebuf_t buf)
{
	GseProcBlk* pProc = (GseProcBlk*)pBlk;
	JWavesSfduBlk* pThis = (JWavesSfduBlk*)pBlk;
	
	byte offset;
	
	offset = JWavesRdrBlk_decode(pBlk, buf);
	
	if(pProc->uStepID != JW_SFDU_PROC_STEP_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x%hhX for the block step "
		             "ID, but recieved 0x%hhX", JW_SFDU_PROC_STEP_ID,
				       pProc->uStepID);
	
	pThis->uVer = buf.ptr[offset];
	if(pThis->uVer != 0x01)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Only Version 1 SFDU Reader blocks"
				       " are readable by this class, but version is 0x%02hhX", 
		             pThis->uVer);
	
	pThis->bTrunc = buf.ptr[offset + 1] == 0x01 ? true : false;
	memcpy(pThis->sSrc, buf.ptr + offset + 2, 64);
	pThis->sSrc[64] = '\0';  /* Insure Null Termination */
	pThis->uOffset = gse_u4FromMSBbytes(buf.ptr + offset + 67);
	
	pThis->uFlags = buf.ptr[offset + 71];
	pThis->uCraft = gse_u2FromMSBbytes(buf.ptr + offset + 72);
	pThis->uDsnID = buf.ptr[offset + 74];
	memcpy(pThis->uERT, buf.ptr + offset + 75, 6);
	pThis->fBitRate = gse_f4FromMSBbytes(buf.ptr + offset + 81);
	memcpy(pThis->uRecTm, buf.ptr + offset + 85, 6);
	memcpy(pThis->sPub, buf.ptr + offset + 91, 6);
	pThis->uNonFillLen = gse_u2FromMSBbytes(buf.ptr + offset + 97);
	
	return offset + 99;
}

/*****************************************************************************/
/* Merging */
bool JWavesSfduBlk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{
	JWavesSfduBlk* pThis = (JWavesSfduBlk*)pBase;	
	JWavesSfduBlk* pSother = (JWavesSfduBlk*)pOther;
	
	if(! JWavesRdrBlk_canMerge(pBase, pOther, uFlags) ) return false;
	
	/* Allow a merge if the source name is the same for both blocks,
	   and the uCraft is the same. */
	if( strcmp(pThis->sSrc, pSother->sSrc ) != 0) return false;
	
	if( pThis->uCraft != pSother->uCraft) return false;
	
	return (pThis->uDsnID == pSother->uDsnID);
}

int JWavesSfduBlk_cmpDsnTime(const byte* pMyTm, const byte* pOthTm)
{
	uint2 uMyDays, uOthDays;
	uint4 uMyMilli, uOthMilli;
	
	uMyDays   = gse_u2FromMSBbytes(pMyTm);
	uOthDays  = gse_u2FromMSBbytes(pOthTm);
	
	uMyMilli  = gse_u4FromMSBbytes(pMyTm + 2);
	uOthMilli = gse_u4FromMSBbytes(pOthTm + 2);
	
	if(uMyDays < uOthDays) return -1;
	if(uMyDays > uOthDays) return 1;
	if(uMyMilli < uOthMilli) return -1;
	if(uMyMilli > uOthMilli) return 1;
	
	return 0;
}


bool JWavesSfduBlk_merge(GseBlock* pBthis, const GseBlock* pBother, uint4 uFlags)
{	
	/* GseProcBlk*    pPthis = (GseProcBlk*)pBthis;*/
	JWavesSfduBlk* pThis  = (JWavesSfduBlk*)pBthis;
	
	/*GseProcBlk*    pPother = (GseProcBlk*)pBother;*/
	JWavesSfduBlk* pOther  = (JWavesSfduBlk*)pBother;
	
	if(! GseProcBlk_merge(pBthis, pBother, uFlags) ) return false;

	/* Take the lowest offset */
	if(pOther->uOffset < pThis->uOffset) pThis->uOffset = pOther->uOffset;
	
	pThis->uFlags |= pOther->uFlags;
	
	/* Take the lowest ERT */
	if( JWavesSfduBlk_cmpDsnTime(pThis->uERT, pOther->uERT) > 0 )
		memcpy(pThis->uERT, pOther->uERT, 6);
	
	/* Take the average bit rate */
	pThis->fBitRate *= 0.5;
	pThis->fBitRate += pOther->fBitRate * 0.5;
	
	/* Take the lowest Record time */
	if( JWavesSfduBlk_cmpDsnTime(pThis->uRecTm, pOther->uRecTm) > 0 )
		memcpy(pThis->uRecTm, pOther->uRecTm, 6);
	
	/* The Pub Field is always zero, just keep mine */
	
	/* Add the NonFillLens */
	pThis->uNonFillLen += pOther->uNonFillLen;
	
	return true;
}

/*****************************************************************************/
/* Copy Constructor */
GseBlock* JWavesSfduBlk_copy(GseBlock* pBase, size_t uOver)
{
	JWavesSfduBlk* pNew = NULL;
	JWavesSfduBlk* pThis = (JWavesSfduBlk*)pBase;
	
	if(uOver < sizeof(JWavesSfduBlk))
		pNew = (JWavesSfduBlk*) JWavesRdrBlk_copy(pBase, sizeof(JWavesSfduBlk));
	else
		pNew = (JWavesSfduBlk*) JWavesRdrBlk_copy(pBase, uOver); 
	
	/* Get functions hooked up */
	pNew->del = pThis->del;
	pNew->ert2tv   = pThis->ert2tv;
 	pNew->rectm2tv = pThis->rectm2tv;
	
	/* Copy the values */
	pNew->uVer     = pThis->uVer;
	pNew->bTrunc   = pThis->bTrunc;
	memcpy(pNew->sSrc, pThis->sSrc, 64);  pNew->sSrc[64] = '\0';
	pNew->uOffset  = pThis->uOffset;
	pNew->uFlags   = pThis->uFlags;
	pNew->uCraft   = pThis->uCraft;
	pNew->uDsnID   = pThis->uDsnID;
	memcpy(pNew->uERT, pThis->uERT, 6);
	pNew->fBitRate = pThis->fBitRate;
	memcpy(pNew->uRecTm, pThis->uRecTm, 6);
	memcpy(pNew->sPub, pThis->sPub, 6);  pNew->sPub[6] = '\0';
	pNew->uNonFillLen = pThis->uNonFillLen;
	
	return (GseBlock*)pNew;
}

/*****************************************************************************/
/* No dynamic items, use base class destructor. */

JWavesSfduBlk* new_JWavesSfduBlk(cbytebuf_t buf, size_t uOver)
{
	GseBlock*     pBase = NULL;
	GseProcBlk*   pProc = NULL;
	JWavesRdrBlk* pRdr  = NULL;
	JWavesSfduBlk* pThis = NULL;
	
	if(uOver < sizeof(JWavesSfduBlk))
		pRdr = new_empty_JWavesRdrBlk(sizeof(JWavesSfduBlk));
	else
		pRdr = new_empty_JWavesRdrBlk(uOver); 
	
	pBase = (GseBlock*)pRdr;
	pBase->encode   = JWavesSfduBlk_encode;
	pBase->decode   = JWavesSfduBlk_decode;
	pBase->copy     = JWavesSfduBlk_copy;
	pBase->canMerge = JWavesSfduBlk_canMerge;
	pBase->merge    = JWavesSfduBlk_merge;
	
	pBase->_sName  = "Waves SFDU Reader";
	pBase->_uMinLen += 99;
	
	pProc = (GseProcBlk*)pRdr;
	pProc->uStepID  = JW_SFDU_PROC_STEP_ID;
	
	pRdr->uSource = 0x03;
	
	pThis = (JWavesSfduBlk*)pRdr;
	pThis->ert2tv   = JWavesSfduBlk_ert2tv;
	pThis->rectm2tv = JWavesSfduBlk_rectm2tv;
	pThis->del      = pRdr->del;               /* Use base class's destructor */
	
	/* Calloc in base makes the rest of this 0, as we want */
	
	if(buf.ptr == NULL) return pThis;
	
	except_cleanup_push(pThis->del, pThis);
	
	JWavesSfduBlk_decode(pBase, buf);

	except_cleanup_pop(except_no_call);
	
	return pThis;
}
