/** @file JWavesSfduBlk.h */
#ifndef _JWavesSfduBlk_H_
#define _JWavesSfduBlk_H_

#include <sys/time.h>

#include <jwaves/JWavesRdrBlk.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_jwaves_pkt */


/** Encoder/Decoder for Waves SFDU reader blocks
 * @class JWavesSfduBlk
 * @extends JWavesRdrBlk
 */

typedef struct jwaves_sfdu_blk{

	/** Base structure */
	JWavesRdrBlk base;

	/** FIELD: VER the version of the SFDU reader header, always 1 for now */
	byte uVer;
	
	/** FIELD: TRUNC true if the source filename has been truncated, false if not */
	byte bTrunc;

	/** FIELD: SOURCE_NAME the source filename, always null terminated */
	char sSrc[65];
	
	/** FIELD: OFFSET the 0 based offset, into the source at which the SFDU 
	    packet was read */
	uint4 uOffset;
	
	/** FIELD: FLAGS This is a bit field of status flags.
	 *
	 *   - PAD bit 0: Since SFDUs must be an even number of bytes, odd lengts
	 *     must be padded.  This bit is 1 if the original SFDU's data CHADO had
	 *     a pad byte at the end.
	 *
	 *   - FILL bit 1: When set to 1 the packet contains fill.  This should 
	 *     always be 0 as Juno is not planning to utilize the fill capability.
	 *
	 *   - RETRANS bit 2: When set to 1 the frame containing the start of the
	 *     packet had the retransmission flag set, otherwise 0.
	 */
	byte uFlags;
	
	/** FIELD: SCFT_ID Spacecraft ID:
	 * 
	 *  Values are:
	 *   - 61: Really from Juno
	 *   - 63: from a simulator
	 */
	uint2 uCraft;
	
	/** FIELD: DSN_SRC An unsigned byte indicating the number of the DSN 
	 *  tracking station which received this data.
	 */
	byte uDsnID;
	
	/** FIELD: EARTH_RECEIVE_TIME The time the packet was received by the
	 * ground station, or simulated ground station.
	 *
	 * See 820-013, 173-Telecomm-Time to decode these bytes, or use the
	 * member function tisTime2tv can be used to get this as a standard
	 * unix timeval.
	 */
	byte uERT[6];
	
	/** FIELD: BIT_RATE The bit rate measured by the DIT (JPL acronym, no idea)
	 *
	 * The units of this field are unknown
	 */
	float4 fBitRate;
	
	/** FIELD: TIS_RECORD_TIME The system clock time at which the TIS system
	 * generated this packet.
	 *
	 * See 820-013, 173-Telecomm-Time to decode these bytes, or use the
	 * member function tisTime2tv can be used to get this as a standard
	 * unix timeval.
	 */
	byte uRecTm[6];
	
	/** FIELD: TIS_PUB An ASCII string (optionally) entered into the TIS
	 * interface by user control.
	 *
	 *  This field is handy for recording test numbers in ATLO, but LM doesn't
	 *  actually make use of it.
	 */
	char sPub[7];
	
	/** FIELD: NON_FILL_LEN The number of bytes of valid data in the Data CHADO
	 * which contained the start of this packet.
	 *
	 */
	uint2 uNonFillLen;
	
	/** Output ERT Telecomm-Time bytes as unix time values. */
	bool (*ert2tv)(const struct jwaves_sfdu_blk* pThis, struct timeval* pTv);
	
	
	/** Output TIS Record time bytes as unix time values. */
	bool (*rectm2tv)(const struct jwaves_sfdu_blk* pThis, struct timeval* pTv);

	
	/** Destructor */
	void (*del)(void* vpThis);
} JWavesSfduBlk;

/** Generate a new Juno SFDU reader Processing Block
 * @public
 * @memberof JWavesSfduBlk
 */
JWavesSfduBlk* new_JWavesSfduBlk(cbytebuf_t buf, size_t uOver);


#ifdef __cplusplus
}
#endif

#endif /* _JWavesSfduBlk_H_ */
