/** @file JWavesSideBlk.h */

#ifndef _JWavesSideBlk_H_
#define _JWavesSideBlk_H_
 
#include <rpwgse/GseProcBlk.h>
#include <jwaves/JWavesDefs.h>

#ifdef __cplusplus
extern "C" {
#endif


/**  @addtogroup mod_jwaves_pkt */
/*@{*/

/** Interface for Gse Sideband recovery Processing Blocks.
 * These blocks are stored on disk in the format:
 * <pre>
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r00-r07|  0x70  |  0x5D  |  0x70  | PROGRAM|           BEGIN_TIME_SEC          |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r08-r15|BEG_TM_US_DIV_100| ERRORS |WARNINGS|PR0C_ID1|PROC_ID2|PROC_ID3|PROC_ID4|
 *        +--------+--------+--------+--------+--------+--------+-----------------+
 * r16-r23|       LO_READ_BEGIN_TIME_SEC      |       HI_READ_BEGIN_TIME_SEC      |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r24-r31|  LO_READ_TM_US  |  HI_READ_TM_US  | FLAGS  | NO_PKTS|    I SeqNo. 1   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r32-r39|    I SeqNo. 2   |    I SeqNo. 3   |    I SeqNo. 4   |    I SeqNo. 5   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r40-r47|    I SeqNo. 6   |    I SeqNo. 7   |    I SeqNo. 8   |    I SeqNo. 9   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r48-r55|    I SeqNo. 10  |    I SeqNo. 11  |    I SeqNo. 12  |    I SeqNo. 13  |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r56-r63|    I SeqNo. 14  |    I SeqNo. 15  |    I SeqNo. 16  |    Q SeqNo. 1   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r64-r71|    Q SeqNo. 2   |    Q SeqNo. 3   |    Q SeqNo. 4   |    Q SeqNo. 5   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r72-r79|    Q SeqNo. 6   |    Q SeqNo. 7   |    Q SeqNo. 8   |    Q SeqNo. 9   |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r80-r87|    Q SeqNo. 10  |    Q SeqNo. 11  |    Q SeqNo. 12  |    Q SeqNo. 13  |
 *        +--------+--------+--------+--------+--------+--------+--------+--------+
 * r88-r95|    Q SeqNo. 14  |    Q SeqNo. 15  |    Q SeqNo. 16  |
 *        +--------+--------+--------+--------+--------+--------+
 *
 * NO_PKTS:
 *    The upper bits are the number of I segments, the lower bits are the number
 *    of Q segments.
 *
 * </pre>
 *
 * @class JWavesSideBlk
 * @extends GseProcBlk
 */

typedef struct jwaves_side_blk{

	/** Base structure */
	GseProcBlk base;
	
	/** Up to 4 distinct processing program ID's that have been observed in
	 * other processing blocks.  A processing program ID is only stored here
	 * if it is different from the value of base.uProgID.  This is provided
	 * as a means of capturing information that is about to be lost.  Since
	 * 0x00 is never a legal program ID, the usused values are zero filled. 
	 */
	byte uObsProgIDs[4];
	
	/** The earliest time of reading for data contained in this packet.  This
	 * is a 32 bit unsigned MSB first integer.  It is the number of seconds
	 * elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time
	 * (UTC). 
	 * Taken from the useg block.
	 */
	uint4 uLoReadSec;
	
	/** The earliest time of reading for data contained in this packet. This is
	 * an unsigned MSB first 16-bit twos compliment integer containing 1/100th 
	 * the number of microseconds since the LO_READ_BEGIN_TIME_SEC.  To get the
	 * earliest read time microseconds, multiply this value by 100. 
	 * Taken from the useg block.
	 */
	uint2 uLoReadUS;
		
	/** The latest time of reading for data contained in this packet.  This is
	 * a 32 bit unsigned MSB first integer.  It is the number of seconds
	 * elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time
	 * (UTC). 
	 * Taken from the useg block.
	 */
	uint4 uHiReadSec;
	
	/** The latest time of reading for data contained in this packet. This is
	 * an unsigned MSB first 16-bit twos compliment integer containing 1/100th
	 * the number of microseconds since the LO_READ_BEGIN_TIME_SEC.  To get the
	 * latest read time microseconds, multiply this value by 100. 
	 * Taken from the useg block.
	 */
	uint2 uHiReadUS;
	
	/** The number of I channel segments combined to generate the complete
	 * capture. 
	 */
	byte uNumISegs;
	
	/** The number of Q channel segments combined to generate the complete
	 * capture. 
	 */
	byte uNumQSegs;
	
	/** Contains the original sequence and telemetry source numbers of the 
	 * I channel IDP packets used to generate this unsegmented data. 
    * These 16 bits contain two fields:
    * 
    *   15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
    * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    * |   OS  |       Big Endian Sequence Number 0 to  16383          |
    * +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    * 
    * The values for TS (Telemetry Source) are:
    * 
    *    0 - Housekeeping
    *    1 - Low Rate Science
    *
    * Up to 16 mini-packets can be combined into a single unsegmented mini-packet.
	 * If mini-packet data  can be up to 4 KB then this provides for 64 KB data
	 * sections.  Note that 0 is a legal value for at most one of the sequence
	 * numbers.  To distinguish between trailing zeros and a real sequence number
	 * consult the uNumSegs member. */
	uint2 uISeqTSary[16];
	
	/** Contains the original sequence and telemetry source numbers of the 
	 * Q channel IDP packets used to generate this unsegmented data.
	 * See JWavesSeqBlk::uISeqTSary for a bit level description.
	 */
	uint2 uQSeqTSary[16];
	
	/** Add unique program IDs to the stored list.
	 * Repeat values are ignored, values that equal my progID are ignored
	 * Returns the new number of non-zero program ids that were stored.
	 */
	int4 (*addProgID)(struct jwaves_side_blk* pThis, byte uProgID);
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} JWavesSideBlk;


/** Generate a new Juno Waves Sideband Recovery Processing Block.
 * @memberof JWavesSideBlk
 * @public
 * @todo Add more error checking to the decoding step 
 */
JWavesSideBlk* new_JWavesSideBlk(cbytebuf_t buf, size_t uOverMalloc);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesSideBlk_H_ */
