#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <math.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseLog.h> 

#include <fftw3.h>

#include "JWavesRdrBlk.h"
#include "JWavesSDSblk.h"
#include "JWavesUsegBlk.h"
#include "JWavesSideBlk.h"

#include "JWavesSideFilter.h"
#include "JWavesExcept.h"
#include "JWavesConv.h"

#define MAGNITUDE(r, i) sqrt((r)*(r) + (i)*(i))

#ifndef M_PI
#define M_PI 3.1415926535897931
#endif

/****************************************************************************
 * Actual Sideband recovery algorithim 
 *
 * WARNING:  If the output is *NOT* in 32 bit floats, change the format 
 *           field in JWavesSideFilter_merge.
 */
bytebuf_t JWavesSideFilter_recover(JWavesSideFilter* pThis, 
      byte uFmtI, cbytebuf_t bufI, byte uFmtQ, cbytebuf_t bufQ)
{
	fftw_complex* pFTin;
	fftw_complex* pFTout;
	fftw_plan pPlan;
	bytebuf_t bufOut;
	struct jwaves_float4_conv convI;
	struct jwaves_float4_conv convQ;
	float8 dNorm, dAvgI, dAvgQ, dShiftI, dShiftQ;
	float4 f4Tmp;
	uint4  u4Tmp;
	int N, Nover2, i, j;
	
	/* Convert the input packet data to floats */
	jwaves_bytes2float4(uFmtI, bufI, 0, &convI);
	jwaves_bytes2float4(uFmtQ, bufQ, 0, &convQ);
	
	gse_assert(convI.len == convQ.len);
	
	/* Setup the plan and wnd are needed */
	
	if(pThis->nFFTlen != convI.len){
		if(pThis->vpPlan != NULL) fftw_destroy_plan( (fftw_plan)pThis->vpPlan);
		if(pThis->vpFTin != NULL) fftw_free( pThis->vpFTin);
		if(pThis->vpFTout != NULL) fftw_free( pThis->vpFTout);
		if(pThis->pWnd != NULL) free(pThis->pWnd);
		
		pThis->nFFTlen = convI.len;
		
		pThis->vpFTin = fftw_malloc(pThis->nFFTlen * sizeof(fftw_complex));	
		pFTin = (fftw_complex*)pThis->vpFTin;
		
		pThis->vpFTout = fftw_malloc(pThis->nFFTlen * sizeof(fftw_complex));	
		pFTout = (fftw_complex*)pThis->vpFTout;
		
		pPlan = fftw_plan_dft_1d(pThis->nFFTlen, (fftw_complex*)pThis->vpFTin,
		                         (fftw_complex*)pThis->vpFTout,
		                         FFTW_FORWARD, FFTW_MEASURE);
		pThis->vpPlan = (void*) pPlan;
		
		pThis->pWnd = (float8*)gse_calloc(convI.len, sizeof(float8));
		
		for(i=0;i<convI.len;i++)
			pThis->pWnd[i] = 1.0 + cos(2.0*M_PI*(i - (convI.len/2.0))/convI.len);
	}
	else{
		pPlan  = (fftw_plan)     pThis->vpPlan;
		pFTin  = (fftw_complex*) pThis->vpFTin;
		pFTout = (fftw_complex*) pThis->vpFTout;
	}
	
	
	/* Load the data I in Reals, Q in imaginary, and do a hanning window, and
	   shift out the DC component */
	dAvgI = 0;
	dAvgQ = 0;
	for(i=0; i<convI.len; i++){
		dAvgI += convI.ptr[i];
		dAvgQ += convQ.ptr[i];
	}
	dShiftI = - dAvgI / convI.len;
	dShiftQ = - dAvgQ / convI.len;
	
	for(i=0; i<convI.len; i++){
		pFTin[i][0] = (convI.ptr[i] + dShiftI )* pThis->pWnd[i];
		pFTin[i][1] = (convQ.ptr[i] + dShiftQ )* pThis->pWnd[i];
	}
	
	/* Do the FFT */
	fftw_execute(pPlan);
	
	/* Copy and reorder the output */
	bufOut.len = pThis->nFFTlen * 4;
	bufOut.ptr = (byte*)gse_calloc(bufOut.len, sizeof(byte));
	
	dNorm = 2.0 / pThis->nFFTlen;
	
	gse_assert(pThis->nFFTlen % 2 == 0);
	
	N = pThis->nFFTlen;
	Nover2 = pThis->nFFTlen / 2.0;
	
	/* Reorder the data as follows (A listing if N = 1024)
	
	   i      j
	----   ----
	
for: i = 0  to  N/2-1
	
	   0    511   j = (N/2 - 1) - i = 511 - i
	   1    510
	   2    509
	 ...    ...
	 
	 509      2
	 510      1
	 511      0   

for i = N/2  to  N-1
	 512   1023   j = (N/2 - 1) + N - i =  1535 - i
	 513   1022
	 514   1021
	 ...    ...
	 
	1021	  510
	1022    511
	1023    512  */
	
	for(i=0; i< Nover2; i++){
		f4Tmp = dNorm * MAGNITUDE(pFTout[i][0], pFTout[i][1]);
		u4Tmp = *( (uint4*) ((void*) &f4Tmp) );
		
		j = Nover2 - 1 - i;
		
		/* Save the data as little endian */
		bufOut.ptr[j*4] = u4Tmp & 0xFF;
		bufOut.ptr[j*4+1] = (u4Tmp >> 8) & 0xFF;
		bufOut.ptr[j*4+2] = (u4Tmp >> 16) & 0xFF;
		bufOut.ptr[j*4+3] = (u4Tmp >> 24) & 0xFF;
	}
	
	for(i=Nover2; i< N; i++){
		f4Tmp = dNorm * MAGNITUDE(pFTout[i][0], pFTout[i][1]);
		u4Tmp = *( (uint4*) ((void*) &f4Tmp) );
		
		j = Nover2 - 1 + N - i;
		
		/* Save the data as little endian */
		bufOut.ptr[j*4] = u4Tmp & 0xFF;
		bufOut.ptr[j*4+1] = (u4Tmp >> 8) & 0xFF;
		bufOut.ptr[j*4+2] = (u4Tmp >> 16) & 0xFF;
		bufOut.ptr[j*4+3] = (u4Tmp >> 24) & 0xFF;
	}
	
	return bufOut;	
}

/****************************************************************************
 * Dumb helpers for the dumb helper */
void JWavesSideFilter_copyNonSegSci(GsePkt* pDest, GsePkt* pSrc)
{
	int i;
	GseBlock* pBlk;
	
	for(i = 0; i<pSrc->pHdr->numBlocks(pSrc->pHdr); i++){
		pBlk = pSrc->pHdr->getBlockAt(pSrc->pHdr, i);
		
		if(pBlk->uType == JW_SDS_BLK_TYPE_ID) continue;
		
		
		if( (pBlk->uType == GSE_PROC_BLK_TYPE) &&
			 ( ((GseProcBlk*)pBlk)->uStepID == JW_USEG_PROC_STEP_ID) )
			continue;
		
		pDest->pHdr->addBlock(pDest->pHdr, pBlk->copy(pBlk, 0));
	}
}

#define JW_USEG_LO_RTI(A) ( ((uint8)A->uLoReadSec)*40 + A->uLoReadUS )
#define JW_USEG_HI_RTI(A) ( ((uint8)A->uHiReadSec)*40 + A->uHiReadUS )
#define JW_RDR_RTI(A) ( ((int8) ((GseProcBlk*) A )->uBegSec)*40 + ((GseProcBlk*)A)->uBegUSecDiv100 )

void JWavesSideFilter_setReadRange(
		JWavesSideBlk* pSide, JWavesRdrBlk* pRdrI, JWavesUsegBlk* pSeqI, 
		JWavesRdrBlk* pRdrQ, JWavesUsegBlk* pSegQ)
{
	
	/* Init values to extreams */
	int i;
	uint8 loRead = 0xFFFFFFFFFFFFFFFF;
	uint8 hiRead = 0;
	
	pSide->uLoReadSec = 0xFFFFFFFF;
	pSide->uLoReadUS  = 0xFFFF;
	pSide->uHiReadSec = 0;
	pSide->uHiReadUS  = 0;
		
	JWavesUsegBlk* pSeg[2]; pSeg[0] = pSeqI; pSeg[1] = pSegQ;
	JWavesRdrBlk*  pRdr[2]; pRdr[0] = pRdrI; pRdr[1] = pRdrQ;
	
	
	for(i=0;i<2;i++){
	
		if(pSeg[i] != NULL){
			if( JW_USEG_LO_RTI(pSeg[i]) < loRead){
				loRead = JW_USEG_LO_RTI(pSeg[i]);
				pSide->uLoReadSec = pSeg[i]->uLoReadSec;
				pSide->uLoReadUS  = pSeg[i]->uLoReadUS;
			}
			
			if( JW_USEG_HI_RTI(pSeg[i]) > hiRead){
				hiRead = JW_USEG_HI_RTI(pSeg[i]);
				pSide->uHiReadSec = pSeg[i]->uHiReadSec;
				pSide->uHiReadUS  = pSeg[i]->uHiReadUS;
			}
		}
	}
	
	if((pSide->uLoReadSec != 0xFFFFFFFF) && (pSide->uHiReadSec != 0))
		return;
	
	for(i=0;i<2;i++){
		
		if(pRdr[i] != NULL){
	
			if( JW_RDR_RTI(pRdr[i]) < loRead ){
				loRead = JW_RDR_RTI(pRdr[i]);
				pSide->uLoReadSec = ((GseProcBlk*)(pRdr[i]))->uBegSec;
				pSide->uLoReadUS = ((GseProcBlk*)(pRdr[i]))->uBegUSecDiv100;
			}
				
			if( JW_RDR_RTI(pRdr[i]) > hiRead ){
				hiRead = JW_RDR_RTI(pRdr[i]);
				pSide->uHiReadSec = ((GseProcBlk*)(pRdr[i]))->uBegSec;
				pSide->uHiReadUS = ((GseProcBlk*)(pRdr[i]))->uBegUSecDiv100;
			}
		}
	}
	
	if((pSide->uLoReadSec == 0xFFFFFFFF) || (pSide->uHiReadSec == 0))
		gse_throw_m(JW_XCEPT, JW_XINVHDRFIELD, "Could not set input time "
		            "boundaries from I and Q packet pair.");
}


/****************************************************************************
 * This is a dumb helper, before calling this check that:
 *
 * 1. The ID's are propely matched.
 * 2. The data lengths are the same.
 * 3. The collect times are the same.
 */
GsePkt* JWavesSideFilter_merge(
		JWavesSideFilter* pThis, GsePkt* pPktI, int4 iPktIdxI,
                               GsePkt* pPktQ, int4 iPktIdxQ
)
{
	/* GsePktFilter* pBase = (GsePktFilter*)pThis; */
	GsePktSrc*    pRoot = (GsePktSrc*)pThis;
	GsePkt* pNew;
	int i;
	GseBlock* pBlk;
	
	JWavesSDSblk* pSciI;
	JWavesSDSblk* pSciQ;
	JWavesSDSblk* pSci;
	
	JWavesRdrBlk* pRdrI;
	JWavesRdrBlk* pRdrQ;
	
	JWavesUsegBlk* pSegI;
	JWavesUsegBlk* pSegQ;
	
	JWavesSideBlk* pSide;
	GseProcBlk*    pSideBase;
	
	bytebuf_t      bufData;
	
	
	pNew = new_empty_GsePkt(0);
	except_cleanup_push(pNew->del, pNew);    /* <-- cleanup */
	pNew->setHdrLimits(pNew, 256, 65535);
	
	pNew->pHdr->nSpaceCraft = pPktI->pHdr->nSpaceCraft;    /* spacecraft ID */
	pNew->pHdr->uContent    = 0x85;          /* Set to complete measurement */
	
	/* Copy all non SDS and useg blocks into this header */
	JWavesSideFilter_copyNonSegSci(pNew, pPktI);
	JWavesSideFilter_copyNonSegSci(pNew, pPktQ);
	
	pSide = new_JWavesSideBlk(CBBUF_NULL, 0);
	except_cleanup_push(pSide->del, pSide);  /* <-- cleanup */
	pSideBase = (GseProcBlk*)pSide;
	pSideBase->uProgID = ((GsePktFilter*)pThis)->uProgID;
	
	/* Fill in the uObsProcIDs before the merge */
	for(i = 0; i<pNew->pHdr->numBlocks(pNew->pHdr); i++){
		pBlk = pNew->pHdr->getBlockAt(pNew->pHdr, i);
		if(pBlk->uType != GSE_PROC_BLK_TYPE) continue;
		
		pSide->addProgID(pSide, ((GseProcBlk*)pBlk)->uProgID);
	}
	
	/* Merge similar blocks, it is up to each header block type to define
	 * it's own merge policy and code. */
	pNew->pHdr->mergeBlocks(pNew->pHdr, 0, 0);
	
	
	/* Now make the new SDS block ****************/
	pSciI = (JWavesSDSblk*)pPktI->getBlkByType(pPktI, JW_SDS_BLK_TYPE_ID, 0);
	pSciQ = (JWavesSDSblk*)pPktQ->getBlkByType(pPktQ, JW_SDS_BLK_TYPE_ID, 0);
	
	pSci = new_JWavesSDSblk(CBBUF_NULL);  /* This uses calloc, so don't have */
   except_cleanup_push(pSci->del, pSci); /* to set items to 0, it's handled */
	
	pSci->uModStep = JW_SBR_PROC_STEP_ID;
	pSci->uTsLo = (pSciI->uTsLo < pSciQ->uTsLo) ? pSciI->uTsLo : pSciQ->uTsLo;
	pSci->uTsHi = (pSciI->uTsHi > pSciQ->uTsHi) ? pSciI->uTsHi : pSciQ->uTsHi;

	pSci->uColSCLK = pSciI->uColSCLK;  /* Caller must verify this */
	pSci->uColRTI  = pSciI->uColRTI;
	
	pSci->uAvg       = (pSciI->uAvg > pSciQ->uAvg)          ? pSciI->uAvg : pSciQ->uAvg;
	pSci->uSeqNoHi   = (pSciI->uSeqNoHi > pSciQ->uSeqNoHi)  ? pSciI->uSeqNoHi : pSciQ->uSeqNoHi;
	pSci->uRepSclkHi = (pSciI->uRepSclkHi>pSciQ->uRepSclkHi)? pSciI->uRepSclkHi : pSciQ->uRepSclkHi;
	
	pSci->uSeqNoLo   = (pSciI->uSeqNoLo < pSciQ->uSeqNoLo)  ? pSciI->uSeqNoLo : pSciQ->uSeqNoLo;
	
	/* Handle the things that just follow around the 1st reported minipkt */
	if(pSciI->uRepSclkLo < pSciQ->uRepSclkLo){
		pSci->uRepSclkLo = pSciI->uRepSclkLo;
		pSci->uIDP_CRC   = pSciI->uIDP_CRC;
		pSci->uMPlen	  = pSciI->uMPlen;
	}
	else{
		pSci->uRepSclkLo = pSciQ->uRepSclkLo;
		pSci->uIDP_CRC   = pSciQ->uIDP_CRC;
		pSci->uMPlen	  = pSciQ->uMPlen;
	}
	
	/* Set new packet types:  0xF1 for HFWBR I & Q, 0xF2 for HFR I & Q */
	if((pSciI->uPSID == 0x88)||(pSciI->uPSID == 0x89)) 
		pSci->uPSID = 0xF1;
	else
		pSci->uPSID = 0xF2;
	
	if(pSciI->uMPrti != pSciQ->uMPrti){
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "RTI field mismatch,\n     I: 0x%04X (no. "
				       "%d, sclk: 0x%08X),\n     Q: 0x%04X (no. %d, sclk: 0x%08X)", 
		             pSciI->uMPrti, iPktIdxI, pSciI->uColSCLK,
				       pSciQ->uMPrti, iPktIdxQ, pSciQ->uColSCLK);
	}
	pSci->uMPrti = pSciI->uMPrti;
	
	gse_assert( (pSciI->bGlop == false)&&(pSciQ->bGlop == false) );
	pSci->bGlop = false;
	
	gse_assert(pSciI->uAttn == pSciQ->uAttn);
	pSci->uAttn = pSciI->uAttn;
	
	gse_assert( pSciI->bEOC && pSciQ->bEOC );
	pSci->bEOC = true;
	
	gse_assert( (pSciI->uSegNo == 0)&&(pSciQ->uSegNo == 0) );
	/* pSci->uSegNo = 0; */
	
	gse_assert( (pSciI->uCycles == 0)&&(pSciQ->uCycles == 0) );
	/* pSci->uCycles = 0; */
	
	gse_assert(pSciI->uMixer == pSciQ->uMixer);
	pSci->uMixer = pSciI->uMixer;
	
	gse_assert((pSciI->uCycleSec == 0)&&(pSciQ->uCycleSec == 0));
	/* pSci->uCycleSec = 0; */
	
	if((pSciI->uMSFlen != 0)||(pSciQ->uMSFlen != 0)){
		gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Handling I/Q sideband recovery "
		            "on packets with the MSF bits set has not been implemented ");
	}
	/* pSci->uMSFlen = 0; */
	pSci->uFmt    = 0x02; /* 32-bit in 32-bits */
	
	pNew->addHdrBlk(pNew, (GseBlock*)pSci);
	except_cleanup_pop(except_no_call);     /* pop pSci cleanup */
	
	/* Finish the Side Band Recovery Block *****************************/
	pSideBase->setNow(pSideBase);
	
	gse_assert( pPktI->pHdr->numBlksByType(pPktI->pHdr,0x70, JW_USEG_PROC_STEP_ID) <= 1);
	gse_assert( pPktQ->pHdr->numBlksByType(pPktQ->pHdr,0x70, JW_USEG_PROC_STEP_ID) <= 1);
	
	
	/* Get the Sequence numbers from the Useg blocks (if present) or the SDS */
	pSegI = NULL;
	if(pPktI->pHdr->numBlksByType(pPktI->pHdr,0x70, JW_USEG_PROC_STEP_ID) > 0){
		pSegI = (JWavesUsegBlk*)pPktI->getBlkByType(pPktI, 0x70, JW_USEG_PROC_STEP_ID);
		
		pSide->uNumISegs = pSegI->uNumSegs ;
		for(i = 0; i<pSegI->uNumSegs; i++)
			pSide->uISeqTSary[i] = pSegI->uSeqTSary[i];		
	}
	else{
		if(pSciI->uSeqNoLo != pSciI->uSeqNoHi)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "I Channel Packet with Collect "
			             "SCLK 0x%x and RTI %d in %s is missing the unsegmentor "
			             "block but was segmented originally.", 
			             pSciI->uColSCLK, pSciI->uColRTI, pRoot->name(pRoot) );
		
		pSide->uNumISegs = 1;					
		pSide->uISeqTSary[0] = pSciI->uSeqNoLo;
	}
	
	pSegQ = NULL;
	if(pPktQ->pHdr->numBlksByType(pPktQ->pHdr,0x70,JW_USEG_PROC_STEP_ID) > 0){
		pSegQ = (JWavesUsegBlk*)pPktQ->getBlkByType(pPktQ,0x70,JW_USEG_PROC_STEP_ID);
	
		pSide->uNumQSegs = pSegQ->uNumSegs;
		for(i = 0; i<pSegQ->uNumSegs; i++)
			pSide->uQSeqTSary[i] = pSegQ->uSeqTSary[i];
		
	}
	else{
		if(pSciQ->uSeqNoLo != pSciQ->uSeqNoHi)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Q Channel Packet with Collect "
			             "SCLK 0x%x and RTI %d in %s is missing the unsegmentor "
			             "block but was segmented originally.", 
			             pSciQ->uColSCLK, pSciQ->uColRTI, pRoot->name(pRoot) );
		
		pSide->uNumQSegs = 1;
		pSide->uQSeqTSary[0] = pSciQ->uSeqNoHi;
	}
	
	/* Add in the lowest and highest read times, from the Rdr or Useg blocks */
	pRdrI = jwaves_get_rdr_blk(pPktI);
	pRdrQ = jwaves_get_rdr_blk(pPktQ);
	
	JWavesSideFilter_setReadRange(pSide, pRdrI, pSegI, pRdrQ, pSegQ);
	
	/* Update our warning count to take into account warnings in the
	   useg blocks */
	if(pSegI != NULL){
		((GseProcBlk*)pSide)->uWarnings += ((GseProcBlk*)pSegI)->uWarnings;
		((GseProcBlk*)pSide)->uErrors += ((GseProcBlk*)pSegI)->uErrors;
	}

	if(pSegQ != NULL){
		((GseProcBlk*)pSide)->uWarnings += ((GseProcBlk*)pSegQ)->uWarnings;
		((GseProcBlk*)pSide)->uErrors += ((GseProcBlk*)pSegQ)->uErrors;
	}
	
	/* Add in the SBR block */
	pNew->addHdrBlk(pNew, (GseBlock*)pSide);
	except_cleanup_pop(except_no_call);     /* pop pSide cleanup */
	
	
	/* Now handle merging the data */
	bufData = JWavesSideFilter_recover(pThis, 
			                            pSciI->uFmt, pPktI->getData(pPktI),
	                                  pSciQ->uFmt, pPktQ->getData(pPktQ));
	
	pNew->setData(pNew, bufData);
	except_cleanup_pop(except_no_call);     /* pop pNew cleanup */
	
	gse_info_v("Narrow band spectra packet created from packets %d (I) and %d "
	           "(Q)", iPktIdxI, iPktIdxQ);
	
	return pNew;
}


/****************************************************************************/
/* Packet matching helpers for readNext */

int JWavesSideFilter_waitCat(GsePkt* pPkt){
	
	JWavesSDSblk* pSDS;
	
	pSDS = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	/* See if it has a compression processing block */
	if(pPkt->pHdr->numBlksByType(pPkt->pHdr, GSE_PROC_BLK_TYPE, JW_DCMP_PROC_STEP_ID) > 0){
		
		switch(pSDS->uPSID){
			case 0x88:  return JW_IQ_CAT_WIR;
			case 0x89:  return JW_IQ_CAT_WQR;
			case 0x8C:  return JW_IQ_CAT_HIR;
			case 0x8D:  return JW_IQ_CAT_HQR;
		}
	}
	else{
		switch(pSDS->uPSID){
			case 0x88:  return JW_IQ_CAT_WI;
			case 0x89:  return JW_IQ_CAT_WQ;
			case 0x8C:  return JW_IQ_CAT_HI;
			case 0x8D:  return JW_IQ_CAT_HQ;
		}
	}
	
	/* gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Packet is not an I or Q waveform "
	             "packet, PSID = 0x%02X", pSDS->uPSID); */
	
	return JW_IQ_WAIT_CATEGORIES;
}

int JWavesSideFilter_mateWaitCat(GsePkt* pPkt){
	
	JWavesSDSblk* pSDS;
	
	pSDS = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	/* See if it has a compression processing block */
	if(pPkt->pHdr->numBlksByType(pPkt->pHdr, GSE_PROC_BLK_TYPE, JW_DCMP_PROC_STEP_ID) > 0){
		
		switch(pSDS->uPSID){
			case 0x88:  return JW_IQ_CAT_WQR;
			case 0x89:  return JW_IQ_CAT_WIR;
			case 0x8C:  return JW_IQ_CAT_HQR;
			case 0x8D:  return JW_IQ_CAT_HIR;
		}
	}
	else{
		switch(pSDS->uPSID){
			case 0x88:  return JW_IQ_CAT_WQ;
			case 0x89:  return JW_IQ_CAT_WI;
			case 0x8C:  return JW_IQ_CAT_HQ;
			case 0x8D:  return JW_IQ_CAT_HI;
		}
	}
	
	gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Packet is not an I or Q waveform "
	             "packet, PSID = 0x%02X", pSDS->uPSID);
	
	return JW_IQ_WAIT_CATEGORIES;
}

bool JWavesSideFilter_getMatch(JWavesSideFilter* pThis, int* iCat, int* iMate)
{
	GsePkt** p = pThis->pWaiting;
	
	if( (p[JW_IQ_CAT_WI] != NULL) && (p[JW_IQ_CAT_WQ] != NULL) ){
		*iCat = JW_IQ_CAT_WI;
		*iMate = JW_IQ_CAT_WQ;
		return true;
	}
	
	if( (p[JW_IQ_CAT_WIR] != NULL) && (p[JW_IQ_CAT_WQR] != NULL) ){
		*iCat = JW_IQ_CAT_WIR;
		*iMate = JW_IQ_CAT_WQR;
		return true;
	}
	
	if( (p[JW_IQ_CAT_HI] != NULL) && (p[JW_IQ_CAT_HQ] != NULL) ){
		*iCat = JW_IQ_CAT_HI;
		*iMate = JW_IQ_CAT_HQ;
		return true;
	}
	
	if( (p[JW_IQ_CAT_HIR] != NULL) && (p[JW_IQ_CAT_HQR] != NULL) ){
		*iCat = JW_IQ_CAT_HIR;
		*iMate = JW_IQ_CAT_HQR;
		return true;
	}
	
	return false;
}


/****************************************************************************/
GsePkt* JWavesSideFilter_readNext(GsePktSrc* pBase)
{
	GsePkt* pPkt;
	GsePkt* pPkt1;
	GsePkt* pPkt2;
	GsePkt* pTmpPkt;
	JWavesSDSblk* pSDS;
	int iCat, idx1, idx2;
	JWavesSDSblk* pSDS1;
	JWavesSDSblk* pSDS2; 
	int8 nTime1, nTime2;
	cbytebuf_t cbbuf1;
	cbytebuf_t cbbuf2;
	
	static const except_id_t catch_spec[] = {{XCEPT_GROUP_ANY, XCEPT_CODE_ANY}};
	except_t* pExcept = NULL;
	
	char* sErrMsg;
	char sBuf[1024];
	sBuf[1023] = '\0';
	bool bContI, bContQ;
	
	GsePktFilter* pGseFltr  = (GsePktFilter*)pBase;
	JWavesSideFilter* pThis = (JWavesSideFilter*)pBase;
	
	
	pPkt = NULL;
	while(true){
		
		/* 1st see if the inputs from a previous merge are waiting to go out */
		if( pThis->pPktIMerged != NULL){
			pPkt = pThis->pPktIMerged;
			pThis->pPktIMerged = NULL;
			break;
		}
		
		if( pThis->pPktQMerged != NULL){
			pPkt = pThis->pPktQMerged;
			pThis->pPktQMerged = NULL;
			break;
		}
		
		/* 2st see if any normal waiting items can hit the door.  Note: the
		   I packets is always in idx1 and the q packet is always in idx2 */
		if( JWavesSideFilter_getMatch(pThis, &idx1, &idx2) ){
			
			
			/* Check the times, if they don't match, emit the packet with the
			   older time value. */
			pPkt1 = pThis->pWaiting[idx1];
			pSDS1 = (JWavesSDSblk*) pPkt1->getBlkByType(pPkt1,JW_SDS_BLK_TYPE_ID,0);
			
			pPkt2 = pThis->pWaiting[idx2];
			pSDS2 = (JWavesSDSblk*) pPkt2->getBlkByType(pPkt2,JW_SDS_BLK_TYPE_ID,0);
			
			nTime1 = ((int8)pSDS1->uColSCLK)*40 + pSDS1->uColRTI;
			nTime2 = ((int8)pSDS2->uColSCLK)*40 + pSDS2->uColRTI;
			
			if(nTime1 != nTime2){
				gse_error("I/Q packet pair time mismatch");
				if(nTime1 < nTime2){
					pPkt = pPkt1;
					pThis->pWaiting[idx1] = NULL;
					pThis->pWaitIdx[idx1] = -1;
				}
				else{
					pPkt = pPkt2;
					pThis->pWaiting[idx2] = NULL;
					pThis->pWaitIdx[idx2] = -1;
				}
				break;
			}
			
			/* Check the lengths, if they don't match get 'em out of here because the
			   times are the same. */
			cbbuf1 = pPkt1->getData(pPkt1);
			cbbuf2 = pPkt2->getData(pPkt2);
			
			if(cbbuf1.len != cbbuf2.len){
				gse_error_v("I/Q packet pair data length mismatch, I len: %zu (no. %zu)"
				            ", Q len: %zu (no. %d).", cbbuf1.len, pThis->pWaitIdx[idx1],
				            cbbuf2.len, pThis->pWaitIdx[idx2]);
				
				gse_assert(pThis->pPktIMerged == NULL);
				pThis->pPktIMerged = pPkt1;
				pThis->pWaiting[idx1] = NULL;
				pThis->pWaitIdx[idx1] = -1;
				
				gse_assert(pThis->pPktQMerged == NULL);
				pThis->pPktIMerged = pPkt2;
				pThis->pWaiting[idx2] = NULL;
				pThis->pWaitIdx[idx2] = -1;
				
				continue;
			}
			
			/* Check the glob bit, if it's on output one of the packets */
			if(pSDS1->bGlop){
				gse_error("Glopped packet encountered");
				pPkt = pPkt1;
				pThis->pWaiting[idx1] = NULL;
				pThis->pWaitIdx[idx1] = -1;
				break;
			}
			
			if(pSDS2->bGlop){
				gse_error("Glopped packet encountered");
				pPkt = pPkt2;
				pThis->pWaiting[idx2] = NULL;
				pThis->pWaitIdx[idx2] = -1;
				break;
			}
			
			
			/* Try to get an I/Q set, if that fails, store the packets as
			   merged as normal, and go around again. */
			if(pGseFltr->pCatcher){
				
				bContI = true;
				bContQ = true;
				except_try_push(catch_spec, 1, &pExcept);
				if(pExcept == NULL){
					pPkt = JWavesSideFilter_merge(
					          pThis, pThis->pWaiting[idx1], pThis->pWaitIdx[idx1],
					          pThis->pWaiting[idx2], pThis->pWaitIdx[idx2]
							);
				}
				else{
					gse_shortprn_except(sBuf, 399, pExcept);
					gse_warning_v("Bad I/Q packet set: %s", sBuf);
					gse_sprint_except(sBuf, 1023, pExcept);
					sErrMsg = NS("Bad I/Q packet set (I member):\n%s\n\n", sBuf);
					bContI = pGseFltr->pCatcher(
							     pGseFltr->pCatchCtx,
					           pThis->pWaiting[idx1]->copy(pThis->pWaiting[idx1], 0),
							     sErrMsg);
					
					if(bContI){
						sErrMsg = NS("Bad I/Q packet set (Q member):\n%s\n\n", sBuf);
						bContQ = pGseFltr->pCatcher(
						            pGseFltr->pCatchCtx,
					               pThis->pWaiting[idx2]->copy(pThis->pWaiting[idx2], 0),
						            sErrMsg);
					}
					
					pPkt = NULL;
				}
				except_try_pop();
				if( (!bContI)||(!bContQ) ){
					gse_error("Error packet handler requested a processing halt");
					return NULL;
				}
			}
			else{
				pPkt = JWavesSideFilter_merge(
				          pThis, pThis->pWaiting[idx1], pThis->pWaitIdx[idx1],
				          pThis->pWaiting[idx2], pThis->pWaitIdx[idx2]
				       );
			}
			
			
			if(pThis->uOpts & JW_FLTR_SBR_KEEP_IQ){
				pThis->pPktIMerged = pThis->pWaiting[idx1];
				pThis->pPktQMerged = pThis->pWaiting[idx2];
			}
			else{
				pThis->pWaiting[idx1]->del(pThis->pWaiting[idx1]);
				pThis->pWaiting[idx2]->del(pThis->pWaiting[idx2]);
			}
			
			pThis->pWaiting[idx1] = NULL;
			pThis->pWaiting[idx2] = NULL;
			
			/* If the merge failed, we don't have a packet and need to loop again*/
			if(pPkt != NULL) break;
		}
		
		/* Okay, then get a packet from the source */
		if((pPkt = pThis->base.pUpSrc->readNext(pThis->base.pUpSrc)) == NULL)
			break;
		
		/* Is this packet at the right processing level? */
		if(pPkt->pHdr->uContent != 0x84)
			break;
		
		pSDS = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		/* If this is not one of the packet types I care about, upgrade it */
		if((iCat = JWavesSideFilter_waitCat(pPkt)) == JW_IQ_WAIT_CATEGORIES){
			pPkt->pHdr->uContent = 0x85;
			break;
		}
		
		/* Hook for debugging */
		/* if(pSDS->uColSCLK > 0x19E7C3D2 && pSDS->uColSCLK < 0x19E7C3E0){
			fprintf(stderr, "Debug Hook Triggered\n");
		}
		*/
		
		/* Someone is in my parking spot, kick them out and store me */
		if(pThis->pWaiting[iCat] != NULL){
			pTmpPkt = pThis->pWaiting[iCat];
			pThis->pWaiting[iCat] = pPkt;
			pPkt = pTmpPkt;
			break;
		}
		
		pThis->pWaiting[iCat] = pPkt;  /* Sweet a parking spot! */
		pThis->pWaitIdx[iCat] = pGseFltr->pUpSrc->pktIdx(pGseFltr->pUpSrc);
	}
	
	return pPkt;
}

/****************************************************************************/
void delete_JWavesSideFilter(void* vpThis)
{	
	JWavesSideFilter* pThis = (JWavesSideFilter*)vpThis;
	int i;	
	
	for(i=0; i<JW_IQ_WAIT_CATEGORIES; i++)
		if(pThis->pWaiting[i] != NULL)
			pThis->pWaiting[i]->del(pThis->pWaiting[i]);
	
	if(pThis->vpPlan != NULL) fftw_destroy_plan( (fftw_plan)pThis->vpPlan);
	if(pThis->vpFTin != NULL) fftw_free( pThis->vpFTin);
	if(pThis->vpFTout != NULL) fftw_free( pThis->vpFTout);
	if(pThis->pWnd != NULL) free(pThis->pWnd);
	
	if(pThis->pPktIMerged != NULL)
		pThis->pPktIMerged->del(pThis->pPktIMerged);
	
	if(pThis->pPktQMerged != NULL)
		pThis->pPktQMerged->del(pThis->pPktQMerged);
		
	free(vpThis);
}

/****************************************************************************/
JWavesSideFilter* new_JWavesSideFilter(GsePktSrc* pSrc, byte uProgID, 
		                                 uint4 uFlags)
{
	JWavesSideFilter* pThis;
	int i;
	
	pThis = (JWavesSideFilter*)new_GsePktFilter(
			    pSrc, 
	          uProgID,
	          sizeof(JWavesSideFilter),
	          JWavesSideFilter_readNext,
	          delete_JWavesSideFilter
	        );
	
	pThis->del = delete_JWavesSideFilter;
	pThis->uOpts = uFlags;
	
	for(i=0;i<JW_IQ_WAIT_CATEGORIES;i++) pThis->pWaitIdx[i] = -1;
	
	/* The rest of the stuff inits to zero since rpwgse always uses calloc */
	 
	return pThis;
}
