#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <time.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
#include <dirent.h>

#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseTime.h>


#include "JWavesDefs.h"
#include "JWavesConv.h"
#include "JWavesMroBlk.h"
#include "JWavesSDSblk.h"
#include "JWavesHrbhBlk.h"
#include "JWavesStore.h"


/*****************************************************************************/
/* Data Group Rules */

bool jwaves_ckGroup(const char* sGroup, bool bThrow)
{
	int i, nLen;
	
	if(sGroup == NULL){
		if(bThrow){
			gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sGroup is NULL");
		}
		else{
			gse_error("Group name is null");
			return false;
		}
	}
	
	nLen = strlen(sGroup);
	if(nLen < 1){
		if(bThrow){
			gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Group name is empty.");
		}
		else{
			gse_error("Group name is empty.");
			return false;
		}
	}
	
	for(i=0; i<nLen; i++){
		
		if(isspace(sGroup[i])){
			if(bThrow){
				gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Spaces not allowed in Juno"
				            " Waves Group Names.");
			}
			else{
				gse_error("Spaces not allowed in Juno Waves Group Names.");
				return false;
			}
		}
			
		if(isupper(sGroup[i])){
			if(bThrow){
				gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Upper case letters not allowed "
				            "in Juno Waves Group Names.");
			}
			else{
				gse_error("Upper case letters not allowed in Juno Waves Group Names.");
				return false;
			}
		}		
		
	}

	return true;
}

/*****************************************************************************/
/* Walking a dir tree */

struct jwaves_latest_state{
	GseStrAry* sufficies;
	
	int4 iRank;
	int8 nTime;
	char* sItem;
};

int jwaves_latest(void* vpState, const char* sPath, const struct stat* pStat)
{
	size_t i, nPath;
	int8 nTime;
	bool bFlag;
	struct jwaves_latest_state* pState;
	const char* pBasename;
	int4 iRank;
	int year = 0, month = 0, day = 0, hour = 0, min = 0, sec = 0;
	
	if( (pState = (struct jwaves_latest_state*)vpState) == NULL)
		return 0;       /* Not interested in if no state info */
	
	if(pStat == NULL)
		return 0;       /* Not interested in un stat-able file */
	
	if(! S_ISREG( pStat->st_mode) ) return 0;   /* want files */
	
	nPath = strlen(sPath);
	bFlag = false;
	for(i = 0; i<pState->sufficies->num; i++){
		if(gse_endswith(sPath, pState->sufficies->pAry[i])){
			bFlag = true;
			iRank = i;
			break;
		}
	}
	if(!bFlag) return 0;       /* Not interested if no suffix matches */
	
	/* Get time fields */	
	pBasename = NULL;
	for(i = nPath - 1; i > 0; i--){
		if((sPath[i] == '/')&&(i < nPath - 1)){
			pBasename = sPath + i + 1;
			break;
		}
	}
	if(pBasename == NULL)
		pBasename = sPath;
	
	if( sscanf(pBasename, "%4d%*c%2d%*c%2dT%2d%*c%2d%*c%2d", &year, &month, &day,
	           &hour, &min, &sec) < 1)
		return 0;
	
	nTime = year*10000000000LL + month*100000000LL + day*1000000 + hour*10000 + 
			  min*100 + sec;

	if((nTime > pState->nTime)||
		((nTime == pState->nTime)&&(iRank > pState->iRank)) )
	{
		if(pState->sItem != NULL) free(pState->sItem);
		pState->sItem = gse_strcopy(sPath);
		pState->nTime = nTime;
		pState->iRank = iRank;
	}

	return 0;
}

/* Break sSufficies at | boundaries */
char* jwaves_latest_file(const char* sGroup, const char* sSufficies,
		                   int nNotFound)
{
	char* sTmp;
	char* sTop;
	struct jwaves_latest_state state;
	
	jwaves_ckGroup(sGroup,JWAVES_THROW);
		
	if(sSufficies == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sSufficies is NULL");
	if(strlen(sSufficies) < 1)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sSufficies is empty");
	
	state.sufficies = gse_strsplit(sSufficies, "|");
	except_cleanup_push(state.sufficies->del, state.sufficies);
	
	if(getenv("WAVES_DATA") != NULL){
		sTop = gse_string("%s/%s", getenv("WAVES_DATA"), sGroup);
	}
	else{
		gse_warning("Environment variable WAVES_DATA is not defined.");
		if(getenv("HOME") == NULL)
			gse_throw_m(GSE_XCEPT, GSE_XMISSINGVAR, 
					      "The HOME Environmet variable is not defined.");
		sTop = gse_string("%s/waves_data/%s", getenv("HOME"), sGroup);
	}
	
	except_cleanup_push(free, sTop);
	
	state.nTime = 0;
	state.sItem = NULL;
	state.iRank = -1;
	
	gse_walk(sTop, jwaves_latest, &state);
	
	if(state.sItem == NULL){
		/* This triggers a bug,   FIND IT!
		if(state.sufficies->num > 1)
			gse_warning_v("Couldn't locate a file under %s with one of the "
			              "sufficies %s (| is a seperater).", sTop, sSufficies);
		else
			gse_warning_v("Couldn't locate a file under %s with the suffix %s.", sTop,
					       sSufficies);
		*/
		if(state.sufficies->num > 1)
			sTmp = gse_string("Couldn't locate a file under %s with one of the "
			                  "sufficies %s (| is a seperater).", sTop, sSufficies);
		else
			sTmp = gse_string("Couldn't locate a file under %s with the suffix %s.",
			                  sTop, sSufficies);
		
		if(nNotFound != GSE_XCEPT)
			gse_log(nNotFound, __FILE__, __LINE__, sTmp);
		else
			gse_throw_m(GSE_XCEPT, GSE_XMISSINGFILE, sTmp);
		free(sTmp);
	}
	
	except_cleanup_pop(except_call); /* sTop */
	except_cleanup_pop(except_call); /* sufficicies */
	return state.sItem;
}

bool jwaves_latest_file_s(const char* sGroup, const char* sSufficies,
		                    char* sFile, size_t nLen, int nNotFound)
{
	char* sRet;
	
	if((sFile == NULL)||(nLen < 2))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sFile NULL, or nLen too short");
	
	if( (sRet = jwaves_latest_file(sGroup, sSufficies, nNotFound)) == NULL)
		return false;
	
	strncpy(sFile, sRet, nLen - 1);
	sFile[nLen - 1] = '\0';
	free(sRet);
	
	return true;
}


char* jwaves_newfile(const char* sGroup, const char* sSuffix,
                     const time_t* pTime)
{
	char* sTop;
	char* sRet;
	struct tm* p;
	time_t t;
	
	jwaves_ckGroup(sGroup, JWAVES_THROW);
	
	if(getenv("WAVES_DATA") != NULL){
		sTop = gse_string("%s/%s", getenv("WAVES_DATA"), sGroup);
	}
	else{
		gse_warning("Environment variable WAVES_DATA is not defined defaulting"
				     " to $HOME/waves_data for the storage directory");
		if(getenv("HOME") == NULL)
			gse_throw_m(GSE_XCEPT, GSE_XMISSINGVAR, 
					      "The HOME Environmet variable is not defined.");
		sTop = gse_string("%s/waves_data/%s", getenv("HOME"), sGroup);
	}
	
	except_cleanup_push(free, sTop);	
	
	if(pTime == NULL){
		t = time(NULL);
		p = gse_gmtime(&t);
	}
	else{
		p = gse_gmtime(pTime);
	}
	
	if(sSuffix != NULL){
		sRet = gse_string("%s/%04d_%02d/%02d/%04d_%02d_%02dT%02d_%02d_%02d%s",
				            sTop, p->tm_year+1900, p->tm_mon+1, p->tm_mday,
		                  p->tm_year+1900, p->tm_mon+1, p->tm_mday, p->tm_hour,
		                  p->tm_min, p->tm_sec, sSuffix);
	}
	else{
		gse_debug("NULL suffix in jwaves_newfile");
		sRet = gse_string("%s/%04d_%02d/%02d/%04d_%02d_%02dT%02d_%02d_%02d",
				            sTop, p->tm_year+1900, p->tm_mon+1, p->tm_mday,
		                  p->tm_year+1900, p->tm_mon+1, p->tm_mday, p->tm_hour,
		                  p->tm_min, p->tm_sec);
	}
	
	except_cleanup_pop(except_call); /* sClean */
	return sRet;
}

/*****************************************************************************/

/* WARNING: DO NOT USE jwaves_pkt2sclk HERE!  The time it reports is a MP
   time, but we  DON'T want that *.raw packet storage !! */

char* jwaves_packet_dest(void* pCtx, const GsePkt* pPkt)
{
	/* Packets are sent to hourly files as follows:
	   
		  HSK Packets -> Causes an exception
		  
		  LRS Packets -> YYYY_MM_DDTHH.lrs.SUFFIX
		  
		  MRS Packets -> YYYY_MM_DDTHH.lrs.SUFFIX
		  
	     HRS Packets w bin header: 
		     If have Q-Factor: YYYY_MM_DDTHH.hrs_bin.SUFFIX
			  Without a Q-Fact: YYYY_MM_DDTHH.hrs_rec.SUFFIX
			  
		  HRS Packets wo bin header:
		     YYYY_MM_DDTHH.hrs_rec.SUFFIX
	*/
	jwaves_dest_ctx_t* pDest = (jwaves_dest_ctx_t*)pCtx;
	
	const char* sMid = NULL;
	const JWavesMroBlk* pMro = NULL;
	const JWavesSDSblk* pSds = NULL;
	const JWavesHrbhBlk* pBinHdr = NULL;
	bool bHrsData = false;
	struct timeval tv;
	int i;
	uint4 uSclk;
	char* sTime;
	char* sFile;
	char* sDir;
	byte uPSID;
	cbytebuf_t cbuf;
	
	if(pPkt->pHdr->nSpaceCraft != -61)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Can output locations for Juno "
				       "(Spice ID -61) Waves data, spacecraft id was %hd.",
		             pPkt->pHdr->nSpaceCraft);
						 
	/* HSK Packets... */
	if((pPkt->pHdr->uContent & 0xF0) == 0x60){
		cbuf = pPkt->getData(pPkt);
		uSclk = gse_u4FromMSBbytes(cbuf.ptr + 4);
		
		sMid = "hsk";
		goto DEST_READY;
	}
	
	/* Burst Mode Data Packets... */
	if((pPkt->pHdr->uContent & 0xF0) == 0x70){
		cbuf = pPkt->getData(pPkt);
		uSclk = gse_u4FromMSBbytes(cbuf.ptr + 4);
		
		sMid = "bst";
		goto DEST_READY;
	}
	
	/* Error if not MRO or SCI */
	if( ((pPkt->pHdr->uContent & 0xF0) != 0xE0)&&
		 ((pPkt->pHdr->uContent & 0xF0) != 0x80) ){
		
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Unknown data kind: 0x%X",
				       (pPkt->pHdr->uContent >> 4) & 0xF);
	}
	
		
	/* Handle *.raw packets */
	if((pPkt->pHdr->uContent & 0xF) <= 0x1){
		
		/* If this is a raw packet, then get bytes 4-7 of the data section */
		cbuf = pPkt->getData(pPkt);
		uSclk = gse_u4FromMSBbytes(cbuf.ptr + 4);
		
		
		if( (cbuf.ptr[8] & 0xF0 ) == 0xE0 )
			sMid = "lrs";
		else
			/* The packet id is found in bytes 8 & 13 of the data section */
			uPSID = ((cbuf.ptr[8]) & 0xF0) | ((cbuf.ptr[13]) & 0x0F);
			
	}
	
	/* Or they are *.ucal packets */
	else{
		if((pPkt->pHdr->uContent & 0xF0) == 0xE0){
			pMro = (const JWavesMroBlk*) pPkt->getBlkByType(pPkt, JW_MRO_BLK_TYPE_ID, 0);
			uSclk = pMro->uColSCLK;
			sMid = "lrs";
		}
		else{
			pSds = (const JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
			uSclk = pSds->uColSCLK;
			uPSID = pSds->uPSID;
		}
	}
	
	
	/* If sMid is still not known, this must be a science data packet, not
	   an MRO */
	if(sMid == NULL){
				
		i = 0;
		while(g_hrs_data[i] != 0){
			if(uPSID == g_hrs_data[i]){
				bHrsData = true;
				break;
			}
			i++;
		}
		
		if( ! bHrsData ){
			sMid = "lrs";
		}
		else{
			if(pPkt->blkCount(pPkt, JW_HRS_BIN_HDR_ID, 0) == 0){
				sMid = "hrs_rec";
			}
			else{
				pBinHdr = (const JWavesHrbhBlk*) 
						pPkt->getBlkByTypeIdx(pPkt, JW_HRS_BIN_HDR_ID, 0, 0);
				
				if(pBinHdr->uQFsec == 0 && pBinHdr->uQFsubSec == 0)
					sMid = "hrs_rec";
				else
					sMid = "hrs_bin";
			}
		}
	}

DEST_READY:
	
	/* Okay, pick a file name */
	jwaves_sclk2tv((double)uSclk, &tv);
	
	sTime = gse_tv2iso_gm(&tv,GSE_HOURS);
	sTime[4] = '_';
	sTime[7] = '_';
	
	sFile = NS("%s.%s%s", sTime, sMid, pDest->sExt);
	sTime[7] = '\0';
	sTime[10] = '\0';
	
	if(getenv("WAVES_DATA") == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XMISSINGVAR, "The WAVES_DATA Environmet "
		            "variable is not defined.");
	
	sDir = NS("%s/%s/rpwgse/%s/%s/%s", getenv("WAVES_DATA"), pDest->sGroup,
			    sTime, sTime+8, sFile);
	free(sTime);
	free(sFile);
	return sDir;
}













