/** @file JWavesStore.h */
#ifndef _JWavesStore_H_
#define _JWavesStore_H_

#include <stdbool.h>
#include <rpwgse/GsePkt.h>

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_junogse */
/*@{*/

/** @defgroup mod_jwaves_store Storage 
 * Convienance functions to impliment the Juno Waves Pre-PDS storage hierachy.
 * <pre>
 * 3. STORAGE HIERARCHY
 * ====================
 * The following storage hierarchy will be used for automatic file placement 
 * underneath some user-specified top level directory.
 * 
 * Note: Times indicated below are in UTC.
 * 
 * Top
 *  |
 *  +--YYYY_MM                              Year and month directory, may be
 *     |                                    derived from workstation time, or
 *     |                                    once it is reliable, from SCLK fields
 *     |
 *     +--DAY_OF_MONTH                      Individual data files, at most
 *        |                                 one file is generated per data type
 *        |                                 for each run of the waves_sce_reader.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.cmd.raw    Raw command bytes as transmitted to
 *        |                                 the SCE bin_cmd block.          
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.strm   Unaltered bytes as received from the
 *        |                                 Lockheed Martin spacecraft emulator's
 *        |                                 low speed interface.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hrs.strm   Unaltered bytes as received from the
 *        |                                 Lockheed Martin spacecraft emulator's
 *        |                                 high speed interface.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.raw    Raw Science data packets primarily
 *        |                                 originating with the instrument's low
 *        |                                 rate science interface.  These may
 *        |                                 be collected from a variety of 
 *        |                                 interfaces.  Data section is raw 
 *        |                                 Waves IDP data.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hrs.raw    Raw Science data packets primarily
 *        |                                 originating with the instrument's high
 *        |                                 rate science interface.  These may
 *        |                                 be collected from a variety of 
 *        |                                 interfaces.  Data section is raw 
 *        |                                 Waves IDP data.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hsk.raw    Raw Housekeeping data packets.  Data
 *        |                                 is raw Waves IDP data.  These originate
 *        |                                 from the instrument's low rate science
 *        |                                 interface, but are seperated out from
 *        |                                 the science date.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.XX.raw     For packets that are not known to 
 *        |                                 be science or hsk, where XX is some
 *        |                                 identifier, probably the CIP type
 *        |                                 integer if available.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.upk    Unpacked Low Rate Science data.  There
 *        |                                 exists one minipkt's data in each GSE
 *        |                                 packet.  This may be only part of a
 *        |                                 data set (capture).  Data may be
 *        |                                 segmented, compressed and or glopped.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hrs.upk    Unpacked High Rate Science data.  
 *        |                                 Data is in the same format as LRS.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.useg   Unsegmented Low Rate Science data. 
 *        |                                 There exists one or more complete data
 *        |                                 sets (captures) in each GSE packet.
 *        |                                 This data may be compressed.  If there
 *        |                                 is more than one capture in a GSE
 *        |                                 packet the data is glopped.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hrs.useg   Unsegmented Low Rate Science data. 
 *        |                                 Data is in the same format as LRS.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.dcmp   Decompressed Low Rate Science data.
 *        |                                 There exists one or more complete data
 *        |                                 sets (captures) in each GSE packet.  If 
 *        |                                 there is more than one capture in a
 *        |                                 GSE packet that packet is glopped.
 *        | 
 *        +--YYYY_MM_DDThh_mm_ss.hrs.dcmp   Decompressed High Rate Science data.
 *        |                                 Data is in the same format as LRS.
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.lrs.ucal   Uncalibrated Low Rate Science data.
 *        |                                 There is one and only one data set
 *        |                                 (capture) per GSE packet.  Data is not 
 *        |                                 glopped, compressed, segmented, or
 *        |                                 packed
 *        |
 *        +--YYYY_MM_DDThh_mm_ss.hrs.ucal   Uncalibrated High Rate Science data.
 *                                          Data is in the same format as LRS.
 * 
 * 3.1 Times Used for File and Directory Names
 * -------------------------------------------
 * For all data in the STREAM, RAW, and UCAL trees, times are taken form the
 * workstation time of the original data collection and files cover an arbitrary
 * duration.  
 * 
 * 3.2 Processing Chain
 * --------------------
 * The storage hierarchy above is listed in dependency order not alphabetical order,
 * the basic processing steps, in order, are:
 * 
 *   1. Command data is sent to waves and optionally copied under CMD.
 * 
 *   2. Data is read from some source and packetized.  The packetized data is placed
 *      under RAW.  Optionally a copy of the input is placed under STREAM.
 * 
 *   3. Data is read from RAW and after running through a series of programs goes
 *      under UCAL.  
 * 
 *   4. Calibration is applied to produce PDS compliant files which go under PDS.
 * </pre>
 */
/*@{*/

/** Get the complete name of a file for depositing data.
 *
 * The path is determined by the current UTC time, and the environment
 * variable WAVES_DATA.  The name is determined by the current UTC time,
 * and the given suffix.  The suffix must contian a dot if that is desired.
 *
 * The gse_openpath() function can be used to generate all the directories 
 * needed before opening the file for writing.
 * An example:
 * @code
 * 
 * FILE* pFile = NULL;
 * char* sFile = jwaves_newfile(".sci.raw");
 * 
 * gse_notice("Outputting data to %s.", sFile);
 *
 * //Generate needed intermediate directories, group writable
 * gse_mkdirsto(sFile, GSE_MODE_770);
 *
 * //Open the file
 * pFile = gse_fopen(sFile, "wb");
 *
 * @endcode
 *
 * @param sGroup The data group to for this new file.  Data groups are used
 *        to keep data from different instruments and from overlapping
 *        SCLKs of the same instrument separated.
 *
 * @param sSuffix Text to append to the base filename and path.
 *
 * @param pTime The epoch time to use for determining the path, if NULL the
 *        the current time will be used.
 *
 * @return A new string allocated on the heap.
 */
char* jwaves_newfile(const char* sGroup, const char* sSuffix,
                     const time_t* pTime);


/** Get the name of the most recent file by its suffix.
 *
 * @par Environment
 * The environment variable @c WAVES_DATA defines the top level directory for
 * file lookup.  If @c WAVES_DATA is not defined, @c $HOME/waves_data is used.
 *
 * Example of getting the newest data at any processing stage:
 * @code
 * 
 * char* sFile;
 * file* pFile;
 * 
 *                          // Increasing priority -->
 * sFile = jwaves_latest_file(".upk|.useg|.dcmp|.ucal", GSE_XCEPT);
 * pFile = gse_fopen(sFile, "rb");
 * 
 * @endcode
 *
 * @param sGroup The data group to for this new file.  Data groups are used
 *        to keep data from different instruments and from overlapping
 *        SCLKs of the same instrument separated.		 
 *
 * @param sSufficies A Juno Waves file type suffix.  See @ref mod_jwaves_store
 *   for a listing of the conventional suffixies.  Multiple sufficies may
 *   be checked by putting pipe characters between each one.  
 *		 
 *	  In the instance that two files indicate the same time, the one with the
 *   <b> right most </b> suffix in @a sSuffix will be returned.
 *
 * @param nNotFound The logging level for messages emitted if a file
 *        can not be found automatically.  This is one of:
 *        - @c GSE_DEBUG
 *        - @c GSE_NOTICE
 *        - @c GSE_WARNING
 *        - @c GSE_ERROR
 *        - @c GSE_CRITICAL
 *       or @c GSE_XCEPT.  If @a nNotFound equals @c GSE_XCEPT, then an
 *       exception is thrown when no file can be found.
 *
 * @return A file path string allocated on the heap, or NULL if no suitable
 *        file could be found.
 */
char* jwaves_latest_file(const char* sGroup, const char* sSufficies, int nNotFound);


/** Get the name of the most recent file by its suffix, stack variable version.
 *
 * @param sGroup The data group to for this new file.  Data groups are used
 *        to keep data from different instruments and from overlapping
 *        SCLKs of the same instrument separated.		 
 *
 * @param sSuffix see @ref jwaves_latest_file()
 * @param sFile a buffer to recieve the filename
 * @param nLen a length of the sFile buffer.  If the filename will not fit
 *        in this length, an exception is thrown.
 * @param sGroup The data group to for this new file.  Data groups are used
 *        to keep data from different instruments and from overlapping
 *        SCLKs of the same instrument separated.
 * @param nNotFound see @ref jwaves_latest_file() for the definition of this
 *        parameter.
 * @return @c true if a file could be found, @c false otherwise.
 */
bool jwaves_latest_file_s(const char* sGroup, const char* sSuffix, char* sFile, size_t nLen,
                          int nNotFound);

/** Check a string to see if it's a legal Juno Waves group name.
 *
 */
bool jwaves_ckGroup(const char* sGroup, bool bThrow);

/** Holds the data group name and a file extension for the jwaves_packet_dest 
 *  function.
 */
typedef struct jwaves_dest_ctx{
	const char* sGroup;
	const char* sExt;
} jwaves_dest_ctx_t;

/** Determining the file destination by packet contents 
 *
 * This can be supplied as the destination callback for GseMultiFileSink
 * objects.
 *
 * @param pCtx a void pointer to a jwaves_dest_ctx_t structure.
 *
 * @param pPkt the GsePkt object which needs a home.
 */
char* jwaves_packet_dest(void* pCtx, const GsePkt* pPkt);
		

/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesStore_H_ */
