#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseLog.h>


#include "JWavesUpkBlk.h"
#include "JWavesUpkFilter.h"

#include "JWavesConv.h"  /* <-- Only added to deal with NRD spectra */

#include "JWavesIDP.h"
#include "JWavesSciMP.h"
#include "JWavesMroMP.h"

#include "JWavesSDSblk.h"
#include "JWavesMroBlk.h"

/*****************************************************************************
 * If this is a long LFR spectrum, update the PSID field as follows:
 *
 *   0x91 -> 0xF0
 *   0x92 -> 0xF4
 *   0x93 -> 0xF5
 */

void jwaves_fixup_nrd_spec_pkt(GsePkt* pPkt)
{
	JWavesSDSblk* pSci = NULL;
	int nBlkBits;
	int nExpected_lo, nExpected_hi, nTrigBig_lo, nTrigBig_hi;
	cbytebuf_t cbuf;
	
	pSci = (JWavesSDSblk*) pPkt->pHdr->getBlkByType(pPkt->pHdr, JW_SDS_BLK_TYPE_ID, 0);
	if((pSci->uPSID != 0x91)&&(pSci->uPSID != 0x92)&&(pSci->uPSID != 0x93))
		return;

	
	nBlkBits = jwaves_dataBlockBits(pSci->uFmt);
	cbuf = pPkt->getData(pPkt);
		
	if(jwaves_isCompressed(pSci->uFmt))
		gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "Cannot determine if LFR spectra "
		            "are Noise Reduction Diagnostic data.  Number of samples is "
				      "unknown because the data has been rice compressed, packet "
	               "SCLK is %s", pSci->uColSCLK);
	
	/* Find out how many bytes we should expect, and how many bytes triggers
	   flagging the packet based on:
		
	   1. The data block size
		2. The number of samples
		3. The need to pad to 4-byte words for each dataset
	*/
	switch(nBlkBits){
	case 8:   
		nExpected_lo = (43)*3;
		nTrigBig_lo  = 43;
		nExpected_hi = (18)*3;
		nTrigBig_hi  = 18;
		break;
		
	case 9:
		nExpected_lo = 52*3;  /* ceil(43 * (9.0/8.0) ) + 3 */
		nTrigBig_lo  = 49;    /* ceil(43 * (9.0/8.0) )     */
		nExpected_hi = 24*3;  /* ceil(18 * (9.0/8.0) ) + 3 */
		nTrigBig_hi  = 21;    /* ceil(18 * (9.0/8.0) )     */
		break;

	case 16:
		nExpected_lo = (43*2)*3;
		nTrigBig_lo  = 86;
		nExpected_hi = (18*2)*3;
		nTrigBig_hi  = 36;
		break;
		
	case 32:
		nExpected_lo = (43*4)*3;
		nTrigBig_lo  = 172;
		nExpected_hi = (18*4)*3;
		nTrigBig_hi  = 76;
		break;
		
	default:
		gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "Handling %d bit items is "
			             "not yet implimented.", nBlkBits);
	}
	
	/* Switch ID's based on the packet length */
	switch(pSci->uPSID){
		case 0x91:
			if(cbuf.len > nTrigBig_lo){
				if(cbuf.len != nExpected_lo)
					gse_error_v("Read %zd bytes for LFR B NRD data, expected "
					            "%d, data not marked as NRD", cbuf.len, nExpected_lo);
				else
					pSci->uPSID = 0xF0;
			}
			break;
			
		case 0x92:
			if(cbuf.len > nTrigBig_lo){
				if(cbuf.len != nExpected_lo)
					gse_error_v("Read %zd bytes for LFR Lo E NRD data, expected "
					            "%d, data not marked as NRD", cbuf.len, nExpected_lo);
				else
					pSci->uPSID = 0xF4;
			}
			break;
						
		case 0x93:
			if(cbuf.len > nTrigBig_hi){ 
				if(cbuf.len != nExpected_hi)
					gse_error_v("Read %zd bytes for LFR Hi E NRD data, expected "
					            "%d, data not marked as NRD", cbuf.len, nExpected_hi);
				else
					pSci->uPSID = 0xF5;
			}
			break;		
	}
	
	
}

/******************************************************************************/
/* Take a single science data minipacket and place it in a stand alone packet */
GsePkt* jwaves_mk_scipkt(byte uProgID, const GsePkt* pOrgPkt,JWavesSciIDP* pIDP, 
                      int iMP, byte uNoMP, bool bConvIDs)
{
	JWavesSciMP* pMP;
	JWavesMP*    pBaseMP;
	GseHdr* pHdr;
	GseHdr* pOrgHdr;
	JWavesSDSblk* pSci;
	GseProcBlk* pUpkBase;
	JWavesUpkBlk* pUpk;
	GseBlock* pBlk;
	int i;

	GsePkt* pPkt = new_empty_GsePkt(0);
	except_cleanup_push(pPkt->del, pPkt);
	pPkt->setHdrLimits(pPkt, 256, 65535);
	
	gse_assert( pIDP->pMPs->pAry[iMP]->uType == JW_SCI_MINIPACKET);
	
	pBaseMP = pIDP->pMPs->pAry[iMP];
	pMP = (JWavesSciMP*)pBaseMP;
	pHdr = pPkt->pHdr;
	pOrgHdr = pOrgPkt->pHdr;
	
	pHdr->nSpaceCraft = pOrgHdr->nSpaceCraft;   /* Preserve the spacecraft ID */
	pHdr->uContent = pOrgHdr->uContent & 0xF0;  /* Preserve the "CIP" */
	
	/* Preserve any existing blocks by coping then from the old header
	   and adding them to the new */
	for(i = 0; i<pOrgHdr->numBlocks(pOrgHdr); i++){
		pBlk = pOrgHdr->getBlockAt(pOrgHdr, i);
		pHdr->addBlock(pHdr, pBlk->copy(pBlk, 0));
	}
	
	/* Make a science data status block ... */
	pSci = new_JWavesSDSblk(CBBUF_NULL);
	
	pSci->uModStep   = JW_UPACK_PROC_STEP_ID;
	pSci->uTsLo      = pIDP->ts;
	pSci->uTsHi      = pIDP->ts;
	
	/* Time for the great clock unmangling, get the data collection time */
	pSci->uColSCLK   = pBaseMP->collectSCLK(pBaseMP, pIDP->uSCLK);
	pSci->uColRTI    = pBaseMP->uRTI % 40;
	pSci->uAvg       = 1;   /** TODO: Should this be a lookup? */
	pSci->uSeqNoHi   = pIDP->uSeq;
	pSci->uRepSclkHi = pIDP->uSCLK;
	
	pSci->uIDP_CRC   = pIDP->uCRC;
	pSci->uSeqNoLo   = pIDP->uSeq;
	pSci->uRepSclkLo = pIDP->uSCLK;
	
	pSci->uPSID     = ((pBaseMP->uID << 4) & 0xF0) | ((pMP->uSrc) & 0x0F);
	
	pSci->uMPlen    = pBaseMP->uLen;
	pSci->uMPrti    = pBaseMP->uRTI;
	pSci->bGlop     = pMP->bGlop;
	pSci->uAttn     = pMP->bufAttn.ptr[0];
	pSci->bEOC      = pMP->bEOC;
	pSci->uSegNo    = pMP->uSegNo;
	pSci->uCycles   = pMP->uCycles;
	pSci->uMixer    = pMP->uMixer;
	pSci->uCycleSec = pMP->uCycleSec;
	pSci->uMSFlen   = pBaseMP->bufMSF.len;
	pSci->uFmt      = pMP->uFmt;
	
	for(i=0; i < pBaseMP->bufMSF.len; i++)
		pSci->MSFary[i] = pBaseMP->bufMSF.ptr[i];
	
	for(i = 0; i < pMP->bufAttn.len - 1; i++)
		pSci->GlopAttnAry[i] = pMP->bufAttn.ptr[i+1];
	
	pSci->uFlags   = 0;
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pSci);
	
	/* Set the content type lower bits */
	if(pSci->isSegmented(pSci) || pSci->isGlopped(pSci) || 
	   pSci->isCompressed(pSci))
		pHdr->uContent = (pHdr->uContent & 0xF0) | 0x03;
	else
		pHdr->uContent = (pHdr->uContent & 0xF0) | 0x04;
	
	
	/* Make an Unpacker process block */
	pUpk = new_JWavesUpkBlk(CBBUF_NULL, 0);
	pUpk->uMiniPkts = uNoMP;
	pUpk->iMiniPkt = iMP;
	
	pUpkBase = (GseProcBlk*)pUpk;
	pUpkBase->setNow(pUpkBase);
	pUpkBase->uProgID = uProgID;
	/* ((GseBlock*)pUpk)->uLen = 15; */ /* remember uLen field is length - 1 */
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pUpk);
	
	/* Set the data */
	pPkt->setData(pPkt, pBaseMP->takeData(pBaseMP));
	
	/* Converting LFR Noise packet ID's to new ones */
	if((pSci->uPSID == 0x91)||(pSci->uPSID == 0x92)||(pSci->uPSID == 0x93))
		jwaves_fixup_nrd_spec_pkt(pPkt);
	
	except_cleanup_pop(except_no_call);
	return pPkt;
}

/******************************************************************************/
/* Take a single MRO data minipacket and place it in a stand alone packet */
GsePkt* jwaves_mk_mropkt(byte uProgID, const GsePkt* pOrgPkt,
                         JWavesSciIDP* pIDP, int iMP, byte uNoMP)
{
	/* Cast in order of appearence */
	GsePkt* pPkt;
	JWavesMroMP* pMP;
	JWavesMP*    pBaseMP;
	GseHdr* pHdr;
	GseHdr* pOrgHdr;
	int i;
	GseBlock* pBlk;
	JWavesMroBlk* pMro;
	JWavesUpkBlk* pUpk;
	GseProcBlk* pUpkBase;
	
	
	pPkt = new_empty_GsePkt(0);
	except_cleanup_push(pPkt->del, pPkt);
	
	pPkt->setHdrLimits(pPkt, 256, 65535);
	
	gse_assert( pIDP->pMPs->pAry[iMP]->uType == JW_MRO_MINIPACKET);
	
	pBaseMP = pIDP->pMPs->pAry[iMP];
	pMP     = (JWavesMroMP*) pBaseMP;
	pHdr = pPkt->pHdr;
	pOrgHdr = pOrgPkt->pHdr;
	
	pHdr->nSpaceCraft = pOrgHdr->nSpaceCraft;   /* Preserve the spacecraft ID */
	pHdr->uContent = 0xE0;       /* Set the "CIP" to 0xE for Memory read outs */
	
	/* Preserve any existing blocks by coping then from the old header
	   and adding them to the new */
	for(i = 0; i<pOrgHdr->numBlocks(pOrgHdr); i++){
		pBlk = pOrgHdr->getBlockAt(pOrgHdr, i);
		pHdr->addBlock(pHdr, pBlk->copy(pBlk, 0));
	}
	
	/* Make a new Memory Readout Description block */
	pMro = new_JWavesMroBlk(CBBUF_NULL);
	
	pMro->uModStep = JW_UPACK_PROC_STEP_ID;
	pMro->uColRTI  = pBaseMP->uRTI % 40;
	pMro->uColSCLK = pBaseMP->collectSCLK(pBaseMP, pIDP->uSCLK);
	pMro->uIDP_CRC = pIDP->uCRC;
	pMro->uSeqNo   = pIDP->uSeq;
	pMro->uTs      = pIDP->ts;
	pMro->uRepSclk = pIDP->uSCLK;
	pMro->uMPlen   = pBaseMP->uLen;
	pMro->uMPrti   = pBaseMP->uRTI;
	pMro->uFrame   = pMP->uFrame;
	pMro->uReg     = pMP->uReg;
	pMro->uAddress = pMP->uAddress;
	pMro->uMSFlen  = pBaseMP->bufMSF.len;
	
	for(i=0; i < pBaseMP->bufMSF.len; i++)
		pMro->MSFary[i] = pBaseMP->bufMSF.ptr[i];
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pMro);
	
	/* Set the data contents indicator, Fortunatly, MROs can't be segmented, or
	   compressed, so this is the end of the line */
	pHdr->uContent |= 0x04;
	
	/* Make an Unpacker process block */
	pUpk = new_JWavesUpkBlk(CBBUF_NULL, 0);
	pUpk->uMiniPkts = uNoMP;
	pUpk->iMiniPkt = iMP;
	
	pUpkBase = (GseProcBlk*)pUpk;
	pUpkBase->setNow(pUpkBase);
	pUpkBase->uProgID = uProgID;
	/* ((GseBlock*)pUpk)->uLen = 15;  */ /* remember uLen field is length - 1 */
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pUpk);
	
	/* Set the data */
	pPkt->setData(pPkt, pBaseMP->takeData(pBaseMP));
	
	except_cleanup_pop(except_no_call);
	return pPkt;
}

/****************************************************************************/
/* Unpack raw data packets into 1-N packets, don't catch exceptions */

GsePktAry* JWavesUpkFilter_idp2GseAry(
		JWavesUpkFilter* pThis, GsePkt* pPkt, int nPktIdx, size_t uPktOff)
{
	GsePktFilter* pBase = (GsePktFilter*)pThis;
	GsePktAry* pPA;
	GsePkt* pNewPkt;
	JWavesSciIDP* pIDP;
	int i;
	
	pPA = new_GsePktAry(0);
	except_cleanup_push(pPA->del, pPA);      /* the packet array */
	
	pIDP = new_JWavesSciIDP(pPkt->getData(pPkt), uPktOff);
	except_cleanup_push(pIDP->del, pIDP);    /* the temp IDP object */
	
	for(i = 0; i < pIDP->pMPs->num; i++){
		
		if(pIDP->pMPs->pAry[i]->uType == JW_SCI_MINIPACKET)			
			pNewPkt = jwaves_mk_scipkt(pBase->uProgID, pPkt, pIDP, i, pIDP->pMPs->num, 
					                     pThis->bConvIDs);
		else
			pNewPkt = jwaves_mk_mropkt(pBase->uProgID, pPkt, pIDP, i, pIDP->pMPs->num);
		
		pPA->add(pPA, pNewPkt);
	}
	
	except_cleanup_pop(except_call);         /* Always del the IDP Object */
	except_cleanup_pop(except_no_call);      /* del the pkt array on xecpt */ 
	return pPA;
}

/****************************************************************************/
/* Unpack raw data packets into 1-N packets, routes errors to the catcher,  */
/* doesn't throw but rather returns NULL on an error.                       */

GsePktAry* JWavesUpkFilter_idp2GseAry_catch(
	JWavesUpkFilter* pThis, GsePkt* pPkt, int nPktIdx, size_t uPktOff, bool* pContinue)
{
	GsePktFilter* pBase = (GsePktFilter*)pThis;
	GsePktAry* pPA = NULL;
	GsePkt* pNewPkt = NULL;
	JWavesSciIDP* pIDP = NULL;
	int i = 0;
	bool bHitException = false;
	static const except_id_t catch_spec[] = {{XCEPT_GROUP_ANY, XCEPT_CODE_ANY}};
	except_t* pExcept = NULL;
	char sBuf[1024];
	sBuf[1023] = '\0';
	char* sErrMsg = NULL;
	
	except_try_push(catch_spec, 1, &pExcept);
	if(pExcept == NULL){
		pIDP = new_JWavesSciIDP(pPkt->getData(pPkt), uPktOff);
	}
	else{
		gse_warning_v("Bad packet: %s", gse_shortprn_except(sBuf, 399, pExcept));
		gse_sprint_except(sBuf, 1023, pExcept);
		sErrMsg = NS("Error in packet %d at offset 0x%08X:\n%s\n\n", 
				       nPktIdx + 1, uPktOff, sBuf);
		*pContinue =  pBase->pCatcher(pBase->pCatchCtx, pPkt, sErrMsg);
	}
	except_try_pop();
	
	if(pIDP == NULL)
		return NULL;  /* This packet is an error, caller can send another one */
	
	
	/* Have the IDP, now grind into packets */
	
	except_cleanup_push(pIDP->del, pIDP);    /* the temp IDP object */
	
	
	pPA = new_GsePktAry(0);
	except_cleanup_push(pPA->del, pPA);      /* the packet array */
	
	
	bHitException = false;
	for(i = 0; i < pIDP->pMPs->num; i++){
		pNewPkt = NULL;
		
		/* Okay, if one or more of these converts fail, sent the input to the
		   catcher, and stop trying to get the other mini-packets */
		
		except_try_push(catch_spec, 1, &pExcept);
		if(pExcept == NULL){
		
			if(pIDP->pMPs->pAry[i]->uType == JW_SCI_MINIPACKET)			
				pNewPkt = jwaves_mk_scipkt(pBase->uProgID, pPkt, pIDP, i,
						                     pIDP->pMPs->num, pThis->bConvIDs);
			else
				pNewPkt = jwaves_mk_mropkt(pBase->uProgID, pPkt, pIDP, i, 
						                     pIDP->pMPs->num);
		}
		else{
			gse_warning_v("Bad packet: %s", gse_shortprn_except(sBuf, 399, pExcept));
			gse_sprint_except(sBuf, 1023, pExcept);
			sErrMsg = NS("Error in packet %d at offset 0x%08X:\n%s\n\n", 
				       nPktIdx+1, uPktOff, sBuf);
			*pContinue =  pBase->pCatcher(pBase->pCatchCtx, pPkt, sErrMsg);
			bHitException = true;
		}
		except_try_pop();
		
		if(pNewPkt) pPA->add(pPA, pNewPkt);
		else break;
	}
	
	except_cleanup_pop(except_no_call);      /* del the pkt array on xecpt */ 
	except_cleanup_pop(except_call);         /* Always del the IDP Object */
	
	if(bHitException){
		pPA->del(pPA);
		return NULL;  /* This packet is an error, caller can send another one */
	}
	else{
		return pPA;
	}
} 

/****************************************************************************/
bool JWavesUpkFilter_false(GsePktSrc* pIface)
{
	return false;
}

/****************************************************************************/
GsePkt* JWavesUpkFilter_readNext(GsePktSrc* pBase)
{
	GsePktFilter*    pGseFltr;
	JWavesUpkFilter* pThis;
	GsePkt* pRet = NULL;
	GsePkt* pPkt = NULL;
	GsePktAry* pNewPkts = NULL;
	size_t nOffset;
	int nPktIdx;
	bool bContinue = true;
	
	pThis = (JWavesUpkFilter*) pBase;
	pGseFltr = (GsePktFilter*) pBase;
	
	/* If something is waiting, send it out.  A Queue would work better for this
	   operation */
	if((pThis->pWaiting != NULL)&&(pThis->pWaiting->num > 0)){
		pRet = pThis->pWaiting->pAry[0];
		pThis->pWaiting->pAry[0] = NULL;
		pThis->pWaiting->rm_ptr(pThis->pWaiting, NULL);
		
		/* If that was the last one out of the wait queue, ditch the queue */
		if(pThis->pWaiting->num == 0){
			pThis->pWaiting->del(pThis->pWaiting);
			pThis->pWaiting = NULL;
		}
		return pRet;
	}
	
	
	/* Unpack the array, either storing errors or letting exceptions occur */
	while( pNewPkts == NULL){
		
		if( (pPkt = pGseFltr->pUpSrc->readNext(pGseFltr->pUpSrc)) == NULL)
			return NULL;
		
		nOffset = pGseFltr->pUpSrc->offset(pGseFltr->pUpSrc);
		nPktIdx = pGseFltr->pUpSrc->pktIdx(pGseFltr->pUpSrc);
	
		/* Ignore packets where the content type is not Raw IDP */
		if( (pPkt->pHdr->uContent & 0x0F) != JWAVES_CNT_RAW_IDP){
			gse_debug("Unpacking not required.");
			return pPkt;
		}
		
		if(pGseFltr->pCatcher == NULL){	
			except_cleanup_push(pPkt->del, pPkt);
			pNewPkts = JWavesUpkFilter_idp2GseAry(pThis, pPkt, nPktIdx, nOffset);
			except_cleanup_pop(except_call);	 /* Cleans up the packet */
		}
		else{
			pNewPkts = JWavesUpkFilter_idp2GseAry_catch(pThis, pPkt, nPktIdx, 
			                                            nOffset, &bContinue);
			
			/* Only del our packets, not those given to the catcher */
			if(pNewPkts != NULL){ 
				pPkt->del(pPkt);
				pPkt = NULL;
			}
			
			if(! bContinue) /* User supplied callback wishes to stop */
				return NULL;
		}
	}
		
	
	pRet = pNewPkts->pAry[0];
	pNewPkts->pAry[0] = NULL;
	
	if(pNewPkts->num == 1){
		pNewPkts->del(pNewPkts);
	}
	else{
		/* More than one returned, some are going to have to wait */
		assert(pThis->pWaiting == NULL);
		pNewPkts->rm_ptr(pNewPkts, NULL);
		pThis->pWaiting = pNewPkts;
	}
	
	return pRet;
}


/****************************************************************************/
void delete_JWavesUpkFilter(void* vpThis)
{
	JWavesUpkFilter* pThis = (JWavesUpkFilter*)vpThis;
	if(pThis->pWaiting != NULL) pThis->pWaiting->del(pThis->pWaiting);
	free(vpThis);
}


/****************************************************************************/
JWavesUpkFilter* new_JWavesUpkFilter(GsePktSrc* pSrc, byte uProgID, uint4 uFlags)
{
	JWavesUpkFilter* pThis;
	
	pThis = (JWavesUpkFilter*)new_GsePktFilter(
			    pSrc, 
	          uProgID,
	          sizeof(JWavesUpkFilter),
	          JWavesUpkFilter_readNext,
	          delete_JWavesUpkFilter
	        );
	
	/* My stuff */
	
	pThis->del            = delete_JWavesUpkFilter;
	
	if(uFlags & JW_FLTR_UPK_CONV_IDS)
		pThis->bConvIDs    = true;
	else
		pThis->bConvIDs    = false;
	
	pThis->pWaiting       = NULL;
	
	return pThis;		
}

