/** @file JWavesUpkFilter.h */
#ifndef _JWavesUpkFilter_H_
#define _JWavesUpkFilter_H_

#include <rpwgse/GsePktFilter.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_jwaves_proc Processing */
/*@{*/

/** Source that Unpacks Packets containing Raw IDP data into GSE packets.
 *
 * The expected input is GSE packets with content type 0x81.  Packets with
 * a different content type are merely passed to the output un-altered.
 * This class in just a more convient interface to the jwaves_unpack()
 * routine.
 *
 * @class JWavesUpkFilter
 * @extends GsePktFilter
 */
typedef struct jwaves_upk_filter
{
	/** Enclosed base structure */
	GsePktFilter base;
		
	/** We may have changed the offical packet ID and Source values
    * for various packet types.  If true, old packet and source IDs will
    * be converted to the new value as unpacking is preformed.
	 */
	bool bConvIDs;
	
	/* New Unpacked packets that are waiting thier turn to head out the door */
	GsePktAry* pWaiting;
	
	/** Destructor */
	void (*del)(void* vpThis);
} JWavesUpkFilter;


/** Constructor for an unpacking packet source.
 *
 * Generates a new unpackting file source.  This filter unpacks recoginzed
 * packets before passing them on.  Raw IDP data may contian 1 to N minipackets.
 * This function generates new packets that contain only one minipacket per GSE
 * packet.  The @ref ::JWavesSDSblk "Science Data Status" header block is
 * defined for all output packets.  
 *
 * @note All packets with content type <b>Raw IDP</b> (0x1 in lower 4 bits
 * of GseHdr::uContent) @b must move through this processing stage
 * <i>even if they contain only a single minipacket</i>.  This function is
 * the only one that understands the Raw IDP format and is the only stage 
 * that generates the Science Data Status block.
 *
 * @public
 * @memberof JWavesUpkFilter
 *
 * @param pSrc A Source of Juno GSE packets with content type 0x81, other
 *        packet types maybe mixed in, but these are merely passed on 
 *        un-altered
 *
 * @param uProgID The program ID to attach to any generated packets.
 *
 * @param uFlags We may have changed the offical packet ID and Source values
 *        for various packet types.  If the flag @a JW_FLTR_UPK_CONV_IDS is set, 
 *        old packet and source IDs will be converted to the new value as unpacking
 *        is preformed.
 *
 * @return a pointer to a new JWavesUpkFilter structure allocated on the heap.
 */
JWavesUpkFilter* new_JWavesUpkFilter(GsePktSrc* pSrc, byte uProgID, uint4 uFlags);


/* Unpack Raw IDP data into GSE packets.
 *
 
 *  
 * @param pPkt a pointer to the GsePkt to unpack.  If the content type if the
 *        packet is not Raw IDP (0x1) no action is taken.
 *
 * @param uProgID The ID of calling program.  All programs altering Juno
 *        Waves data have a unique ID.
 *
 * @param bConvIDs We may have changed the offical packet ID and Source values
 *        for various packet types.  If this parameter is @a true old packet
 *        and source IDs will be converted to the new value as unpacking
 *        is preformed.
 * 
 * @param uPktOff The offset in the source to the input packet, used for
 *        error reporting.
 *
 * @retval "New packets"  An array of new packets where both the GsePkt
 *         structures and the pointer array are allocated on the heap.
 *         One output packet in generated for each @ref JWavesMP "Minipacket"
 *         in the data section of the input.
 *
 * @retval NULL if the input packet did not require unpacking.  Note that NULL
 *         does @b not indicate an error.
 *
 * @throws varies  Almost any type of exception may be thrown during packet
 *         processing.
 */
/* GsePktAry* jwaves_unpack(const GsePkt* pPkt, byte uProgID, bool bConvIDs,
                         size_t uPktOff); */

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesUpkFilter_H_ */
