/** \file JWavesUsegBlk.h */

#ifndef _JWavesUsegBlk_H_
#define _JWavesUsegBlk_H_

#include <rpwgse/GseProcBlk.h>
#include <jwaves/JWavesDefs.h>

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_jwaves_pkt */
/*@{*/

/** Interface for Gse Unsegmentation Processing Blocks.
 * Note that when this block is present in a GSE header, reading and unpacking
 * blocks will not be present.  Data in those blocks are collapsed into this
 * one.
 *
 * @class JWavesUsegBlk
 * @extends GseProcBlk
 */
typedef struct jwaves_useg_blk{
	
	/** Base structure */
	GseProcBlk base;
	
	/** The number of segments combined to generate the complete capture. */
	byte uNumSegs;
	
	/** The lowestest program ID that preformed unpacking on data contained 
	 * in this packet. */
	byte uLoUnpakID;

	/** The greatest program ID that preformed unpacking on data contained in
	 * this packet. */
	byte uHiUnpakID;
	
	/** The lowest program ID that read data contained in this packet from its 
	 * original source. */
	byte uLoRdrID;
	
	/** The lowest source ID for data contained in this packet. */
	byte uLoSource;
	
	/** The earliest time of reading for data contained in this packet.  This
	 * is a 32 bit unsigned MSB first integer.  It is the number of seconds
	 * elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time
	 * (UTC). */
	uint4 uLoReadSec;
	
	/** The earliest time of reading for data contained in this packet. This is
	 * an unsigned MSB first 16-bit twos compliment integer containing 1/100th 
	 * the number of microseconds since the LO_READ_BEGIN_TIME_SEC.  To get the
	 * earliest read time microseconds, multiply this value by 100. */
	uint2 uLoReadUS;
	
	/** The greatest program ID that read data contained in this packet from
	 * its original source. */
	byte uHiRdrID;
	
	/** The greatest source ID for data contained in this packet. */
	byte uHiSource;
	
	/** The latest time of reading for data contained in this packet.  This is
	 * a 32 bit unsigned MSB first integer.  It is the number of seconds
	 * elapsed since 00:00:00 on January 1, 1970, Coordinated Universal Time
	 * (UTC). */
	uint4 uHiReadSec;
	
	/** The latest time of reading for data contained in this packet. This is
	 * an unsigned MSB first 16-bit twos compliment integer containing 1/100th
	 * the number of microseconds since the LO_READ_BEGIN_TIME_SEC.  To get the
	 * latest read time microseconds, multiply this value by 100. */
	uint2 uHiReadUS;
	
	
	/** Contains the original sequence and telemetry source numbers of the IDP
    packets used to generate this unsegmented data.  These 16 bits contain
    two fields:
    
      15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
    +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    |   OS  |       Big Endian Sequence Number 0 to  16383          |
    +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
    
    The values for TS (Telemetry Source) are:
    
       0 - Housekeeping
       1 - Low Rate Science
    
    Up to 16 mini-packets can be combined into a single unsegmented mini-packet.
	 If mini-packet data  can be up to 4 KB then this provides for 64 KB data
	 sections.  Note that 0 is a legal value for at most one of the sequence
	 numbers.  To distinguish between trailing zeros and a real sequence number
	 consult the uNumSegs member. */
	uint2 uSeqTSary[16];
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} JWavesUsegBlk;

/** Generate a new Juno Waves unsegmentation Processing Block.
 * @memberof JWavesUsegBlk
 * @public
 * @todo Add more error checking to the decoding step 
 */
JWavesUsegBlk* new_JWavesUsegBlk(cbytebuf_t buf, size_t uOverMalloc);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _JWavesUsegBlk_H_ */

