#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseTime.h>
#include <rpwgse/GseLog.h>

#include "JWavesDefs.h"
#include "JWavesRdrBlk.h"
#include "JWavesUpkBlk.h"
#include "JWavesSDSblk.h"
#include "JWavesUsegBlk.h"

#include "JWavesUsegFilter.h"


/***************************************************************************/
/* Make a new packet from an existing data channel */

GsePkt* JWavesUsegFilter_newPkt(JWavesUsegFilter* pThis, byte uPSID, byte uSegs,
		                          byte uErrs)
{
	GsePkt* pPkt = NULL;                   /* Existing item references */
	JWavesSDSblk* pSDS = NULL;
	JWavesRdrBlk* pRdr = NULL;
	GseProcBlk*   pProcRdr = NULL;
	GseProcBlk* pProcUpk  = NULL;
	GseBlock*   pBlk = NULL;
	cbytebuf_t  cbuf = {NULL, 0};
	
	GsePkt* pNew = NULL;                   /* Things made here */
	JWavesSDSblk* pNewSDS = NULL;
	JWavesUsegBlk* pUseg = NULL; 
	GseProcBlk* pProcUseg = NULL;
	
	bytebuf_t data = {NULL, 0};
	int i;
	byte u;
	size_t uOffset = 0;
	bool bHasRdrInfo = false;
	
	if(uSegs < 1)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "No. Segments < 1");
	
	pNew = new_empty_GsePkt(0);
	except_cleanup_push(pNew->del, pNew);    /* <-- cleanup */
	pNew->setHdrLimits(pNew, 256, 65535);
	
	pUseg = new_JWavesUsegBlk(CBBUF_NULL, 0);
	except_cleanup_push(pUseg->del, pUseg);  /* <-- cleanup */
	
	pProcUseg = (GseProcBlk*)pUseg;
	pProcUseg->setNow(pProcUseg);           /* Record the op time */
	pProcUseg->uProgID = pThis->base.uProgID;    /* Set program ID */
	pProcUseg->uErrors = uErrs;             /* Record the inconsistance count */
	

	pPkt = pThis->channel[uPSID][0];	
	
	pNew->pHdr->nSpaceCraft = pPkt->pHdr->nSpaceCraft;   /* spacecraft ID */

	/* Copy all blocks from the 0th header into the new header */
	for(i = 0; i<pPkt->pHdr->numBlocks(pPkt->pHdr); i++){
		pBlk = pPkt->pHdr->getBlockAt(pPkt->pHdr, i);
		pNew->pHdr->addBlock(pNew->pHdr, pBlk->copy(pBlk, 0));
	}
	
	/* Set the content byte */
	pNewSDS = (JWavesSDSblk*)pNew->getBlkByType(pNew, JW_SDS_BLK_TYPE_ID, 0);
	if(pNewSDS->isCompressed(pNewSDS) || pNewSDS->isGlopped(pNewSDS))
		pNew->pHdr->uContent = 0x83;
	else
		pNew->pHdr->uContent = 0x84;
	
	/* Set the new SDS block to indicate unsegmented data, and set
	   the content type */
	pNewSDS->bEOC = true;
	pNewSDS->uSegNo = 0;
	
	
	/* State which step last modified the SDS block */
	pNewSDS->uModStep   = JW_USEG_PROC_STEP_ID;
	
	/* Init the high and low trackers to extream opposite values */
	pNewSDS->uRepSclkLo = 0xFFFFFFFF;
	pNewSDS->uRepSclkHi = 0;
	pNewSDS->uSeqNoLo   = 0xFFFF;
	pNewSDS->uSeqNoHi   = 0;
	pNewSDS->uTsLo      = hsk;
	pNewSDS->uTsHi      = invalid;
	
	pUseg->uNumSegs     = uSegs;
	pUseg->uLoUnpakID   = 0xFF;
	pUseg->uHiUnpakID   = 0;
	
	pUseg->uLoRdrID     = 0xFF;
	pUseg->uLoSource    = 0xFF;
	pUseg->uLoReadSec   = 0xFFFFFFFF;
	pUseg->uLoReadUS    = 0xFFFF;
	
	pUseg->uHiRdrID     = 0;
	pUseg->uHiSource    = 0;
	pUseg->uHiReadSec   = 0;
	pUseg->uHiReadUS    = 0;
	
	/* allocate the data buffer */
	for(u = 0; u<uSegs; u++){
		pPkt = pThis->channel[uPSID][u];
		cbuf = pPkt->getData(pPkt);
		data.len += cbuf.len;
	}
	data.ptr = (byte*)gse_calloc(data.len, sizeof(byte));
	except_cleanup_push(free, data.ptr);  /* <-- cleanup */
	
	/* Consolidate... */
	uOffset = 0;
	bHasRdrInfo = false;
	for(u=0; u<uSegs; u++){
		pPkt = pThis->channel[uPSID][u];
		
		/* Update SDS with souce info */
		pSDS = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		if(pNewSDS->uRepSclkLo > pSDS->uRepSclkLo) 
			pNewSDS->uRepSclkLo = pSDS->uRepSclkLo;
		
		if(pNewSDS->uRepSclkHi < pSDS->uRepSclkHi) 
			pNewSDS->uRepSclkHi = pSDS->uRepSclkHi;
		
		if(pNewSDS->uSeqNoLo > pSDS->uSeqNoLo) pNewSDS->uSeqNoLo = pSDS->uSeqNoLo;
		if(pNewSDS->uSeqNoHi < pSDS->uSeqNoHi) pNewSDS->uSeqNoHi = pSDS->uSeqNoHi;
		
		if(pNewSDS->uTsLo > pSDS->uTsLo) pNewSDS->uTsLo = pSDS->uTsLo;
		if(pNewSDS->uTsHi < pSDS->uTsHi) pNewSDS->uTsHi = pSDS->uTsHi;
		
		
		/* Add reader info to Useg Blk */
		if( (pRdr = jwaves_get_rdr_blk(pPkt)) != NULL){
			
			if(pUseg->uLoSource > pRdr->uSource)
				pUseg->uLoSource = pRdr->uSource;
			
			if(pUseg->uHiSource < pRdr->uSource)
				pUseg->uHiSource = pRdr->uSource;
			
			pProcRdr = (GseProcBlk*)pRdr;
			
			if(pUseg->uLoRdrID > pProcRdr->uProgID)
				pUseg->uLoRdrID = pProcRdr->uProgID;
			
			if(pUseg->uHiRdrID < pProcRdr->uProgID)
				pUseg->uHiRdrID = pProcRdr->uProgID;
	
			
			if( (((int8)pUseg->uLoReadSec)*1000000 + pUseg->uLoReadUS*100) > 
				 (((int8)pProcRdr->uBegSec)*1000000 + pProcRdr->uBegUSecDiv100*100) )
			{
				pUseg->uLoReadSec = pProcRdr->uBegSec;
				pUseg->uLoReadUS = pProcRdr->uBegUSecDiv100;
			}
			
			if( (((int8)pUseg->uHiReadSec)*1000000 + pUseg->uHiReadUS*100) < 
				 (((int8)pProcRdr->uBegSec)*1000000 + pProcRdr->uBegUSecDiv100*100) )
			{
				pUseg->uHiReadSec = pProcRdr->uBegSec;
				pUseg->uHiReadUS = pProcRdr->uBegUSecDiv100;
			}
			
			bHasRdrInfo = true;
		}
		
		/* Add unpack info to Useg Blk */
		pProcUpk = (GseProcBlk*) pPkt->getBlkByType(pPkt, GSE_PROC_BLK_TYPE, JW_UPACK_PROC_STEP_ID);
		
		if(pUseg->uLoUnpakID > pProcUpk->uProgID)
			pUseg->uLoUnpakID = pProcUpk->uProgID;
		
		if(pUseg->uHiUnpakID < pProcUpk->uProgID)
			pUseg->uHiUnpakID = pProcUpk->uProgID;
	
		/* Add in the sequence number and telem source for this segment */
		switch(pSDS->uTsLo){
			case hrs: pUseg->uSeqTSary[u] = pSDS->uSeqNoLo | 0x40; break;
			case lrs: pUseg->uSeqTSary[u] = pSDS->uSeqNoLo | 0x80; break;
			case hsk: pUseg->uSeqTSary[u] = pSDS->uSeqNoLo | 0xC0; break;
			default: pUseg->uSeqTSary[u] = pSDS->uSeqNoLo;
		}
		
		/* Copy this chunk of data */
		cbuf = pPkt->getData(pPkt);
		memcpy(data.ptr + uOffset, cbuf.ptr, cbuf.len);
		uOffset += cbuf.len;
	}
	
	/* If didn't get reader info, zero those parts out, bump warning count */
	if(! bHasRdrInfo ){
		pUseg->uLoRdrID     = 0;
		pUseg->uLoSource    = 0;
		pUseg->uLoReadSec   = 0;
		pUseg->uLoReadUS    = 0;
		
		/* High records were already set to zero above */
		pProcUseg->uWarnings += 1;
	}
	
	pNew->setData(pNew, data);
	except_cleanup_pop(except_no_call); /* pop data.ptr cleanup */
	
	pNew->addHdrBlk(pNew, (GseBlock*)pUseg);
	except_cleanup_pop(except_no_call); /* pop pUseg cleanup */
	
	except_cleanup_pop(except_no_call); /* pop pNew cleanup */
	return pNew;
}

/****************************************************************************/
/* Collapse a channel to a single packet if possible, cleans the            */
/* Channel entries as well                                                  */
GsePkt* JWavesUsegFilter_chanEmpty(JWavesUsegFilter* pThis, byte uPSID)
{
	GsePkt* pPkt;
	GsePkt* p1stPkt;
	GsePkt* pNewPkt;
	
	int iByte, iSeg;
	byte uSegs = 0;
	
	JWavesSDSblk* pSDS = NULL;
	JWavesSDSblk* p1stSDS = NULL;
	
	char* s1stUTC = NULL;
	
	cbytebuf_t buf1st;
	cbytebuf_t buf;
	
	struct timeval tv1st;
	byte uErrs = 0;
	
	
	/* First check to see if the channel is complete... */
	for(iSeg = 0; iSeg < 16; iSeg++){
		pPkt = pThis->channel[uPSID][iSeg];
	
		if(pPkt == NULL)
			return NULL;
		
		pSDS = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		
		if(pSDS->bEOC){
			uSegs = (byte)(iSeg + 1);
			break;
		}
	}
	
	if(uSegs < 1) return NULL;
	
	/* Okay, I have all the segments, check consistancy. */
			
	p1stPkt = pThis->channel[uPSID][0];
	p1stSDS = (JWavesSDSblk*)pPkt->getBlkByType(p1stPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	p1stSDS->collectUTC(p1stSDS, &tv1st);
	s1stUTC = gse_tv2iso_gm(&tv1st, GSE_MILLISECS);
	buf1st  = ((GseBlock*)p1stSDS)->original((GseBlock*)p1stSDS);
	
	if(p1stSDS->isCompressed(p1stSDS)){
		gse_info_v("Unsegmenting compressed packet type 0x%02hhX @ UTC %s", uPSID, s1stUTC);
	}
	else{
		if(p1stSDS->isGlopped(p1stSDS))
			gse_info_v("Unsegmenting glopped packet type 0x%02hhX @ UTC %s", uPSID, s1stUTC);
		else
			gse_info_v("Unsegmenting packet type 0x%02hhX @ UTC %s", uPSID, s1stUTC);
	}
	
	
	for(iSeg = 1; iSeg < uSegs; iSeg++){
		pPkt = pThis->channel[uPSID][iSeg];
		pSDS = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
		buf  = ((GseBlock*)pSDS)->original((GseBlock*)pSDS);
		
		/* Changed 2013-12-05 to use MP rti, since uColRTI allows measurements
		   from different instrument cycles to get mixed together */
		if(p1stSDS->uMPrti != pSDS->uMPrti){
			gse_error_v("Packet type 0x%02hhX at UTC %s: RTI field error. seg 0 "
					      "RTI: 0x%hX  seg %hhu RTI  0x%hX\n", uPSID, s1stUTC,
					      p1stSDS->uMPrti, pSDS->uMPrti);
			free(s1stUTC);
			return NULL;
		}
		
		if(p1stSDS->uFmt != pSDS->uFmt){
			gse_error_v("Packet type 0x%02hhX at UTC %s: FMT field error. seg 0 "
					     "FMT: 0x%hhX  seg %hhu FMT  0x%hhX\n", uPSID, s1stUTC,
					      p1stSDS->uFmt, pSDS->uFmt);
			free(s1stUTC);
			return NULL;
		}
		
		/* Record an error num for any deviations other than the LEN, SEGNO and
		   EOS header bits */
		for(iByte=26; iByte < buf1st.len; iByte++){
			
			/* Check all bytes in the MP header for similarity except for 
			   the lower 5 bits of byte 4 of the mph */
			if( iByte == 28 ){
				if((buf1st.ptr[iByte] & 0xE0) != (buf.ptr[iByte] & 0xE0)){
					if(uErrs < 0xFF) uErrs += 1;
				}
			}
			else{
				if(buf1st.ptr[iByte] != buf.ptr[iByte]){
					if(uErrs < 0xFF) uErrs += 1;
				}
			}
		}
		
		if(uErrs)
			gse_warning_v("Minipacket inconsistancies detected in segmented "
					        "capture.  Packet Type 0x%02hhX,  UTC: %s  Seq: %hu",
		      	        uPSID, s1stUTC, p1stSDS->uSeqNoLo);
		
		/* Get Seq and Rep SCLK bounds */
		
	}
	free(s1stUTC);
	
	/* They are consistant enough, make a single new output packet */
	if( (pNewPkt = JWavesUsegFilter_newPkt(pThis, uPSID, uSegs, uErrs)) == NULL)
		return NULL;	
	
	/* Remove the old packets */
	for(iSeg=0; iSeg<uSegs; iSeg++){
		pThis->channel[uPSID][iSeg]->del(pThis->channel[uPSID][iSeg]);
		pThis->channel[uPSID][iSeg] = NULL;
	}
	
	return pNewPkt;
}

/****************************************************************************/
GsePkt* JWavesUsegFilter_readNext(GsePktSrc* pBase)
{
	GsePkt* pPkt       = NULL;
	JWavesSDSblk* pSDS = NULL;
	GsePkt* pRmPkt     = NULL;
	JWavesSDSblk* pRmSDS = NULL;
	struct timeval rmTV;
	char* sRmUTC       = NULL;
	
	JWavesUsegFilter* pThis = (JWavesUsegFilter*)pBase;
	
	/* Algorithim mentality:
	 *
	 * 1. EOSEG = 1,  SEGNO = 0:
	 *
	 *   Simple case no unsegmentation needed, pass to output
	 *
	 * 2. not (EOSEG = 1 and SEGNO = 0):
	 *
	 *    If there is an existing packet in this spot, Throw it out and log
	 *    as an error. 
	 *
	 *    Save the new packet.
	 *
	 *    Output the channel, if it's complete and constistant enough
	 */
	 
	 while(true){
		
		if( (pPkt = pThis->base.pUpSrc->readNext(pThis->base.pUpSrc)) == NULL)
			return NULL;
		
		/* JWAVES_KIND_SCI_DATA + JWAVES_CNT_MP_DATA */
		
		if(pPkt->pHdr->uContent != 0x83)
			return pPkt;
		
		if( ! pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0) ){
			gse_error_v("Missing Science Data Status Block in packet at "
					      "offset 0x%X, source: %s", 
					      pThis->base.pUpSrc->offset(pThis->base.pUpSrc),
					      pThis->base.pUpSrc->name(pThis->base.pUpSrc)  );
			continue;
		}
		
		pSDS = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	 
		/* 1. Simple case, not segmented */
		if(pSDS->bEOC && (pSDS->uSegNo == 0))
			return pPkt;
		
		/* Temporary Hack, adding a breakpoint location for packets with a sclk
		   above 1A2B78F4 */
	 	if(0x1A2B78F4 <= pSDS->uRepSclkHi && pSDS->uRepSclkHi <= 0x1A2B7900){
			gse_info_v("Hit packet with report sclk 0x%08X", pSDS->uRepSclkHi);
		}
		
		/* 2. Is part of a segment, add to the channel */
		if(pSDS->uSegNo > 15) gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		
		/* Throwing out segments from incomplete channel */
		if( pThis->channel[pSDS->uPSID][pSDS->uSegNo] != NULL){
			pRmPkt = pThis->channel[pSDS->uPSID][pSDS->uSegNo];
			pRmSDS = (JWavesSDSblk*)pRmPkt->getBlkByType(pRmPkt,JW_SDS_BLK_TYPE_ID,0);
			pRmSDS->collectUTC(pRmSDS, &rmTV);
			
			sRmUTC = gse_tv2iso_gm(&rmTV, GSE_MILLISECS);		
			gse_error_v("Incomplete capture, pkt Type 0x%02X, at 0x%08X (UTC %s), "
					      "Segment %d was dropped to make room for the next "
					      "capture\n", pRmSDS->uPSID, pRmSDS->uColSCLK, sRmUTC,
					      pRmSDS->uSegNo); 
			free(sRmUTC);
			pRmPkt->del(pRmPkt);
			pRmPkt = NULL;
		}
		
		pThis->channel[pSDS->uPSID][pSDS->uSegNo] = pPkt;
		
		/* 3. Output unsegmented packet if needed */
		if( (pPkt = JWavesUsegFilter_chanEmpty(pThis, pSDS->uPSID)) != NULL)
			return pPkt;
		
		/* ... Go round again, didn't have a complete channel */
		
	 }
	 
	  
	return NULL;
}


/****************************************************************************/
void delete_JWavesUsegFilter(void* vpThis)
{	
	JWavesUsegFilter* pThis = (JWavesUsegFilter*)vpThis;
	int i, j;
	
	for(i=0; i<256; i++)
		for(j=0; j<16; j++)
			if(pThis->channel[i][j] != NULL)
				pThis->channel[i][j]->del(pThis->channel[i][j]);
	
	free(vpThis);
}


/****************************************************************************/
JWavesUsegFilter* new_JWavesUsegFilter(GsePktSrc* pSrc, byte uProgID)
{
	JWavesUsegFilter* pThis = NULL;
	
	pThis = (JWavesUsegFilter*)new_GsePktFilter(
			    pSrc, 
	          uProgID,
	          sizeof(JWavesUsegFilter),
	          JWavesUsegFilter_readNext,
	          delete_JWavesUsegFilter
	        );
	
	/* My stuff */
	pThis->del            = delete_JWavesUsegFilter;
	
	/* The calloc call zero filled the channel array for us! */
	
	return pThis;	
}
