#include <stdlib.h>
#include <assert.h>
#include <stdbool.h>

#include "BfdContext.h"


struct bfd_find_data
{
	const char* func;
	asymbol** symbol_table;
	bfd_vma counter;
};

/*****************************************************************************/
/* Name lookup */

void find_func_name_in_sec(bfd *abfd, asection *sec, void *opaque_data)
{
	const char* sFunc = NULL;
	const char* sFile = NULL;
	unsigned uLine = 0;
	bfd_vma vma;
	struct bfd_find_data* data = (struct bfd_find_data*)opaque_data;
	
	assert(sec);
	assert(opaque_data);

	if( data->func != NULL) return;  /* already have it */

	if (!(bfd_get_section_flags(abfd, sec) & SEC_ALLOC)) return;

	vma = bfd_get_section_vma(abfd, sec);
	if(data->counter < vma || vma + bfd_get_section_size(sec) <= data->counter) return;

	if (bfd_find_nearest_line(abfd, sec, data->symbol_table, data->counter - vma, 
			                    &sFile, &sFunc, &uLine) && sFunc)
	 	data->func = sFunc;
}

const char* BfdContext_getFuncName(struct bfd_context* pThis, unsigned long offset)
{
	struct bfd_find_data data;
	data.symbol_table = pThis->symbol_table_;
	data.counter = offset;
	data.func = NULL;
	
	bfd_map_over_sections(pThis->abfd_, &find_func_name_in_sec, &data);

	return data.func;
}
	
	
/*****************************************************************************/
/* Destructor */
void delete_BfdContext(void* vpThis)
{
	BfdContext* pThis = (BfdContext*)vpThis;
	
	free(pThis->symbol_table_);
	bfd_close(pThis->abfd_);
	free(vpThis);
}


/*****************************************************************************/
/* Constructor */

BfdContext* new_BfdContext(const char* sFileName)
{
	BfdContext* pThis;
	char **formats = NULL;
	bool b1, b2, b3;
	unsigned int dummy = 0;
	
	pThis = (BfdContext*)calloc(1, sizeof(BfdContext));
	pThis->getFuncName = BfdContext_getFuncName;
	pThis->del         = delete_BfdContext;
	
	bfd_init();
	if( ! (pThis->abfd_ = bfd_openr(sFileName, 0) )){
		free(pThis);
   	return NULL;
	}
	
	b1 = bfd_check_format(pThis->abfd_, bfd_object);
	b2 = bfd_check_format_matches(pThis->abfd_, bfd_object, &formats);
	b3 = bfd_get_file_flags(pThis->abfd_) & HAS_SYMS;
	
	if( !(b1 && b2 && b3) ){
		
		bfd_close(pThis->abfd_);
		free(formats);
		free(pThis);
		return NULL;
	}
	free(formats);
	
	/* Load the symbol table */
	if( bfd_read_minisymbols(pThis->abfd_, FALSE, (void **)&(pThis->symbol_table_), 
			                   &dummy) == 0 &&
	    bfd_read_minisymbols(pThis->abfd_, TRUE, (void **)&(pThis->symbol_table_),
		                      &dummy) < 0){
		
		free(pThis->symbol_table_);
		bfd_close(pThis->abfd_);
		return NULL;
	}
	
	return pThis;
}
