#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseBlock.h"
#include "GseConv.h"
#include "GseLog.h"

bytebuf_t GseBlock_encode(GseBlock* pThis, byte uExtraLen, byte* pOffset){
	bytebuf_t ret = {NULL, 0};
	byte uTmp;
	
	/* Checks for over run */
	gse_storeUinU(uExtraLen + 1, &uTmp, sizeof(byte));
	
	/* Use the length in uLen, unless that is less than what is needed */
	ret.len = 2 + uExtraLen;
	if(ret.len < (pThis->uLen + 1))
		ret.len = pThis->uLen + 1;
	else
		pThis->uLen = ret.len - 1;
	
	ret.ptr = gse_calloc(ret.len, sizeof(byte));
	
	ret.ptr[0] = pThis->uType;
	ret.ptr[1] = pThis->uLen;
	
	/* If I was decoded, go a head and prefill the buffer with orig bites,
	   derived classes may over-write these bytes but that's thier perogative */
	if(pThis->_bufOrg.ptr != NULL){
		
		if(ret.len < pThis->_bufOrg.len)
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, 
			   "uExtraLen is too short, cannot copy all the original bytes "
				"into the allocated buffer, ret.len = %zd, but _bufOrg.len = %zd", 
				ret.len, pThis->_bufOrg.len);
		
		memcpy(ret.ptr + 2, pThis->_bufOrg.ptr + 2, pThis->_bufOrg.len - 2);
	}
	
	if(pOffset != NULL) *pOffset = 2;
	return ret; 
}

const char* GseBlock_name(GseBlock* pThis){
	return pThis->_sName;
}


/* Decoding: checks uType and uLen */
byte GseBlock_decode(GseBlock* pThis, cbytebuf_t buf){
	if(buf.ptr == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Cannot decode a NULL buffer");
	
	if(buf.len < 2)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Buffer is to short to be a GSE "
		            "header block");
	
	pThis->uType = buf.ptr[0];	
	pThis->uLen = buf.ptr[1];
		
	if(buf.len < pThis->_uMinLen)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "At least %hhu bytes are required"
				" to make a %s header block", pThis->_uMinLen, pThis->_sName);
		
	if(buf.ptr[1] + 1 < pThis->_uMinLen)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE,
		            "Length field (byte idx 1) was %hhu, it must be at least %hhu"
		            "to make a %s header block", buf.ptr[1], pThis->_uMinLen,
		            pThis->_sName);
	
	if(pThis->uLen + 1 > buf.len)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Didn't get the entire block, the "
		             "length field + 1 was %hhu, but the buffer is only %hhu "
		             "bytes long.", pThis->uLen + 1, buf.len);
	
	/* Save the original bytes */
	pThis->_bufOrg = copy_cbytebuf(buf);
	
	return 2;
}

cbytebuf_t GseBlock_original(const GseBlock* pThis){
	return const_bytebuf(pThis->_bufOrg);
}

/* Destructor */
void delete_GseBlock(void* vpThis){
	GseBlock* pThis = (GseBlock*)vpThis;
	
	if(pThis->_bufOrg.ptr != NULL) free(pThis->_bufOrg.ptr);
	free(pThis);
}

/* Base class merge, does nothing */
bool GseBlock_merge(GseBlock* pThis, const GseBlock* pOther, uint4 uFlags)
{
	gse_debug("Merge call defaulted to the base GseBlock do nothing merge.");
	return false;
}

bool GseBlock_canMerge(GseBlock* pThis, const GseBlock* pOther, uint4 uFlags)
{
	return false;
}


/* Copy constructor */
GseBlock* GseBlock_copy(GseBlock* pThis, size_t uOverMalloc){
	GseBlock* pNew = NULL;
	size_t uSize = 0;
	 
 	if(uOverMalloc > sizeof(GseBlock))
		uSize = uOverMalloc;
	else
		uSize = sizeof(GseBlock);
	
	/* Get the contained items */
	pNew = (GseBlock*)gse_calloc(1, uSize);
	memcpy(pNew, pThis, sizeof(GseBlock));
	
	/* Copy the buffers */
	if(pNew->_bufOrg.ptr == NULL){
		gse_assert(pNew->_bufOrg.len == 0);
		return pNew;
	}
	
	gse_assert(pNew->_bufOrg.len != 0);
	pNew->_bufOrg.ptr = (byte*)gse_calloc(pNew->_bufOrg.len, sizeof(byte));
	memcpy(pNew->_bufOrg.ptr, pThis->_bufOrg.ptr, pNew->_bufOrg.len);
	
	/* fprintf(stderr, "GseBlock_copy called\n"); */
	return pNew;
}

/* Constructor */
GseBlock* new_GseBlock(cbytebuf_t buf, size_t uOverMalloc){
	GseBlock* pThis = NULL;
	size_t uSize = 0;
	bool bIsTopLevel;
	 
 	if(uOverMalloc > sizeof(GseBlock)){
		uSize = uOverMalloc;
		bIsTopLevel = false;
	}
	else{
		uSize = sizeof(GseBlock);
		bIsTopLevel = true;
	}
	
	pThis = (GseBlock*)gse_calloc(1, uSize);
	
	except_cleanup_push(free, pThis);
	
	/* Attach the member functions */
	pThis->del = delete_GseBlock;
	pThis->encode = GseBlock_encode;
	pThis->name = GseBlock_name;
	pThis->decode = GseBlock_decode;
	pThis->original = GseBlock_original;
	pThis->copy     = GseBlock_copy;
	pThis->merge    = GseBlock_merge;
	pThis->canMerge = GseBlock_canMerge;
	
	pThis->_sName = "Generic";
	pThis->_uMinLen = 2;
	pThis->_bufOrg = BBUF_NULL;
	
	if(buf.ptr == NULL){
		pThis->uLen = 1;
		pThis->uType = 0;
	}
	else{
		GseBlock_decode(pThis, buf);
	}
	
	except_cleanup_pop(except_no_call);
	/* fprintf(stderr, "new_GseBlock called\n");*/
	return pThis;
}



/*****************************************************************************/
/* Need explicit instanitaiton of the GseBlockAry macro template,            */
/* In this case when a pointer is removed from the array make sure that      */
/* the del member function of the structure in called                        */

#define C GseBlkAry
#define T GseBlock
#define CASCADE_DEL

#include "ptrAryTplt.c"

#undef CASCADE_DEL
#undef C
#undef T
