/** @file GseBlock.h Generic Packet Header Blocks */
#ifndef _GseBlock_H_
#define _GseBlock_H_

#include <stdlib.h>
#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @addtogroup mod_packet Packets */
/*@{*/

/** Basic interface for all Gse Header Blocks, each GseHdr contains one or 
 * more of these. @n
 *
 * @class GseBlock
 */  
typedef struct gse_block{
	/** FIELD: The type of block */
	byte uType;
	
	/** FIELD: The length in bytes of this block's data on disk - 1.
	 *
	 * Set this if you want to use up more room on disk than is needed. @n
	 * @n
	 * The @a encode function automatically increases this field if it is
	 * too small to hold all the derived structure's bytes.  If less room
	 * is needed on disk then indicated by this field, zero padding is added
	 * to the encoded output.
	 */
	byte uLen;
	
	/** Make a byte stream from the current fields.
	 * 
	 * This regenerates a new byte string from the current fields.
	 * it does not necessarily regenerate decoded bytes.  To get the
	 * bytes, if any, that were decoded use original()
	 *
	 * @param uExtraLen extra bytes of memory to allocate, useful for
	 * derived classes.  Set this to zero for calling code.
	 *
	 * @param pOffset If this is not null, it is set to the number
	 * of bytes actully written into the buffer, as opposed the buffer's
	 * actual length.  The pointed to value is always set to the length
	 * of the returned buffer if uExtraLen = 0.  Useful for derived 
	 * classes. 
	 *
	 * @return a buffer containing the encoded bytes, plus zero padding
	 * out to uExtraLen
	 */
	bytebuf_t (*encode)(struct gse_block* pThis, byte uExtraLen, byte* pOffset);
	
	
	/** Get the original byte string use during decoding. 
	 *
	 * @return a constant byte buffer (not a new heap allocation) to the
	 *         original decoded bytes, or an empty buffer if this block
	 *         was not generated via decoding.
	 */
	cbytebuf_t (*original)(const struct gse_block* pThis);
	
	/** Decode bytes into fields and save a copy.
	 *
	 * Note if the buf arugment to new_GseBlock in not null, decode is
	 * called automaticlly.  The 2nd byte of the buffer is tested against
	 * _uMinLen.  If in is smaller than _uMinLen, an exception is thrown.
	 *
	 * @return the nunber of bytes that were decoded.  This is useful for 
	 *    derived classes that may wish to decode the rest of the bytes
	 *    in buf.
	 */
	byte (*decode)(struct gse_block* pThis, cbytebuf_t buf);
	
	/** Get a const string describing this block type */
	const char* (*name)(struct gse_block* pThis);
	
	/** Get a copy of this Header block
	 *
	 * @param pThis a pointer to a GsePkt object
	 * @param uOverMalloc is used by derived classes, just set to zero.
	 * @return A new GseBlock derived object allocated on the heap with the
	 *         same values and internal state as this one.  All sub objects
	 *         are copied as well
	 */
	struct gse_block* (*copy)(struct gse_block* pThis, size_t uOverMalloc);
	
	/** Merge the data from two similar blocks.
	 * This alters this block's data data, the input block is unchanged.
	 *
	 * @param pThis a pointer to a GsePkt object
	 * @param pBlock a pointer to a block to merge into this one.
	 * @param uFlags parameters that affect the merge operation.  The may
	 *        defined by various derived structures.
	 *
	 * @return true if the given block was succesfully merged into this
	 *         one, false if not.
	 * @public 
	 */
	bool (*merge)(struct gse_block* pThis, const struct gse_block* pOther,
	              uint4 uFlags);
	
	/** See if this block is mergeable with another block, but don't actually
	 * change anything.
	 */
	bool (*canMerge)(struct gse_block* pThis, const struct gse_block* pOther,
	              uint4 uFlags);
	 
	/** The long name of this type of block 
	 * @protected 
	 */
	const char* _sName;
	
	/** The original decoded bytes pointer. 
	 * @protected
	 */
	bytebuf_t _bufOrg;
	
	/** Minimum number of bytes required to hold all encoded fields
	 * for this block.
	 * @protected
	 */
	byte _uMinLen;
	
	/** Destructor */
	void (*del)(void* vpThis);
		
} GseBlock;

/** Generate a new GseBlock on the heap.
 *
 * @param buf  Parse this buffer into fields.  Note the bytes in buf.ptr 
 * are parsed into fields, but no reference is saved.  In is safe to free 
 * @c buf.ptr immediatly after calling this function.
 *
 * @param uOverMalloc Used by derived classes to allocate extra memory for
 * the new struture.
 *
 * @memberof GseBlock
 */
GseBlock* new_GseBlock(cbytebuf_t buf, size_t uOverMalloc);

/** Used by derived classes for thier copy constructors
 * @protected
 * @memberof GseBlock
 */
GseBlock* GseBlock_copy(GseBlock* pThis, size_t uOverMalloc);

/** Same as GseBlock::del, used by derived classes 
 * @memberof GseBlock
 * @protected
 */
void delete_GseBlock(void* vpThis);

/** Same as GseBlock::encode, used by derived classes 
 * @memberof GseBlock
 */
bytebuf_t GseBlock_encode(GseBlock* pThis, byte uExtraLen, byte* pOffset);

/** Same as GseBlock::decode, used by derived classes
 * @memberof GseBlock
 * @protected 
 */
byte GseBlock_decode(GseBlock* pThis, cbytebuf_t buf);

/** Type definition of a block factory function. 
 *
 *  @see gse_generic_blocks() for more information on binary blob to 
 *  GseBlock generators.
 * 
 */
typedef GseBlock* (*gse_blk_factory)(int2 nSpaceCraft, byte uType, 
		                               byte uProcess, cbytebuf_t buf);

/** The default block factory function.
 *
 * Block factory functions are responsible for taking in blobs of bytes in
 * parsing them into GseBlock derived structures.  Each spacecraft project
 * defines it's own set of header blocks.
 *
 * This function only knows how to make GseBlock and GseProcBlk objects from
 * raw bytes.  This is useful for testing, and for handling unknown header
 * blocks.  It will accept data for any spacecraft ID.  Spacecraft
 * specific header block factory functions should always hand off any
 * un-recognized header blocks to this function for parsing.  All block
 * factories must have the function signiture defined by the @a typedef 
 * @ref gse_blk_factory.
 * 
 * @param nSpaceCraft The NAIF ID of the spacecraft in question.  Since this
 *        is a generic block factory function this parameter is ignored.
 * 
 * @param uType The block type, if this is not also be the first byte of
 *        @a buf most block factories will throw an error.  Will
 *        be stored in the member GseBlock::uType of the generated structure.
 *        field.
 *
 * @param uProcess If @a uType is @a 0x70, then a process block is indicated,
 *        if so then this determines which type of process.  If this is not
 *        also the second byte of a process block, most block factories will
 *        throw an error.
 *
 * @param buf A cbytebuf containing the bytes to parse into a GseBlock 
 *        derived structure.
 *
 * @return A pointer to a GseProcBlk structure if the block type is @a 0x70,
 *         otherwise a pointer to a GseBlock structure is returned.  In either
 *         case memory for the structure is allocated on the heap, and may
 *         be free'ed by calling the del member function.
 *
 * @throws (GSE_XCEPT, GSE_XDECODE) if there is an error decoding the 
 *         block bytes.
 */
GseBlock* gse_generic_blocks(int2 nSpaceCraft, byte uType, byte uProcess,
		                       cbytebuf_t buf);


/* An array of pointers to GseBlock structures */
#define S gse_blk_ary
#define C GseBlkAry
#define T GseBlock

/* Note, the included file has doxygen documentation that is automatically
   extracted via doc_filter.py at the top level */
#include "ptrAryTplt.h"

#undef S
#undef C
#undef T

/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseBlock_H_ */

