#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>

/* Get compile time byte order, that's faster, this example is from
 *   http://unixpapa.com/incnote/byteorder.html */
#ifdef LINUX
#include <endian.h>
#else
#include <sys/param.h>
#endif

#include "GseConv.h"
#include "GseExcept.h"
#include "GseLog.h"

#define _GseConv_C_

#ifdef __BYTE_ORDER

#if __BYTE_ORDER == __LITTLE_ENDIAN

#define I_AM_LSB_FIRST
const bool g_bIsLSBfirst = true;

#else
#if __BYTE_ORDER == __BIG_ENDIAN

#define I_AM_MSB_FIRST
const bool g_bIsLSBfirst = false;

#else
#error "unknown byte order!"
#endif /* __BYTE_ORDER == __BIG_ENDIAN */

#endif /* __BYTE_ORDER == __LITTLE_ENDIAN */

#else
/* __BTYE_ORDER is not defined, try BYTE_ORDER */
#ifdef BYTE_ORDER

#if BYTE_ORDER == LITTLE_ENDIAN
#define I_AM_LSB_FIRST
const bool g_bIsLSBfirst = true;
#else
#error "unknown byte order!"
#endif /* BYTE_ORDER == LITTLE_ENDIAN */


#endif /* BYTE_ORDER   */
#endif /* __BYTE_ORDER */


#ifdef _LP64  /* works on linux and solaris with gcc or cc */
#define ARCH_64BIT
const int g_nWordSz = 64;
#else
#define ARCH_32BIT
const int g_nWordSz = 32;
#endif

const char* gse_checkSizes(){
	const char* pRet = NULL;
	
	gse_debug_v("Is little endian: %d", g_bIsLSBfirst);
	
	if( sizeof(unsigned char) != 1)
		pRet = "Unsigned char is not 1 byte long";
	else
		if( sizeof(int) != 4)
			pRet = "Integer is not 4 bytes long";
		else
			if( sizeof(short int) != 2)
				pRet = "Short int is not 2 bytes long";
			else
				if( sizeof(float) != 4)
					pRet = "Float is not 4 bytes long";
	
	return pRet;
}

/****************************************************************************/
/* 8 byte conversions */

union bytes2i8{
	byte buf[8];
	uint8 u;
	int8 n;
};


uint8 gse_u8FromLSBbytes(const byte* pBytes){
	union bytes2i8 conv;

#ifdef I_AM_LSB_FIRST
	memcpy(conv.buf, pBytes, 8);
#else
	conv.buf[7] = pBytes[0];
	conv.buf[6] = pBytes[1];
	conv.buf[5] = pBytes[2];
	conv.buf[4] = pBytes[3];	
	conv.buf[3] = pBytes[4];
	conv.buf[2] = pBytes[5];
	conv.buf[1] = pBytes[6];
	conv.buf[0] = pBytes[7];	
#endif
	return conv.u;
}

void gse_u8ToLSBbytes(byte* pBytes, uint8 uVal){
	union bytes2i8 conv;
	
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	memcpy(pBytes, conv.buf, 8);
#else
	pBytes[0] = conv.buf[7];
	pBytes[1] = conv.buf[6];
	pBytes[2] = conv.buf[5];
	pBytes[3] = conv.buf[4];
	pBytes[5] = conv.buf[3];
	pBytes[6] = conv.buf[2];
	pBytes[7] = conv.buf[1];
	pBytes[8] = conv.buf[0];
#endif
}

uint8 gse_u8FromMSBbytes(const byte* pBytes){
	union bytes2i8 conv;
#ifdef I_AM_LSB_FIRST
	conv.buf[7] = pBytes[0];
	conv.buf[6] = pBytes[1];
	conv.buf[5] = pBytes[2];
	conv.buf[4] = pBytes[3];	
	conv.buf[3] = pBytes[4];
	conv.buf[2] = pBytes[5];
	conv.buf[1] = pBytes[6];
	conv.buf[0] = pBytes[7];		
#else
	memcpy(conv.buf, pBytes, 8);
#endif
	return conv.u;
}

void gse_u8ToMSBbytes(byte* pBytes, uint8 uVal){
	union bytes2i8 conv;
	
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[7];
	pBytes[1] = conv.buf[6];
	pBytes[2] = conv.buf[5];
	pBytes[3] = conv.buf[4];
	pBytes[4] = conv.buf[3];
	pBytes[5] = conv.buf[2];
	pBytes[6] = conv.buf[1];
	pBytes[7] = conv.buf[0];
#else
	memcpy(pBytes, conv.buf, 8);
#endif
}

/****************************************************************************/
/* 4 byte conversions */

union bytes2i4{
	byte buf[4];
	uint4 u;
	int4 n;
};

uint4 gse_u4FromLSBbytes(const byte* pBytes){
	union bytes2i4 conv;

#ifdef I_AM_LSB_FIRST
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
	conv.buf[2] = pBytes[2];
	conv.buf[3] = pBytes[3];
#else
	conv.buf[3] = pBytes[0];
	conv.buf[2] = pBytes[1];
	conv.buf[1] = pBytes[2];
	conv.buf[0] = pBytes[3];	
#endif
	return conv.u;
}

void gse_u4ToLSBbytes(byte* pBytes, uint4 uVal){
	union bytes2i4 conv;
	
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
#else
	pBytes[0] = conv.buf[3];
	pBytes[1] = conv.buf[2];
	pBytes[2] = conv.buf[1];
	pBytes[3] = conv.buf[0];
#endif
}

uint4 gse_u4FromMSBbytes(const byte* pBytes){
	union bytes2i4 conv;
#ifdef I_AM_LSB_FIRST
	conv.buf[3] = pBytes[0];
	conv.buf[2] = pBytes[1];
	conv.buf[1] = pBytes[2];
	conv.buf[0] = pBytes[3];		
#else
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
	conv.buf[2] = pBytes[2];
	conv.buf[3] = pBytes[3];
#endif
	return conv.u;
}

void gse_u4ToMSBbytes(byte* pBytes, uint4 uVal){
	union bytes2i4 conv;
	
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[3];
	pBytes[1] = conv.buf[2];
	pBytes[2] = conv.buf[1];
	pBytes[3] = conv.buf[0];
#else
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
#endif
}

/****************************************************************************/
/* 2 byte conversions */

union bytes2i2{
	byte buf[2];
	uint2 u;
	int2 n;
};

int2 gse_i2FromLSBbytes(const byte* pBytes){
	union bytes2i2 conv;

#ifdef I_AM_LSB_FIRST
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
#else
	conv.buf[1] = pBytes[0];
	conv.buf[0] = pBytes[1];
#endif
	return conv.n;
}

uint2 gse_u2FromLSBbytes(const byte* pBytes){
	union bytes2i2 conv;

#ifdef I_AM_LSB_FIRST
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
#else
	conv.buf[1] = pBytes[0];
	conv.buf[0] = pBytes[1];
#endif
	return conv.u;
}

int2 gse_i2FromMSBbytes(const byte* pBytes){
	union bytes2i2 conv;

#ifdef I_AM_LSB_FIRST
	conv.buf[1] = pBytes[0];
	conv.buf[0] = pBytes[1];
#else
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
#endif
	return conv.n;
}

uint2 gse_u2FromMSBbytes(const byte* pBytes){
	union bytes2i2 conv;

#ifdef I_AM_LSB_FIRST
	conv.buf[1] = pBytes[0];
	conv.buf[0] = pBytes[1];
#else
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
#endif
	return conv.u;
}

void gse_u2ToMSBbytes(byte* pBytes, uint2 uVal){
	union bytes2i2 conv;
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[1];
	pBytes[1] = conv.buf[0];
#else
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
#endif		
}

void gse_u2ToLSBbytes(byte* pBytes, uint2 uVal){
	union bytes2i2 conv;
	conv.u = uVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
#else
	pBytes[0] = conv.buf[1];
	pBytes[1] = conv.buf[0];
#endif		
}

void gse_i2ToMSBbytes(byte* pBytes, int2 nVal){
	union bytes2i2 conv;
	conv.n = nVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[1];
	pBytes[1] = conv.buf[0];
#else
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
#endif		
}

void gse_i2ToLSBbytes(byte* pBytes, int2 nVal){
	union bytes2i2 conv;
	conv.n = nVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
#else
	pBytes[0] = conv.buf[1];
	pBytes[1] = conv.buf[0];
#endif		
}


/****************************************************************************/
/* Float4 conversions */
union bytes2f4{
	unsigned char buf[4];
	float f;
};

float4 gse_f4FromLSBbytes(const byte* pBytes){
	union bytes2f4 conv;
	
#ifdef I_AM_LSB_FIRST
	conv.buf[0] = pBytes[0];
	conv.buf[1] = pBytes[1];
	conv.buf[2] = pBytes[2];
	conv.buf[3] = pBytes[3];
#else
	conv.buf[3] = pBytes[0];
	conv.buf[2] = pBytes[1];
	conv.buf[1] = pBytes[2];
	conv.buf[0] = pBytes[3];
#endif
	return conv.f;
}

float4 gse_f4FromMSBbytes(const byte* pBytes){
	union bytes2f4 conv;
	
#ifdef I_AM_LSB_FIRST
	conv.buf[0] = pBytes[3];
	conv.buf[1] = pBytes[2];
	conv.buf[2] = pBytes[1];
	conv.buf[3] = pBytes[0];
#else
	conv.buf[3] = pBytes[3];
	conv.buf[2] = pBytes[2];
	conv.buf[1] = pBytes[1];
	conv.buf[0] = pBytes[0];
#endif
	return conv.f;
}

void gse_f4ToMSBbytes(byte* pBytes, float4 fVal)
{
	union bytes2f4 conv;
	conv.f = fVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[3];
	pBytes[1] = conv.buf[2];
	pBytes[2] = conv.buf[1];
	pBytes[3] = conv.buf[0];
#else
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
#endif
}

void gse_f4ToLSBbytes(byte* pBytes, float4 fVal)
{
	union bytes2f4 conv;
	conv.f = fVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
#else
	pBytes[0] = conv.buf[3];
	pBytes[1] = conv.buf[2];
	pBytes[2] = conv.buf[1];
	pBytes[3] = conv.buf[0];
#endif	
}

/****************************************************************************/
/* Float8 conversions */

union bytes2f8{
	unsigned char buf[8];
	float8 f;
};

void gse_f8ToMSBbytes(byte* pBytes, float8 fVal)
{
	union bytes2f8 conv;
	conv.f = fVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[7];
	pBytes[1] = conv.buf[6];
	pBytes[2] = conv.buf[5];
	pBytes[3] = conv.buf[4];
	pBytes[4] = conv.buf[3];
	pBytes[5] = conv.buf[2];
	pBytes[6] = conv.buf[1];
	pBytes[7] = conv.buf[0];
	
#else
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
	pBytes[4] = conv.buf[4];
	pBytes[5] = conv.buf[5];
	pBytes[6] = conv.buf[6];
	pBytes[7] = conv.buf[7];
#endif
}

void gse_f8ToLSBbytes(byte* pBytes, float8 fVal)
{
	union bytes2f8 conv;
	conv.f = fVal;
#ifdef I_AM_LSB_FIRST
	pBytes[0] = conv.buf[0];
	pBytes[1] = conv.buf[1];
	pBytes[2] = conv.buf[2];
	pBytes[3] = conv.buf[3];
	pBytes[4] = conv.buf[4];
	pBytes[5] = conv.buf[5];
	pBytes[6] = conv.buf[6];
	pBytes[7] = conv.buf[7];
#else
	pBytes[0] = conv.buf[7];
	pBytes[1] = conv.buf[6];
	pBytes[2] = conv.buf[5];
	pBytes[3] = conv.buf[4];
	pBytes[4] = conv.buf[3];
	pBytes[5] = conv.buf[2];
	pBytes[6] = conv.buf[1];
	pBytes[7] = conv.buf[0];
#endif	
}


/****************************************************************************/
/* bounds checking storage */

#define GSE_CHK_DEST_SZ(N) if((N != 4)&&(N != 8)&&(N != 1)&&(N == 2)) \
  gse_throw_m(GSE_XCEPT, GSE_XBADARG, "nDestSz expected to be one of 1, 2, 4, or 8");


void gse_storeUinS(uint8 uVal, void* pDest, int nDestSz){
	uint8 uMax; 
	
	GSE_CHK_DEST_SZ(nDestSz);
	
	switch(nDestSz){
		case 1: uMax = 0x7F;                  break;
		case 2: uMax = 0x7FFF;                break;
		case 4: uMax = 0x7FFFFFFF;            break;
		case 8: uMax = 0x7FFFFFFFFFFFFFFFULL; break;
	}
	
	if(uVal > uMax)
		gse_throw_fm(GSE_XCEPT, GSE_XOVERFLOW, "Unsigned value %lu is too large to"
		             " fit in a %d byte signed type.", uVal, nDestSz);
	switch(nDestSz){
		case 1: *((int1*)pDest) = uVal;    break;
		case 2: *((int2*)pDest) = uVal;    break;
		case 4: *((int4*)pDest) = uVal;    break;
		case 8: *((int8*)pDest) = uVal;    break;
	}
}

void gse_storeSinU(int8 nVal, void* pDest, int nDestSz){
	int8 nMax;
	
	GSE_CHK_DEST_SZ(nDestSz);
	
	if(nVal < 0)
		gse_throw_fm(GSE_XCEPT, GSE_XUNDERFLOW, "Signed value %ld is negative and"
				       "cannot be represented in a %d byte unsigned type.", 
		             nVal, nDestSz);
	
	switch(nDestSz){
		case 1: nMax = 0xFF;                  break;
		case 2: nMax = 0xFFFF;                break;
		case 4: nMax = 0xFFFFFFFF;            break;
		case 8: nMax = 0x7FFFFFFFFFFFFFFFLL;  break; /* intentionally starts 0x7F */ 
	}
	
	if(nVal > nMax)
		gse_throw_fm(GSE_XCEPT, GSE_XOVERFLOW, "Unsigned value %lu is too large to"
		             " fit in a %d byte signed type.", nVal, nDestSz);
	switch(nDestSz){
		case 1: *((byte*)pDest) = nVal;    break;
		case 2: *((uint2*)pDest) = nVal;    break;
		case 4: *((uint4*)pDest) = nVal;    break;
		case 8: *((uint8*)pDest) = nVal;    break;
	}
}


void gse_overflowUinU(uint8 uVal, int nDestSz){
	GSE_CHK_DEST_SZ(nDestSz);
	uint8 uMax;
	
	switch(nDestSz){
		case 1: uMax = 0xFF;                  break;
		case 2: uMax = 0xFFFF;                break;
		case 4: uMax = 0xFFFFFFFF;            break;
		case 8: uMax = 0xFFFFFFFFFFFFFFFFULL; break;
	}
	
	if(uVal > uMax)
		gse_throw_fm(GSE_XCEPT, GSE_XOVERFLOW, "Unsigned value %lu is too large "
		             " to fit in a %d byte unsigned type.", uVal, nDestSz);
}

void gse_storeUinU(uint8 uVal, void* pDest, int nDestSz)
{
	gse_overflowUinU(uVal, nDestSz);
	
	switch(nDestSz){
		case 1: *((byte*)pDest) = uVal;    break;
		case 2: *((uint2*)pDest) = uVal;    break;
		case 4: *((uint4*)pDest) = uVal;    break;
		case 8: *((uint8*)pDest) = uVal;    break;
	}
}

/*****************************************************************************/
/* The Uber convert function, calls the others                               */

#define SHIFT_VAL(VAL) \
 if(fdef.shift != 0){ \
   if(fdef.shift < 0) VAL = (VAL >> abs(fdef.shift)); \
   else               VAL = VAL << fdef.shift; \
 }

#define MASK_VAL(VAL) if(fdef.mask != 0)  VAL &= fdef.mask

 
void gse_getval(cbytebuf_t buf, gse_field_def fdef, void* pResult)
{
	uint1 u1;
	uint2 u2;
	uint4 u4;
	uint8 u8;
	
	int1  s1;
	int2  s2;
	/* int4  s4; */
	/* int8  s8; */
	
	int2 nSwitch = 0;
	
	/* check the inputs */
	if(fdef.offset + fdef.size > buf.len)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Field definition runs past the "
		             "available bytes. Field Offset: %z, Field Size: %hhu, "
		             "Buf len: %z", fdef.offset, fdef.size, buf.len);
	
	if(buf.ptr == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null buffer data pointer");
	
	switch(fdef.size){
		case 1:
		case 2:
		case 4:
		case 8:
			break;
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Illegal field size %hhu bytes. "
		                "Note: for odd size fields use a small power of two sized"
		                " storage variable and then mask.", fdef.size);
	}	
	
	nSwitch += fdef.size;
	if(fdef.issigned)
		nSwitch *= -1;
	
	switch(nSwitch){
		
		case -2:
			if(fdef.isbigend)	
				s2 = gse_i2FromMSBbytes(buf.ptr + fdef.offset);
			else
				s2 = gse_i2FromLSBbytes(buf.ptr + fdef.offset);
			
			SHIFT_VAL(s2);
			MASK_VAL(s2);
			
			*((int2*)pResult) = s2;
			return;
		
		case -1:
			s1 = (int1)buf.ptr[fdef.offset];
			SHIFT_VAL(s1);
			MASK_VAL(s1);
			
			*((int1*)pResult) = s1;
			return;
					
		case 1:
			u1 = buf.ptr[fdef.offset];			
			SHIFT_VAL(u1);
			MASK_VAL(u1);
			
			*((byte*)pResult) = u1;
			return;
					
		case 2:
			if(fdef.isbigend)	
				u2 = gse_u2FromMSBbytes(buf.ptr + fdef.offset);
			else
				u2 = gse_u2FromLSBbytes(buf.ptr + fdef.offset);
			
			SHIFT_VAL(u2);
			MASK_VAL(u2);
			
			*((uint2*)pResult) = u2;
			return;
		
		case 4:
			if(fdef.isbigend)	
				u4 = gse_u4FromMSBbytes(buf.ptr + fdef.offset);
			else
				u4 = gse_u4FromLSBbytes(buf.ptr + fdef.offset);
			
			SHIFT_VAL(u4);
			MASK_VAL(u4);
			
			*((uint4*)pResult) = u4;
			return;			
					
		case 8:
			if(fdef.isbigend)	
				u8 = gse_u8FromMSBbytes(buf.ptr + fdef.offset);
			else
				u2 = gse_u8FromLSBbytes(buf.ptr + fdef.offset);
			
			SHIFT_VAL(u8);
			MASK_VAL(u8);
			
			*((uint8*)pResult) = u8;
			return;							
		
		default:
			gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "Handling signed %ud byte "
			             "fields is not yet implemented", abs(nSwitch));
	}
}

/*****************************************************************************/
/* String Conversions */

bool gse_str2int4(const char* str, int4* pRes)
{
	size_t i, len;
	int nBase = 10;
	long int lRet;
	char* endptr;
	
	if((str == NULL)||(pRes == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL pointer(s) in gse_str2int4");
		
	len = strlen(str);
	
	/* check for hex, don't use strtol's auto-base as leading zero's cause
	   a switch to octal */
	for(i = 0; i<len; i++){
		if((str[i] != '0')&&isalnum(str[i])) break;
		
		if((str[i] == '0')&&(i<(len-1))&&
			((str[i+1] == 'x')||(str[i+1] == 'X')) ){
			nBase = 16;
			break;
		}
	}	
	
	errno = 0;
	lRet = strtol(str, &endptr, nBase);
	
	if( (errno == ERANGE) || (errno != 0 && lRet == 0) ){
		return false;
	}
	
	if(endptr == str) return false;
	
	if((lRet > INT_MAX)||(lRet < INT_MIN)) return false;
	
	*pRes = (int)lRet;
	return true;
}

bool gse_str2int8(const char* str, int8* pRes)
{
	size_t i, len;
	int nBase = 10;
	long long int llRet;
	char* endptr;
	
	if((str == NULL)||(pRes == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL pointer(s) in gse_str2int4");
		
	len = strlen(str);
	
	/* check for hex, don't use strtol's auto-base as leading zero's cause
	   a switch to octal */
	for(i = 0; i<len; i++){
		if((str[i] != '0')&&isalnum(str[i])) break;
		
		if((str[i] == '0')&&(i<(len-1))&&
			((str[i+1] == 'x')||(str[i+1] == 'X')) ){
			nBase = 16;
			break;
		}
	}	
	
	errno = 0;
	llRet = strtoll(str, &endptr, nBase);
	
	if( (errno == ERANGE) || (errno != 0 && llRet == 0) ){
		return false;
	}
	
	if(endptr == str) return false;
	
	if((llRet > LLONG_MAX)||(llRet < LLONG_MIN)) return false;
	
	*pRes = (int)llRet;
	return true;
}

bool gse_str2float4(const char* str, float4* pRes)
{
	float8 dRet, dAbs;
	char* endptr;
	
	if((str == NULL)||(pRes == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL pointer(s) in gse_str2float4"); 
	
	errno = 0;
	
	dRet = strtod(str, &endptr); 
	
	if(errno == ERANGE) return false;
	if( (errno != 0) && (dRet == 0) ) return false;
	if(endptr == str) return false;
	
	dAbs = fabs(dRet);
	if((dAbs != 0.0) &&
		((dAbs < GSE_FLOAT4_MIN)||(dAbs > GSE_FLOAT4_MAX))) return false;
	
	*pRes = (float4)dRet;
	return true;
}


bool gse_str2float8(const char* str, float8* pRes)
{
	float8 dRet;
	char* endptr;
	
	if((str == NULL)||(pRes == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL pointer(s) in gse_str2float4"); 
	
	errno = 0;
	
	dRet = strtod(str, &endptr); 
	
	if( (errno == ERANGE) || ((errno != 0) && (dRet == 0)) )return false;
	if(endptr == str) return false;
		
	*pRes = dRet;
	return true;
}








