/** \file GseConv.h Binary data conversion utilities */
#ifndef _GseConv_h_
#define _GseConv_h_

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/
/** @defgroup mod_conversion Binary Conversions
 *
 * C-Numeric Types to and from Byte Strings. @n
 * @todo This module makes heavy use of unions, these may not be portable.
 */
/*@{*/

#ifndef _GseConv_C_
/** Set at compile time to @c true if this is a LSB first machine. */
extern const bool g_bIsLSBfirst;

/** Get the word size examples include, 64, 32, 16, 8. */
extern const int g_nWordSz;
#endif

/** Checks to see if character and integer sizes are what we think they are.
  * This is called as part of library initialization in gse_thread_init().
  *
  * @return a pointer to a null terminated error string if the sizes are 
  *  unexpected, 0 is returned otherwise.   
  */
const char* gse_checkSizes(void);

/** Definiting how to pull a field out of a chunk of bytes */
typedef struct _gse_fieled_def{
	size_t offset;
	byte    size;
	bool    isbigend;
	bool    issigned;
	int2    shift;
	uint8   mask;
}gse_field_def;

/** Read a field out of a chunk of bytes
 * 
 * @param buf The buffer from which to extract the bytes.
 * @param fdef a field description structure.
 * @param pResult A void pointer to the location that is to store the converted 
 *        result.  This location must be at least @a fsize bytes long.
 *        For correct reading it should also have the same signedness.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) for a number of reasons.
 */
void gse_getval(cbytebuf_t buf, gse_field_def fdef, void* pResult);


/** Convert a string value to an 4 byte integer, similar to strtol(3).
 *
 * @param str the string to convert.  Conversion stops at the first improper
 *        character.  Whitespace and leading 0's are ignored in the input.
 *        The numder is assumed to be in base 10, unless the first non-whitespace
 *        characters after the optional '+' or '-' sign are '0x'.
 *
 * @param pRes The location to store the resulting 4-byte integer.
 *
 * @returns @c true if the conversion succeeded, @c false otherwise.  Among 
 *        other reason, conversion will fail if the resulting integer won't fit
 *        in a 4 byte integer.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if either of the arguments are NULL. 
 */
bool gse_str2int4(const char* str, int4* pRes);


/** Convert a string value to an 8 byte integer, similar to strtoll(3).
 *
 * @param str the string to convert.  Conversion stops at the first improper
 *        character.  Whitespace and leading 0's are ignored in the input.
 *        The numder is assumed to be in base 10, unless the first non-whitespace
 *        characters after the optional '+' or '-' sign are '0x'.
 *
 * @param pRes The location to store the resulting 8-byte integer.
 *
 * @returns @c true if the conversion succeeded, @c false otherwise.  Among 
 *        other reason, conversion will fail if the resulting integer won't fit
 *        in a 4 byte integer.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if either of the arguments are NULL. 
 */
bool gse_str2int8(const char* str, int8* pRes);


/** Convert a string value to a 4-byte float, similar to strtod(3).
 *
 * @param str the string to convert.  Conversion stops at the first improper
 *        character.  Whitespace and leading 0's are ignored in the input.
 *        The numder is assumed to be in base 10, unless the first non-whitespace
 *        characters after the optionas '+' or '-' sign are '0x'.
 *
 * @param pRes The location to store the resulting 4-byte float.
 *
 * @returns @c true if the conversion succeeded, @c false otherwise.  Among 
 *        other reason, conversion will fail if the resulting value won't fit
 *        in a 4 byte float.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if either of the arguments are NULL. 
 */
bool gse_str2float4(const char* str, float4* pRes);

/** Convert a string value to a 8-byte float, similar to strtod(3).
 *
 * @param str the string to convert.  Conversion stops at the first improper
 *        character.  Whitespace and leading 0's are ignored in the input.
 *        The numder is assumed to be in base 10, unless the first non-whitespace
 *        characters after the optionas '+' or '-' sign are '0x'.
 *
 * @param pRes The location to store the resulting 8-byte float.
 *
 * @returns @c true if the conversion succeeded, @c false otherwise.  Among 
 *        other reason, conversion will fail if the resulting integer won't fit
 *        in a 8 byte float.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if either of the arguments are NULL. 
 */
bool gse_str2float8(const char* str, float8* pRes);


/** Convert eight unsigned bytes in MSB first order to an unsigned integer.
 *
 * Assumes that integers are 4-bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 8 bytes.
 * @return the unsigned integer.
 */
uint8 gse_u8FromMSBbytes(const byte* pBytes);

/** Convert four unsigned bytes in LSB first order to an unsigned integer. 
 *
 * Assumes that integers are 4-bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 4 bytes.
 * @return the unsigned integer.
 */
uint8 gse_u8FromLSBbytes(const byte* pBytes);

/** Write out an 8 byte unsigned integer in MSB first order 
 *
 * @param pBytes  Eight bytes are writtin to this address
 * @param uVal The value to encode.
 */ 
void gse_u8ToMSBbytes(byte* pBytes, uint8 uVal);

/** Write out an 8 byte unsigned integer in LSB first order 
 *
 * @param pBytes  Eight bytes are writtin to this address
 * @param uVal The value to encode.
 */ 
void gse_u8ToLSBbytes(byte* pBytes, uint8 uVal);


/** Convert four unsigned bytes in LSB first order to an unsigned integer. 
 *
 * Assumes that integers are 4-bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 4 bytes.
 * @return the unsigned integer.
 */
uint4 gse_u4FromLSBbytes(const byte* pBytes);

/** Convert four unsigned bytes in MSB first order to an unsigned integer.
 *
 * Assumes that integers are 4-bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 4 bytes.
 * @return the unsigned integer.
 */
uint4 gse_u4FromMSBbytes(const byte* pBytes);


/** Write out a 4 byte unsigned integer in MSB first order 
 *
 * @param pBytes  Four bytes are writtin to this address
 * @param uVal The value to encode.
 */ 
void gse_u4ToMSBbytes(byte* pBytes, uint4 uVal); 


/** Write out a 4 byte unsigned integer in LSB first order 
 *
 * @param pBytes  Four bytes are written to this address
 * @param uVal The value to encode.
 */ 
void gse_u4ToLSBbytes(byte* pBytes, uint4 uVal); 


/** Convert two unsigned bytes in LSB first order to a signed short integer.
 *
 * Assumes that short integers are two bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 2 bytes
 * @return the signed short integer
 */
int2 gse_i2FromLSBbytes(const byte* pBytes);

/** Convert two unsigned bytes in LSB first order to an unsigned short integer.
 *
 * Assumes that short integers are two bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 2 bytes
 * @return the signed short integer
 */
uint2 gse_u2FromLSBbytes(const byte* pBytes);

/** Convert two unsigned bytes in MSB first order to a signed short integer.
 *
 * Assumes that short integers are two bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 2 bytes
 * @return the signed short integer
 */
int2 gse_i2FromMSBbytes(const byte* pBytes);

/** Convert two unsigned bytes in MSB first order to an unsigned short integer.
 *
 * Assumes that short integers are two bytes long, check with gse_checkSizes().
 * @param pBytes a pointer to 2 bytes
 * @return the signed short integer
 */
uint2 gse_u2FromMSBbytes(const byte* pBytes);

/** Write out a 2 byte unsigned integer in MSB first order 
 *
 * @param pBytes two bytes are writtin to this address
 * @param uVal The value to encode.
 */ 
void gse_u2ToMSBbytes(byte* pBytes, uint2 uVal); 

/** Write out a 2 byte unsigned interger in LSB first order
 *
 * @param pBytes two bytes are writtin to this address
 * @param uVal The value to encode.
 */ 
void gse_u2ToLSBbytes(byte* pBytes, uint2 uVal); 

/** Write out a 2 byte signed integer in MSB first order 
 *
 * @param pBytes two bytes are writtin to this address
 * @param nVal The value to encode.
 */ 
void gse_i2ToMSBbytes(byte* pBytes, int2 nVal); 

/** Write out a 2 byte signed integer in LSB first order 
 *
 * @param pBytes two bytes are writtin to this address
 * @param nVal The value to encode.
 */ 
void gse_i2ToLSBbytes(byte* pBytes, int2 nVal); 


/** Convert four unsigned bytes in LSB first order to a float.
 *
 * Assumes that floats are 4 bytes long, check with gse_checkSizes()
 * @param pBytes a pointer to a byte array least 4 items long.
 * @return the float
 */
float4 gse_f4FromLSBbytes(const byte* pBytes);

/** Convert four unsigned bytes in MSB first order to a float.
 *
 * Assumes that floats are 4 bytes long, check with gse_checkSizes()
 * @param pBytes a pointer to a byte array least 4 items long.
 * @return the float
 */
float4 gse_f4FromMSBbytes(const byte* pBytes);

/** Write out a 4 byte float in LSB first order.
 *
 * @param pBytes a pointer to a byte array at least 4 items long.
 * @param fVal The value to encode.
 */
void gse_f4ToLSBbytes(byte* pBytes, float4 fVal);

/** Write out a 4 byte float in MSB first order.
 *
 * @param pBytes a pointer to a byte array at least 4 items long.
 * @param fVal The value to encode.
 */
void gse_f4ToMSBbytes(byte* pBytes, float4 fVal);

/** Write out a 4 byte float in LSB first order.
 *
 * @param pBytes a pointer to a byte array at least 4 items long.
 * @param fVal The value to encode.
 */
void gse_f8ToLSBbytes(byte* pBytes, float8 fVal);

/** Write out a 4 byte float in MSB first order.
 *
 * @param pBytes a pointer to a byte array at least 4 items long.
 * @param fVal The value to encode.
 */
void gse_f8ToMSBbytes(byte* pBytes, float8 fVal);

/** Store an unsigned value in a signed type with overflow checking.
 *
 * @param uVal can be any value storable in an 8 byte unsigned integer
 * @param nDestSz the sizeof the signed type that is to recieve the value.
 * @param pDest (output) a pointer to a signed integer nDestSz bytes wide
 *        to recieve the value.
 * @throws (GSE_XCEPT, GSE_XOVERFLOW) if uVal is too large to fit
 *        without truncation and/or sign bit wrapping.
 */
void gse_storeUinS(uint8 uVal, void* pDest, int nDestSz);

/** Store an unsigned value in a signed type with overflow checking.
 *
 * @param uVal can be any value storable in an 8 byte unsigned integer
 * @param nDestSz the sizeof the signed type that is to recieve the value.
 * @param pDest (output) a pointer to an unsigned integer nDestSz bytes wide
 *        to recieve the value.
 * @throws (GSE_XCEPT, GSE_XOVERFLOW) if uVal is too large to fit
 *        without truncation and/or sign bit wrapping.
 */
void gse_storeUinU(uint8 uVal, void* pDest, int nDestSz);

/** Overflow check for storing an unsigned value in an unsigned variable. */
void gse_overflowUinU(uint8 uVal, int nDestSz);

/** Store an unsigned value in a signed type with overflow checking.
 *
 * @param nVal can be any value storable in an 8 byte unsigned integer
 * @param nDestSz the sizeof the signed type that is to recieve the value.
 * @param pDest (output) a pointer to an unsigned integer nDestSz bytes
 *        wide to recieve the value. 
 * @throws (GSE_XCEPT, GSE_XOVERFLOW) if uVal is too large to fit
 *        without truncation and/or sign bit wrapping.
 * @throws (GSE_XCEPT, GSE_XUNDERFLOW) if the sign is negative.
 */
void gse_storeSinU(int8 nVal, void* pDest, int nDestSz);

/** Store an signed value in a signed type with overflow checking.
 *
 * @param nVal can be any value storable in an 8 byte unsigned integer
 * @param nDestSz the sizeof the signed type that is to recieve the value.
 * @param pDest (output) a pointer to a signed integer nDestSz bytes
 *        wide to recieve the value.
 * @throws (GSE_XCEPT, GSE_XOVERFLOW) if the sizeof the dest is too small
 *         to hold the positive value. 
 * @throws (GSE_XCEPT, GSE_XUNDERFLOW) if the sizeof the dest is too small 
 *        to hold the negative value.
 */
void gse_storeSinS(int8 nVal, void* pDest, int nDestSz);



/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /*GseConv_h_*/
