/** @file GseData.h */
#ifndef _GseData_H_
#define _GseData_H_

#include <time.h>

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_data Generic Datasets
 *
 * An interface for display and analysis tools
 */
/*@}*/

/** @addtogroup mod_data */
/*@{*/

/** An enumeration of data types, to be grown as needed */
typedef enum {
	/** Unitless numbers */
	NUMBER, 
			
	/** Unitless numbers as a log ratio */
	DB,
			
	/** 1/DB */
	INV_DB,
			
	/** Time values */
	TIME, 
	
	/** Inverse time values */
	FREQUENCY,
	
	/** Spectral Density, usually (V/m)^2 / Hz */
	POWER_PER_HZ,
	
	/** Inv Spectral Density, not used often, usually (V/m)^2 / sec */
	POWER_PER_SEC
	
} gse_unittype_t;

/** Get a string name for the data type */
const char* gse_str_unittype(gse_unittype_t type);

/** Get the inverse type for a given type */
gse_unittype_t gse_unittype_invert(gse_unittype_t type);

/** Given a unit string, try to make its inverse unit string in a sane way.
 * @return a new unit string allocated on the heap.
 */ 
char* gse_invert_units(const char* sUnit);


#define _GSE_HNT_SET_X_LOG 0x00000400
#define _GSE_HNT_SET_Y_LOG 0x00000800

#define GSE_HNT_X_SIG(x) (x & 0x0000001F)
#define GSE_HNT_Y_SIG(x) ((x & 0x000003D0) >> 5)
#define GSE_HNT_X_LOG(x) ((x & _GSE_HNT_SET_X_LOG) >> 10)
#define GSE_HNT_Y_LOG(x) ((x & _GES_HNT_SET_Y_LOG) >> 11)

	
/** A source neutral set of XY data
 *
 * This is defined as an abstract class, instead of an interface because it 
 * has a destructor.@n
 *
 * @class GseXYdata
 */
typedef struct gse_xy_data{

	/** A short, but human readable source definition, must be unique for
	 * each instrument capture type */
	const char*  (*channel)(const struct gse_xy_data* pThis);
	
	/** A longer version of the channel name */
	const char*  (*channelName)(const struct gse_xy_data* pThis);
	
	/** Information about the data channel that produced this data sample */
	const char*  (*channelInfo)(const struct gse_xy_data* pThis);
	
	/** The upper limit for Y data from this channel */
	float4  (*yHiLimit)(const struct gse_xy_data* pThis);
	
	/** The lower limit for Y data from this channel */
	float4  (*yLoLimit)(const struct gse_xy_data* pThis);
	
	/** The UTC time at which capture of this data sample was initiated.
	 * @param pThis Structure pointer.
	 * @param pTm Will be filled in the broken down time see gm_time(2)
	 * @param pNano Will be filled in with the nanoseconds of the 
	 *        capture.
	 */
	void (*captureUTC)(const struct gse_xy_data* pThis,
	                   struct tm* pTm, uint4* pNano);
	
	/** The UTC time at which capture of this data sample ended 
	 * This is an @b exclusive upper bound 
	 */
	const char* (*captureSCLK)(const struct gse_xy_data* pThis);
	
	/** The length of both the X and Y arrays */
	size_t  (*length)(const struct gse_xy_data* pThis);
	
	/** The X points for the sample */
	const float4* (*Xdata)(const struct gse_xy_data* pThis);
	
	/** X axis data type */
	gse_unittype_t (*Xtype)(const struct gse_xy_data* pThis);
	
	/** The base unit for the X axis, ex. 'm', 's', 'Hz', 'MHz' */
	const char*   (*Xunit)(const struct gse_xy_data* pThis);
	
	/** The Y points for the sample */
	const float4* (*Ydata)(const struct gse_xy_data* pThis);
	
	/** The base unit for the Y axis, ex. 'count', 'V^2/m^2/Hz' */
	const char*   (*Yunit)(const struct gse_xy_data* pThis);
	
	/** Y axis data type */
	gse_unittype_t (*Ytype)(const struct gse_xy_data* pThis);
	
	/** Hints, an or'ed set of flags suggesting display and/or formatting hints.
	 * 
	 * Currently the following test macros for hints are defined:
	 *
	 *   - GSE_HNT_X_LOG(hint) - X Not zero if X data is log spaced
	 *   - GSE_HNT_Y_LOG(hint) - Y Not zero if Y data is log spaced
	 *   - GSE_HNT_X_SIG(hint) - X number of X significant digits, 0 if not defined
	 *   - GSE_HNT_Y_SIG(hint) - Y number of Y significant digits, 0 if not defined
	 */
	uint4 (*hints)(const struct gse_xy_data* pThis);
	
	/** Copy constructor */
	struct gse_xy_data* (*copy)(const struct gse_xy_data* pThis);
	
	/** The destructor */	
	void (*del)(void* vpThis);
	
}GseXYdata;

/* An array of pointers to GseXYdata structures */
#define S gse_xydata_ary
#define C GseXYdataAry
#define T GseXYdata

#include "ptrAryTplt.h"

#undef S
#undef C
#undef T

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseData_H_ */
