/* Extenstions to Kaz Kylheku exception library to add traceback
 * capibility.  This is not portable ANSI C, nor can it be */

#define _POSIX_C_SOURCE 200112

#if defined(__linux)
#define _GNU_SOURCE
#include <execinfo.h>
#include <dlfcn.h>

#elif defined(__MINGW32__)
#include "GseWinSvc.h"

#endif

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>

#include <netdb.h>  /* Does windows have this? */

#include "GseExcept.h"
#include "GseWrap.h"

#define MAX_STACK_DEPTH 256
#define MAX_EXCEPT_MSG_SZ 1024


#ifdef GSE_POSIX_THREADS /* IS THREADED */

static pthread_mutex_t s_mtx_init = PTHREAD_MUTEX_INITIALIZER;
static int s_nThreads = 0;
static pthread_key_t s_keyBacktrace;
static pthread_key_t s_keyFileNames;
static pthread_key_t s_keyFunctions;


bool gse_except_init()
{
	bool BackOk, FNOk, FuncOk, bRet = true;
	 
	if(except_init() != 1) return false;
	
	pthread_mutex_lock(&s_mtx_init);
	
	s_nThreads += 1;
	
	if(s_nThreads == 1){
		BackOk = (pthread_key_create(&s_keyBacktrace, free) == 0);
		FNOk = (pthread_key_create(&s_keyFileNames, free) == 0);
		FuncOk = (pthread_key_create(&s_keyFunctions, free) == 0);
		bRet = BackOk && FNOk && FuncOk;
		
		if(!bRet){
			s_nThreads -= 1;
			if(BackOk) pthread_key_delete(s_keyBacktrace);
			if(FNOk)   pthread_key_delete(s_keyFileNames);
			if(FuncOk) pthread_key_delete(s_keyFunctions);
		}
	}
	
	pthread_mutex_unlock(&s_mtx_init);
	
	except_unhandled_catcher(gse_unhandled_catcher);
	except_set_allocator(malloc, gse_except_data_free);
	return bRet;
}

void gse_except_deinit()
{
	pthread_mutex_lock(&s_mtx_init);
	
	if(s_nThreads > 0){
		s_nThreads -= 1;
		
		if(s_nThreads == 0){
			pthread_key_delete(s_keyBacktrace);
			pthread_key_delete(s_keyFileNames);
			pthread_key_delete(s_keyFunctions);
		}
	}
	
	pthread_mutex_unlock(&s_mtx_init);
	except_deinit();
}

/* Make backtrace buffer for this thread if it doesn't exist */
void** get_BacktraceBuf(){
	void** pBuf = (void**) pthread_getspecific(s_keyBacktrace);
	if(pBuf == NULL){
		pBuf = (void**)calloc(MAX_STACK_DEPTH, sizeof(void*));
		pthread_setspecific(s_keyBacktrace, (void*)pBuf);
	}
	return pBuf;
}

/* Make file name array for this thread if it doesn't exist */
const char** get_FileNames(){
	const char** pAry = (const char**) pthread_getspecific(s_keyFileNames);
	if(pAry == NULL){
		pAry = (const char**)calloc(MAX_STACK_DEPTH, sizeof(const char**));
		pthread_setspecific(s_keyFileNames, (void*)pAry);
	}
	return pAry;
}

/* Make function name array for this thread if it doesn't exist */
const char** get_Functions(){
	const char** pAry = (const char**) pthread_getspecific(s_keyFunctions);
	if(pAry == NULL){
		pAry = (const char**)calloc(MAX_STACK_DEPTH, sizeof(const char**));
		pthread_setspecific(s_keyFunctions, (void*)pAry);
	}
	return pAry;
}

#else /* NOT THREADED */

bool gse_except_init()
{
	if(except_init() != 1) return false;
	except_unhandled_catcher(gse_unhandled_catcher);
	except_set_allocator(malloc, gse_except_data_free);
	
	return true;
}

void gse_except_deinit(){

	except_deinit();
}

static void*       s_nBacktrace[MAX_STACK_DEPTH];
static const char* s_sFileNames[MAX_STACK_DEPTH];
static const char* s_sFunctions[MAX_STACK_DEPTH];

#define get_BacktraceBuf() ((void**)s_nBacktrace)
#define get_FileNames() ((const char**)s_sFileNames)
#define get_Functions() ((const char**)s_sFunctions)

#endif /* THREADED */


/*****************************************************************************/
void delete_GseExceptData(void* vpThis){
	GseExceptData* pThis = (GseExceptData*)vpThis;
	free(pThis->sMsg);
	free(pThis->sBacktrace);
	free(pThis);
}

GseExceptData* new_GseExceptData(char* sMsg, char* sBacktrace, 
                                 const char* sSrcFile, int nLine){
	GseExceptData* pThis = (GseExceptData*)except_alloc(sizeof(GseExceptData));
	pThis->del = delete_GseExceptData;
	pThis->sMsg = sMsg;
	pThis->sSrcFile = sSrcFile;
	pThis->nLine = nLine;
	pThis->sBacktrace = sBacktrace;
	pThis->sDataCtx = NULL;
	return pThis;
}

void gse_except_data_free(void* vpExceptData){
	GseExceptData* pData = (GseExceptData*)vpExceptData;
	if(pData != NULL)
		pData->del(pData);
}

void gse_unhandled_catcher(except_t* pExcept){
	fprintf(stderr, "Unhandled ");
	gse_fprint_except(stderr, pExcept);
	exit(99);
}

const char* gse_except_message(except_t* pExcept){
	const char* pRet;
	pRet = except_message(pExcept);
	if(pRet == NULL) pRet = "(no message";
	return pRet;
}

const char* gse_except_backtrace(except_t* pExcept){
	GseExceptData* pData = (GseExceptData*) except_data(pExcept);
	if(pData != NULL)
		return pData->sBacktrace;
	else
		return NULL;
}

const char* gse_except_srcfile(except_t* pExcept){
	GseExceptData* pData = (GseExceptData*) except_data(pExcept);
	if(pData != NULL)
		return pData->sSrcFile;
	else
		return NULL;
}

int gse_except_line(except_t* pExcept){
	GseExceptData* pData = (GseExceptData*) except_data(pExcept);
	if(pData != NULL)
		return pData->nLine;
	else
		return -1;
}

/****************************************************************************/
const char* clib_code_str(unsigned long uErrno, char* buf, size_t len){
#ifdef _GNU_SOURCE
	return strerror_r(uErrno, buf, len);
#elif WIN32
	return strerror(uErrno);
#else
	int nRet;
	buf[len] = '\0';
	if( (nRet = strerror_r(uErrno, buf, len - 1)) != 0)
		snprintf(buf, len, "code %lu", uErrno);
	return buf;
#endif
}

const char* host_code_str(int nCode, char* buf, size_t len){
	const char* pRet;
	
	switch(nCode){
	case HOST_NOT_FOUND:
		pRet = "Specified host is unknown"; break;
		
	case NO_ADDRESS:
		pRet = "Host name is valid but does not have an IP address"; break;
	
	case NO_RECOVERY:
		pRet = "Non-recoverable name server error"; break;
	
	case TRY_AGAIN:
		pRet = "Temporary error on authoritative name server, try again later";
		break;
	
	default:
		buf[127] = '\0';
		snprintf(buf, 127, "Unknown code %d", nCode);
		pRet = buf;
		break;
	}
	
	return pRet;
}

const char* gse_code_str(unsigned long uCode, char* buf, size_t len){
	const char* pRet;
	
	switch(uCode){
	case GSE_XBADARG:
		pRet = "Bad Arugment"; break;
				
	case GSE_XENCODE:
		pRet = "Cannot encode structure to valid byte stream"; break;
		
	case GSE_XDECODE:
		pRet = "Cannot decode byte stream to structure"; break;
			
	case GSE_XINITFAIL:
		pRet = "Failed to initialize library"; break;
	
	case GSE_XNOIMPL:
		pRet = "Missing function implimentation"; break;

	case GSE_XIOFAIL:
		pRet = "Unexpected I/O error without a CLIB exception"; break;
		
	case GSE_XASSERT:
		pRet = "Unexpected condition."; break;
				
	case GSE_XARYINDEX:
		pRet = "Array index is out of bounds"; break;
	
	case GSE_XDUPKEY:
		pRet = "Duplicate dictionary key"; break;

	case GSE_XNOKEY:
		pRet = "Missing dictionary key"; break;
		
	case GSE_XKEYTYPE:
		pRet = "Invalid dictionary key type"; break;
		
	case GSE_XVALTYPE:
		pRet = "Invalid dictionary value type"; break;
		
	case GSE_XUNDERFLOW:
		pRet = "Value too small (neg) to store in given variable"; break;
	
	case GSE_XOVERFLOW:
		pRet = "Value too large to store in given variable"; break;

	case GSE_XDIVZERO:
		pRet = "Division by zero"; break;
		
	case GSE_XPKTTOBIG:
		pRet = "Buffer exceeds maximum allowed packet size"; break;
	
	case GSE_XCANTITERATE:
		pRet = "Iterable is size 0"; break;

	case GSE_XOVERWRITE:
		pRet = "Operation would overwrite existing file"; break;
		
	case GSE_XNOTUNIQUE:
		pRet ="Unique non-NULL value does not exist"; break;
		
	case GSE_XMISSINGFILE:
		pRet = "File not found"; break;
		
	case GSE_XMISSINGDIR:
		pRet = "Directory not found"; break;

	case GSE_XMISSINGVAR:
		pRet = "Environment variable not set"; break;
		
	case GSE_XNULLPTR:
		pRet = "NULL Pointer"; break;
		
	case GSE_XNOTREADY:
		pRet = "Required operation has not finished"; break;

	case GSE_XSYNTAX:
		pRet = "Syntax Error"; break;
	
	case GSE_XLIBINCOMPAT:
		pRet ="Incompatibe version of the rpwgse lib"; break;

	case GSE_XNOTYETIMP:
		pRet = "Feature not implimented"; break;
						
	case GSE_XINCONCEIVABLE:
		pRet = "In-con-cievable"; break;
			
	default:
		buf[127] = '\0';
		snprintf(buf, 127, "Unknown code %lu", uCode);
		pRet = buf;
		break;
	}
	return pRet;
}

void gse_fprint_except(FILE* pOut, except_t* pExcept){
	
	char xcept_buf[1024];
	xcept_buf[1023] = '\0';
	
	if(pOut == NULL) pOut = stderr;
	if(pExcept == NULL) exit(135);
	fputs(gse_sprint_except((char*)xcept_buf, 1023, pExcept), pOut);
}

char* gse_sprint_except(char* sBuf, size_t len, except_t* pExcept)
{
	char code_buf[128];
	char group_buf[64];
	const char* pGroup = NULL;
	const char* pMsg = NULL;
	size_t nWrote = 0;
	
	if(len < 1) exit(137);
	if(sBuf == NULL) exit(136);
	if(pExcept == NULL) exit(135);
	sBuf[len-1] = '\0';

	switch(except_group(pExcept)){
		case CLIB_XCEPT: 
			pGroup = "CLIB";
			pMsg = clib_code_str(except_code(pExcept), code_buf, 127);
			break;
		case GSE_XCEPT:
			pGroup = "GSE";
			pMsg = gse_code_str(except_code(pExcept), code_buf, 127);
			break;
		case HOST_XCEPT:
			pGroup = "HOST";
			pMsg = host_code_str(except_code(pExcept), code_buf, 127);
			break;
		case SPICE_XCEPT:	
			pGroup = "SPICE";
			pMsg = "details follow";
			break;
		default:
			snprintf(group_buf, 63, "Group %lu", except_group(pExcept));
			pGroup = group_buf;
			snprintf(code_buf, 128, "Code %lu", except_code(pExcept));
			pMsg = code_buf;
	}
	nWrote = snprintf(sBuf, len - nWrote, "%s exception, %s.\n", pGroup, pMsg);
	if(nWrote > len) goto sprintf_except_fin;
	
	if(gse_except_srcfile(pExcept) != NULL)
		nWrote += snprintf(sBuf + nWrote, len - nWrote, "\tFrom %s line %d:  ", 
		                   gse_except_srcfile(pExcept), gse_except_line(pExcept));
	else
		nWrote += snprintf(sBuf + nWrote, len - nWrote, "\tFrom (unknown file) "
		                   "line %d:  ", gse_except_line(pExcept));
	if(nWrote > len) goto sprintf_except_fin;
	
	if(except_message(pExcept) != NULL)
		nWrote += snprintf(sBuf + nWrote, len - nWrote,"%s\n", 
		                   except_message(pExcept));
	else
		nWrote += snprintf(sBuf + nWrote, len - nWrote, "\t(no message)\n");
	if(nWrote > len) goto sprintf_except_fin;
	
	if(gse_except_backtrace(pExcept)){
		nWrote += snprintf(sBuf + nWrote, len - nWrote, "\n\tbacktrace\n"
		                   "\t---------\n%s", gse_except_backtrace(pExcept));
	}

sprintf_except_fin:
	sBuf[len-1] = '\0';
	return sBuf;
}

char* gse_shortprn_except(char* sBuf, size_t len, except_t* pExcept)
{
	char code_buf[128];
	char group_buf[64];
	const char* pGroup = NULL;
	const char* pMsg = NULL;
	size_t nWrote = 0;
	
	if(len < 1) exit(137);
	if(sBuf == NULL) exit(136);
	if(pExcept == NULL) exit(135);
	sBuf[len-1] = '\0';

	switch(except_group(pExcept)){
		case CLIB_XCEPT: 
			pGroup = "CLIB";
			pMsg = clib_code_str(except_code(pExcept), code_buf, 127);
			break;
		case GSE_XCEPT:
			pGroup = "GSE";
			pMsg = gse_code_str(except_code(pExcept), code_buf, 127);
			break;
		default:
			snprintf(group_buf, 63, "Group %lu", except_group(pExcept));
			pGroup = group_buf;
			snprintf(code_buf, 128, "Code %lu", except_code(pExcept));
			pMsg = code_buf;
	}
	nWrote = snprintf(sBuf, len - nWrote, "%s exception, %s:  ", pGroup, pMsg);
	if(nWrote < len){
	
		if(except_message(pExcept) != NULL)
			nWrote += snprintf(sBuf + nWrote, len - nWrote,"%s", 
			                   except_message(pExcept));
		else
			nWrote += snprintf(sBuf + nWrote, len - nWrote, "(no message)");
		}
	sBuf[len-1] = '\0';
	return sBuf;
}


/****************************************************************************/
/* Note the last two functions, getBacktrace and gse_throwf are excluded,   */
/* as well as the 1st Malloc that holds the program code on startup         */
char* makeBackTrace(){
	int i, nDepth;
	size_t uStrLen;
	char* sBuf;
	Dl_info info;
	
	nDepth = backtrace(get_BacktraceBuf(), MAX_STACK_DEPTH);
	
	uStrLen = 0;
	for(i = 2; i < nDepth - 1; i++){
		get_Functions()[i] = "unknown function";
		get_FileNames()[i] = "unknown file";
				
		if(dladdr(get_BacktraceBuf()[i], &info)){
			if(info.dli_sname) get_Functions()[i] = info.dli_sname;
			if(info.dli_fname) get_FileNames()[i] = info.dli_fname;
		}
		
		uStrLen += strlen(get_Functions()[i]) + strlen(get_FileNames()[i]) + 10;
	}
	
	sBuf = (char*) except_alloc(uStrLen + 1);
		
	uStrLen = 0;
	for(i = nDepth - 2; i >= 2; i--){
	
		sprintf(sBuf + uStrLen, "\t%s (from %s)\n", get_Functions()[i], get_FileNames()[i]);
		uStrLen += strlen(get_Functions()[i]) + strlen(get_FileNames()[i]) + 10;
#ifdef __MINGW32___
		if(get_Functions()[i] != NULL) free(get_Functions()[i]);
		if(get_FileNames()[i] != NULL) free(get_FileNames()[i]);
#endif
	}
	
	return sBuf;
}

/****************************************************************************/
void gse_throw_vmsg(long group, long code, const char* sFile, int nLine,
                const char *fmt, ...){
	GseExceptData* pData;
	char* sMsg = NULL;
	va_list ap;

	if(fmt != NULL){
		sMsg = except_alloc(MAX_EXCEPT_MSG_SZ);
		sMsg[MAX_EXCEPT_MSG_SZ -1] = '\0';
	
		va_start(ap, fmt);
		vsnprintf(sMsg, MAX_EXCEPT_MSG_SZ - 1, fmt, ap);
		va_end(ap);
	}
	
	pData = new_GseExceptData(sMsg, makeBackTrace(), sFile, nLine);
	except_throwd(group, code, sMsg, pData);
}

/* Throwing when the message is (maybe) present but the string contians no
   formating arguments */
void gse_throw_msg(long group, long code, const char* sFile, int nLine,
                const char *msg){
	GseExceptData* pData;
	char* sMsg = NULL;
	
	if(msg != NULL){
		sMsg = except_alloc(MAX_EXCEPT_MSG_SZ);
		sMsg[MAX_EXCEPT_MSG_SZ -1] = '\0';
	
		strncpy(sMsg, msg, MAX_EXCEPT_MSG_SZ - 1);
	}
	
	pData = new_GseExceptData(sMsg, makeBackTrace(), sFile, nLine);
	except_throwd(group, code, sMsg, pData);
}


