/** \file GseExcept.h Exception handling for the GSE library.
 *
 * @todo Add thread safe versions of file static data.
 *
 * @todo Investigate installing signal handlers for common situations such as
 *       SIGthat act as if on exception
 *       has been thrown.  
*/
#ifndef _GseExcept_H_
#define _GseExcept_H_

#include <stdbool.h>
#include <stdio.h>

#include <rpwgse/except.h>  /* Kazlib exception componet */

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_exceptions Exceptions
 * 
 * Thread safe exception handling with stack unwinding and heap cleanup.
 *
 * Thread safe exception handling with stack unwinding (via setjmp and
 * longjmp) and heap cleanup (via except_cleanup_push and except_cleanup_pop).
 * This module is a layer on the exception component of Kazlib.  This layer 
 * adds stacktrace information to the exception data.
 *
 * See @link except1.c @endlink for exception handling code patterns. 
 */
/*@{*/

/** @example except1.c
 * Basic code structure for exception handling and heap cleanup.
 */

/** C lib wrappers exception group id. 
 * For all exceptions in this group, the exception value is the associated
 * errno which can be converted to a string using strerror_r (3) */
#define CLIB_XCEPT 1

/** Win32 wrapper programs exception group id. (Only visible on Windows 
 * platforms.)  For all exceptions in this group the exception value is
 * the return from the Win32 API function, GetLastError(), and can thus
 * be converted to a string using FormatMessage().
 */
#define WIN32_XCEPT 2

/** Unix host database errors.  For all exceptions in this group the
 * exception value is the value of the external variable h_errno
 */
#define HOST_XCEPT 3


/** GSE lib Exception Group */
#define GSE_XCEPT   10


/** Spice lib Exception Group */
#define SPICE_XCEPT    20

/** Spice doesn't provide error codes, so all errors are considered the
    same */
#define SPICE_XERR     1

/** If an improper or un-useable argument was passed to a function */
#define GSE_XBADARG    1

/** If an improper header field is detected, usualy when encoding to 
 * byte stream */
#define GSE_XENCODE    2

/** If an improper input is detected on a byte stream to be decoded
  * into fields */
#define GSE_XDECODE    3

/** Library Initialization failure */
#define GSE_XINITFAIL  4 

/** Derived class did not impliment a required abstract function definition */
#define GSE_XNOIMPL    5

/** An unexpected IO error */
#define GSE_XIOFAIL    6 

/** An unexpected codition has occured, this error should never be seen in
 * production code. */
#define GSE_XASSERT    7

/** An array index is out of bounds */
#define GSE_XARYINDEX  8

/** Attempting to insert a duplicate key into an associative array (dictionary)*/
#define GSE_XDUPKEY    9

/** A key doesn't exist in a dictionary */
#define GSE_XNOKEY     10

/** Attempting to insert a dictionary pair with an invalid key type */
#define GSE_XKEYTYPE   11

/** Attempting to insert a dictionary pair with an invalid value type */
#define GSE_XVALTYPE   12

/** Value is too small to fit in storage varible, usually from trying to
 * put a negative value in an unsigned variable */
#define  GSE_XUNDERFLOW  13

/** Value is too large to fit in storage area, can come from trying to put
 * an 8 byte value in a 4 byte variable without loosing significate bits */
#define GSE_XOVERFLOW  14

/** A calculation would cause a divide by zero fault */
#define GSE_XDIVZERO   15

/** Value of a buffer exceeds the maximum packet size.   
 *@see gse_setMaxPktSz
 */
#define GSE_XPKTTOBIG 16

/** Iteration is impossible, because iterable item has size 0, 
 * or the end of the colloction has already been hit
 */
#define GSE_XCANTITERATE 17

/** An operation would overwrite an existing file */
#define GSE_XOVERWRITE 20

/** A unique non-null value could not be found */
#define GSE_XNOTUNIQUE 21

/** An expected file could not be found */
#define GSE_XMISSINGFILE 22

/** An expected environment variable is not set */
#define GSE_XMISSINGVAR 23

/** Infamous NULL pointer error */
#define GSE_XNULLPTR    24

/** Some required operation has not yet been preformed */
#define GSE_XNOTREADY   25

/** An expected directory could no be found */
#define GSE_XMISSINGDIR 26

/** Syntax error occured during parsing of human editable data */
#define GSE_XSYNTAX     30

/** current RPWGSE lib is not compatible with application code */
#define GSE_XLIBINCOMPAT 90

/** A developer intends to make this work, someday... */
#define GSE_XNOTYETIMP 99

#define GSE_XNOTDONE 99

/** Exceptions that should never ever occur. */
#define GSE_XINCONCEIVABLE 999

/** Shorter version of exceptions that should never occur. */
#define GSE_XWTF 999




/** Exception data object for this library.
 * If the library user wishes to call except_throwd directly for there own
 * exception data, thier exception data must derive from this and override
 * the #del member.
 * @class GseExceptData
 */
typedef struct gse_except_data{
	/** Destructor function */
	void (*del)(void* pThis);
	
	/** Message given at the throw site. */
	char* sMsg;
	
	/** Source code file containing the throw site. 
	 @note that the macros #gse_throw, #gse_throw_m, or #gse_throw_fm
	 automatically add this.*/
	const char* sSrcFile;
	
	/** Source code line number of the throw site 
	@note that the macros #gse_throw, #gse_throw_m, or #gse_throw_fm
	 automatically add this */
	int nLine;
	
	/** Automatically appended backtrace */
	char* sBacktrace;
	
	/** User supplied data processing context string, if not null cleanup
	 * routine will free this. */
	char* sDataCtx;
} GseExceptData;

/** Make a new exception data object.
 * Called by indirectly by #gse_throw, #gse_throw_m and @#gse_throw_fm.
 * @memberof GseExceptData
 */
GseExceptData* new_GseExceptData(char* sMsg, char* sBacktrace, 
                                 const char* sSrcFile, int nLine);

/** Get the backtrace from an exception */
const char* gse_except_backtrace(except_t* pExcept);

/** Get the source file from which the exception was thrown */
const char* gse_except_srcfile(except_t* pExcept);

/** Get the line number from which the exception was thrown */
int gse_except_line(except_t* pExcept);

/** Throw an exception that contains error code, a message and a stack trace.
 * 
 * There is little need to call this function directly, use the wrapper 
 * macro #gse_throw_fm.
 *
 * Note: Do NOT call the Kazlib function except_throwf directly as this
 * library adds a custom exception cleanup handler that will treat the
 * string as a GseExceptData structure and will thus SEGFAULT. */
void gse_throw_vmsg(long group, long code, const char* sFile, int nLine, 
		              const char* fmt, ...);
					 
/** Throw an exception that contains error code, a message and a stack trace.
 * 
 * There is little need to call this function directly, use the wrapper 
 * macros #gse_throw or #gse_throw_m.
 *
 * Note: Do NOT call the Kazlib function except_throwf directly as this
 * library adds a custom exception cleanup handler that will treat the
 * string as a GseExceptData structure and will thus SEGFAULT. */					 
void gse_throw_msg(long group, long code, const char* sFile, int nLine, 
		              const char* msg);

/** Get the message generated at the call site from an exception object.
 * 
 * @param pExcept a pointer to an exception object.
 *
 * @returns Always returns a usable string.  If no message was generated
 *          at the call site, a pointer to the string "(no message)" is
 *          returned
 */
const char* gse_except_message(except_t* pExcept);


/** Get the backtrace out of an exception */
const char* gse_except_backtrace(except_t* pExcept);

/** Pretty print an exception to a stream.
 * 
 * Used by the default unhandled exception handler.
 */
void gse_fprint_except(FILE* pOut, except_t* pExcept);

/** Pretty print an exception to a string.
 * 
 * @param sBuf buffer to contain the exception message
 * @param len the length of the buffer.
 * @param pExcept the exception to print.
 * @return The pretty printed exception message, this is guarranteed to be
 *         NULL terminated.
 */
char* gse_sprint_except(char* sBuf, size_t len, except_t* pExcept);

/** Pretty print an exception to a string, without backtrace or soure file
 * 
 * @param sBuf buffer to contain the exception message
 * @param len the length of the buffer.
 * @param pExcept the exception to print.
 * @return The pretty printed exception message, this is guarranteed to be
 *         NULL terminated.
 */
char* gse_shortprn_except(char* sBuf, size_t len, except_t* pExcept);

/** Wrapper for gse_throwf that adds the file name and line number to 
 * exception messages.  
 *
 * This macro does not take a caller supplied message
 */
#define gse_throw(G,C) gse_throw_msg(G, C, __FILE__, __LINE__, NULL)

/** Compatible with gse_throwf, but adds file name and line number to 
 * exception messages.  This macro expects a user message as either a string
 * literal or as a const char* pointer. */
#define gse_throw_m(G,C,F)  gse_throw_msg(G, C, __FILE__, __LINE__, F) 

/** Compatible with gse_throwf, but adds file name and line number to 
 * exception messages.  This varargs macro expects a caller supplied format
 * string literal, as well as values in the style of fprintf. */ 
#define gse_throw_fm(G,C,F,...) \
  gse_throw_vmsg(G, C, __FILE__, __LINE__, F, __VA_ARGS__)

/** Assert-like macro statement that throws an exception instead of halting
 * the program */
#define gse_assert(B) \
  if(!(B)) gse_throw_msg(GSE_XCEPT, GSE_XASSERT, __FILE__, __LINE__, NULL)

/** Structure for adding to the set of builtin exception definitions.
 * @see gse_thread_xcept_def()
 */
struct gse_except_def{
	int code;
	const char* msg;
};

/** Given a Clib errno, convert it into an error string.
 * @param uErrno the errno value
 * @param buf a pointer to a string buffer
 * @param len the length of the string buffer
 * 
 * @return A pointer to the error string, which may be the same as 
 *         @b buf or may point to some constant string inside the system's
 *         clib.  Always use the return value, do not depend on @b buf
 *         containing the return value.
 *
 * @todo: Collapse this to a single function that works with the new
 *        exception definition function gse_thread_xcepts_defs()
 */
const char* clib_code_str(unsigned long uErrno, char* buf, size_t len);

/** Given an execption code in the GSE_XCEPT group return an error string.
 *
 * @param uCode the exception code obtained via except_code(pExcept)
 * @param buf a pointer to a string buffer
 * @param len the length of the string buffer
 * 
 * @return A pointer to the error string, which may be the same as 
 *         @b buf or may point to some constant string inside the system's
 *         clib.  Always use the return value, do not depend on @b buf
 *         containing the return value.
 */
const char* gse_code_str(unsigned long uCode, char* buf, size_t len);

/** Override of the default kazlib unhandled catcher that expects the pData
 * to be a GseExceptData structure */
void gse_unhandled_catcher(except_t* pExcept);

/** GseExceptData destructor, call del member so that destructors can
 * be virtual */
void gse_except_data_free(void* vpExceptData);
		
/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseExcept_H_ */

