#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <assert.h>
#include <strings.h>

#ifdef LINUX
#include <endian.h>
#else
#include <sys/param.h>
#endif

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseString.h"
#include "GseLog.h"
#include "GseMath.h"
#include "GseConv.h"

#include "GseField.h"
#include "GseParser.h"
#include "GseFieldPrn.h"


/*****************************************************************************/
/* Little Wrapers */

const char* GseField_name(const GseField* pThis)
{
	return (const char*)pThis->sName;
}

const char* GseField_desc(const GseField* pThis)
{
	return gse_pd_getstr(pThis->pProps, "desc");
}

char GseField_valType(const GseField* pThis)
{
	if(pThis->cEffType == '\0')
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, field validation required",
				       pThis->sName);

	return pThis->cEffType;
}

/*****************************************************************************/
/* Validation */

char* GseField_checkOrAndHiBit(GseField* pThis)
{
	const int8* pnTmp = NULL;
	const char* sTmp;
	int8 nSz   = 0;
	int8 nSft  = 0;
	byte b     = 0;
	uint8 uTmp = 0;
	uint8 uOR  = 0;
	char cType = '\0';
	
	sTmp = gse_pd_getstr(pThis->pProps, "ty");
	cType = sTmp[0];
	
	/* This stuff only matters for stuff that is turned into a Numeric value */
	if((cType == 'C')||(cType == 'B')){
		pThis->uHiBitIdx = 0;
		return NULL;
	}
	
	nSz   = *( gse_pd_getint(pThis->pProps, "sz") );
	
	if( (pnTmp = gse_pd_getint(pThis->pProps, "sft")) != NULL)
		nSft  = *pnTmp;
	
	if( ! gse_pd_cktype(pThis->pProps, pt_int, "or") )
		return CS("Type error in property 'or'");
		
	if( (pnTmp = gse_pd_getint(pThis->pProps, "or")) != NULL)
		uOR = *( (uint8*) pnTmp);
	
	/* If this field is not of type B or C, calculate the high-bit that we will
	   have after read, shift and or */
	if((cType != 'B')&&(cType != 'C')){
			
		pThis->uHiBitIdx = nSz*8 - nSft - 1;
		
		if(uOR != 0){
			b = 63;
			uTmp = uOR;
			while((b > 0) && !(uTmp & 0x8000000000000000)){
				b--;
				uTmp = uTmp << 1;
			}
			if(b > pThis->uHiBitIdx) pThis->uHiBitIdx = b;
		}
	}
	else
		pThis->uHiBitIdx = 0;
	
	/* Check for a legal hi bit index for Float types */
	if(cType == 'F'){
	
		/* Allow for slop with an implied sign bit */
		if((pThis->uHiBitIdx != 30)&&(pThis->uHiBitIdx != 31)&&
		   (pThis->uHiBitIdx != 62)&&(pThis->uHiBitIdx != 63)){
		
			if(nSft < 0)
				return NS("A %lld byte field left shifted %lld bits and OR'ed with "
						    " %llX has a high bit index of %hhu, which aligns "
					       " with neither a 4 byte nor 8 byte IEEE float.",
			             nSz, nSft, uOR, pThis->uHiBitIdx);
			else
				return NS("A %lld byte field OR'ed with %llX has a high bit "
				          "index of %hhu, which aligns with neither a 4 byte nor "
				          "8 byte IEEE float.", nSz, uOR, pThis->uHiBitIdx);
		}
	}
	
	if(cType == 'S'){
		
		/* Allow for slop with an implied sign bit */
		if((pThis->uHiBitIdx != 6 )&&(pThis->uHiBitIdx != 7 )&&
		   (pThis->uHiBitIdx != 14)&&(pThis->uHiBitIdx != 15)&&
		   (pThis->uHiBitIdx != 30)&&(pThis->uHiBitIdx != 31)&&
		   (pThis->uHiBitIdx != 62)&&(pThis->uHiBitIdx != 63)){
			
			if(nSft < 0)
				return NS("A %lld byte field left shifted %lld bits and OR'ed with "
						    " %llX has a high bit index of %hhu, which aligns "
					       " with neither a 1, 2, 4 nor 8 byte signed integer.",
			             nSz, nSft, uOR, pThis->uHiBitIdx);
			else
				return NS("A %lld byte field OR'ed with %llX has a high bit "
				          "index of %hhu, which aligns with neither a 1, 2, "
				          "4 nor 8 byte signed integer.", nSz, uOR, 
				          pThis->uHiBitIdx);
		}
	}
	
	return NULL;
}

char* GseField_getAlarmVal(const GseProp* pProp, int4 item, float8* pRes)
{
	if((pProp->ptype != pt_flt)&&(pProp->ptype != pt_int))
		return NS("Element %d of 'alarm' list is not a numeric property.", item);
	
	if(pProp->ptype == pt_flt)
		*pRes = ((const GsePropFloat*)pProp)->fVal;
	else
		*pRes = ((const GsePropInt*)pProp)->nVal;
	
	return NULL;
}


char* GseField_checkAlarm(const GseField* pThis)
{
	const GseProp*     pAlarm  = NULL;
	const GsePropList* pList   = NULL;
	const GsePropEnum* pEnum   = NULL;
	const GseProp*     pMember = NULL;
	
	float8 fCur, fNext;
	char* sErr = NULL;
	const char* sVal;
	const int8* pnKey = NULL;
	
	int4 i;
	
	if( (pAlarm = pThis->pProps->get(pThis->pProps, "alarm")) == NULL)
		return NULL;
	
	
	switch(pAlarm->ptype){
		
	case pt_list:
		pList = (GsePropList*)pAlarm;
	
		/* Make sure that there are 4 values */
		if(pList->length(pList) != 4)
			return NS("Alarm list contains %d values, expected 4",pList->length(pList));
		
		pMember = pList->get(pList, 0);
		if( (sErr = GseField_getAlarmVal(pMember, 1, &fCur)) != NULL)
			return sErr;
	
		for(i=1; i<4; i++){
			pMember = pList->get(pList, i);
		
			if( (sErr = GseField_getAlarmVal(pMember, i+1, &fNext)) != NULL)
				return sErr;
		
			if(fNext <= fCur)
				return CS("Alarm values are not monotonically increasing");
			fCur = fNext;
		}	
		break;
	
	case pt_enum:
		pEnum = (GsePropEnum*)pAlarm;
		
		/* Go through the alarm dictionary and make sure the values are legal */
		for(pnKey = pEnum->pDict->first(pEnum->pDict);
		 pnKey != NULL;
		 pnKey = pEnum->pDict->next(pEnum->pDict, *pnKey))
		{
			sVal = pEnum->pDict->get(pEnum->pDict, *pnKey);		
		
			if(gse_str2alarm(sVal) == inv_alarm_stat)
				return NS("Unrecognized condition '%s' for alarm value %d", sVal, *pnKey); 
		}
	
		/* Check the value for the default, if present */
		if(pEnum->hasDefault(pEnum)){
			if(gse_str2alarm(pEnum->sDefault) == inv_alarm_stat)
				return NS("Unrecognize condition '%s' for alarm default",
						     pEnum->sDefault); 
		}	
		break;
		
	default:
		return CS("Alarm property is not a list or enumeration.");
		break;
	}
	
	return NULL;
}


char* GseField_validate(GseField* pThis)
{
	char* sErr = NULL;
	
	const char* sTmp = NULL;
	const int8* pnTmp = NULL;
	const GseProp* pProp = NULL;
	const GsePropList* pList = NULL;
	const char* sPFunc;
	
	char cType = '\0';
	int8 nSz = 0;
	int4 n;
	
	const GseStrPropDict* pD = pThis->pProps;
	
	/* I need to know: The data type, the offset, the byteorder, the sz
	 * and they need to be inside the proper range */
	
	/* Type */
	if( ! gse_pd_cktype(pD, pt_str, "ty"))
		return CS("Type error for property 'ty'");
	
	if( (sTmp = gse_pd_getstr(pD, "ty")) == NULL)
		return CS("Type, 'ty', property missing");
	
	if(sTmp[0] != 'B' && sTmp[0] != 'U' && sTmp[0] != 'S' && sTmp[0] != 'F' &&
	   sTmp[0] != 'C')
		return NS("Unknown value for the type property: '%s'", sTmp);
	else
		cType = sTmp[0];
	
	
	/* Offset */
	if( ! gse_pd_cktype(pD, pt_int, "off"))
		return CS("Type error for property 'off'");
		
	if( (pnTmp = gse_pd_getint(pD, "off")) == NULL)
		return CS("Offset, 'off', property missing");
	
	if(*pnTmp < 0)
		return CS("Negative offset properties are not supported");
	
	if(*pnTmp > 65536)
		gse_warning_v("Unusual offset, %lld for field %s.", *pnTmp, pThis->sName);
	
	
	/* Byte Order */
	if( ! gse_pd_cktype(pD, pt_str, "bo"))
		return CS("Type error for property 'bo'");

	if( (sTmp = gse_pd_getstr(pD, "bo")) == NULL)
		return CS("Byte order, 'bo', property missing");
	
	if( !(strcasecmp(sTmp, "BE") == 0) && !(strcasecmp(sTmp, "LE")==0))
		return NS("Unknown value for the byte order property: '%s'", sTmp);
	
	
	/* Size (in bytes) */
	if( ! gse_pd_cktype(pD, pt_int, "sz"))
		return CS("Type error for property 'sz'");

	if( (pnTmp = gse_pd_getint(pD, "sz")) == NULL)
		return CS("Size in bytes, 'sz', property missing");
	
	if(*pnTmp < 1)
		return NS("Invalid field size, %lld", *pnTmp);
	
	if((cType != 'B') && (cType != 'C') && (*pnTmp > 8))
		return CS("Fields that are greater than 8 bytes long must be of type "
				    "'B'inary, or type 'C'haracter");
	nSz = *pnTmp;
	
	/* Printing function */
	if( (sPFunc = gse_pd_getstr(pD, "pFunc")) == NULL)
		return CS("Type, 'pfunc', property missing.");
	
	if( ! gse_has_field_printer(sPFunc) )
		return NS("Field printing function %s is unknown.", sPFunc);
	
	
	/* There is a corner case here where a 64 bit unsigned can not be
	   evaluated, but a 63 bit one can.  Knowing the difference would
		require knowledge of the shift and mask fields, punt for now */	
	
	/* Now check optional items */
	
	/* Shift */
	if( ! gse_pd_cktype(pD, pt_int, "sft"))
		return CS("Type error for property 'sft'");

	if( (pnTmp = gse_pd_getint(pD, "sft")) != NULL){
		
		if( llabs(*pnTmp) >= 64)
			return NS("Shift value, %lld is too large. Must be within the range"
					    " -63 (up) to +63 (down)", *pnTmp);
	}
	
	/* Mask */
	if( ! gse_pd_cktype(pD, pt_int, "msk"))
		return CS("Type error for property 'msk'");

	if( (pnTmp = gse_pd_getint(pD, "msk")) != NULL){
		if(*pnTmp == 0)
			gse_warning_v("Field %s, mask of 0 always produces no data",
					        pThis->sName);
		if(*pnTmp < 0)
			return NS("Mask value, %lld is negative.", *pnTmp);
	}
	
	/* Or, and set the uHiBitIdx */
	if( (sErr = GseField_checkOrAndHiBit(pThis)) != NULL)
		return sErr;
	
	
	/* Polynomial checking, polynomials can cause type promotion */
	pThis->cEffType = cType;
	if( ! gse_pd_cktype(pD, pt_list, "cal") )
		return CS("cal property syntax error");
	
	if( (pList = gse_pd_getlist(pD, "cal")) != NULL){
		
		if((cType == 'B')||(cType == 'C'))
			return NS("'cal' property not understood for type '%c' fields", cType);
		
		for(n = 0; n<pList->length(pList); n++)
		{
			pProp = pList->get(pList, n);
			if( (pProp->ptype != pt_int) && (pProp->ptype != pt_flt))
				return NS("Coefficient %d of the 'cal' property is neither a float"
				          " nor an integer.", n);
			
			if(pProp->ptype == pt_flt) 
				pThis->cEffType = 'F';
			
			if((pProp->ptype == pt_int)&&(pThis->cEffType != 'F')){
				if( ( ((GsePropInt*)pProp)->nVal < 0) && (pThis->cEffType == 'U') )
					pThis->cEffType = 'S';
			}
		}
	}
	
	/* Alarm checking */
	if( (sErr = GseField_checkAlarm(pThis)) != NULL)
		return sErr;
	
	/* Format validataion (includes validating the lack of a format) */
	if( (sErr = gse_validate_fmt(sPFunc, pThis)) != NULL)
		return sErr;
		
	pThis->bNeedsVal = false;
	return NULL;
}

/*****************************************************************************/
/* Converting Bits */

uint8 GseField_convUnsign(const GseField* pThis, cbytebuf_t cbuf){
	
	bool  bFieldIsLSBfirst = true;
	uint8 uMsk             = 0xffffffffffffffff;
	uint8 uOR              = 0;
	int8  nSft             = 0;
	
	uint8 uRet = 0;
	byte uShift = 0;
	byte* pPut = NULL;
	const int8* pnTmp;
	int i;
	
	if(cbuf.len > 8)
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	if((pnTmp = gse_pd_getint(pThis->pProps, "sft")) != NULL)
		nSft = *pnTmp;
	
	if((pnTmp = gse_pd_getint(pThis->pProps, "msk")) != NULL)
		uMsk = *( (uint8*)pnTmp );
	
	if((pnTmp = gse_pd_getint(pThis->pProps, "or")) != NULL)
		uOR = *( (uint8*)pnTmp );
	
	if( strcasecmp( gse_pd_getstr(pThis->pProps, "bo"), "BE") == 0)
		bFieldIsLSBfirst = false;
	
	/* Start put in the lowest position */
	if(g_bIsLSBfirst)
		pPut = (byte*)(&uRet);
	else
		pPut = (byte*)( &uRet + 7);
	
	/* Read from the lowest order to the highest order bytes of the field */
	if(bFieldIsLSBfirst){
		for(i = 0; i< cbuf.len; i++){
			*pPut = cbuf.ptr[i];
			
			if(g_bIsLSBfirst) 
				pPut++;
			else
				pPut--;
		}
	}
	else{
		for(i = cbuf.len - 1; i >= 0; i--){
			*pPut = cbuf.ptr[i];
			
			if(g_bIsLSBfirst) 
				pPut++;
			else
				pPut--;
		}
	}
	
	/* uRet is loaded */
	
	if(nSft != 0){
		if(nSft < 0){
			uShift = (byte)( nSft * -1);
			
			uRet = uRet << uShift;
		}
		else{
			uShift = (byte)nSft;
			uRet = uRet >> uShift;
		}
	}
	
	uRet &= uMsk;
	uRet |= uOR;
	
	return uRet;	
}

float8 GseField_convFloat(const GseField* pThis, cbytebuf_t cbuf)
{
	uint8 u8;
	float4 f4;
	float8 f8;
	
	u8 = GseField_convUnsign(pThis, cbuf);
	
	/* Figure out if this is a 32 bit float, or a 64 bit float, 
	 * For 32 bit floats, make sure that the bits get up to at least the 30th
	 * index (assuming it's okay to take the sign bit as 0 = always positive)
	 *	and that for 64 bit floats that the bits get up to at least the 62nd 
	 * position.
	 */
	
	if((pThis->uHiBitIdx == 30)||(pThis->uHiBitIdx == 31)){
		/* convert as a 4 byte float, remember, for MSB 1st machines
		 * the float is gonig to start in the middle of the uint8 */
		if(g_bIsLSBfirst)
			f4 = *( (float4*) &u8 );
		else
			f4 = *( (float4*) (&u8 + 4) );
		
		f8 = f4;
		return f8;
	}
	
	if((pThis->uHiBitIdx == 62)||(pThis->uHiBitIdx == 63)){
		f8 = *( (float8*) &u8);
		return f8;
	}
	
	gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	return 0;
}

int8 GseField_convSign(const GseField* pThis, cbytebuf_t cbuf)
{
	uint8 u8;
	int1 i1;
	int2 i2;
	int4 i4;
	int8 i8;
	
	u8 = GseField_convUnsign(pThis, cbuf);
	
	if((pThis->uHiBitIdx == 6)||(pThis->uHiBitIdx == 7)){
		/* Convert this as a 1 byte int. */
		if(g_bIsLSBfirst)
			i1 = *( (int1*) &u8);
		else
			i1 = *( (int1*) (&u8 + 7));
		i8 = i1;
		return i8;
	}
	
	if((pThis->uHiBitIdx == 14)||(pThis->uHiBitIdx == 15)){
		/* Convert this as a 2 byte int. */
		if(g_bIsLSBfirst)
			i2 = *( (int2*) &u8);
		else
			i2 = *( (int2*) (&u8 + 6));
		i8 = i2;
		return i8;
	}
	
	if((pThis->uHiBitIdx == 30)||(pThis->uHiBitIdx == 31)){
		/* Convert this as a 4 byte int. */
		if(g_bIsLSBfirst)
			i4 = *( (int4*) &u8);
		else
			i4 = *( (int4*) (&u8 + 4));
		i8 = i4;
		return i8;
	}
	
	if((pThis->uHiBitIdx == 62)||(pThis->uHiBitIdx == 63)){
		/* Convert this as an 8 byte int. */
		i8 = *( (int8*) &u8);
		return i8;
	}
	
	gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	return 0;
	
}

#define GSE_FLOAT_VAL(L, I, A) pProp = L->get(L, I); \
   if(pProp->ptype == pt_flt) \
		A = ((const GsePropFloat*)pProp)->fVal; \
	else \
		A = ((const GsePropInt*)pProp)->nVal

float8 GseField_calibrate(float8 fVal, const GsePropList* pPoly)
{
	int n;
	const GseProp* pProp;
	float8 fCoeff;
	float8 fRes = 0;
	
	for(n = 0; n < pPoly->length(pPoly); n++){
		GSE_FLOAT_VAL(pPoly, n, fCoeff);
	
		fRes += fCoeff * gse_pow(fVal, n);
	}
	
	return fRes;
}

gse_alarm_stat_t GseField_setListAlarm(float8 fVal, const GsePropList* pAlarm)
{
	/* Function assumes that the alarm property has been validated!!! */
	float8 fTest;
	const GseProp* pProp;
	
	/* Work in from the extreams */
	
	GSE_FLOAT_VAL(pAlarm, 0, fTest);
	if(fVal <= fTest)
		return err_lo;
	
	GSE_FLOAT_VAL(pAlarm, 1, fTest);
	if(fVal <= fTest)
		return warn_lo;

	GSE_FLOAT_VAL(pAlarm, 3, fTest);
	if(fVal >= fTest)
		return err_hi;
	
	GSE_FLOAT_VAL(pAlarm, 2, fTest);
	if(fVal >= fTest)
		return warn_hi;

	return no_alarm;
}

gse_alarm_stat_t GseField_setEnumAlarm(int8 nVal, const GsePropEnum* pAlarm)
{
	if(! pAlarm->hasInt(pAlarm, nVal) ){
		if( ! pAlarm->hasDefault(pAlarm))
			return err_lo;  /* Arbitrary decision, but will get noticed */
		else
			return gse_str2alarm( pAlarm->sDefault);
	}
	
	return gse_str2alarm( pAlarm->i8ToStr(pAlarm, nVal) );
}

char* GseField_read(GseField* pThis, cbytebuf_t cbuf)
{
	char* sErr;
	const GsePropList* pCal;
	const GseProp*     pAlarm;
	
	int8 nOff;
	int8 nSz;
	bool bBE;
	char cType;
	
	const char* sPFunc = NULL;
	
	const GseStrPropDict* pD = pThis->pProps;
		
	if(cbuf.ptr == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null buffer data pointer");
	
	/* first validate if need be */
	if(pThis->bNeedsVal){
		if( (sErr = pThis->validate(pThis)) != NULL )
			return sErr;
	}
	
	/* Store the raw bytes */
	free(pThis->buf.ptr);
	nSz  = *( gse_pd_getint(pD, "sz") );
	
	pThis->buf.ptr = gse_calloc(nSz, sizeof(byte));
	pThis->buf.len = nSz;
	
	nOff = *( gse_pd_getint(pD, "off") );
	bBE = false;
	if( strcasecmp( gse_pd_getstr(pD, "bo"), "BE") == 0)
		bBE = true;
	
	if(bBE){
		if(nOff + 1 - nSz < 0)
			return NS("field is defined such that it starts before the beginning "
			          "of the blob.  Offset to byte 0: %lld, Field length: %lld",
			          nOff, nSz);
		if(nOff >= cbuf.len)
			return NS("field is defined such that in runs past the end of the"
					    "available bytes.  Field Offset: %lld, Blob length: %z",
			          nOff, cbuf.len);
		
		memcpy(pThis->buf.ptr, cbuf.ptr + nOff + 1 - nSz,  nSz);
	}
	else{
		if(nOff + nSz > cbuf.len)
			return NS("field is defined such that in runs past the end of the"
					    "available bytes.  Field Offset: %lld, Field Size: %lld, "
		      	    "Blob length: %z", nOff, nSz, cbuf.len);
		
		memcpy(pThis->buf.ptr, cbuf.ptr + nOff,  nSz);
	}
	
	/* Convert the bytes to numbers, if needed */
	cType = gse_pd_getstr(pD, "ty")[0];
	switch(cType){				
	case 'F':
		pThis->fVal = GseField_convFloat(pThis, const_bytebuf(pThis->buf));
		break;
	
	case 'S':
		pThis->iVal = GseField_convSign(pThis, const_bytebuf(pThis->buf));
		pThis->fVal = pThis->iVal;
		break;
			
	case 'U':
		pThis->uVal = GseField_convUnsign(pThis, const_bytebuf(pThis->buf));
		pThis->fVal = pThis->uVal;
		break;

	default:  /* Nothing to convert for 'B' and 'C' types */
		break;
	}
	
	if( (pCal = gse_pd_getlist(pD, "cal")) != NULL){
		switch(pThis->cEffType){
		case 'F': pThis->fVal = GseField_calibrate(pThis->fVal,pCal); break;
		case 'S': pThis->iVal = GseField_calibrate(pThis->iVal,pCal); break;
		case 'U': pThis->uVal = GseField_calibrate(pThis->uVal,pCal); break;
		default: break;	
		}
	}
	
	if( (pAlarm = pD->get(pD, "alarm")) != NULL){
		if(pAlarm->ptype == pt_list)
			pThis->alrm = GseField_setListAlarm(pThis->fval(pThis),
					                              (GsePropList*)pAlarm);
		else
			pThis->alrm = GseField_setEnumAlarm(pThis->ival(pThis),
					                              (GsePropEnum*)pAlarm);
	}
	
	/* Convert to a string */
	free(pThis->sVal);
	sPFunc = gse_pd_getstr(pD, "pfunc");
	pThis->sVal = gse_print_field(sPFunc, pThis);
	pThis->bNeedsRead = false;
	return NULL;
}

/*****************************************************************************/
/* Reading converted values */

cbytebuf_t GseField_bval(const GseField* pThis)
{
/*	if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required",
				       pThis->sName); */

	return const_bytebuf(pThis->buf);
}

const char* GseField_sval(const GseField* pThis)
{
/*	if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required",
				       pThis->sName); */
	
	return (const char*) pThis->sVal;
}

float8 GseField_fval(const GseField* pThis)
{
	float8 fRet;
	
	/* if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required",
				       pThis->sName); */

	switch(pThis->valType(pThis))
	{
		case 'B':
		case 'C': fRet = pThis->buf.ptr[0]; break;
		
		case 'F':
		case 'S':
		case 'U': fRet = pThis->fVal; break;
		default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	}
	
	return fRet;
}

uint8 GseField_uval(const GseField* pThis)
{
	uint8 uRet;
	float8 fCompare;
	
	/* if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required", 
				       pThis->sName); */

	switch(pThis->valType(pThis))
	{
		case 'B':
		case 'C': uRet = pThis->buf.ptr[0]; break;
		
		case 'F': 
			fCompare = GSE_UINT8_MAX;
			if( pThis->fVal > fCompare){
				uRet = GSE_UINT8_MAX;
			}
			else{
				if( pThis->fVal < 0)
					uRet = 0;
				else
					uRet = pThis->fVal;
			}
			break;
		
		case 'S': 
			if(pThis->iVal < 0)
				uRet = 0;
			else
				uRet = pThis->iVal;
			break;		
			
		case 'U': uRet = pThis->uVal; break;
		default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	}
	
	return uRet;	
}

int8 GseField_ival(const GseField* pThis)
{
	int8 iRet;
	float8 fCompare;
	
	/* if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required", 
				       pThis->sName); */
	
	switch(pThis->valType(pThis))
	{
		case 'B':
		case 'C': iRet = pThis->buf.ptr[0]; break;
		
		case 'F': 
			fCompare = GSE_INT8_MAX;
			if( pThis->fVal > fCompare){
				iRet = GSE_INT8_MAX;
			}
			else{
				fCompare = GSE_INT8_MIN;
				if( pThis->fVal < fCompare)
					iRet = GSE_INT8_MIN;
				else
					iRet = pThis->fVal;
			}
			break;
		
		case 'S': 
			iRet = pThis->iVal;
			break;		
			
		case 'U': 
			if(pThis->uVal > GSE_INT8_MAX)
				iRet = GSE_INT8_MAX;
			else
				iRet = pThis->uVal;
			break;
			
		default:
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
	}
	
	return iRet;	
}

gse_alarm_stat_t GseField_alarm(const GseField* pThis)
{
	if(pThis->bNeedsRead)
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "%s, data read required", 
				       pThis->sName);
	
	return pThis->alrm;
}


/*****************************************************************************/
/* Manipulating properties */

const GseProp* GseField_first(const GseField* pThis)
{
	const char* sKey;
	
	if( (sKey = pThis->pProps->first(pThis->pProps)) == NULL)
		return NULL;
	else
		return pThis->pProps->get(pThis->pProps, sKey);
}

const GseProp* GseField_next(const GseField* pThis, const GseProp* pProp)
{
	const char* sKey;
	
	if( (sKey = pThis->pProps->next(pThis->pProps, pProp->sName)) == NULL)
		return NULL;
	else
		return pThis->pProps->get(pThis->pProps, sKey);
}

bool GseField_hasProp(const GseField* pThis, const char* sName)
{
	return pThis->pProps->hasKey(pThis->pProps, sName);	
}

const GseProp* GseField_getProp(const GseField* pThis, const char* sName)
{
	return pThis->pProps->get(pThis->pProps, sName);
}

void GseField_addProp(GseField* pThis, GseProp* pProp)
{
	GseProp* pOld;
	
	pThis->bNeedsVal = true;
	pThis->bNeedsRead = true;
			
	if(pThis->pProps->hasKey(pThis->pProps, pProp->sName)){
		pOld = pThis->pProps->replace(pThis->pProps, pProp->sName, pProp);
		pOld->del(pOld);
	}
	else{
		pThis->pProps->insert(pThis->pProps, pProp->sName, pProp);
	}
}

GseProp* GseField_takeProp(GseField* pThis, const char* sName)
{
	pThis->bNeedsVal = true;
	pThis->bNeedsRead = true;
	
	return pThis->pProps->take(pThis->pProps, sName);
}


/*****************************************************************************/
/* Creation, destruction */

void delete_GseField(void* vpThis)
{
	GseField* pThis = (GseField*)vpThis;
	
	free(pThis->sVal);
	free(pThis->sName);
	free(pThis->buf.ptr);
	
	pThis->pProps->del(pThis->pProps);
}

GseField* new_GseField(const char* sName, GseStrPropDict* pProps,
                       GseStrPropDict* pDefs, int nLineNo)
{
	GseField* pThis;
	const char* sKey;
	const GseProp* pProp;
	GseProp* pCopy;
	
	pThis = (GseField*)gse_calloc(1, sizeof(GseField));
	pThis->del = delete_GseField;
	
	except_cleanup_push(pThis->del, pThis);
	
	if(sName != NULL)
		pThis->sName = CS(sName);
	else
		pThis->sName = "unnamed";
	
	/* Make or take ownership of the property dictionary */
	if(pProps != NULL)
		pThis->pProps = pProps;
	else
		pThis->pProps = new_GseStrPropDict(pThis->sName);
	
	pThis->pProps->setRetNull(pThis->pProps);
	
	/* Fold in the defaults if there are any */
	if(pDefs != NULL){
		for(sKey=pDefs->first(pDefs); sKey!=NULL; sKey=pDefs->next(pDefs,sKey)){
			if( ! pThis->pProps->hasKey(pThis->pProps, sKey) ){
				pProp = pDefs->get(pDefs, sKey);
				pCopy = pProp->copy(pProp);
				pThis->pProps->insert(pThis->pProps, pCopy->sName, pCopy);
			}
		}
	}
	
	pThis->name     = GseField_name;
	pThis->desc     = GseField_desc;
	pThis->valType  = GseField_valType;
	
	pThis->validate = GseField_validate;	
	pThis->read     = GseField_read;
	
	pThis->bval     = GseField_bval;
	pThis->sval     = GseField_sval;
	pThis->fval     = GseField_fval;
	pThis->uval     = GseField_uval;
	pThis->ival     = GseField_ival;
	pThis->alarm    = GseField_alarm;
	
	pThis->first    = GseField_first;
	pThis->next     = GseField_next;
	pThis->hasProp  = GseField_hasProp;
	pThis->getProp  = GseField_getProp;
	pThis->addProp  = GseField_addProp;
	pThis->takeProp = GseField_takeProp;
	
	pThis->del      = delete_GseField;
	
	pThis->nLine      = nLineNo;
	pThis->bNeedsVal  = true;
	pThis->bNeedsRead = true;
	
	except_cleanup_pop(except_no_call);
	return pThis;
}


/*****************************************************************************/
/* Dictionary of fields */

#define C GseStrFieldDict
#define K char
#define V GseField

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcasecmp(a, b)

#include "ptrDictTplt.c"

#undef FREE_KEYS
#undef DEL_VALS
#undef C
#undef K
#undef V

void gse_fd_setconfig(GseStrFieldDict* pDict, const char* sConfFile)
{
	GseField* pField;
	const char* sKey;
	
	for(sKey = pDict->first(pDict); sKey != NULL; sKey = pDict->next(pDict,sKey))
	{
		pField = pDict->getMutable(pDict, sKey);
		pField->sConf = sConfFile;
		
		gse_pd_setconfig(pField->pProps, sConfFile);
	}
}


void gse_fd_setparser(GseStrFieldDict* pDict, const struct gse_parser* pParent)
{
	GseField* pField;
	const char* sKey;
	
	for(sKey = pDict->first(pDict); sKey != NULL; sKey = pDict->next(pDict,sKey))
	{
		pField = pDict->getMutable(pDict, sKey);
		pField->pParent = pParent;
	}
}

/*****************************************************************************/
/* Misc */

gse_alarm_stat_t gse_str2alarm(const char* sVal)
{
	if(strcmp(sVal, "normal") == 0) return no_alarm;
	if(strcmp(sVal, "err_lo") == 0) return err_lo;
	if(strcmp(sVal, "warn_lo") == 0) return warn_lo;
	if(strcmp(sVal, "warn_hi") == 0) return warn_hi;
	if(strcmp(sVal, "err_hi") == 0) return err_hi;
	
	return inv_alarm_stat;
}
