/** @file GseField.h Extracting data from binary blobs */
#ifndef _GseField_H_
#define _GseField_H_

#include <rpwgse/GseProp.h>


#ifdef __cplusplus
extern "C" {
#endif

/* typedef enum _gse_byteorder{ msb_first, lsb_first } gse_byteorder_t; */

/** @addtogroup mod_bin_parser */
/*@{*/

/* Alarm status enum.
 *
 * @note @a no_alarm is 0
 */
typedef enum gse_alarm_stat_t{
	no_alarm=0, err_lo, warn_lo, warn_hi, err_hi, inv_alarm_stat
}gse_alarm_stat_t;

/** Returns the alarm enum value that corresponds to a given string.
 *
 * @param sVal a pointer to one of the following strings:
 *   - "err_lo"
 *   - "warn_lo"
 *   - "normal"
 *   - "warn_hi"
 *   - "err_hi"
 *
 * @return The corresponding alarm enum, or @a inv_alarm_stat if the string is
 *         not one of the above.
 */
gse_alarm_stat_t gse_str2alarm(const char* sVal);

#ifndef _GseParser_H_
struct gse_parser;
#endif

/** Single Field information extractor.
 *
 * These are owned by parsers.  GseFields in turn own GseProp objects.
 * @n
 * @class GseField
 */
typedef struct gse_field{
	 	
	/** Get the name of this field.
	 * @returns The name of the field or a pointer to the string "unnamed".
	 */
	const char* (*name)(const struct gse_field* pThis);
	
	/** Gse the description for this field.
	 *
	 * @returns A pointer to the description string, or NULL in no 
	 * description has been defined
	 */
	const char* (*desc)(const struct gse_field* pThis);
	
	
	/** Is this field ready for byte to text conversion?
	 *  
	 *
	 * @returns A pointer to an error message allocated on the heap if the
	 *          field is not ready for use, or NULL if there are no errors to
	 *          report.
	 */
	char* (*validate)(struct gse_field* pThis);

	
	/* Getting Data from the binary blob.
	 *
	 * @returns NULL if the read was successful, or an error message
	 *          allocated on the heap if not.
	 */
	char* (*read)(struct gse_field* pThis, cbytebuf_t cbuf);
	
	/** Return the type of data.
	 *
	 * @note This is @b NOT necessarily the same as the read type in
	 *       a config file.  Calibration polynomials can cause type
	 *       promotion of integers to floats if they contain floating
	 *       point values.
	 *
	 * @return One of:
	 *   - 'B' The data is unconvertable, but raw bytes may be retrieved
	 *   - 'U' the data from the binary blob is an unsigned integer
	 *   - 'S' the data from the binary blob is a signed integer
	 *   - 'F' the data from the binary blob is an IEEE float
	 *   - 'C' the data from the binary blob is ascii charaters
	 *  or 0 if the 'ty' property has not been set.
	 */
	char (*valType)(const struct gse_field* pThis);
	
	/** Get the field data as a byte buffer.
	 * This always works after a read.  Shifting and masking are applied
	 * but the field has the same with as intitally described.
	 */
	cbytebuf_t (*bval)(const struct gse_field* pThis);	
	
	/** Get a string representation of the data.
	 *
	 * This always works after a read, for binary data ASCII Hex Bytes
	 * are returned.
	 */
	const char* (*sval)(const struct gse_field* pThis);
	
	/** Get a floating point representation for the data.
	 *
	 * This works after a read.  
	 * @note This is the apporiate function to use for 'S' or 'U' types that
	 *       have the 'poly' property.  Polynominal calibration promotes
	 *       integer types to floating types internally.
	 *
	 * @warning This reads integer types as floats, which may cause a loss
	 *          of precision.  As indicated by the "const" pointer,
	 *          the stored value is not affected.  For 'B' and 'C' types
	 *          the first byte is cast as an integer then stored in a float.
	 */
	float8 (*fval)(const struct gse_field* pThis);
	
	/** Get an unsigned int representation of the data.
	 * 
	 * This works after a @a read.  
	 *
	 * @note Use this to get the @b uncalibrated values for integer fields
	 *       with a 'poly'nomial property, and to get the un 'enum'erated
	 *       values.
	 *
	 * @warning For negative values truncates to 0, for floating point
	 *          values > than @c GSE_UINT8_MAX, @c GSE_UINT8_MAX is returned.
	 *          For 'B' and 'C' data types, the first byte is cast and returned.
	 */
	uint8 (*uval)(const struct gse_field* pThis); 
	
	/** Get an integer representation of the data.
	 * 
	 * This works after a @a read.
	 *
	 * @note Use this to get the @b uncalibrated values for integer fields
	 *       with a 'poly'nomial property, and to get the un 'enum'erated
	 *       values.
	 *
	 * @warning For floating point values < GSE_INT8_MIN, GSE_INT8_MIN is
	 *          returned.  For unsigned or floating point values > than 
	 *          @c GSE_UINT8_MAX, @c GSE_UINT8_MAX is returned.  For 'B' and 'C'
	 *          data types, the first byte is cast and returned.
	 */
	int8 (*ival)(const struct gse_field* pThis); 

	
	/** Get the alarm status of converted values
	 *
	 * This works after a read for fields of type 'F' 'U' and 'S'.
	 *
	 * @note the first enum element of gse_alarm_stat_t is no_alarm which
	 *       has value 0, thus the following is permissible:
	 * @code
	 *  if( pField->alarm(pField) ){
	 *   // Handled alarm conditions
	 *  }
	 * @endcode
	 *
	 * @returns the alarm status. 
	 */
	gse_alarm_stat_t (*alarm)(const struct gse_field* pThis);

	
	/* Raw property functions */
	 
	/** Get the 1st property in this field's dictionary of properties.
	 * 
	 * This is useful when iterating over all the properties of a field.
	 * See the example in @ref GseField::next().
	 */
	const GseProp* (*first)(const struct gse_field* pThis);
	
	/** Get the next property in this field's dictionary of properties.
	 *
	 * Used along with GseField::first to iterate over all the properties
	 * of a field.
	 * @code
	 *
	 * cosnt GseField* pField;
	 * const GseProp* pProp;
	 *
    * for(pProp = pField->first(pField); pProp != NULL; 
	 *     pProp = pField->next(pField, pProp))
	 * {
	 *    //Do something with pProp...
	 *
	 * }
	 * @endcode
	 */
	const GseProp* (*next)(const struct gse_field* pThis, const GseProp* pProp);
	
	/** Does this field have a named property */
	bool (*hasProp)(const struct gse_field* pThis, const char* sName);
	
	/** Get a named property */
	const GseProp* (*getProp)(const struct gse_field* pThis, 
			                         const char* sName);
	
	/** Adds a property to the internal dictionary of properties.*/
	void (*addProp)(struct gse_field* pThis, GseProp* pProp);
	
	/** Remove a property from this field. */
	GseProp* (*takeProp)(struct gse_field* pThis, const char* sName);
	
	
		
	/* Destructor */
	void (*del)(void* vpThis);
	
	/* PRIVATE ***************************************************** */
	/* Don't screw with this stuff directly, use the functions above */
	
	/* Usually doesn't change for the life of the instance */
	
	char* sName;             /* The name of this field */
	
	const struct gse_parser* pParent;  /* The parser owning this field */
	
	const char* sConf;       /* File containing the field definition */
	
	int4 nLine;              /* Line number of the field definition */
	
	
	/* Can change upto once per validation */
	
	GseStrPropDict* pProps;  /* The property collection */
	
	bool bNeedsVal;          /* Has this been validated? */
	
	byte uHiBitIdx;          /* What is the highest non-zero bit in 'or' */
	
	char cEffType;           /* That is the effective type of this data 
	                            after type promotion from cal polynomials */
	
	/* Changes once per read */
	
	bool bNeedsRead;         /* Has data been read? */
	
	bytebuf_t buf;           /* Raw byte buffer for holding read in values */
	
	char* sVal;              /* Pointer for holding converted values */
	
	gse_alarm_stat_t alrm;   /* Current alarm status */
	
	float8 fVal;             /* Stored float conversion, ty 'F' and 
	                            'U' or 'S' when 'poly' is defined */
	
	int8   iVal;             /* Stored signed conversion */
	
	uint8  uVal;             /* Stored unsigned conversion */
	
} GseField;

/** Make a new field
 * 
 * @warning The GseField takes owership of the property array, so deleting
 *          it outside the extrator destructor will cause a propram crash!
 *          Also, the property array must be allocated using HEAP memory.
 * 
 * @param sName The name of this field, can not be null
 *
 * @param pProps A starting dictionary of properties, note the GseField
 *        takes ownership of this dictionary.
 *
 * @param pDefs A set of default properties, to be used if a property with
 *        that key is not in pProps.  Note that the field does @b NOT take
 *        ownership of the defaults, instead properties in @a pDefs are
 *        copied to the internal property array.
 * 
 * @param nSrcLine If the field is built from a configuration file you
 *        can include the line number here and it will be included in any
 *        error messages.
 *
 * @return A new GseField allocated on the Heap.
 *
 * @public
 * @memberof GseField
 */
GseField* new_GseField(const char* sName, GseStrPropDict* pProps, 
                       GseStrPropDict* pDefs, int4 nSrcLine);
							  

/* A dictionary of fields, accessed by strings */
#define S gse_str_field_dict
#define C GseStrFieldDict
#define K char
#define V GseField

#include "ptrDictTplt.h"

#undef S
#undef C
#undef K
#undef V

/** Sets the config file name for all the fields in the dictionary.
 *
 * This recursively sets the config file name for all the parameters within
 * each field
 *
 * @param pDict the property dictionary in question.
 * @param sConfFile a constant pointer to the name of the config file.
 * 
 * @warning The string @a sFile is @b NOT copied to internal memory. It is
 *          the responsibility of the caller to insure that pointer is valid
 *          for the lifetime of the fields in the dictionary.
 *
 * @public
 * @memberof GseStrFieldDict
 */
void gse_fd_setconfig(GseStrFieldDict* pDict, const char* sConfFile);


void gse_fd_setparser(GseStrFieldDict* pDict, const struct gse_parser* pParent);
 
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseField_H_ */

