#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseString.h"

#include "GseFieldPrn.h"


/*****************************************************************************/
/* The direct printer, for fixed length character data */

char* gse_fmtval_retnull(const GseField* pField){ return NULL; }

char* gse_fldprn_direct(const GseField* pField)
{
	char* sRet;
	int i;
	
	sRet =  (char*) gse_calloc(pField->buf.len + 1, sizeof(char)); /* Get's the null */
	memcpy(sRet, pField->buf.ptr, pField->buf.len);
	
	for(i=0; i<pField->buf.len; i++){
		if( (sRet[i] > 126) ||
			                  /* Tab            Newline        Carrage Return */
		    ((sRet[i] < 32) && (sRet[i]!=9) && (sRet[i]!=10) && (sRet[i]!=13) )
		)
			sRet[i] = '*';
	}
	
	return sRet; 
}

/*****************************************************************************/
/* The single line hex printer */

char* gse_fldprn_hexprn(const GseField* pField)
{
	return gse_dyn_hexprn( pField->bval(pField) );
}

/*****************************************************************************
 * The printf printer.  This is the default printer and it will just make
 * up a format if none is present for the field types it supports.
 */

const char* gse_fldfmt_float_specs[] = {"e","E","f","F","g","G", NULL};
const char* gse_fldfmt_usign_specs[] = {"llo","llu","llx","llX", NULL};
const char* gse_fldfmt_sign_specs[]  =  {"lld","lli", NULL};

char* gse_fmtval_printf(const GseField* pField)
{
	char cType = '\0';
	const char* sFmt  = NULL;
	const char* pSpec  = NULL;
	const char* pStart = NULL;
	const char* pStop  = NULL;
	
	const char** ppSpecs = NULL;
	const char* sValidSpecs = NULL;
	
	
	/* Get the effective type (print type) of the field */
	cType = pField->valType(pField);
	
	if( (cType == 'B')||(cType == 'C'))
		return CS("Currently the 'printf' pfunc does not support 'B'inary or "
		          "'C'haratcer fields, use the 'direct' or 'hex' printer instead");
	
	if( ! gse_pd_cktype(pField->pProps, pt_str, "pfmt") )
		return CS("'printf' printer expects property 'pfmt' to be a string.");
	
	if( (sFmt = gse_pd_getstr(pField->pProps, "pfmt")) == NULL)
		return NULL;
		
		
	/* Make sure % is in the string */
	if( (pStart = strchr(sFmt, '%')) == NULL)
		return NS("Format value, '%s', is missing the '%%' character", sFmt);
				
	/* Make sure a conversion specifier is present */
	if(cType == 'F' ||
		(((cType == 'U')||(cType == 'S'))&&pField->hasProp(pField, "cal")) ){
		ppSpecs = gse_fldfmt_float_specs;
		sValidSpecs = "e, E, f, F, g, G";
	}
	else{
		if(cType == 'U'){
			ppSpecs = gse_fldfmt_usign_specs;
			sValidSpecs = "llo, llu, llx, llX";
		}
		else{
			ppSpecs = gse_fldfmt_sign_specs;
			sValidSpecs = "lld, lli";
		}
	}
	
	pSpec = *ppSpecs;
	while(pSpec != NULL){
		if( (pStop = strstr(sFmt, pSpec)) != NULL)
			break;
		pSpec = *( ppSpecs++) ;
	}
	
	if(! pStop )
		return NS("Format value, '%s', is missing a valid specifier, expecting "
		          "one of %s", sFmt, sValidSpecs);
	
	if( pStop <= pStart)
		return NS("Invalid format property, '%s'", sFmt);
	
	/* We are no where near guarunting a valid format specifier at this point,
	   but this should catch many casual errors */
	
	return NULL;
}

char* gse_fldprn_printf(const GseField* pField)
{
	char cType;
	const char* sFmt;
	
	cType = pField->valType(pField);
	
	if( (cType == 'B')||(cType == 'C'))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "The 'printf' printer is not "
				       "compatible with type %c fields ", cType);
	
	sFmt = gse_pd_getstr(pField->pProps, "pfmt");
	
	if(sFmt != NULL){
		switch(cType){
			case 'F': return NS(sFmt, pField->fval(pField));
			case 'U': return NS(sFmt, pField->uval(pField));
			case 'S': return NS(sFmt, pField->ival(pField));
			default: gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		}
	}
	else{
		switch(cType){
			case 'F': return NS("%e",   pField->fval(pField));
			case 'U': return NS("%llX", pField->uval(pField));
			case 'S': return NS("%lld", pField->ival(pField));
			default: gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		}
	}
	return NULL;
}

/*****************************************************************************/
/* The enum printer */

char* gse_fmtval_enum(const GseField* pField)
{
	char cType;
	const GseProp* pProp;
	
	/* Make sure efftype is S or U */ 
	cType = pField->valType(pField);
	
	if( (cType == 'B')||(cType == 'C')||(cType == 'F'))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "The 'enum' printer is not "
				       "compatible with type %c fields ", cType);
	
	/* Make sure the fromat property is present */
	if( (pProp = pField->getProp(pField, "pfmt")) == NULL)
		return CS("Missing 'pfmt' property for the 'enum' printer");
	
	/* Make sure the format property is an enum */	
	if( ! gse_pd_cktype(pField->pProps, pt_enum, "pfmt") )
		return CS("'enum' printer expects property 'pfmt' to be of type enum");
	
	return NULL;
}

char* gse_fldprn_enum(const GseField* pField)
{
	char cType;
	const GsePropEnum* pFmt;
	
	cType = pField->valType(pField);
	
	pFmt = gse_pd_getenum(pField->pProps, "pfmt");
	
	if( (cType == 'B')||(cType == 'C')||(cType == 'F') )
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "The 'enum' printer is not "
				       "compatible with type %c fields ", cType);
	
	if(cType == 'U')
		return pFmt->u8ToStr(pFmt, pField->uVal);
	else
		return pFmt->i8ToStr(pFmt, pField->iVal);
}		

/*****************************************************************************/
/* The gmtime printer */ 
 
char* gse_fmtval_gmtime(const GseField* pField)
{
	char cType;
	const GseProp* pFmt;
	const char* sFmt;
	struct tm test;
	char sTest[60];
	sTest[59] = '\0';
	
	cType = pField->valType(pField);
	
	if( (cType == 'B')||(cType == 'C'))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "The 'gmtime' printer is not "
				       "compatible with type %c fields ", cType);
	
	/* If the format field is not specified, make one up */
	if( (pFmt = pField->getProp(pField, "pfmt")) == NULL)
		return NULL;
	
	/* If present, must be a string */
	if( pFmt->ptype != pt_str)
		return CS("'gmtime' printer expects the 'pfmt' property to be a string");
	
	sFmt = ((GsePropStr*)pFmt)->sVal;
	
	memset(&test, '\0', sizeof(struct tm));
	test.tm_mday = 1;
	
	if( strftime(sTest, 59, sFmt, &test) == 0)
		return NS("Error in time format string: '%s'", sFmt);
	
	return NULL;
}

char* gse_fldprn_gmtime(const GseField* pField)
{
	char cType;
	const GseProp* pFmt;
	const char* sFmt;
	struct tm bdt;
	time_t epoch;
	char sBuf[128];
	sBuf[127] = '\0';
	
	cType = pField->valType(pField);
	
	if( (cType == 'B')||(cType == 'C'))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "The 'gmtime' printer is not "
				       "compatible with type %c fields ", cType);
	
	epoch = pField->ival(pField);
	
	if( gmtime_r(&epoch, &bdt) == NULL)
		return CS("gmtime_r ERR");
	
	/* You did validate 1st right? */
	if( (pFmt = pField->getProp(pField, "pfmt")) == NULL)
		sFmt = "%Y-%m-%dT%H:%M:%S";
	else
		sFmt = ((GsePropStr*)pFmt)->sVal;
	
	strftime(sBuf, 127, sFmt, &bdt);
	
	return CS(sBuf);
}


/*****************************************************************************/
/* Gobal Stuff, not editable in this version of the library */

struct gse_field_prn_set{
	const char* sName;
	gse_field_printer* printer;
	gse_format_validator* validator;
};

const struct gse_field_prn_set gse_printerAry[] = {
	{"printf", gse_fldprn_printf,  gse_fmtval_printf },
	{"enum"  , gse_fldprn_enum,    gse_fmtval_enum   },
	{"direct", gse_fldprn_direct,  gse_fmtval_retnull},
	{"hex"   , gse_fldprn_hexprn,  gse_fmtval_retnull},
	{"gmtime", gse_fldprn_gmtime,  gse_fmtval_gmtime },
	{0}
};

bool gse_has_field_printer(const char* sName)
{
	int i = 0;
	
	while(true){
		if(gse_printerAry[i].sName == NULL) return false;
		if(strcmp(gse_printerAry[i].sName, sName) == 0) return true;
		i++;
	}
	return false;
}

void gse_add_field_printer(const char* sName, gse_field_printer* pPrinter, 
		                     gse_format_validator* pValidator)
{
	gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Adding new field printers at run-time"
	            " would require per thread field printer arrays.  This has been "
	            " skipped for now.");
}

char* gse_print_field(const char* sFunc, const GseField* pField)
{
	int i = 0;
	
	if(! sFunc || !pField)
	gse_throw(GSE_XCEPT, GSE_XBADARG);

	
	while(true){
		if( gse_printerAry[i].sName == NULL){
			gse_throw_fm(GSE_XCEPT, GSE_XNOKEY, "Printing function '%s' is not "
					       "defined.", sFunc);
		}
		
		if( strcmp(gse_printerAry[i].sName, sFunc) == 0)
			return gse_printerAry[i].printer(pField);
		
		i++;
	}
	return NULL;
}


char* gse_validate_fmt(const char* sFunc, const GseField* pField)
{
	int i = 0;
	
	if(! sFunc || !pField)
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	while(true){
		if( gse_printerAry[i].sName == NULL){
			gse_throw_fm(GSE_XCEPT, GSE_XNOKEY, "Printing function '%s' is not "
					       "defined.", sFunc);
		}
		
		if( strcmp(gse_printerAry[i].sName, sFunc) == 0)
			return gse_printerAry[i].validator(pField);
		
		i++;
	}
	return NULL;
}
