#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <limits.h>
#include <utime.h>

#include "GseFile.h"
#include "GseWrap.h"
#include "GseExcept.h"
#include "GseString.h"


bool gse_at_eof(int fd)
{
	size_t uLast;
	size_t uPos = gse_lseek(fd, 0, SEEK_CUR);
	uLast = gse_lseek(fd, 0, SEEK_END);
	gse_lseek(fd, uPos, SEEK_SET);
	return (uLast == uPos);
}

size_t gse_lseek_bnds(int fd, size_t offset, int whence){
	size_t cur, new, end;
	
	if( (cur = gse_ltell(fd)) == (off_t)-1)
		gse_throw(CLIB_XCEPT, errno);
	
	end = lseek(fd, 0, SEEK_END);
	
	switch(whence){
		case SEEK_SET: new = offset; break;
		case SEEK_CUR: new = cur + offset; break;
		case SEEK_END: new = end + offset; break;
		default:
				lseek(fd, cur, SEEK_SET);
				gse_throw(GSE_XCEPT, GSE_XBADARG);
	}
	
	if(new > end) new = end;
	if(new < 0)   new = 0;
	
	return lseek(fd, new, SEEK_SET);
}

/*****************************************************************************/
/* Restart Controlled reads/writes */

size_t gse_read(int fd, void *buf, size_t count, const bool* restart)
{
	ssize_t nRet;
	off_t uPos;
			
	uPos = lseek(fd, 0, SEEK_CUR);
		
	while(1){
		lseek(fd, uPos, SEEK_SET);
		
		if( (nRet = read(fd, buf, count)) >= 0)
			break;
		
		if(errno != EINTR)
			gse_throw_m(CLIB_XCEPT,errno,"Could not read from file.");
			
		if((restart == NULL) || (*restart == false)){
			lseek(fd, uPos, SEEK_SET);
			return 0;
		}
	}
	
	return nRet;
}

size_t gse_write(int fd, const void *buf, size_t count, const bool* restart)
{
	ssize_t nRet;
	off_t uPos;
			
	uPos = lseek(fd, 0, SEEK_CUR);
		
	while(1){
		lseek(fd, uPos, SEEK_SET);
		
		if( (nRet = write(fd, buf, count)) >= 0)
			break;
		
		if(errno != EINTR)
			gse_throw_m(CLIB_XCEPT,errno,"Could not write to file.");
			
		if((restart == NULL) || (*restart == false)){
			lseek(fd, uPos, SEEK_SET);
			return 0;
		}
	}
	
	return nRet;
}

/*****************************************************************************/
/* Path item info */
bool gse_isdir(const char* path)
{
	struct stat stbuf;
	if(stat(path, &stbuf) != 0) return false;
	
	if(S_ISDIR(stbuf.st_mode)) return true;
	else return false;
}

bool gse_isfile(const char* path)
{
	struct stat stbuf;
	if(stat(path, &stbuf) != 0) return false;
	
	if(S_ISREG(stbuf.st_mode)) return true;
	else return false;
}

bool gse_exists(const char* path)
{
	struct stat stbuf;
	
	if(stat(path, &stbuf) == 0) return true;
	else return false;
}

/*****************************************************************************/
/* Making directories */

void gse_mkdirs(const char* path, mode_t mode)
{
	char* sPath = NULL;
	size_t i, len;
	int nErr;
	
	if((path == NULL)||(path[0] == '\0'))
		gse_throw_m(CLIB_XCEPT, ENOENT, "path is NULL or empty");
	
	/* Get all the parent directories */
	sPath = CS(path);
	len = strlen(path);
	
	for(i = 0; i<len; i++){
		if((sPath[i] == '/') && (i>0) && (i < len-1)){
			sPath[i] = '\0';
			
			if(! gse_isdir(sPath) ){
#ifndef WIN32
				if(mkdir(sPath, mode) != 0)
#else
				if(mkdir(sPath) != 0)
#endif
				{
					nErr = errno;
					gse_throw_fm(CLIB_XCEPT, nErr, "Cannot make directory '%s'",
					             sPath);
				}
			}
			
			sPath[i] = '/';
		}
	}
	
	/* Make the leaf */
	if(! gse_isdir(sPath) ){
				
#ifndef WIN32
		if(mkdir(sPath, mode) != 0)
#else
		if(mkdir(sPath) != 0)
#endif
		{
			nErr = errno;
			gse_throw_fm(CLIB_XCEPT, nErr, "Cannot make directory '%s'", sPath);
		}
	}
}

void gse_mkdirsto(const char* path, mode_t mode)
{
	char* sPath = NULL;
	size_t i, len;
	int nErr;
	
	/* Determine the dirmode from the file mode */
#ifndef WIN32
	mode_t dirmode;
	dirmode = mode | S_IRWXU;
	if(dirmode & S_IRGRP) dirmode |= S_IXGRP;
	if(dirmode & S_IROTH) dirmode |= S_IXOTH;
#endif

	if((path == NULL)||(path[0] == '\0'))
		gse_throw_m(CLIB_XCEPT, ENOENT, "path is NULL or empty");
	
	/* Get all the parent directories */
	sPath = CS(path);
	except_cleanup_push(free, sPath);
	
	len = strlen(path);
	
	for(i = 0; i<len; i++){
		if((sPath[i] == '/') && (i>0) && (i < len-1)){
			sPath[i] = '\0';
			
			if(! gse_isdir(sPath) ){
#ifndef WIN32
				if(mkdir(sPath, mode) != 0)
#else
				if(mkdir(sPath) != 0)
#endif
				{
					nErr = errno;
					gse_throw_fm(CLIB_XCEPT, nErr, "Cannot make directory '%s'",
					             sPath);
				}
			}
			
			sPath[i] = '/';
		}
	}
	except_cleanup_pop(except_call);
}

/*****************************************************************************/
/* Read/Write buffers to files */

bytebuf_t gse_readfile(const char* sPath)
{
	bool bRestart = true;
	return gse_readfile_i(sPath, &bRestart);
}

bytebuf_t gse_readfile_i(const char* sPath, const bool* restart)
{
	int fd;
	bytebuf_t buf;
	size_t nRet;
	buf.ptr = NULL;
	
	/* Note, the first call to read will throw for us if this is a dir. */
	
	fd = gse_open_ro(sPath);
	buf.len = gse_lseek(fd, 0, SEEK_END);
	gse_lseek(fd, 0, SEEK_SET);
	buf.ptr = gse_calloc(buf.len, 1);
	
	except_cleanup_push(free, buf.ptr);
	
	nRet = gse_read(fd, buf.ptr, buf.len, restart); 
	if((nRet != buf.len)&&(restart != NULL)&&(*restart == true))
		gse_throw(GSE_XCEPT, GSE_XASSERT);
	
	/* Save some memory if we can */
	if(nRet != buf.len){
		buf.len = nRet;
		
		if(nRet > 0){
			buf.ptr = gse_realloc(buf.ptr, buf.len);
		}
		else{
			free(buf.ptr);
			buf.ptr = NULL;
		} 
	}
	
	except_cleanup_pop(except_no_call);
	return buf;
}

void gse_writefile(const char* path, cbytebuf_t data, mode_t mode)
{
	int fd;
	bool bRestart = true;
	
	gse_mkdirsto(path, mode);
	
	/* Make the file */
	fd = gse_open(path, O_CREAT|O_RDWR, mode);
	
	gse_write(fd, data.ptr, data.len, &bRestart);
}

/*****************************************************************************/
/* Walking the file tree */

int _gse_walk_helper(char* fullpath, gse_walk_action action, void* user_data)
{
	struct stat stbuf;
	int nRet;
	char* ptr;
	DIR* pDir;
	struct dirent* pEnt;
	
	if(stat(fullpath, &stbuf) < 0)
		return action(user_data, fullpath, NULL);
	
	if(S_ISDIR(stbuf.st_mode) == 0)
		return action(user_data, fullpath, &stbuf);
	
	/* First do the action for the directory then iterate over all the items
	   in the directory */
	
	if( (pDir = opendir(fullpath)) == NULL)
		return action(user_data, fullpath, NULL);
	
	if( (nRet = action(user_data, fullpath, &stbuf)) != 0)
		return nRet;
	
	ptr = fullpath + strlen(fullpath); /* point to the end of fullpath */
	*ptr = '/';
	ptr += 1;
	*ptr = '\0';
	
	while( (pEnt = readdir(pDir)) != NULL){
		
		if((strcmp(pEnt->d_name, ".") == 0)||(strcmp(pEnt->d_name, "..") == 0))
			continue;
		
		strcpy(ptr, pEnt->d_name);
		
		if( (nRet = _gse_walk_helper(fullpath, action, user_data)) != 0)
			break;
	}
	*(ptr - 1) = '\0'; /* erase from the slash onwards */ 
	
	closedir(pDir);
	
	return nRet;
} 

int gse_walk(const char* path, gse_walk_action action, void* user_data)
{
	char fullpath[PATH_MAX + 1];
   fullpath[PATH_MAX] = '\0';
	
	if((path == NULL)||(action == NULL))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	strncpy(fullpath, path, PATH_MAX-1); 
	
	return _gse_walk_helper((char*)fullpath, action, user_data);
}

/*****************************************************************************/
bool gse_basename(const char* sPath, char* sBasename, size_t nLen)
{
	int4 i, iSlash = -1;
	int4 nPathLen = 0;
	
	if(nLen < 2)
		return false;
	
	memset(sBasename, 0, nLen);
	
	/* If sBasename contains no characters, return "NULL" */
	if(sPath == NULL || sPath[0] == '\0'){
		return false;
	}
	
	nPathLen = strlen(sPath);
	
	/* If last character is '/' return NULL */
	if( sPath[nPathLen - 1] == '/')
		return false;
	
		
	/* If has no '/' chars just copy everthing */
	iSlash = -1;
	for(i = nPathLen - 1; i >= 0; i--){
		if(sPath[i] == '/'){
			iSlash = i;
			break;
		}
	}
	
	if(iSlash == -1){
		if(nLen > nPathLen){
			strncpy(sBasename, sPath, nPathLen);
			return true;
		}
		return false;
	}
	
	/* Copy over the chars back to the '/' character */
	if(nLen > nPathLen - (iSlash + 1) ){
		strncpy(sBasename, sPath + iSlash + 1, nPathLen - (iSlash + 1));
		return true;
	}
	return false;
}

/*****************************************************************************/
char* gse_bname(const char* sPath)
{
	
	int4 i, iSlash, nPathLen = 0;
	
	/* If sPath contains no characters, return "NULL" */
	if(sPath == NULL || sPath[0] == '\0'){
		return NULL;
	}
	
	nPathLen = strlen(sPath);
	
	/* If last character is '/' return NULL */
	if( sPath[nPathLen - 1] == '/')
		return NULL;
	
	/* Find position of last '/' */
	iSlash = -1;
	for(i = nPathLen - 1; i >= 0; i--){
		if(sPath[i] == '/'){
			iSlash = i;
			break;
		}
	}
	
	/* If has no '/' chars just copy everthing */
	if(iSlash == -1)
		return CS(sPath);
	
	/* Copy over the chars back to the '/' character */
	return CS(sPath + iSlash + 1);
}



/*****************************************************************************/
char* gse_dname(const char* sPath)
{
	int4 i, iSlash, nPathLen = 0;
	bool bAllSlash = true;
	char* sDir = NULL;
	
	/* If sPath contains no characters, return "NULL" */
	if(sPath == NULL || sPath[0] == '\0'){
		return NULL;
	}
	
	nPathLen = strlen(sPath);
	
	/* Special Case, if input is all /'s return an empty string 
	   (Not a null string) */
	
	for(i = 0; i<nPathLen; i++){
		if(sPath[i] != '/'){
			bAllSlash = false;
			break;
		}
	}
	if(bAllSlash)
		return (char*) gse_calloc(4, sizeof(char));
	
	
	
	/* Find position of last '/' */
	iSlash = -1;
	for(i = nPathLen - 1; i >= 0; i--){
		if(sPath[i] == '/'){
			iSlash = i;
			break;
		}
	}
	
	/* If has no '/' chars there is no directory info */
	if(iSlash == -1)
		return NULL;
	
	/* Copy over the chars up to the '/' character */
	sDir = (char*)gse_calloc(iSlash + 1, sizeof(char));
	for(i = 0; i<iSlash; i++)
		sDir[i] = sPath[i];
	
	return sDir;
}


/*****************************************************************************/
char* gse_getcwd()
{
	size_t nSz = 128; /* start small */
	char* pBuf = NULL;
	
	
	while(pBuf == NULL){
		
		pBuf = (char*) gse_calloc(nSz, sizeof(char)); 
		
		errno = 0;
		if(!getcwd(pBuf, nSz)){
			free(pBuf);
			pBuf = NULL;
			
			if(errno == ERANGE)
				nSz *= 2;
			else
				gse_throw(CLIB_XCEPT, errno);	
			
		}
	}
	
	return pBuf;
}


/*****************************************************************************/
char* gse_abspath(const char* sInput)
{
	size_t nLen;
	size_t nCwdLen;
	char* sBuf = NULL;
	char* sCwd = NULL;
	char* pChar = NULL;
	bool bSimplify = false;
	bool pKeep[30];
	GseStrAry* pParts = NULL;
	char* sBuf2 = NULL;
	int i, iLastGood, j;
	
	if((sInput == NULL)||(sInput[0] == '\0'))
		return NULL;
	
	nLen = strlen(sInput);
	
	/* If doesn't start with /, add the cwd */
	if(sInput[0] != '/'){
		sCwd = gse_getcwd();
		nCwdLen = strlen(sCwd);
		sBuf = (char*) gse_calloc(strlen(sCwd) + nLen + 2, sizeof(char));
		strcpy(sBuf, sCwd);
		free(sCwd);
		sCwd = NULL;
		sBuf[nCwdLen] = '/';
		strcpy(sBuf + nCwdLen + 1, sInput);
	}
	else{
		sBuf = CS(sInput);
	}
	
	/* Okay, now to see if we need reductions */
	pChar = sBuf + 1;
	while(*pChar != '\0'){			
		if( *(pChar-1) == '.' && (*pChar == '/' || *pChar == '.')){
			bSimplify = true;
			break;
		}
		pChar++;
	}
	
	if(!bSimplify) return sBuf;
	
	except_cleanup_push(free, sBuf);
	
	
	pParts = gse_strsplit(sBuf, "/");
	except_cleanup_push(pParts->del, pParts);
	
	if(pParts->num > 30)
		gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "Currently gse_abspath cannot "
		            "handle an arbitrary number of path componets.");
	
	
	/* Figure out which path components are worth keeping */
	iLastGood = -1;
	memset(pKeep, false, 30);
	
	for(i=0;i<pParts->num;i++){
		if(strcmp(pParts->pAry[i], ".") == 0)
			continue;
		
		if(strcmp(pParts->pAry[i], "..") == 0){
			if(iLastGood < 0){
				
				/* Make a reasonable error message */
				pChar = sBuf;
				for(j=0; j<=i; j++){
					*pChar = '/';
					pChar++;
					strcpy(pChar, pParts->pAry[j]);
					pChar += strlen(pParts->pAry[j]);
				}
				*pChar = '\0';
				
				gse_throw_fm(CLIB_XCEPT, ENOTDIR, "Directory %s is above the "
				             "root directory.", sBuf);
			}
			
			pKeep[iLastGood] = false;
			
			/* Lost that one, try to find a previous good one */
			for(j = iLastGood - 1; j> -1; j--){
				if(pKeep[j]){
					iLastGood = j;
					break;
				}
			}
			if(j == -1) iLastGood = j;  /* went to 1st of list, nothing's okay */
				
		}
		else{
			pKeep[i] = true;
			iLastGood = i;
		}
	}
	
	
	/* Figure out how long the string will be */
	nLen = 1;
	for(i=0;i<pParts->num;i++){
		if(pKeep[i])
			nLen += strlen(pParts->pAry[i]) + 1;
	}
	
	/* Add extra byte to handle the case of the root directory */
	sBuf2 = (char*)gse_calloc( nLen + 1, sizeof(char));
				
	if(pParts->num == 0){
		sBuf2[0] = '/';
	}
	else{
		
		pChar = sBuf2;

		for(i=0;i<pParts->num;i++){
			if(! pKeep[i]) continue;
			
			*pChar = '/';
			pChar++;
				
			strcpy(pChar, pParts->pAry[i]);
			pChar += strlen(pParts->pAry[i]);
		}
	}
	
	except_cleanup_pop(except_call); /* pParts */
	except_cleanup_pop(except_call); /* sBuf */
	
	return sBuf2;
}

/*****************************************************************************/

GseStrList* gse_listdir(const char* path, int nType)
{
	GseStrList* pList = NULL;
	int nErrno = 0;
	DIR* pDir = NULL;
	struct dirent* pEnt = NULL;
	char* sItemPath;
	
	if( (pDir = opendir(path)) == NULL){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Can't read directory %s", path);
	}
	
	while( (pEnt = gse_readdir(pDir)) != NULL){
		
		if((strcmp(pEnt->d_name, ".") == 0)||(strcmp(pEnt->d_name, "..") == 0))
			continue;
		
		sItemPath = NS("%s/%s", path, pEnt->d_name);

		if(gse_isdir(sItemPath) && ((nType & 0x2) == 0)){
			free(sItemPath);
			continue;
		}
		
		if(gse_isfile(sItemPath) && ((nType & 0x1) == 0)){
			free(sItemPath);
			continue;
		}
		
		if(pList == NULL) pList =  new_GseStrList();
		
		pList->append(pList, sItemPath);
	}
	
	closedir(pDir);
	
	if(pList != NULL) pList->sort(pList, strcmp);
	
	return pList;
}

/*****************************************************************************/
/* Boy this is messy, seems to take too much code */

size_t gse_copyfile(const char* sPath, const char* sDest, mode_t mode)
{
	char sPathBase[256];
	char* sOut = NULL;
	int nOut = -1;
	
	FILE* pIn;
	FILE* pOut;
	byte buf[4096];
	
	size_t uSz = 0;
	size_t uRead = 0;
	size_t uWritten = 0;
	
	int nErr; 
	
	if((sPath == NULL)||(sDest == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null source or Dest path");

	if(mode == 0)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "mode == 0");
	
	pIn = gse_fopen(sPath, "rb");
	
	
	if(gse_isdir(sDest)){
		if( ! gse_basename(sPath, sPathBase, 255) )
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Can't determine the basename "
			             "of '%s'.", sPath);
		
		sOut = NS("%s/%s", sDest, sPathBase);
	}
	else{
		sOut = (char*)sDest;
	}
	
	nOut = gse_open(sOut, O_CREAT|O_TRUNC|O_WRONLY, mode);
	
	if( (pOut = fdopen(nOut, "wb")) == NULL){
		nErr = errno;
		fclose(pIn);
		if(sOut != sDest) free(sOut);
		
		gse_throw_fm(CLIB_XCEPT, nErr, "Error opening %s", sOut);
	}
	
	uSz = 0;
	while( (uRead = fread(buf, sizeof(byte), 4096, pIn)) != 0){
	
		uWritten = fwrite(buf, sizeof(byte), uRead, pOut);
		
		if(uWritten != uRead){
			fclose(pIn);
			fclose(pOut);
			if(sOut != sDest) free(sOut);
			gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "Write to '%s' failed.", sOut); 
		}
		
		uSz += uWritten;
	}
	
	if( ! feof(pIn) ){
		fclose(pIn);
		fclose(pOut);
		if(sOut != sDest) free(sOut);
		gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "Failed to read all bytes from '%s'", 
		             sPath);
	}
	
	fclose(pIn);
	fclose(pOut);	
	if(sOut != sDest) free(sOut);
	
	return uSz;
}

/*****************************************************************************/
/* Copy file and preserve c and m timestamps */

size_t gse_copyfile_p(const char* sPath, const char* sDest, mode_t mode){
	
	int nErrno;
	size_t uSz;
	struct stat sbuf;
	struct utimbuf tbuf;
	char* sOut = NULL;
	char sPathBase[256];
		
	
	uSz = gse_copyfile(sPath, sDest, mode);
	
	if(gse_isdir(sDest)){
		if( ! gse_basename(sPath, sPathBase, 255) )
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Can't determine the basename "
			             "of '%s'.", sPath);
		
		sOut = NS("%s/%s", sDest, sPathBase);
	}
	else{
		sOut = (char*)sDest;
	}
	
	if( stat(sPath, &sbuf) != 0){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Can't stat %s", sPath);
	}
	
	tbuf.actime = sbuf.st_atime;
	tbuf.modtime = sbuf.st_mtime;
	
	if( utime(sOut, &tbuf) != 0){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Can't set mod time on %s", sOut);
	}
	
	if(sOut != sDest) free(sOut);
	return uSz;
}



