/** @file GseFile.h Utilities for file handling */

#ifndef _GseFile_H_
#define _GseFile_H_

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>

#include <rpwgse/GseTypes.h>
#include <rpwgse/GseString.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_file Files 
 * Utilities to assist with file and directory handling.  These module
 * extends some of the simple wrapper funtion defined in @see mod_wrap.
 */
/*@{*/

#ifndef _WIN32
/** A mode_t flag set for Read/Write user and group, and no permissions for 
 * other. 
 */
#define GSE_MODE_660 S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP
#else
#define GSE_MODE_660 0
#endif

#ifndef _WIN32
/** A mode_t flag set for Read/Write user and group, and read-only for other */ 
#define GSE_MODE_664 S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH
#else
#define GSE_MODE_664 0
#endif

#ifndef _WIN32
/**  A mode_t flag set for Read/Write/Execute user and group, and no
 * permissions for other. 
 */
#define GSE_MODE_770 S_IRWXU|S_IRWXG
#else
#define GSE_MODE_770 0
#endif

#ifndef _WIN32
/**  A mode_t flag set for Read/Write/Execute user and group, and Read/Exec
 * permissions for other. 
 */
#define GSE_MODE_775 S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH
#else
#define GSE_MODE_775 0
#endif

/** A basename fuction that is thread safe and doesn't alter the input path. 
 * @param sPath the path from which to get the basename.
 * @param sBasename a buffer to hold the result.
 * @param nLen the actual size of the buffer, result will be NULL terminated
 *        or false is returned.
 * @returns true if the basename could be gathered, false if not.  Eitherway
 *          the output buffer will most likely be altered.
 */
bool gse_basename(const char* sPath, char* sBasename, size_t nLen);

/** A simplified interface basename function, just return an new string
 * that is the basename.
 * @param The path from which to get the basename
 * @returns The filename portion of the path.
 * throws (GSE_XCEPT, GSE_XBADARG) if underlying call to gse_basename fails
 */
char* gse_bname(const char* sPath);


/** A simplified interface directory name function, just return a new string
 * that is the dirname
 *
 */
char* gse_dname(const char* sPath);


/** Grind down a generic file name into an absolute path.
 *
 * @todo make a static buffer version of this function (gse_abspath_s)
 *
 * @param sInput the filename to be converted to an abs path.
 * @returns The corresponding absolute pathname allocated on the heap,
 *          which may be NULL if .the input is null.
 */
char* gse_abspath(const char* sInput);


/** Is the internal pointer for a file descriptor at the end of file? 
 *
 * Note this function moves and resets the file pointer, if an exception is
 * throw, the file pointer might be left in at the EOF.
 *
 * @param fd the file descriptor in question.
 * @return true if at EOF, false if not.
 * @throws (CLIB_XCEPT, errno) for all the reasons lseek(2) sets errno
 * @throws (GSE_XCEPT, GSE_XOVERFLOW) if the file is larger than 2 Gigabytes
 *         and rpwgse was built against a 32-bit clib.
 */
bool gse_at_eof(int fd);

/** Set the file pointer, but don't go out of bounds 
 *
 * @throws nothing  This function does not throw do to bad offsets, and thus
 *         may be used in an exception handler.
 */
size_t gse_lseek_bnds(int fd, size_t offset, int whence);


/** Reading from a file descripter with restart control.
 * 
 * This version of read is intended for use when one wishes to interrupt the
 * read via a signal using either pthread_kill(P), or raise(3).  Works as
 * standard read(2), with the following exceptions:
 *
 *  - Cannot be used on a Pipe, FIFO, or socket.
 *  
 *  - There are no negative return values, errors are indicated by throwing
 *    an exception.  Before throwing, the file pointer is reset to the 
 *    position it hand before this call.
 *
 *  - An errno of EINTR will not cause an exception see below.
 *
 *  - If the pointer to the boolean value in not null, then upon interuption
 *    due to a signal the value at @b restart will be evaluated.  If it is
 *    true read will be called again, if false the function will return 0.
 *
 * @b NOTE: that gse_thread_init() installs a do nothing signal handler for
 * SIGUSR2, thus sending SIGUSR2 to a thread is the default way to interrupt a
 * read. 
 *
 * @param fd the file descriptor from which to read.
 * @param buf where to put the data
 * @param count the maximum number of bytes to read into @b buf
 * @param restart a pointer to a bool.  Checked to see if interruption
 *        calls to read(2) should be restarted.  If restart is NULL calls
 *        are not restarted.
 * @return The number of bytes read from the file.
 */
size_t gse_read(int fd, void *buf, size_t count, const bool* restart);

/** Writing to a file descripter with restart control.
 * 
 * This version of write is intended for use when one wishes to interrupt the
 * write via a signal using either pthread_kill(P), or raise(3).  Works as
 * standard write(2), with the following exceptions:
 *
 * - Cannot be used on a Pipe, FIFO, or socket.
 * 
 * - There are no negative return values, errors are indicated by throwing
 *   an exception.  Before throwing, the file pointer is reset to the 
 *   position it hand before this call.
 *
 * - An errno of EINTR will not cause an exception see below.
 *
 * - If the pointer to the boolean value in not null, then upon interuption
 *   due to a signal the value at @b restart will be evaluated.  If it is
 *   true read will be called again, if false the function will return 0.
 *
 * @b NOTE: that gse_thread_init() installs a do nothing signal handler for
 * SIGUSR2, thus sending SIGUSR2 to a thread is the default way to interrupt a
 * read. 
 *
 * @param fd the file descriptor from which to read.
 * @param buf from which data is read.
 * @param count the maximum number of bytes to read into @b buf
 * @param restart a pointer to a bool.  Checked to see if interruption
 *        calls to write(2) should be restarted.  If restart is NULL calls
 *        are not restarted.
 * @return The number of bytes read from the file.
 */
size_t gse_write(int fd, const void *buf, size_t count, const bool* restart);


/** Read the entire contents of a file into a buffer.
 * This version restarts all reads that have been interrupted via a signal.
 * 
 * @param path the file name, passed to clib open.
 *
 * @return a bytebuf pointing to the file bytes that have been allocated on
 * the heap. If this function returns properly, the bytebuf contains all the
 * file's bytes.
 * 
 * @throws (CLIB_XCEPT, errno) may be thrown for all the reasons that the 
 * clib open, malloc, or read functions might fail. 
 */
bytebuf_t gse_readfile(const char* path);

/** Write the contents of a buffer as a file.
 *
 * This version restarts all writes that have been interupted via a signal.
 * Directories are made as required.  
 *
 * @note If directories are to be created, then in each spot that read
 * permissions are set for the file, execute permissions will be set for the
 * new directories. For example if the file mode is: @c 640 then the directory
 * mode will be @c 750.
 *
 * @param path the file name may include directory components.
 * @param data the data buffer to write.
 * @param mode the premission bits for the file as described in open (2).
 * @throws (CLIB_EXCEPT, errno) if the system C lib sets errno.
 */
void gse_writefile(const char* path, cbytebuf_t data, mode_t mode);

/** Read the entire contents of a file into a buffer, interruptable version.
 * 
 * @param path the file name, passed to clib open.
 *
 * @param restart if not null and deferences to true, interrupted reads are
 * restarted.
 *
 * @return a bytebuf pointing to the file bytes that have been allocated on
 * the heap.  If this function is interrupted via a signal and @b pbRestart
 * is NULL or differences to false, the bytebuf may be shorter than the file.
 * 
 * @throws (CLIB_XCEPT, errno) may be thrown for all the reasons that the 
 * clib open, malloc, or read functions might fail. 
 */
bytebuf_t gse_readfile_i(const char* path, const bool* restart);


/** Is the path a directory.
 * @param path The directory in question, passed to stat(2)
 * @return true if @b path can be determined to be a directory, false otherwise 
 * @throws This function does not throw.
 */
bool gse_isdir(const char* path);

/** Is the path a file.
 * @param path The file in question, passed to stat(2)
 * @return true if @b path can be determined to be a file, false otherwise 
 * @throws This function does not throw.
 */
bool gse_isfile(const char* path);

/** Does this path entry exists
 * @param path The directory in question, passed to stat(2)
 * @return true if @b path is a filesystem entry, false otherwise
 * @throws This function does not throw.
 */
bool gse_exists(const char* path);

/** Make a directory and its parents.
 *
 * @param path the directory to create may have intermediate directories tha
 * are created as well.
 *
 * @param mode the permission mode for the new directories.  At a minimum
 * S_IRWXU is required.  Set the definition of the @c stat function in your
 * system C library documentation.
 * 
 * @throws (CLIB_XCEPT, errno) for all the reasons that stat(2) or mkdir(2)
 * may fail.
 *
 * @note Windows users should use the alternate path seperator '/', instead of
 *       the primary seperator '\\'.
 */
void gse_mkdirs(const char* path, mode_t mode);

/** Make the directories leading to a file.
 *
 * Will make any missing directories leading to a path element.  If all 
 * containing directories are present, no action is taken.
 *
 * @param path to a file.  If the path ends in '/', then all elements of the
 *        path will be considered as directories.  If @a path does not end in
 *        '/', the last path element will be considered a file or link, not a
 *        directory.
 *
 * @param mode The file mode, this value is ignored on windows.
 * 
 * @throws (CLIB_XCEPT, errno) for all the reasons that stat(2) or mkdir(2)
 *        may fail.
 *
 * @note Windows users should use the alternate path seperator '/', instead of
 *       the primary seperator '\\'.
 */
void gse_mkdirsto(const char* path, mode_t mode);


/** Get a directory listing 
 *
 * Return a listing of all the items in a directory. May select only files or
 * only directories, or both.  No the '.' and '..' directories are not 
 * returned.
 *
 * @param A directory path.
 * 
 * @param What to return 1 = files, 2 = sub-directories, 3 = both
 *
 * @return A sorted GseStrList containing the directory items allocated on
 *         the heap, or NULL if the directory was empty.
 *
 * @throws (CLIB_XCEPT,errno) if there is an error in an system CLIB function.
 *         (GSE_XCEPT,GSE_XBADARG) if nType & 0x3 is zero.
 */

GseStrList* gse_listdir(const char* path, int nType);


/** Copy a file from one location to another.
 *
 * @warning This function does not check to see if the destination exists and
 *          will overwrite files.
 *
 * @param sPath the file to copy
 *
 * @param sDest the directory or filename to create
 *
 * @param mode the filesystem premissions mode of the destination file.  These
 *        are defined in man 2 open, but the convience macros GSE_MODE_660 and
 *        GSE_MODE_664 may also be used.
 *
 * @return the number of bytes copied to the new location.  This is only zero
 *         if the size of the input file is zero.
 */
size_t gse_copyfile(const char* sPath, const char* sDest, mode_t mode);

/** Copy a file from one location to another, preserving the mod time.
 *
 * @warning This function does not check to see if the destination exists and
 *          will overwrite files.
 *
 * @param sPath the file to copy
 *
 * @param sDest the directory or filename to create with the same mtime as sPath
 *
 * @param mode the filesystem premissions mode of the destination file.  These
 *        are defined in man 2 open, but the convience macros GSE_MODE_660 and
 *        GSE_MODE_664 may also be used.
 *
 * @return the number of bytes copied to the new location.  This is only zero
 *         if the size of the input file is zero.
 */
size_t gse_copyfile_p(const char* sPath, const char* sDest, mode_t mode);


/* Open a file, and make directories if needed
 *
 * @param path the path (optional) and filename of the file to open.
 * @param mode The standard fopen (3) mode string.  If the file must be opened for
 *        writing, and the intermediate directories to the file do not exist
 *        they will be created.
 * @return An FILE pointer ready for use.  Does not return NULL, instead errors
 *         cause an exception to be thrown.
 *
 * @throws (CLIB_XCEPT, errno) for any number of reasons
 */


#define GSE_ENT_FILE             1
#define GSE_ENT_DIR              2
#define GSE_ENT_UNSTATABLE_FILE  3
#define GSE_ENT_UNREADABLE_DIR   4

/** Definition of a callback function for use in gse_walk()
 *
 * @param pUser The user data pointer provided to gse_walk()
 * @param sPath The directory or file in question
 * @param pStat Information from stat(2) for filesystem entry @a sPath.
 *        note that pStat may be @b NULL if stat(2) returns an error. 
 *
 * @return 0 if walking should continue, a non-zero value otherwise.
 * @throws It is permissible to throw exceptions from a walk action function.
 */
typedef int (*gse_walk_action)(void* pUser, const char* sPath, 
                               const struct stat* pStat);

/** Walk a directory tree and invoke a user defined action for each item.
 * This function follows symlinks, if needed a "non symlink following"
 * version can be created. Note that @c gse_walk does not change the 
 * current working directory of calling program.  
 *
 * @note Guidance for gse_walk was taken from "Advanced Programming in the
 *       UNIX environment", section 4.21 by W. Richard Stevens. 
 *
 * @param path The path to walk, may be a file, link or directory.
 * @param action The callback to invoke for each filesystem item.  Not that
 *        action is called even if the filesystem item is not stat'able.  This
 *        includes the top level @a path. 
 * @param user_data An optional pointer to be passed to the walk action 
 *        callback.
 * @return 0 unless the walk action returned non-zero
 * @throws (GSE_XCEPT, GSE_XBADARG) if path or action is NULL.
 * @throws walk_action any exceptions thrown by the @a action callback.
 */
int gse_walk(const char* path, gse_walk_action action, void* user_data); 
		       
                


/*@}*/
/*@}*/	
#ifdef __cplusplus
}
#endif

#endif /* _GseFile_H_ */
