#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <libgen.h>

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseFile.h"
#include "GseString.h"
#include "GseFilePktSink.h"
#include "GseLog.h"

/*****************************************************************************/
void GseFilePktSink_open(GsePktSink* pBase)
{
	GseFilePktSink* pThis = (GseFilePktSink*)pBase;
	int openflags = 0;
	
	openflags = O_RDWR|O_CREAT;
	if(pThis->_uFlags&GSE_OVERWRITE) openflags |= O_TRUNC;
	pThis->_fd = gse_open(pThis->_sPath, openflags, pThis->_mode);
	
	if(pThis->_uFlags&GSE_APPEND)
		gse_lseek(pThis->_fd, 0, SEEK_END);
}


int4 GseFilePktSink_received(const GsePktSink* pBase)
{
	const GseFilePktSink* pThis = (GseFilePktSink*)pBase;
	return pThis->_recieved;
}
	
int4 GseFilePktSink_flushed(const GsePktSink* pBase)
{
	const GseFilePktSink* pThis = (GseFilePktSink*)pBase;
	return pThis->_flushed;
}
	
void GseFilePktSink_writeNext(GsePktSink* pBase, GsePkt* pPkt)
{
	bytebuf_t buf;
	bool bRestart = true;
	GseFilePktSink* pThis = (GseFilePktSink*)pBase;
/*	static const except_id_t catch_spec[] = {{CLIB_XCEPT, */
	
	buf = pPkt->encode(pPkt);
	except_cleanup_push(free, buf.ptr);
	
	if(pThis->_fd == -1) GseFilePktSink_open(pBase);
	
	/* This function backs up the file pointer in the case of a problem */
	gse_write(pThis->_fd, buf.ptr, buf.len, &bRestart);
	
	/* Flush to disk after each packet */
#ifndef __MINGW32__
	fdatasync(pThis->_fd);
#endif
	pThis->_flushed += 1;
	gse_debug_v("Wrote packet %d to %s", pThis->_flushed, pThis->_sPath);
	
	except_cleanup_pop(except_call);       /* Removes the local buf data */
}

void GseFilePktSink_consumeNext(GsePktSink* pBase, GsePkt* pPkt)
{
	GseFilePktSink_writeNext(pBase, pPkt);
	pPkt->del(pPkt);
}

/*****************************************************************************/
void delete_GseFilePktSink(void* vpThis)
{
	GseFilePktSink* pThis = (GseFilePktSink*)vpThis;
	
	if(pThis->_fd != -1) close(pThis->_fd);
	if(pThis->_sPath != NULL) free(pThis->_sPath);
	free(pThis);
}

GseFilePktSink* new_GseFilePktSink(const char* sPath, uint4 uFlags, mode_t mode,
                                   size_t nOver)
{
	GsePktSink*     pBase = NULL;
	GseFilePktSink* pThis = NULL;
	size_t nAlloc = 0;
	char* sPathCpy = NULL;
	char* sDir = NULL;
	mode_t dirmode = 0;
	
	if((sPath == NULL)||(sPath[0] == '\0'))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	if((sPath[strlen(sPath) - 1] == '/')||   /* Can't end in "/" */
	   (strcmp(".", sPath) == 0)        ||   /* Can't be '.'     */
	   (strcmp("..", sPath) == 0))           /* Can't be '..'    */
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "'%s' is not a file path", sPath);
	
   if(nOver > sizeof(GseFilePktSink))
		nAlloc = nOver;
	else
		nAlloc = sizeof(GseFilePktSink);
	
	pThis = (GseFilePktSink*)gse_calloc(1, nAlloc);
	pBase = (GsePktSink*)pThis;
	
	/* Interface implimentiation */
	pBase->_pIfaceOwner = (void*)pThis;
	pBase->received     = GseFilePktSink_received;
	pBase->flushed      = GseFilePktSink_flushed;
	pBase->consumeNext  = GseFilePktSink_consumeNext;
	pBase->writeNext    = GseFilePktSink_writeNext;
	pBase->del          = delete_GseFilePktSink;
	
	pThis->del = delete_GseFilePktSink;
	
	except_cleanup_push(delete_GseFilePktSink, pThis);
	
	/* Make the directory tree if need be */
	sPathCpy = CS(sPath);                /* arrg, why did POSIX allow dirname */
	except_cleanup_push(free, sPathCpy); /* and basename to modify thier      */
	                                     /* arguments?                        */
	sDir = dirname(sPathCpy);               
	
	if(! gse_isdir(sDir)){ 
		
		if( !(uFlags & GSE_MAKEDIRS))
			gse_throw_fm(CLIB_XCEPT, ENOTDIR,"Directory '%s' does not exist",sDir);
	
		dirmode = mode;
#ifndef __MINGW32__
		if(mode &  S_IRUSR) dirmode |= S_IXUSR;
		if(mode &  S_IRGRP) dirmode |= S_IXGRP;
		if(mode &  S_IROTH) dirmode |= S_IXOTH;
#endif
		gse_mkdirs(sDir, dirmode);
	}
	
	/* Okay, see if the file exists */
	if(gse_isfile(sPath) && !(uFlags&GSE_OVERWRITE) && !(uFlags&GSE_APPEND))
		gse_throw_fm(GSE_XCEPT, GSE_XOVERWRITE, 
		             "Cowardly refusing to overwrite '%s', use GSE_OVERWRITE or "
		             "GSE_APPEND flags to overwrite or append to an existing "
		             "file.", sPath);
		
	except_cleanup_pop(except_call);   /* free sPathCpy no matter what */
	
	/* Save the settings, but don't open the file until it's needed */
	pThis->_fd = -1;
	pThis->_sPath = CS(sPath);
	pThis->_mode = mode;
	pThis->_uFlags = uFlags;
		
	except_cleanup_pop(except_no_call); /* pThis */
	return pThis;
}
