/** @file GseFilePktSink.h Write packets to a file. */
#ifndef _GseFilePktSink_H_
#define _GseFilePktSink_H_

#include <rpwgse/GsePktSink.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_processing Processing */
/*@{*/

#define GSE_APPEND    0x1
#define GSE_OVERWRITE 0x2
#define GSE_MAKEDIRS  0x4

/** Send all packets recieved to disk. 
 *
 * Implements the GsePktSink interface.  Packets are encoded via GsePkt::encode
 * and the resulting byte stream is writen to disk.
 * @n
 * @n
 * @class GseFilePktSink
 * @implements GsePktSink
 */
typedef struct gse_file_ptk_sink{
	
	/** Pointers to this struct can be cast to a generic GsePktSink struct */
	GsePktSink base;
	
	char* _sPath;
	
	int _fd;
	
	uint4 _recieved;
	
	uint4 _flushed;
	
	mode_t _mode;
	
	uint4 _uFlags;
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} GseFilePktSink;

/** Make a new GseFilePktSink.  
 * 
 * @note This implementation of the GsePktSink interface @b does define the
 *       @a GsePktSink::del member.  So it is safe to store a pointer to an
 *       instance of this structure as a GsePktSink pointer. 
 *
 * @param sPath the output file name
 *
 * @param uFlags 0 or more of the following may be specified:
 *    - @b GSE_APPEND:    Append packets if the file exists
 *    - @b GSE_OVERWRITE: Overwrite existing files
 *    - @b GSE_MAKEDIRS:  Make directories as needed before opening sPath for
 *         writing.  @b Warning, this may have the side effect of making
 *         directories but not creating a file. 
 * 
 * @param mode The file creation mode, see man stat for modes.  Note that
 *        if the flag GSE_MAKEDIRS is specified, directory modes will add
 *        the execute bit in each place that the read bit is seen in the mode
 *
 * @param nOver Used by derived structures, set this to 0.
 *
 * @return The new packet sink object.  Never returns NULL.
 * 
 * @throws (CLIB_XCEPT, errno) if a clib call fails.
 * 
 * @throws (GSE_XCEPT, GSE_XOVERWRITE) if creating the source would overwrite
 * an existing file and the GSE_OVERWRITE flag was not specified.
 *
 * @throws (GSE_XCEPT, GSE_XBADARG) if there is a problem with the function args
 *
 * @public
 * @memberof GseFilePktSink
 */
 
GseFilePktSink* new_GseFilePktSink(const char* sPath, uint4 uFlags, mode_t mode,
                                   size_t nOver);
 
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseFilePktSink_H_ */

