/** @file GseFilePktSrc.h Generate packets by reading a file. */

#ifndef _GseFilePktSrc_H_
#define _GseFilePktSrc_H_

#include <rpwgse/GsePktSrc.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_processing Processing */
/*@{*/

/** Constants for the choice of how to handle EOF on input.
 * @see new_GseFilePktSrc()
 */
enum gse_eof_disp {at_eof_stop, at_eof_retry, at_eof_block};
typedef enum gse_eof_disp gse_eof_disp;

/** Generate packets from a disk file.  
 * 
 * @b Note: <b>Behavior</b>. Reading starts from the current file pointer.
 * If an exception is thrown the file pointer is reversed back to <b> 
 * immediatly after the initial sync bytes</b> and a search for the
 * sync bytes begins anew.
 *
 * The file contents or lack there of shall not cause an exception. @n
 *
 * @todo Make the linux specific version that uses the inotify(7) 
 * functianality to avoid polling.
 *
 * @todo Make the solaris version that uses the File Events Notification
 * system to avoid polling.  See 
 *    http://blogs.sun.com/praks/entry/file_events_notification
 * for more details
 *
 * @class GseFilePktSrc
 * @extends GsePktSrc
 */
typedef struct gse_file_pkt_src{
	
	/** interface type is a GsePktSrc */
	GsePktSrc base;
	
	/* The block factory function used to generate the
	 * proper header block decoders */
	gse_blk_factory _factory;
	
	/* EOF disposition */
	gse_eof_disp _eof_disp; 
	
	int _fd;
	
	char* _sName;
	
	int4 _pktIdx;
	
	int4 _nPkts; /* Not set until hit EOF */
	
	bool _bRestartReads;
	
	/** Destructor, deallocates memory and closes the underlying file. */
	void (*del)(void* vpThis);
	
} GseFilePktSrc;


/** Create packet source that reads from a file. 
 * 
 * @param sFilename The file to read
 *
 * @param factory The GseBlock factory function to use to interperate 
 * header bytes
 *
 * @param disp The End of file disposition.  Can have one of the following
 * enum'erated values:
 *
 *   - \b at_eof_stop  if specified, once the EOF is found, no attempt will
 *     be made to read past it in GsePktSrc::nextPkt.  GsePktSrc::numPkts()
 *     will return the number of packets encountered once EOF is reached.
 *
 *   - \b at_eof_retry if specificed, each call to GsePktSrc::nextPkt will 
 *     attempt to read the next packet's data, and return immediatly if there
 *     are insufficient bytes to form the next packet.  GsePktSrc::numPkts()
 *     will always return -1.   This can be used to read a file that may be
 *     growing.
 *
 *   - \b at_eof_block if specificed, each call to GsePktSrc::nextPkt will 
 *     attempt to read the next packet's data, and will \b block if there
 *     are insufficient bytes to form the next packet.  GsePktSrc::numPkts()
 *     will always return -1.   This can be used to read a file that may be
 *     growing.  No that in the multithreaded version of the library, 
 *     GsePktSrc::interrupt may be called from a different thread to stop 
 *     a blocked read.
 *
 * @memberof GseFilePktSrc
 */
 
GseFilePktSrc* new_GseFilePktSrc(const char* sFilename, gse_blk_factory factory,
		                           gse_eof_disp disp);
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseFilePktSrc_H_ */
