/** @file GseFileStr.h Using a character buffer as a file */
#ifndef _GseFileStr_H_
#define _GseFileStr_H_

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_string */
/*@{*/

/** A file-like string buffer.
 * If you need to write to an internal buffer as if it was a file, this is
 * the class for you... @n
 * @class GseFileStr
 */
typedef struct gse_file_str{
	/** Get a constant pointer to the underlying null terminated string array */
	const char* (*getstr)(struct gse_file_str* pThis);
	
	/** Set the internal buffer to have all 0 bytes, but don't reallocate it */
	void (*clear)(struct gse_file_str* pThis);
	
	/** Take the data string away from the object.
	 *
	 * The destructor will no longer delete the buffer, and any future calls
	 * to add data will allocate a new internal buffur.
	 * This resets all internal buffers, indicies etc to defaults.
	 */
	char* (*takestr)(struct gse_file_str* pThis);
	
	/** Take the data string away, but start a new one.
	 *
	 * Caller is responsible for freeing the returned memory, as it's no
	 * longer owned by this class.
	 */
	char* (*reuse)(struct gse_file_str* pThis, size_t nInitialBuf);
	
	/** write a single character at the current index */
	void (*putchar)(struct gse_file_str* pThis, char c);
	
	/** write a string at the current index */
	void (*putstr)(struct gse_file_str* pThis, const char* s);
	
	/** write a string at the current index with printf fanciness */
	int  (*printf)(struct gse_file_str* pThis, const char* format, ...);
	
	/** write raw bytes to a string.
	 * Use buflen instead of stringlen when retrieving raw bytes from the
	 * string, to avoid the problem of embedded nulls.
	 */
	void (*write)(struct gse_file_str* pThis, const byte* ptr, size_t len);
	
	/** set the write index */
	void (*seek)(struct gse_file_str* pThis, size_t offset, int whence);
	
	/** read the write index */
	size_t (*tell)(struct gse_file_str* pThis);
	
	/** get the current string len (does not include terminating NULL) */ 
	size_t (*strlen)(struct gse_file_str* pThis);
	
	/** get the current number of bytes written to the string.
	 * @note This returns the same value as strlen @b only @b if there are
	 *       no embedded NULL bytes.
	 * @return Get the number of bytes written to the string (does not include
	 *         the terminating NULL)
	 */
	size_t (*buflen)(struct gse_file_str* pThis);
	
	
	/** destructor */
	void (*del)(void* vpThis);
	
	char* _buf;
	size_t _idx;
	size_t _len;
	
}GseFileStr;


/** Construct a new file string.
 *
 * @param nInitialBuf Set the initial size of the internal buffer.  This buffer
 *        will be reallocated to larger sizes as needed.  The default is 10
 *        bytes if nInitialBuf < 10.
 *
 * @returns A new GseFileStr structure allocated on the heap.
 * @public
 * @memberof GseFileStr */
GseFileStr* new_GseFileStr(size_t nInitialBuf);

/** Stack variable initializer
 * @public
 * @memberof GseFileStr */
void init_GseFileStr(GseFileStr* pThis, size_t nInitialBuf);

/** Stack variable finalizer
 * @public
 * @memberof GseFileStr 
 */
void fin_GseFileStr(void* vpThis);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseFileStr_H_ */
