/** @file GseHdr.h Handle GSE packet headers */
#ifndef _GseHdr_H_
#define _GseHdr_H_

#include <rpwgse/GseProcBlk.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_packet Packets */
/*@{*/

/** Contains a complete GSE Header, including all header blocks.
 * The @ref page_spec_4_2_1 "16 GSE Header Prefix" fields are owned by
 * this class.
 * @n
 * @class GseHdr 
 */
typedef struct gse_hdr{

	/** @b FIELD: The CRC-CCITT of the entire packet's data except for the
	 *  the sync bytes and the CRC inself.
	 * 
	 * In order to avoid extra hidden dependencies, this header structure does
	 * not have direct access to the packet data, and thus CRCs can not be 
	 * auto calculated.  So for packets that are created instead of decoded,
	 * this field is not set.
	 */
   uint2 uCRC;
	
	/** @b FIELD: The total number of non-data bytes in the packet.
	 * 
	 * For created packets this field is auto-calculated on #encode.
	 */
	uint2 uHeadTailLen;
	
	/** @b FIELD: The total number of bytes in the entire packet
	 *
	 * @warning This field is not auto calculated on #encode. 
	 * @see #encode for more details.
	 */
	uint4 uTotalLen; 
	
	/** @b FIELD: If available this is the NAIF spacecraft code.
	 * NAIF spacecraft IDs are negative numbers.  For non-spacecraft
	 * applications, use positive IDs.
	 */
    int2 nSpaceCraft;
	 
	/** @b FIELD: This understands how to read version 4 of the GSE
	 * format specification.  This value should always be 0x04.
	 */
   byte uHdrVer;
	
	/** @b FIELD: An identifier for the contents of the packet.
	 * The meaning of this field application specific.  This for each
	 * value of GseHdr::SpaceCraft a the meaning of this field requires
	 * external documentation. 
	 */
	byte uContent;
	
	/** Copy constructor 
	 *
	 * Generates a new GseHdr on the heap as a deep copy of this one. 
	 * @param pThis The GseHdr to copy
	 * @param uOver Usually set this to 0, it's for derived classes
	 * @return A pointer to a GsePkt structure.
	 */
	struct gse_hdr* (*copy)(const struct gse_hdr* pThis, size_t uOver);
	
	/** Get the original byte string use during decoding. 
	 *
	 * @return a constant byte buffer (not a new heap allocation) to the
	 *         original decoded bytes, or an empty buffer if this block
	 *         was not generated via decoding.
	 */
	cbytebuf_t (*original)(struct gse_hdr* pThis);
	
	/** Adding a block to an existing header.  
	 *
	 * This is used when creating a new header instead of decoding one.
	 * 
	 * @param pBlock The block to add.  Note: NULL pointers will silently 
	 *        be ignored.
	 *
	 * @return Returns the new number of blocks in the header.
	 * @throws (CLIB_XCEPT, ENOMEM)
	 */
	size_t (*addBlock)(struct gse_hdr* pThis, GseBlock* pBlock);
	
	/** Repacing a block at a given index.
	 *
	 * @attention The caller becomes responsible for the memory of the returned
	 *            block.  Likewise, the GseHdr assumes responsibility for
	 *            the memory of the passed in block.
	 * @param pNewBlk The new block to place at the given index, may be NULL.
	 * @return The previous block at this index.  
	 */
	GseBlock* (*replaceBlockAt)(struct gse_hdr* pThis, size_t idx,
	                            GseBlock* pNewBlk);
	
	/** Get a block at an index
	 *
	 * @param idx the 0 based index of the block to retrieve.
	 * @return The Block pointer at the specified index, may be null.
	 * @throws (GSE_XCEPT, GSE_XBADARG) if idx is larger than the number of
	 *         blocks in this header minus one.
	 */
	GseBlock* (*getBlockAt)(struct gse_hdr* pThis, size_t idx);
	
	/** Get the number of blocks in the header.
	 * 
	 * @return The number of GseBlock pointers held by this header, suitable
	 *         for use in iterating over all blocks.  Note that this value
	 *         @b includes any NULL pointers.
	 */
	size_t (*numBlocks)(struct gse_hdr* pThis);
	
	/** Retrieve a header block based on it's type and process IDs.
	 * 
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block requested.
	 * @param uStep The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        proceessing step ID.  For Non-process blocks use 0.
	 * @return A pointer to GseBlock derived structure corresponding to this
	 *         block type.
	 * @throws (GSE_XCEPT, GSE_XNOTUNIQUE) if there is not one and only one
	 *         block of this type in the header
	 * @ref Overview for a description of different block structures are
	 *      generated based on type and process IDs.  
	 */
	GseBlock* (*getBlkByType)(struct gse_hdr* pThis, byte uType, byte uStep);
	
	/** Retrieve a header block based on it's type and process IDs.
	 * 
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block requested.
	 * @param uStep The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        proceessing step ID.  For Non-process blocks use 0.
	 * @param idx The index of the block to return.  In the case were there
	 *        is only one block of the given type the only valid value of
	 *        this parameter is 0.  Blocks are not sorted, until GsePkt::encode
	 *        is called, so there is no-garruntee on the block ordering.
	 * @return A pointer to GseBlock derived structure corresponding to this
	 *         block type.
	 * @throws (GSE_XCEPT, GSE_XNOTUNIQUE) if there is not one and only one
	 *         block of this type in the header
	 * @ref Overview for a description of different block structures are
	 *      generated based on type and process IDs.  
	 */
	GseBlock* (*getBlkByTypeIdx)(struct gse_hdr* pThis, byte uType, byte uStep,
	                             int4 idx);
	
	/** Get the number of header blocks of a particular type and process ID.
	 *
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block requested.
	 * @param uStep The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        proceessing step ID.  For Non-process blocks use 0.
	 * @return The number of header blocks matching @p uType and optionally 
	 *         @p uStep.
	 * @throws None, This function does not throw exceptions.
	 * @ref Overview for a description of different block structures are
	 *      generated based on type and process IDs.  
	 */
	size_t (*numBlksByType)(struct gse_hdr* pThis, byte uType, byte uStep);
	
	/** Attempt to merge similar blocks.
	 *
	 * Cycle over blocks in the header.  Whenever two blocks with the
	 * same type and step id's are encontered, call the block merge function.
	 * If the merge succeeds, only the merged block is retained.  If more
	 * than two blocks of a particular type remain, merging is repeated until
	 * it fails or only one remains.
	 *
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType the type of blocks to merge, if @b 0 merge all blocks
	 *        of the same type.
	 * @param uStep for processing blocks, this is the step ID.  If @b 0 merge
	 *        all blocks with the same type and step.  Note, this parameter
	 *        is not used unless uType is @b 0x70.
	 * @return The number of header blocks that were sucessfully merged, i.e.
	 *         the number of blocks removed from the header.
	 *
	 * @throws None.
	 */
	size_t (*mergeBlocks)(struct gse_hdr* pThis, byte uType, byte uStep);
				 

	/** Encode the header to a byte stream.
	 * 
	 * This version can be used to specify exactly how long the returned
	 * byte stream should be.  This function updates the uHeadTailLen field.
	 *
	 * @note: The returned byte string is @b not suitable for output to disk as
	 * the uCRC and uTotalLen fields are @b never encoded.  It is up to the
	 * caller to fill in these fields which are at offset 4 (CRC) and offset
	 * 8 (TotalLen) in the output string.  GsePkt::encode handles this
	 * automatically.
	 * 
	 * @param pThis poniter to the relavent GseHdr structure.
	 * 
	 * @param uMinLen The minimum length of the encoded byte stream.  If less
	 * than uMinLen bytes are required the remainder will be zero padded.
	 * 
	 * @param uMaxLen the maximum length -1 allowed for the encoded byte stream.
	 * If more than uMaxLen bytes are needed to encode the header, a 
	 * (GSE_XCEPT, GSE_XENCODE) exception is thrown.  To indicate the maximum
	 * possible length use the constants 0xFFFF or 65535.  
	 * 
	 * @return a pointer to the byte string suitable for use in free.  The
	 * header does not own the pointer after the encoding process and the
	 * the caller is responsible for freeing the memory.
	 */
	bytebuf_t (*encode)(struct gse_hdr* pThis, uint2 uMinLen, uint2 uMaxLen);
	
	/** @public Destructor */
	void (*del)(void* vpThis);
	
	/* Encode the header to a byte stream.
	 * 
	 * This version can be used to a chunk size instead of a maximum len.
	 * 
	 * @param pThis poniter to the relavent GseHdr structure.
	 * 
	 * @param uMinLen The minimum length of the encoded byte stream.  If less
	 * than uMinLen bytes are required the remainder will be zero padded.
	 * 
	 * @param uMaxLen the maximum length allowed for the encoded byte stream.
	 * If more than uMaxLen bytes are needed to encode the header, a 
	 * (GSE_XCEPT, GSE_XENCODE) exception is thrown.
	 * 
	 * @return a pointer to the byte string suitable for use in free.  The
	 * header does not own the pointer after the ecoding process and the
	 * the caller is responsible for freeing the memory.
	 */
	/* byte* (*getBytes2)(struct gse_hdr* pThis, uint2 uChunkSz); */
	
	/** @protected an array of pointers to GseBlocks */
	GseBlkAry* _pBlocks;
	
} GseHdr;


/** Generate a new GseHdr from a string of bytes.
 * 
 * @param buf a buffer conting bytes to decode into header fields and header
 *   blocks.  Only header bytes are read.  The actual number of bytes read is
 *   determined by the @c HEAD_TAIL_LEN encoded in the byte string.
 *
 * @param factory The block factory function used to determine which types
 *   of blocks to generate based on block IDs.
 *
 * @param uOverMalloc used by derived classes to allocate more space than is
 *  needed for just a GseHdr structure.
 *
 * @return a GseHdr allocated on the heap.  If buf is not empty all fields
 * and blocks are decoded.
 *
 * @throws (GSE_XCEPT, GSE_XDECODE) if buf.len is shorter than the 
 *    @c HEAD_TAIL_LEN from the byte stream.
 *
 * @memberof GseHdr 
 */
GseHdr* new_GseHdr(cbytebuf_t buf, gse_blk_factory factory, size_t uOverMalloc);

/** Generate a new empty GseHdr suitable for use in generating data.
 * 
 * @param uOverMalloc used by derived classes to allocate more space than is
 *  needed for just a GseHdr structure.
 *
 * @return a GseHdr allocated on the heap with all fields except uHdrVer
 *  set to zero and without any GseBlocks.
 *
 * @memberof GseHdr
 */
GseHdr* new_empty_GseHdr(size_t uOverMalloc);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseHdr_H_ */
  
