#define _POSIX_C_SOURCE 200112

#if defined(__linux)
#define _GNU_SOURCE
#include <execinfo.h>
#include <dlfcn.h>

#elif defined(__sun)
#define __EXTENSIONS__
#include <dlfcn.h>
#include "GseSunSvc.h"

#elif defined(WIN32)
#include "GseWinSvc.h"

#endif

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include "GseExcept.h"
#include "GseConv.h"
#include "GseLog.h"
#include "GseInit.h"


bool gse_except_init();
void gse_except_deinit();

bool gse_log_init(int nLevel, gse_log_handler_t func);
void gse_log_deinit();

void gse_time_init();
void gse_time_deinit();


#define MAX_STACK_DEPTH 20
#define PRETTY_PRN_BUF_SZ   1024

/*****************************************************************************/
static void gse_sigusr2_handler(int nSignal){
	return;  /* May interrupt a system call */
}

static void gse_sigsegv_handler(int nSignal){
	int i, nDepth;
	size_t nAdd, nStrLen;
	Dl_info info;
	void* nStackTraceBuf[MAX_STACK_DEPTH];
	const char* sFileNames[MAX_STACK_DEPTH];
	const char* sFunctions[MAX_STACK_DEPTH];
	char* sBuf[PRETTY_PRN_BUF_SZ];
	sBuf[0] = '\0';
		
	nDepth = backtrace(nStackTraceBuf, MAX_STACK_DEPTH);
	
	/* Getting the function names */
	nStrLen = 0;
	for(i = 2; i < nDepth ; i++){
		sFunctions[i] = "unknown function";
		sFileNames[i] = "unknown file";
				
		if(dladdr(nStackTraceBuf[i], &info)){
			if(info.dli_sname) sFunctions[i] = info.dli_sname;
			if(info.dli_fname) sFileNames[i] = info.dli_fname;
		}
		
		nStrLen += strlen(sFunctions[i]) + strlen(sFileNames[i]) + 15;
	}
	
	/* Printing them */	
	nStrLen = 0;
	for(i = nDepth - 2; i >= 2; i--){
		nAdd = strlen(sFunctions[i]) + strlen(sFileNames[i]) + 15;
		
		if(nStrLen + nAdd >  PRETTY_PRN_BUF_SZ) 
			break;
		
		sprintf( ((char*)sBuf)+nStrLen, "   in %s (from %s)\n",
				  sFunctions[i], sFileNames[i]);
		nStrLen += nAdd;
	}
	
	write(STDERR_FILENO, "Segfault in:\n", 14);
	if(sBuf[0] == '\0')
		write(STDERR_FILENO, "   Stack lost!!!  This is usually caused by a bad"
				" function pointer.\n   Check for unassigned function pointers.\n",
				112);
	else
		write(STDERR_FILENO, sBuf, nStrLen);
	exit(EXIT_FAILURE);
}


/****************************************************************************/
void gse_thread_init(int level, gse_log_handler_t handler){
	const char* sCheckSizes = NULL;

#ifdef __MINGW32__	
	gse_winsvc_init();
#endif
	
	if(! gse_except_init() ){
		fprintf(stderr, "Failure to initialize the gse except component\n");
		exit(134);
	}
	
	/* Set a handler for segfaults */
	if( signal(SIGSEGV, gse_sigsegv_handler) == SIG_ERR)
		gse_throw_m(GSE_XCEPT, GSE_XINITFAIL, "Failure to init SIGSEGV handler");
	
	/* Set a handler for SIGUSR2 (SIGBREAK on windows) that does nothing, it is
	 * just available to interrupt things */
#ifdef __MINGW32__
	if( signal(SIGBREAK, gse_sigusr2_handler) == SIG_ERR)
		gse_throw_m(GSE_XCEPT, GSE_XINITFAIL, "Failure to init SIGBREAK handler");
#else
	if( signal(SIGUSR2, gse_sigusr2_handler) == SIG_ERR)
		gse_throw_m(GSE_XCEPT, GSE_XINITFAIL, "Failure to init SIGUSR2 handler");
#endif
		


	/* Set our timezone offset */
	tzset();
	
	gse_log_init(level, handler);
			
	if((sCheckSizes = gse_checkSizes()) != NULL)
		gse_throw_m(GSE_XCEPT, GSE_XINITFAIL, sCheckSizes);
	
	gse_time_init();
}

void gse_thread_xcept_def(int group, const struct gse_except_def* newdefs)
{
	gse_warning("gse_thread_xcept_def is currently a do nothing function");
	/* Dose nothing for now */
}

void gse_thread_deinit(){
	/* Free pthread stuff here */
	
	gse_except_deinit();
	gse_log_deinit();
#ifdef __MINGW32__
	gse_winsvc_deinit();
#endif
	gse_time_deinit();
}
