#define _POSIX_C_SOURCE 200112
#define _XOPEN_SOURCE 500

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include "GseExcept.h"
#include "GseLog.h"
#include "GseString.h"

#include <time.h>

void gse_def_log_handler(int nLevel, const char* sMsg, const char* sDataStatus, 
                         const char* sStackTrace, bool bPrnTime);

static int gse_nMinLevel = GSE_WARNING;
static int gse_nMinLineLvl = GSE_ERROR;
static bool gse_bLogWithTimes   = false;
gse_log_handler_t gse_curMsgHandler = gse_def_log_handler;


#ifdef GSE_POSIX_THREADS /* IS THREADED */

static pthread_mutex_t s_gse_log_mtx = PTHREAD_MUTEX_INITIALIZER;
#define LOCK()    pthread_mutex_lock(&s_gse_log_mtx)
#define UNLOCK()  pthread_mutex_unlock(&s_gse_log_mtx)

#else /* NOT THREADED */

#define LOCK() 
#define UNLOCK()

#endif /* THREADED */


/* No throwing during init! */
void gse_log_init(int nLevel, gse_log_handler_t func)
{
	LOCK();
	if((nLevel > GSE_EVERYTHING) || (nLevel < GSE_NOTHING))
		gse_nMinLevel = nLevel;
	if(func != NULL)
		gse_curMsgHandler = func;
	UNLOCK();
}

void gse_log_deinit()
{
	/* This function does nothing right now, but is a place holder
	   for future implementation of per thread logging */
	return;
};


bool gse_log_set_showline(int nLevel){ 
	int old;
	LOCK();
	old = gse_nMinLineLvl;
	
	if((nLevel < GSE_EVERYTHING) || (nLevel > GSE_NOTHING))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Source line display threshold "
		             "level %d is not in the range %d to %d.", nLevel,
		             GSE_EVERYTHING, GSE_NOTHING);
	
	gse_nMinLineLvl = nLevel;
	UNLOCK();
	return old;	
}

int gse_log_setlevel(int nLevel){ 
	int old;
	LOCK();
	old = gse_nMinLevel;
	
	if((nLevel < GSE_EVERYTHING) || (nLevel > GSE_NOTHING))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Message level %d is not in the "
				       "range %d to %d.", nLevel, GSE_EVERYTHING, GSE_NOTHING);
	
	gse_nMinLevel = nLevel;
	UNLOCK();
	return old;	
}

int gse_log_getlevel(){ return gse_nMinLevel; }


void gse_log_include_time(bool bPrnTime)
{
	LOCK();
	gse_bLogWithTimes = bPrnTime;
	UNLOCK();
}


void gse_log(int nLevel, const char* sSrcFile, int nLine, const char* sFmt, ...){
	char* sMsg, * sTmp;
	va_list ap;
	
	if(nLevel < gse_nMinLevel) return;
	
	va_start(ap, sFmt);
	sMsg = gse_vstring(sFmt, ap);
	va_end(ap);
	
	if((nLevel >= gse_nMinLineLvl)&&(sSrcFile != NULL)&&(nLine > 0)){
		sTmp = sMsg;
		sMsg = gse_string("%s\n\t(Reported from %s, line %d)", sMsg, sSrcFile, nLine);
		free(sTmp);
	}
	
	gse_curMsgHandler(nLevel, sMsg, NULL, NULL, gse_bLogWithTimes);
	free(sMsg);
}

void gse_log_msg(int nLevel, const char* sMsg){
	if(nLevel < gse_nMinLevel) return;
		
	gse_curMsgHandler(nLevel, sMsg, NULL, NULL, gse_bLogWithTimes);
}


void gse_log_except(int nLevel, except_t* pExcept)
{
	char code_buf[128];
	char group_buf[64];
	char xcept_buf[1024];
	const char* pGroup = NULL;
	const char* pMsg = NULL;
	size_t nWrote = 0;
	xcept_buf[1023] = '\0';
	
	if(nLevel < gse_log_getlevel())
		return;

	/* Make message string */
	switch(except_group(pExcept)){
		case CLIB_XCEPT: 
			pGroup = "CLIB";
			pMsg = clib_code_str(except_code(pExcept), code_buf, 127);
			break;
		case GSE_XCEPT:
			pGroup = "GSE";
			pMsg = gse_code_str(except_code(pExcept), code_buf, 127);
			break;
		default:
			snprintf(group_buf, 63, "Group %lu", except_group(pExcept));
			pGroup = group_buf;
			snprintf(code_buf, 128, "Code %lu", except_code(pExcept));
			pMsg = code_buf;
	}
	nWrote = snprintf(xcept_buf, 1023, "%s exception, %s.\n", pGroup, pMsg);

	
	if(gse_except_srcfile(pExcept) != NULL)
		nWrote += snprintf(xcept_buf + nWrote, 1023 - nWrote,
		                   "\tFrom %s, line %d: ", gse_except_srcfile(pExcept),
		                   gse_except_line(pExcept));
	else
		nWrote += snprintf(xcept_buf + nWrote, 1023 - nWrote, 
		                   "\tFrom (unknown file), line %d: ", 
		                   gse_except_line(pExcept));
			
	
	if(nWrote < 1023){
		if(except_message(pExcept) != NULL)
			nWrote += snprintf(xcept_buf + nWrote, 1023 - nWrote,
			                   "%s", except_message(pExcept));
		else
			nWrote += snprintf(xcept_buf + nWrote, 1023 - nWrote, "(no message)");		
	}
	
	/* Prn the backtrace */
	xcept_buf[1023] = '\0';
	/* gse_curMsgHandler(nLevel, xcept_buf, NULL, gse_except_backtrace(pExcept));*/
	gse_curMsgHandler(nLevel, xcept_buf, NULL, NULL, gse_bLogWithTimes);
}

gse_log_handler_t gse_log_sethandler(gse_log_handler_t func){
	gse_log_handler_t old;
	
	if(func == NULL)
		func = gse_def_log_handler;
					
	old = gse_curMsgHandler;
	gse_curMsgHandler = func;
	return old;
}

void gse_def_log_handler(int nLevel, const char* sMsg, const char* sDataStatus, 
                         const char* sStackTrace, bool bPrnTime){
	const char* sLvl;
	char buf[32];
	char sTime[32];
#ifndef __MINGW32__
	struct tm bdTime;
#endif
	time_t tEpoch;
	buf[31] = '\0';
	sTime[31] = '\0';
	
	switch(nLevel){
		case GSE_TRACE: sLvl = "TRACE"; break;
		case GSE_DEBUG: sLvl = "DEBUG"; break;
		case GSE_NOTICE: sLvl = "INFO"; break; 
		case GSE_WARNING: sLvl = "WARNING"; break; 
		case GSE_ERROR: sLvl = "ERROR"; break;	
		case GSE_CRITICAL: sLvl = "CRITICAL"; break;
		default:
			snprintf(buf, 31, "LEVEL %d MSG", nLevel); sLvl = buf; break;
	}
	
	if(bPrnTime){
		tEpoch = time(NULL);
#ifdef __MINGW32__
		/* Note: localtime is thread safe on win32, see note:
		 * http://sources.redhat.com/ml/pthreads-win32/2005/msg00011.html
		 */
		strftime(sTime, 31, "%Y-%m-%dT%H:%M:%S", localtime( &tEpoch));
#else
		strftime(sTime, 31, "%Y-%m-%dT%H:%M:%S", localtime_r( &tEpoch, &bdTime));
#endif
	
		fprintf(stderr, "(%s, %8s) ", sTime, sLvl);
	}
	else{
		fprintf(stderr, "%8s: ", sLvl);
	}
	fprintf(stderr, "%s\n", sMsg);
	if(sDataStatus != NULL)
		fprintf(stderr, "\t\n\tdata status\n\t-----------\n%s", sDataStatus);
	if(sStackTrace != NULL)
		fprintf(stderr, "\t\n\tbacktrace\n\t---------\n%s", sStackTrace);
}
