/** @file GseLog.h Simple message logging. */
#ifndef _GseLog_H_
#define _GseLog_H_

#include <rpwgse/GseExcept.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_log Logging 
 *
 * Generic thread safe logging.
 * By default messages are simply printed to standard error, use
 * gse_set_log_handler() to send messages some where else.  All log messages 
 * are sent via gse_log(), however the following convience macros make for
 * less typing:
 * 
 *  - gse_trace(): Log a GSE_TRACE level message.
 *  - gse_trace_v(): Log a GSE_TRACE level message using fprintf style varargs.
 *  - gse_debug(): Log a GSE_DEBUG level message.
 *  - gse_debug_v(): Log a GSE_DEBUG level message using fprintf style varargs.
 *  - gse_notice(): Log a GSE_NOTICE level message.
 *  - gse_notice_v(): Log a GSE_NOTICE level message using fprintf style varargs.
 *  - gse_warning(): Log a GSE_WARNING level message.
 *  - gse_warning_v(): Log a GSE_WARNING level message using fprintf style varargs.
 *  - gse_error(): Log a GSE_ERROR level message.
 *  - gse_error_v(): Log a GSE_ERROR level message using fprintf style varargs.
 *  - gse_critical(): Log a GSE_CRITICAL level message, these should be reserved for program exit conditions.
 *  - gse_critical_v(): Log a GSE_CRITICAL level message using fprintf style varargs, these should be reserved for program exit conditions.
 *
 * For example a log line such as:
 * @code
 *   gse_warning_v("File %s, Pkt %05d: Header Block > 256 bytes", sFile, nIdx);
 * @endcode
 * is equivalent to:
 * @code
 *   gse_log(GSE_WARNING, __FILE__, __LINE__, 
 *           "File %s, Pkt %05d: Header Block > 256 bytes", sFile, nIdx);
 * @endcode
 * but shorter.
 */
/*@{*/

#define GSE_EVERYTHING   0

#define GSE_TRACE       20
#define GSE_DEBUG       30
#define GSE_NOTICE      40
#define GSE_INFO        40 /* Alias for GSE_NOTICE */
#define GSE_WARNING     50
#define GSE_ERROR       60
#define GSE_CRITICAL    70

#define GSE_NOTHING    100


/** Output source file and line numbers for messages at or above this level */
bool gse_log_set_showline(int nLevel);

/** Set the logging level for this thread.
 *
 * @param nLevel Set to one of
 *   - GSE_TRACE
 *   - GSE_DEBUG
 *   - GSE_NOTICE
 *   - GSE_WARNING
 *   - GSE_ERROR
 *   - GSE_CRITICAL
 *   - GSE_NOTHING
 *
 * @return The previous log level.
 */
int gse_log_setlevel(int nLevel);

/** Get the current logging level, messages with nLevel at or higher than
 * this value should be logged. */
int gse_log_getlevel(void);

/** State that timestamps should, or should not be included in the log 
 * messages.  The default is not to include timestamps
 */
void gse_log_include_time(bool bPrnTime);

/** basic logging function, no provision for printing backtraces */
void gse_log(int nLevel, const char* sSrcFile, int nLine, const char* sFmt, ...);


/** Even more basic logging function, good for use in python code */
void gse_log_msg(int nLevel, const char* sMsg);

/** Macro wrapper around gse_log() for TRACE messages with out variable args */
#define gse_trace(M) gse_log(GSE_TRACE, __FILE__, __LINE__, M)
/** Macro wrapper around gse_log() for DEBUG messages with out variable args */
#define gse_debug(M) gse_log(GSE_DEBUG, __FILE__, __LINE__, M)
/** Macro wrapper around gse_log() for NOTICE messages with out variable args */
#define gse_notice(M) gse_log(GSE_NOTICE, __FILE__, __LINE__, M)
/** Alias for gse_notice */
#define gse_info(M) gse_log(GSE_NOTICE, __FILE__, __LINE__, M) 
/** Macro wrapper around gse_log() for WARNING messages with out variable args */
#define gse_warning(M) gse_log(GSE_WARNING, __FILE__, __LINE__, M) 
/** Macro wrapper around gse_log() for ERROR messages with out variable args */
#define gse_error(M) gse_log(GSE_ERROR, __FILE__, __LINE__, M) 
/** Macro wrapper around gse_log() for CRITICAL messages with out variable args */
#define gse_critical(M) gse_log(GSE_CRITICAL, __FILE__, __LINE__, M)
  
/** Macro wrapper around gse_log() for TRACE messages with variable arguments */
#define gse_trace_v(F, ...)\
  gse_log(GSE_TRACE, __FILE__, __LINE__, F, __VA_ARGS__)
/** Macro wrapper around gse_log() for DEBUG messages with variable arguments */
#define gse_debug_v(F, ...)\
  gse_log(GSE_DEBUG, __FILE__, __LINE__, F, __VA_ARGS__)
/** Macro wrapper around gse_log() for NOTICE messages with variable arguments */
#define gse_notice_v(F, ...)\
  gse_log(GSE_NOTICE, __FILE__, __LINE__, F, __VA_ARGS__)
/** Alias for gse_notice_v */
#define gse_info_v(F, ...)\
  gse_log(GSE_NOTICE, __FILE__, __LINE__, F, __VA_ARGS__) 
/** Macro wrapper around gse_log() for WARNING messages with variable arguments */
#define gse_warning_v(F, ...)\
  gse_log(GSE_WARNING, __FILE__, __LINE__, F, __VA_ARGS__) 
/** Macro wrapper around gse_log() for ERROR messages with variable arguments */
#define gse_error_v(F, ...)\
  gse_log(GSE_ERROR, __FILE__, __LINE__, F, __VA_ARGS__) 
/** Macro wrapper around gse_log() for CRITICAL messages with variable arguments */
#define gse_critical_v(F, ...)\
  gse_log(GSE_CRITICAL, __FILE__, __LINE__, F, __VA_ARGS__)

/** Log an exception.
 * Formats a message and backtrace and hands them off to the logging function
 * set via @ref gse_log_sethandler, or te the default logging function if none
 * has been set.
 * @param nLevel the log level for the message.  When in doubt just use
 *        GSE_ERROR for the log level.
 * @param pExcept the exception pointer as returned from #except_try_push.
 */
void gse_log_except(int nLevel, except_t* pExcept);

/** Definition of a message handler function pointer.
 * Message handlers need to be prepared for any of the string pointers 
 * sMsg, sDataStatus, or sStackTrace to be null.
 *
 * @param nLevel The message level.  If nLevel is equal to or greater than
 *  gse_log_getlevel() then the message should be logged.
 * 
 * @param sMsg The message, usually not null.
 * 
 * @param sDataStatus the current data status (Always null not yet implimented)
 *
 * @param sStackTrace A string giving the current stacktrace if an
 *        exception triggered the log request
 * 
 * @param bPrnTime The current system time should be included in the log
 *        output.
 */ 
typedef void (*gse_log_handler_t)(int nLevel, const char* sMsg, 
									 const char* sDataStatus, 
                            const char* sStackTrace, bool bPrnTime);

/** Install a new message handler function for this thread.
 * The default message handler just prints to stderr, which is not very 
 * effecient, nor is it appropriate for GUI applications.
 * 
 * @param new_handler The new message handler, or NULL to set to the default
 *        handler.
 * @return The previous message handler function pointer
 */
gse_log_handler_t gse_log_sethandler(gse_log_handler_t new_handler);

/*@}*/
/*@}*/	
#ifdef __cplusplus
}
#endif

#endif /* _GseLog_H_ */
