/** @file GseMD5.h Generating MD5 sums */

#ifndef _GseMD5_H_
#define _GseMD5_H_

/*****************************************************************************
 * MD5.H - header file for MD5C.C
 *
 *
 * Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
 * rights reserved.
 *
 * License to copy and use this software is granted provided that it
 * is identified as the "RSA Data Security, Inc. MD5 Message-Digest
 * Algorithm" in all material mentioning or referencing this software
 * or this function.
 *
 * License is also granted to make and use derivative works provided
 * that such works are identified as "derived from the RSA Data
 * Security, Inc. MD5 Message-Digest Algorithm" in all material
 * mentioning or referencing the derived work.
 *
 * RSA Data Security, Inc. makes no representations concerning either
 * the merchantability of this software or the suitability of this
 * software for any particular purpose. It is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this
 * documentation and/or software.
 *
 *
 * MD5 context. 
 *
 * This is the original file, edited version is below //
 * #ifdef __cplusplus
 * extern "C" {
 * #endif
 *
 * typedef struct {
 *   UINT4 state[4];                                   * state (ABCD) *
 *   UINT4 count[2];        * number of bits, modulo 2^64 (lsb first) *
 *   unsigned char buffer[64];                         * input buffer *
 * } MD5_CTX;
 *
 *
 *
 * void MD5Init PROTO_LIST ((MD5_CTX *));
 * void MD5Update PROTO_LIST
 *   ((MD5_CTX *, unsigned char *, unsigned int));
 * void MD5Final PROTO_LIST ((unsigned char [16], MD5_CTX *));
 *
 * #ifdef __cplusplus
 * } //extern
 * #endif
 *
 ****************************************************************************/

#include <rpwgse/GseTypes.h> 
		 		 
#ifdef __cplusplus
extern "C" {
#endif

   
/* Size of an md5 hash string */
#define GSE_MD5HASH_SZ 32
   
/* POINTER -> byte* */

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_hash Hashes
 *
 * Various CRC and Message Digest routines 
 */
/*@{*/

/** @example hash1.c
 * Basic code structure for getting a file's MD5 sum.
 */

/** Calculate an MD5 Hash on some set of bytes
 * This object does not copy the bytes internally, instead the hash is
 * calculated a "chunk" at a time, so indefinitly long memory blocks 
 * may be hashed.  This code is just a convienent C++ wrapper around the
 * "RSA Data Security, Inc. MD5 Message-Digest Algorithm".  See the header
 * file: GseMD5.h for the license statement.
 *
 * See @link hash_example.c @endlink for an example of using the md5 
 * hashing routines.
 */
typedef struct gse_md5hasher{
	
	/** Add to the set of bytes over which to calcuate the hash. */
	void (*update)(struct gse_md5hasher* pThis, byte* input, uint4 inputLen);
			
	/** State that you are finished adding bytes, and get the binary hash.
	 *
	 * @returns the hash as a pointer to a 16 byte array.  The return value
	 *          points to an internal data structure, so you must copy 
	 *          the hash somewhere else before calling @a del.
	 */
	const byte* (*final)(struct gse_md5hasher* pThis);

	/** State that you are finished adding bytes, and get a string hash.
	 *
	 * @returns the final hash as a C string. This is a pointer to an
	 *          internal data structure, so you must copy it somewhere
	 *          else before calling @a del.
	 */
	const char* (*finalStr)(struct gse_md5hasher* pThis);
	
	byte digest[16];   /* binary version of the digest     */
	char sDigest[33];  /* character version of the digest  */
	
	/* This stuff used to be called the MD5_CONTEXT */
	
	uint4 state[4];   /* m_state - state (ABCD) */
	uint4 count[2];   /* m_count - number of bits, modulo 2^64 (LSB first) */
	byte  buffer[64]; /* input buffer */
	
	/* Remember when final is called, don't finallize twice */
	bool bFinal;
	
	/** Desturctor, frees what little memory the hasher uses */
	void (*del)(void* vpThis);	
	
} GseMD5Hasher;

/** MD5 Hasher Constructor
 * Make a new hasher allocated on the heap
 * @public 
 * @memberof GseMD5Hasher
 */
GseMD5Hasher* new_GseMD5Hasher();

/** Convienence function to hash an entire file
 *
 * @param sPath The file to hash
 * @returns The GseMD5Hasher object used to calculate the hash with the
 *          @a digest and @a sDigest members in their final state after all
 *          file bytes have been passed through.  The caller is responsible 
 *          for free'ing the hasher.
 */
GseMD5Hasher* gse_get_file_md5(const char* sPath);


/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _gsemd5_H_ */
