/** @file GseMath.h Simple Math Routines */
#ifndef _GseMath_H_
#define _GseMath_H_

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_simp_math Simple Math 
 *
 * Wrappers for libm functionality, as well as simple math routines such as
 * linear interpolation.
 */
/*@}*/

/** @addtogroup mod_simp_math */
/*@{*/

/** Wrapper for sqrt
 *
 * @param x input value
 * @returns the Non-negative square root of x.
 * @throws (CLIB_XCEPT, EDOM) if X is negative.
 */
float8 gse_sqrt(float8 x);

/** Wrapper for log10
 *
 * @param x input value
 * @returns the base 10 log of @a x
 * @throws (CLIB_XCEPT, errno) if X is 0 or negative 
 */
float8 gse_log10(float8 x);


/** Wrapper for pow.
 *
 * @returns The value of @a x raised to the power of @a y.
 * @throws (CLIB_XCEPT,EDOM) if @a x is negative and @a y is not an 
 *         integral value as this would result in a complex number.
 */
float8 gse_pow(float8 x, float8 y); 

/** Simple linear interpolation of a value.
 *  @warning This function does not check to make sure the X value to be
 *           interpolated actually lies within the given X boundaries.
 * 
 * @param x0 The first x value.
 * @param x1 The second x value.
 * @param y0 The value of Y at x0.
 * @param y1 The value Y at x1
 * @param x_interp The value of X at which to linearly interpolte a Y value.
 * @returns The interpolated Y value.
 */
float8 gse_lerp(float8 x0, float8 x1, float8 y0, float8 y1, float8 x_interp);


/** Log base 10 with substitution for log10(0).
 *
 * Convienence routine to allow for substitutions for log of 0.
 *
 * @param val The value for which the log base 10 must be found.
 * @param sub The value to return if val is 0.
 * @returns log10( @a val ) or @a sub ib @a val is 0.
 */
float8 gse_log10_sub(float8 val, float8 sub);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseMath_H_ */
