#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <libgen.h>

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseFile.h"
#include "GseLog.h"
#include "GseString.h"
#include "GseMultiFileSink.h"

/*****************************************************************************/
int4 GseMultiFileSink_received(const GsePktSink* pBase)
{
	const GseMultiFileSink* pThis = (GseMultiFileSink*)pBase;
	return pThis->_recieved;
}

/*****************************************************************************/
int4 GseMultiFileSink_flushed(const GsePktSink* pBase)
{
	const GseMultiFileSink* pThis = (GseMultiFileSink*)pBase;
	return pThis->_flushed;
}

/*****************************************************************************/
void GseMultiFileSink_consumeNext(GsePktSink* pBase, GsePkt* pPkt)
{
	pBase->writeNext(pBase, pPkt);
	pPkt->del(pPkt);
}

/*****************************************************************************/
void GseMultiFileSink_open(GseMultiFileSink* pThis, const char* sPath)
{
	int openflags = 0;
	char* sPathCopy = NULL;
	char* sDir = NULL;
	mode_t dirmode = 0;
	
	/* Make sure the sPath arg is not smoking crack. */
	if((sPath == NULL)||(sPath[0] == '\0'))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	if((sPath[strlen(sPath) - 1] == '/')||   /* Can't end in "/" */
	   (strcmp(".", sPath) == 0)        ||   /* Can't be '.'     */
	   (strcmp("..", sPath) == 0))           /* Can't be '..'    */
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "'%s' is not a file path", sPath);
	
	/* Close the previously open file, if any. */
	if( pThis->_fdCur != -1){
		free(pThis->_sCurPath);
		close(pThis->_fdCur);
		pThis->_sCurPath = NULL;
	}
	
	/* Make the directory tree if needed, since the dir doesn't exist, we
	   won't need to worry about overwrite failures. */
	sPathCopy = CS(sPath);
	except_cleanup_push(free, sPathCopy);  /* Damn POSIX dirname */
	
	sDir = dirname(sPathCopy);
	
	if(! gse_isdir(sDir)){
		dirmode = pThis->_mode;
#ifndef __MINGW32__
		if(pThis->_mode &  S_IRUSR) dirmode |= S_IXUSR;
		if(pThis->_mode &  S_IRGRP) dirmode |= S_IXGRP;
		if(pThis->_mode &  S_IROTH) dirmode |= S_IXOTH;
#endif
		gse_mkdirs(sDir, dirmode);	
			
	}
	else{
		/* Well we could have a problem... */
		if(gse_isfile(sPath) && !(pThis->_uFlags&GSE_OVERWRITE) && 
		   !(pThis->_uFlags&GSE_APPEND))
			gse_throw_fm(GSE_XCEPT, GSE_XOVERWRITE, 
			             "Cowardly refusing to overwrite '%s'.  Use GSE_OVERWRITE "
			             "or GSE_APPEND flags to overwrite or append to an "
		   	          "existing file.", sPath);
	}
	
	except_cleanup_pop(except_call);   /* free sPathCopy no matter what */
	
	openflags = O_RDWR|O_CREAT;
	if(pThis->_uFlags&GSE_OVERWRITE) openflags |= O_TRUNC;
	pThis->_fdCur = gse_open(sPath, openflags, pThis->_mode);
	
	pThis->_sCurPath = CS(sPath);
	
	if(pThis->_uFlags&GSE_APPEND)
		gse_lseek(pThis->_fdCur, 0, SEEK_END);
}

/*****************************************************************************/
void GseMultiFileSink_writeNext(GsePktSink* pBase, GsePkt* pPkt)
{
	bytebuf_t buf;
	bool bRestart = true;
	char* sPath = NULL;
	
	GseMultiFileSink* pThis = (GseMultiFileSink*)pBase;

	/* See if the packet encodes */	
	buf = pPkt->encode(pPkt);
	except_cleanup_push(free, buf.ptr);
	
	/* Alright, find out where it goes */
	sPath = pThis->_getDest(pThis->_pDestCtx, pPkt);
	except_cleanup_push(free, sPath);
	
	if( (pThis->_sCurPath == NULL)||(strcmp(pThis->_sCurPath, sPath) != 0))
		pThis->open(pThis, sPath);
	
	/* This function backs up the file pointer in the case of a problem */
	gse_write(pThis->_fdCur, buf.ptr, buf.len, &bRestart);
	
	/* Flush to disk after each packet */
#ifndef __MINGW32__
	fdatasync(pThis->_fdCur);
#endif
	pThis->_flushed += 1;
	gse_debug_v("Wrote packet %d to %s", pThis->_flushed, pThis->_sCurPath);
	
	except_cleanup_pop(except_call);       /* Removes the local sPath String */
	except_cleanup_pop(except_call);       /* Removes the local buf data */
}

/*****************************************************************************/
void delete_GseMultiFileSink(void* vpThis)
{
	GseMultiFileSink* pThis = (GseMultiFileSink*)vpThis;
	
	if(pThis->_fdCur != -1) close(pThis->_fdCur);
	if(pThis->_sCurPath != NULL) free(pThis->_sCurPath);
	free(pThis);
}

/*****************************************************************************/
GseMultiFileSink* new_GseMultiFileSink(
    gse_pkt_dest_file_t func, void* pCtx, uint4 uFlags, mode_t mode, 
	 size_t nOver)
{
	GsePktSink*       pBase = NULL;
	GseMultiFileSink* pThis = NULL;
	size_t nAlloc = 0;

	if(func == NULL) gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	if(nOver > sizeof(GseMultiFileSink))
		nAlloc = nOver;
	else
		nAlloc = sizeof(GseMultiFileSink);
	
	pThis = (GseMultiFileSink*)gse_calloc(1, nAlloc);
	pBase = (GsePktSink*)pThis;
	
	/* Interface implimentiation */
	pBase->_pIfaceOwner = (void*)pThis;
	pBase->received     = GseMultiFileSink_received;
	pBase->flushed      = GseMultiFileSink_flushed;
	pBase->consumeNext  = GseMultiFileSink_consumeNext;
	pBase->writeNext    = GseMultiFileSink_writeNext;
	pBase->del          = delete_GseMultiFileSink;
	
	pThis->open = GseMultiFileSink_open;
	pThis->del  = delete_GseMultiFileSink;
	
	/* Save the settings, but don't open the file until it's needed */
	pThis->_getDest = func;
	pThis->_pDestCtx = pCtx;
	pThis->_uFlags = uFlags;	
	pThis->_mode = mode;
	
	pThis->_recieved = 0;
	pThis->_flushed = 0;
	pThis->_fdCur = -1;
	pThis->_sCurPath = NULL;
	
	return pThis;
}
