/** @file GseMultiFileSink.h Write packets to multiple files */
#ifndef _GseMultiFileSink_H_
#define _GseMultiFileSink_H_

#include <rpwgse/GseFilePktSink.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_processing Processing */
/*@{*/

/** The type of a callback function which produces a project defined index
    for each packet */

/* typedef char* (*gse_pkt2idx_t)(void* pCtx, const GsePkt* pPkt); */
/* typedef char* (*gse_idx2file_t)(void* pCtx, const char* sIndex); */

typedef char* (*gse_pkt_dest_file_t)(void* pCtx, const GsePkt* pPkt);




/** Send all packets receieved to 1-N files on disk.
 *
 * Implements the GsePktSink interface.  A destination file is selected via a
 * callback, if the indicated file is not the current file, a new file is opened
 * in append mode.  Then Packets are encoded via GsePkt::encode and the
 * resulting byte stream is writen to disk.
 *
 * @note The default is to fail if the an output file exists, usually one
 *       will want to use this sink in append mode.
 *
 * @n
 * @n
 *
 * @class GseMultiFileSink
 * @implements GsePktSink
 */
typedef struct gse_multi_file_sink{
	
	/** Pointers to this structure can be cast to a generic GsePktSink struct */
	GsePktSink 	base;
	
	/** @protected Return a file pointer given a file name.  You can override
	 *  this, and add new flags to get file locking.
	 *
	 * @param pThis A pointer to GseMultiFileSink object.
	 *
	 * @param sFile the file to open, uses the stored @a uFlags to decide
	 *        if existing files must be appended to or overwritten
	 *
	 * @throws (GSE_XCEPT, GSE_XBADARG) if there is a problem with sFile
	 *
	 *         (GSE_XCEPT, GSE_XOVERWRITE) if opening would overwrite an
	 *         existing file and the overwrite flag is not set.
	 * 
	 */
	void (*open)(struct gse_multi_file_sink* pThis, const char* sFile);
	
	gse_pkt_dest_file_t _getDest;
	
	void* _pDestCtx;
	
	char* _sCurPath;
	
	int _fdCur;
	
	uint4 _recieved;
	
	uint4 _flushed;
	
	mode_t _mode;
	
	uint4 _uFlags;
	
	/** Destructor */
	void (*del)(void* vpThis);
		
} GseMultiFileSink;


/** Make a new GseMultiFileSink
 *   
 * @note This implementation of the GsePktSink interface @b does define the
 *       @a GsePktSink::del member.  So it is safe to store a pointer to an
 *       instance of this structure as a GsePktSink pointer. 
 *
 * @param func A callback used to get the destination filename for a packet
 *
 * @param pCtx This pointer will be supplied to the callback to carry 
 *        whatever context informaiton the library user decides, may be NULL.
 *
 * @param uFlags 0 or more of the following may be specified:
 *    - @b GSE_APPEND:    Append packets if the file exists
 *    - @b GSE_OVERWRITE: Overwrite existing files
 *
 * @param mode The file creation mode, see man stat for modes.  Note that
 *        if the flag GSE_MAKEDIRS is specified, directory modes will add
 *        the execute bit in each place that the read bit is seen in the mode
 *
 * @param nOver Used by derived structures, set this to 0.
 *
 * @return The new packet since object.  Never returns NULL.
 * 
 * @throws (GSE_XCEPT, GSE_XBADARG) if there is a problem with the function args
 *
 * @public
 * @memberof GseMultiFileSink
 */
 
GseMultiFileSink* new_GseMultiFileSink(
    gse_pkt_dest_file_t func, void* pCtx, uint4 uFlags, mode_t mode, 
	 size_t nOver);

         
 /*@}*/

#ifdef __cplusplus
}
#endif
 
#endif /* _GseMultiFileSink_H_ */
