#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include "GseExcept.h"
#include "GseLog.h"
#include "GseWrap.h"
#include "GseTime.h"
#include "dict.h"

#include "GseMultiFileSrc.h"


/******************************************************************************/
/******************************************************************************/

/* String - String dictionary for tracking highest sent index by category */

#define S gse_str_str_dict
#define C GseStrStrDict
#define V char
#define K char

#include "ptrDictTplt.h"

#undef S
#undef C
#undef V
#undef K


#define C GseStrStrDict
#define V char
#define K char
#define FREE_KEYS
#define FREE_VALS
#define KEY_COMPARE strcmp

#include "ptrDictTplt.c"

#undef KEY_COMPARE
#undef FREE_KEYS
#undef FREE_VALS
#undef C
#undef V
#undef K

/******************************************************************************/
/******************************************************************************/

/* Used to keep track of where I'm at on reading packets from a channel */
typedef struct gse_file_chan_state{
	
	/* A packet waiting to go out */
	GsePkt* pPkt;
	
	/* Lookup indicies of the last catorigized packet sent */
	GseStrStrDict* pSentIdx;
	
	/* The last read filename */
	char* sCurPath;
	
	/* Flag to restart reading from the start index, this is part of the
	   deleted file interruption handling. */
	bool bStartFromIdx;
	
	/* Flag to say this channel is done */
	bool bDone;
	
	/* The file reader */
	GsePktSrc* pSrc;
	
	/* The delete function */
	void (*del)(void* vpThis);
	
} GseFileChanState;

/******************************************************************************/
void delete_GseFileChanState(void* vpThis)
{
	GseFileChanState* pThis = (GseFileChanState*)vpThis;
	
	if(pThis->pPkt != NULL) pThis->pPkt->del(pThis->pPkt);
	if(pThis->pSrc != NULL) pThis->pSrc->del(pThis->pSrc);
	if(pThis->pSentIdx != NULL) pThis->pSentIdx->del(pThis->pSentIdx);
	free(pThis->sCurPath);
	free(vpThis);
}

GseFileChanState* new_GseFileChanState()
{
	GseFileChanState* pThis;
	pThis = (GseFileChanState*)gse_calloc(1, sizeof(GseFileChanState));
	pThis->del = delete_GseFileChanState;
	pThis->bDone = false;
	pThis->bStartFromIdx = false;
	pThis->pSentIdx = new_GseStrStrDict(NULL);
	
	return pThis;
}

/******************************************************************************/

#define S gse_file_ch_state_ary
#define C GseFileChStateAry
#define T GseFileChanState

#include "ptrAryTplt.h"

#undef S
#undef C
#undef T


#define C GseFileChStateAry
#define T GseFileChanState
#define CASCADE_DEL

#include "ptrAryTplt.c"

#undef CASCADE_DEL
#undef C
#undef T


/******************************************************************************/
/******************************************************************************/


/******************************************************************************/
/* Helper for readNext when reseting the read, may return NULL */

const char* GseMultiFileSrc_getLowestSent(GseMultiFileSrc* pThis, int4 iChan)
{
	const char* sCat = NULL;
	const char* sCur = NULL;
	const char* sLowest = NULL;
	
	GseStrStrDict* pDict = pThis->pStateAry->pAry[iChan]->pSentIdx;
	
	for(sCat = pDict->first(pDict); sCat != NULL; sCat = pDict->next(pDict, sCat)){
		
		sCur = pDict->get(pDict, sCat);
		if( sCur == NULL )
			continue;
		
		if( ! sLowest ){
			sLowest = sCur;
		}
		else{
			if( pThis->pMgr->compare(pThis->pMgr, sCur, sLowest) < 0 )
				sLowest = sCur;
		}
	}
	
	return sLowest;
}

/******************************************************************************/
/* This is where the rubber hits the road */
GsePkt* GseMultiFileSrc_readNext(GsePktSrc* pBase)
{
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	
	bool bAllValid = false;
	bool bBackToTop = false;
	int4 iCh = -1, jCh = -1;
	GsePkt* pPkt = NULL;
	GsePkt* pEmit = NULL;
	GseFileChanState* pState = NULL;
	GseFileChanState* pJState = NULL;
	const char* sNewPath = NULL;
	char* sIdx = NULL;
	const char* sBegIdx = NULL;
	const char* sEndIdx = NULL;
	char* sCat = NULL;            /* data category */
	char* sLowest = NULL;
	char* sCur = NULL;
	int4 iLowest = -1;
	const char* sSentIdx;
	
	static const except_id_t catch_spec[] = {{CLIB_XCEPT, XCEPT_CODE_ANY}};
	except_t* pExcept = NULL;

	bAllValid = true;
	
	while(pEmit == NULL){	
		
		/* The channel prep loop */
		iCh = 0;
		while(iCh < pThis->pStateAry->num){
		
			pThis->pMgr->update(pThis->pMgr);  /* Give chan mgr chance to check */
			
			pState = pThis->pStateAry->pAry[iCh];
		
			/* Channel needs no prep, go to next */
			if(pState->bDone || pState->pPkt){ 
				iCh++;
				continue;
			}
		
			
			/* Check to see if *any* file channel is invalid, not just the current
			   one. */
			bAllValid = true;
			for(jCh = 0; jCh < pThis->pStateAry->num; jCh++){
				
				if(! pThis->pMgr->isValid(pThis->pMgr, jCh)){
					pJState = pThis->pStateAry->pAry[jCh];
					pJState->bStartFromIdx = true;
					if(pJState->pSrc != NULL){
						pJState->pSrc->del(pJState->pSrc);
						pJState->pSrc = NULL;
						pThis->pMgr->clearCurFile(pThis->pMgr, jCh);
					}
							
					bAllValid = false;
				}
			}
			
			if(!bAllValid){
				gse_info_v("Current file moved, waiting %d secs for filesystem to"
				           " settle...", pThis->nHoldOff);
				gse_millisleep(1000* pThis->nHoldOff, true);
	
				for(jCh = 0; jCh < pThis->pStateAry->num; jCh++)
					pThis->pMgr->clearInvalid(pThis->pMgr, jCh);
				
				/* Don't advance iCh... */
				continue;
			}
		
			
			/* Start a reader if needed */
		
			if(pState->pSrc == NULL){
			
				if(pState->bStartFromIdx == true){
					sSentIdx = GseMultiFileSrc_getLowestSent(pThis, iCh);
					sNewPath = pThis->pMgr->getFile(pThis->pMgr, iCh, sSentIdx);
				}
				else{
					sNewPath = pThis->pMgr->nextFile(pThis->pMgr, iCh, pState->sCurPath);
				}
				
				if(sNewPath == NULL){
					iCh++;  /* Can't setup this channel, go to the next one */	
					if(pThis->_eof_disp == at_eof_stop)	pState->bDone = true;
					continue;
				}
				else{
				
					/* The file could disappear before I can read it, be ready for that */
					except_try_push(catch_spec, 1, &pExcept);
					if(pExcept == NULL){
						gse_info_v("Opening '%s'", sNewPath);
						pState->pSrc = (GsePktSrc*)
								new_GseFilePktSrc(sNewPath,pThis->factory,at_eof_retry);
						free(pState->sCurPath);
						pState->sCurPath = CS(sNewPath);
						pState->bStartFromIdx = false;
					}
					else{
						gse_log_except(GSE_WARNING, pExcept);
						pState->pSrc = NULL;
					}
					except_try_pop();
					
					if(pState->pSrc == NULL){
						iCh++;  /* Can't setup this channel, go to the next one */
						if(pThis->_eof_disp == at_eof_stop) pState->bDone = true;
						continue;
					}
				}
			}
			
			/* I've done all I can to get a reader setup, now use it */
			
			bBackToTop = false;
			while(pState->pPkt == NULL){
				
				/* Check for invalid channels during the read */
				pThis->pMgr->update(pThis->pMgr);
				
				for(jCh = 0; jCh < pThis->pStateAry->num; jCh++){
					if(! pThis->pMgr->isValid(pThis->pMgr, jCh)){
						bBackToTop = true;
						break;
					}
				}
			
				/* Read a packet from this source, or decide you can't get one. */
				pPkt = NULL;
				if( (pPkt = pState->pSrc->readNext(pState->pSrc)) == NULL){
					
					/* See if there is a new file to go to */
					sNewPath = pThis->pMgr->nextFile(pThis->pMgr, iCh, pState->sCurPath);
				
					if(sNewPath != NULL){
						gse_debug_v("Closing '%s'", pState->pSrc->name(pState->pSrc));
						pState->pSrc->del(pState->pSrc);
						pState->pSrc = NULL;
						/* Do not advance the channel index */
						/* Do not change the current file string (sPath) */
					}
					else{
						if(pThis->_eof_disp == at_eof_stop) pState->bDone = true;
						iCh++; /* go to the next channel there are no more files here */
					}
					bBackToTop = true;
					break;
				}
				
				
				/* Okay is this packet index > start ? */
				sIdx = pThis->pMgr->getPktIdx(pThis->pMgr, pPkt);
				sBegIdx = pThis->pMgr->begIdx(pThis->pMgr);
				
				if(pThis->pMgr->compare(pThis->pMgr, sIdx, sBegIdx) < 0){
					pPkt->del(pPkt);
					pPkt = NULL;
					free(sIdx);
					continue;
				}
				
				/* Is this pkt idx > the high water mark for this category? */
				sCat = pThis->pMgr->getPktCat(pThis->pMgr, iCh, pPkt);
				
				sSentIdx = NULL;
				if( pState->pSentIdx->hasKey(pState->pSentIdx, sCat) )
					sSentIdx = pState->pSentIdx->get(pState->pSentIdx, sCat);	
					
				free(sCat);
						
				if(sSentIdx != NULL){
					if(pThis->pMgr->compare(pThis->pMgr, sIdx, sSentIdx) <= 0){
						pPkt->del(pPkt);
						pPkt = NULL;
						free(sIdx);
						continue;
					}
				}
				
				
				/* Is this pkt idx > the ending index? If so end the channel */
				sEndIdx = pThis->pMgr->endIdx(pThis->pMgr);
				if(pThis->pMgr->compare(pThis->pMgr, sIdx, sEndIdx) > 0){
					pPkt->del(pPkt);
					free(sIdx);
					pState->bDone = true;
					gse_debug_v("Closing '%s'", pState->pSrc->name(pState->pSrc));
					pState->pSrc->del(pState->pSrc);
					pState->pSrc = NULL;
					iCh++;
					bBackToTop = true;
					break;
				}
				
				pState->pPkt = pPkt;
				free(sIdx);
			}
			
		
			if(bBackToTop) continue;
			
		} /* End of the make channel ready loop */
			
		
		/* I've done everything I can to get a packet from all the channels, 
			emitt the lowest one */
		
		sLowest = NULL;
		iLowest = -1;
		sCur = NULL;
		for(iCh = 0; iCh < pThis->pStateAry->num; iCh++){
			pState = pThis->pStateAry->pAry[iCh];
			if(pState->pPkt == NULL) continue;
			
			if(sLowest == NULL){
				sLowest = pThis->pMgr->getPktIdx(pThis->pMgr, pState->pPkt);
				iLowest = iCh;
			}
			else{
				sCur = pThis->pMgr->getPktIdx(pThis->pMgr, pState->pPkt);
				
				if(pThis->pMgr->compare(pThis->pMgr, sLowest, sCur) > 0){
					free(sLowest);
					sLowest = sCur;
					iLowest = iCh;
				}
				else{
					free(sCur);
				}
			}
			
		}
		
		if(iLowest != -1){
			pState = pThis->pStateAry->pAry[iLowest]; 
			
			pEmit = pState->pPkt;  /* <------------ Will stop the loop */
			
			pThis->iLastSentCh = iLowest;
			pState->pPkt = NULL;
			
			sCat = pThis->pMgr->getPktCat(pThis->pMgr, iLowest, pEmit);
			
			if( pState->pSentIdx->hasKey(pState->pSentIdx, sCat) ){
				free( pState->pSentIdx->replace(pState->pSentIdx, sCat, sLowest) );
			}
			else{
				pState->pSentIdx->insert(pState->pSentIdx, sCat, sLowest);
			}

		}
		else{
			/* If all pPkts are NULL wait and try again, or just exit */
			if((pThis->_eof_disp == at_eof_stop)||
				(pThis->_eof_disp == at_eof_retry)){
				break;
			}
			else{
				gse_millisleep(500, false);
			}
		}
	}
	
	if(pEmit != NULL){
		if(gse_log_getlevel() <= GSE_DEBUG){
			
			const char* sChan; 
			
			sChan = pThis->pMgr->chanName(pThis->pMgr, pThis->iLastSentCh);
			sIdx = pThis->pMgr->getPktIdx(pThis->pMgr, pEmit);
			sCat = pThis->pMgr->getPktCat(pThis->pMgr, pThis->iLastSentCh, pEmit);
			
			gse_debug_v("Emiting from chan: %s  Pkt with idx: %s & cat: %s", sChan, sIdx, sCat);
			
			free(sIdx);
			free(sCat);
		}
	}
	return pEmit;
}

/******************************************************************************/
void GseMultiFileSrc_rewind(GsePktSrc* pBase)
{
	/* reset to go */
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	int iCh;
	
	for(iCh = 0; iCh < pThis->pStateAry->num; iCh++){
		pThis->pStateAry->pAry[iCh]->del(pThis->pStateAry->pAry[iCh]);
		pThis->pStateAry->pAry[iCh] = new_GseFileChanState();
	}
}

/******************************************************************************/
#ifdef GSE_POSIX_THREADS
void GseMultiFileSrc_interrupt(GsePktSrc* pBase, pthread_t threadID)
{
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	GsePktSrc* pSrc = NULL;
	int4 iCh;
	
	for(iCh = 0; iCh < pThis->pStateAry->num; iCh++){
		if( (pSrc = pThis->pStateAry->pAry[pThis->iLastSentCh]->pSrc) != NULL){
			pSrc->interrupt(pSrc, threadID);
		}
	}
}
#endif /* GSE_POSIX_THREADS */

/******************************************************************************/
size_t GseMultiFileSrc_offset(GsePktSrc* pBase){
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	GsePktSrc* pSrc = NULL;
	
	if(pThis->iLastSentCh == -1) return -1;
	
	if( (pSrc = pThis->pStateAry->pAry[pThis->iLastSentCh]->pSrc) == NULL)
		return -1;
	
	return pSrc->offset(pSrc);
}

const char* GseMultiFileSrc_name(GsePktSrc* pBase){
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	GsePktSrc* pSrc = NULL;

	if(pThis->iLastSentCh == -1) return NULL;
	
	if( (pSrc = pThis->pStateAry->pAry[pThis->iLastSentCh]->pSrc) == NULL)
		return NULL;
	
	return pSrc->name(pSrc);
}

int4 GseMultiFileSrc_pktIdx(GsePktSrc* pBase){
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)pBase;
	GsePktSrc* pSrc = NULL;

	if(pThis->iLastSentCh == -1) return -1;
	
	if( (pSrc = pThis->pStateAry->pAry[pThis->iLastSentCh]->pSrc) == NULL)
		return -1;
	
	return pSrc->pktIdx(pSrc);
}

/******************************************************************************/
bool GseMultiFileSrc_false(const GsePktSrc* pBase){   return false;  }
bool GseMultiFileSrc_true(const GsePktSrc* pBase){    return true;   }
int4 GseMultiFileSrc_numPkts(GsePktSrc* pBase){ return -1;     }

/******************************************************************************/
void delete_GseMultiFileSrc(void* vpThis)
{
	GseMultiFileSrc* pThis = (GseMultiFileSrc*)vpThis;
	pThis->pMgr->del(pThis->pMgr);
	pThis->pStateAry->del(pThis->pStateAry);
	free(vpThis);
}

/******************************************************************************/
GseMultiFileSrc* new_GseMultiFileSrc(
	GseFileChanMgr* pMgr, gse_blk_factory factory, gse_eof_disp disp,
	int4 nHoldOff)
{
	GseMultiFileSrc* pThis = NULL;
	GsePktSrc* pBase = NULL;
	int4 i = 0;
	
	
	/* Check args */
	if((pMgr == NULL)||(factory == NULL))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	/* if(disp == at_eof_retry)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "GseMultiFileSrc doesn't implement "
		            "the at_eof_retry end-of-file disposition.");
	*/
		
	pThis = (GseMultiFileSrc*)gse_calloc(1, sizeof(GseMultiFileSrc));
	pBase = (GsePktSrc*)pThis;
	
	
	pBase->pktIdx = GseMultiFileSrc_pktIdx;   /*Return's info for most recent pkt*/
	pBase->numPkts = GseMultiFileSrc_numPkts;
	pBase->offset  = GseMultiFileSrc_offset;  /*Return's info for most recent pkt*/
	pBase->name    = GseMultiFileSrc_name;    /*Return's info for most recent pkt*/
#ifdef GSE_POSIX_THREADS
	pBase->interrupt = GseMultiFileSrc_interrupt;
#endif
	pBase->readNext = GseMultiFileSrc_readNext;
	pBase->supportsPrev = GseMultiFileSrc_false;
	pBase->supportsRewind = GseMultiFileSrc_true;
	pBase->readPrev = NULL;
	pBase->supportsSkipAhead = GseMultiFileSrc_false;
	pBase->supportsSkipBack = GseMultiFileSrc_false;
	pBase->skip = NULL;
	pBase->skipToEnd = NULL;
	pBase->rewind = GseMultiFileSrc_rewind;
	pBase->setMaxPktSz = NULL;
	pBase->getMaxPktSz = NULL;
	pBase->del = delete_GseMultiFileSrc;
	
	pThis->factory = factory;
	pThis->iLastSentCh = -1;
	pThis->_eof_disp = disp;
	
	
	pThis->pStateAry = new_GseFileChStateAry(0);
	for(i = 0; i < pMgr->channels(pMgr); i++)		
		pThis->pStateAry->add(pThis->pStateAry, new_GseFileChanState());
	
	pThis->pMgr = pMgr;
	
	if(nHoldOff < 5){
		gse_warning_v("Ignoring on-delete-retry timeout of %d seconds, using 5 "
		              "seconds instead", nHoldOff);
		pThis->nHoldOff = 5;	
	}
	else{
		pThis->nHoldOff = nHoldOff;
	}
	
	return pThis;
}



