/** @file GseMultiFileSrc.h  Read packets for multiple growing files */
#ifndef _GseMultiFileSrc_H_
#define _GseMultiFileSrc_H_

#include <rpwgse/GseString.h>
#include <rpwgse/GseFilePktSrc.h>


#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_processing Processing */
/*@{*/


/** Base class for an object that keeps track of 1-N channels of indexed 
 * files containing GsePkts.
 *
 * Time is the natural index for radio reciever data, but need not be
 * the one choosen.  It is up to the implementer to decide what the
 * abstract index strings actually mean.
 */
typedef struct gse_file_chan_mgr{
		
	/** Provide the starting index that is supposed to be read */
	const char* (*begIdx)(const struct gse_file_chan_mgr* pThis);
	
	/** Provide the ending index that is supposed to be read, may be never */
	const char* (*endIdx)(const struct gse_file_chan_mgr* pThis);
	
	
	/** Given a GsePkt, provide the index used by the other functions */
	char* (*getPktIdx)(const struct gse_file_chan_mgr* pThis, const GsePkt* pPkt);
	
	/** Compare two indicies.
	 * @return -1 if first is lower, +1 if first is higher, 0 if the same.
	 */
	int4 (*compare)(const struct gse_file_chan_mgr* pThis, 
			          const char* sIdx1, const char* sIdx2);
	
	/** Check to see if any of the files listed in the channels needs updating
	 *
	 * @returns true if all channels are still valid, false if one or more
	 *          have been invalided.
	 */
	bool (*update)(struct gse_file_chan_mgr* pThis);
	
	/** Returns the number of file channels being watched */
	size_t (*channels)(const struct gse_file_chan_mgr* pThis);

	/** Returns the name of a channel */
	const char* (*chanName)(const struct gse_file_chan_mgr* pThis, int4 iChan);
	
	/** Check to see if a channel has been invalidated */
	bool (*isValid)(struct gse_file_chan_mgr* pThis, int4 iChan);
	
	/** Clear the invalid channel flag */
	void (*clearInvalid)(struct gse_file_chan_mgr* pThis, int4 iChan);	
	
	/** Get the category for a packet.  The category can be the same as
	 * the channel name, or it may be some sub-specilization
	 *
	 * @param pThis a pointer to a GseFileChanMgr object
	 * 
	 * @param iChan the channel index
	 *
	 * @param pPkt the packet to catorigize
	 *
	 * @returns The category of the packet as a string allocated on the heap,
	 *          may NOT return NULL.
	 */
	char* (*getPktCat)(const struct gse_file_chan_mgr* pThis, int4 iChan, 
	                   const GsePkt* pPkt);
	
	/** Gets the file at or before a given index */
	const char* (*getFile)(struct gse_file_chan_mgr* pThis, int4 iChan,
			                 const char* sIdx);
	
	/** Gets the next file in the pile.
	 *
	 * @param sCurr the path to the currently read file or NULL if there
	 *        you haven't started reading anything.
	 */
	const char* (*nextFile)(struct gse_file_chan_mgr* pThis, int4 iChan,
	                        const char* sCurFile);
	
	/** Let channel manage know we are nolonger reading a file. 
	 * Internally the channel manager tracks the last filename handed out
	 * for a channel, so that if that file is deleted, the channel can be
	 * marked as invalid.
	 */
	void (*clearCurFile)(struct gse_file_chan_mgr* pThis, int4 iChan);
	
	
	/** Delete the file channel manager and free it's resources */
	void (*del)(void* vpThis);
	
} GseFileChanMgr;


/** A multi-file packet source.
 *
 * Reads from one or more groups of files, called channels.  The number
 * of channels is set in the GseFileChanMgr derived object that is 
 * passed in.
 */
typedef struct gse_multi_file_src{

	/** interface type is a GsePktSrc */
	GsePktSrc base;
	
	/* File channel manager, project specific plugin */
	GseFileChanMgr* pMgr;
	
	/* The block factory function used to generate the
	 * proper header block decoders, project specific plugin */
	gse_blk_factory factory;
	
	/* EOF disposition */
	gse_eof_disp _eof_disp; 
	
	/* How long to wait after an in-process file has been deleted before
	   re-starting the read */
	int4 nHoldOff;
	
	/* An array of file channel read state objects */
	struct gse_file_ch_state_ary* pStateAry;
	
	/* Store the last sent channel so the that the file offset
	   can be queried */
	int4 iLastSentCh;
	
	/* The destructor */
	void (*del)(void* vpThis);
	
} GseMultiFileSrc;



/** Creat a packet source that feeds from 1-N files
 *
 * Given a project specific file name provider, create a packet source that
 * can read from 1-N indexed files.
 *
 * @param pMgr A pointer to a GseFileChanMgr object
 *
 * @param factory The GseBlock factory function to use to interperate 
 *                project specific header header blocks.
 *
 * @param disp  The End of file disposition.  Can have one of the following
 * enum'erated values:
 *
 *   - \b at_eof_stop  if specified, once the end of the file is hit and
 *     the channel manager reports that there are no new files to read in
 *     all channels, GsePktSrc::nextPkt will return NULL.
 *
 *   - \b at_eof_retry if specificed, each call to GsePktSrc::nextPkt will 
 *     attempt to read the next packet's data, and return immediatly if there
 *     are insufficient bytes to form the next packet form any of the file
 *     channels.  GsePktSrc::numPkts() will always return -1.   This can be
 *     used to read a file that may be growing.
 *
 *   - \b at_eof_block if specificed, each call to GsePktSrc::nextPkt will 
 *     attempt to read the next packet's data, and will \b block if there
 *     are insufficient bytes to form the next packet in any of the file
 *     channels.  GsePktSrc::numPkts() will always return -1.   This can
 *     be used to read a file that may be growing, and to block the caller in
 *     a read loop.  Not that in the multithreaded version of the library, 
 *     GsePktSrc::interrupt may be called from a different thread to stop 
 *     a blocked read.
 *
 * @prama nHoldOff The number of seconds to halt packet reading if the 
 *        currently read file is deleted.  This gives whatever external
 *        process is mucking with things time to get finished before we
 *        reopen the updated version of the file.
 *
 * @public
 * @memberof GseMultiFileSrc
 */
GseMultiFileSrc* new_GseMultiFileSrc(
  GseFileChanMgr* pMgr, gse_blk_factory factory, gse_eof_disp disp,
  int4 nHoldOff
);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseMultiFileSrc_H_ */
