#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <strings.h>

#include "GseExcept.h"
#include "GseString.h"
#include "GseFileStr.h"
#include "GseWrap.h"

#include "GseParser.h"

/*****************************************************************************/

const char* GseParser_name(const GseParser* pThis)
{
	return (const char*)pThis->sName;
}

const GseField* GseParser_getField(const GseParser* pThis, const char* sKey)
{
	return pThis->pFields->get(pThis->pFields, sKey);
}

const GseField* GseParser_firstField(const GseParser* pThis)
{
	const char* sKey;
	sKey = pThis->pFields->first(pThis->pFields);
	if(sKey != NULL)
		return pThis->pFields->get(pThis->pFields, sKey);
	else
		return NULL;
}

const GseField* GseParser_nextField(const GseParser* pThis, 
                                    const GseField* pField)
{
	const char* sKey;
	sKey = pThis->pFields->next(pThis->pFields, pField->sName);
	if(sKey != NULL)
		return pThis->pFields->get(pThis->pFields, sKey);
	else
		return NULL;
}

char* GseParser_read(GseParser* pThis, cbytebuf_t cbuf)
{
	const char* sKey;
	const int8* pExpects;
	const GseProp* pTmpProp;
	GseFileStr* pOut;
	char* sErr = NULL;
	bool bErrs = false;
	
	GseStrFieldDict* pF = pThis->pFields;
	GseField* pField;
	
	/* First check the size of the buffer, against the expects parameter */
	if( (pExpects = gse_pd_getint(pThis->pProps, "expects")) != NULL){
		if(cbuf.len < *pExpects){
			pTmpProp = pThis->pProps->get(pThis->pProps, "expects");
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Input buffer is shorter than"
				          " %lld bytes.  Expected length defined at %s,%d.",
			             *pExpects, pTmpProp->sConf, pTmpProp->nLine);
		}
	}
	
	pOut = new_GseFileStr(128);
	
	except_cleanup_push(pOut->del, pOut);
	
	/* Loop over all of my fields asking them to read the data, build up
	   an error string while this is going on. */
	for(sKey = pF->first(pF); sKey != NULL; sKey = pF->next(pF, sKey))
	{
		pField = pF->getMutable(pF, sKey);
		if( (sErr = pField->read(pField, cbuf)) != NULL){
			pOut->printf(pOut, "    Field %s Error: %s\n",pField->name(pField),sErr);
			free(sErr);
			bErrs = true;
			sErr = NULL;
		}
	}
	
	except_cleanup_pop(except_no_call);
	
	if(bErrs){
		sErr = pOut->takestr(pOut);
		pOut->del(pOut);
	}
	return sErr;
}

uint4 GseParser_addField(GseParser* pThis, GseField* pField)
{
	pThis->pFields->insert(pThis->pFields, (char*)pField->name(pField), pField);
	return pThis->pFields->length(pThis->pFields);
}

void delete_GseParser(void* vpThis)
{
	GseParser* pThis = (GseParser*)vpThis;
	
	if(pThis->pProps != NULL) pThis->pProps->del(pThis->pProps);
	if(pThis->pFields != NULL) pThis->pFields->del(pThis->pFields);
	if(pThis->sName != NULL) free(pThis->sName);
	free(vpThis);
}

GseParser* new_GseParser(const char* sName, GseStrPropDict* pProps, 
		                   GseStrFieldDict* pFields, const char* sSrcFile)
{
	GseParser* pThis;
	
	pThis = (GseParser*)gse_calloc(1, sizeof(GseParser));
	
	
	except_cleanup_push(delete_GseParser, pThis);
	
	pThis->name       = GseParser_name;
	pThis->addField   = GseParser_addField;
	pThis->getField   = GseParser_getField;
	pThis->firstField = GseParser_firstField;
	pThis->nextField  = GseParser_nextField;
	pThis->read       = GseParser_read;
	pThis->del        = delete_GseParser;
	
	if(sSrcFile != NULL)
		pThis->sConf = sSrcFile;
	else
		pThis->sConf = "unnamed";
	
	if(sName != NULL)
		pThis->sName = CS(sName);
	else
		pThis->sName = "unnamed";
	
	if(pProps != NULL)
		pThis->pProps = pProps;    /* Takes ownership */
	else
		pThis->pProps = new_GseStrPropDict(pThis->sName);
	
	pThis->pProps->setName(pThis->pProps, pThis->sName);
	pThis->pProps->setRetNull(pThis->pProps);
	pThis->pProps->disownAllKeys(pThis->pProps);
	gse_pd_setconfig(pThis->pProps, pThis->sConf);
	
	if(pFields != NULL)
		pThis->pFields = pFields;  /* Takes ownership */
	else
		pThis->pFields = new_GseStrFieldDict(pThis->sName);
	
	pThis->pFields->setName(pThis->pFields, pThis->sName);
	pThis->pFields->setRetNull(pThis->pFields);
	pThis->pFields->disownAllKeys(pThis->pFields);
	gse_fd_setconfig(pThis->pFields, pThis->sConf);
	gse_fd_setparser(pThis->pFields, pThis);
	
	except_cleanup_pop(except_no_call);
	return pThis;
}

/*****************************************************************************/
/* A dictionary of parsers */

#define C GseStrPsrDict
#define K char
#define V GseParser

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcasecmp(a, b)

#include "ptrDictTplt.c"

#undef FREE_KEYS
#undef DEL_VALS
#undef C
#undef K
#undef V

