/** @file GseParser.h Generic Binary Blob Parsers */
#ifndef _GseParser_H_
#define _GseParser_H_

#include <rpwgse/GseTypes.h>
#include <rpwgse/GseField.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_rpwgse */
/*@{*/
/** @defgroup mod_bin_parser Binary Blob Parsers */
/*@}*/

/** @addtogroup mod_bin_parser */
/*@{*/

/** Binary Blob Parser 
 * Given a blob of bytes, this class extracts bytes from the defined fields
 * and saves it as text strings. @n
 * @class GseParser
 */
typedef struct gse_parser{

	const char* (*name)(const struct gse_parser* pThis);
	
	/** Add a new field to this parser
	 *
	 * @returns The total number of fields in this parser, including the
	 *          newly added item.
	 */
	uint4 (*addField)(struct gse_parser* pThis, GseField* pField);
	
	/** Get a field by name */
	const GseField* (*getField)(const struct gse_parser* pThis, const char* sName);
	
	/** Get the first field */
	const GseField* (*firstField)(const struct gse_parser* pThis);
	
	/** Get the next field */
	const GseField* (*nextField)(const struct gse_parser* pThis, 
			                       const GseField* pField);
	
	/** Read a blob of data an parse it. 
	 * 
	 * When this function is called the GseParser instructs all it's GseFields
	 * to read and convert thier part of the blob.  
	 *
	 * @todo Add info on how to retrieve each field's data.
	 *
	 * @param pThis the GseParser object
	 * @param cbuf a buffer pointing to the blob to parse. 
	 *
	 * @returns NULL if there were no errors parsing the data, or an string
	 *          listing all the errors (one per line) allocated on the heap.
	 */
	char* (*read)(struct gse_parser* pThis, cbytebuf_t cbuf);
		
	/** Destructor - Also deletes all contained structures */
	void (*del)(void* vpThis);
	
	/* A dictionary of properties */
	GseStrPropDict* pProps;
	
	/* A dictioary of fields */
	GseStrFieldDict* pFields;
	
	const char* sConf; /* The name of the configuration file, or "unnamed" */
		
	char* sName;
	
} GseParser;


/** Generates a new empty GseBinPsr.
 * This is usually called by gse_build_parsers() as it reads a configuration
 * file.
 * @public
 * @memberof GseParser
 */
GseParser* new_GseParser(const char* sName, GseStrPropDict* pProps, 
		                   GseStrFieldDict* pFields, const char* sSrcFile);


/* A dictionary of parsers, accessed by strings */
#define S gse_str_psr_dict
#define C GseStrPsrDict
#define K char
#define V GseParser

#include "ptrDictTplt.h"

#undef S
#undef C
#undef K
#undef V


/** Generate Blob Parsers given a config file.
 * 
 * The easiest way to generate a set of blob parsers, is to feed a
 * configuration file to this function.
 *
 * @ref page_psrbldr_config
 *
 * @param sConf The configuation data string.
 *
 * @param sFile (Optional) The name of the configuration file from which the
 *        config string was read.  This is only used when printing error
 *        messages.  If NULL filename info will not be included in error
 *        output. 
 *        @warning If sFile is supplied, do @b NOT delete it for the 
 *        life of the returned parsers and parameters.
 *
 * @param ppParams The location to receive a pointer to a GseStrPropDict 
 *        object allocated on the heap.  This will contian GseProp objects
 *        defined in the 'parameters' section.
 *
 * @param ppParsers The location to recieve a pointer to a GseStrPsrDict 
 *        object allocated on the heap.  This will contian the GseParser 
 *        objects defined through out the configuration data.
 *
 * @throws varies Generating Blob parsers is a non-trival task that may
 *                throw a veriety of errors, more details will be added
 *                later.
 */
void gse_build_parsers(const char* sConf, const char* sFile, 
                       GseStrPropDict** ppParams, GseStrPsrDict** ppParsers);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseParser_H_ */
