/* GseParser.i - Interface definition to binary blob parsers */

%rename(GseParser) cGseParser;

%newobject cGseParser::getField() const;
%newobject cGseParser::firstField() const;
%newobject cGseParser::nextField(const cGseField* pField) const;
%newobject cGseParser::read();

%header %{
#include <rpwgse/GseParser.hpp>

cGseParser::cGseParser(GseParser* pWrap){
	if(pWrap == NULL)
		gse_throw_m(GSE_XCEPT,GSE_XBADARG,"Null pointer to cGseParser constructor");
	m_pParser = pWrap;
}

cGseParser::~cGseParser(){
	m_pParser->del(m_pParser);
}

const char* cGseParser::name() const{
	return m_pParser->name(m_pParser);
}

bool cGseParser::hasField(const char* sName) const {
	return m_pParser->pFields->hasKey(m_pParser->pFields, sName);
}

cGseField* cGseParser::getField(const char* sName) const{
	const GseField* pCobj = m_pParser->getField(m_pParser, sName);
	if(pCobj == NULL) return NULL;
	return new cGseField(pCobj);
}

cGseField* cGseParser::firstField() const{
	const GseField* pCobj = m_pParser->firstField(m_pParser);
	if(pCobj == NULL) return NULL;
	return new cGseField(pCobj);
}

cGseField* cGseParser::nextField(const cGseField* pField) const{
	const GseField* pCobj = m_pParser->nextField(m_pParser, pField->m_pField);
	if(pCobj == NULL) return NULL;
	return new cGseField(pCobj);
}

char* cGseParser::read(cbytebuf_t cbuf){
	return m_pParser->read(m_pParser, cbuf);
}

bool cGseParser::hasProp(const char* sName) const{
	return m_pParser->pProps->hasKey(m_pParser->pProps, sName);
}

cGseProp* cGseParser::getProp(const char* sName) const{
	if(! m_pParser->pProps->hasKey(m_pParser->pProps, sName) )
		return NULL;
	
	/* Okay to cast, I'm not going to own the object anyway */
	GseProp* pCobj = (GseProp*) (m_pParser->pProps->get(m_pParser->pProps, sName));
	cGseProp* pNew = NULL;
	
	switch(pCobj->ptype){
	case pt_int:
	case pt_str:
	case pt_flt:   pNew = (cGseProp*)new cGsePropSimple(pCobj); break;
	case pt_list:  pNew = (cGseProp*)new cGsePropList(pCobj);   break;
	default:       pNew = (cGseProp*)new cGsePropEnum(pCobj);   break;
	}
	pNew->m_bOwner = false;
	return pNew;
}

%}

%include rpwgse/GseParser.hpp

/* Here goes the big ugly typemaps needed for the massive build parsers
   furction */
#if defined(SWIGPYTHON)

%typemap(in,numinputs=0) (GseStrPropDict** ppParams, GseStrPsrDict** ppParsers) 
{
	GseStrPropDict* pPropDict;
	GseStrPsrDict* pParserDict;
	$1 = &pPropDict;
	$2 = &pParserDict;
}


%typemap(argout) (GseStrPropDict** ppParams, GseStrPsrDict** ppParsers)
{
	
	PyObject* $1_dict = PyDict_New();
	PyObject* $2_dict = PyDict_New();
	
	$result = PyTuple_New(2);
	PyTuple_SetItem($result, 0, $1_dict);
	PyTuple_SetItem($result, 1, $2_dict);
	
	/* Loop over all the properties inserting GseProps into the dictionary */
	{
		const char* sKey;
		GseStrPropDict* pDict = *$1;
		for(sKey = pDict->first(pDict); sKey != NULL; sKey=pDict->next(pDict,sKey)){
			PyObject* pKey = PyString_FromString(sKey);
			GseProp* pCobj = pDict->replace(pDict,CS(sKey),NULL);
			cGseProp* pProp = NULL;
			switch(pCobj->ptype){
			case pt_int:
			case pt_str:
			case pt_flt:   pProp = (cGseProp*)new cGsePropSimple(pCobj); break;
			case pt_list:  pProp = (cGseProp*)new cGsePropList(pCobj);   break;
			default:       pProp = (cGseProp*)new cGsePropEnum(pCobj);   break;
			}
			PyObject* pVal = SWIG_NewPointerObj(pProp, $descriptor(cGseProp*), SWIG_POINTER_OWN);
			PyDict_SetItem($1_dict, pKey, pVal);
		}
		
		pDict->del(pDict);
	}
	
	/* Loop over all the parsers inserting GseParsers into the dictionary */
	{
		const char* sKey;
		GseStrPsrDict* pDict = *$2;
		for(sKey = pDict->first(pDict); sKey != NULL; sKey=pDict->next(pDict,sKey)){
			PyObject* pKey = PyString_FromString(sKey);
			cGseParser* pParser = new cGseParser(pDict->replace(pDict,CS(sKey),NULL));
			PyObject* pVal = SWIG_NewPointerObj(pParser, $descriptor(cGseParser*), SWIG_POINTER_OWN);
			PyDict_SetItem($2_dict, pKey, pVal);
		}
		pDict->del(pDict);
	}
}

#else
#error Extend the typemaps in GseParser.i for your target language
#endif

void gse_build_parsers(const char* sConf, const char* sFile,
                       GseStrPropDict** ppParams, GseStrPsrDict** ppParsers);


