#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <string.h>

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseConv.h"
#include "GseString.h"
#include "GseLog.h"
#include "GseCrc.h"

#define _GsePkt_C_
#include "GsePkt.h"
#include "GseInit.h"

/* Const Sync Bytes */
/* WARNING!  WARNING! WARNING! 

   if the bytes in _g_sSyncPtrn are not unique (i.e. there are repeats)
	the function GsePktSrc_search will fail !!!
	
*/
static const byte _g_sSyncPtrn[] = {0xFA, 0x6C, 0x27, 0x41};
const byte* gse_sSyncPtrn = _g_sSyncPtrn;

/****************************************************************************/
static uint4 gse_uMaxPktSz = 0xFFFFFFFF;

/****************************************************************************/
/* Field Definitions in any GSE packet */

const gse_field_def GSE_FD_CRC      = { 4, 2, true, false, 0, 0};
const gse_field_def GSE_FD_HT_LEN   = { 6, 2, true, false, 0, 0};
const gse_field_def GSE_FD_PKT_LEN  = { 8, 4, true, false, 0, 0};
const gse_field_def GSE_FD_APP_ID   = {12, 2, true, true,  0, 0};
const gse_field_def GSE_FD_VERSION  = {14, 1, true, false, 0, 0};
const gse_field_def GSE_FD_CONTENTS = {15, 1, true, false, 0, 0};

/****************************************************************************/
#ifdef GSE_POSIX_THREADS
static pthread_mutex_t s_mtx_maxPktSz = PTHREAD_MUTEX_INITIALIZER;

uint4 gse_maxPktSz(){ 
	uint4 uRet;
	pthread_mutex_lock(&s_mtx_maxPktSz);
	uRet = gse_uMaxPktSz;
	pthread_mutex_unlock(&s_mtx_maxPktSz);
	return uRet;
}

uint4 gse_setMaxPktSz(uint4 max){
	uint4 old;
	
	if(max < 21)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "The smallest possible GSE packet "
		             "that can contain data is 21 bytes");
	
	pthread_mutex_lock(&s_mtx_maxPktSz);
	old = gse_uMaxPktSz;
	gse_uMaxPktSz = max;
	pthread_mutex_unlock(&s_mtx_maxPktSz);
	return old;
}

#else
uint4 gse_maxPktSz(){ return gse_uMaxPktSz; }

uint4 gse_setMaxPktSz(uint4 max){
	uint4 old;
	
	if(max < 21)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "The smallest possible GSE packet "
		             "that can contain data is 21 bytes");
	 old = gse_uMaxPktSz;
	 gse_uMaxPktSz = max;
	 return old;
}

#endif

/*****************************************************************************/
bytebuf_t GsePkt_encode(GsePkt* pThis){
	bytebuf_t hdrBuf = BBUF_NULL; 
	bytebuf_t pktBuf = BBUF_NULL;
	uint4 u4PktLen;
	uint2 uCRC;
	
	/* Check to see if we have any data first */
	
	hdrBuf = pThis->pHdr->encode(pThis->pHdr, pThis->_uMinHdr,
	                             pThis->_uMaxHdr);
	except_cleanup_push(free, hdrBuf.ptr);
	
	pktBuf.len = hdrBuf.len + pThis->_data.len + 4;
	if((pktBuf.len > gse_maxPktSz())||(pktBuf.len < 21))
		gse_throw_fm(GSE_XCEPT, GSE_XPKTTOBIG, "Encoded Packet size, %d "
		             "is outside range of 21 through %d bytes.", pktBuf.len,
		             gse_maxPktSz());
	
	
	/* Copy all data into a single buffer */
	pktBuf.ptr = gse_calloc(pktBuf.len, sizeof(byte));
	except_cleanup_push(free, pktBuf.ptr);
	
	memcpy(pktBuf.ptr, hdrBuf.ptr, hdrBuf.len);
	memcpy(pktBuf.ptr + hdrBuf.len, pThis->_data.ptr, pThis->_data.len);
	
	/* Insert the total pkt length at the beginnig and end */
	gse_storeSinU(pktBuf.len, &u4PktLen, 4);
	gse_u4ToMSBbytes(pktBuf.ptr + 8,              u4PktLen);
	gse_u4ToMSBbytes(pktBuf.ptr + pktBuf.len - 4, u4PktLen);
	
	/* Generate and store the CRC */
	uCRC = gse_crc_ccitt(pktBuf.ptr + 6, pktBuf.len - 6);
	gse_u2ToMSBbytes(pktBuf.ptr + 4, uCRC);
	
	except_cleanup_pop(except_no_call); /* << only frees on an exception    */
	except_cleanup_pop(except_call);    /* << always frees the temp hdrBuf  */
	return  pktBuf;
}

bool GsePkt_testEncode(GsePkt* pThis, int nLevel){
	static const except_id_t catch_spec[] = {{XCEPT_GROUP_ANY, XCEPT_CODE_ANY}};
	except_t* pExcept;
	bool bOkay = false;
	bytebuf_t buf = BBUF_NULL;
	
	except_try_push(catch_spec, 1, &pExcept);
		
	if(pExcept == NULL){
		except_cleanup_push(free, buf.ptr);
		buf = pThis->encode(pThis);
		bOkay = true;
		except_cleanup_pop(except_call);
	}
	else{
		gse_log_except(nLevel, pExcept);
	}
		
	except_try_pop();
	return bOkay;
}

int GsePkt_addHdrBlk(GsePkt* pThis, GseBlock* pBlk)
{
	/* Any relavent checks should be in GseHdr */
	return pThis->pHdr->addBlock(pThis->pHdr, pBlk);
}

const GseBlock* GsePkt_getBlkByType(const GsePkt* pThis, byte uType, byte uProc)
{
	return pThis->pHdr->getBlkByType(pThis->pHdr, uType, uProc);
}

int4 GsePkt_blkCount(const GsePkt* pThis, byte uType, byte uProc)
{
	if(uType == 0)
		return pThis->pHdr->numBlocks(pThis->pHdr);
	else
		return pThis->pHdr->numBlksByType(pThis->pHdr, uType, uProc);
}

const GseBlock* GsePkt_getBlkByTypeIdx(const GsePkt* pThis, byte uType,
		                                 byte uProc, int4 idx)
{
	return pThis->pHdr->getBlkByTypeIdx(pThis->pHdr, uType, uProc, idx);
}
	
bool GsePkt_hasUniqueBlkByType(const struct gse_pkt* pThis, byte uType, byte uProc){
	bool bRet;
	bRet = pThis->pHdr->numBlksByType(pThis->pHdr, uType, uProc) == 1;
	return bRet;
}

void GsePkt_setHdrLimits(GsePkt* pThis, uint2 uMinHdr, uint2 uMaxHdr)
{
	if(uMinHdr > uMaxHdr)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Min. Header Length, %hu, exceeds "
		          "the max. header length, %hu.", uMinHdr, uMaxHdr);
	pThis->_uMinHdr = uMinHdr;
	pThis->_uMaxHdr = uMaxHdr;
}

void GsePkt_setData(GsePkt* pThis, bytebuf_t dataBuf)
{
	if(dataBuf.len > (gse_maxPktSz() - 20))
		gse_throw_fm(GSE_XCEPT, GSE_XPKTTOBIG, "Data buffer length, %u, plus"
				       "the smallest legal header (20 bytes) is greater than the "
		             "maximum packet size, %u.", dataBuf.len, gse_maxPktSz());
	
	if(pThis->_data.ptr != NULL) free(pThis->_data.ptr);
	
	pThis->_data = dataBuf;
}

cbytebuf_t GsePkt_getData(const GsePkt* pThis)
{
	cbytebuf_t cbuf;
	cbuf.ptr = (const byte*) pThis->_data.ptr;
	cbuf.len = pThis->_data.len;
	return cbuf;
}

/*****************************************************************************/
void delete_GsePkt(void* vpThis)
{
	GsePkt* pThis = (GsePkt*)vpThis;
	if(pThis->pHdr != NULL) pThis->pHdr->del(pThis->pHdr);
	if(pThis->_data.ptr != NULL) free(pThis->_data.ptr);
	free(pThis);
}

GsePkt* GsePkt_copy(const GsePkt* pThis, size_t uOver)
{
	GsePkt* pNew = NULL;
	
	if(uOver > sizeof(GsePkt))
		pNew = (GsePkt*) gse_calloc(1, uOver);
	else
		pNew = (GsePkt*) gse_calloc(1, sizeof(GsePkt));
	
	/* Get most things using memcpy */
	memcpy(pNew, pThis, sizeof(GsePkt));
	
	/* Deep copy the header pointer */
	pNew->pHdr = pThis->pHdr->copy(pThis->pHdr, 0);
	
	/* Deep copy the data */
	pNew->_data.ptr = (byte*)gse_calloc(pThis->_data.len, sizeof(byte));
	memcpy(pNew->_data.ptr, pThis->_data.ptr, pThis->_data.len);
	
	return pNew;
}

GsePkt* new_GsePkt(cbytebuf_t buf, gse_blk_factory factory, size_t uOverMalloc)
{
	uint4 uTailLen;
	uint2 uCRC;
	GsePkt* pThis = NULL;
	size_t nLen, nSize = 0;
	
	/* Make one function handle both jobs */
	if(buf.len == 0 && factory == NULL) 
		return new_empty_GsePkt(0);
	 
   if(uOverMalloc > sizeof(GsePkt))
		nSize = uOverMalloc;
	else
		nSize = sizeof(GsePkt);
	
	pThis = (GsePkt*)gse_calloc(1, nSize);
	except_cleanup_push(delete_GsePkt, pThis);
	
	pThis->del = delete_GsePkt;
	pThis->copy = GsePkt_copy;
	pThis->encode = GsePkt_encode;
	pThis->testEncode = GsePkt_testEncode;
	pThis->setData = GsePkt_setData;
	pThis->getData = GsePkt_getData;
	pThis->addHdrBlk = GsePkt_addHdrBlk;
	pThis->getBlkByType = GsePkt_getBlkByType;
	pThis->setHdrLimits = GsePkt_setHdrLimits;
	pThis->hasUniqueBlkByType = GsePkt_hasUniqueBlkByType;
	pThis->blkCount = GsePkt_blkCount;
	pThis->getBlkByTypeIdx = GsePkt_getBlkByTypeIdx;
	
	pThis->_uMinHdr = 16;
	pThis->_uMaxHdr = 0xFFFF;

	/* Need at least 22 bytes to do anything */
	if(buf.len < 22)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "An absolute minimum of 22 bytes "
		             "are needed to make a GsePkt, recieved %d.", buf.len);
	
	
	/* First decode the header bytes */
	pThis->pHdr = new_GseHdr(buf, factory, 0);
	
	
	/* Make sure that buf.ptr is long enough to point to the whole packet */
	if(buf.len < pThis->pHdr->uTotalLen)
		gse_throw_m(GSE_XCEPT, GSE_XDECODE, "Encoded packet length is longer "
				       "than then the data recieved.");

	
	/* Check the CRC, this covers byte 6 to the end */
	gse_storeUinS(pThis->pHdr->uTotalLen - 6, &nLen, sizeof(size_t));
	uCRC = gse_crc_ccitt(buf.ptr + 6, nLen);
	
	if(uCRC != pThis->pHdr->uCRC){
		gse_error_v("GSE Header CRC mismatch, File CRC: 0x%02hX, Calc CRC: "
		            "0x%02hX", pThis->pHdr->uCRC, uCRC);
		/* pThis->addMsg(error, GSE_BAD_CRC, NULL); */
	}
	
	
	/* Store the data, if any */
	pThis->_data.len = pThis->pHdr->uTotalLen - pThis->pHdr->uHeadTailLen;
	gse_storeUinS(pThis->_data.len, &nLen, sizeof(nLen));
	
	pThis->_data.ptr = (byte*)gse_calloc(nLen, sizeof(byte));
	memcpy(pThis->_data.ptr, buf.ptr + pThis->pHdr->uHeadTailLen - 4, nLen);
	
	
	/* Check the trailing length */
	uTailLen = gse_u4FromMSBbytes(buf.ptr + pThis->pHdr->uTotalLen - 4);
	if(uTailLen != pThis->pHdr->uTotalLen)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Trialing length value, %u is in not"
		             " equal to the initial length value, %u", uTailLen,
		             pThis->pHdr->uTotalLen);
 	
	except_cleanup_pop(except_no_call);
	return pThis;
}


GsePkt* new_empty_GsePkt(size_t uOverMalloc)
{
	GsePkt* pThis = NULL;
	size_t uSize = 0;
	 
   if(uOverMalloc > sizeof(GsePkt))
		uSize = uOverMalloc;
	else
		uSize = sizeof(GsePkt);
	
	pThis = (GsePkt*)gse_calloc(1, uSize);
	except_cleanup_push(delete_GsePkt, pThis);
	
	pThis->copy = GsePkt_copy;
	pThis->del = delete_GsePkt;
	pThis->encode = GsePkt_encode;
	pThis->testEncode = GsePkt_testEncode;
	pThis->setData = GsePkt_setData;
	pThis->getData = GsePkt_getData;
	pThis->addHdrBlk = GsePkt_addHdrBlk;
	pThis->getBlkByType = GsePkt_getBlkByType;
	pThis->setHdrLimits = GsePkt_setHdrLimits;
	pThis->hasUniqueBlkByType = GsePkt_hasUniqueBlkByType;
	pThis->blkCount = GsePkt_blkCount;
	pThis->getBlkByTypeIdx = GsePkt_getBlkByTypeIdx;
	
	pThis->_uMinHdr = 16;
	pThis->_uMaxHdr = 0xFFFF;
	
	pThis->pHdr = new_empty_GseHdr(0);
	
	except_cleanup_pop(except_no_call);
	return pThis;
}


/*****************************************************************************/
/* Need explicit instanitaiton of the GsePktPtrAry macro template,           */
/* In this case when a pointer is removed from the array make sure that      */
/* the del member function of the structure in called                        */

#define C GsePktAry
#define T GsePkt
#define CASCADE_DEL

#include "ptrAryTplt.c"

#undef CASCADE_DEL
#undef C
#undef T
