/** \file GsePkt.h General Packet Structures and Functions */
#ifndef _GsePkt_H_
#define _GsePkt_H_

#include <rpwgse/GseHdr.h>
#include <rpwgse/GseConv.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_packet Packets
 *
 * Ground Support Packet Framework
 */
/*@{*/

#ifndef _GsePkt_C_
/* The sync pattern at the front of GSE packets */
extern const byte* gse_sSyncPtrn;

/* Location and definition of the CRC field in a GSE packets */
extern const gse_field_def GSE_FD_CRC;

/* Location and definition of the Head Tail Length field in a GSE packets */
extern const gse_field_def GSE_FD_HT_LEN;

/* Location and definition of the Total Packet Length field in GSE packets */
extern const gse_field_def GSE_FD_PKT_LEN;

/* Location and definition of the Application ID field in GSE packets */
extern const gse_field_def GSE_FD_APP_ID;

/* Location and definition of the Header Version field in GSE packets */
extern const gse_field_def GSE_FD_VERSION;

/* Location and definition of the Contents field in GSE packets */
extern const gse_field_def GSE_FD_CONTENTS;
	
#endif /* _GsePkt_C_ */

/** Error number for Bad CRC */
#define GSE_BAD_CRC 0x1

/* The message classification types */
enum gse_msg_type {error, warning, notice};
typedef enum gse_msg_type gse_msg_type;


typedef struct gse_msg{
	gse_msg_type type;
	byte         code;
	byte         prog_id;
	byte         proc_id;
	char*        sMsg;
	void (*del)(void* vpThis);
} GseMsg;	


/** GSE Packet
 * The overall container for one Packet's headers and data.  On disk and in
 * memory a packet consists of:
 *
 * -# A General Ground Support Packet Header.
 * -# Instrument and Mission specific Headers.
 * -# Instrument data.
 * -# A copy of the total packet length.
 *
 * See @ref page_spec_4_2 for an overview of GSE packets.
 * @n
 * @class GsePkt
 */
typedef struct gse_pkt{

	/** The header structure */
	GseHdr* pHdr;
	
	/** Copy constructor 
	 *
	 * Generates a new GsePkt on the heap as a deep copy of this one. 
	 *
	 * @param pThis The GsePkt to copy
	 * @param uOver Usually set this to 0, it's for derived classes
	 * @return A pointer to a GsePkt structure.
	 */
	struct gse_pkt* (*copy)(const struct gse_pkt* pThis, size_t uOver);
	
	/* Read in data from a byte stream */
	/* void (*decode)(cbytebuf_t buf, gse_blk_factory factory); */
	
	/** Add an error, warning or notice for this packet 
	 *
	 * @return the new number of messages for the packet 
	 */
	/*int (*addMsg)(gse_msg_type type, byte code, const char* sMsg);*/
	
	/** Get all the data as a chunk of bytes, this makes a copy.
	 *
	 * @param pThis, a GsePkt pointer.
	 *
	 * @return A byte buffer whose @b ptr member is suitable for use in free.
	 * 
	 * @throws (GSE_XCEPT, GSE_XPKTTOBIG) if total packet size exceeds
	 *          gse_maxPktSz().
	 *
	 * @throws (GSE_XCEPT, GSE_XENCODE) if GseHdr::encode() fails.
	 *
	 * @throws (CLIB_XCEPT, ENOMEM) if memory cannot be allocated.
	 *
 	 * @see GseHdr::encode(struct gse_hdr* pThis, uint2 uMinLen, uint2 uMaxLen).
	 */
	bytebuf_t (*encode)(struct gse_pkt* pThis);
	
	
	/** Test encoding.
	 *
	 * Encode the bytes using @a encode but throw away the output buffer.
	 *
	 * @param pThis, a GsePkt pointer.
	 * @param nLevel, the log level for any caught exceptions. See 
	 *        @ref mod_log for more information on log levels.
	 *
	 * @return @c true if encoding succeeds for this packet, @c false 
	 *         if it fails.
	 *
	 * @throws nothing  This function internally catches all exceptions.
	 */
	bool (*testEncode)(struct gse_pkt* pThis, int nLevel);
	
	
	/** Retrieve a header block based on it's type and process IDs.
	 * 
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block requested.
	 * @param uProc The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        proceessing step ID.  For Non-process blocks use 0.
	 * @return A pointer to GseBlock derived structure corresponding to this
	 *         block type.
	 * @throws (GSE_XCEPT, GSE_XNOTUNIQUE) if there is not one and only one
	 *         block of this type in the header
	 * @ref page_overview for a description of different block structures are
	 *      generated based on type and process IDs.  
	 * @public
	 */
	const GseBlock* (*getBlkByType)(const struct gse_pkt* pThis, byte uType,
	                                byte uProc);
											  
	/** Retrieve a header block based on it's type, process ID and index
	 * 
	 * This function is similar to getBlkByType, except that it can be used
	 * in situations where more than 1 of a particular block type may be
	 * present.
	 *
	 * @throws (GSE_XCEPT, GSE_XBADARG) if the specified block type doesn't 
	 *         exist.
	 *
	 * @public
	 */
	const GseBlock* (*getBlkByTypeIdx)(const struct gse_pkt* pThis, byte uType,
	                                byte uProc, int4 idx);

	/** Determine the number of header blocks of a particular type in a header
	 *
	 * @param  pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block to search for.  Since 0 is not a legal
	 *        block type, use 0 to get a count of all the blocks in the header.
	 * @param uProc The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        processing step ID.  For Non-process blocks use 0.
	 * @return The count of blocks of the specified type. 
	 * 
	 * @public
	 */
	int4 (*blkCount)(const struct gse_pkt* pThis, byte uType, byte uProc);


	/** Determine if there is a unique header block based on it's type and process IDs.
	 * 
	 * @param pThis the pointer to the GseHdr that contains this function.
	 * @param uType The type of block to search for.
	 * @param uProc The processing step ID of the block requested.  Processing
	 *        blocks all have the same type, but are distinguished by thier
	 *        processing step ID.  For Non-process blocks use 0.
	 * @return True if the block uniquely exists in the packet headers, false otherwise.
	 * @public
	 */
	bool (*hasUniqueBlkByType)(const struct gse_pkt* pThis, byte uType, byte uProc);

	/** Add an information block to the header. 
	 *
	 * Use this along with #addHdrBlk to generate packets instead of reading
	 * them in.
	 */
	int (*addHdrBlk)(struct gse_pkt* pThis, GseBlock* pBlk);
	
	/** Set but <b>do not copy</b> the data bytes for this packet.
	 *
	 * The old data section memory, if any is free'ed.
	 *
	 * @note The size in the header section is not re-calculated when this
	 * function is called.  Even if this call succeeds for a particular size
	 * data buffer, later use of GsePkt::encode may fail if the header plus
	 * data size exceeds gse_maxPktSz().
	 * 
	 * @param dataBuf a buffer containing 0 to gse_maxPktSz() - 16 bytes.
	 * The packet takes ownership of this buffer so don't delete it elsewhere. 
	 *
	 * @throws (GSE_XCEPT, GSE_XPKTTOBIG) is thrown if dataBuf is more
	 * than 4 Gigabytes.
	 */
	void (*setData)(struct gse_pkt* pThis, bytebuf_t dataBuf);
	
	/** Get the data bytes for this packet 
	 * @param pThis the pointer to the GsePkt that contains this function.
	 * @return a constant byte buffer pointing at the data bytes.
	 */
	cbytebuf_t (*getData)(const struct gse_pkt* pThis);
	
	/** Set the upper and lower bounds for the header length.
	 *
	 * @param pThis a GsePkt structure pointer.
	 *
	 * @param uMinHdr The minimum length of the encoded header.  If less
	 *    than uMinHdr bytes are required for the header data, the remainder
	 *    of the header will be zero padded.  Set to 0 to have no minimum 
	 *    header length.
	 *
	 * @param uMaxHdr the maximum length header allowed.  The maximum header
	 *    length is 2^16 - 1 or 65535
	 */
	void (*setHdrLimits)(struct gse_pkt* pThis, uint2 uMinHdr, uint2 uMaxHdr);
	
	/** Destructor */
	void (*del)(void* vpThis);

	uint2 _uMinHdr;
	uint2 _uMaxHdr;
	
	/* Raw data byte buffer, use getDataBytes instead */
	bytebuf_t _data;

} GsePkt;

/** Generate a new GsePkt, with plugin header block interpertation.
 * @public
 * @memberof GsePkt
 */
GsePkt* new_GsePkt(cbytebuf_t buf, gse_blk_factory factory, size_t uOverMalloc);

/** Generate a new empty GsePkt.
 * Headers and data must be filled in manually when using this constructor.
 * @public
 * @memberof GsePkt
 */
GsePkt* new_empty_GsePkt(size_t uOverMalloc);

/** Set maximum allowed packet size for this particular application.
 *
 * By the files format definitions, the maximum allowed packet size
 * is 4 Gigabytes, (0xFFFFFFFF).  This is fantastically large for most
 * applications, thus the max allowed size may be lowered thread wide
 * via this function.  Any library routine that is asked to read or
 * write a packet larger than the size set here will throw 
 * (GSE_XCEPT, GSE_XPKTTOBIG).
 *  
 * @param max Can be any value up to 4 Gigabytes, (0xFFFFFFFF).
 *
 * @return the old max packet size for this thread.
 */  
uint4 gse_setMaxPktSz(uint4 max);

/** Get the maximum allowed size (header plus data) for a GsePkt in this
 * thread. 
 */
uint4 gse_maxPktSz(void);


/* An array of pointers to GseBlock structures */
#define S gse_pkt_ary
#define C GsePktAry
#define T GsePkt

/* Note, the included file has doxygen documentation that is automatically
   extracted via doc_filter.py at the top level */
#include "ptrAryTplt.h"

#undef S
#undef C
#undef T

/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GsePkt_H_ */

