/** @file GsePktFilter.h Interface description for GsePktFilter */
#ifndef _GsePktFilter_H_
#define _GsePktFilter_H_

#include <rpwgse/GsePktSrc.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_processing Processing */
/*@{*/

/** Type definition for callback to catch packets that are not final. 
 * Functions of this type are expected to take memory ownership of passed
 * in packets.
 *
 * @param pCtx a user defined object passed to GsePktFilter::setCatcher,
 *        may be NULL.
 * @param pPkt the erroneous packet, not NULL.
 * @param sMsg the reason the packet was considered an error, may be NULL.
 * 
 * @return @b true if processing should continue, @b false if it should stop.
 *
 */
typedef bool (*gse_errpkt_hndlr_t)(void* pCtx, GsePkt* pPkt, char* sMsg);

/** Basic functionality for single input packet filters.
 *
 * Most packet filters need to maitain a link to an upstream packet source.
 * Most of the time queries for source name and position information are
 * passed back to the source.  This class handles that functionality so
 * that the code need not be repeated for most common single input packet
 * filters.  
 *
 * This is a partially abstract base class.  A constructor and destructor are
 * provided, they do not fill in the critical @a readNext function pointer.
 * @n
 * @class GsePktFilter
 * @extends GsePktSrc
 */
typedef struct gse_pkt_filter{
	
	/** A filter is a type of packet source */
	GsePktSrc base;
	
	/** A pointer to the upstream packet source */
	GsePktSrc* pUpSrc;
	
	/** The ID of the program running this filter */
	byte uProgID;
	
	
	/** Set a Callback to receive erroneous input.
	 *
	 * Normally if a packet can't be processed, an error message is
	 * emmitted and an exception is thrown.
	 *
	 * Registering a callback with this function will allow the input packet
	 * that triggered the error to be retained and for processing to continue.
	 *
	 * If the catcher callback returns false then processing is halted, if it
	 * returns true processing continues.  In either case the catcher is
	 * @b still responsible for freeing the packet's memory.
	 *
	 * @param pThis A pointer to the GsePktFilter
	 * @param pFunc The callback function.
	 * @param pCatchCtx A pointer to a structure that will be passed back to
	 *        the setCatcher caller, this may be null
	 * 
	 */
	void (*setCatcher)(struct gse_pkt_filter* pThis, gse_errpkt_hndlr_t pFunc, 
	                   void* pCatchCtx);
	
	/** Get the current catcher callback function for handling error packets.
	 *
	 * If GsePktFilter::setCatcher has not been called, returns NULL.
	 */
	gse_errpkt_hndlr_t (*getCatcher)(const struct gse_pkt_filter* pThis);
	
	/** Get the current context object that will be passed to any error packet
	 * catchers.
	 * 
	 * If GsePktFilter::setCatcher has not been called, returns NULL.
	 */
	void* (*getCatchCtx)(const struct gse_pkt_filter* pThis);
	
	
	/** Registered error packet callback. 
	 * @protected
	 */
	gse_errpkt_hndlr_t pCatcher;
	
	/** Registered error packet callback context (user data)
	 * @protected
	 */
	void* pCatchCtx;
	
	/** Virtual destructor.
	 * This is needed as calls using pointers cast as a GsePktFilter should be
	 * supported
	 */
	void (*del)(void* vpThis);

} GsePktFilter;


/** Convienence constructor for GsePktFilter derived classes.  
 *
 * This function is not ment to be used directly in application programs,
 * rather it should be wrapped by the derived class constructor. 
 * 
 * @param pSrc A GsePktSrc implementing the readNext function.
 *
 * @param uProgID The program ID to attach to any generated packets.
 *
 * @param uMalloc Since this is a base class, provide the actual amount of
 *        memory to calloc here.
 *
 * @param del The destructor function for the derived class.  This is copied
 *        to GsePktFilter::del and to GsePktSrc::del.
 *
 * @param readNext The readNext function for the derived class.
 * 
 * @protected
 * @memberof GsePktFilter
 */
GsePktFilter* new_GsePktFilter(GsePktSrc* pSrc, byte uProgID, size_t uMalloc,
                          GsePkt* (*readNext)(GsePktSrc*), void (*del)(void*));

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GsePktFilter_H_ */
