/** @file GsePktSrc.h Interface description for packet generators. */
#ifndef _GsePktSrc_H_
#define _GsePktSrc_H_

#include <rpwgse/GsePkt.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_processing Processing
 *
 * Packet Processing Framework
 */
/*@}*/

/** @addtogroup mod_processing Processing */
/*@{*/

/** Packet source interface, all things that support generating packets
 * needs to impliement this.
 * @n
 * @class GsePktSrc
 */
typedef struct gse_pkt_src{
	
	/** The packet offset of the source */
	int4 (*pktIdx)(struct gse_pkt_src* pThis);
	
	/** The number of packets in the source if known.
	 *
	 * Note, this is usually only known if the last packet has been read from
	 * the source and the source has a way of knowing that I cannot produce any
	 * new packets.
	 *
	 * @return The number of distict packets returnable by this source if known,
	 *  -1 if not known.
	 */
	int4 (*numPkts)(struct gse_pkt_src* pThis);
				
	/** The current byte offset of the source */
	size_t (*offset)(struct gse_pkt_src* pThis);
	
	/** Return the name of the packet source.  For a file this could be
	 * the filename. */
	const char* (*name)(struct gse_pkt_src* pThis);
	
	
#ifdef GSE_POSIX_THREADS
	/** <b>(Multi-thread only)</b> Interrupt the packet source, this causes any functions in this source
	 * that are blocked waiting to return. 
	 *
	 * @note This function only exists in @b multi-threaded versions of the 
	 *       library.
	 * 
	 * Obviously this in only useful if the application has more than one
	 * thread.  Does nothing if the source is not blocked waiting on data.
	 *
	 * Note: MT Safe but not Async Signal safe */
	void (*interrupt)(struct gse_pkt_src* pThis, pthread_t threadID);
#endif

	/* Toggle reading after a signal.
	 *
	 * Sets the "continue reading after an OS signal" state.  Though this 
	 * function cannot be used to immediatly interrupt a packet source, it
	 * can be used Signal handlers, unlike interrupt(). 
	 *
	 * @param  pThis pointer to the structure containing this function.
	 *
	 * @param bContinue If @c true, the packet source will restart any reads
	 *        that are interrupted by a signal.  If @c false, the packet source
	 *        will not restart interrupted reads.  
	 *
	 * @return The provious continue state.
	 */
	 /* void (*continueReads)(struct gse_pkt_src* pThis, bool bCont); */
	
	/** Produce the next packet from the source.
	 * 
	 * The function pointed to may or may not block depending on the details
	 * of the specific packet source.  In the multi-threaded version of the
	 * library the GsePktSrc::close function may to stop a blocked read.
	 *
    * This function only throws in the CLIB_XCEPT group.  The input contents
	 * or lack there of shall not cause an exception.
	 *
	 * @param pThis pointer to the structure containing this function
	 *
	 * @return The packet pointer or NULL if no more packets could be gleaned
	 *         from the source.
	 *
	 * @throws None.  All exceptions are converted to error messages.
    */
	GsePkt* (*readNext)(struct gse_pkt_src* pThis);
	
	/** Does the prevPacket function work for this source? */
	bool (*supportsPrev)(const struct gse_pkt_src* pThis);
	
	/** Does the rewind function work for this source? */
	bool (*supportsRewind)(const struct gse_pkt_src* pThis);
	
	/** Get the previous packet in this source, if supported.
	 * @return The previous or null if the current packet in the 1st one */
	GsePkt* (*readPrev)(struct gse_pkt_src* pThis);
	
	/** Can skip be given a negative argument for nSkip ? */
	bool (*supportsSkipAhead)(const struct gse_pkt_src* pThis);
	
	/** Can skip be given a positive argument for nSkip ? */
	bool (*supportsSkipBack)(const struct gse_pkt_src* pThis);
	
	/** Skip over packets in the input stream 
	 * Skips over packets *without* checking the CRC!  Hitting the end of the
	 * file does not block, it just causes a short skip count to be returned.
	 *
	 * @param pThis the GsePktSrc derived object.
	 *
	 * @param nSkip the number of packets to skip.  Positive numbers cause
	 *        a forward skip, negative numbers a backwards skip.  A particular
	 *        source may not support back skips.
	 *
	 * @returns The number of packets actually skipped, 0 indicates no change.
	 */
	int (*skip)(struct gse_pkt_src* pThis, int nSkip);
	
	/** Skips past the last complete packet in the source.
	 * This is supported if supportsSkipAhead returns true.
	 * 
	 * return the number of parkets actually skipped.
	 */
	int (*skipToEnd)(struct gse_pkt_src* pThis);
	
	/** Return to the beginning of the packet source */
	void (*rewind)(struct gse_pkt_src* pThis);
	
	/** Set the max legal packet size */
	uint4 (*setMaxPktSz)(struct gse_pkt_src* pThis, uint4 uMax);
	
	/** Get the max legal packet size */
	uint4 (*getMaxPktSz)(const struct gse_pkt_src* pThis);
	
	/** Destructor function pointer:
	 * @warning Check to see if this is NULL.  Classes supporting multiple 
	 * interfaces will not implement the destructor at the interface level. */
	void (*del)(void* vpThis);
	
	/* Owner Pointer, useful to classes that may support multiple interfaces */
	void* _pIfaceOwner;
	
	/* Maximum legal packet size, used to avoid jumping way across the file
	   when garbled data is enconteret */
	uint4 uMaxPktSz;
	
} GsePktSrc;

/** Constructor for use by derived classes.
 *
 * @memberof GsePktSrc
 * @protected
 */
GsePktSrc* new_GsePktSrc(size_t uMalloc);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GsePktSrc_H_ */
