#define _POSIX_C_SOURCE 200112

#include <pthread.h>
#include <stdlib.h>

#ifdef USE_CUPS
#include <cups/cups.h>
#endif


#include "GseExcept.h"
#include "GseWrap.h"
#include "GsePrnSvc.h"
#include "GseLog.h"
#include "GseSys.h"

/*****************************************************************************/
/* The one, the only the default printer setting */
static char* gse_sDefPrinter = NULL;

/* Use the mutex variable to when accessing gse_sDefPrinter */
#ifdef GSE_POSIX_THREADS /* Is Threaded */
static pthread_mutex_t s_mtx_prn = PTHREAD_MUTEX_INITIALIZER;
#endif


/*****************************************************************************/
/* Setting a new default printer */

bool gse_set_def_printer(const char* sDefPrn)
{
	int i;
	bool bRet = false;
	GseStrAry* pPrinters = gse_get_printers();
	
	for(i = 0; i< pPrinters->num; i++){
		if( strcmp(sDefPrn, pPrinters->pAry[i]) == 0){
			
#ifdef GSE_POSIX_THREADS
			pthread_mutex_lock(&s_mtx_prn);
#endif
			free(gse_sDefPrinter);
			gse_sDefPrinter = CS(sDefPrn);
			
#ifdef GSE_POSIX_THREADS
			pthread_mutex_unlock(&s_mtx_prn);
#endif			
			bRet = true;
		}
	}
	
	pPrinters->del(pPrinters);
	return bRet;
}

/****************************************************************************/
/* Picking the a default printer 
 * If the env var PRINTER is defined use that, then use the CUPS/Windows
 * default, then just use "FILE".
 */

void _gse_reset_def_printer(bool bLockMutex){
	GseStrAry* pPrinters = NULL;
	int i;

#ifdef GSE_POSIX_THREADS
	if(bLockMutex) pthread_mutex_lock(&s_mtx_prn);
#endif

	pPrinters = gse_get_printers();
	
	if(gse_sDefPrinter != NULL) free(gse_sDefPrinter);
	gse_sDefPrinter = NULL;
	
		
	if(getenv("PRINTER") != NULL){
		/* Use this if it is a real printer */
		for(i = 0; i < pPrinters->num; i++){
			if( strcmp(pPrinters->pAry[i], getenv("PRINTER") ) == 0){
				gse_sDefPrinter = CS( getenv("PRINTER") );
				break;
			}
		}
	}
		
#ifdef USE_CUPS
	/* Still NULL, try the cups default */
	if(gse_sDefPrinter == NULL){ 
			
		cups_dest_t* pDests;
		int nDests = cupsGetDests(&pDests);
	
		for(i = 0; i<nDests; i++){
				
			if(pDests[i].is_default){
				gse_sDefPrinter = CS( pDests[i].name );
				break;
			}
		}
		cupsFreeDests(nDests, pDests);
	}
#endif
		
	/* Fallback, just use "FILE" */
	gse_sDefPrinter = CS("FILE");
		
#ifdef GSE_POSIX_THREADS
	if(bLockMutex) pthread_mutex_unlock(&s_mtx_prn);
#endif			

	pPrinters->del(pPrinters);				
}

void gse_reset_def_printer()
{
	_gse_reset_def_printer(true);
	
}

/*****************************************************************************/
/* Returns the default printer, set's it if it's not defined */
char* gse_get_def_printer()
{ 
	char* sRet;
#ifdef GSE_POSIX_THREADS
	pthread_mutex_lock(&s_mtx_prn);
#endif
	
	/* 1st Try to set the default printer using the env var */
	if(gse_sDefPrinter == NULL)
		_gse_reset_def_printer(false);	
	
	sRet = CS(gse_sDefPrinter);
	
#ifdef GSE_POSIX_THREADS
	pthread_mutex_unlock(&s_mtx_prn);
#endif	
	return sRet;	
}

/****************************************************************************/
/* Getting the list of printers */

GseStrAry* gse_get_printers()
{
	GseStrAry* pPrinters = NULL;

#ifdef WIN32
#error gse_get_printers not yet implemented for Windows systems
#else 

#ifdef USE_CUPS
	cups_dest_t* pDests;
	int i;
	int nDests = cupsGetDests(&pDests);
	
	pPrinters = new_GseStrAry(nDests + 1);
	
	pPrinters->pAry[0] = CS("FILE");
	
	for(i = 0; i<nDests; i++) 
		pPrinters->pAry[i+1] = CS(pDests[i].name);
	
	cupsFreeDests(nDests, pDests);
	
#else /* General POSIX printer getting here, not sure what to do */
	
#error gse_get_printers not yet implemented for non-cups unix environments
	
#endif /* CUPS */
#endif /* windows */

	return pPrinters;
}

/****************************************************************************/
/* Checking to see if this is the default printer */

bool gse_is_default_prn(const char* sPrinter)
{
	bool bRet = false;
#ifdef GSE_POSIX_THREADS
	pthread_mutex_lock(&s_mtx_prn);
#endif
	
	if(gse_sDefPrinter == NULL)
		_gse_reset_def_printer(false);
	
	if( strcmp(sPrinter, gse_sDefPrinter) == 0) bRet = true;
	
#ifdef GSE_POSIX_THREADS
	pthread_mutex_unlock(&s_mtx_prn);
#endif
	return bRet;
}


/*****************************************************************************/
/* Sending a file to a printer, in all it's encarnations */

#ifdef WIN32
#error GsePrnJob_sendJob not yet implemented for Windows systems
#else 


#ifdef USE_CUPS
int GsePrnJob_sendJob(GsePrnJob* pThis)
{	
	if(strcmp(pThis->sPrinter, "FILE") == 0){
		gse_debug("GsePrnJob::sendJob does nothing when printing to a file\n");
		pThis->nJobID = -1;
	}
	else{
	
		/*TODO: make a nice title for the printout 
		sTitle = NS("%s @ %s from %s", pThis->sFile */ 
		pThis->nJobID = cupsPrintFile(pThis->sPrinter, pThis->sFile, pThis->sFile,
			                        0, NULL);
	}
		
	pThis->setTitle(pThis);
	return pThis->nJobID;
}

#else /* Generic POSIX system version */
int GsePrnJob_sendJob(GsePrnJob* pThis)
{
	/* This is cheap, can't get job ID without a pipe but it does something */
	char sCmd;
	int nRet;
	
	if(strcmp(pThis->sPrinter, "FILE") == 0){
		gse_debug("GsePrnJob::sendJob does nothing when printing to a file\n");
		pThis->nJobID = -1;
	}
	else{
		sCmd = NS("lp -d %s %s", pThis->sPrinter, pThis->sFile);
	
	
		gse_info_v("Starting print job via: %s", sCmd);
	
		if( system(sCmd) != 0) pThis->nJobID = 0;
		else pThis->nJobID = 1;
	
		gse_warning("TODO: Read output of lp, real print job ID is not available, using 1");
	}
	
	pThis->setTitle(pThis);
	return pThis->nJobID;
}

#endif /* CUPS */
#endif /* windows */

/*****************************************************************************/
const char* GsePrnJob_useTempFile(GsePrnJob* pThis, const char* sSuffix)
{
	pThis->sFile = gse_tempname(sSuffix);
	pThis->bTmp = true;
	return (const char*)pThis->sFile;
}

void GsePrnJob_setFile(GsePrnJob* pThis, const char* sPath, bool bTmp)
{
	if(pThis->sFile) free(pThis->sFile);
	pThis->sFile = CS(sPath);
	pThis->bTmp = bTmp;
}

const char* GsePrnJob_getFile(const GsePrnJob* pThis)
{
	return (const char*)(pThis->sFile);
}

/*****************************************************************************/
/* Helper, set the printer to use */

bool GsePrnJob_setPrinter(GsePrnJob* pThis, const char* sPrinter)
{
	GseStrAry* pPrinters;
	int i;
		
	if(!sPrinter){
		free(pThis->sPrinter);
		pThis->sPrinter = gse_get_def_printer();
		return true;
	}
	
	pPrinters = gse_get_printers();
	for(i=0; i< pPrinters->num; i++){
		
		if(strcmp(sPrinter, pPrinters->pAry[i]) == 0){
			free(pThis->sPrinter);
			pThis->sPrinter = CS(pPrinters->pAry[i]);
			
			pPrinters->del(pPrinters);
			return true;
		}	
	}
	
	return false;
}

const char* GsePrnJob_getPrinter(const GsePrnJob* pThis)
{
	return (const char*)(pThis->sPrinter);
}

/*****************************************************************************/
/* Helper, set the current job title based on the object state               */

void GsePrnJob_setTitle(GsePrnJob* pThis)
{	
	char sBuf[256];
	char* sAbsPath = NULL;
	bool bFreeAbs = false;
	char* sPre;
	
	sBuf[255] = '\0';
	
	if((pThis->sFile != NULL)&&(pThis->sFile[0] != '/')&&(getcwd(sBuf, 255))){
		sAbsPath = NS("%s/%s", sBuf, pThis->sFile);
		bFreeAbs = true;
	}
	else{
		sAbsPath = pThis->sFile;
	}
	
	gse_assert(pThis->sPrinter != NULL);
	if(pThis->sTitle != NULL) free(pThis->sTitle);
	
	if(strcmp(pThis->sPrinter, "FILE") == 0){
		if(sAbsPath) sPre = NS("file %s", sAbsPath);
		else sPre = CS("unknown file");
	}
	else{
		if(sAbsPath)
			sPre = NS("file %s on %s", pThis->sFile, pThis->sPrinter);
		else 
			sPre = CS(pThis->sPrinter);
	}
	
	if(pThis->nJobID != 0){
		pThis->sTitle = NS("%s [job id: %d]", sPre, pThis->nJobID);
		free(sPre);
	}
	else
		pThis->sTitle = sPre;
	
	if(bFreeAbs) free(sAbsPath);
}

char* GsePrnJob_getTitle(const GsePrnJob* pThis)
{
	return CS(pThis->sTitle);
}

/*****************************************************************************/
/* Destructor */

void delete_GsePrnJob(void* vpThis)
{
	GsePrnJob* pThis = (GsePrnJob*)vpThis;
	
	free(pThis->sTitle);
	free(pThis->sPrinter);
	
	if((pThis->bTmp)&&(pThis->sFile)){
		unlink(pThis->sFile);
	}
	
	free(pThis->sFile);
	free(vpThis);
}
	
/*****************************************************************************/
/* Print job object */

GsePrnJob* new_GsePrnJob(const char* sFile, const char* sPrinter, size_t uOver)
{
	GsePrnJob* pThis;
	
	if(uOver > sizeof(GsePrnJob))
		pThis = (GsePrnJob*)gse_calloc(1 , uOver);
	else
		pThis = (GsePrnJob*)gse_calloc(1 , sizeof(GsePrnJob));
	
	pThis->del = delete_GsePrnJob;
	except_cleanup_push(pThis->del, pThis);
	
	pThis->nJobID = 0;
	if(sFile != NULL) pThis->sFile  = CS(sFile);
	pThis->bTmp   = false;
		
	pThis->sendJob     = GsePrnJob_sendJob;
	pThis->setPrinter  = GsePrnJob_setPrinter;
	pThis->getPrinter  = GsePrnJob_getPrinter;
	pThis->getFile     = GsePrnJob_getFile;
	pThis->setFile     = GsePrnJob_setFile;
	pThis->useTempFile = GsePrnJob_useTempFile;
	pThis->getTitle    = GsePrnJob_getTitle;
	pThis->setTitle    = GsePrnJob_setTitle;
	
	pThis->setPrinter(pThis, sPrinter); /* If null, picks the default printer */
	pThis->setTitle(pThis);
	
	except_cleanup_pop(except_no_call);
	return pThis;
}


