/** @file GsePrnSvc.h Helpers for printing hardcopy files */
#ifndef _GsePrnSvc_H_
#define _GsePrnSvc_H_

#include <rpwgse/GseString.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_sys */
/*@{*/


/** Get the default printer for the library. 
 * @return a string allocated on the heap wit the name of the 
 *         current default printer.  The setting may not be current
 *         if gse_set_def_printer or gse_reset_def_printer is called.*/
char* gse_get_def_printer();

/** Set the default printer for the library.
 *
 * @note This is a library wide global.  Change with care if you are a
 * multi-threaded user of the library.  
 *
 * @param sDefPrn The new default printer, set this to NULL to indicate
 * printing only to files.
 *
 * @returns @a true if the new setting is valid, @a false if the new
 *          setting was ignored.
 */
bool gse_set_def_printer(const char* sDefPrn);

/** Reset to the system default printer, whatever the libary thinks that
 * is.
 */
void gse_reset_def_printer();


/* Get a list of the printing destinations by name.
 * Note, only the destination "Print to file" is guarunteed to exist.  And
 * it's always index 0 in the list.
 */
GseStrAry* gse_get_printers();


/** Return true if sPrinter in the current default, false otherwise */
bool gse_is_default_prn(const char* sPrinter);


/** Print job work item.
 *
 * The mentality of this work item is that some work creator fills
 * out what they want done and the handler makes it happen.  Thus
 * the output file may not exist when the handler recieves the 
 * work order, but the printer should exist. 
 *
 * Note the printing services can be compiled against more than
 * one backend, the best supported one is CUPS on Linux.
 *
 * There is no higher level applicaiton information in the class,
 * unless the reciever of this structure just knows what to do
 * you will need to subclass inorder to make this useful.
 * @n @n
 */
typedef struct gse_prn_job{	
	
	/** Set the filename, may reference a non existant one that the job
	  * handler is supposed create.
	  * 
	  * @param bTmp if true, the file is temporary an the destructor
	  *        should clean it up.
	  */
	void (*setFile)(struct gse_prn_job* pThis, const char* sPath, bool bTmp);
	
	/** Set the internal filename to be a temporary name.
	 * The print job object makes up a name and storage location in a manner
	 * apporpriate for the operating system in use.  This is not bulletproof
	 * but it will get you by if the sys-api's are less than helpful.
	 * (I'm looking at you Linux).
	 *
	 * @param sSuffix a file name suffix, starting with a '.' (dot).
	 *        If this is null, then '.tmp' is used.
	 *
	 * @returns The temporary filename selected.
	 */
	const char* (*useTempFile)(struct gse_prn_job* pThis, const char* sSuffix);
	
	/** Get the filename to print. 
	 * returns NULL if no filename has been defined 
	 */
	const char* (*getFile)(const struct gse_prn_job* pThis);
	
	
	/** Set the printer for the print job.
	 *
	 *  @param sThis pointer to a the print job structure.
	 *  @param sPrinter if NULL, use the default printer
	 *  @return true if the printer exists, false if not.
	 */
	bool (*setPrinter)(struct gse_prn_job* pThis, const char* sPrinter);
	
	/** Get the printer to use, if NULL just print to a file */
	const char* (*getPrinter)(const struct gse_prn_job* pThis);
	
	
	/** Send the print job, return the job ID or 0 on an error */
	int (*sendJob)(struct gse_prn_job* pThis);
	
	/** Get the title of the print job, varies depending on if
	 * sendJob has been called.  Returns a string on the heap.
	 */
	char* (*getTitle)(const struct gse_prn_job* pThis);
	
	
	/* Protected member sets title from internal state, may be overridden */
	void (*setTitle)(struct gse_prn_job* pThis);
	
	int  nJobID;
	char* sTitle;
	char* sFile;
	char* sPrinter;
	bool bTmp;
	
	/** Print job destructor */
	void (*del)(void* vpThis);
	
} GsePrnJob;


/** Generate a new print job object
 *
 * @param sFile the name of the file to print.  If this is null at construction,
 *        the GsePrnJob::setFile member must be called prior to calling
 *        GsePrnJob::sendJob.
 *
 * @param sPrinter the name of the printer to send the job to, if NULL the
 *        default printer may be used.
 *
 * @param uOver set to 0, memory management used by derived classes 
 *
 * @returns a new print job object allocated on the heap
 *
 * @public 
 * @memberof GsePrnJob
 */
GsePrnJob* new_GsePrnJob(const char* sFile, const char* sPrinter, size_t uOver);


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GsePrnSvc_H_ */
