#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <sys/time.h>
#include <errno.h>

#include "GseExcept.h"
#include "GseConv.h"
#include "GseProcBlk.h"
#include "GseLog.h"

void GseProcBlk_setNow(GseProcBlk* pThis)
{
	struct timeval tv;
	
	if( gettimeofday(&tv, NULL) != 0)
		gse_throw(CLIB_XCEPT, errno);
	
	pThis->uBegSec = tv.tv_sec;
	pThis->uBegUSecDiv100 = tv.tv_usec / 100;
}

uint8 GseProcBlk_microEpoch(const GseProcBlk* pThis)
{
	return ( (uint8)pThis->uBegSec) * 1000000 + pThis->uBegUSecDiv100*100.0;
}

bytebuf_t GseProcBlk_encode(GseBlock* pBase, byte uExtraLen, byte* pOffset){
	bytebuf_t buf;
	byte uBaseOff;
	
	GseProcBlk* pThis = (GseProcBlk*)pBase;
	
	if((pBase->uType != 0x70)||(pThis->uStepID == 0)||
	   (pThis->uProgID == 0)||(pThis->uBegSec == 0)   )
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Incorrect value in one of, "
		   "BLK_TYPE: %hhu, PROCESS: %hhu, PROGRAM: %hhu, BEGIN_TIME_SEC: %u",
			pBase->uType, pThis->uStepID, pThis->uProgID, pThis->uBegSec);

	gse_overflowUinU(uExtraLen + 10, sizeof(byte)); 
	
	buf = GseBlock_encode(pBase, uExtraLen + 10, &uBaseOff);
	
	buf.ptr[uBaseOff]      = pThis->uStepID;
	buf.ptr[uBaseOff + 1]  = pThis->uProgID;
	gse_u4ToMSBbytes(buf.ptr + uBaseOff + 2, pThis->uBegSec);
	gse_u2ToMSBbytes(buf.ptr + uBaseOff + 6, pThis->uBegUSecDiv100);
	buf.ptr[uBaseOff + 8]  = pThis->uErrors;
	buf.ptr[uBaseOff + 9]  = pThis->uWarnings;
	
	if(pOffset != NULL) *pOffset = uBaseOff + 10;
	return buf;
}

byte GseProcBlk_decode(GseBlock* pBase, cbytebuf_t buf)
{
	GseProcBlk* pThis = (GseProcBlk*)pBase;
	byte uBaseOff = GseBlock_decode(pBase, buf);
	
	/* Check decoded type, make sure it is processing */
	if(pBase->uType != GSE_PROGRAM_BLK_TYPE_ID)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Expected 0x70 for the block type "
		             "ID, but recieved 0x%hhX", pBase->uType);
	
	/* Decoding, 1st 2 bytes handled by base class */
	pThis->uStepID =    *(buf.ptr + uBaseOff);
	pThis->uProgID =    *(buf.ptr + uBaseOff + 1);
	pThis->uBegSec =    gse_u4FromMSBbytes(buf.ptr + uBaseOff + 2);
	pThis->uBegUSecDiv100 = gse_u2FromMSBbytes(buf.ptr + uBaseOff + 6);
	
	if(pThis->uBegUSecDiv100 > 10000){
		gse_warning("Invalid microsecond times from GesProcBlk::setNow bug, "
		            "regenerate input file with a post 03/2010 version of librpwgse.");
		pThis->uBegUSecDiv100 = 0;
	}
		
	pThis->uErrors =    *(buf.ptr + uBaseOff + 8);
	pThis->uWarnings =  *(buf.ptr + uBaseOff + 9);
	
	return uBaseOff + 10;
}

bool GseProcBlk_merge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{	
	GseProcBlk* pThis = (GseProcBlk*)pBase;
	GseProcBlk* pProc = (GseProcBlk*)pOther;

	uint8 uThis = pThis->microEpoch(pThis);
	uint8 uOther = pThis->microEpoch(pProc);
	
	if(! pBase->canMerge(pBase, pOther, uFlags)) return false;
	
	if( ((uFlags&GSE_PROC_MERGE_LATE)&&(uOther > uThis)) || (uOther < uThis)){
		pThis->uBegSec = pProc->uBegSec;
		pThis->uBegUSecDiv100 = pProc->uBegUSecDiv100;
	}
	
	pThis->uErrors += pProc->uErrors;
	pThis->uWarnings += pProc->uWarnings;
	return true;
}

bool GseProcBlk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags)
{	
	if((pBase->uType != pOther->uType) || (pBase->uLen != pOther->uLen))
		return false;
	
	if( ((GseProcBlk*)pBase)->uStepID != ((GseProcBlk*)pOther)->uStepID )
		return false;
	
	if( strcmp(pBase->_sName, pOther->_sName) != 0) return false;
	
	return true;
}

GseBlock* GseProcBlk_copy(GseBlock* pBase, size_t uOverMalloc)
{
	GseProcBlk* pNew;
	GseProcBlk* pThis = (GseProcBlk*)pBase;	
	size_t uSize = 0;
	 
 	if(uOverMalloc > sizeof(GseProcBlk))
		uSize = uOverMalloc;
	else
		uSize = sizeof(GseProcBlk);
	
	pNew = (GseProcBlk*)GseBlock_copy(pBase, uSize);
	
	pNew->uStepID        = pThis->uStepID;
	pNew->uProgID        = pThis->uProgID;
	pNew->uBegSec        = pThis->uBegSec;
	pNew->uBegUSecDiv100 = pThis->uBegUSecDiv100;;
	pNew->uErrors        = pThis->uErrors;
	pNew->uWarnings      = pThis->uWarnings;
	pNew->setNow         = pThis->setNow;
	pNew->microEpoch     = pThis->microEpoch;
	pNew->del            = pThis->del;
	
	/* fprintf(stderr, "GseProcBlk_copy called\n");*/
	
	return (GseBlock*)pNew;
} 

GseProcBlk* new_GseProcBlk(cbytebuf_t buf, size_t uOverMalloc)
{
	size_t uSize;
	GseBlock* pBase = NULL;
	GseProcBlk* pThis = NULL;
	bool bIsTopLevel;
	
	if(uOverMalloc > sizeof(GseProcBlk)){
		uSize = uOverMalloc;
		bIsTopLevel = false;
	}
	else{
		uSize = sizeof(GseProcBlk);
		bIsTopLevel = true;
	}
	
	pBase = new_GseBlock(CBBUF_NULL, uSize);
	pThis = (GseProcBlk*)pBase;
	
	/* Override a couple base class functions */
	/* Del override not needed */
	pBase->encode = GseProcBlk_encode;
	pBase->decode = GseProcBlk_decode;
	pBase->_sName = "Processing";
	pBase->uType  = GSE_PROGRAM_BLK_TYPE_ID;
	pBase->copy   = GseProcBlk_copy;
	pBase->merge  = GseProcBlk_merge;
	pBase->canMerge = GseProcBlk_canMerge;
			
	pBase->_uMinLen += 10;
	
	/* Set new items */
	pThis->del = delete_GseBlock;
	pThis->setNow = GseProcBlk_setNow;
	pThis->microEpoch = GseProcBlk_microEpoch;
	
	/* Decoding or defaults */
	if(buf.ptr == NULL)
		return pThis;
	
	except_cleanup_push(delete_GseBlock, pThis);
	
	GseProcBlk_decode(pBase, buf);
	
	except_cleanup_pop(except_no_call);
	
	/*fprintf(stderr, "new_GseProcBlk called\n");*/
	return pThis;
}

GseProcBlk* new_GseProcBlk2(byte uStepID, byte uProgID, uint4 uBegSec,
                           uint2 uBegUSecDiv100)
{
	GseBlock* pBase = NULL;
	GseProcBlk* pThis = NULL;
	
	pBase = new_GseBlock(CBBUF_NULL, sizeof(GseProcBlk));
	pThis = (GseProcBlk*)pBase;
	
	/* Override a couple base class functions */
	/* Del override not needed */
	pBase->encode    = GseProcBlk_encode;
	pBase->decode    = GseProcBlk_decode;
	pBase->_sName    = "Processing";
	pBase->uType     = GSE_PROGRAM_BLK_TYPE_ID;
	pBase->copy      = GseProcBlk_copy;
	pBase->merge     = GseProcBlk_merge;
	pBase->canMerge  = GseProcBlk_canMerge;
	
	pBase->_uMinLen += 10;
	
	/* Set new items */
	pThis->del        = delete_GseBlock;
	pThis->setNow     = GseProcBlk_setNow;
	pThis->microEpoch = GseProcBlk_microEpoch;
	
	if((uStepID == 0)||(uProgID == 0))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	pThis->uStepID = uStepID;
	pThis->uProgID = uProgID;
	
	if(uBegSec != 0){
		pThis->uBegSec = uBegSec;
		pThis->uBegUSecDiv100 = uBegUSecDiv100;
	}
	else{
		GseProcBlk_setNow(pThis);
	}
	
	/*fprintf(stderr, "new_GseProcBlk2 called\n");*/
	return pThis;
}

/*****************************************************************************/
GseBlock* gse_generic_blocks(int2 nSpaceCraft, byte uType, byte uProcess,
		                        cbytebuf_t buf)
{
	if(uType == GSE_PROGRAM_BLK_TYPE_ID)
		return (GseBlock*)new_GseProcBlk(buf, 0);
	
	/* Add other generic block types such as message and aux blocks here... */
	
	/* the default just store the raw bytes */
	return new_GseBlock(buf, 0);
}
