/** @file GseProcBlk.h */
#ifndef _GseProcBlk_H_
#define _GseProcBlk_H_

#include <rpwgse/GseBlock.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_packet Packets */
/*@{*/
#define GSE_PROGRAM_BLK_TYPE_ID 0x70
#define GSE_PROC_BLK_TYPE       0x70

#define GSE_PROC_MERGE_EARLY   0x00000000
#define GSE_PROC_MERGE_LATE    0x00000001

/** Basic reading and writing of Gse Processing Blocks @n
 * @class GseProcBlk
 * @extends GseBlock
 */
typedef struct gse_proc_blk{
	
	/** Base structure */
	GseBlock base;
	
	/** FIELD: The processing step ID, ex: Unpacking, Unsegmenting, etc. */
	byte uStepID;
	
	/** FIELD: The ID of the program that preformed this processing step */
	byte uProgID;
	
	/** FIELD: The UNIX epoch time when this processing step was attempted */
	uint4 uBegSec;
	
	/** FIELD: The hundreds of micro seconds for the start of the 
	 * processing step */
	uint2 uBegUSecDiv100; 

	/** FIELD: Number of errors encountered during the processing step.
	@see getMsg to get message text */
	byte uErrors;
	
	/** FIELD: Number of warnings encountered during the processing step.
	@see getMsg to get message text */
	byte uWarnings;
	
	/** Update the @a uBegSec and @a uBegUSecDiv100 to have the current time */
	void (*setNow)(struct gse_proc_blk* pThis);
	
	/** Get the processing time in @b microseconds since 1970 */
	uint8 (*microEpoch)(const struct gse_proc_blk* pThis);

	/** Destructor */
	void (*del)(void* vpThis);
   

} GseProcBlk;

/** Generate a new Gse Processing block
 * @memberof GseProcBlk
 * @public
 */
GseProcBlk* new_GseProcBlk(cbytebuf_t buf, size_t uOverMalloc);

/** Generate a new Generic GSE processing block by specifing fields
 *
 * @param uStepID The processing step Emumeration from your project
 *        documentation (you do have project documentation right?)
 *
 * @param uProgID The ID for the program calling this function.
 *
 * @param uBegSec The UNIX epoch time when this processing step was attempted.
 *        If 0 will be set to @b now.
 *
 * @param uBegUSecDiv100 The hundenths of micro seconds for the start
 *        of the processing step.  Value is ignored if @a uBegSec is @c 0.
 *
 * @memberof GseProcBlk
 * @public
 */
GseProcBlk* new_GseProcBlk2(byte uStepID, byte uProgID, uint4 uBegSec,
                            uint2 uBegUSecDiv100);

/** Used by derived classes for thier copy constructors
 * @protected
 * @memberof GseProcBlk
 */
GseBlock* GseProcBlk_copy(GseBlock* pThis, size_t uOverMalloc);

/** Same as GseProcBlk::encode, used by derived classes
 * @memberof GseProcBlk
 * @protected
 */
bytebuf_t GseProcBlk_encode(GseBlock* pBase, byte uExtraLen, byte* pOffset);

/** Same as GseBlock::decode, used by derived classes 
 * @memberof GseProcBlk
 * @protected
 */
byte GseProcBlk_decode(GseBlock* pBase, cbytebuf_t buf);

/** Override of GseBlock::merge
 *
 * The merge operation will set @a uBegSec and @a uBegUSecDiv100 to
 * the earliest or latest of the two blocks, depending on the setting of
 * @a uFlags and will set uWarnings and uErrors to the sum of values in
 * both blocks.
 *
 * @param pBase A GseProcBlk cast as a GseBlock
 * @param pOther the other GseProcBlk derived block to merge with this one.
 * @param uFlags The only understood flag is @b GSE_PROC_MERGE_LATE.  For
 *        readability the flag GSE_PROC_MERGE_EARLY is also defined which
 *        just evaluates to zero.
 * @protected
 * @memberof GseProcBlk
 */
bool GseProcBlk_merge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags);

/** Override of GseBlock::canMerge
 *
 * Returns true if the two blocks have the same type IDs, the same length
 * and the same block name, false otherwise.
 * 
 * @param pBase A GseProcBlk cast as a GseBlock
 * @param pOther the other GseProcBlk derived block to merge with this one.
 * @param uFlags This parameter is ignored.
 *
 * @protected
 * @memberof GseProcBlk
 */
bool GseProcBlk_canMerge(GseBlock* pBase, const GseBlock* pOther, uint4 uFlags);
		
/** Set the time fields in a Gse Processing block to the current time 
 * @memberof GseProcBlk
 * @protected
 */
void GseProcBlk_setNow(GseProcBlk* pProcBlk);

/*@}*/


#ifdef __cplusplus
}
#endif

#endif /* _GseProcBlk_H_ */

