#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <strings.h>

#include "GseExcept.h"
#include "GseString.h"
#include "GseWrap.h"
#include "GseLog.h"

#include "GseProp.h"

/*****************************************************************************/
/* Property Base Class */

const char* GseProp_name(const GseProp* pThis)
{ 
	return (const char*)pThis->sName; 
}

int4 GseProp_line(const GseProp* pThis)
{
	return pThis->nLine;
}

gse_prop_type_t GseProp_type(const GseProp* pThis)
{ 
	return pThis->ptype;
}

void GseProp_setSrc(GseProp* pThis, const char* sFile, int4 nLine)
{
	pThis->nLine = nLine;
	pThis->sConf = sFile;
}

void GseProp_setConf(GseProp* pThis, const char* sConf)
{
	pThis->sConf = sConf;
}

void delete_GseProp(void* vpThis)
{
	GseProp* pThis = (GseProp*)vpThis;
	free(pThis->sName);
	free(vpThis);
}

GseProp* GseProp_copy(const GseProp* pThis, size_t uMalloc)
{
	GseProp* pNew = (GseProp*)gse_calloc(1, uMalloc);
	memcpy(pNew, pThis, uMalloc);
	pNew->sName = CS(pThis->sName);
	return pNew;
}

GseProp* new_GseProp(const char* sName, gse_prop_type_t type,  int4 nLine,
		               size_t uMalloc)
{
	GseProp* pThis = gse_calloc(1, uMalloc);
	pThis->name = GseProp_name;
	pThis->type = GseProp_type;
	pThis->del  = delete_GseProp;
	pThis->copy = NULL;           /* <--WARNING, MUST OVERRIDE THIS */
	pThis->setConf = GseProp_setConf;
	if(sName != NULL)
		pThis->sName = gse_strcopy(sName);
	pThis->ptype = type;
	pThis->nLine = nLine;
	
	return pThis;
}

/*****************************************************************************/
/* INTEGER PROPS */

GseProp* GsePropInt_copy(const GseProp* pBase)
{
	return GseProp_copy(pBase, sizeof(GsePropInt));
}
 
GsePropInt* new_GsePropInt(const char* sName, int8 nVal, int4 nSrcLine){
	GsePropInt* pThis;
	GseProp*    pBase;
	
	pBase = new_GseProp(sName, pt_int, nSrcLine, sizeof(GsePropInt));
	pBase->copy = GsePropInt_copy;
	
	pThis = (GsePropInt*) pBase;
	pThis->del  = delete_GseProp;
	pThis->nVal = nVal;
	return pThis;
}

/*****************************************************************************/
/* FLOAT PROPS */

GseProp* GsePropFloat_copy(const GseProp* pBase)
{
	return GseProp_copy(pBase, sizeof(GsePropFloat));
}

GsePropFloat* new_GsePropFloat(const char* sName, float8 fVal, int4 nSrcLine){
	GsePropFloat* pThis;
	GseProp*      pBase;
	
	pBase = new_GseProp(sName, pt_flt, nSrcLine, sizeof(GsePropFloat));
	pBase->copy = GsePropFloat_copy;
			
	pThis = (GsePropFloat*)	pBase;
	pThis->del  = delete_GseProp;
	pThis->fVal = fVal;
	return pThis;
}

/*****************************************************************************/
/* STRING PROPS */

GseProp* GsePropStr_copy(const GseProp* pBase)
{
	GsePropStr* pProp;
	pProp = (GsePropStr*)GseProp_copy(pBase, sizeof(GsePropStr));
	pProp->sVal = CS( ((GsePropStr*)pBase)->sVal);
	return (GseProp*)pProp;
}

void delete_GsePropStr(void* vpThis)
{
	free( ((GsePropStr*)vpThis)->sVal );
	free( vpThis );
}

GsePropStr* new_GsePropStr(const char* sName, const char* sVal, int nSrcLine){
	GseProp*    pBase;
	GsePropStr* pThis;
	
	pBase = new_GseProp(sName, pt_str, nSrcLine, sizeof(GsePropStr));
	pBase->del  = delete_GsePropStr;
	pBase->copy = GsePropStr_copy;
	
	pThis = (GsePropStr*)pBase;  
	if(sVal != NULL)
		pThis->sVal = gse_strcopy(sVal);
	
	pThis->del = delete_GsePropStr;
	
	return pThis;
}

/*****************************************************************************/
/* LIST PROPS */


void GsePropList_setConf(GseProp* pBase, const char* sConf)
{
	int i;
	GseProp* pProp;
	GsePropList* pThis;
	
	pThis = (GsePropList*)pBase;
	pBase->sConf = sConf;
	
	for(i=0; i<pThis->props->num; i++){
		pProp = pThis->props->pAry[i];	
		pProp->setConf(pProp, sConf);            /* Handles the recursion */
	}
}

int GsePropList_length(const GsePropList* pThis)
{
	if(pThis->props == NULL) return 0;
	else return pThis->props->num;
}

int GsePropList_add(GsePropList* pThis, GseProp* pProp)
{
	pThis->props->add(pThis->props, pProp);
	return pThis->props->num;
}

GseProp* GsePropList_get(const GsePropList* pThis, int4 idx)
{
	GseProp* pBase;
	const char* sMyName = NULL;
	pBase = (GseProp*) pThis;	
	
	if((idx < 0)||(idx >= pThis->props->num)){
		
		if(pBase->sName != NULL) sMyName = pBase->sName;
		else                      sMyName = "(unnamed)";
		
		gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "List property %s has %d items, "
		             "request for item at index %d is out of bounds.", sMyName,
		             pThis->props->num, idx);
	}
	
	return pThis->props->pAry[idx];
	
} 

GseProp* GsePropList_getByName(const GsePropList* pThis, const char* sName)
{
	int i;
	GseProp* pProp;
	GsePropList* pSubList;
	
	if((sName == NULL)||(sName[0] == '\0'))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sName must not be NULL or empty in "
		            "GsePropList::getPropByName"); 
	
	if(pThis->props == NULL) return NULL;
	
	for(i = 0; i < pThis->props->num; i++){
		pProp = pThis->props->pAry[i];
		
		/* check this properties name */
		if(pProp->sName != NULL){
			if( strcmp(pProp->sName, sName) == 0)
				return pProp;
		}
		
		/* Recursively search into lists */
		if(pProp->ptype == pt_list){
			pSubList = (GsePropList*)pProp;
			
			if( (pProp = pSubList->getByName(pSubList, sName)) != NULL)
				return pProp;
		}
	}
	
	return NULL;
}

GseProp* GsePropList_copy(const GseProp* pBase)
{
	int i;
	GsePropList* pNew;
	const GsePropList* pThis = (const GsePropList*)pBase;
	
	pNew = new_GsePropList(pBase->sName, NULL, pBase->nLine);
	
	for(i = 0; i<pThis->props->num; i++)
		pNew->add(pNew, pThis->props->pAry[i]->copy(pThis->props->pAry[i]) );
	
	return (GseProp*)pNew;
}

void delete_GsePropList(void* vpThis)
{
	GsePropList* pThis = (GsePropList*)vpThis;
	
	if(pThis->props != NULL)
		pThis->props->del(pThis->props);
	delete_GseProp(vpThis);
}

GsePropList* new_GsePropList(const char* sName,GsePropAry* pProps,int nSrcLine){
	GseProp*    pBase;
	GsePropList* pThis;
	
	pBase = new_GseProp(sName, pt_list, nSrcLine, sizeof(GsePropList));
	pBase->del = delete_GsePropList;
	pBase->copy = GsePropList_copy;
	pBase->setConf = GsePropList_setConf;
	
	pThis = (GsePropList*)pBase;
	
	if(pProps != NULL)
		pThis->props = pProps;
	else
		pThis->props = new_GsePropAry(0);
	
	pThis->length    = GsePropList_length;
	pThis->add       = GsePropList_add;
	pThis->get       = GsePropList_get;
	pThis->getByName = GsePropList_getByName;
	pThis->del       = delete_GsePropList;
	
	return pThis;	
}


/*****************************************************************************/
/* ENUM PROPS */

int4 GsePropEnum_length(const GsePropEnum* pThis)
{
	return pThis->pDict->length(pThis->pDict);
}

bool GsePropEnum_hasDefault(const GsePropEnum* pThis)
{
	if(pThis->sDefault)
		return true;
	else
		return false;
}

char* GsePropEnum_u8ToStr(const GsePropEnum* pThis, uint8 uVal)
{
	int8 i8;
	if(uVal > 0x8000000000000000){
		gse_warning("Enum range error %llu > 0x8000000000000000");
		return CS("RNG ERR");
	}
	
	i8 = uVal;
	
	return pThis->i8ToStr(pThis, i8);
}

char* GsePropEnum_i8ToStr(const GsePropEnum* pThis, int8 nVal)
{
	if(pThis->pDict->hasKey(pThis->pDict, nVal))
		return CS( pThis->pDict->get(pThis->pDict, nVal) );
	
	if(pThis->sDefault)
		return CS( pThis->sDefault );
	
	return NS("%lld", nVal);		
}

void GsePropEnum_add(GsePropEnum* pThis, int8 nVal, const char* sVal)
{
	pThis->pDict->insert(pThis->pDict, nVal, gse_strcopy(sVal) );
}

bool GsePropEnum_hasInt(const GsePropEnum* pThis, int8 nVal)
{
	return pThis->pDict->hasKey(pThis->pDict, nVal);
}

const char* GsePropEnum_get(const GsePropEnum* pThis, int8 nVal)
{
	const char* sRet;
	if((sRet = pThis->pDict->get(pThis->pDict, nVal)) == NULL){
		return (const char*) pThis->sDefault;
	}		
	else
		return sRet;
}

void GsePropEnum_setDefault(GsePropEnum* pThis, const char* sDef)
{
	free(pThis->sDefault);
	if(sDef)
		pThis->sDefault = gse_strcopy(sDef);
}

GseProp* GsePropEnum_copy(const GseProp* pBase)
{
	GsePropEnum* pNew;
	const GseIntStrDict* pSrc;
	const int8* pKey;
	const GsePropEnum* pThis = (const GsePropEnum*)pBase;
	
	pNew = new_GsePropEnum(pBase->sName, pBase->nLine);
	
	pSrc = pThis->pDict;
	
	for(pKey = pSrc->first(pSrc); pKey != NULL; pKey = pSrc->next(pSrc, *pKey)){
		pNew->add(pNew, *pKey, pSrc->get(pSrc, *pKey));
	}
	
	return (GseProp*)pNew;
}

void delete_GsePropEnum(void* vpThis)
{
	GsePropEnum* pThis = (GsePropEnum*)vpThis;
	
	free(pThis->sDefault);
	pThis->pDict->del(pThis->pDict);
	delete_GseProp(vpThis);
}

GsePropEnum* new_GsePropEnum(const char* sName, int4 nSrcLine)
{
	GseProp*     pBase;
	GsePropEnum* pThis;
	
	pBase = new_GseProp(sName, pt_enum, nSrcLine, sizeof(GsePropEnum));
	pBase->del = delete_GsePropEnum;
	pBase->copy = GsePropEnum_copy;
	
	pThis = (GsePropEnum*)pBase;
	pThis->pDict      = new_GseIntStrDict(pBase->sName);
	
	pThis->length     = GsePropEnum_length;
	pThis->hasDefault = GsePropEnum_hasDefault;
	pThis->u8ToStr    = GsePropEnum_u8ToStr;
	pThis->i8ToStr    = GsePropEnum_i8ToStr;
	pThis->add        = GsePropEnum_add;
	pThis->hasInt     = GsePropEnum_hasInt;
	pThis->setDefault = GsePropEnum_setDefault;
	pThis->get        = GsePropEnum_get;
	pThis->del        = delete_GsePropEnum;
	
	return pThis;	
}

/*****************************************************************************/
/* Property Array */

#define C GsePropAry
#define T GseProp
#define CASCADE_DEL

#include "ptrAryTplt.c"

#undef CASCADE_DEL
#undef C
#undef T

/*****************************************************************************/
/* Int to Str dictionary */

#define C GseIntStrDict
#define K int8
#define V char

#define FREE_VALS
#define KEY_FMT  "%d"

#include "../rpwgse/dictTplt.c"

#undef FREE_VALS
#undef C
#undef K
#undef V

/*****************************************************************************/
/* Str to Prop dictionary */

#define C GseStrPropDict
#define K char
#define V GseProp

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcasecmp(a, b)

#include "ptrDictTplt.c"

#undef FREE_KEYS
#undef DEL_VALS
#undef C
#undef K
#undef V

/*****************************************************************************/
/* Prop dictionary helpers */

bool gse_pd_cktype(const GseStrPropDict* pDict, gse_prop_type_t chtype, 
                   const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) == NULL)
		return true;
	
	return (pProp->ptype == chtype) ? true : false;
}

const int8* gse_pd_getint(const GseStrPropDict* pDict, const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) != NULL){
		if(pProp->type(pProp) != pt_int)
			gse_throw_fm(GSE_XCEPT, GSE_XSYNTAX, "Property %s, defined at %s, line"
					       " %d is not an integer property.", pProp->name(pProp),
					       pProp->sConf, pProp->nLine);
		return &( ((const GsePropInt*)pProp)->nVal );
	}
	
	return NULL;
}

const float8* gse_pd_getfloat(const GseStrPropDict* pDict, const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) != NULL){
		if(pProp->type(pProp) != pt_flt)
			gse_throw_fm(GSE_XCEPT, GSE_XSYNTAX, "Property %s, defined at %s, line"
					       " %d is not an floating point property.", 
			             pProp->name(pProp), pProp->sConf, pProp->nLine);
		return &( ((const GsePropFloat*)pProp)->fVal );
	}
	
	return NULL;
}

const char* gse_pd_getstr(const GseStrPropDict* pDict, const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) != NULL){
		if(pProp->type(pProp) != pt_str)
			gse_throw_fm(GSE_XCEPT, GSE_XSYNTAX, "Property %s, defined at %s, line"
					       " %d is not a string property.", pProp->name(pProp),
					       pProp->sConf, pProp->nLine);
		return (const char*) ( ((const GsePropStr*)pProp)->sVal );
	}
	
	return NULL;
}

const GsePropList* gse_pd_getlist(const GseStrPropDict* pDict, const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) != NULL){
		if(pProp->type(pProp) != pt_list)
			gse_throw_fm(GSE_XCEPT, GSE_XSYNTAX, "Property %s, defined at %s, line"
					       " %d is not a list property.", pProp->name(pProp),
					       pProp->sConf, pProp->nLine);
		return (const GsePropList*)pProp;
	}
	
	return NULL;
}

const GsePropEnum* gse_pd_getenum(const GseStrPropDict* pDict, const char* sKey)
{
	const GseProp* pProp;
	
	if( (pProp = pDict->get(pDict, sKey)) != NULL){
		if(pProp->type(pProp) != pt_enum)
			gse_throw_fm(GSE_XCEPT, GSE_XSYNTAX, "Property %s, defined at %s, line"
					       " %d is not an enum property.", pProp->name(pProp),
					       pProp->sConf, pProp->nLine);
		return (const GsePropEnum*)pProp;
	}
	
	return NULL;
}

void gse_pd_setconfig(GseStrPropDict* pDict, const char* sConfFile)
{
	GseProp* pProp;
	const char* sKey;
	
	for(sKey = pDict->first(pDict); sKey != NULL; sKey = pDict->next(pDict,sKey))
	{
		pProp = pDict->getMutable(pDict, sKey);
		pProp->setConf(pProp, sConfFile);
	}
}
