/** @file GseProp.h Extrator properties */
#ifndef _GseProp_H_
#define _GseProp_H_

#include <rpwgse/dict.h>
#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_bin_parser */
/*@{*/

/** Binary blob parser property types */
typedef enum _gse_prop_type{ 
	pt_invalid = 0, pt_int, pt_flt, pt_str, pt_list, pt_enum
} gse_prop_type_t;

/** Base class of binary blob parser properties
 * @class GseProp
 */
typedef struct gse_prop{
	/** Get the name of the property, if it has one.
	 * Anonymous properties do not have names
	 * @return A pointer to the name of this property, or null if the
	 *         property is unnammed. */
	const char* (*name)(const struct gse_prop* pThis);
	
	/** Get the type of a property, all properties have a defined type */
	gse_prop_type_t (*type)(const struct gse_prop* pThis);
	
	/** Get a copy of (new object on the heap) of this property */
	struct gse_prop* (*copy)(const struct gse_prop* pBase);
	
	/** Destructor */
	void (*del)(void* vpBase);
	
	/* Private Stuff */
	void (*setConf)(struct gse_prop* pBase, const char* sConf);
	
	
	char* sName;
	gse_prop_type_t ptype;
	int4 nLine;
	const char* sConf;  /* The config file name pointer */
	
} GseProp;

/* An array of properties, accessed by indicies, this is used
   by the list property. */
#define S gse_prop_ary
#define C GsePropAry
#define T GseProp

#include "ptrAryTplt.h"

#undef S
#undef C
#undef T


/****************************************************************************/
/* Different Types of Properties... */

/** Binary blob parser integer property 
 * @class GsePropInt
 * @extends GseProp
 */ 
typedef struct gse_prop_int{
	
	/** Base Structure */
	GseProp base;
	
	/** The value */
	int8 nVal;
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} GsePropInt;

/** GsePropInt constructor
 * @public
 * @memberof GsePropInt
 */
GsePropInt* new_GsePropInt(const char* sName, int8 nVal, int4 nSrcLine);



/** Field Extractor floating point number property 
 * @class GsePropFloat
 * @extends GseProp
 */ 
typedef struct gse_prop_float{
	
	/** Base structure */
	GseProp base;
	
	/** The value */
	float8 fVal;
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} GsePropFloat;

/** GsePropFloat constructor
 * @public
 * @memberof GsePropFloat
 */
GsePropFloat* new_GsePropFloat(const char* sName, float8 fVal, int4 nSrcLine);



/** Field Extractor string property 
 * @class GsePropStr
 * @extends GseProp
 */ 
typedef struct gse_prop_str{
	
	/** Base Structure */
	GseProp base;
	
	/** The value (freed on delete) */
	char* sVal;
	
	/** Destructor */
	void (*del)(void* vpThis);
	
} GsePropStr;

/** GsePropStr constructor
 * @public
 * @memberof GsePropStr
 */
GsePropStr* new_GsePropStr(const char* sName, const char* sVal, int4 nSrcLine);



/** Field Extractor list properties 
 * Lists can contain any other property type, including other lists
 * @class GsePropList
 * @extends GseProp
 */
typedef struct gse_prop_list{
	GseProp base;
	/** The number of items in this list.
	 * The function is not recursive, it doesn't count the items in sub lists
	 */
	int4 (*length)(const struct gse_prop_list* pThis);
	
	/** Add a named or anonymous property to this list */
	int4 (*add)(struct gse_prop_list* pThis, GseProp* pProp);
	
	/** Do a deep search for a property by name.
	 *
	 * Lists embedded in this list are searched recursively.  Anonymous
	 * properties connot be found with this function.  But a list property
	 * embedded in another list property could be looked up in this manner.
	 */
	GseProp* (*getByName)(const struct gse_prop_list* pThis,
			                    const char* sName);
	
	/** Do a shallow search for a property by index.
	 * Lists embeded in lists are not searched recursively
	 */
	GseProp* (*get)(const struct gse_prop_list* pThis, int4 idx);
	
	/** Destructor */
	void (*del)(void* vpThis);
	
	GsePropAry* props;
} GsePropList;

/** Make a new empty list property
 *
 * @param sName The name of this property
 *
 * @param pProps an optional starting array of properties.  If specified,
 *        the GsePropList takes ownership of the array.
 *
 * @param  nSrcLine If the property list is built from a configuration file 
 *        you can include the line number here and it will be included in
 *        any error messages.
 *
 * @return A new GsePropList allocated on the heap.
 *
 * @public
 * @memberof GsePropList
 */
GsePropList* new_GsePropList(const char* sName, GsePropAry* pProps,
                             int4 nSrcLine);



/* A dictionary of integer to string lookups */
#define S gse_int_str_dict
#define C GseIntStrDict
#define K int8
#define V char

#include "dictTplt.h"

#undef S
#undef C
#undef K
#undef V

/** Emum property, used for enums, and enum like items.
 * An emum is a mapping of integer values to text strings.
 * @class GsePropEnum
 * @extends GseProp
 */
typedef struct gse_prop_enum{
	GseProp base;
	
	/** Get the length of the enumeration (not including the default value) */
	int4 (*length)(const struct gse_prop_enum* pThis);
	
	bool (*hasDefault)(const struct gse_prop_enum* pThis);
	
	/** Convert an usigned integer to an enumeration string */
	char* (*u8ToStr)(const struct gse_prop_enum* pThis, uint8 uVal);
	
	/** Convert a signed integer to an enumeration string */
	char* (*i8ToStr)(const struct gse_prop_enum* pThis, int8 nVal);
	
	/* See if a key (the int8) already exists in this enumeration */
	bool (*hasInt)(const struct gse_prop_enum* pThis, int8 key);
	
	/* Add key value pair to the enum */
	void (*add)(struct gse_prop_enum* pThis, int8 key, const char* sVal);
	
	/* Set a string to return if key is not found.
	 */
	void (*setDefault)(struct gse_prop_enum* pThis, const char* sDef);
	
	/** Given an integer value, get the enumeration string.
	 *  @return NULL if there is no string for this value.
	 */
	const char* (*get)(const struct gse_prop_enum* pThis, int8 nVal);
	
	/** Destructor */
	void (*del)(void* vpThis);
	
	char* sDefault;
	
	GseIntStrDict* pDict;
	
} GsePropEnum;

/** GsePropEnum constructor
 * @public
 * @memberof GsePropEnum
 */
GsePropEnum* new_GsePropEnum(const char* sName, int4 nSrcLine);

/* Property dictionaries *****************************************************/

#define S gse_str_prop_dict
#define C GseStrPropDict
#define K char
#define V GseProp

#include "ptrDictTplt.h"

#undef S
#undef C
#undef K
#undef V

/** Sets the config file name for all properties in the dictionary.
 *
 * @param pDict the property dictionary in question.
 * @param sConfFile a constant pointer to the name of the config file.
 * 
 * @warning The string @a sFile is @b NOT copied to internal memory. It is
 *          the responsibility of the caller to insure that pointer is valid
 *          for the lifetime of the properties in the dictionary.
 * @public
 * @memberof GseStrPropDict
 */
void gse_pd_setconfig(GseStrPropDict* pDict, const char* sConfFile);


/** check the type of a property, if it exists
 *
 * @returns One of the following:
 *   - @a false if the property exists but is not of type @a chtype
 *   - @a true  if the property exists and is of type @a chtype
 *   - @a  @b true if the proerty does not exist.
 *
 * @public
 * @memberof GseStrPropDict
 */
bool gse_pd_cktype(const GseStrPropDict* pDict, gse_prop_type_t chtype, 
                   const char* sKey);

/** Get an integer value out of a dictionary of properties.
 *
 * @param pDict the property dictionary in question.
 * @param sKey the string key used to lookup the property.
 *
 * @returns a pointer to the interger value, or NULL if no property exists
 *          at key @a sKey.
 *
 * @throws (GSE_XCEPT, GSE_XSYNTAX) if the property at the given key is
 *         not a GsePropInt structure.
 * @public
 * @memberof GseStrPropDict
 */
const int8* gse_pd_getint(const GseStrPropDict* pDict, const char* sKey);

/** Get a floating point value out of a dictionary of properties 
 *
 * @param pDict the property dictionary in question.
 * @param sKey the string key used to lookup the property.
 *
 * @returns a pointer to the floating point value, or NULL if no property
 *          exists at key @a sKey.
 *
 * @throws (GSE_XCEPT, GSE_XSYNTAX) if the property at the given key is
 *         not a GsePropFloat or a GsePropInt structure.
 * @public
 * @memberof GseStrPropDict
 */
const float8* gse_pd_getfloat(const GseStrPropDict* pDict, const char* sKey);

/** Get a string value out of a dictionary of properties
 *
 * @param pDict the property dictionary in question.
 * @param sKey the string key used to lookup the property.
 *
 * @returns a pointer to the string value, or NULL if no property exists
 *          at key @a sKey.
 *
 * @throws (GSE_XCEPT, GSE_XSYNTAX) if the property at the given key is
 *         not a GsePropStr structure.
 * @public
 * @memberof GseStrPropDict
 */
const char* gse_pd_getstr(const GseStrPropDict* pDict, const char* sKey);

/** Get a sub property out of a list property that lives in a dictionary.
 *
 * @param pDict the property dictionary in question.
 * @param sKey the string key used to lookup the property.
 *
 * @returns a pointer to the list property, or NULL if no property exists
 *          at key @a sKey.
 *
 * @throws (GSE_XCEPT, GSE_XSYNTAX) if the property at the given key is
 *         not a GsePropList structure.
 * @public
 * @memberof GseStrPropDict
 */
const GsePropList* gse_pd_getlist(const GseStrPropDict* pDict, const char* sKey);

/** Get a string that maps to an integer out of an enum property in a
 * dictionary
 *
 * @param pDict the property dictionary in question.
 * @param sKey the string key used to lookup the property.
 *
 * @returns a pointer to a GsePropEnum, or NULL if no property exists
 *          at key @a sKey.
 *
 * @throws (GSE_XCEPT, GSE_XSYNTAX) if the property at the given key is
 *         not a GsePropEnum structure.
 * @public
 * @memberof GseStrPropDict
 */
const GsePropEnum* gse_pd_getenum(const GseStrPropDict* pDict, const char* sKey);


/*@}*/

#ifdef __cplusplus
}
#endif


#endif /* _GseExtorProps_H_ */
