
%{
/* A lexer for the field extraction from blobs for the rpwgse library */
#include <stdio.h>
#include <strings.h>

#include <rpwgse/GseTypes.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseLog.h>

#include <rpwgse/GseParser.h>

#include <rpwgse/gse_psrbldr_int.h>

%}

%pure-parser
%name-prefix="gse_psrbldr_"

%locations
%error-verbose

%parse-param { struct gse_psrbldr_state* pState }
%lex-param {void* scanner}

   /* The first three types represent the value types according to the
	   lexer (which has a much simpiler view of the world) the rest are
		the value types according to the parser. */ 
%union {
	float8 f8val;
	int8 i8val;
	char* sVal;
	void* pVoid;
	
	GseStrPsrDict*   pParsers;
	GseParser*       pParser;
	GseStrFieldDict* pDictField;
	GseField*        pField;
	GseStrPropDict*  pDictProp;
	GseProp*         pProp;
	GsePropInt*      pPropInt;
	GsePropFloat*    pPropFloat;
	GsePropStr*      pPropStr;
	GsePropList*     pPropList;
	GsePropAry*      pAryProp;
	GsePropEnum*     pPropEnum;
	
	GseStrAry*       pStrings;
}

%token <sVal>  STRING
%token <i8val> INT
%token <f8val> FLOAT

%token PARSER         "parser"
%token DEFAULT        "default"
%token RMDEFAULT      "rmdefault"
%token PARAMETERS     "parameters"
%token END         0  "End of input"

%type <pDictProp>  parameters
%type <pParsers>   parsers

%type <pParser>    parser
%type <pDictField> fields
%type <pField>     field

%type <pDictProp>  properties
%type <pProp>      property
%type <pProp>      value
%type <pAryProp>   list
%type <pAryProp>   listitems
%type <pPropEnum>  enumeration
%type <pPropEnum>  enumitems

%type <pStrings>   keys
%type <sVal>       key

%{
typedef void* yyscan_t;

char* gse_psrbldr_get_text(yyscan_t yyscanner);

int gse_psrbldr_get_lineno(yyscan_t yyscanner);

int gse_psrbldr_lex(YYSTYPE * yylval_param,YYLTYPE * yylloc_param,
                    yyscan_t yyscanner);

void gse_psrbldr_error(YYLTYPE* locp, struct gse_psrbldr_state* pState,
                       const char* sMsg);

void gse_psrbldr_update_defs(GseStrPropDict* pDefs, GseStrPropDict* pNew);

void gse_psrbldr_rm_defs(GseStrPropDict* pDefs, GseStrAry* keys);

bool gse_psrbldr_hasstr(GseStrAry* strs, const char* sVal);

#define scanner pState->scanner
%}

%%

everything: parameters parsers {
	pState->pParams = $1;
	pState->pParsers = $2;
};

everything: parsers {
	pState->pParsers = $1;
};

parameters: PARAMETERS '{' properties '}' {
	$$ = $3;
};

parsers: parsers parser {
	if($1->hasKey($1, $2->sName)){
		char* sErr = gse_string("Duplicate parser definition '%s'", $2->sName);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;
	}	
	$1->insert($1, $2->sName, $2);
	$$ = $1;
};

parsers: parsers default {
	/* Shift-Reduce as a no op */
	$$ = $1;
};

parsers: parsers rmdefault {
	/* Shift-Reduce as a no op */
	$$ = $1;
};
	
parsers: parser {
	$$ = new_GseStrPsrDict(NULL);
	$$->disownAllKeys($$);
	$$->insert($$, $1->sName, $1);
};

parsers: default {
	$$ = new_GseStrPsrDict(NULL);
	$$->disownAllKeys($$);
};

parsers: rmdefault {
	$$ = new_GseStrPsrDict(NULL);
	$$->disownAllKeys($$);
};

parser: PARSER STRING '{' properties fields '}' {
	$$ = new_GseParser( $2, $4, $5, pState->sFile);
}

parser: PARSER STRING '{' fields '}' {
	$$ = new_GseParser( $2, NULL, $4, pState->sFile);
}

fields: fields field {
	if($1->hasKey($1, $2->sName)){
		char* sErr = gse_string("Duplicate field definition '%s'", $2->sName);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;
	}
	$1->insert($1, $2->sName, $2);
	$$ = $1
};

fields: fields default {
	/* Shift-Reduce as a no op */
	$$ = $1;
}

fields: fields rmdefault {
	/* Shift-Reduce as a so op */
	$$ = $1;
}

fields: field {
	$$ = new_GseStrFieldDict(NULL);
	$$->disownAllKeys($$);
	$$->insert($$, $1->sName, $1);
};

fields: default {
	$$ = new_GseStrFieldDict(NULL);
	$$->disownAllKeys($$);
}

fields: rmdefault {
	$$ = new_GseStrFieldDict(NULL);
	$$->disownAllKeys($$);
}

field: STRING '{' properties '}'{
	$$ = new_GseField($1, $3, pState->pDefs,  gse_psrbldr_get_lineno(scanner));
	/* Validate it */
	char* sTmp;
	sTmp = $$->validate($$);
	if(sTmp != NULL){
		char* sErr;
		sErr = gse_string("Field '%s' is invalid; %s.\n", $$->sName, sTmp);
		free(sTmp);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;
	}
};

default: DEFAULT '{' properties '}' {
	gse_psrbldr_update_defs(pState->pDefs, $3);
	$3->del($3);
};

rmdefault: RMDEFAULT '{' keys '}' {
	/* Error if not set as a default */
	int i;
	for(i = 0; i < $3->num; i++){
		if( pState->pDefs->hasKey(pState->pDefs, $3->pAry[i]) == false){
			char* sErr;
			sErr = gse_string("No default has been set for property %s\n", 
			                  $3->pAry[i]);
			gse_psrbldr_error(&yylloc, pState, sErr);
			free(sErr);
			YYERROR;	
		}
	}
	gse_psrbldr_rm_defs(pState->pDefs, $3);
	$3->del($3);
}

keys: keys key {
	/* Error if key is mentioned more than once */
	if(gse_psrbldr_hasstr($1, $2)){
		char* sErr;
		sErr = gse_string("Property %s listed twice rmdefault list\n", $2);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;	
	}
	$1->add($1, $2);
	$$ = $1;
}

keys: key {
	$$ = new_GseStrAry(0);
	$$->add($$, $1);
}

key: STRING {
	$$ = $1;
}

/* Strict Hieractical stuff is below here */

properties: properties property{
	if($1->hasKey($1, $2->sName)){
		char* sErr;
		sErr = gse_string("Property %s is listed more than once.\n", $2->sName);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;
	}
	$1->insert($1, $2->sName, $2);
	$$ = $1;
};

properties: property {
	$$ = new_GseStrPropDict(NULL);
	$$->disownAllKeys($$);
	$$->insert($$, $1->sName, $1);
};

property: STRING '=' value {
	$3->sName = gse_string($1);
	$$ = $3;
};
	
value: STRING { 
	$$ = (GseProp*) new_GsePropStr(NULL, $1, gse_psrbldr_get_lineno(scanner)); 
};

value: INT { 
	$$ = (GseProp*) new_GsePropInt(NULL, $1, gse_psrbldr_get_lineno(scanner)); 
};

value: FLOAT  { 
	$$ = (GseProp*) new_GsePropFloat(NULL, $1, gse_psrbldr_get_lineno(scanner)); 
};

 /* List Reduction */
value: list   { 
	$$ = (GseProp*) new_GsePropList(NULL, $1, gse_psrbldr_get_lineno(scanner));
};
					  
list: '(' listitems ')' { $$ = $2 } ;

listitems: value {
	$$ = new_GsePropAry(0);
	$$->add($$, $1);
};

listitems: property {
	$$ = new_GsePropAry(0);
	$$->add($$, $1);
};

listitems: listitems ',' value {
	$1->add($1, $3);
	$$ = $1;
};

listitems: listitems ',' property{
	$1->add($1, $3);
	$$ = $1;
};

  /* Enumeration Reduction */

value: enumeration { 
	$$ = (GseProp*)$1;
};

enumeration: '[' enumitems ']' { $$ = $2 };

enumitems: enumitems enumitem {
	
	if(pState->bEnumDef){
		$1->setDefault($1, pState->sEnumVal);
	}
	else{
		if( $1->hasInt($1, pState->iEnumVal) ){
			char* sErr;
			sErr = gse_string("Value %d already exists in enumeration\n", 
		   	               pState->iEnumVal);
			gse_psrbldr_error(&yylloc, pState, sErr);
			free(sErr);
			YYERROR;	
		}
		$1->add($1, pState->iEnumVal, pState->sEnumVal);
	}
		
	pState->iEnumVal = 0;
	pState->sEnumVal = NULL;
	pState->bEnumDef = false;
	$$ = $1;
};

enumitems: enumitem {
	$$ = new_GsePropEnum(NULL, gse_psrbldr_get_lineno(scanner));
	
	if(pState->bEnumDef)
		$$->setDefault($$, pState->sEnumVal);
	else
		$$->add($$, pState->iEnumVal, pState->sEnumVal);
		
	pState->iEnumVal = 0;
	pState->sEnumVal = NULL;
	pState->bEnumDef = false;
};

enumitem: STRING ':' STRING {
	/* Setting the default value for an enumeration */
	if(strcasecmp("def", $1) != 0){
		char* sErr;
		sErr = gse_string("Unexpected string '%s' in enumeration", $1);
		gse_psrbldr_error(&yylloc, pState, sErr);
		free(sErr);
		YYERROR;
	}
	
	pState->sEnumVal = $3;
	pState->bEnumDef = true;
};

enumitem: INT ':' STRING { 
	pState->iEnumVal = $1;
	pState->sEnumVal = $3;
	pState->bEnumDef = false;
};
	
%%

/*****************************************************************************/
/* Parsing the config file */

bool gse_psrbldr_hasstr(GseStrAry* strs, const char* sVal)
{
	int i;
	for(i=0; i<strs->num; i++)
		if(strcmp( strs->pAry[i], sVal) == 0)
			return true;
	return false;
}

void gse_psrbldr_update_defs(GseStrPropDict* pDefs, GseStrPropDict* pNew)
{
	const GseProp* pNewProp;
	GseProp* pCopy;
	GseProp* pOldProp;
	const char* sKey;
		
	for(sKey = pNew->first(pNew); sKey != NULL; sKey = pNew->next(pNew, sKey))
	{	
		pNewProp = pNew->get(pNew, sKey);
		pCopy = pNewProp->copy(pNewProp);
		
		if(pDefs->hasKey(pDefs, sKey)){
			pOldProp = pDefs->replace(pDefs, pCopy->sName, pCopy);
			pOldProp->del(pOldProp);
		}
		else{
			pDefs->insert(pDefs, pCopy->sName, pCopy);
		}
	}	
}

void gse_psrbldr_rm_defs(GseStrPropDict* pDefs, GseStrAry* keys){
	int i;
	GseProp* p;
	for(i = 0; i < keys->num; i++){
		if(pDefs->hasKey(pDefs, keys->pAry[i])){
			p = pDefs->take(pDefs, keys->pAry[i]);
			p->del(p);
		}
	}
}

void gse_build_parsers(const char* sConf, const char* sFile, 
                       GseStrPropDict** ppParams, GseStrPsrDict** ppParsers)
{
	struct gse_psrbldr_state state;
	GseStrPropDict* d;
	char sStopLeak[256];
	sStopLeak[255] = '\0';
	
	/* Setup our parts of the state structure */
	state.iEnumVal = 0;
	state.sEnumVal = NULL;
	state.pParams = NULL;
	state.pParsers = NULL;
	state.sFile = sFile;
	state.sErr = NULL;
	state.pDefs = new_GseStrPropDict(NULL);
	state.pDefs->disownAllKeys(state.pDefs);
	
	if(sConf == NULL || ppParams == NULL || ppParsers == NULL)
		gse_throw(GSE_XCEPT, GSE_XNULLPTR);
		
	/* Set the default properties:
	 *
	 * default{ bo=BE sz=1 off=0 ty=U sft=0 fmt="%u" } 
	 */
	d = state.pDefs;
	d->setName(d, "Defaults");
	d->insert(d, CS("bo"), (GseProp*)new_GsePropStr("bo", "BE", -1));
	d->insert(d, CS("sz"), (GseProp*)new_GsePropInt("sz", 1, -1));
	d->insert(d, CS("off"), (GseProp*)new_GsePropInt("off", 0, -1));
	d->insert(d, CS("ty"), (GseProp*)new_GsePropStr("ty", "U", -1));
	d->insert(d, CS("sft"), (GseProp*)new_GsePropInt("sft", 0, -1));
	d->insert(d, CS("pfunc"), (GseProp*)new_GsePropStr("pfunc", "printf", -1));
	
	/* No default format is needed */
	
	except_cleanup_push(state.pDefs->del, state.pDefs);
	
	/* Let the lexer init what it needs */
	gse_psrbldrlxr_beg(&state, sConf);
	
	except_cleanup_push(free, state.sErr);
	
	/* gse_psrbldr_parse is generated by bison */
	if( gse_psrbldr_parse(&state) != 0){
		if(state.sErr != NULL){
			strncpy(sStopLeak, state.sErr, 255); 
			gse_throw_m(GSE_XCEPT, GSE_XSYNTAX, sStopLeak);
		}
		else
			gse_throw_m(GSE_XCEPT, GSE_XSYNTAX, 
			             "Error reading parser configuration data.");
	}
	
	gse_psrbldrlxr_end(&state);
	
	*ppParams = state.pParams;
	*ppParsers = state.pParsers;

	except_cleanup_pop(except_call); /* always free the default prop. dict. */
	except_cleanup_pop(except_call); /* always free the error msg if present */
}	

void gse_psrbldr_error(YYLTYPE* locp, struct gse_psrbldr_state* pState,
                       const char* sMsg){

   if(pState->sErr != NULL) free(pState->sErr);

   if(pState->sFile)
      pState->sErr = gse_string("%s, line %d: %s at token '%s'", 
      	                  pState->sFile, locp->first_line, sMsg,
                           gse_psrbldr_get_text(scanner));      
   else
      pState->sErr = gse_string("Line %d: %s at token '%s'", 
      	                  locp->first_line, sMsg,
                          gse_psrbldr_get_text(scanner));
}


