  /* Make a multi-thread safe lexer */
%option reentrant
  
  /* Have all non-static symbols start with gse_psrbldr_, allows for 
     safe linking with other flex generated code */
%option prefix="gse_psrbldr_"

  /* Add the argument yylval to the generated gse_psrbldr_lex function */
%option bison-bridge

  /* Add the argument yylloc to the generated gse_psrbldr_lex function */
%option bison-locations

  /* Use Flex's support for input line numbers */
%option yylineno

  /* Don't generate unused static functions */
%option nounput

  /* %option noyy_top_state */

  /* Extra C-code in the top of the generated lexer */
%{
#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

/* Include token codes from the parser */
#include <rpwgse/GseTypes.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseParser.h>

#include <rpwgse/GsePsrBldr.h>
#include <rpwgse/gse_psrbldr_int.h>

/* char* gse_psrbldr_get_text(void); */
char* strdup(const char* s);
int fileno(FILE* stream);

/* Define the extra type to be a pointer to a char buf read state structure */
#define YY_EXTRA_TYPE struct gse_psrbldr_state*

/* Define a user action to set the line no. each time a token is recognized */
#define YY_USER_ACTION yylloc->first_line = yylineno;

/* Define input to be string reading form yyextra pointer */
#define YY_INPUT(buf, result, max_size) \
{\
   size_t nRemainder = (yyextra->pStart + yyextra->length) - yyextra->pCur;\
   if(nRemainder <= 0){ \
      result = YY_NULL; \
   } \
   else{ \
      if(nRemainder <= max_size){ \
         memcpy(buf, yyextra->pCur, nRemainder); \
         yyextra->pCur += nRemainder; \
         result = nRemainder; \
      } \
      else{ \
         memcpy(buf, yyextra->pCur, max_size); \
         yyextra->pCur += max_size; \
         result = max_size; \
      } \
   } \
}
%}


  /* Name our patterns for easier readability */ 
str_ptrn   [a-zA-Z_][a-zA-Z0-9_\-]*
qstr_ptrn  \"[^"\n]*["\n]
int_ptrn   (-?[0-9]+)|(0x[0-9a-fA-F]+)
flt_ptrn   -?((([0-9]+)|([0-9]*\.[0-9]+))([eE][-+]?[0-9]+)?)

%x IN_COMMENT


%%
   	
<INITIAL>{
"/*"    BEGIN(IN_COMMENT);
}

<IN_COMMENT>{
"*/"      BEGIN(INITIAL);
[^*\n]+   /* eat comment in chunks */
"*"       /* eat the lone star */
"\n"      /* eat the newline */
}

{qstr_ptrn} { 
   yyget_lval(yyscanner)->sVal = strdup(yytext+1);  /* Skip the open quote */
   if(yyget_lval(yyscanner)->sVal[yyget_leng(yyscanner)-2] != '"')
      fprintf(stderr, "Line %d: Unterminated string: '%s'\n", yylineno, 
              yytext);
   else
      yyget_lval(yyscanner)->sVal[yyget_leng(yyscanner)-2] = '\0'; 
   return STRING;
} 

  /* Now for some keywords, must be before strings and adter comments */ 

parser { 
   return PARSER; 
}

default { 
   return DEFAULT; 
}

rmdefault {
   return RMDEFAULT;
}

parameters { 
   return PARAMETERS; 
}

  /* Generic strings */

{str_ptrn} { 
   yyget_lval(yyscanner)->sVal = strdup(yytext);
   return STRING;
}

{int_ptrn} {
   yyget_lval(yyscanner)->i8val = strtoll(yytext, NULL, 0);
   return INT;
}

{flt_ptrn} {
   yyget_lval(yyscanner)->f8val = atof(yytext);
   return FLOAT;
}


[[:space:]]  { /* Do nothing with spaces, newlines tabs or carrage returns */ }

. { 
   /* Send all other characters out as return values */ 
   return yytext[0];
}

%%

/* Replacement for yywrap as -Pgse_psrbldr_ makes this undefined */
int gse_psrbldr_wrap(yyscan_t yyscanner){
   return 1;
}

void gse_psrbldrlxr_beg(struct gse_psrbldr_state* pState, const char* sConfig)
{
   pState->pStart  = sConfig;
   pState->pCur    = sConfig;
   pState->length  = strlen(sConfig);
   pState->sErr    = NULL;
   
   yylex_init(&(pState->scanner));
   yyset_extra(pState, pState->scanner);
}

void gse_psrbldrlxr_end(struct gse_psrbldr_state* pState){
   yylex_destroy(pState->scanner);
   pState->scanner = NULL;
}

