#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <strings.h>

#include "GseExcept.h"
#include "GseWrap.h"
#include "GseRunAvg.h"

/*****************************************************************************/

void GseRunAvg_add(GseRunAvg* pThis, const float* pData){
	
	float* pA;
	float* pS;
	float fAverages;
	unsigned int u;
	
	assert(pThis->uSamples > 0);
	
	/* The N = 1 case */
	if(pThis->uSets == 1){
		if(pThis->pAvg == NULL)
			pThis->pAvg = (float*) gse_calloc( pThis->uSamples, sizeof(float));
		
		memcpy(pThis->pAvg, pData, sizeof(float) * pThis->uSamples);
		
		pThis->bNeedsCalc = 0;
		return;
	}
	
	
	/* The general case */
	
	pThis->pCur = pThis->pCur->pNext;
	pA = pThis->pAvg;

	if(pThis->pCur->pData == NULL){ /* same as uAverages < uSets */
		assert(pThis->uAverages < pThis->uSets);
		
		pThis->pCur->pData = (float*) gse_calloc(pThis->uSamples, sizeof(float));
	
		pThis->uAverages += 1;
		pS = pThis->pCur->pData;
		
		memcpy((void*)pS, (void*)pData, sizeof(float) * pThis->uSamples);
		
		fAverages = (float) pThis->uAverages;
		
		for(u = 0; u<pThis->uSamples; u++)
			pA[u] = ((fAverages - 1.0) * pA[u] + pS[u] ) / fAverages;
		
	}
	else{
		assert(pThis->uAverages == pThis->uSets);
		
		pS = pThis->pCur->pData;
		fAverages = (float) pThis->uSets;
		
		/* Subtract out the old averages, thier memory will be reused */
		for(u = 0; u < pThis->uSamples; u++)
			pA[u] -= pS[u] / fAverages;
		
		memcpy((void*)pS, (void*)pData, sizeof(float) * pThis->uSamples);
		
		/* Add in the new averages */
		for(u = 0; u < pThis->uSamples; u++)
			pA[u] += pS[u] / fAverages;
	}
	
	pThis->bNeedsCalc = 0;
}


float* GseRunAvg_getAddPtr(GseRunAvg* pThis){
	float* pA;
	float* pS;
	float fAverages;
	unsigned int u;
	
	assert(pThis->uSamples > 0);
	
	/* The N = 1 case */
	if(pThis->uSets == 1){
		pThis->bNeedsCalc = 0;
		return pThis->pAvg;
	}
	
	/* The general case */
	pThis->pCur = pThis->pCur->pNext;
	
	if(pThis->pCur->pData == NULL){ /* same as uAverages < uSets */
		assert(pThis->uAverages < pThis->uSets);
		
		pThis->pCur->pData = (float*) gse_calloc(pThis->uSamples, sizeof(float));		
	}
	else{
		assert(pThis->uAverages == pThis->uSets);
		
		/* first subtract out the old data */
		pA = pThis->pAvg;
		pS = pThis->pCur->pData;
		fAverages = (float) pThis->uSets;
		
		for(u = 0; u < pThis->uSamples; u++)
			pA[u] -= pS[u] / fAverages;
		
		/* Empty the array so the user will notice something amiss, if they
		   don't fill it */
		/* bzero((void*)pS, pThis->uSamples * sizeof(float)); */
	}
	
	pThis->bNeedsCalc = 1;
	return pThis->pCur->pData;
}


/****************************************************************************/
/* Getting averages */

const float* GseRunAvg_getAvg(GseRunAvg* pThis){
	float* pA;
	float* pS;
	float fAverages;
	unsigned int u; 
	
	
	if(! pThis->bNeedsCalc) 
		return (const float*) pThis->pAvg;
	
	
	/* The rest of the function is a follow on to getAddPtr, it completes it*/
	pA = pThis->pAvg;
	pS = pThis->pCur->pData;
	
	if(pThis->uAverages < pThis->uSets){
		pThis->uAverages += 1;
		fAverages = (float) pThis->uAverages;
				
		for(u = 0; u < pThis->uSamples; u++)
			pA[u] = ((fAverages - 1.0) * pA[u]  + pS[u]) / fAverages;
		
	}
	else{
		
		fAverages = (float) pThis->uSets;
		
		/* Add in the new averages */
		for(u = 0; u < pThis->uSamples; u++)
			pA[u] += pS[u] / fAverages;
	}
	
	pThis->bNeedsCalc = 0;
	return (const float*) pThis->pAvg;	
}


/****************************************************************************/
/* Destructor */


void _clearCircleBuf(CirBuf* pBuf){
	CirBuf* pFirst;
	CirBuf* pNext;
	
	if(pBuf == NULL) return;

	pFirst = pBuf;
	do{
		pNext = pBuf->pNext;
		if(pBuf->pData != NULL) free(pBuf->pData);
		pBuf->pData = NULL;
		pBuf = pNext;
	}
	while(pBuf != pFirst);
}

void _freeCircleBuf(CirBuf* pBuf){
	CirBuf* pFirst;
	CirBuf* pNext;
	
	if(pBuf == NULL) return;

	pFirst = pBuf;
	do{
		pNext = pBuf->pNext;
		if(pBuf->pData != NULL) free(pBuf->pData);
		free(pBuf);
		pBuf = pNext;
	}
	while(pBuf != pFirst);
	
}

void GseRunAvg_del(void* vpThis){
	GseRunAvg* pThis;
	pThis = (GseRunAvg*)vpThis;
	
	if(pThis->pAvg != NULL) free(pThis->pAvg);
	_freeCircleBuf(pThis->pCur);
	free(vpThis);
}

/****************************************************************************/
/* Constructors, init'ers */

CirBuf* _newCircleBuf(size_t uSets){
	size_t u;
	CirBuf* pCur = NULL;
	CirBuf* pFirst = NULL;
	CirBuf* pPrev = NULL;
	
	u = 0;
	do{
		
		pPrev = pCur;
		
		pCur = (CirBuf*) gse_malloc(sizeof(CirBuf));
		pCur->pData = NULL;
		pCur->pNext = NULL;
		
		if(pPrev != NULL) pPrev->pNext = pCur;
		
		
		if(pFirst == NULL) pFirst = pCur;
		u++;
	}while(u<uSets);
	
	pCur->pNext = pFirst;
	
	return pFirst;
}


void GseRunAvg_reInitSamples(GseRunAvg* pThis, size_t uSamples){
	
	if(pThis->pAvg != NULL){
		free(pThis->pAvg);
		pThis->pAvg = NULL;
	}
	
	pThis->uSamples = uSamples;
	if(pThis->uSamples > 0)
		pThis->pAvg = (float*) gse_calloc(uSamples, sizeof(float));
	
	_clearCircleBuf(pThis->pCur);
	
	pThis->uAverages = 0;
	pThis->bNeedsCalc = 0;
}

void GseRunAvg_reInitSets(GseRunAvg* pThis, size_t uSets){
	assert(uSets > 0);
	
	if(pThis->pAvg != NULL)
		bzero(pThis->pAvg, pThis->uSamples * sizeof(float));
	
	_freeCircleBuf(pThis->pCur);
	pThis->pCur = NULL;
	
	pThis->uSets = uSets;
	if(uSets > 1)
		pThis->pCur = _newCircleBuf(uSets);
	
	pThis->uAverages = 0;
	pThis->bNeedsCalc = 0;
}

GseRunAvg* new_GseRunAvg(size_t uSets, size_t uSamples, int nFlags,
                         size_t uMalloc)
{
	GseRunAvg* pThis;
	size_t uSize;
	
	if(uMalloc > sizeof(GseRunAvg))
		uSize = uMalloc;
	else
		uSize = sizeof(GseRunAvg);
	
	pThis = (GseRunAvg*)gse_malloc(uSize);
	pThis->del = &GseRunAvg_del;
	pThis->reInitSets = &GseRunAvg_reInitSets;
	pThis->reInitSamples = &GseRunAvg_reInitSamples;
	pThis->add = &GseRunAvg_add;
	pThis->getPtrForAdding = &GseRunAvg_getAddPtr;
	pThis->getAvg = &GseRunAvg_getAvg;
			
	pThis->uSets = uSets;
	assert(uSets > 0);
	pThis->uSamples = uSamples; /* Don't care if this is zero right now just
	                               needs to be defined before the 1st add */
	
	/* If N = 1, then there is no circle buf */
	if(uSets == 1)
		pThis->pCur = NULL;
	else
		pThis->pCur = _newCircleBuf(uSets);
	
	if(uSamples > 0)
		pThis->pAvg = (float*)calloc(uSamples, sizeof(float));
	else
		pThis->pAvg = NULL;
	
	pThis->uAverages = 0;
	pThis->bNeedsCalc = 0;
	
	if(pThis->uSets > 1 && (nFlags & RUN_AVG_CLEAR_ON_STEP))
		gse_throw_m(GSE_XCEPT, GSE_XNOTYETIMP, "RUN_AVG_CLEAR_ON_STEP: not yet "
				      "implemented.");
	
	pThis->bClearOnStep = 0;
	
	return pThis;
}
