/** @file GseRunAvg.h Utility class to help with running averages */

#ifndef GseRunAvg_H_
#define GseRunAvg_H_

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef struct circle_buf_item{
	struct circle_buf_item* pNext;
	float* pData;
} CirBuf;


#define RUN_AVG_CLEAR_ON_STEP 0x1

/** @addtogroup mod_simp_math */
/*@{*/

/** Structure to compute running averages with memory management @n
 * @class GseRunAvg
 */
typedef struct gse_run_avg {
	
	/** clear the structure for reuse, but don't resize */
	/* void (*clear)(struct gse_run_avg* pThis); */
	
	/** clear the structure for reuse and resize the internal buffers */
	void (*reInitSets)(struct gse_run_avg* pThis, size_t uSets);
	
	/** clear the structure for reuse and resize internal buffer */
	void (*reInitSamples)(struct gse_run_avg* pThis, size_t uSamples);
	
	/* clear the struture for reuse and resize all internal buffers */
	/* void (*reInitAll)(struct gse_run_avg* pThis, size_t uSets, size_t uSamples); */
	
	/** Add a sample to the average
	
	This will cause the last item in the circle to fall off
	
   Calculating running averages with no data drop outs:

<pre>
Given n = number of captures (from 1 to whatever)
      N = captures in running average
      A = the average
      c = the current sample


                     (n - 1)A     + c
                             n-1     n
 If n <= N:   A  =  -------------------
               n           n

                              c  -  c
                               n     n-N
 If n > N:    A  =  A      + --------------
               n      n-1         N
</pre>

  For the code below, A ~ pAvg,  c ~ pCur,  n ~ uAverages,  N ~ uSamples  */
	void (*add)(struct gse_run_avg* pThis, const float4* pData);
	
	/** Another way to add a sample to average, good for loops
	
	The assumption is that the resulting pointer will be used to
	store the same number of points specified via \a uSetSize in
	reInit or ::newRunAvg.  Example usage:
	\code
	GseRunAvg* pAvg;
	float* pInsert;
	
	pAvg = new_GseRunAvg(1, 256, 0);
	pInsert = (*(pAvg->getPtrForAdding))(pAvg);
	
	for(i = 0; i< 256; i++)
		pInsert[i] = // Some assignment; 
	
	\endcode
	*/
	float4* (*getPtrForAdding)(struct gse_run_avg* pThis);
	
	/** Get the average
	
	WARNING: this class still owns the pointer 
	
	@return a pointer to the average, or NULL if no data has been appended */
	const float4* (*getAvg)(struct gse_run_avg* pThis); 

	
	/** Destructor, frees memory */
	void (*del)(void* vpThis);
	
	/** Number of samples in a set, set via #reInitSamples and ::newRunAvg */
	size_t uSamples;
	
	/** Nuber of sets in an average, set via #reInitSets and ::newRunAvg */
	size_t uSets;
	
	/** Privte stuff, don't access directly */
	/** @cond PRIVATE */
	
	/* Current insertion point in a circular buffer */
	CirBuf* pCur;
	
	/* Currently calculated average */
	float4* pAvg;
	
	/** Number of caputres currently averaged together */
	size_t uAverages;
	
	/* Flag, avg needs calculation */
	int bNeedsCalc;
	
	/* weather to clear the average after each collection of uSets */
	int bClearOnStep;
	
	/** @endcond */

} GseRunAvg;

/** Generate a new running average structure 
 *
 * @param uSets number of data sets in the running average.  If this is 1 then
 *        no circular buffer is allocated and lookup via GseRunAvg::getAvg should incure
 *        little or no penalty.
 *
 * @param uSamples number of samples in a single data set.  All data sets to be
 *        averaged together need to have a constant sample size.
 *
 * @param nFlags set this to 0, no flags have been implemented.
 *
 * @param uMalloc is used by derived classes, always set to 0
 *
 * @return a new GseRunAvg structure.
 *
 * @public
 * @memberof GseRunAvg
 */
GseRunAvg* new_GseRunAvg(size_t uSets, size_t uSamples, int nFlags, 
                         size_t uMalloc);

/*@}*/


#ifdef __cplusplus
}
#endif

#endif /* _GseRunAvg_H_ */
