#define _POSIX_C_SOURCE 200112
#define _XOPEN_SOURCE 500

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "GseString.h"
#include "GseExcept.h"
#include "GseWrap.h"
#include "GseFileStr.h"
#include "GseLog.h"
#include "GseConv.h"

/****************************************************************************/
/* String functions */

char* gse_vstring(const char* fmt, va_list ap){
	/* Guess we need no more than 50 bytes. */
	int n, size = 50;
	char *p;
	va_list _ap;
	
	p = gse_malloc(size);
	
	except_cleanup_push(free, p);

	while (1) {
		/* Try to print in the allocated space. */
		va_copy(_ap, ap);
		n = vsnprintf (p, size, fmt, _ap);
		va_end(_ap);
		
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			break;
		
		/* Else try again with more space. */
		if (n > -1) 	/* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else  			/* glibc 2.0 */
			size *= 2;  /* twice the old size */
		
		p = gse_realloc(p, size);
	}
	
	except_cleanup_pop(except_no_call);
	return p;
}

char* gse_string(const char *fmt, ...){
	/* Guess we need no more than 50 bytes. */
	int n, size = 50;
	char *p;
	va_list ap;

	p = gse_malloc(size);
	
	except_cleanup_push(free, p);

	while (1) {
		/* Try to print in the allocated space. */
		va_start(ap, fmt);
		n = vsnprintf (p, size, fmt, ap);
		va_end(ap);
		
		/* If that worked, return the string. */
		if (n > -1 && n < size)
			break;
		
		/* Else try again with more space. */
		if (n > -1) 	/* glibc 2.1 */
			size = n+1; /* precisely what is needed */
		else  			/* glibc 2.0 */
			size *= 2;  /* twice the old size */
		
		p = gse_realloc(p, size);
	}
	
	except_cleanup_pop(except_no_call);
	return p;
}

char* gse_strcopy(const char* sSrc){
	char* sDest;
	
	if(sSrc == NULL) return NULL;
	sDest = (char*)(gse_calloc(strlen(sSrc) + 1, sizeof(char)));
	strcpy(sDest, sSrc);
	return sDest;
}

/* try to be efficient and use realloc */
char* gse_strcat(char* str, const char* suffix)
{
	char* sRet;
	size_t nStr, nSuffix;
	
	if((suffix == NULL)||(suffix[0] == '\0')) return str;	
	if(str == NULL) return gse_strcopy(suffix);
	
	nStr = strlen(str);
	nSuffix = strlen(suffix);
	sRet = (char*)gse_realloc(str, nStr + nSuffix + 1);
	
	memcpy(sRet + nStr, suffix, nSuffix + 1);
	return sRet; 
}

char* gse_strrep(const char* src, const char* find, const char* rep){
	const char *pSrc, *pFind;
	size_t nSrc, nFind;
	int nReplacements = 0;
	char* sRet = NULL;
	char* sBuf, *pBuf, *p;
	
	GseFileStr fileStr;

	if((src == NULL)||(find == NULL)||(rep == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL string supplied to gse_strrep");
	
	if((src[0] == '\0')||(find[0] == '\0')||(rep[0] == '\0'))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Empty string supplied to gse_strrep");
	
	nSrc = strlen(src);
	nFind = strlen(find);
	
	init_GseFileStr(&fileStr, nSrc + 11);
	except_cleanup_push(fin_GseFileStr, &fileStr);
	
	sBuf = gse_calloc(nFind + 1, sizeof(char));
	except_cleanup_push(free, sBuf);
	pBuf = sBuf;
	
	pFind = find;
	for(pSrc = src; pSrc < src + nSrc; pSrc++){
		
		if( *pSrc == *pFind){
			/* In the find string */
			
			if(pFind < (find + nFind - 1)){
				/* Not at end of find string */
				*pBuf = *pSrc;
				pBuf++;
				pFind++;
			}
			else{
				/* An the end of find string */
				nReplacements += 1;
				pFind = find;
				fileStr.putstr(&fileStr, rep);
				sBuf=pBuf;
			}
		}
		else{
			/* not in the find string */
			if(pBuf != sBuf){
				
				/* Flush the partial match buffer */
				for(p = sBuf; p < pBuf; p++)
					fileStr.putchar(&fileStr, *p);
				
				pBuf = sBuf;
				
				/* reset the find pointer */
				pFind = find;
			}
			
			fileStr.putchar(&fileStr, *pSrc);
		}
	}
	
	fileStr.putstr(&fileStr,sBuf);
	/* See if have conversions */
	if(nReplacements != 0) sRet = fileStr.takestr(&fileStr);
	
	except_cleanup_pop(except_call);     /* sBuf */
	except_cleanup_pop(except_no_call);  /* fileStr */
	return sRet;
}

bool gse_endswith(const char* str, const char* suffix)
{
	int i, nStr, nSuffix;
	
	if((str == NULL)||(str[0] == '\0')||(suffix == NULL)||(suffix[0] == '\0'))
		gse_throw_m(GSE_XCEPT,GSE_XBADARG,"Null or empty string in gse_endswith");
	
	nStr = strlen(str);
	nSuffix = strlen(suffix);
	
	if(nStr < nSuffix) return false; /* if str len < suffix len, can't match */
	
	for(i = nSuffix - 1; i >= 0; i--){
		if(str[nStr - (nSuffix - i)] != suffix[i]) 
			return false;
	}
	
	return true;
}

bool gse_endswith_n(const char* str, const char* suffix)
{
	int i, nStr, nSuffix;
	
	if((str == NULL)||(str[0] == '\0')||(suffix == NULL)||(suffix[0] == '\0'))
		gse_throw_m(GSE_XCEPT,GSE_XBADARG,"Null or empty string in gse_endswith_n");
	
	nStr = strlen(str);
	nSuffix = strlen(suffix);
	
	if(nStr < nSuffix) return false; /* if str len < suffix len, can't match */
	
	for(i = nSuffix - 1; i >= 0; i--){
		if(tolower(str[nStr - (nSuffix - i)]) != tolower(suffix[i])) 
			return false;
	}
	
	return true;
}

bool gse_isoneof(char c, const char* sList)
{
	size_t i;
	size_t n = strlen(sList);
	for(i = 0; i < n; i++)
		if(c == sList[i])
			return true;
	
	return false;
}

GseStrAry* gse_strsplit(const char* str, const char* toks)
{
	size_t i, nLen;
	char* _str;
	
	GseStrAry* pAry = new_GseStrAry(0);
	
	if(str == NULL)
		return pAry;
	
	_str = gse_strcopy(str);
	
	nLen = strlen(str);
	
	/* Turn all tokens into nulls */
	for(i = 0; i<nLen; i++) 
		if( gse_isoneof(_str[i], toks) )
			_str[i] = '\0';
	
	/* Iterate until finding the start of a word */
	i = 0;
	while(i<nLen){
		
		/* Advance past nulls */
		while((_str[i] == '\0')&&(i<nLen)) i++;
		
		/* Copy word or quit */
		if(i<nLen)
			pAry->add(pAry, gse_strcopy(_str+i));
		else
			break;
		
		/* Advance till null */
		while((_str[i] != '\0')&&(i<nLen)) i++;
	}
	
	free(_str);
	return pAry;
}

bool gse_isin(char c, const char* str)
{
	const char* p = str;
	while((*p != '\0') && (c != *p)) p++;
	
	return (c == *p); 
}

char* gse_strip(const char* str, const char* strip)
{
	size_t i, j, iStart, iEnd, nLen, nNewLen;
	char* sRet;
	
	if(str == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL string pointer in gse_strip");
	
	nLen = strlen(str);
	nNewLen = nLen;
	
	for(i = 0; i<nLen; i++){
		iStart = i;
		if( ! gse_isin(str[i], strip) ) break;
		nNewLen--;
	}
	
	if(nNewLen){
		for(i = nLen - 1; i >= 0; i--){
			iEnd = i;
			if( ! gse_isin(str[i], strip) ) break;
			nNewLen--;
		}
	}
	
	if(nNewLen < 1) return NULL;
	
	sRet = gse_calloc(nNewLen + 1, sizeof(char));
	
	j = 0;
	for(i = iStart; i <= iEnd; i++)
	{
		sRet[j] = str[i];
		j++;
	}
	
	return sRet;
}

void gse_toupper(char* str)
{
	size_t i, len;
	if(str == NULL) return;
	
	len = strlen(str);
	for(i = 0; i<len; i++) str[i] = toupper(str[i]);
}

void gse_tolower(char* str)
{
	size_t i, len;
	if(str == NULL) return;
	
	len = strlen(str);
	for(i = 0; i<len; i++) str[i] = tolower(str[i]);
}

char* gse_strip_cmnt(const char* sLine, char cTok){
	
	const char* pEnd;
	const char* pStart;
	char* pOut;
	int nStrLen, i;
	
	if(sLine == NULL) return NULL;
	
	pEnd = sLine;
	while((*pEnd != cTok)&&(*pEnd != '\n')&&(*pEnd != '\r')&&(*pEnd != '\0'))
		pEnd++;
	
	pEnd--; /* Pointing at character before the stop char */
	if(pEnd < sLine) return NULL;
	
	/* Now ignore all spaces at end of line */
	while((pEnd > sLine)&& isspace(*pEnd) ) pEnd--;
		
	pStart = sLine;
	while((pStart <= pEnd) && isspace(*pStart) ) pStart++;
		
	if(pStart >= pEnd) return NULL;
	
	nStrLen = pEnd - pStart + 1;
	pOut = (char*)(gse_malloc(nStrLen + 1));
	
	for(i=0; i<nStrLen; i++) pOut[i] = pStart[i];
	pOut[nStrLen] = '\0';
	
	return pOut;
}

/*****************************************************************************/
/* String array instantiation */

#define C GseStrAry
#define T char
#define CASCADE_FREE

#include "ptrAryTplt.c"

#undef CASCADE_FREE
#undef C
#undef T

/*****************************************************************************/
/* Constant string array instantiation, doesn't free jack */

#define C GseCStrAry
#define T const char

#include "ptrAryTplt.c"

#undef C
#undef T

/*****************************************************************************/
/* String list instantiation */

#define C GseStrList
#define I char
#define CASCADE_FREE

#include "listTplt.c"

#undef CASCADE_FREE
#undef I
#undef C


/*****************************************************************************/
/* printing hex bytes */
static const char* gse_nibble_chars = "0123456789ABCDEF"; 

char* gse_hexprn(const byte* pBytes, size_t nIn, char* sDest, size_t nOut)
{
	size_t i, nLastIn;
	
	if((nOut < 3)||(pBytes == NULL)||(nIn < 1)||(sDest == NULL))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	nLastIn = nOut/3;
	
	for(i = 0; i<nLastIn; i++){
		 
		sDest[i*3]     = gse_nibble_chars[ (pBytes[i] >> 4) & 0xF ];
		sDest[i*3 + 1] = gse_nibble_chars[ pBytes[i] & 0xF ];
				
		if(i < nLastIn - 1)
			sDest[i*3 + 2] = ' ';
		else
			sDest[i*3 + 2] = '\0';
	}
	
	return sDest;
}

char* gse_dyn_hexprn(cbytebuf_t cbuf)
{
	char* sRet;
	size_t i;
	
	sRet = gse_calloc(cbuf.len * 3, sizeof(char));
	
	for(i = 0; i<cbuf.len; i++){
		
		sRet[i*3]      = gse_nibble_chars[ (cbuf.ptr[i] >> 4) & 0xF ];
		sRet[i*3 + 1] = gse_nibble_chars[ cbuf.ptr[i] & 0xF ];
		
		if(i < cbuf.len - 1)
			sRet[i*3 + 2] = ' ';
		else
			sRet[i*3 + 2] = '\0';
	}
	
	return sRet;
}

void gse_fmt_hexprn(FILE* stream, cbytebuf_t buf, uint4 indent, uint4 group)
{
	size_t i, nGrpPerLine, nBytesPerLine;
	char sCh[3], sIndent[21];

	if((indent < 0)||(indent > 20))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "indent was %d, expected a value"
		             "between 0 and 20", indent);
	
	if((group < 1)||(group > (80 - indent)))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "group size was %d, expected a "
		             "value between 1 and %d (80 - indent) inclusive.", 
		             group, 80 - indent);
	
	sCh[2] = '\0';
	sIndent[indent] = '\0'; 
	for(i=0; i<indent; i++) sIndent[i] = ' ';
	
	nGrpPerLine = (81 - indent) / (group*2 + 1);
	nBytesPerLine = nGrpPerLine * group;
	
	for(i = 0; i < buf.len; i++){
		
		if(i % nBytesPerLine == 0){
			
			if(i != 0)        fputc('\n', stream);
			if(indent > 0)    fputs(sIndent, stream);
		}
		else{
			if(i % group == 0) fprintf(stream, " ");
		}
		
		sCh[0] = gse_nibble_chars[ (buf.ptr[i] >> 4) & 0xF ];
		sCh[1] = gse_nibble_chars[ buf.ptr[i] & 0xF ];
		
		fputs(sCh, stream);
	}
	
	fputc('\n', stream);
}   

/*****************************************************************************/
/* CSV file helpers, move to thier own file as they become more developed    */

int gse_csv_fields(const char* sLine)
{
	const char* pChar = NULL;
	bool bInQuotes = false;
	int4 nFields = 0;
	
	if(sLine == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null input string");
		
	/* Allow \r \n or \0 to terminate the last field */
	while(true){
		
		if(pChar == NULL) 
			pChar = sLine;
		else
			pChar++;
		
		if(*pChar == '"'){
			bInQuotes = ! bInQuotes;
			continue;
		}
		
		if(*pChar == '\0'){
			if(bInQuotes){
				gse_warning("Unclosed quote in CSV file text");
			}
			nFields += 1;
			break;
		}
		
		if((*pChar == '\r' || *pChar == '\n') && !bInQuotes){
			nFields += 1;
			break;
		}
		
		if((*pChar == ',') && !bInQuotes)
			nFields += 1;
	}
	
	return nFields;
}


void _gse_csv_getbounds(const char* sLine, int4 iCol, int4* pStart, 
                        int4* pTo)
{
	int4 i;
	int4 iStart;
	int4 iTo;
	
	if(sLine == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null input string");
	
	/* Find the boundaries of the requested field */
	
	/* Allow \r \n or \0 to terminate the last field */
	
	/* iStart points to the start of the field, iTo points to the 
	   terminating character, if the first character encountered
		is a terminating character, iStart = iTo */
	
	bool bInQuotes = false;
	bool bBreak = false;
	int4 iCurField = -1;
	const char* pChar = NULL;
	
	while(!bBreak){
		
		if(pChar == NULL){ 
			i = 0;
			pChar = sLine;
			iStart = 0;
			iTo =  -1;  /* -1 in iTo is a flag that the terminator
			               has not been found */
		}
		else{
			i++;
			pChar++;
		}
		
		if(*pChar == '"'){
			bInQuotes = ! bInQuotes;
			continue;
		}
		
		if(*pChar == '\0'){
			if(bInQuotes){
				gse_warning("Unclosed quote in CSV file text");
			}
			iTo = i;
			iCurField++;
			bBreak = true;
		}
		
		if((*pChar == '\r' || *pChar == '\n') && !bInQuotes){
			iCurField++;
			iTo = i;
			bBreak = true;
		}
		
		if((*pChar == ',') && !bInQuotes){
			iCurField++;
			iTo = i;
		}
		
		/* If I've found a terminator, quit if this is the desired 
		   field or setup for the next on */
		if( iTo != -1){
			
			if(iCurField == iCol){
				bBreak = true;
			}
			else{
				iStart = i + 1;
				iTo = -1;
			}
		}
	}
	
	/* Did I stop because I got the field I needed? Or for another reason */
	if(iCurField != iCol)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG,"Field index %d not present in "
		             "CSV row.", iCol);
	
	*pStart = iStart;
	*pTo = iTo;
}


char* gse_csv_getField(const char* sLine, int4 iCol)
{
	int4 i;
	int4 iStart;
	int4 iTo;
	char* sField = NULL;
	const char* pChar;
	
	_gse_csv_getbounds(sLine, iCol, &iStart, &iTo);
	
	if(iStart == iTo) return NULL;
	
	sField = gse_calloc(iTo - iStart + 1, sizeof(char));
	pChar = sLine + iStart;
	
	for(i = 0; i < iTo - iStart; i++){	
		sField[i] = *pChar;
		pChar++;
	}
	sField[i] = '\0';
	return sField;
}
	

char* gse_csv_getField_s(const char* sLine, int4 iCol, char* sField,
		int4 nLen)
{
	int4 iStart, iTo, i;
	const char* pChar;
	
	_gse_csv_getbounds(sLine, iCol, &iStart, &iTo);
	
	if(iStart == iTo) return NULL;
	
	if(iTo - iStart + 1 > nLen ){
		gse_error_v("Buffer is too small to hold field number %d, must be "
		            "at least %d bytes long", iCol, iTo - iStart + 1);
		return NULL;		            
	}
		
	pChar = sLine + iStart;
	
	for(i = 0; i < iTo - iStart; i++){	
		sField[i] = *pChar;
		pChar++;
	}
	sField[i] = '\0';
	return sField;	
}

float8 gse_csv_getFloat(const char* sLine, int4 iCol, char* sSrc, int4 iRow){
	
	char sField[128];
	float8 dVal;
	
	if((sSrc != NULL)&&(iRow > -1)){
	
		if(gse_csv_getField_s(sLine, iCol, sField, 127) == NULL)
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE, "%s, Row %d, Col %d: Field is too "
			             "long to be a floating point number", sSrc, iRow, iCol);
				
		if(!gse_str2float8(sField, &dVal))
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s, Row %d, Col %d: Error converting "
			             "'%s' to a floating point number.", sSrc, iRow, iCol, sField);
	}
	else{
		if(gse_csv_getField_s(sLine, iCol, sField, 127) == NULL)
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE, "Field %d: Field is too "
			             "long to be a floating point number", iCol);
				
		if(!gse_str2float8(sField, &dVal))
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"Field %d: Error converting "
			             "'%s' to a floating point number.", iCol, sField);
	}
	
	return dVal;
}

int8 gse_csv_getInt(const char* sLine, int4 iCol, char* sSrc, int4 iRow){
	
	char sField[128];
	int8 nVal;
	
	if((sSrc != NULL)&&(iRow > -1)){
	
		if(gse_csv_getField_s(sLine, iCol, sField, 127) == NULL)
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE, "%s, Row %d, Col %d: Field is too "
			             "long to be an integer", sSrc, iRow, iCol);
				
		if(!gse_str2int8(sField, &nVal))
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s, Row %d, Col %d: Error converting "
			             "'%s' to a floating point number.", sSrc, iRow, iCol, sField);
	}
	else{
		if(gse_csv_getField_s(sLine, iCol, sField, 127) == NULL)
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE, "Field %d: Field is too "
			             "long to be a floating point number", iCol);
				
		if(!gse_str2int8(sField, &nVal))
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"Field %d: Error converting "
			             "'%s' to a floating point number.", iCol, sField);
	}
	
	return nVal;
}











