/** @file GseString.h String handling utilities */
#ifndef _GseString_H_
#define _GseString_H_

#include <stdio.h>
#include <stdarg.h>
#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif
/** @addtogroup mod_rpwgse */
/*@{*/

/** @defgroup mod_string Strings
 * 
 * yet another set of string handling utilities
 */
/*@{*/

/** Allocate a new string on the heap and format it
 *
 * Except for using GSE malloc and realloc wrappers this is a copy of the
 * code out of man 3 printf on Linux.
 *
 * @return a string pointer suitable for use in free().
 */
char* gse_string(const char* fmt, ...); 



/** Allocate a new string on the heap and format it
 *
 * This version is suitable for calling from variable argument functions.
 * Except for using GSE malloc and realloc wrappers this is a copy of the
 * code out of man 3 printf on Linux.
 *
 * @return a string pointer suitable for use in free().
 */
char* gse_vstring(const char* fmt, va_list ap);

/** Allocate a new string on the heap and copy an existing string to it
 *
 * @return a string pointer suitable for use in free()
 */
char* gse_strcopy(const char* sSrc);

/** Append to an existing string, may cause a calloc and free
 *
 * @param str the string to append to, need not have room for then new data
 * @param suffix the new string to add to the old
 * @return the new string pointer, may be the same as the old, may not
 */
char* gse_strcat(char* str, const char* suffix);


/** Find and replace substrings.
 *
 * @param src the string to search
 * @param find the string to find in @a search
 * @param rep the new string to insert in place of @a find
 * 
 * @return Returns a new string allocated on the heap with all occurances of 
 *         @a find in @a src replaced by @a rep, or NULL if @a find was
 *         not present in @a src.
 */
char* gse_strrep(const char* src, const char* find, const char* rep);	


/** Match string ending pattern
 *
 * @param str the string to search
 * @param suffix ending pattern to match
 * @return true if @a str ends with all the characters in suffix
 */
bool gse_endswith(const char* str, const char* suffix);

/** Match string ending pattern, without regards to captilization.
 *
 * @param str the string to search
 * @param suffix ending pattern to case insensitive match 
 * @return true if @a str ends with all the characters in suffix
 */
bool gse_endswith_n(const char* str, const char* suffix);


/* Var String array */
#define S gse_str_ary
#define C GseStrAry
#define T char

#include <rpwgse/ptrAryTplt.h>

#undef S
#undef C
#undef T

/* Const string array */
#define S gse_cstr_ary
#define C GseCStrAry
#define T const char

#include <rpwgse/ptrAryTplt.h>

#undef S
#undef C
#undef T

/* Var string List */
#define S gse_str_list
#define C GseStrList
#define I char

#include <rpwgse/listTplt.h>

#undef I
#undef C
#undef S


/** Short form wrapper for gse_string */
#define NS(F,...) gse_string(F,  __VA_ARGS__)

/** Short form wrapper for gse_strcopy */
#define CS(S) gse_strcopy(S)

/** Split a string on character boundaries.
 *
 * @warning This function merges multiple split tokens into one, thus
 * it is not suitable for reading things like CSV (comma seperated values)
 * files, were adjacent split tokens are taken to mean that an empty value
 * exists between the two tokens.  On the other hand it's great for path
 * reductions as // becomes a single splitter.
 *
 * @param str the string to split into an array of strings
 * @param toks The list characters on which to split
 * @return An array of strings allocated on the heap
 */
GseStrAry* gse_strsplit(const char* str, const char* toks);


/** Strip characters from the left and right of a string.
 *
 * @b Note Characters to remove are considered individually, not as a substring.
 *       Thus specifing a @a strip string of @c "aaa" produces the same
 *       result as a @a strip string of @c "a".
 *
 * @param str the string to strip
 * @param strip the characters to remove from the left and right.
 * @return a new string allocated on the heap, even if there was
 *         nothing to remove.  If the entire input string consists of
 *         characters in @a strip, NULL is returned.
 */
char* gse_strip(const char* str, const char* strip);

/** Convert a string in-place to upper case 
 * @param str the string to convert
 * @throws nothing this function does not throw
 */
void gse_toupper(char* str);

/** Convert a string in place to lower case
 * @param str the string to convert.
 * @throws nothing this function does not throw
 */
void gse_tolower(char* str);

/** Strip whitespace and remove shell style comments.
 *
 * @param sLine The string to strip.
 * @param cTok The comment start character, for shell style comments use '#'.
 *        cTok and all text to the right of cTok is removed.
 *
 * @return A new string allocated an the heap, or NULL if the input string
 *         contained only whitespace and comments.
 */
char* gse_strip_cmnt(const char* sLine, char cTok);

/** Get a count of the fields is a line of CSV file text
 *
 * In a CSV file, quotes are escaped via putting two of them together,
 * and commas that appear within fields are not considered as field
 * seperators.  Given these rules count the number of fields (columns)
 * represented in a string of text.
 *
 * @param sLine The text to parse, using the rules for comma seperated
 *        values fields
 * 
 * @return The number of csv fields encountered, @b Note: it is legal
 *         to have empty fields in a CSV file, so the mimimum
 *         return value is 1.
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) if the input string is NULL
 */
int gse_csv_fields(const char* sLine);


/** Get a particular field from a line of CSV file text
 *
 * In a CSV file, quotes are escaped via putting two of them together,
 * and commas that appear within fields are not considered as field
 * seperators.  Given these rules return the requested field from a
 * line of text.
 *
 * @param sLine The line of CSV file text.
 * 
 * @param iCol The 0 based index of the field to return.
 * 
 * @return a new string allocated on the heap containing the field contents.
 *         @b Note quotes are not removed from text fields.  If the field 
 *         exists, but is empty, then NULL is returned.
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) if @a iField does not exist.
 */
char* gse_csv_getField(const char* sLine, int4 iCol);


/** Get a particular field as a floating point number from a line of CSV
 * text.
 *
 * @param sLine The line of CSV file text.
 * 
 * @param iCol The 0 based index of the field to return.
 *
 * @param sSrc an optional data source name, such as a file, if non-null it
 *        will be used in error reporting.
 *
 * @param iRow an optional data source index such as a line number.  If
 *        sSrc is not null it will be used in error reporting.
 * 
 * @return the value if it was convertable, else throws
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) if @a iField does not exist.
 */
float8 gse_csv_getFloat(const char* sLine, int4 iCol, char* sSrc, int4 iRow);

/** Get a particular field as a integer from a line of CSV text.
 *
 * @param sLine The line of CSV file text.
 * 
 * @param iCol The 0 based index of the field to return.
 *
 * @param sSrc an optional data source name, such as a file, if non-null it
 *        will be used in error reporting.
 *
 * @param iRow an optional data source index such as a line number.  If
 *        sSrc is not null it will be used in error reporting.
 * 
 * @return the value if it was convertable, else throws
 *
 * @throws (GSE_XCEPT,GSE_XBADARG) if @a iField does not exist.
 */
int8 gse_csv_getInt(const char* sLine, int4 iCol, char* sSrc, int4 iRow);


/** Get a particular field from a line of CSV file text
 * Static buffer version of gse_csv_getField.
 */
char* gse_csv_getField_s(const char* sLine, int4 iField, char* sField,
		                   int4 nLen);


/** Print a string of bytes as as ASCII hex pattern. 
 *
 * @param pBytes the byte string to convert.
 * @param nIn how much of the byte string to convert.
 * @param sDest the output buffer.
 * @param nOut the size of the output buffer.  This needs to be 
 *        three times nIn to convert all the input bytes.
 *
 * @return a hex string representation of bytes in sDest.  The string is
 * always null terminated unless lenDest is 0.  The return val is just a 
 * pointer to sDest.
 */
char* gse_hexprn(const byte* pBytes, size_t nIn, char* sDest, size_t nOut);

/** Print a string of bytes as as ASCII hex pattern, dynamic version.
 *
 * @param cbuf The bytes to convert, plus thier length.
 *
 * @return a hex string representation of bytes in cbuf allocated on the
 *         heap.  The string is always null terminated.
 */
char* gse_dyn_hexprn(cbytebuf_t cbuf);

/** Pnint a byte buffer as as multiline ASCII hex pattern.
 *
 * The output lines wrap at 80 columns and wrapping never occurs in the
 * middle of a group.
 *
 * @param stream output file pointer.
 * @param buf the buffer to print.
 * @param indent the number of spaces to indent each line.
 * @param group the number of bytes to represent as ASCII befor printing a
 *    space seperator.
 */
void gse_fmt_hexprn(FILE* stream, cbytebuf_t buf, uint4 indent, uint4 group);

/*@}*/
/*@}*/
#ifdef __cplusplus
}
#endif

#endif /* _GseString_H_ */
