/** @file GseSys.h Process Handling */
#ifndef _GseSys_H_
#define _GseSys_H_

#include <rpwgse/GseTypes.h>
#include <stdio.h>

#ifdef __cplusplus
extern "C" {
#endif


/** @addtogroup mod_rpwgse */
/*@{*/
/** @defgroup mod_sys System Services 
 *
 * Cross platform system functions.
 *
 * This module currently supports getting system time, running sub-processes,
 * sleeping, printing and getting user account information the on POSIX and
 * Win32 based systems. 
 */
/*@}*/

/** @addtogroup mod_sys */
/*@{*/


/** Get the current account's username, not the login name.
 * @return The username in a string allocated on the heap.
 */
char* gse_username(void);

/** Get the current account's username, not the login name, static memory
 *  version.
 * @param buf The buffer to hold the username, at most @a len - 1 bytes of
 *        the user name are written so that the string is always null
 *        terminated.
 * @param len the length of buf.
 * @return the @a buf pointer.
 */
char* gse_username_s(char* buf, size_t len);

#ifdef _WIN32
typedef HANDLE procid_t;
#else
typedef pid_t procid_t;
#endif 

/** Open a subprocess and set FILE* pointers to it's inputs and outputs.
 *
 * @warning This function will deadlock simple single threaded programs
 *          unless all but one of @a ppStdin, @a ppStdout, and @a ppStderr are 
 *          NULL.
 *
 * @param sCmd The command to execute, this command is started as a 
 *        child of "/bin/sh" on UNIX, or run using "cmd.exe /C" on windows.
 *
 * @param ppStdin A pointer to a FILE* object which will be set in write mode
 *        to the subprocess' standard input.  If NULL standard in for the
 *        subprocess will closed down.
 * 
 * @param ppStdout A pointer to a FILE* object which will be set in read mode
 *        to the subprocess' standard output.  If NULL standard out for the
 *        subprocess will be inherited from the parent process.
 * 
 * @param ppStderr A pointer to a FILE* object which will be set in read mode
 *        to the subprocess' standard error.  If NULL standard error for the
 *        subprocess will be inherited from the parent process.
 *
 * @return The process pid on UNIX and a process HANDLE on Windows.
 */
procid_t gse_sopen(const char* sCmd, FILE** ppStdin, FILE** ppStdout, 
                   FILE** ppStderr);


/** Get a temporary filename.
 * Make a temporary filename that does not exist at the time the function
 * is called.  This does not actually create a file, just returns a 
 * suitable name.
 *
 * @note  The filename contains both the pid and the thread id, so it 
 *        shouldn't collide with other programs and threads.
 *
 * @note Don't use this function for high security applications, use
 *       system rountines such as mktemp(3) instead.
 *
 * @param sSuffix if not NULL this string will be appended to the file
 *        name, if not '.tmp' will be appended.  Use a '.' as the first
 *        character of sSuffix if you specify one.
 *
 * @return a string allocated on the heap suitable for use a filename.
 */
char* gse_tempname(const char* sSuffix);


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseSys_H_ */
