#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#ifdef WIN32
#error Win32 version of GseSys.c is not yet implemented
#else
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>
#endif

#include <math.h>
#include <ctype.h>

#include "GseExcept.h"
#include "GseLog.h"
#include "GseString.h"
#include "GseWrap.h"

#define _GseTime_C_
#include "GseTime.h"

extern long timezone;
extern int daylight;
extern char *tzname[2];
		 
/*****************************************************************************/

#ifndef WIN32

int gse_millisleep(int millisec, bool cont)
{
	struct timespec ts;
	struct timespec tr;
	int nRet;
	
	if(millisec < 1)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "millsec < 1 in gse_millisleep");
	
	ts.tv_sec = millisec / 1000;
	ts.tv_nsec = (millisec - (ts.tv_sec*1000))*1000000;
	
	if(!cont){
		
		if( nanosleep(&ts, &tr) == 0){
			nRet = 0;
		}
		else{
			if( errno == EINTR)
				nRet = tr.tv_sec * 1000 + (tr.tv_nsec / 1000000); 
			else
				gse_throw_m(CLIB_XCEPT, errno, "Error in nanosleep call");
		}
	}
	else{
		
		while( nanosleep(&ts, &tr) != 0){
			if(errno == EINTR){
				ts.tv_sec = tr.tv_sec;
				ts.tv_nsec = tr.tv_nsec;
			}
			else{
				gse_throw_m(CLIB_XCEPT, errno, "Error in nanosleep call");
			}
		}
		nRet = 0;
	}
	
	return nRet;
} 



#else
#error TODO: Windows versions of GseTime functions are not yet implemented
#endif

/*****************************************************************************/
/* Getting time strings */

/* Thread local storage for gmtime on Linux, this is automatic on Windows */
#if defined(GSE_POSIX_THREADS) && ! defined(WIN32)
static pthread_mutex_t s_sys_mtx_init = PTHREAD_MUTEX_INITIALIZER;
static int s_sys_nThreads = 0;
static pthread_key_t s_keyGmTime;
static pthread_key_t s_keyLocalTime;
#endif

void gse_time_init()
{
#if defined(GSE_POSIX_THREADS) && ! defined(WIN32)
	bool bLocalOk, bGmOk;
	
	pthread_mutex_lock(&s_sys_mtx_init);
	
	tzset();
	
	s_sys_nThreads += 1;
	
	if(s_sys_nThreads == 1){
		bLocalOk = (pthread_key_create(&s_keyLocalTime, free) == 0);
		bGmOk    = (pthread_key_create(&s_keyGmTime, free) == 0);
		
		if(! (bGmOk && bLocalOk)){
			s_sys_nThreads -= 1;
			if(bLocalOk) pthread_key_delete(s_keyLocalTime);
			if(bGmOk)    pthread_key_delete(s_keyGmTime);
			gse_throw_m(GSE_XCEPT, GSE_XINITFAIL, "GseSys module thread local "
			            "storage key create error"); 
		}
	}
	pthread_mutex_unlock(&s_sys_mtx_init);
#endif
	return;	
}

void gse_time_deinit(){
#if defined(GSE_POSIX_THREADS) && ! defined(WIN32)
	pthread_mutex_lock(&s_sys_mtx_init);
	
	if(s_sys_nThreads > 0){
		s_sys_nThreads -= 1;
		
		if(s_sys_nThreads == 0){
			pthread_key_delete(s_keyLocalTime);
			pthread_key_delete(s_keyGmTime);
		}
	}
	
	pthread_mutex_unlock(&s_sys_mtx_init);
	
#endif	
	return;
}

struct tm* gse_gmtime(const time_t* pTime)
{
#if defined(GSE_POSIX_THREADS) && ! defined(WIN32)
	struct tm* pTm = (struct tm*)pthread_getspecific(s_keyGmTime);
	if(pTm == NULL){
		pTm = (struct tm*)gse_calloc(1, sizeof(struct tm));
		pthread_setspecific(s_keyGmTime, (void*)pTm);
	}
	return gmtime_r(pTime, pTm);
#else
	return gmtime(pTime);
#endif
}

struct tm* gse_localtime(const time_t* pTime)
{
#if defined(GSE_POSIX_THREADS) && ! defined(WIN32)
	struct tm* pTm = (struct tm*)pthread_getspecific(s_keyLocalTime);
	if(pTm == NULL){
		pTm = (struct tm*)gse_calloc(1, sizeof(struct tm));
		pthread_setspecific(s_keyLocalTime, (void*)pTm);
	}
	return localtime_r(pTime, pTm);
#else
	return localtime(pTime);
#endif
}

char* _gse_tv2iso_helper(struct tm* p, int4 nMicro, int prec)
{
	if((prec < GSE_YEARS)||(prec > GSE_MICROSECS))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "prec: %d, is outside the valid "
		             "range 1-8 (years to microseconds)", prec);
	
	if((nMicro < 0)||(nMicro > 999999))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "microsecs: %d, is outside the "
		             "valid range 0 to 999,999", nMicro);
	
	switch(prec){
	case GSE_YEARS:
		return gse_string("%04d", p->tm_year+1900);
	case GSE_MONTHS:
		return gse_string("%04d-%02d", p->tm_year+1900, p->tm_mon+1);
	case GSE_DAYS:
		return gse_string("%04d-%02d-%02d", p->tm_year+1900, p->tm_mon+1, 
		                  p->tm_mday);
	case GSE_HOURS:
		return gse_string("%04d-%02d-%02dT%02d", p->tm_year+1900,
		                  p->tm_mon+1, p->tm_mday, p->tm_hour);
	case GSE_MINS:
		return gse_string("%04d-%02d-%02dT%02d:%02d", p->tm_year+1900,
		                  p->tm_mon+1, p->tm_mday, p->tm_hour, p->tm_min );
	case GSE_SECS:
		return gse_string("%04d-%02d-%02dT%02d:%02d:%02d", p->tm_year+1900,
		                  p->tm_mon+1, p->tm_mday, p->tm_hour, p->tm_min, 
		                  p->tm_sec);
	case GSE_MILLISECS:
		return gse_string("%04d-%02d-%02dT%02d:%02d:%02d.%03d", p->tm_year+1900,
		                  p->tm_mon+1, p->tm_mday, p->tm_hour, p->tm_min, 
		                  p->tm_sec, nMicro/1000);
	}
	
	return gse_string("%04d-%02d-%02dT%02d:%02d:%02d.%06d", p->tm_year+1900,
	                  p->tm_mon+1, p->tm_mday, p->tm_hour, p->tm_min, 
	                  p->tm_sec, nMicro);
}

char* gse_tv2iso_gm(struct timeval* pTv, int prec)
{	
	if(pTv == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL struct timeval* arg in "
		            "gse_tv2iso_gm");
	
	return _gse_tv2iso_helper(gse_gmtime(&(pTv->tv_sec)), pTv->tv_usec, prec);
}

char* gse_tv2iso_local(struct timeval* pTv, int prec)
{	
	if(pTv == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL struct timeval* arg in "
		            "gse_tv2iso_gm");
	
	return _gse_tv2iso_helper(gse_localtime(&(pTv->tv_sec)), pTv->tv_usec, prec);
}

char* gse_tv2iso_gm_s(struct timeval* pTv, int prec, char* buf, int len)
{
	char* sTmp;
	
	if(buf == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL char* arg in gse_tv2iso_gm_s");
	if(len < 5)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "buf len < 5 in gse_tv2iso_gm_s");
	
	sTmp = gse_tv2iso_gm(pTv, prec);
	strncpy(buf, sTmp, len - 1);
	buf[len-1] = '\0';
	free(sTmp);
	
	return buf;	
}

char* gse_tv2iso_local_s(struct timeval* pTv, int prec, char* buf, int len)
{
	char* sTmp;
	
	if(buf == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "NULL char* arg in gse_tv2iso_gm_s");
	if(len < 5)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "buf len < 5 in gse_tv2iso_gm_s");
	
	sTmp = gse_tv2iso_local(pTv, prec);
	strncpy(buf, sTmp, len - 1);
	buf[len-1] = '\0';
	free(sTmp);
	
	return buf;	
}

/*****************************************************************************/
/* Sleeping functions, used to be in ppd source area */

void gse_tv_calcwake(float fSecFromNow, struct timeval* pWake){
	struct timeval tvNow;
	unsigned int uAddSec = 0;
	
/*	fprintf(stderr, "Requested a wakeup time of %.3e seconds from now\n",
			  fSecFromNow); */
	
	gettimeofday(&tvNow, NULL);
/*	fprintf(stderr, "Current time is: %u.%06u\n", 
			  (unsigned int)tvNow.tv_sec, (unsigned int)tvNow.tv_usec); */
	
	pWake->tv_usec = tvNow.tv_usec + (fSecFromNow - floorf(fSecFromNow))*1000000;
	if(pWake->tv_usec > 1000000){
		uAddSec = 1;
		pWake->tv_usec -= 1000000;
	}
	else
		uAddSec = 0;
	
	pWake->tv_sec = tvNow.tv_sec + (unsigned int)floorf(fSecFromNow) + uAddSec;

/*	fprintf(stderr, "calcWakeup: Wake up %.3e seconds from now\n", tvAminusB(pWake, &tvNow)); */
}

int gse_tv_sleepuntil(struct timeval* pWake, int* pbRun){
	/* Sleep until pWake, if awaken by a signal check to see if *bRun has
	   value 0, if not go back to sleeping */
	struct timeval tvNow;
	struct timespec tsReq; /* requires sleep */
	struct timespec tsRem; /* remaining sleep */
	long int nMinusSec = 0;
	long int nTmp;
	
	
	/* Is it already time? */
	gettimeofday(&tvNow, NULL);
	if( gse_tv_cmp(pWake, &tvNow) <= 0)
		return 0;
	
	/* calc sleep needed */
		
	nTmp 	= ((long int)pWake->tv_usec) - ((long int)tvNow.tv_usec);
	nTmp *= 1000;
	
	if(nTmp < 0){
		nTmp += 1000000000;
		nMinusSec = -1;
	}
	assert(nTmp >= 0);
	tsReq.tv_nsec = nTmp;
	
	
	nTmp = ((long int)pWake->tv_sec) - ((long int)tvNow.tv_sec) + nMinusSec;
	tsReq.tv_sec = (time_t)nTmp;
	
	/* fprintf(stderr, "sleepUntil: Require %u.%09u sec of sleep\n",
	  	    (unsigned int)tsReq.tv_sec, (unsigned int)tsReq.tv_nsec); */
	
	/* Start sleeping */
	while(nanosleep(&tsReq, &tsRem) != 0){
		assert(errno == EINTR);
		if(*pbRun == 0){
			return -1;
		}
		else{
			tsReq.tv_nsec = tsRem.tv_nsec;
			tsReq.tv_sec = tsRem.tv_sec;
			/* It was nothing honey, go back to sleep... */
		}
	}
	
	return 0;
}

int gse_tv_cmp(const struct timeval* pOne, const struct timeval* pTwo){
	if(pOne->tv_sec > pTwo->tv_sec) return 1;
	else{
		if(pOne->tv_sec < pTwo->tv_sec) return -1;
		else{
			if(pOne->tv_usec > pTwo->tv_usec) return 1;
			else{
				if(pOne->tv_usec < pTwo->tv_usec) return -1;
				else
					return 0;
			}
		}
	}
}


/*****************************************************************************
 * Larry granroth's parsetime, included here so compiling options are the same
 * would have preferred to keep it seperate
 */

/* ----------------------------------------------------------------------

  parsetime.c written by L. Granroth  04-15-93
  to parse typical ascii date/time strings and return
  year, month, day of month, day of year, hour, minute, second.

  --------------------------------------------------------------------- */
  
#define DELIMITERS " \t/-:,_;\n\r"
#define PDSDELIMITERS " \t/-T:,_;\n\r"

#define DATE 0
#define YEAR 1
#define MONTH 2
#define DAY 3
#define HOUR 4
#define MINUTE 5
#define SECOND 6

char *months[] = {"january", "february", "march", "april", "may", "june",
  "july", "august", "september", "october", "november", "december"};

int4 gse_day_offset[2][14] = {
  {  0,   0,  31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334, 365},
  {  0,   0,  31,  60,  91, 121, 152, 182, 213, 244, 274, 305, 335, 366} };

int4 gse_days_in_month[2][14] = {
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 0},
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 0} };

  
size_t gse_isLeap(int4 nYear){
	return nYear % 4 ? 0 : (nYear % 100 ? 1 : (nYear % 400 ? 0 : 1));
}

void gse_ct_gmnow(gse_caltime_t* pCt)
{
	struct timeval tv;
	struct tm* pTm;
	
	if(pCt == NULL)
		gse_throw(GSE_XCEPT, GSE_XNULLPTR);
	
	if(gettimeofday(&tv,NULL) != 0)
		gse_throw(CLIB_XCEPT, errno);
	
	pTm = gse_gmtime(&(tv.tv_sec));
	
	pCt->year = pTm->tm_year + 1900;
	pCt->month = pTm->tm_mon + 1;
	pCt->dom = pTm->tm_mday;
	pCt->hour = pTm->tm_hour;
	pCt->min = pTm->tm_min;
	pCt->sec = (float8) pTm->tm_sec;
	
	pCt->sec += (tv.tv_usec) / 1000000.0;
	
	gse_ct_setdoy(pCt);
}
  
int4 gse_parsetime(const char *string, int *year, int *month, int *day_month,
                  int *day_year, int *hour, int *minute, double *second)
{
  char s[80];
  char *c;
  char *delimiters;
  char *end_of_date;
  time_t curtime;
  struct tm *curtm;
  int i, j, len, n;
  char *tok[10];
  int want[7] = {0};
  char *ptr;
  int number;
  double value;
  int hold;
  int leap;

  (void)strncpy (s, string, 80);

  /* handle PDS time format */

  delimiters = DELIMITERS;
  if ((c = strchr (s, 'Z'))) *c = (char)0;
  end_of_date = strchr (s, 'T');
  if (end_of_date) {
    c = end_of_date - 1;
    if (isdigit ((int)(*c))) delimiters = PDSDELIMITERS;
    else end_of_date = (char *)0;
  }

  /* if not PDS then count out 3 non-space delimiters */

  if (!end_of_date) {
    n = 0;
    len = strlen (s);
    for (i = 0; i < len; i++) {
      if ((c = strchr (delimiters+2, (int)s[i]))) n++;
      if (n == 3) {
        end_of_date = s + i;
        break;
      }
    }
  }

  /* default to current year, limiting, it means that date before 1970
     can't be generated, even if you provide the complete string! */

  if (time (&curtime) == (time_t)(-1)) return -1;
  if (!(curtm = localtime (&curtime))) return -1;
  *year = curtm->tm_year + 1900;
  *month = 0;
  *day_month = 0;
  *day_year = 0;
  *hour = 0;
  *minute = 0;
  *second = 0.0;

  /* tokenize the time string */

  if (!(tok[0] = strtok (s, delimiters))) return -1;

  for (n = 1; n < 10 && (tok[n] = strtok ((char *)0, delimiters)); n++);

  want[DATE] = want[YEAR] = want[MONTH] = want[DAY] = 1;
  hold = 0;

  for (i = 0; i < n; i++) {

    if (end_of_date && want[DATE] && (tok[i] > end_of_date)) {
      want[DATE] = 0;
      want[HOUR] = want[MINUTE] = want[SECOND] = 1;
    }

    len = strlen (tok[i]);

    /* skip 3-digit day-of-year values in parenthesis */
    if ((len == 5) && (*tok[i] == '(') && (*(tok[i]+4) == ')')) {
      value = strtod (tok[i]+1, &ptr);
      if ((value > 0) && (value < 367)) continue; 
    }

    value = strtod (tok[i], &ptr);
    if (ptr == tok[i]) {
      if (len < 3 || !want[DATE]) return -1;
      for (c = tok[i]; *c; c++) *c = tolower ((int)(*c));
      for (j = 0; j < 12; j++) {
        if (!strncmp (months[j], tok[i], len)) {
	  *month = j + 1;
	  want[MONTH] = 0;
	  if (hold) {
	    if (*day_month) return -1;
	    *day_month = hold;
	    hold = 0;
	    want[DAY] = 0;
	  }
	  break;
	}
      }
      if (want[MONTH]) return -1;
      continue;
    }

    if (fmod (value, 1.0) != 0.0) {
      if (want[SECOND]) {
        *second = value;
        break;
      } else return -1;
    }

    number = value;
    if (number < 0) return -1;

    if (want[DATE]) {

      if (!number) return -1;

      if (number > 31) {

        if (want[YEAR]) {
	  *year = number;
	  if (*year < 1000) *year += 1900;
	  want[YEAR] = 0;
	} else if (want[MONTH]) {
	  want[MONTH] = 0;
	  *month = 0;
	  *day_year = number;
	  want[DAY] = 0;
	} else return -1;

      } else if (number > 12) {

	if (want[DAY]) {
	  if (hold) {
	    *month = hold;
	    want[MONTH] = 0;
	  }
	  if (len == 3) {
	    if (*month) return -1;
	    *day_year = number;
	    *day_month = 0;
	    want[MONTH] = 0;
	  } else *day_month = number;
	  want[DAY] = 0;
	} else return -1;

      } else if (!want[MONTH]) {

	if (*month) {
	  *day_month = number;
	  *day_year = 0;
	} else {
	  *day_year = number;
	  *day_month = 0;
	}
	want[DAY] = 0;

      } else if (!want[DAY]) {

	if (*day_year) return -1;
	*month = number;
	want[MONTH] = 0;

      } else if (!want[YEAR]) {

	if (len == 3) {
	  if (*month) return -1;
	  *day_year = number;
	  *day_month = 0;
	  want[DAY] = 0;
	} else {
	  if (*day_year) return -1;
	  *month = number;
	  if (hold) {
	    *day_month = hold;
	    want[DAY] = 0;
	  }
	}
	want[MONTH] = 0;

      } else if (hold) {

	*month = hold;
	hold = 0;
	want[MONTH] = 0;
	*day_month = number;
	want[DAY] = 0;

      } else hold = number;

      if (!(want[YEAR] || want[MONTH] || want[DAY])) {
        want[DATE] = 0;
        want[HOUR] = want[MINUTE] = want[SECOND] = 1;
      }

    } else if (want[HOUR]) {

      if (len == 4) {
        hold = number / 100;
	if (hold > 23) return -1;
	*hour = hold;
	hold = number % 100;
	if (hold > 59) return -1;
	*minute = hold;
	want[MINUTE] = 0;
      } else {
        if (number > 23) return -1;
	*hour = number;
      }
      want[HOUR] = 0;

    } else if (want[MINUTE]) {

      if (number > 59) return -1;
      *minute = number;
      want[MINUTE] = 0;

    } else if (want[SECOND]) {

      if (number > 61) return -1;
      *second = number;
      want[SECOND] = 0;

    } else return -1;

  } /* for all tokens */

  if (*month > 12) return -1;
  if (*month && !*day_month) *day_month = 1;

  leap = gse_isLeap(*year);

  if (*month && *day_month && !*day_year) {
    if (*day_month > gse_days_in_month[leap][*month]) return -1;
    *day_year = gse_day_offset[leap][*month] + *day_month;
  } else if (*day_year && !*month && !*day_month) {
    if (*day_year > (365 + leap)) return -1;
    for (i = 2; i < 14 && *day_year > gse_day_offset[leap][i]; i++);
    i--;
    *month = i;
    *day_month = *day_year - gse_day_offset[leap][i];
  } else return -1;

  return 0;

} /* parsetime */


bool gse_ct_isnorm(const gse_caltime_t* pCal)
{
	int4 iLeap, nDays;
	
	if( (pCal->month < 1)||(pCal->month > 12)) return false;
	
	iLeap = gse_isLeap(pCal->year);
	
	if( (pCal->dom < 1) ||(pCal->dom > gse_days_in_month[iLeap][pCal->month]))
		return false;
	
	nDays = iLeap ? 366 : 365;
	if( (pCal->doy < 1)   || (pCal->doy > nDays) ) return false;
	
	if( (pCal->hour < 0)  || (pCal->hour > 23)   ) return false;
	if( (pCal->min < 0)   || (pCal->min > 59)    ) return false;
	if( (pCal->sec < 0.0) || (pCal->sec >= 60.0) ) return false;
	
	return true;
}


/*****************************************************************************
 * From Larry Granroth's daslib, thanks!
 *
 * tnorm -- normalize date and time components for the Gregorian calendar
 * ignoring leap seconds
 *
 * (This is the most likely bug nest.)
 *
  ---------------------------------------------------------------------------*/
	
void gse_ct_norm_dom(gse_caltime_t* pCal)
{
	
/* tnorm (int *year, int *month, int *mday, int *yday,
       int *hour, int *minute, double *second) */

  int4 ileap;

  
  /* month is required input -- first adjust month */
  if (pCal->month > 12 || pCal->month < 1) {
    /* temporarily make month zero-based */
    (pCal->month)--;
    pCal->year += pCal->month / 12;
    pCal->month %= 12;
    if (pCal->month < 0) {
      pCal->month += 12;
      (pCal->year)--;
    }
    (pCal->month)++;
  }

  /* index for leap year */
  ileap = gse_isLeap(pCal->year);

  /* day of year is output only -- calculate it */
  pCal->doy = gse_day_offset[ileap][pCal->month] + pCal->dom;
  
  /* At this point the doy reflects the month field, so do this as a 
     doy normalization, which sets the month if other fields require it
	  to change. */
  gse_ct_norm_doy(pCal);
  
};


void gse_ct_norm_doy(gse_caltime_t* pCal)
{
	
	/* month input ignored (output only) for this version */
	
  int4 ileap, ndays;
	
  /* index for leap year */
  ileap = gse_isLeap(pCal->year);

  /* again, we're ignoring leap seconds */
  if (pCal->sec >= 60.0 || pCal->sec < 0.0) {
    pCal->min += (int)(pCal->sec / 60.0);
    pCal->sec = fmod (pCal->sec, 60.0);
    if (pCal->sec < 0.0) {
      pCal->sec += 60.0;
      (pCal->min)--;
    }
  }

  if (pCal->min >= 60 || pCal->min < 0) {
    pCal->hour += pCal->min / 60;
    pCal->min %= 60;
    if(pCal->min < 0) {
      pCal->min += 60;
      (pCal->hour)--;
    }
  }

  if (pCal->hour >= 24 || pCal->hour < 0) {
    pCal->doy += pCal->hour / 24;
    pCal->hour %= 24;
    if (pCal->hour < 0) {
      pCal->hour += 24;
      (pCal->doy)--;
    }
  }

  /* final adjustments for year and day of year */
  ndays = ileap ? 366 : 365;
  if (pCal->doy > ndays || pCal->doy < 1) {
    while(pCal->doy > ndays) {
      (pCal->year)++;
      pCal->doy -= ndays;
      ileap = gse_isLeap(pCal->year);
      ndays = ileap ? 366 : 365;
    }
    while(pCal->doy < 1) {
      (pCal->year)--;
      ileap = gse_isLeap(pCal->year);
      ndays = ileap ? 366 : 365;
      pCal->doy += ndays;
    }
  }

  /* and finally convert day of year back to month and day */
  while (pCal->doy <= gse_day_offset[ileap][pCal->month]) (pCal->month)--;
  while (pCal->doy >  gse_day_offset[ileap][pCal->month + 1]) (pCal->month)++;
  pCal->dom = pCal->doy - gse_day_offset[ileap][pCal->month];

  /* all done */
  return;

} /* tnorm - normalize date and time components */


void gse_ct_setdoy(gse_caltime_t* pCal)
{
	int4 iLeap = gse_isLeap(pCal->year);
	
	if( (pCal->month < 1)||(pCal->month > 12)){ 
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Month value %d in date "
		             "%04d-%02d-%02dT%02d:%02d:%06.3f is outside the valid "
		             "range.", pCal->month, pCal->year, pCal->month, pCal->dom,
				       pCal->hour, pCal->min, pCal->sec);
	}
	
	iLeap = gse_isLeap(pCal->year);
	
	if( (pCal->dom < 1) ||(pCal->dom > gse_days_in_month[iLeap][pCal->month])){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Day value %d in date "
		             "%04d-%02d-%02dT%02d:%02d:%06.3f is outside the valid "
		             "range.", pCal->dom, pCal->year, pCal->month, pCal->dom,
				       pCal->hour, pCal->min, pCal->sec);
	}
	
	pCal->doy = gse_day_offset[iLeap][pCal->month] + pCal->dom;
}

void gse_ct_setdom(gse_caltime_t* pCal)
{
	int4 iLeap = gse_isLeap(pCal->year);
	int4 nDays;
	
	nDays = iLeap ? 366 : 365;
	
	if( (pCal->doy < 1) || (pCal->doy > nDays)){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Day of Year value %d in date "
		             "%04d-%03dT%02d:%02d:%06.3f is outside the valid range. ",
		              pCal->dom, pCal->year, pCal->doy, pCal->hour, pCal->min,
		              pCal->sec);
	}
	
	pCal->month = 1;
	while( pCal->doy > gse_day_offset[iLeap][pCal->month + 1]) pCal->month++;
	pCal->dom = pCal->doy - gse_day_offset[iLeap][pCal->month];
}

/*****************************************************************************/

const gse_caltime_t gse_min_unix_32time = {1901, 12, 13, 20, 45, 52.0};
const gse_caltime_t gse_max_unix_32time = {2038, 01, 19, 3, 14, 7.0};

/* Thanks GNU man pages... */
time_t gse_timegm(struct tm *tm) {
	time_t ret;
	char *tz;

	tz = getenv("TZ");
	setenv("TZ", "", 1);
	tzset();
	ret = mktime(tm);
	if (tz)
		setenv("TZ", tz, 1);
	else
		unsetenv("TZ");
	tzset();
	return ret;
}


bool gse_ct2tv_gm(const gse_caltime_t* pCal, struct timeval* pTv, bool bThrow)
{
	struct tm tmTime;
	
	if(pCal == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null calendar time structure pointer");
	
	if(!gse_ct_isnorm(pCal))
		gse_throw_m(GSE_XCEPT, GSE_XASSERT, "Unnormalized calendar time");
	
	if(pTv == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null timeval struc pointer");
	
	if(sizeof(time_t) == 4){
	
		if(gse_ctcmp(pCal, &gse_min_unix_32time) < 0){
			if(bThrow)
				gse_throw_fm(GSE_XCEPT, GSE_XUNDERFLOW, "%04d-%02d-%02dT%02d:%02d:%06.3f"
				             " is below the minimum 32bit UNIX time.", pCal->year, pCal->month,
				             pCal->dom, pCal->hour, pCal->min, pCal->sec);
			else
				gse_error_v("%04d-%02d-%02dT%02d:%02d:%06.3f"
				            " is below the minimum 32bit UNIX time.", pCal->year, pCal->month,
				            pCal->dom, pCal->hour, pCal->min, pCal->sec);
			return false;
		}
		
		if(gse_ctcmp(pCal, &gse_max_unix_32time) >= 0){
			if(bThrow)
				gse_throw_fm(GSE_XCEPT, GSE_XOVERFLOW, "%04d-%02d-%02dT%02d:%02d:%06.3f"
				             " is after the max UNIX time.", pCal->year, pCal->month,
				             pCal->dom, pCal->hour, pCal->min, pCal->sec);
			else
				gse_error_v("%04d-%02d-%02dT%02d:%02d:%06.3f"
				             " is after the max UNIX time.", pCal->year, pCal->month,
				             pCal->dom, pCal->hour, pCal->min, pCal->sec);
			
			return false;
		}
	}
	
	tmTime.tm_year = pCal->year - 1900;
	tmTime.tm_mon  = pCal->month - 1;
	tmTime.tm_mday = pCal->dom;
	tmTime.tm_hour = pCal->hour;
	tmTime.tm_min  = pCal->min;
	tmTime.tm_sec  = (int)(pCal->sec);
	tmTime.tm_isdst = -1; 

	pTv->tv_sec = gse_timegm(&tmTime);
	pTv->tv_usec = (pCal->sec - floor(pCal->sec)) * 1000000;
	
	return true;
}

bool gse_str2ct_gm(const char* sTime, gse_caltime_t* pCal, bool bThrow)
{
	if(sTime == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null time string pointer");
	
	if(pCal == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null gse_caltime_t struc pointer");
		
	if(gse_parsetime(sTime,&(pCal->year),
	                   &(pCal->month),
	                   &(pCal->dom),
	                   &(pCal->doy),
	                   &(pCal->hour),
	                   &(pCal->min),
	                   &(pCal->sec)) != 0){
		if(bThrow)
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Un-parseable time string, '%s'",
			             sTime);
		else
			gse_warning_v("Un-parseable time string, '%s'", sTime);
					
		return false;
	}
	
	return true;
}

char* gse_ct2iso_gm(const gse_caltime_t* pCt, int prec)
{
	if(pCt == NULL) gse_throw(GSE_XCEPT, GSE_XNULLPTR);
	
	if(!gse_ct_isnorm(pCt))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Only normalized calender times "
		            "can be printed.");
		
	switch(prec){
	case GSE_YEARS:
		return NS("%04d", pCt->year);
	case GSE_MONTHS:
		return NS("%04d-%02d", pCt->year, pCt->month);
	case GSE_DAYS:
		return NS("%04d-%02d-%02d", pCt->year, pCt->month, pCt->dom);
	case GSE_HOURS:
		return NS("%04d-%02d-%02dT%02d", pCt->year, pCt->month, pCt->dom, 
		          pCt->hour);
	case GSE_MINS:
		return NS("%04d-%02d-%02dT%02d:%02d", pCt->year, pCt->month, pCt->dom,
		          pCt->hour, pCt->min);
	case GSE_SECS:
		return NS("%04d-%02d-%02dT%02d:%02d:%02.0f", pCt->year, pCt->month,
				    pCt->dom, pCt->hour, pCt->min, pCt->sec);
	case GSE_MILLISECS:
		return NS("%04d-%02d-%02dT%02d:%02d:%06.3f", pCt->year, pCt->month,
				    pCt->dom, pCt->hour, pCt->min, pCt->sec);
	case GSE_MICROSECS:
		return NS("%04d-%02d-%02dT%02d:%02d:%09.6f", pCt->year, pCt->month,
				    pCt->dom, pCt->hour, pCt->min, pCt->sec);
	default:
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "prec: %d, is outside the valid "
		             "range 1-8 (years to microseconds)", prec);
		break;
	}
	
	return NULL;  /* I'll never get there but compiler doesn't know that */
}

bool gse_str2tv_gm(const char* sTime, struct timeval* pTv, bool bThrow)
{
	gse_caltime_t cal;
	
	if(!gse_str2ct_gm(sTime, &cal, bThrow)) return false;
	if(!gse_ct2tv_gm(&cal, pTv, bThrow)) return false;
	
	return true;
}

int gse_ctcmp(const gse_caltime_t* pCal1, const gse_caltime_t* pCal2)
{
	if((pCal1 == NULL)||(pCal2 == NULL))
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null calender time pointer");
	
	
	if(pCal1->year < pCal2->year) return -1;
	if(pCal1->year > pCal2->year) return 1;
	
	if(pCal1->month < pCal2->month) return -1;
	if(pCal1->month > pCal2->month) return 1;
	
	if(pCal1->dom < pCal2->dom) return -1;
	if(pCal1->dom > pCal2->dom) return 1;	
	
	if(pCal1->hour < pCal2->hour) return -1;
	if(pCal1->hour > pCal2->hour) return 1;
	
	if(pCal1->min < pCal2->min) return -1;
	if(pCal1->min > pCal2->min) return 1;
	
	if(pCal1->sec < pCal2->sec) return -1;
	if(pCal1->sec > pCal2->sec) return 1;
	
	return 0;
}

bool gse_cal_inrange(const gse_caltime_t* pStart, const gse_caltime_t* pEnd, 
                     const gse_caltime_t* pTest)
{
	if(pTest == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Null pTest struc pointer");
	
	if(pEnd == NULL){
		if(pStart == NULL) return true;		
		return (gse_ctcmp(pStart, pTest) <=0) ? true : false;
	}
	
	if(pStart == NULL){
		if(pEnd == NULL) return true;
		return (gse_ctcmp(pTest, pEnd) < 0) ? true : false;
	}
	
	/* Okay, the range is not half open */
	return ( (gse_ctcmp(pStart, pTest) <= 0) &&
			   (gse_ctcmp(pTest, pEnd) < 0));
}




