/** @file GseTime.h Time Handling */
#ifndef _GseTime_H_
#define _GseTime_H_

#include <rpwgse/GseTypes.h>
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_sys */
/*@{*/

/** Tell the current thread to take a nap for a duration.
 *
 * @param millisec Number of milliseconds to sleep.
 *
 * @param cont if @c true automatically resume sleeping the remainder
 *        of any nap time that was interrupted via a signal, if @c false
 *        return on reciept of a signal.
 *
 * @return Returns the number of uslept milliseconds if @a cont was true 
 *         and sleeping was interrupted.
 *
 * @note Underlying implementation uses @c nanosleep on POSIX systems, and
 *       @c SleepEx on windows.
 */
 
int gse_millisleep(int millisec, bool cont);


/** Convert a unix epoch time to broken-down UTC time.
 *
 * Unlike a similar POSIX function, this uses thread local storage, and is
 * safe for use in multi-threaded programs and libraries.
 */
struct tm* gse_gmtime(const time_t* pTime);

/** Convert a unix epoch time to broken-down Local time.
 *
 * Unlike a similar POSIX function, this uses thread local storage, and is
 * safe for use in multi-threaded programs and libraries.
 */
struct tm* gse_localtime(const time_t* pTime);


/** Calculate a wakeup time as an offset from now 
 *
 * @param fSecFromNow how long from now to wakeup
 * @param pWake a timeval structure that will be filled in with the
 *        wakeup time
 *
 */
void gse_tv_calcwake(float4 fSecFromNow, struct timeval* pWake);


/** Tell the current thread to take a nap until some time in the future.
 *
 * @param pWake When to wakeup
 * @param pbRun If a signal interrupts the sleep the value pointed to
 *        by *pbRun is checked, if 
 *
 */
int gse_tv_sleepuntil(struct timeval* pWake, int* pbRun);

/** Compare to timeval structures 
 * @param pOne a pointer to the first timevalue
 * @param pTwo a pointer to the second timevalue
 * @returns 1 if *pOne is > *pTwo, 0 if they are the same and -1 if 
 *            *pOne is < *pTwo.
 */
int gse_tv_cmp(const struct timeval* pOne, const struct timeval* pTwo);


#define GSE_YEARS       1
#define GSE_MONTHS      2
#define GSE_DAYS        3
#define GSE_HOURS       4
#define GSE_MINS        5
#define GSE_SECS        6
#define GSE_MILLISECS   7
#define GSE_MICROSECS   8
		 
/** Convert a time value to a GM time ISO-8601 date string.
 *
 * The resolution of the returned time string can range from years to
 * microseconds.  Any portion below the requested resolution is @b truncated
 * not rounded. 
 *
 * @see gse_tv2iso_local for an example.
 * @see gse_str2tv_gm for the inverse operation.
 *
 * @param pTv the time value to convert to a UTC date string
 *
 * @param prec defines the returned time resolution is one of the
 * defines:
 *   - @a GSE_YEARS
 *   - @a GSE_MONTHS
 *   - @a GSE_DAYS
 *   - @a GSE_HOURS
 *   - @a GSE_MINS
 *   - @a GSE_SECS
 *   - @a GSE_MILLISECS
 *   - @a GSE_MICROSECS
 * 
 * @return A UTC (GM) time string allocated on the heap in the format: @n@n
 *    <b> ccyy-MM-ddThh:mm:SS.sss </b> @n@n
 * Where: @n
 *   - @b cc is the century (00-99)
 *   - @b yy is the year    (00-99)
 *   - @b MM is the month   (01-12)
 *   - @b hh is the hour    (00-23)
 *   - @b mm is the minute  (00-59)
 *   - @b SS is the second  (00-60, account for 1 leap second)
 *   - @b sss is the fraction of a second.  Upto 6 digits for microsecs resolution
 *
 * @throws (CLIB_XCEPT, errno) if pTvUTC is invalid.
 */
char* gse_tv2iso_gm(struct timeval* pTv, int prec);


/** Convert a time value to a GM time ISO-8601 date string, static memory
 * version.
 * @see gse_tv2iso_gm
 */
char* gse_tv2iso_gm_s(struct timeval* pTv, int prec, char* buf, int len);

/** Convert a time value to a local time ISO-8601 date string.
 *
 * This function has the same call interface as gse_tv2iso_gm(), the only
 * difference is that @c pTv is converted to a local time.  Note that the
 * same @c struct @c timeval can be read two different ways as demonstrated 
 * below.
 * @code
 *
 * //Get the current local and GM time and print both
 * struct timeval tv;
 * char *sUtc, *sLocal;
 * 
 * gettimeofday(&tv, NULL);
 *
 * sUtc =   gse_tv2iso_gm(&tv, GSE_TP_SECS);
 * sLocal = gse_tv2iso_local(&tv, GSE_TP_SECS);
 *
 * printf("Current time is: %s (GM), %s (Local)\n", sUtc, sLocal);
 *
 * free(sUtc);
 * free(sLocal);
 * @endcode
 *
 */
char* gse_tv2iso_local(struct timeval* pTv, int prec);

/** Convert a time value to a local time ISO-8601 date string, static memory
 * version.
 * @see gse_tv2iso_local
 */
char* gse_tv2iso_local_s(struct timeval* pTv, int prec, char* buf, int len);


/** A structure for holding calaneder time values */
typedef struct gse_caltime{
	/** The 4 digit calander year, valid range is -2^32 to +2^32 */
	int4 year;
	
	/** The month of the year, jaunary = 1 */
	int4 month;
	
	/** The day of the month, the first day = 1 */
	int4 dom;
	
	/** The day of the year, here January 1st = 1 */
	int4 doy;
	
	/** The hour of the day, starts at 0, here 11 PM = 23 */
	int4 hour;
	
	/** The minute after the hour, starts at 0 */
	int4 min;
	
	/** The decmial seconds after the minute, starts at 0.0 */
	float8 sec;	
	
} gse_caltime_t;


/** Load 'NOW' into a caltime structure
 *
 * @param A pointer to a caltime structure to be initialized with GMT now.
 */
void gse_ct_gmnow(struct gse_caltime* pCt);


/** Is this a normalized calendar time structure?
 *
 */
bool gse_ct_isnorm(const gse_caltime_t* pCal);

/** Set the doy field to match the month and dom fields.
 *
 * @param pCal a calendar time with a legal year, month, and day of month
 *        fields.
 */
void gse_ct_setdoy(gse_caltime_t* pCal);


/** Set the month and dom field to match the doy field.
 *
 * @param pCal a calendar time with a legal year and day of year fields.
 */
void gse_ct_setdom(gse_caltime_t* pCal);


/** Normalize a calender time assuming month and day of month contain data.
 *
 * If any fields have been incremented or decremented outside their
 * valid range, update them.  This version calculates the doy field
 * from the final normalized month and dom fields.
 *
 * @note This was taken from Larry Granroth's tnorm function, Thanks!
 *
 * @param pCal a pointer to the cal structure to normalize.
 */
void gse_ct_norm_dom(gse_caltime_t* pCal);

/** Normalize a calender time the day of year field contains data.
 *
 * If any fields have been incremented or decremented outside their
 * valid range, update them.  This version calculates the month and
 * dom fields from the final normalized day of year field.
 *
 * @param pCal a pointer to the cal structure to normalize.
 */
void gse_ct_norm_doy(gse_caltime_t* pCal);


/** Convert a GMT calender time into a unix struct timeval
 * 
 * @param pCal a pointer to the input calender time structure.
 *
 * @param pTv a pointer to the struct timeval to fill out.
 *
 * @param bThrow if true, the function will throw an exception when the
 *        calendar time cannot be represented as a unix time.
 *
 * @returns true if the calendar time could be represented in a 32bit unix
 *          time range (1970-01-01 to 2037-12-31), false otherwise.
 */
bool gse_ct2tv_gm(const gse_caltime_t* pCal, struct timeval* pTv, bool bThrow);


/** Convert a time string into a calander time value.
 * 
 * @note Time parsing uses Larry Granroth's parsetime function, thanks.
 *
 * @param sTime the string to convert, many formats are accepted.
 *
 * @param pCal the call structure to be filled in.
 *
 * @param bThrow If set to true the function always returns true, as 
 *        any issues recognizing the time string prompt an exception.
 *        If set to false, numskull problems such as null pointers will still
 *        cause a throw, but a string that cannot be parsed causes a false
 *        return and @pCal is untouched.
 *
 * @returns true if the string was parsed into a calendar time, false 
 *          otherwise
 */
bool gse_str2ct_gm(const char* sTime, gse_caltime_t* pCal, bool bThrow);

/** Convert a calendar time to a GM time ISO-8601 date string.
 *
 * The resolution of the returned time string can range from years to
 * microseconds.  Any portion below the requested resolution is @b truncated
 * not rounded. 
 *
 * @see gse_tv2iso_local for an example.
 * @see gse_str2tv_gm for the inverse operation.
 *
 * @param pTv the time value to convert to a UTC date string
 *
 * @param prec defines the returned time resolution is one of the
 * defines:
 *   - @a GSE_YEARS
 *   - @a GSE_MONTHS
 *   - @a GSE_DAYS
 *   - @a GSE_HOURS
 *   - @a GSE_MINS
 *   - @a GSE_SECS
 *   - @a GSE_MILLISECS
 *   - @a GSE_MICROSECS
 * 
 * @return A UTC (GM) time string allocated on the heap in the format: @n@n
 *    <b> sccyy-MM-ddThh:mm:SS.sss </b> @n@n
 * Where: @n
 *   - @b s is the sign if the year is before 1AD
 *   - @b cc is the century (00-99)
 *   - @b yy is the year    (00-99)
 *   - @b MM is the month   (01-12)
 *   - @b hh is the hour    (00-23)
 *   - @b mm is the minute  (00-59)
 *   - @b SS is the second  (00-60, account for 1 leap second)
 *   - @b sss is the fraction of a second.  Upto 6 digits for microsecs resolution
 *
 * @throws (CLIB_XCEPT, errno) if pCt is not normalized.
 */
char* gse_ct2iso_gm(const gse_caltime_t* pCt, int prec);


/** Convert a time string into a unix time value.
 *
 * This is a convienence rollup of gse_str2ct_gm and gse_ct2tv_gm.  The
 * bThrow parameter is passed down to the subfunctions.
 * @see gse_str2ct_gm
 * @see gse_ct2tv_gm
 *
 * @returns True if both subfunctions return true.
 */
bool gse_str2tv_gm(const char* sTime, struct timeval* pTv, bool bThrow);



/** Compare two calander values
 *
 * @returns -1 if cal1 < cal2, 0 if they are equal or 1 if cal1 > cal2 
 */
int gse_ctcmp(const gse_caltime_t* pCal1, const gse_caltime_t* pCal2);

/** Compare a calendar value to a time range
 *
 * @param pStart The start of the time range.  This may be NULL which is
 *        taken to be negative infinity.
 * @param pEnd The end of the time range, this may be NULL which is taken
 *        to be positive infinity.
 * @param pTest The time to be compared to start and end.  May not be NULL.
 * @return true if the test time point is at or equal to start and before
 *         end, false otherwise.
 */

bool gse_ct_inrange(gse_caltime_t* pStart, gse_caltime_t* pEnd, 
                     gse_caltime_t* pTest);

/** Is the a given year a leap year 
 *
 * @param nYear a year with the century component.
 *
 * @returns 0 if this is not a leap year, 1 if it is.  This value is 
 *          suitable for use as the 1st index in the gse_day_offset and
 *          gse_days_in_month arrays.
 */
size_t gse_isLeap(int4 nYear);
					
#ifndef _GseTime_C_
extern int4 gse_day_offset[2][14]; 
extern int4 gse_days_in_month[2][14];
#endif


/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseTime_H_ */
