/** File GseTreeWatch.h */
#ifndef _GseTreeWatch_H_
#define _GseTreeWatch_H_

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_sys */
/*@{*/

/* Changes and Lists of Changes *********************************************/

/** Structure for holding a single filesystem change notice.
    Note: a change of type 'C' can mean that a file you are watching
	       has been modified, as in a new file was copied into it's place
			 the behavior is up to the user of this lib object, but one
			 way to handle this would be to stop reading from the current
			 file, and reopen it */

typedef struct gse_tree_change{
	
	/** This is one of the codes 'C' - Create, 'D' - delete, 'M' - modify,
	 * or the dreaded 'I' meaning that all file names below the given 
	 * directory are now invalid, if you were dealing with any of them
	 * they are now in a new home.
	 */
	char cChange;
	
	/** This is the name of the file that changed, relative to 
	 * some root directory.  Note that directory seperator's are
	 * given as a '/' regardless of the OS type.
	 */
	char* sPath;
	
	/** If true this change applies to a directory */
	bool bIsDir;
	
	/** Destructor for a tree change item */
	void (*del)(void* vpThis);
	
} GseTreeChange;

GseTreeChange* new_GseTreeChange(char cChange, char* sPath, bool bIsDir);


/* Change List Object */
#define S gse_twc_list
#define C GseTwcList
#define I GseTreeChange

#include <rpwgse/listTplt.h>

#undef I
#undef C
#undef S

/** Type definition for the change filter callback.
 * @param pCtx the Context pointer supplied to GseTreeWatch::addFilter
 * @param pChange the change in question.  
 *        @warning Do Not Change the sPath pointed to within the callback.
 * @return If true, allow the change to be reported, otherwise suppress it.
 */
typedef bool (*gse_tw_filter_t)(void* pCtx, const GseTreeChange* pChange);


/** Object to watch a directory tree for changes.
 * 
 * Note: If the underlying file system is unmounted then all changes are
 *       flushed.   Each call to "numChanges" after an unmount will cause
 *       the object to try an open the given directory.  If it can't it will
 *       just report "0 changes".    Each call to waitForChange will just
 *       send the object into a slow polling loop.
 *
 */
typedef struct gse_tree_watch{

   /** Change or add the callback message filter
	 *
	 * By default all changes of a particular type are reported, add a filter
	 * function to limit the changes reported based on file or directory names.
	 * For efficiency, callbacks should not be used to filter all changes of
	 * a particular type such as all modifies or deletes.  See the @a flags
	 * parameter of new_GseTreeWatch for more efficient general filtering.
	 *
	 * @param pThis a pointer to a gse tree watch structure.
	 *
	 * @param func the callback function pointer.
	 *
	 * @param pCtx this pointer will be passed to the callback when it's 
	 *        executed.
	 *
	 * @returns The old callback filter, if any.  Otherwise NULL is returned.
	 */
	gse_tw_filter_t (*setFilter)(struct gse_tree_watch* pThis,
	                             gse_tw_filter_t func, void* pCtx);
	
	/** Wait for changes to a dir tree. 
	 *
	 * This function allows the caller to be put to sleep waiting on
	 * changes to a directory tree.  It is not needed if you merely 
	 * wish to see if changes have been detected right now.
	 *
	 * @parm pThis a pointer to a gse tree watch structure.
	 *
	 * @param pbInter If in the rare instance that this function is called
	 *        to wait on a directory that doesn't exist, it will check 
	 *        the value of pbInter on each loop.  If it point to true, 
	 *        the function will stop checking.
	 */
	void (*waitForChange)(struct gse_tree_watch* pThis);
	
	
#ifdef GSE_POSIX_THREADS
	/** <b>(Multi-thread only)</b> Interrupt the packet source, this causes
	 * waitForChange to return if it's blocking. 
	 *
	 * @note This function only exists in @b multi-threaded versions of the 
	 *       library.
	 * 
	 * Obviously this in only useful if the application has more than one
	 * thread.  Does nothing if the source is not blocked waiting on data.
	 *
	 * Note: MT Safe but not Async Signal safe */
	void (*interrupt)(struct gse_tree_watch* pThis, pthread_t threadID);
#endif
	
	/** Find out how many changes are waiting to be read.
	 *
	 * This function checks with the underlying operating system service to 
	 * see if anything has changed.
	 */
	int4 (*numChanges)(struct gse_tree_watch* pThis);
	
	/** Read a change notice. 
	 *
	 * The change notice is removed from the internal list and the caller is
	 * responsible for free'ing the memory.
	 */
	GseTreeChange* (*getChange)(struct gse_tree_watch* pThis);
	
	GseTwcList* pChanges;        /* The current list of changes */
		
	uint4 uFlags;                /* User specified "what to watch flags" */
	
	char* sTop;                  /* The top level directory */
	
	struct gse_tw_node* pTop;    /* Top level node directory */
	
	struct gse_wd_node_dict* pRev; /* Backwards node lookup dictionary */
	
	bool bRestartSelect;   /* If false, waitForChanges won't restart select calls */
	
	bytebuf_t bBuf;        /* Read buffer for notify discriptor */	
	
	gse_tw_filter_t filter;  /* Change filter callback function */
	
	void* pCtx;              /* context pointer supplied to the callback */
	
	int _fdNotify;  /* Linux specific: Kernel resource handle */
	uint4 _flags;   /* Linux specific: Watch flags */
	
	/** Delete the tree watcher, free's any kernel resources */
	void (*del)(void* vpThis);
	
} GseTreeWatch;

#define GSE_TW_CREATE 0x1
#define GSE_TW_DELETE 0x2
#define GSE_TW_MODIFY 0x4
#define GSE_TW_FILES_ONLY 0x8  /* Only report on changes to files, not dirs */
#define GSE_TW_INITIAL_C  0x10 /* Report initially seen items as creates */

/** Make a new tree watcher 
 *
 * @param sPath the directory to watch, note this directory need not exist at
 *        the time the watcher is created.
 *
 * @param uFlags The kinds of changes to watch for.  May be one or more of
 *        - GSE_TW_CREATE
 *        - GSE_TW_DELETE
 *        - GSE_TW_MODIFY
 *        - GSE_TW_FILES_ONLY
 *        - GSE_TW_INITIAL_C
 *
 * @param filter If this argument is NULL all changes of a particular type are
 *        reported.  Use this parameter filter the changes reported based on
 *        file or directory names.  For efficiency, callbacks should not be
 *        used to filter all changes of a particular type such as all modifies
 *	       or deletes, for that purpose use the @a uFlags parameter.
 *
 * @param pCtx this pointer will be passed to the callback when it's  
 *        executed.
 *  
 * @returns a new tree watch object allocated on the heap.
 */
GseTreeWatch* new_GseTreeWatch(const char* sPath, uint4 uFlags,
                               gse_tw_filter_t filter, void* pFilterCtx);

/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseTreeWatch_H_ */
