#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#ifdef WIN32
#error Win32 Complile GseTreeWatch_win.c instead of this file
#endif

#include <unistd.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <signal.h>

#include "dict.h"
#include "GseTreeWatch.h"
#include "GseExcept.h"
#include "GseString.h"
#include "GseWrap.h"
#include "GseFile.h"
#include "GseLog.h"
#include "GseTime.h"

/*****************************************************************************/
/* Tree change records */

void delete_GseTreeChange(void* vpThis)
{
	GseTreeChange* pThis = (GseTreeChange*)vpThis;
	free(pThis->sPath);
	free(vpThis);
}

/* Warning!!!! I take ownership of sPath ! */
GseTreeChange* new_GseTreeChange(char cChange, char* sPath, bool bIsDir)
{
	GseTreeChange* pThis;
	
	if(sPath == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sPath is Null");
	
	if(sPath[0] != '/')
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Expected an absolute path, recieved "
		             "'%s'", sPath);
				
	pThis = (GseTreeChange*)gse_calloc(1, sizeof(GseTreeChange));
	pThis->cChange = cChange;
	pThis->sPath = sPath;
	pThis->bIsDir = bIsDir;
	pThis->del = delete_GseTreeChange;
	
	return pThis;
}


/* Hide the directory node objects from the rest of the lib, contains OS level
   details */
	
/* Directory Watching Nodes **************************************************/

typedef struct gse_tw_node{

	int wd;            /* Watch discriptor */
	
	char* sKey;        /* Owns the key memory */
	
	struct gse_tw_node* pParent;
	
	struct gse_str_twnode_dict* pSubs;
	
	bool bPrune;
	
	bool bMovePrune;
	
	GseStrAry* pNewSubs;
	
	void (*del)(void* vpThis);
	
} GseTwNode;


/* Dictonary of Directory Watching Nodes by subdir name **********************/

#define S gse_str_twnode_dict
#define C GseStrTwNodeDict
#define K char
#define V GseTwNode

#include "ptrDictTplt.h"

#undef S
#undef C
#undef K
#undef V

/* Dictiorary of Directory Watching Nodes by watch discriptor ****************/

#define S gse_wd_node_dict
#define C GseWdNodeDict
#define K int
#define V GseTwNode

#include "dictTplt.h"

#undef S
#undef C
#undef K
#undef V

/* Node destructor ************************************************************/
void delete_GseTwNode(void* vpThis)
{
	GseTwNode* pThis = (GseTwNode*)vpThis;
	
	/* fprintf(stderr, "Deleting GseTwNode, wd: %d, key: %s\n", pThis->wd,
	        pThis->sKey); */
	free(pThis->sKey);
	pThis->pSubs->del(pThis->pSubs);
	pThis->pNewSubs->del(pThis->pNewSubs);
	free(vpThis);
}

/* Get the abspath of a child item *******************************************/

char* GseTwNode_getChildPath(const GseTwNode* pNode, const char* sItem)
{
	char sPath[PATH_MAX];
	char* pIns = NULL;
	memset(sPath, 0, PATH_MAX);
	
	pIns = sPath + PATH_MAX - 1 - strlen(sItem) - 1;
	if( pIns < sPath) gse_throw(CLIB_XCEPT, ENAMETOOLONG);
	
	memcpy(pIns+1, sItem, strlen(sItem));
	*pIns = '/';
			
	/* Recurse up through the parents until we get to the top */
	while(pNode){
		
		if(pNode->sKey[0] != '/'){
		
			pIns -= strlen(pNode->sKey) + 1;
			if( pIns < sPath) gse_throw(CLIB_XCEPT, ENAMETOOLONG);
			memcpy(pIns+1, pNode->sKey, strlen(pNode->sKey));
			*pIns = '/';
		}
		else{
			pIns -= strlen(pNode->sKey);
			if( pIns < sPath) gse_throw(CLIB_XCEPT, ENAMETOOLONG);
			memcpy(pIns, pNode->sKey, strlen(pNode->sKey));
		}
		
		pNode = pNode->pParent;
	}
	
	return CS(pIns);
}

char* GseTwNode_getPath(const GseTwNode* pNode)
{
	char sPath[PATH_MAX];
	char* pIns = NULL;
	
	memset(sPath, 0, PATH_MAX);
	pIns = sPath + PATH_MAX - 1;
	
	while(pNode){
		
		if(pNode->sKey[0] != '/'){		
			pIns -= strlen(pNode->sKey) + 1;
			if( pIns < sPath) gse_throw(CLIB_XCEPT, ENAMETOOLONG);
			memcpy(pIns+1, pNode->sKey, strlen(pNode->sKey));
			*pIns = '/';
		}
		else{
			pIns -= strlen(pNode->sKey);
			if( pIns < sPath) gse_throw(CLIB_XCEPT, ENAMETOOLONG);
			memcpy(pIns, pNode->sKey, strlen(pNode->sKey));
		}
		
		pNode = pNode->pParent;
	}
	
	return CS(pIns);
}


/*****************************************************************************/
/*****************************************************************************/
/* Tree Watcher Class */


/****************************************************************************/

#define GSE_TW_NEW 0x01
#define GSE_TW_DEL 0x02
#define GSE_TW_MOD 0x04
#define GSE_TW_DIR 0x08

/* WARNING!!!! This function takes control of the sPath argument memory!! */
void GseTreeWatch_addChange(GseTreeWatch* pThis, char* sPath, uint4 uCode)
{
	GseTreeChange change;
	GseTreeChange* pChange;
	
	if((uCode & GSE_TW_DIR )&&(pThis->uFlags & GSE_TW_FILES_ONLY )){
		return;
	}
	
	if(uCode & GSE_TW_DEL){ 
		if( pThis->uFlags & GSE_TW_DELETE)
			change.cChange = 'D';
		else
			return;
	}
	
	if(uCode & GSE_TW_MOD){
		if( pThis->uFlags & GSE_TW_MODIFY)
			change.cChange = 'M';	
		else
			return;
	}
	
	if(uCode & GSE_TW_NEW){
		if( pThis->uFlags & GSE_TW_CREATE)
			change.cChange = 'C';
		else
			return;
	}
		
	change.bIsDir = (uCode & GSE_TW_DIR) ? true : false;
	change.sPath = sPath;
	change.del = NULL;	
	
	if(pThis->filter){
		if( ! pThis->filter(pThis->pCtx, &change) )
			return;
	}
	
	pChange = new_GseTreeChange(change.cChange, sPath, change.bIsDir);
	pThis->pChanges->append(pThis->pChanges, pChange);
}

/****************************************************************************/

GseTwNode* GseTreeWatch_makeNode(GseTreeWatch* pThis, GseTwNode* pParent,
                                  const char* sItem, bool bAsChange)
{
	int wd, nError;
	char* sMyPath = NULL;
	char* sSub;
	bool bFreeMyPath = false;
	GseTwNode* pNode = NULL;
	GseTwNode* pSubNode = NULL;
	DIR* pDir;
	struct dirent* pEnt;
	
	if(sItem == NULL)     gse_throw(GSE_XCEPT, GSE_XBADARG);
	if(strlen(sItem) < 1) gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	if(sItem[strlen(sItem) -1] == '/'){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Directory separator not expected "
			             "at the end of '%s'", sItem);
	}

	
	/* Top level interface, if pParent == NULL assume that sItem is an
	   absolute path */
	if(pParent == NULL){
		
		if(sItem[0] != '/'){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Expecting an absolute path "
			             "for top-level tree directories, recieved '%s'", sItem);
		}
		sMyPath = (char*)sItem;
	}
	else{
		
		if(sItem[0] == '/'){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Expecting only the dir-name "
			             "for sub-directories recieved '%s'", sItem);
		}
		
		sMyPath = GseTwNode_getChildPath(pParent, sItem);
		bFreeMyPath = true;
	}
	
	wd = inotify_add_watch(pThis->_fdNotify, sMyPath, pThis->_flags);
	nError = errno;
	
	
	/* Sometimes a create message for a sub-directory comes in after
	   I've already created the watcher, if that occures just return
		the existing node back again, and don't issue a change notice */
	if(pThis->pRev->hasKey(pThis->pRev, wd)){
		if(bFreeMyPath) free(sMyPath);
		return pThis->pRev->getMutable(pThis->pRev, wd);
	}
	
		
	if(wd < 0){
		switch(nError){
			case EACCES:
				gse_error_v("%s: Cannot setup a watch, the directory is not "
				          "readable\n", sMyPath);
				if(bFreeMyPath) free(sMyPath);
				return NULL;
			case ENOMEM:
				gse_error_v("%s: Cannot setup a watch, kernel memory limit "
				          "reached.\n", sMyPath);
				if(bFreeMyPath) free(sMyPath);
				return NULL;
			case ENOSPC:
				gse_error_v("%s: Cannot setup a watch, user limit on number of "
				          "watches reached\n", sMyPath);
				if(bFreeMyPath) free(sMyPath);
				return NULL;
			default:
				if(bFreeMyPath) free(sMyPath);
				gse_throw(CLIB_XCEPT, nError);
		}
	}
		
	pDir = gse_opendir(sMyPath);  /* <-- This can memory leak! */
	
	/* I have a watch, setup the watch sturcture */
	pNode = (GseTwNode*)gse_calloc(1, sizeof(GseTwNode));
	pNode->wd = wd;
	pNode->pParent = pParent;
	pNode->sKey = CS(sItem);
	pNode->del = delete_GseTwNode;
		
	pNode->pSubs = new_GseStrTwNodeDict(NULL);
	
	/* Init the flag holders */
	pNode->bPrune = false;
	pNode->bMovePrune = false;
	pNode->pNewSubs = new_GseStrAry(0);
	
	/* fprintf(stderr, "Creating GseTwNode, wd: %d, key: %s, path: %s\n", pNode->wd,
	        pNode->sKey, sMyPath); */
	
	
	/* Add to reverse lookup */
	pThis->pRev->insert(pThis->pRev, pNode->wd, pNode);
	
	/* Record as a change if needed */
	if(bAsChange)
		GseTreeWatch_addChange(pThis, CS(sMyPath), GSE_TW_NEW|GSE_TW_DIR);
	
	
	/* see if I have any sub-nodes to make */
	while( (pEnt = gse_readdir(pDir)) != NULL){   /* <-- This can memory leak! */
		
		if(strcmp(".", pEnt->d_name) == 0) continue;
		if(strcmp("..", pEnt->d_name) == 0) continue;
		
		sSub = NS("%s/%s", sMyPath, pEnt->d_name);
		if(gse_isdir(sSub)){
			
			pSubNode = GseTreeWatch_makeNode(pThis, pNode, pEnt->d_name, bAsChange);
			if(pSubNode != NULL)
				pNode->pSubs->insert(pNode->pSubs, CS(pEnt->d_name), pSubNode);
		}
		else{
			if(bAsChange)
				GseTreeWatch_addChange(pThis, CS(sSub), GSE_TW_NEW);
		}
		free(sSub);
	}
	
	closedir(pDir);
	
	if(bFreeMyPath) free(sMyPath);
	return pNode;
}

/****************************************************************************/
/* Setup the top node */

/* TODO: Handle recording each directory and file encountered as a create
         if the directory tree comes into existance after this object is
			created in memory */

void GseTreeWatch_setupTopNode(GseTreeWatch* pThis, bool bAsChange)
{
	/* If the top node is not empty, freak out */
	if(pThis->pTop != NULL)
		gse_throw(GSE_XCEPT, GSE_XASSERT);
	
	/* If the directory we are supposed to watch doesn't exist yet just return*/
	if(! gse_isdir(pThis->sTop)) return;

	pThis->pTop = GseTreeWatch_makeNode(pThis, NULL, pThis->sTop, bAsChange);
}

/*****************************************************************************/
/* Deleting nodes in response to changes */

bool GseTreeWatch_pruneNodes(GseTreeWatch* pThis, GseTwNode* pNode)
{	
	/* According to the caller some of these nodes have the bPrune flag set
	   remove them, but don't record the prune as a change */
	const char* sSub = NULL;
	GseTwNode* pSubNode = NULL;
	char* sPath = NULL;
	size_t iRm = 0;
	size_t nRm = 0;
	const char** ppRm = NULL;
	int*   pWd  = NULL;
	GseTreeChange* pChange = NULL;
	
	/* Note, If I'm going to be pruned, my sub-items are too, so set
	   that automatically */
	
	/* 1st found out how many of my sub-dirs are going to disappear, and
	   make space to hold the keys references */
	for(sSub = pNode->pSubs->first(pNode->pSubs); sSub != NULL;
	    sSub = pNode->pSubs->next(pNode->pSubs, sSub)){
	
		pSubNode = pNode->pSubs->getMutable(pNode->pSubs, sSub);
		
		if(pNode->bPrune) pSubNode->bPrune = true;
		if(pNode->bMovePrune) pSubNode->bMovePrune = true;
		
		if(pSubNode->bPrune) nRm++;
	}
	
	if(nRm > 0){
		ppRm = (const char**)gse_calloc(nRm, sizeof(char*));
		pWd  = (int*)gse_calloc(nRm, sizeof(int));
	}
	
	except_cleanup_push(free, ppRm);
	except_cleanup_push(free, pWd);
	
	/* Recurse for all sub-nodes */
	for(sSub = pNode->pSubs->first(pNode->pSubs); sSub != NULL;
	    sSub = pNode->pSubs->next(pNode->pSubs, sSub)){
	
		pSubNode = pNode->pSubs->getMutable(pNode->pSubs, sSub);
		
		if(GseTreeWatch_pruneNodes(pThis, pSubNode)){
			ppRm[iRm] = sSub;
			pWd[iRm]  = pSubNode->wd;
			iRm++;
			
			/* If that sub-directory had a watch, remove it */
			/* inotify_rm_watch(pThis->_fdNotify, pSubNode->wd); */
			
			sPath = GseTwNode_getChildPath(pNode, sSub);
			GseTreeWatch_addChange(pThis, sPath, GSE_TW_DEL|GSE_TW_DIR);
		}
	}
	nRm = iRm;
	
	/* Okay, remove the one's that say they need pruning from both my list
	   and from the reverse lookup list */
	iRm = 0;
	for(iRm = 0; iRm < nRm; iRm++){
		
		pThis->pRev->delPair(pThis->pRev, pWd[iRm]);
		/* fprintf(stderr, "About to delete subnode: %s\n", ppRm[iRm]); */
		pNode->pSubs->delPair(pNode->pSubs, ppRm[iRm]);
	}
	
	except_cleanup_pop(except_call);  /* pWd */
	except_cleanup_pop(except_call);  /* ppRm */
	
	/* If I'm going to be pruned, my children have been deleted already
	   thus generating the needed delete messages, no need to report on
		deleted children here... 
		unless! we are actually undergoing a move instead */
	
	if(pNode->bMovePrune){		
		/* Generate tree invalidation message, skip the normal filtering
		   on this one, they are getting it. */
		sPath = GseTwNode_getPath(pNode);
		pChange = new_GseTreeChange('I', sPath, true);
		pThis->pChanges->append(pThis->pChanges, pChange);
	}
	
	return pNode->bPrune;
}

/*****************************************************************************/
/* Walk the nodes adding from the top down, for each node dir added the      */
/* change needs to be recorded, as this is not done in readChanges for new   */
/* directories.  This is not called when the top node is created             */

void GseTreeWatch_addFlagged(GseTreeWatch* pThis, GseTwNode* pNode)
{
	char* sPath = NULL;
	const char* sSub = NULL;
	int i;
	GseTwNode* pSubNode = NULL;
	
	/* If I have any sub items, add them as creates and record the change */
	for(i = 0; i < pNode->pNewSubs->num; i++){
		
		sPath = GseTwNode_getChildPath(pNode, pNode->pNewSubs->pAry[i]);
		
		if(gse_isdir(sPath)){
			
			/* Sometimes I get create messages for items I already have as
			   there is a race condition between reading the inotify queue
				and setting up new watches, if that happens, just record the
				create message as a change instead. */
			if(pNode->pSubs->hasKey(pNode->pSubs, pNode->pNewSubs->pAry[i])){

				GseTreeWatch_addChange(pThis, CS(sPath), GSE_TW_MOD|GSE_TW_DIR);
			}
			else{
			
				pSubNode = GseTreeWatch_makeNode(pThis, pNode, pNode->pNewSubs->pAry[i],
	 		                                    true);
				if(pSubNode != NULL)
					pNode->pSubs->insert(pNode->pSubs, CS(pNode->pNewSubs->pAry[i]),
					                     pSubNode);
			}
		}
		free(sPath);
	}
	
	/* Let the child nodes check this as well */
	
	for(sSub = pNode->pSubs->first(pNode->pSubs); sSub != NULL;
	    sSub = pNode->pSubs->next(pNode->pSubs, sSub)){
	
		pSubNode = pNode->pSubs->getMutable(pNode->pSubs, sSub);
		
		GseTreeWatch_addFlagged(pThis, pSubNode);
	}
	
	/* Free this node's new item flags */
	if(pNode->pNewSubs->num > 0){
		pNode->pNewSubs->del(pNode->pNewSubs);
		pNode->pNewSubs = new_GseStrAry(0);
	}
}

/*****************************************************************************/
/* Dagnostic handler, event printing */
void GseTreeWatch_prnEvent(GseWdNodeDict* pRev, struct inotify_event* pEvent, 
	                        FILE* pDest)
{
	int nNodeWd;
	char sChange[10];
	const GseTwNode* pNode = NULL;
	const char* sNode = NULL;
	
	strcpy(sChange, "         ");
	
	if(pRev->hasKey(pRev, pEvent->wd)){
		pNode = pRev->get(pRev, pEvent->wd);
		sNode = pNode->sKey;
		nNodeWd = pNode->wd;
	}
	else{
		sNode = "N/A";
		nNodeWd = -1;
	}
	
	
	if(pEvent->mask & IN_CREATE)     sChange[0] = 'C';
	if(pEvent->mask & IN_MOVED_TO)   sChange[1] = 'c';
			
	if(pEvent->mask & IN_MODIFY)     sChange[2] = 'M';
			
	if(pEvent->mask & IN_DELETE)     sChange[3] = 'D';
	if(pEvent->mask & IN_MOVED_FROM) sChange[4] = 'd';
			
	if(pEvent->mask & IN_DELETE_SELF)sChange[5] = 'S';
	if(pEvent->mask & IN_MOVE_SELF)  sChange[6] = 's';
			
	if(pEvent->mask & IN_IGNORED)    sChange[7] = 'R';
			
	if(pEvent->mask & IN_ISDIR)      sChange[8] = '/';
	
	fprintf(pDest, "Node: %s, Node Wd: %d, Wd: %d, Change: '%s'", 
			   sNode, nNodeWd, pEvent->wd, sChange);
	if(pEvent->len)
		fprintf(pDest, "  File: %s\n", pEvent->name);
	else
		fprintf(pDest, "\n");
}

/*****************************************************************************/
/* Given a set of FD with changes pending, read in the changes */

int4 GseTreeWatch_readChanges(GseTreeWatch* pThis)
{
	unsigned int nQueueLen;
	int nErrno;
	struct inotify_event* pEvent = NULL;
	ssize_t i, nRead;
	const GseTwNode* pNode = NULL;
	const GseTwNode* pSubNode = NULL;
	char* sChangePath;
	uint4 uChangeFlags;
	bool bPrune, bAdd;
	GseStrAry* pNewSubs = NULL;
	GseTreeChange* pChange = NULL;
	
	if(ioctl(pThis->_fdNotify, FIONREAD, &nQueueLen) < 0){
		nErrno = errno;
		gse_throw(CLIB_XCEPT, nErrno);
	}
	
	/* fprintf(stderr, "nQueueLen = %d\n", nQueueLen); */
	
	/* Expand or create the bytebuf as needed */
	if(nQueueLen > pThis->bBuf.len){
		if(pThis->bBuf.ptr != NULL) free(pThis->bBuf.ptr);
		if(nQueueLen < 1024)
			pThis->bBuf.len = 1024;
		else
			pThis->bBuf.len = nQueueLen;
		pThis->bBuf.ptr = (byte*)gse_calloc(pThis->bBuf.len, sizeof(byte));
	}
	
	while(true){
		/* memset(pThis->bBuf.ptr, 0, pThis->bBuf.len); */
		nRead = read(pThis->_fdNotify, pThis->bBuf.ptr, pThis->bBuf.len);
		nErrno = errno;
		if(nRead < 0){
			if(nErrno == EINTR)
				continue;
			else
				gse_throw(CLIB_XCEPT, nErrno);
		}
		
		if(nRead < 16){
			gse_warning_v("GseTreeWatch_readChanges: Unexpected return from read,"
			            " nRead = %d", nRead);
			return pThis->pChanges->length(pThis->pChanges);
		}
		
		break;
	}
	
	i = 0;
	bPrune = false;
	bAdd = false;
	while(i < nRead){
		sChangePath = NULL;
		uChangeFlags = 0;
		pEvent = (struct inotify_event*)( pThis->bBuf.ptr + i);
		
		/* GseTreeWatch_prnEvent(pThis->pRev, pEvent, stderr); */
		
		if((pEvent->wd == 0)&&(pEvent->mask == 0))
			goto handled;
		
		/* Sometimes a delete on a higher directory will trigger down to a lower
		   directory and the wd may refer to a non-existant watch, that's okay
			just ignore the message.
		*/
		if(pThis->pRev->hasKey(pThis->pRev, pEvent->wd))
			pNode = pThis->pRev->get(pThis->pRev, pEvent->wd);
		else
			goto handled;
		
		/* Handle each event type */
		
		/* A file in a directory has been deleted */
		if((pEvent->mask & IN_DELETE)&& !(pEvent->mask & IN_ISDIR)){
			uChangeFlags = GSE_TW_DEL;
			sChangePath = GseTwNode_getChildPath(pNode, pEvent->name);
			goto handled;
		}
		
		/* A file in a directory has been modified */
		if((pEvent->mask & IN_MODIFY)&& !(pEvent->mask & IN_ISDIR)){
			uChangeFlags = GSE_TW_MOD;
			sChangePath = GseTwNode_getChildPath(pNode, pEvent->name);
			goto handled;
		}
		
		/* A file in a directory has been created */
		if((pEvent->mask & IN_CREATE)&& !(pEvent->mask & IN_ISDIR)){
			uChangeFlags = GSE_TW_NEW;
			sChangePath = GseTwNode_getChildPath(pNode, pEvent->name);
			goto handled;
		}
		
		/* A file in a directory has been renamed to */
		if((pEvent->mask & IN_MOVED_TO)&& !(pEvent->mask & IN_ISDIR)){
			uChangeFlags = GSE_TW_NEW;
			sChangePath = GseTwNode_getChildPath(pNode, pEvent->name);
			goto handled;
		}
		
		/* A file in a directory has been renamed from */
		if((pEvent->mask & IN_MOVED_FROM)&& !(pEvent->mask & IN_ISDIR)){
			uChangeFlags = GSE_TW_DEL;
			sChangePath = GseTwNode_getChildPath(pNode, pEvent->name);
			goto handled;
		}
		
		/* Directory handling change notice creation happens in sub-functions */
		
		
		/* A watched directory was removed, there is nothing to do in this case
		   because, we will handle this on the delete self */
		if(pEvent->mask & IN_IGNORED){
			((GseTwNode*)pNode)->bPrune = true;
			bPrune = true;
			goto handled;
		}
			
		/* A watched directory was deleted, mark it for pruning.  The prunner
		   function will handle making change notices as needed. */
		if(pEvent->mask & IN_DELETE_SELF){
			((GseTwNode*)pNode)->bPrune = true;
			bPrune = true;
			goto handled;
		}
		
		/* I have been moved, mark me as deleted, if I'm the top node it's
		   okay, the bPruned section below will handle my deltion */
		if(pEvent->mask & IN_MOVE_SELF){
			((GseTwNode*)pNode)->bPrune = true;
			((GseTwNode*)pNode)->bMovePrune = true;
			bPrune = true;
			goto handled;
		}
		
		/* A subdirectory has been created */
		if((pEvent->mask & IN_CREATE) && (pEvent->mask & IN_ISDIR)){
			if(pEvent->len == 0)
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			pNewSubs = ((GseTwNode*)pNode)->pNewSubs;
			pNewSubs->add(pNewSubs, CS(pEvent->name));
			bAdd = true;
			goto handled;
		}
		
		/* A sub-directory has been coppied in, add watchers */
		if((pEvent->mask & IN_MOVED_TO) && (pEvent->mask & IN_ISDIR)){
			if(pEvent->len == 0)
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			pNewSubs = ((GseTwNode*)pNode)->pNewSubs;
			pNewSubs->add(pNewSubs, CS(pEvent->name));
			bAdd = true;
			goto handled;
		}
		
		/* A sub-directory has been deleted, mark it for pruning */
		if((pEvent->mask & IN_DELETE) && (pEvent->mask & IN_ISDIR)){
			if(pEvent->len == 0)
				gse_throw(GSE_XCEPT, GSE_XASSERT);
			
			if( pNode->pSubs->hasKey(pNode->pSubs, pEvent->name) ){
				pSubNode = pNode->pSubs->get(pNode->pSubs, pEvent->name);
				((GseTwNode*)pSubNode)->bPrune = true;
				bPrune = true;
			}
			goto handled;
		}
		
		/* A sub-directory has been renamed, the old name needs no be marked 
		   as deleted, set it up for pruning.  If it had a watch the 
			IN_MOVE_SELF message will handle it, else it goes unnoticed! */
		if((pEvent->mask & IN_MOVED_FROM) && (pEvent->mask & IN_ISDIR))
			goto handled;
		
		
		gse_warning("Unhandled inotify message");
		        
handled:
		if(sChangePath != NULL)
			GseTreeWatch_addChange(pThis, sChangePath, uChangeFlags);
				
		i += (sizeof(struct inotify_event)) + pEvent->len;
	}
	
	/* Walk the nodes deleting from the bottom up */
	if(bPrune){
		if( GseTreeWatch_pruneNodes(pThis, pThis->pTop) ){
			
			/* inotify_rm_watch(pThis->_fdNotify, pThis->pTop->wd); */

			GseTreeWatch_addChange(pThis, CS(pThis->sTop), GSE_TW_DEL|GSE_TW_DIR);
			
			if(pThis->pTop->bMovePrune){				
				pChange = new_GseTreeChange('I', CS(pThis->sTop), true);
				pThis->pChanges->append(pThis->pChanges, pChange);
			}
			
			pThis->pRev->delPair(pThis->pRev, pThis->pTop->wd);
			pThis->pTop->del(pThis->pTop);
			pThis->pTop = NULL;
		}
	}
	
	/* Walk the nodes adding from the top down */
	if(bAdd){
		GseTreeWatch_addFlagged(pThis, pThis->pTop);
	}
	
	return pThis->pChanges->length(pThis->pChanges);
}

/*****************************************************************************/
/* Waiting for a file change */

void GseTreeWatch_waitForChange(GseTreeWatch* pThis)
{
	int nReady, nErrno;
	fd_set fdSet;
	
	if(pThis->pChanges->length(pThis->pChanges) > 0)
		return;
	
	/* Handle waiting for the named directory to come into being */
	if(pThis->pTop == NULL){
		while(!gse_isdir(pThis->sTop) && pThis->bRestartSelect){
			gse_millisleep(1000, false);
		}
		if(! pThis->bRestartSelect) return;
		
		GseTreeWatch_setupTopNode(pThis, true);
		return;
	}
	
	
	FD_ZERO(&fdSet);
	FD_SET(pThis->_fdNotify, &fdSet);
	
	while(true){
		nReady = select(pThis->_fdNotify+1, &fdSet, NULL, NULL, NULL);
		nErrno = errno;
		
		if((nReady == -1)&&(nErrno == EINTR)){ 
			
			if(pThis->bRestartSelect)
				continue;
			else
				break;
		}
		
		if(nReady == -1) 
			gse_throw(CLIB_XCEPT, nErrno);
		
		if(nReady == 0)
			gse_throw(GSE_XCEPT, GSE_XINCONCEIVABLE);
		
		break;
	}
	
	GseTreeWatch_readChanges(pThis);
}

/*****************************************************************************/
/* Check for changes without sleeping */

int4 GseTreeWatch_numChanges(GseTreeWatch* pThis)
{
	/* Check for changes without waiting */
	int nReady, nErrno;
	fd_set fdSet;
	struct timeval tv;
	
	/* Handle bringing the top directory into being */
	if(pThis->pTop == NULL){
		if(!gse_isdir(pThis->sTop))
			return pThis->pChanges->length(pThis->pChanges);
			
		GseTreeWatch_setupTopNode(pThis, true);
	}
	
	/* Now check to see if any changes were recorded */

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	
	FD_ZERO(&fdSet);
	FD_SET(pThis->_fdNotify, &fdSet);
	
	while(true){
		nReady = select(pThis->_fdNotify+1, &fdSet, NULL, NULL, &tv);
		nErrno = errno;
		
		if((nReady == -1)&&(nErrno == EINTR)){ 
			
			if(pThis->bRestartSelect)
				continue;
			else
				break;
		}
		
		if(nReady == -1) 
			gse_throw(CLIB_XCEPT, nErrno);
				
		break;
	}
	
	if(nReady == 0)
		return pThis->pChanges->length(pThis->pChanges);
	else
		return GseTreeWatch_readChanges(pThis);
}

/*****************************************************************************/
/* Who interrups my slumber... */
#ifdef GSE_POSIX_THREADS
void GseTreeWatch_interrupt(GseTreeWatch* pThis, pthread_t threadID)
{
	/* Set the halt waiting flag then interrupt any system calls */
	pThis->bRestartSelect = false;
	pthread_kill(threadID, SIGUSR2);
}
#endif

/*****************************************************************************/
/* If you want change, vote for Ron Paul or Dennis Kucinich, Obama is a */
/* centrist */

GseTreeChange* GseTreeWatch_getChange(GseTreeWatch* pThis)
{
	if( pThis->pChanges->length(pThis->pChanges) > 0)
		return pThis->pChanges->pop(pThis->pChanges, 0);
	else
		return NULL;
}


/****************************************************************************/
/* destructor */

void delete_GseTreeWatch(void* vpThis)
{
	GseTreeWatch* pThis = (GseTreeWatch*)vpThis;
	
	pThis->pChanges->del(pThis->pChanges);               /* free change list */
	
	if( pThis->pTop != NULL) 
		pThis->pTop->del(pThis->pTop);                          /* free nodes */
	
	if(pThis->_fdNotify > -1) 
		close(pThis->_fdNotify);                          /* close discriptor */
	
	if(pThis->bBuf.ptr != NULL)
		free(pThis->bBuf.ptr);                               /* free read buf */
	
	if(pThis->sTop != NULL)
		free(pThis->sTop);                                    /* free dir str */
	
	pThis->pRev->del(pThis->pRev);                   /* free rev lookup dict */
	
	free(pThis);                                                  /* free me */
}

/****************************************************************************/
/* Change filter setting */

gse_tw_filter_t GseTreeWatch_setFilter(GseTreeWatch* pThis,
	                                    gse_tw_filter_t func, void* pCtx)
{
	gse_tw_filter_t pRet = pThis->filter;
	pThis->filter = func;
	pThis->pCtx = pCtx;
	
	return pRet;
}

/****************************************************************************/
/* Constructor */

GseTreeWatch* new_GseTreeWatch(const char* sPath, uint4 uFlags,
                               gse_tw_filter_t filter, void* pFilterCtx)
{
	int nErrno = 0;
	bool bExists = false;
	GseTreeWatch* pThis;
	char* sName;
	
	/* make sure sPath is an absolute path */
	if(sPath == NULL)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sPath is null");
	
	if(sPath[0] != '/')
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "'%s' is not an absolute path", sPath);
	
	
	/* If sPath exists, make sure it's a directory */
	if( gse_exists(sPath)){
		bExists = true;
		if(! gse_isdir(sPath))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "'%s' is not a directory.", sPath);
	}
			
	pThis = (GseTreeWatch*) gse_calloc(1, sizeof(GseTreeWatch));
	
	pThis->waitForChange = GseTreeWatch_waitForChange;
	pThis->numChanges    = GseTreeWatch_numChanges;
	pThis->getChange     = GseTreeWatch_getChange;	
#ifdef GSE_POSIX_THREADS
	pThis->interrupt     = GseTreeWatch_interrupt;
#endif
	pThis->setFilter     = GseTreeWatch_setFilter;
	
	pThis->del = delete_GseTreeWatch;
	
	pThis->sTop = CS(sPath);
	pThis->pTop = NULL;
	pThis->uFlags = uFlags;
	pThis->bBuf.len = 0;
	pThis->bBuf.ptr = NULL;
	pThis->filter = filter;
	pThis->pCtx = pFilterCtx;
			
	/* Setup the change list array */
	pThis->pChanges = new_GseTwcList();
	
	/* Setup the reverse lookup dictionary */
	sName = NS("'%s' reverse lookup", sPath);
	pThis->pRev = new_GseWdNodeDict(sName);
	free(sName);
	pThis->pRev->disownAllVals(pThis->pRev);  /* Only has pointers, not objs */
	
	pThis->bRestartSelect = true;
			
	/* Setup the app-global notify event queue, or throw */
	pThis->_fdNotify = -1;
	pThis->_fdNotify = inotify_init();
	nErrno = errno;
	if(pThis->_fdNotify < 0){
		pThis->del(pThis);
		gse_throw_m(CLIB_XCEPT, nErrno, "Couldn't init the kernel inotify event queue.");
	}
	
	/* Setup the inotify flags, change loging is filtered based on the
	   user flags, but we need all of these to work properly */
	pThis->_flags = IN_CREATE|IN_MOVED_TO|IN_DELETE|IN_MOVED_FROM|IN_MODIFY|
	                IN_DELETE_SELF|IN_MOVE_SELF;
	
	if(pThis->_flags == 0){
		pThis->del(pThis);
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "uFlags contains no options");
	}
		
	/* If bExists, setup the watch tree */
	if(bExists){ 
		except_cleanup_push(pThis->del, pThis);
		if(uFlags & GSE_TW_INITIAL_C)
			GseTreeWatch_setupTopNode(pThis, true);
		else
			GseTreeWatch_setupTopNode(pThis, false);
		except_cleanup_pop(except_no_call);
	}
	
	return pThis;
}

/****************************************************************************/
/* instantiate the tree-watch node dictionaries */


#define C GseStrTwNodeDict
#define K char
#define V GseTwNode

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcmp(a, b)

#include "ptrDictTplt.c"

#undef C
#undef K
#undef V

/****************************************************************************/
/* Instantiate the wd-node reverse lookup dictionairy */

#define C GseWdNodeDict
#define K int
#define V GseTwNode

#define DEL_VALS
#define KEY_FMT "fd %d"
#include "dictTplt.c"

#undef C
#undef K
#undef V

/****************************************************************************/
/* instantiate the change list object */

#define C GseTwcList
#define I GseTreeChange

#define CASCADE_DEL

#include "listTplt.c"

#undef CASCADE_DEL
#undef I
#undef C

