/** @file GseTypes.h An aid for remembering type widths. */
#ifndef _GSE_TYPES_H_
#define _GSE_TYPES_H_

#include <float.h>
#include <stdbool.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/
/** @defgroup mod_types Types 
 * Unambiguous Data Types.
 *
 * In addition to the typedefs below, @b bool and @b bytebuf_t are used
 * throughout the library.
 *
 * The @c bool type was introduced in the C99 standard.  It has two legal
 * values, @c true and @c false and is defined in the system header 
 * @c <stdbool.h>.
 *
 * The @c bytebuf_t type and it associated @c const version are defined as:
 * @code
 * typedef struct bytebuf_t{ byte* ptr; size_t len; } bytebuf_t;
 * typedef struct cbytebuf_t { const byte* ptr; size_t len; } cbytebuf_t;
 * @endcode
 *
 * This structure is often used to return a buffer and it's length as a 
 * single entity from function call.  This allows for cleaner code as the
 * outputs of many functions no longer need to appear in thier argument lists.
 */
/*@{*/

/* Specifing int always gives a 4 byte integer on 32 and 64 bit, and 
 long long int always gives an 8 byte integer, but this is hard to
 remember so just use typedefs */

#define GSE_UINT8_MAX ULLONG_MAX

/** Always a 64-bit unsigned integer on 32 or 64 platforms */
typedef unsigned long long int uint8;

#define GSE_INT8_MIN LLONG_MIN
#define GSE_INT8_MAX LLONG_MAX

/** Always a 64-bit signed integer on 32 or 64 platforms */
typedef long long int int8;

/** Always a 32-bit unsigned integer on 32 or 64 platforms */
typedef unsigned int uint4;

/** Always a 32-bit signed integer on 32 or 64 platforms */
typedef int int4;

/** Always a 16-bit unsigned integer on 32 or 64 platforms */
typedef unsigned short int uint2;

/** Always a 16-bit signed integer on 32 or 64 platforms */
typedef short int int2;

/** Always an 8-bit unsigned integer on 32 or 64 platforms */
typedef unsigned char byte;

/** Always an 8-bit unsigned integer on 32 or 64 platforms */
typedef unsigned char uint1;

/** Always an 8-bit unsigned integer on 32 or 64 platforms */
typedef signed char int1;

/** Always a 32-bit floating point value on 32 or 64 platforms */
typedef float float4;

/** Maximum representable finite number in a 4 byte float */
#define GSE_FLOAT4_MAX FLT_MAX

/** Minimum normalized positive number in a 4 byte float */
#define GSE_FLOAT4_MIN FLT_MIN

/** Always a 64-bit floating point value 32 or 64 platforms */
typedef double float8;

/** Maximum representable finite number in an 8 byte float */
#define GSE_FLOAT8_MAX DBL_MAX

/** Minimum normalized positive number in an 8 byte float */
#define GSE_FLOAT8_MIN DBL_MIN


/* A byte pointer and it's associate length */
typedef struct bytebuf_t{ byte* ptr; size_t len; } bytebuf_t;

/* A constant byte pointer and it's associated length */
typedef struct cbytebuf_t { const byte* ptr; size_t len; } cbytebuf_t;

/** "Cast" a bytebuf_t to the constant version cbytebuf_t */
cbytebuf_t const_bytebuf(bytebuf_t buf);

/** Make a copy of the contents of a bytebuf 
 *
 * Space for the copy is allocated on the heap.
 */
bytebuf_t copy_cbytebuf(cbytebuf_t src);

/** Make a copy of the contents of a bytebuf 
 *
 * Does not alter the original. Space for the copy is allocated on the heap.
 */
bytebuf_t copy_bytebuf(bytebuf_t src);


/* A character pointer and it's associated length */
typedef struct { char* ptr; size_t len; } charbuf_t;

/* A constant character pointer and it's associated length */
typedef struct { const char* ptr; size_t len; } ccharbuf_t;

/* A set of strings pointer and it's associated length */
typedef struct { char** ptr; size_t len; } strbuf_t;


/* A 4-byte float array and it's associated length */
typedef struct { float4* ptr; size_t len; } float4buf_t;

/* A constant 4-byte float array and it's associated length */
typedef struct { const float4* ptr; size_t len; } cfloat4buf_t;

/* An 8-byte float array and it's associated length */
typedef struct { float8* ptr; size_t len; } float8buf_t;

/* A constant 8-byte float array and it's associated length */
typedef struct { const float8* ptr; size_t len; } cfloat8buf_t;

/* A 4-byte int array and it's associated length */
typedef struct { int4* ptr; size_t len; } int4buf_t;

/* A constant 4-byte int array and it's associated length */
typedef struct { const int4* ptr; size_t len; } cint4buf_t;


#ifndef _GseTypes_C_
/** A NULL byte buffer */
extern cbytebuf_t CBBUF_NULL;

/** A NULL constant byte buffer */
extern bytebuf_t BBUF_NULL;

#endif


/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GSE_TYPES_H_ */


