/* Some helpful routines for porting to Windows, these are POSIX look alikes
   that help the Gse Libraries do something useful and are not athoritative in
	any way */
	
#ifndef __MINGW32__
#error These functions are only for use on the Win32 API
#endif

#include <windows.h>

/* The following is only needed for dladdr replacement. It is a free download
 * to get the Dbghelp SDK from Microsoft see:
 *   
 */
#include <imagehlp.h>  

#include "GseExcept.h"
#include "GseWinSvc.h"
#include "BfdContext.h"

uid_t getuid()
{

	HANDLE hToken = NULL;
	DWORD dwBufferSize = 0;
	PTOKEN_USER pTokenUser = NULL;
	int uid;

	/* Open the access token associated with the calling process. */
	if(!OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
		gse_throw_m(WIN32_XCEPT, GetLastError(), "OpenProcessToken failed.");

	/* get the size of the memory buffer needed for the SID */
	GetTokenInformation(hToken, TokenUser, NULL, 0, &dwBufferSize);

	pTokenUser = (PTOKEN_USER)malloc(dwBufferSize);
	memset(pTokenUser, 0, dwBufferSize);

	/* Retrieve the token information in a TOKEN_USER structure. */
	if (!GetTokenInformation(hToken, TokenUser, pTokenUser, dwBufferSize,
	    &dwBufferSize))
		 gse_throw_m(WIN32_XCEPT, GetLastError(), "GetTokenInformation failed.");
		
	CloseHandle(hToken);

	if (!IsValidSid(pTokenUser->User.Sid)) {
		free(pTokenUser);
		gse_throw_m(WIN32_XCEPT, ERROR_INVALID_SID, "The owner SID is invalid.");
	}

	uid = *GetSidSubAuthority(pTokenUser->User.Sid,
	*GetSidSubAuthorityCount(pTokenUser->User.Sid) - 1);
	
	free(pTokenUser);
	CloseHandle(hToken);

	return uid;
}

/*****************************************************************************/
/* GNU backtrace and dlladdr replacements for win32                          */
/*                                                                           */
/* Some handy glibc -> Win32 equivalencies                                   */
/* dlopen  --> HMODULE LoadLibrary(char *path)                               */
/* dlsym   --> void *GetProcAddress(HMODULE lib, char *sym)                  */
/* dlclose --> void FreeLibrary(HMODULE lib)                                 */
/*                                                                           */

/* Note: This is supposed to be defined in WinNT.h if WINVER >= WindowsXP:
 *
 * VOID WINAPI RtlCaptureContext(
 *   __out  PCONTEXT ContextRecord
 * );
 *
 * 
 * but it ain't, so load the dll manually and use a typedef.
 */
		 
typedef void (*RtlCaptureContext)(CONTEXT* ContextRecord);
 
int g_winsvc_init = 0;
BfdContext* g_pBfd = NULL;
RtlCaptureContext g_RtlCaptureContext = NULL;


#ifdef GSE_POSIX_THREADS
#error thread save version of dladdr is not yet implimented
#else
#define LOCK()
#define UNLOCK()
#endif

void gse_winsvc_init(){
	char sExeName[MAX_PATH + 1];
	HMODULE hKernel32;
	
	if(g_winsvc_init > 1){
		g_winsvc_init += 1;
		return;
	}
	
	/* Initialize searching in non-dlls */
	sExeName[MAX_PATH] = '\0';
	GetModuleFileName(NULL, sExeName, MAX_PATH);
	if( (g_pBfd = new_BfdContext(sExeName)) == NULL){
		fprintf(stderr, "CRITICAL: Cannot Initalize thi Bfd Context, exiting\n");
		exit(66);
	}
	
	/* Initialize searching in dlls */
	if( (hKernel32 = LoadLibrary("Kernel32.dll")) == NULL){
		fprintf(stderr, "CRITICAL: Cannot load Kernel32.dll, exiting\n");
		exit(66);
	}
	if( (g_RtlCaptureContext = (RtlCaptureContext)GetProcAddress(hKernel32,
		  "RtlCaptureContext")) == NULL){
		fprintf(stderr, "CRICITAL: Cannot get address of RtlCaptureContext"
			     "in Kernel32.dll, exiting...\n");
		exit(66);
	}
	
	SymInitialize(GetCurrentProcess(), 0, true);
}

void gse_winsvc_deinit(){
	g_winsvc_init -= 1;
	if(g_winsvc_init > 0) return;
	
	SymCleanup(GetCurrentProcess());
	
	g_RtlCaptureContext = NULL;
			
	g_pBfd->del(g_pBfd);
	g_pBfd = NULL;
}

/*****************************************************************************/
/* Getting the function pointer stack                                        */
/* Thanks to Edd Dawson. for the C++ version                                 */
int backtrace(void** buffer, int size)
{
	STACKFRAME frame;
	CONTEXT context;
	DWORD mType;
	HANDLE hProcess;
	HANDLE hThread;
	
	int nFrame;
			
	memset(&context, 0, sizeof(CONTEXT));
	context.ContextFlags = CONTEXT_FULL;
	
	/* Get the current thread context */
	g_RtlCaptureContext(&context);
	
	/* init STACKFRAME for first call */
	memset(&frame, 0, sizeof(frame));
#ifdef _M_IX86
	/* normally, call ImageNtHeader() and use machine info from PE header */
	mType = IMAGE_FILE_MACHINE_I386;
	frame.AddrPC.Offset = context.Eip;
	frame.AddrPC.Mode = AddrModeFlat;
	frame.AddrFrame.Offset = context.Ebp;
	frame.AddrFrame.Mode = AddrModeFlat;
	frame.AddrStack.Offset = context.Esp;
	frame.AddrStack.Mode = AddrModeFlat;
#elif _M_X64
	mType = IMAGE_FILE_MACHINE_AMD64;
	frame.AddrPC.Offset = context.Rip;
	frame.AddrPC.Mode = AddrModeFlat;
	frame.AddrFrame.Offset = context.Rsp;
	frame.AddrFrame.Mode = AddrModeFlat;
	frame.AddrStack.Offset = context.Rsp;
	frame.AddrStack.Mode = AddrModeFlat;
#elif _M_IA64
	mType = IMAGE_FILE_MACHINE_IA64;
	frame.AddrPC.Offset = context.StIIP;
	frame.AddrPC.Mode = AddrModeFlat;
	frame.AddrFrame.Offset = context.IntSp;
	frame.AddrFrame.Mode = AddrModeFlat;
	frame.AddrBStore.Offset = context.RsBSP;
	frame.AddrBStore.Mode = AddrModeFlat;
	frame.AddrStack.Offset = context.IntSp;
	frame.AddrStack.Mode = AddrModeFlat;
#else
#error "Platform not supported!"
#endif

	hProcess = GetCurrentProcess();
	hThread  = GetCurrentThread();
	
	nFrame = 0;
	while(StackWalk(mType, hProcess, hThread, &frame, &context, 0, 
			          SymFunctionTableAccess, SymGetModuleBase, 0)){
		nFrame += 1;
		
		if(nFrame > size){
			nFrame -= 1;
			break;
		}
		
		buffer[nFrame - 1] = (void*)frame.AddrPC.Offset;
	}
	
	
	/* Maybe but this in a thread free function: 
	 * FreeLibrary(hKernel32);
	 */
	return nFrame;
}

/*****************************************************************************/
/* Getting a function and file name given an address                         */
		
int dladdr(void *addr, Dl_info *info){
	HANDLE hProcess;
	DWORD dwModuleBase; /* <-- Won't work on 64bit windows */
	DWORD displacement;
	char sModuleName[MAX_PATH + 1];
	sModuleName[MAX_PATH] = '\0';
	
	hProcess = GetCurrentProcess();
	
	/* Init the structure to nothing */
	info->dli_fname = NULL;
	info->dli_fbase = NULL;
	info->dli_sname = NULL;
	info->dli_saddr = NULL;
	info->dli_salloc = 0;
	
	LOCK();
	
	dwModuleBase = SymGetModuleBase(hProcess, (DWORD)addr);
	info->dli_fbase = (void*)dwModuleBase;
	if(! GetModuleFileNameA((HMODULE)dwModuleBase, sModuleName, MAX_PATH)) return 0;
	
	info->dli_fname = (const char*) calloc(MAX_PATH + 1, sizeof(char));
	memcpy((char*)info->dli_fname, sModuleName, MAX_PATH);
	
	/* First assume that name is in the current mingw compiled executable */
	info->dli_sname = g_pBfd->getFuncName(g_pBfd, (unsigned long)addr);
   
	/* Name might be in a DLL, try to get it */
   if(!(info->dli_sname)){
		
		displacement = 0;
		char symbol_buffer[sizeof(IMAGEHLP_SYMBOL) + 255];
		symbol_buffer[sizeof(IMAGEHLP_SYMBOL) + 254] = '\0';
		
		IMAGEHLP_SYMBOL* pSymbol = (IMAGEHLP_SYMBOL*)symbol_buffer;
		
		pSymbol->SizeOfStruct = sizeof(IMAGEHLP_SYMBOL) + 255;
      pSymbol->MaxNameLength = 254;
		if(SymGetSymFromAddr(hProcess, (DWORD)addr, &displacement, pSymbol)){
			
			info->dli_sname = (const char*) calloc(pSymbol->MaxNameLength + 1, 1);
			memcpy((char*)info->dli_sname, pSymbol->Name, pSymbol->MaxNameLength);
			info->dli_salloc = 1;
		}
	}
	
	UNLOCK();
	
	return 1; /* True */
}

#ifdef GSE_TEST
int main(int argc, char** argv)
{
	gse_thread_init(GSE_WARNING, NULL);
	
	printf("Current user ID: %d", getuid());
	return 0;
}

#endif
