#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#ifdef __MINGW32__
#include <windows.h>
/* Use windows API millisecond sleep */
#define sleep(n) Sleep(1000 * n)
#endif

#include <errno.h>
#include <stdlib.h> 
#include <stdio.h>
#include <math.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "GseWrap.h"
#include "GseExcept.h"
#include "GseConv.h"

void* gse_malloc(size_t size){
	void* pRet;
	
	if( (pRet = malloc(size)) == NULL ) 
		gse_throw_fm(CLIB_XCEPT, ENOMEM, "Could not allocate %lu bytes of memory.",
		           size);
	
	return pRet;
}

void* gse_calloc(size_t nmemb, size_t size){
	void* pRet;
	
	if( (pRet = calloc(nmemb, size)) == NULL ) 
		gse_throw_fm(CLIB_XCEPT, ENOMEM, "Could not allocate %lu bytes of memory.",
		           nmemb * size);
	
	return pRet;
}

void* gse_realloc(void *ptr, size_t size){
	void* pRet;
	
	pRet = realloc(ptr, size);
	if((pRet == NULL)&&(size != 0)) 
		gse_throw_fm(CLIB_XCEPT, ENOMEM, "Could not reallocate %lu bytes of "
				       "memory.", size);
	return pRet;
}


size_t gse_poll_fread(void *ptr, size_t size, size_t nmemb, FILE *stream){
	long idx;
	size_t uRead;
	
	if( (idx = ftell(stream)) < 0)
		gse_throw_m(CLIB_XCEPT, errno, "Could not get file pointer position");
	
	while( (uRead = fread(ptr, size, nmemb, stream)) < nmemb ){
		if(ferror(stream)) 
			clearerr(stream);
		
		if( fseek(stream, idx, SEEK_SET) < 0)
			gse_throw_m(CLIB_XCEPT, errno, "Could not set file pointer position");

		sleep(1);	
	}
	
	return uRead;
}

/* Since errors cause a throw, there is no reason to introduce a signed
 * type to the user */
size_t gse_lseek(int fildes, size_t offset, int whence){
	off_t nPos;
	off_t nOffset;
	size_t uPos;
			
	gse_storeUinS(offset, &nOffset, sizeof(nOffset));
	
	if( (nPos = lseek(fildes, offset, whence)) == (off_t)-1){
		if( (whence == SEEK_CUR)&&(offset == 0))
			gse_throw_m(CLIB_XCEPT,errno,"Could not get the file pointer position");
		else
			gse_throw_m(CLIB_XCEPT,errno,"Could not set the file pointer position");
	}
	
	gse_storeSinU(nPos, &uPos, sizeof(uPos));
	
	return uPos;
}


FILE* gse_fopen(const char *path, const char *mode){
	FILE* pFile;
	int nErrno;
	
	if( (pFile = fopen(path, mode)) == NULL){
		nErrno = errno;
		
		if(path != NULL)
			gse_throw_fm(CLIB_XCEPT, nErrno, "Could not open file '%s'", path);
		else
			gse_throw_m(CLIB_XCEPT, nErrno, "Path string pointer was NULL"); 
	}
	return pFile;
}

int gse_open(const char *pathname, int flags, mode_t mode){
	int fd;
	int nErrno;
	if( (fd = open(pathname, flags, mode)) == -1){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not open file '%s'", pathname);
	}
	return fd; 
}

int gse_open_ro(const char *pathname){
	int fd;
	int nErrno;
	if( (fd = open(pathname, O_RDONLY)) == -1){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not open file '%s'", pathname);
	}
	return fd; 
}

void gse_remove(const char *pathname)
{
	int nErrno;
	if(remove(pathname) != 0){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not remove file '%s'", pathname);
	}
}

void gse_rename(const char* oldpath, const char* newpath)
{
	int nErrno;
	if(rename(oldpath, newpath) != 0){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not rename '%s' to '%s'",
		             oldpath, newpath);
	}
}

DIR* gse_opendir(const char* name)
{
	DIR* pDir;
	int nErrno;
	
	if( (pDir = opendir(name)) == NULL){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not open directory '%s'", name);
	}
	
	return pDir;
}

struct dirent* gse_readdir(DIR* pDir)
{
	int nErrno;
	struct dirent* pRet;
	errno = 0;
	pRet = readdir(pDir);
	
	if((pRet == NULL)&&(errno != 0)){
		nErrno = errno;
		gse_throw_m(CLIB_XCEPT,nErrno,"Could not get the next directory entry.");
	}
	return pRet;
}

void gse_stat(const char* path, struct stat* buf)
{
	int nErrno;
	
	if(stat(path, buf) < 0){
		nErrno = errno;
		gse_throw_fm(CLIB_XCEPT, nErrno, "Could not stat file '%s'", path);
	}
}

