/** @file GseWrap.h Error to exception wrappers for common clib functions */
#ifndef _GseWrap_H_
#define _GseWrap_H_

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>

#include <rpwgse/GseTypes.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @addtogroup mod_rpwgse */
/*@{*/
/** @defgroup mod_wrap CLib Wrappers
 * Exception wrappers for common clib functions.
 *
 * Wrappers for common Clib functions so that exceptions are thrown instead
 * of returning error values.  This allows for application code which does
 * not have to check error returns for every call.  
 * 
 * On an error condition, functions in this module throw exceptions in the 
 * group CLIB_EXCEPT with the exception code set to the value of @b errno.
 * Thus when catching exceptions from these functions the following is valid:
 * 
 * @code 
 *   strerror_r(except_code(except), buf, len)
 * @endcode
 * 
 * The default exception printer gse_print_except() takes this into account.
 *
 * @see mod_exceptions for more details on exception group and type codes.
 */ 
/*@{*/


/** Exception based wrapper for malloc(3).
 * @throws (CLIB_XECPT, ENOMEM) on an error instead of returning NULL.
 */
void* gse_malloc(size_t size);

/** Exception based wrapper for calloc(3).
 * @throws (CLIB_XECPT, ENOMEM) on an error instead of returning NULL.
 */
void* gse_calloc(size_t nmemb, size_t size);

/** Exception based wrapper for realloc(3).
 * @throws (CLIB_XECPT, ENOMEM) on an error instead of returning NULL
 * when @b size is > 0.
 */
void* gse_realloc(void *ptr, size_t size);

/** Works as standard fread, but can keep polling for data after EOF */
size_t gse_poll_fread(void *ptr, size_t size, size_t nmemb, FILE *stream);


/** Exception based wrapper for fopen(3).
 * @throws (CLIB_XCEPT, errno) instead of returning NULL on an error 
 */
FILE* gse_fopen(const char *path, const char *mode);

/** Exception based wrapper for open(2).
 * @throws (CLIB_XCEPT, errno) instead of returning -1 on an error 
 */
int gse_open(const char *pathname, int flags, mode_t mode);

/** Exception based wrapper for open(2), with read only options
 * @throws (CLIB_XCEPT, errno) instead of returning -1 on an error 
 */
int gse_open_ro(const char *pathname);

/** Exception based wrapper for remove(3).
 * @throws (CLIB_XCEPT, errno) instead of returning -1 on an error.
 */
void gse_remove(const char *pathname);

/** Exception based wrapper for rename(2).
 * @throws (CLIB_XCEPT, errno) instead of returning -1 on an error.
 */
void gse_rename(const char* oldpath, const char* newpath);

/** Exception based wrapper for opendir(3).
 * @throws (CLIB_XCEPT, errno) instead of returning NULL on an error.
 */
DIR* gse_opendir(const char* name);

/** Exception baesd wrapper for readdir(3).
 * Alleviates the need to set errno to 0 before each call inorder to
 * detect errors.  A NULL return is still possible, it just means that
 * the end of the directory has been reached.
 *
 * @throws (CLIB_XCEPT, errno) instead of returning NULL and setting errno.
 */
struct dirent* gse_readdir(DIR* pDir);

/** Exception based wrapper for stat(2).
 * @throws (CLIB_XCEPT, errno) for all the reasons stat(2) returns -1 
 */
void gse_stat(const char* path, struct stat* buf);

/** Wrapper for lseek.
 *
 * @return Exceptions are thrown on errors, thus the return is aways
 * a value >= 0
 */
size_t gse_lseek(int fildes, size_t offset, int whence);

/** Macro simplifier for gse_lseek */
#define gse_ltell(F) gse_lseek(F, 0, SEEK_CUR)


/** Wrapper for fwrite
 * 
 * @throws (GSE_XCEPT, GSE_XIOFAIL) on an error, instead of returning 
 *         a short item count.
 */
void gse_fwrite(const void* ptr, size_t size, size_t nmemb, FILE* stream);

/** Wrapper for fputc
 *
 * @throws (GSE_XCEPT, GSE_XIOFAIL) on an error, instead of returning EOF
 */
void gse_fputc(int c, FILE* stream);

/** Wrapper for fputs
 *
 * @throws (GSE_XCEPT, GSE_XIOFAIL) on an error instead of returning EOF
 */
void gse_fputs(const char* s, FILE* stream);
 

/*@}*/
/*@}*/

#ifdef __cplusplus
}
#endif

#endif /* _GseWrap_H_ */
