#ifndef _MASH_DEF
#define _MASH_DEF
#define _MASH_IN(x, y)  x ## y
#define _MASH(x, y) _MASH_IN(x, y)
#endif /* _MASH_DEF */


/* Put this inside a .page file describing using the ptrDictTplt
 *
 *  - S = Struct Definition
 *  - C = Class Type
 *  - K = Key Type  Can be anything of width size_t or smaller.
 *  - V = Data Type
 *
 * The declaration include, "dictTplt.h" requires that dict.h
 * and GseTypes.h have already been included.
 *
 * The instatiation include, "dictTplt.c" requires that GseWrap.h
 * and GseExcept.h have already been included.
 */


/** Generic typed 1 to 1 dictionary structure.
 * 
 * This dictionary assumes ownership of all data pointers, and 
 * frees or deletes them when the @a del member is called.  To prevent
 * a specific data item from being deleted use @a disown_data.  To prevent
 * an instance of this dictionary from managing the memory for all
 * current and future data use @a disown_all_data.
 *
 * Generated from the Dictionary Template in "dictTplt.h"
 * @n @n
 */
typedef struct S{

	/** Get the dictionary name, as defined during construction.
	 * @param pThis A pointer to the dictionary instance.
	 * @return The user defined name of the dictionary, or a pointer to
	 *         the string "unnamed".  Never returns NULL.
	 */
	const char* (*name)(const struct S* pThis);
	
	/** Is the current dictionary empty */
	bool (*isEmpty)(const struct S* pThis);
	
	/** Get the number of entries in the dictionary */
	uint4 (*length)(const struct S* pThis);
	
	/** Does the dictionary contain a specific key */
	bool (*hasKey)(const struct S* pThis, K key);
	
	/** Get the first key in a dictionary
	 *
	 * @param pThis A pointer to the dictionary instance.
	 *
	 * @returns A pointer to the first key in the dictionary, which may also be
	 *         the last key in the dictionary if @a length is 1.  For a zero
	 *         length dictionary, NULL in returned.
	 */
	const K* (*first)(const struct S* pThis);
	
	/** Get the last key in a dictionary
	 *
	 * @param pThis A pointer to the dictionary instance.
	 *
	 * @returns A pointer to the first key in the dictionary, which may also be
	 *         the last key in the dictionary if @a length is 1.  For a zero
	 *         length dictionary, NULL in returned.
	 */
	const K* (*last)(const struct S* pThis);
	
	/** Get the next key
	 *
	 * @param pThis A pointer to the dictionary instance.
	 *
	 * @param key the reference key, the key after this one is returned.
	 *
	 * @returns A pointer to the next key, or NULL if @a key is the last
	 *          key in the this dictionary.
	 */
	const K* (*next)(const struct S* pThis, K key);
	
	/** Get the previous key.
	 *
	 * @param pThis A pointer to the dictionary instance.
	 *
	 * @param key the reference key, the key before this one is returned.
	 *
	 * @returns A pointer to the next key, or NULL if @a key is the first
	 *          key in the this dictionary.
	 */
	const K* (*prev)(const struct S* pThis, K key);
	
	/** Get a constant pointer to a value.
	 * @param pThis The dictionary instance.
	 * @param key The key for the item to retrieve.
	 * @return a constant pointer to the value, or NULL if the value
	 *         does not exist.
	 */
	const V* (*get)(const struct S* pThis, K key);
	
	/** Add a key value pair to the dictionary 
	 *
	 * @param pThis A pointer to the dictionary instance.
	 * @param key The key value. 
	 * @param pVal A pointer to the value.  Unless @a disownVal or @a disownAllVals
	 *        is called, the memory associated with the value will be deleted when
	 *        the dictionary is deleted.
	 *
	 * @returns the key pointer cast as constant.
	 *
	 * @throws (GSE_XCEPT,GSE_XDUPKEY) If a key is already in the dictionary.
	 * @throws (GSE_XCEPT,GSE_XBADARG) If @a pKef is null
	 */
	K (*insert)(struct S* pThis, K key, V* pVal);
	
	/** Get a mutable pointer to a value.
	 *
	 * @warning Careful how you use this pointer.  It is ment for
	 *          mutating dictionary values but @b NOT deleting them.
	 *          Deleting the return value will cause a "double free"
	 *          error in Clib when the @a del member is invoke.
	 *
	 * @param pThis The dictionary instance.
	 * @param key The lookup key for the new item.
	 * @returns a pointer to the value, or NULL if the value
	 *         does not exist.
	 *
	 * @throws (GSE_XCEPT,GSE_XNOKEY) if @a key is not in the dictionary.
	 */
	V* (*getMutable)(struct S* pThis, K key);
	
	/** Take a value out of the dictionary, and delete it's key. 
	 *
	 * @note Taking an item causes it's key to be deleted, unless
	 *       @a disownKey or @a disownAllKeys has been called.
	 * 
	 * @param pThis The dictionary instance.
	 * @param key The key for the item to take.
	 *
	 * @returns The requested value.
	 *
	 * @throws (GSE_XCEPT,GSE_XNOKEY) if @a key is not in the dictionary.
	 */
	V* (*take)(struct S* pThis, K key);
		
	/** Replace one value with another, keeping the key.
	 *
	 * The new value takes on the ownership disposition of the old value.
	 * Thus if the old value was owned by the dictionary then the new value
	 * is as well, and vice versa.
	 
	 * @warning The caller is expected to take ownership of the memory
	 *          occupied by the returned value.  The old value is 
	 *          @b NOT deleted by this call.
	 *
	 * @param pThis The dictionary instance.
	 * @param key the key for the location of the new value.
	 * @param pNewVal The new value, this may be null.
	 *
	 * @returns a Pointer to the old value.  The caller is now reponsible
	 *          for the old value.
	 *
	 * @throws (GSE_XCEPT,GSE_XNOKEY) if @a key is not in the dictionary.
	 */
	V* (*replace)(struct S* pThis, K key, V* pNewVal);
	
	/** Delete a key,value pair from the dictionary.
	 *
	 * @note If the value has been disowned, its memory is not freed.
	 *
	 * @param pThis The dictionary instance.
	 * @param key The key for the pair to delete.
	 * 
	 * @throws (GSE_XCEPT,GSE_XNOKEY) if @a key is not in the dictionary.
	 */
	void (*delPair)(struct S* pThis, K key);

	
	/** Do not delete a specific value item when the destructor is invoked 
	 *
	 * @param pThis the dictionary instance.
	 * @param key the key that should not be deleted. 
	 * @returns A pointer to the val that has been disowned.
	 *
	 * @throws (GSE_XCEPT,GSE_XNOKEY) if key is not in the dictionary.	 
	 */
	V* (*disownVal)(struct S* pThis, K key);
	
	/** Do not delete any value items when the destructor is invoked */
	void (*disownAllVals)(struct S* pThis);
	
	/** Destructor.
	 * Recusively deletes all keys and values unless they have been 
	 * disowned.
	 */
   void (*del)(void* vpThis);
	
	char* sName;
	
	bool bOwnsVals;
	
	dict_t dict;

} C;

/** Creates a new dynamic typed dictionary structure on the heap. */
C* _MASH(new_, C) (const char* sName);

/** Initilizing a stack dictionary object */
void _MASH(init_, C)(C* pThis, const char* sName);

/** Finishes a stack dictionary object */
void _MASH(finish_, C)(C* pThis);

