/** @file except.h Portable Exception Handling for ANSI C from Kaz Kylheku.
 *
 * Copyright (C) 1999 Kaz Kylheku <kaz@ashi.footprints.net>
 *
 * Free Software License:
 *
 * All rights are reserved by the author, with the following exceptions:
 * Permission is granted to freely reproduce and distribute this software,
 * possibly in exchange for a fee, provided that this copyright notice appears
 * intact. Permission is also granted to adapt this software to produce
 * derivative works, as long as the modified versions carry this copyright
 * notice and additional notices stating that the work has been modified.
 * This source code may be translated into executable form and incorporated
 * into proprietary software; there is no requirement for such software to
 * contain a copyright notice related to this source.
 *
 * $Id: except.h,v 1.18.2.1 2001/07/25 03:37:06 kaz Exp $
 * $Name: kazlib_1_20 $
 */

#ifndef XCEPT_H
#define XCEPT_H

#include <setjmp.h>
#include <stdlib.h>
#include <assert.h>

#define XCEPT_GROUP_ANY	0
#define XCEPT_CODE_ANY	0
#define XCEPT_BAD_ALLOC 1

#ifdef __cplusplus
extern "C" {
#endif

/**  @addtogroup mod_exceptions */
/*@{*/

/** Flag values for #except_cleanup_pop @p E argument.
 *
 * - @p except_no_call do not call the cleanup handler unless an exception
 *   has been raised within the cleanup region.
 * - @p except_call call the cleanup handler even if no exception is raised
 *   within the cleanup region. 
 */
enum { except_no_call, except_call };

typedef struct {
    unsigned long except_group;
    unsigned long except_code;
} except_id_t;

typedef struct {
    except_id_t volatile except_id;
    const char *volatile except_message;
    void *volatile except_dyndata;
} except_t;

struct except_cleanup {
    void (*except_func)(void *);
    void *except_context;
};

struct except_catch {
    const except_id_t *except_id;
    size_t except_size;
    except_t except_obj;
    jmp_buf except_jmp;
};

enum except_stacktype {
    XCEPT_CLEANUP, XCEPT_CATCHER
};

struct except_stacknode {
    struct except_stacknode *except_down;
    enum except_stacktype except_type;
    union {
	struct except_catch *except_catcher;	
	struct except_cleanup *except_cleanup;
    } except_info;
};

/* private functions made external so they can be used in macros */
void except_setup_clean(struct except_stacknode *,
	struct except_cleanup *, void (*)(void *), void *);
void except_setup_try(struct except_stacknode *,
	struct except_catch *, const except_id_t [], size_t);
struct except_stacknode *except_pop(void);

/* public interface functions */
int except_init(void);
void except_deinit(void);
void except_rethrow(except_t *);
void except_throw(long, long, const char *);
void except_throwd(long, long, const char *, void *);
void except_throwf(long, long, const char *, ...);
void (*except_unhandled_catcher(void (*)(except_t *)))(except_t *);
unsigned long except_code(except_t *);
unsigned long except_group(except_t *);
const char *except_message(except_t *);
void *except_data(except_t *);
void *except_take_data(except_t *);
void except_set_allocator(void *(*)(size_t), void (*)(void *));
void *except_alloc(size_t);
void except_free(void *);

#define except_code(E) ((E)->except_id.except_code)
#define except_group(E) ((E)->except_id.except_group)
#define except_message(E) ((E)->except_message)
#define except_data(E) ((E)->except_dyndata)

#ifdef __cplusplus
}
#endif

/*
 * void except_cleanup_push(void (*)(void *), void *); 
 * void except_cleanup_pop(int);
 * void except_checked_cleanup_pop(void (*)(void *), int);
 * void except_try_push(const except_id_t [], size_t, except_t **);
 * void except_try_pop(void);
 */

/** Start a cleanup region.
 * 
 * Sections of code that allocate memory on the heap usually need to free
 * that memory when an exception occurs.  This macro and the 
 * #except_cleanup_pop provide for global resource cleanup in the event than
 * an exception is raised.
 *
 * @par From the kazlib documentation:
 * The call to except_cleanup_push shall be matched with a call to 
 * #except_cleanup_pop which must occur in the same statement block at the 
 * same level of nesting. @n
 * The except_cleanup_push macro registers a cleanup handler that will be
 * called if an exception subsequently occurs before the matching
 * #except_cleanup_pop is executed, and is not intercepted and handled by a
 * try-catch region that is nested between the two. @n
 * The first argument to except_cleanup_push is a pointer to the cleanup 
 * handler, a function that returns nothing and takes a single argument of 
 * type void *. The second argument is a void * value that is registered along 
 * with the handler. This value is what is passed to the registered handler,
 * should it be called. @n
 * Cleanup handlers are called in the reverse order of their nesting: inner 
 * handlers are called before outer handlers. @n
 * The program shall not leave the cleanup region between the call to the 
 * macro except_cleanup_push and the matching call to #except_cleanup_pop by
 * means other than throwing an exception, or calling #except_cleanup_pop. @n
 * Within the call to the cleanup handler, it is possible that new exceptions 
 * may happen. Such exceptions must be handled before the cleanup handler
 * terminates. If the call to the cleanup handler is terminated by an
 * exception, the behavior is undefined.
 *
 * @hideinitializer
 */
#define except_cleanup_push(F, C) 				\
    {								\
	struct except_stacknode except_sn;			\
	struct except_cleanup except_cl;			\
	except_setup_clean(&except_sn, &except_cl, F, C)

/** End a cleanup region.
 *
 * @see except_cleanup_push
 * @hideinitializer
 */
#define except_cleanup_pop(E)					\
	except_pop();						\
	if (E)							\
	    except_cl.except_func(except_cl.except_context);	\
    }

#define except_checked_cleanup_pop(F, E)			\
    	except_pop();						\
	assert (except_cl.except_func == (F));			\
	if (E)							\
	    except_cl.except_func(except_cl.except_context);	\
    }

/** Start an exception handling region.
 * 
 * @par From the kazlib documentation:
 * The except_try_push macro marks the beginning of a try-catch region of the
 * program. It must be matched by a #except_try_pop written in the same
 * statement block at the same level of nesting, which terminates the 
 * try-catch region. Regions may be nested. The program shall not leave a
 * try-catch region other than by throwing an exception or by executing the except_try_pop.
 *
 * @param ID an array of except_id_t structures defining the "catch spec", a 
 *        list the exceptions to catch, by group id and code id.  The special
 *        exception group XCEPT_GROUP_ANY means catch exceptions in any group
 *        and the special exception code XCEPT_CODE_ANY means catch 
 *        exceptions with any code id. 
 * @param NUM the number of "catch spec"s.
 * @param PPE a pointer to an exception structure pointer.
 *
 * @hideinitializer
 */
#define except_try_push(ID, NUM, PPE)				\
     {								\
	struct except_stacknode except_sn;			\
	struct except_catch except_ch;				\
	except_setup_try(&except_sn, &except_ch, ID, NUM);	\
	if (setjmp(except_ch.except_jmp))			\
	    *(PPE) = &except_ch.except_obj;			\
	else							\
	    *(PPE) = 0

/** End an exception handling region.
 *
 * @par From the kazlib documentation:
 * The except_try_pop macro terminates a try-catch region. It must match a 
 * previous except_try_push macro in the same statement block at the same
 * level of nesting which is not already matched by an earlier except_try_pop.
 *
 * @hideinitializer
 */
#define except_try_pop()					\
	except_free(except_ch.except_obj.except_dyndata);	\
	except_pop();						\
    } 

#endif /* XCEPT_H */

/*@}*/

