#ifndef QDataSet_H_
#define QDataSet_H_

#ifdef __cplusplus
extern "C" {
#endif


/* Forward declaration placeholder */
struct qdataset_t;


/** QDataset values may be strings, integers, floats, or other QDataSets.
 *
 *  No matter the type each QDS value occupies 64-bits of memory.
 */
typedef union qvalue_t{
	char* s;               /* <--This should probably be wide characters for i18n */
	const char* cs;               /* Used when slicing */
	
	double d;
	
	long long int i;
	
	struct qdataset_t* q;
	const struct qdataset_t* cq;  /* Used when slicing */
} QValue;

#define QDS_VT_INVALID   0  

#define QDS_VT_STRING    0x10
#define QDS_VT_CONST_STR 0x11

#define QDS_VT_INT       0x20

#define QDS_VT_FLT       0x30

#define QDS_VT_QDS       0x40
#define QDS_VT_CONST_QDS 0x41

#define QDS_VT_MIXED     0xFE /* <-- Do you want to allow this ? */
#define QDS_VT_UNKNOWN   0xFF 


/** Property values may be of any type, the first member is used to hold a type
 * identifer
 */
typedef struct qpropval_t{
	int nType;
	void* pVal;
} QPropVal;

/** Structure for QDataSet
 *
 * @warning This is subject to change, use the accessor functions below instead
 *          of directly manipulating or reading QDataSet structures.
 */
typedef struct qdataset_t{
	int           nValType;
   QValue*       pVals;
	unsigned int  uLen;
   unsigned int  uRank;
   unsigned int* pQubeDims;
	int*          pValTypes;  /* <-- Usually null, only need if 
	                                  nValType == QDS_VT_MIXED */
   char**        pPropNames;
   QPropVal**    pPropVals;
} QDataSet;

/** Generate a new QDataSet with pre-allocated storage space.
 * 
 * @param nRank The number of diminsions for the qube of values.
 * @param pDims The size of each diminsion.
 * @param nType The type of data stored in this qdataset may be one of:
 *    
 *    - @b QDS_VT_STRING - Data will consist of null terminated strings.
 *    - @b QDS_VT_INT - Data will consist of 8-byte signed integers.
 *    - @b QDS_VT_FLT - Data will consist of 8-byte IEEE floats.
 *    - @b QDS_VT_QDS - Data will consist of other QDataSets.
 *    - @b QDS_VT_MIXED - Data will be of mixed types.
 *    - @b QDS_VT_UNKNOWN - Data type will be set when the first values
 *                          are added to the QDataSet
 * 
 * @returns A new QDataSet allocated on the heap.  If nRank is not
 *         zero, and pDims is valid then storages for the data is
 *         allocated as well. NULL is returned if nType is invalid
 *         or if one or more entries in pDims is 0 for a given rank.
 */
QDataSet* new_QDataSet(unsigned int nRank, unsigned int* pDims, int nType);

/** Initialize a QDataSet stack variable.
 *
 * @returns @a true if successful, @a false otherwise.
 */
bool init_QDataSet(QDataSet* qds, unsigned int nRank, unsigned int* pDims);


/** Free a QDataSet and all contained data
 */
void delete_QDataSet(const QDataSet* qds);

/** Finalize a QDataSet stack variable
 * This free's any 


/** Deep Copy a QDataSet 
 *
 * @return A copy of the QDataSet and all contained items allocated on the heap.
 */
QDataSet* QDataSet_copy(const QDataSet* qds);


/** Slice a QDataSet
 *
 */
const QDataSet* QDataSet_slice(const QDataSet* qds);

 
/** Retrieve a single value.
 *
 * This is a variable length argument function.  If the number of indicies
 * does not equal the rank of the corresponding QDataSet, it is an error.
 * 
 * @return A pointer to the QValue or NULL on an error.
 */
const QValue* QDataSet_get(const QDataSet* qds, ...);

/** Retrieve a single value, and be able to change it.
 * 
 * @warning Use with care.
 */
QValue* QDataSet_getMutable(QDataSet* qds, ...);

/** Add a single value to a QDataSet 
 *
 * @note The QDataSet takes ownership of qVal and will delete it when 
 *       delete_QDataSet() or finish_QDataSet() are called. 
 */
bool QDataSet_set(QDataSet* qds, QValue* qVal, ...);

/** Convience function for adding floating point value */
bool QDataSet_setD(QDataSet* qds, double dVal, ...);

/** Convience function for adding an integer value */
bool QDataSet_setI(QDataSet* qds, long long int iVal, ...);

/** Convience function for adding a string value.
 *
 * @note The QDataSet takes ownership of sVal and will delete it when 
 *       delete_QDataSet() or finish_QDataSet() are called. 
 */
bool QDataSet_setS(QDataSet* qds, char* sVal, ...);

/** Convience function for adding a copy of a string value */
bool QDataSet_setCS(QDataSet* qds, const char* csVal, ...);

/** Convience function for adding QDataSet value
 * @note The QDataSet takes ownership of qVal and will delete it when 
 *       delete_QDataSet() or finish_QDataSet() are called. 
 */
bool QDataSet_setQ(QDataSet* qds, QDataSet* qVal, ...);

/** Retrieve a single property */
const QPropVal* QDataSet_getProp(const QDataSet* qds, const char* sName);


/** Retrive a single mutatable property 
 * @warning use with care.
 */
QPropVal* QDataSet_getMutableProp(QDataSet* qds, const char* sName);

/* Many more methods would need to be defined, but you get the idea... */

#ifdef __cplusplus
}
#endif

#endif /* QDataSet_H_ */
