/*
 * fen.c
 *
 * cc  fen.c -o fen
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <strings.h>
#include <port.h>
#include <errno.h>
#include <pthread.h>
#include <thread.h>

struct fileinfo {
        struct file_obj fobj;
        int events;
        int port;
};

void
printevent(int event, char *pname)
{
        printf("%s :",pname);
        if (event & FILE_ACCESS) {
                printf(" FILE_ACCESS");
        }
        if (event & FILE_MODIFIED) {
                printf(" FILE_MODIFIED");
        }
        if (event & FILE_ATTRIB) {
                printf(" FILE_ATTRIB");
        }
        if (event & FILE_DELETE) {
                printf(" FILE_DELETE");
        }
        if (event & FILE_RENAME_TO) {
                printf(" FILE_RENAME_TO");
        }
        if (event & FILE_RENAME_FROM) {
                printf(" FILE_RENAME_FROM");
        }
        if (event & UNMOUNTED) {
                printf(" UNMOUNTED");
        }
        if (event & MOUNTEDOVER) {
                printf(" MOUNTEDOVER");
        }
        printf("\n");
}


/*
 * event handler for file events source.
 */
void
process_file(struct fileinfo *finf, int revents)
{
        struct file_obj *fobjp = &finf->fobj;
        int port = finf->port;
        struct stat sb;


        if (!(revents & FILE_EXCEPTION) &&
                stat(fobjp->fo_name, &sb) == -1) {
                fprintf(stderr, "Failed to stat file : %s - errno %d\n",
                        fobjp->fo_name, errno);
                free(finf->fobj.fo_name);
                free(finf);
                return;
        }

        /*
         * Add what ever processing that needs to be done
         * here. Process received events.
         */
        if (revents) {
                printevent(revents, fobjp->fo_name);
                /*
                 * If exception, no need to re-register.
                 */
                if (revents & FILE_EXCEPTION) {
                        free(finf->fobj.fo_name);
                        free(finf);
                        return;
                }
        }
        /*
         * re-register.
         */
        fobjp->fo_atime = sb.st_atim;
        fobjp->fo_mtime = sb.st_mtim;
        fobjp->fo_ctime = sb.st_ctim;
        if (port_associate(port, PORT_SOURCE_FILE, (uintptr_t)fobjp,
                 finf->events, (void *)finf) == -1) {
                /*
                 * Add error processing as required, file may have been
                 * deleted/moved.
                 */
                fprintf(stderr, "Failed to register file :%s - errno %d\n",
                        fobjp->fo_name, errno);
                free(finf->fobj.fo_name);
                free(finf);
        }
}

/*
 * worker threads wait here for event.
 */
void *
waitforevents(void *pn)
{
        int port = *((int *)pn);
        port_event_t pe;

        while (!port_get(port, &pe, NULL)) {
                /*
                 * Can add cases for other sources if this
                 * port is used to collect events from multiple sources.
                 */
                switch (pe.portev_source) {
                case PORT_SOURCE_FILE:
                        /* Call file events event handler */
                        process_file((struct fileinfo *)pe.portev_object,
                        pe.portev_events);
                        break;
                default:
                        perror("Event from unexpected source");
                }
        }
        printf("worker thread exiting\n");
        return(NULL);
}


#define STRLEN 256
char str[STRLEN];

int
main ()
{
        struct fileinfo *finf;
        int port;
        char *stp;
        pthread_t tid;
        if ((port = port_create()) == -1) {
                perror("port_create");
        }

        /*
         * create a worker thread to process events. Can add as many
         * worker threads as required.
         */
        pthread_create(&tid, NULL, waitforevents, (void *)&port);

        while(1) {
                if(fgets(str, STRLEN, stdin) == NULL) {
                        /* EOF */
                        break;
                }
                if ((stp = strstr(str, " ")) != NULL ||
                        (stp = strstr(str, "\n")) != NULL) {
                        stp[0]='\0';
                }

                finf = malloc(sizeof(struct fileinfo));
                if (finf == NULL) {
                        perror("memory alloc");
                        continue;
                }

                if ((finf->fobj.fo_name = strdup(str)) == NULL) {
                        perror("strdup");
                        free(finf);
                        continue;
                }

                /*
                 * Event types to watch.
                 */
                finf->events = FILE_ACCESS|FILE_MODIFIED|FILE_ATTRIB;
                finf->port = port;

                /*
                 * Start monitor this file.
                 */
                process_file(finf, 0);
        }
        /*
         * close port, will de-activate all file events watches associated
         * with the port.
         */
        close(port);

        /*
         * wait for threads to exit.
         */
        while (thr_join(0, NULL, NULL) == 0)
                ;
        return(0);
}
