/* test_write.c 
 * Test making generic packets, uses the GSE lib to replicate tar-like writing.
 */
#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <libgen.h>

/* NOTE: Test programs have to insure that headers are from the source 
 *        tree.  For end user programs: 
 *			
 *       #include <rpwgse/Gse.h> 
 *			
 *	      is fine.
 */

#include "../rpwgse/GseInit.h"
#include "../rpwgse/GseExcept.h"
#include "../rpwgse/GseWrap.h"
#include "../rpwgse/GseTypes.h"
#include "../rpwgse/GseConv.h"
#include "../rpwgse/GseString.h"
#include "../rpwgse/GseFile.h"
#include "../rpwgse/GseLog.h"

#include "../rpwgse/GseBlock.h"
#include "../rpwgse/GseHdr.h"
#include "../rpwgse/GsePkt.h"
#include "../rpwgse/GsePktSink.h"
#include "../rpwgse/GseFilePktSink.h"

/* Custom block type implimentation for test code */
#include "FileInfoBlk.c"

GsePkt* make_pkt(const char* sPath);

/*****************************************************************************/
int main(int argc, char** argv)
{
	int i;
	GsePkt* pPkt;
	GsePktSink* pSink;
	
	if((argc < 3)||(strcmp(argv[1], "-h") == 0)||
	   ( strncmp(argv[1], "--help", 6) == 0)){
		
		printf("\n"
"Test writing packets to a file.  Each file specified on the command line\n"
"is wraped as a generic GSE packet.  For each packet, the header contains\n"
"an original filename block and a generic processing block with process \n"
"ID 0xFF.\n"
"\n"
"     USAGE:  %s  file1 [file2 file3 ...] outfile\n\n", basename(argv[0]));
		return 1;
	}
	
	gse_thread_init(GSE_DEBUG, NULL);
	
	pSink = (GsePktSink*) new_GseFilePktSink(
	           argv[argc - 1], 
	           GSE_OVERWRITE|GSE_MAKEDIRS, 
	           GSE_MODE_664, 
	           0
	         );
	
	for(i = 1; i< argc-1; i++){

		pPkt = make_pkt(argv[i]);         /* allocates pkt */
		
		pSink->consumeNext(pSink, pPkt);  /* frees pkt */
	}
	
	pSink->del(pSink);
	return 0;
}

/*****************************************************************************/
/* Makes a packet containing a single file's data */

GsePkt* make_pkt(const char* sPath)
{
	struct stat stbuf;
	GsePkt* pPkt = NULL;
	FileInfoBlk* pBlk = NULL;
	bytebuf_t fileBuf = BBUF_NULL;

	pPkt = new_empty_GsePkt(0);
	except_cleanup_push(pPkt->del, pPkt);     /* >> Free pPkt mem on execpt */
	
	gse_stat(sPath, &stbuf);	
	
	pBlk = new_FileInfoBlk(CBBUF_NULL);
	pPkt->setHdrLimits(pPkt, 128, 128);
	pPkt->pHdr->nSpaceCraft = 999;            /* Real NAIF IDs are neg.     */
	
	except_cleanup_push(pBlk->del, pBlk);     /* >> Free pBlk mem on except */
	
	pBlk->st_mode  = stbuf.st_mode;
	gse_debug_v("st_mode: 0%o (0x%x)", stbuf.st_mode, stbuf.st_mode);
	pBlk->st_uid   = stbuf.st_uid;  
	pBlk->st_gid   = stbuf.st_gid;  
	pBlk->st_atime = stbuf.st_atime;
	pBlk->st_mtime = stbuf.st_mtime;
	pBlk->st_ctime = stbuf.st_ctime;
	pBlk->sPath    = CS(sPath);
	
	pPkt->addHdrBlk(pPkt, (GseBlock*)pBlk);
	except_cleanup_pop(except_no_call);       /* << pPkt owns pBlk now */
	
	fileBuf = gse_readfile(sPath);
	pPkt->setData(pPkt, fileBuf);
	
	except_cleanup_pop(except_no_call);       /* << Okay no exceptions      */
	return pPkt;
}

