#!/usr/bin/env python
import sys
import os
import struct
import time
import optparse
import logging
import os.path
import csv
import glob

from os.path import basename as bname
from os.path import join as pjoin

from rpwgse import *
from jwaves import *

##############################################################################
# Globals 

# Converting HFRH channel names into mixer frequencies in kHz:

g_dChan2Mixer = {
  'Chan 000':3500,  'Chan 001':4500,  'Chan 002':5500,  'Chan 003':6500, 
  'Chan 004':7500,  'Chan 005':8500,  'Chan 006':9500,  'Chan 007':10500, 
  'Chan 008':11500, 'Chan 009':12500, 'Chan 010':13500, 'Chan 011':14500,
  'Chan 012':15500, 'Chan 013':16500, 'Chan 014':17500, 'Chan 015':18500,
  'Chan 016':19500, 'Chan 017':20500, 'Chan 018':21500, 'Chan 019':22500,
  'Chan 020':23500, 'Chan 021':24500, 'Chan 022':25500, 'Chan 023':26500,
  'Chan 024':27500, 'Chan 025':28500, 'Chan 026':29500, 'Chan 027':30500,
  'Chan 028':31500, 'Chan 029':32500, 'Chan 030':33500, 'Chan 031':34500,
  'Chan 032':35500, 'Chan 033':36500, 'Chan 034':37500, 'Chan 035':38500,
  'Chan 036':39500, 'Chan 037':40500
}

##############################################################################
# Copied in to make sure script has no dependencies
def setLogLevel(sLevel):
	
	sLevel = sLevel.lower()
	nLevel = logging.WARNING
	nLibLevel = GSE_INFO
	
	sDateFmt = '%Y-%m-%d %H:%M:%S'
	sConFmt = '%(levelname)-8s: %(message)s'
	
	if sLevel.startswith("c"):
		nLevel = logging.CRITICAL
		nLibLevel = GSE_CRITICAL
		
	elif sLevel.startswith("e"):
		nLevel = logging.ERROR
		nLibLevel = GSE_ERROR
		
	elif sLevel.startswith("i"):
		nLevel = logging.INFO
		nLibLevel = GSE_INFO
		
	elif sLevel.startswith("d"):
		nLevel = logging.DEBUG
		sFileFmt = '%(asctime)s %(levelname)-8s: %(message)s'
		sConFmt = '%(asctime)s %(levelname)-8s: %(message)s'
		nLibLevel = GSE_DEBUG
	
	rootLogger = logging.getLogger('')
	rootLogger.setLevel(nLevel)

	conHdlr = logging.StreamHandler(sys.stderr)
	formatter = logging.Formatter(sConFmt, sDateFmt)
	conHdlr.setFormatter(formatter)
	rootLogger.addHandler(conHdlr)
	
	gse_log_setlevel(nLibLevel)

	return rootLogger

##############################################################################
def load_srv_cal(sCalFile):
	
	# key's are the mixer frequency, in kHz
	# Items are lists whose index is the data number
	fIn = file(sCalFile, 'rb')
	rdr = csv.DictReader(fIn)
	
	# Get past the 1st row
	dRow = rdr.next()
	
	dOut = {}
	for (key,val) in g_dChan2Mixer.iteritems():
		dOut[val] = []
	
	n = 0
	for dRow in rdr:
		
		if int(dRow['Data Value']) != n:
			raise ValueError("%s: Expected to be on data number %d, but on %d"%(
			                 sCalFile, n, int(dRow['Data Value'])) )
		
		for (sChan, sVal) in dRow.iteritems():
		
			if sChan == 'Data Value':
				continue
			
			dOut[ g_dChan2Mixer[sChan] ].append( float(sVal) )
			
			iIdx = len(dOut[ g_dChan2Mixer[sChan] ]) - 1
			
			if iIdx != n:
				raise ValueError("%s: List index %d doesn't correspond to data "
				                 " number %d"%(sCalFile, iIdx, n))
	
		n += 1
	
	return dOut
	
##############################################################################
def load_amp_ref(sNbsAmpRef):
	"""Loads a dictionary of frequencies by channel at which the amplitude 
	calibration is applicable"""
	
	dOut = {}
	
	fIn = file(sNbsAmpRef, 'rb')
	rdr = csv.DictReader(fIn)
	
	for dRow in rdr:
		
		sChan = dRow['Channel']
		sAmpCalFreq = dRow['Amp Cal Offset (kHz)']
		
		dOut[sChan] = int(sAmpCalFreq)
		
	return dOut
	

##############################################################################
# Main

def main(argv):
	
	
	psr = optparse.OptionParser(
	   usage="%prog [options] LOGAMP_CAL1 [LOGAMP_CAL2 ...]",
	   prog="mk_nrb_freq_cal.py",
		version="1",
		description="""
Read 1-N HFR_logamp_off/HFWBR_logamp_off calibration files and output a
WAV_CAL_BST_NBS_FREQ file.  Note it takes 38 * 2 input files to complete the
output table.  If more that one run of a particular channel is input then the
results are collapsed by simple averaging."""
	)

	psr.add_option('-l', "--log-level", dest="sLevel", metavar="LOG_LEVEL",
	               help="Logging level one of [critical, error, warning, "+\
	               "info, debug].  The default is info.", type="string",
	               action="store", default="info")
						
	psr.add_option('-d', '--alt-dir', dest="sCalDir", metavar="DIR",
	               default=None, type="string", action="store", 
						help="Look in this DIRectory for the survey data HFR "
						"calibration files: WAV_CAL_SRV_HFR_HI_???V??.CSV.  This "
						"program uses the SURVEY mode frequency response data as "
						"BURST frequency respsonse data are not available")
						
	(opts, lTables) = psr.parse_args(argv[1:])
	
	log = setLogLevel(opts.sLevel)
	
	if len(lTables) == 0:
		log.error("No input logamp calibration files specified, use -h for more info")
		return 1
	
	if opts.sCalDir == None:
		opts.sCalDir="/opt/project/juno/data/flight/JNOWAV_1000/CALIB"
	
	lTmp = glob.glob( pjoin(opts.sCalDir, 'WAV_CAL_SRV_HFR_HI_45_V??.CSV') )
	if len(lTmp) < 1:
		log.error("Can't find WAV_CAL_SRV_HFR_HI_45_V??.CSV in %s"%opts.sCalDir)
		return 1
	lTmp.sort()
	sHfwbr = lTmp[-1]
	
	# key's are the mixer frequency, in kHz
	# Items are lists whose index is the data number
	dHfwbrSrvCal = load_srv_cal(sHfwbr)
		
	lTmp = glob.glob( pjoin(opts.sCalDir, 'WAV_CAL_SRV_HFR_HI_44_V??.CSV') )
	if len(lTmp) < 1:
		log.error("Can't find WAV_CAL_SRV_HFR_HI_44_V??.CSV in %s"%opts.sCalDir)
		return 1
	lTmp.sort()
	sHfr = lTmp[-1]
	
	dHfrSrvCal = load_srv_cal(sHfr)
	
	# Open the accompaining NBS AMP cal data and find out which frequency
	# point should be the reference
	dNbsAmpRef = {}
	
	lTmp = glob.glob( pjoin(opts.sCalDir, 'WAV_CAL_BST_NBS_AMP_V??.CSV') )
	if len(lTmp) < 1:
		log.error("Can't find WAV_CAL_BST_NBS_AMP_V??.CSV in %s"%opts.sCalDir)
		return 1
	lTmp.sort()
	sNbsAmpRef = lTmp[-1]
	
	dNbsAmpRef = load_amp_ref(sNbsAmpRef)
	
	# Initialize the output channels
	dFreqCalTbl = {}
	
	for i in xrange(3,41):
		sKey = 'HFWBR_%d.5'%i
		dFreqCalTbl[sKey] = {}
		
		for j in xrange(-750, 800, 50):
			dFreqCalTbl[sKey][j] = None
			
	for i in xrange(3,41):
		sKey = 'HFR_%d.5'%i
		dFreqCalTbl[sKey] = {}
		
		for j in xrange(-750, 800, 50):
			dFreqCalTbl[sKey][j] = None
	
	
	# Read all the inputs add them into the global calc
	for sTable in lTables:
		
		fIn = file(sTable, 'rb')
		
		log.info("%s: Reading survey cal"%sTable)
		
		rdr = csv.DictReader(fIn)
		
		for dRow in rdr:
		
			# See if I'm at the end of the data section
			if dRow['UTC Cal'] == None:
				break
			if len(dRow['UTC Cal']) == 0:
				break
			if dRow['UTC Cal'].lower() == 'tot cal pts:':
				break
				
			#print dRow['UTC Cal']
			
			# Figure out which channel this row corresponds to
			dSrvCalTbl = None
			if dRow['Pkt Desc'].startswith('HFR Log Det'):
				sRec = 'HFR_'
				dSrvCalTbl = dHfrSrvCal
			elif dRow['Pkt Desc'].startswith('HFWBR Log Det'):
				sRec = 'HFWBR_'
				dSrvCalTbl = dHfwbrSrvCal
			else:
				raise ValueError("Unknown value for column 'Pkt Desc': "
				                 "'%s'"%dRow['Pkt Desc'])
									  
			# Figure out which survey channel we're calibrating from
			nSrvCalChan = int(dRow['Bin KHz'])
			
	
			# Figure out which output mixer frequency channel we're in
			sOutChan = '%s%.1f'%(sRec, float(dRow['Bin KHz']) / 1000.0) 
			
			if not dFreqCalTbl.has_key(sOutChan):
				raise ValueError("Unknown NBS channel key: %s"%sChan)
			
			# Figure out the offset from the mixer frequency:
			nFreq = int( float(dRow['FGen KHz']) )
			nOffset = nFreq - int(dRow['Bin KHz']) 
			
			if not dFreqCalTbl[sOutChan].has_key(nOffset):
				log.debug('%s: Freq. offset %d is outside the range of interest'%(
				          sTable, nOffset))
				continue
			
			# Okay, convert to a Vrms
			nIv = int( dRow['Data Val'] )
			rVrms = dSrvCalTbl[nSrvCalChan][nIv]
			
			if dFreqCalTbl[sOutChan][nOffset] == None:
				dFreqCalTbl[sOutChan][nOffset] = (rVrms, 1)
				
			else:
			
				rCur = 	dFreqCalTbl[sOutChan][nOffset][0]
				nWeight = dFreqCalTbl[sOutChan][nOffset][1]
				#print "rCur:", rCur
				#print "nWeight:", nWeight
				rNew = (rCur * nWeight + rVrms) / (nWeight + 1)
				
				dFreqCalTbl[sOutChan][nOffset] = (rNew, nWeight + 1)
				
	
	# Okay, now go through the dFreqCalTbl and normalize so that the 
	# multipictive voltage gain is 1.0 at the amplitude cal frequency.
	# From the board level tests this is always at +250 kHz above the 
	# mixer frequency.
	
	lGoodChans = []
	
	for sChan in dFreqCalTbl.keys():
		
		dChan = dFreqCalTbl[sChan]
		
		nAmpFreqOffset = dNbsAmpRef[sChan]
		
		# Find the amplitude cal freq item
		if dChan[nAmpFreqOffset] == None:
			
			log.error("Can't normalize channel %s, %d kHz freq point not "
			          "read"%(nAmpFreqOffset, sChan))
			continue
		else:
			rNorm = dChan[nAmpFreqOffset][0]
			log.info("Using offset %d for %s channel normalization"%(
			         nAmpFreqOffset, sChan))
		
		
		for nOffset in dChan.keys():
		
			if dChan[nOffset] == None:
				log.error("Missing offset %d in channel %s"%(nOffset, sChan))
			else:
				dChan[nOffset] = ( dChan[nOffset][0] / rNorm, dChan[nOffset][1] )
		
		

	# Okay, now make a table of the results
	
	pout = sys.stdout.write
	lChans = []
	
	for i in xrange(3,41):
		lChans.append( 'HFWBR_%d.5'%i )
		
	for i in xrange(3,41):
		lChans.append( 'HFR_%d.5'%i )
		
	pout('"Offset (kHz)"')
	for sChan in lChans:
		pout(',"%s"'%sChan)
	pout('\r\n')

	bMissing = False
	for nOffset in xrange(-750, 800, 50):
		pout('%d'%nOffset)
		
		for sChan in lChans:
			
			pout(',')
			
			if dFreqCalTbl[sChan][nOffset] != None:
				pout('%.4e'%dFreqCalTbl[sChan][nOffset][0])
			else:
				bMissing = True
		
		pout('\r\n')
		
	
	if not bMissing:
		log.info("All frequency ratios output")
	else:
		log.error("One or more channels are missing gain ratio")
		
	return 13

##############################################################################
if __name__ == "__main__":
	sys.exit(main(sys.argv))


