#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <strings.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseMath.h>

#include "BstCalTbl.h"

/*****************************************************************************/

bool BstCalTbl_isLoaded(const BstCalTbl* pThis)
{
	return (pThis->uFinState == pThis->uLoadState);
}

/*****************************************************************************/
/* Getting a frequency correction */

float8 BstCalTbl_freqFact(const BstCalTbl* pThis, float8 dFreq)
{
	int4 i = 0, iLo = 0, iHi = 0;
	float8 dRet = 0;
	
	if((pThis->uLoadState & BST_FREQ_CAL) == 0)
		gse_throw(GSE_XCEPT, GSE_XNOTREADY);
	
	for(i = 0; i < pThis->nFreqs; i++){
		
		/* If within a kilo hertz, just use this value */
		if( abs( (pThis->pFOffsets[i] + pThis->dMixerHz) - dFreq) < 1000.0)
			return pThis->pFGains[i];
		
		/* Okay, if the table frequency > than the target freq save this as 
		   the lower bound */
		
		if( dFreq > (pThis->pFOffsets[i] + pThis->dMixerHz) ){
			iLo = i;
			continue;
		}
		
		if(dFreq < (pThis->pFOffsets[i] + pThis->dMixerHz) ){
			iHi = i;
			break;
		}
	}
	
	if(iHi == 0)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Frequency %.3e Hz is outside the "
		             "frequency gain table range of %.3e to %.3e Hz "
		             "(the row for mixer = %.3e Hz)", 
		             dFreq, pThis->dMixerHz + pThis->pFOffsets[0], 
		             pThis->dMixerHz + pThis->pFOffsets[pThis->nFreqs - 1],
				       pThis->dMixerHz);
	
	dRet = gse_lerp(
			pThis->dMixerHz + pThis->pFOffsets[iLo],
			pThis->dMixerHz + pThis->pFOffsets[iHi],
			pThis->pFGains[iHi], pThis->pFGains[iHi],
			dFreq);
	
	return dRet;
};


/*****************************************************************************/
/* Destructor */

void delete_BstCalTbl(void* vpThis)
{
	BstCalTbl* pThis = (BstCalTbl*)vpThis;
	
	if(pThis->sChannel) free(pThis->sChannel);
	if(pThis->sAttnChan) free(pThis->sAttnChan);
	if(pThis->pAttns) free(pThis->pAttns);
	if(pThis->pFOffsets) free(pThis->pFOffsets);
	if(pThis->pFGains) free(pThis->pFGains);
	
	free(vpThis);
}

/*****************************************************************************/
/* Constructor */

BstCalTbl* new_BstCalTbl(const char* sChanName, const char* sAttnChName,
                         uint4 uFinState, int4 nNumAttns)
{
	BstCalTbl* pThis = NULL;

	if((sChanName == NULL)||(uFinState == 0)||
		(uFinState > 0x15)||((uFinState & BST_AMP_CAL) == 0) )
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Bad argument to BstCalTbl constructor");
	
	if((uFinState & BST_ATTN_CAL) && (sAttnChName == NULL))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Missing attenuator calibration "
		             "file key for analog channel %s.", sChanName);
	
	pThis = (BstCalTbl*)gse_calloc(1, sizeof(BstCalTbl));
	pThis->del         = delete_BstCalTbl;
	pThis->freqFact    = BstCalTbl_freqFact;
	pThis->isLoaded    = BstCalTbl_isLoaded;
	pThis->sChannel    = CS(sChanName);
	
	if(sAttnChName)
		pThis->sAttnChan   = CS(sAttnChName);
	
	pThis->uFinState   = uFinState;
	
	pThis->dPagf       = 1.0;  /* Default pre-amp gain factor to 1 */
	
	/* Set all cal versions to -1 to indicate un-initialized */
	pThis->nAmpVer     = -1;
	pThis->nAttnVer    = -1;
	pThis->nFreqVer    = -1;
	
	/* Default to the rate of sampling for the downmixed data as there are
	   so many tables that are for mixed data */
	pThis->dSampHz     = BST_DOWN_MIXED_SAMPLE_RATE_HZ; 
	
	pThis->nAttns = nNumAttns;
	if(pThis->nAttns > 0)
		pThis->pAttns =(attn_cal_t*) gse_calloc(pThis->nAttns, sizeof(attn_cal_t));
		
	
	/* Not really complete until cal files have been read */
	return pThis;
}

/*****************************************************************************/
/* Load attenuator settings for all channels with attenuators */

void load_bstcal_attns(StrBstCalTblDict* pDict, const char* sPath)
{
	
	char        sFile[128]   = {'\0'};
	int4        nCalVer      = -1;
	FILE*       fIn          = NULL;
	const char* sChannel     = NULL;
	BstCalTbl*  pTbl         = NULL;
	int4        iInsPos      = 0;
	char        sLine[1024]  = {'\0'};
	int4        nLine        = 0;
	int4        nLen         = 0;
	char        sRawField[128] = {'\0'};
	char*       sField       = NULL;
	attn_cal_t* pAtn         = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading attenuator calibrations", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading attenuator calibrations", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
	/* Will be reading this file multiple times, once for each cal-table in
	   the dictionary */
	
	for(sChannel = pDict->first(pDict); sChannel != NULL; 
	    sChannel = pDict->next(pDict, sChannel) ){
	
		pTbl = pDict->getMutable(pDict, sChannel);
		iInsPos = 0;
		
		if(pTbl->nAttns < 1)   /* This channel doesn't have attenuators (B) */
			continue;
		
		pTbl->nAttnVer = nCalVer;
		
		fseek(fIn, 0, SEEK_SET);
		nLine = 0;
		
		while( fgets(sLine, 1023, fIn) != NULL){
			nLine += 1;	
			
			nLen = strlen(sLine);
			
			if(sLine[nLen - 2] != '\r' || sLine[nLen - 1] != '\n')
				gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d has a non PDS "
				             "complient line ending.", sFile, nLine);
			
			if(nLine < 3)      /* 1st 2 lines are headers */
				continue;
			
			gse_csv_getField_s(sLine, 0, sRawField, 128);
			sField = gse_strip(sRawField, "\"");
			
			/* This line describes a different analog channel than the current 
			   cal table */
			if( strcasecmp(sField, pTbl->sAttnChan) != 0 ){
				free(sField);
				continue;
			}
			free(sField);

			
			if(iInsPos >= pTbl->nAttns)
				gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Too many "
				             "attenuation settings for analog path %s", sFile,
				             nLine, pTbl->sAttnChan);
						
			pAtn = pTbl->pAttns + iInsPos;
			
			pAtn->nPaVal = gse_csv_getInt(sLine, 1, sFile, nLine);
			pAtn->nRecVal = gse_csv_getInt(sLine, 2, sFile, nLine);
			pAtn->dDb = gse_csv_getFloat(sLine, 3, sFile, nLine);
			pAtn->dFactor = gse_csv_getFloat(sLine, 4, sFile, nLine);
			
			iInsPos += 1;
		}
		
		/* Make sure we got 'em all */
		if(pTbl->nAttns != iInsPos)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Missing expected attenuation"
			             "calibration(s) for analog path", sFile, 
			             pTbl->sAttnChan);	
		
		
		pTbl->uLoadState |= BST_ATTN_CAL;
	}
	
	fclose(fIn);
}


/*****************************************************************************/
/* Load amplitude calibrations for all directly sampled channels */

void load_bstcal_direct_amp(StrBstCalTblDict* pDict, const char* sPath)
{
	
	char sFile[128]; memset(sFile, 0, 128);
	int4        nCalVer      = -1;
	FILE*       fIn          = NULL;
	BstCalTbl*  pTbl         = NULL;
	char        sLine[1024]; memset(sLine, 0, 1024);
	size_t      nLine        = 0;
	int4        nLen         = 0;
	char        sRawField[128]; memset(sRawField, 0, 128);
	char*       sField       = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading amplitude calibrations for directly "
				       "sampled channels", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading amplitude calibrations for directly "
		             "sampled channels", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
	/* Will be reading this file once, and picking which cal tables to 
	   update */
	
	nLine = 0;
		
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;	
			
		nLen = strlen(sLine);
			
		if(sLine[nLen - 2] != '\r' || sLine[nLen - 1] != '\n')
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d has a non PDS "
			             "complient line ending.", sFile, nLine);
			
		if(nLine < 2)      /* 1st line is a header */
			continue;
			
		gse_csv_getField_s(sLine, 0, sRawField, 128);
		sField = gse_strip(sRawField, "\"");
		except_cleanup_push(free, sField);
		
		/* See if we have a cal table that goes with this line */
		if( (pTbl = pDict->getMutable(pDict, sField)) == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Unknown analog "
			             "path name %s", sFile, nLine, sField);
		
		if(pTbl->uLoadState & BST_AMP_CAL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Duplicate analog "
					       "path name %s", sFile, nLine, sField);
		
		except_cleanup_pop(except_call); /* sField */
		
		pTbl->dSampHz = gse_csv_getFloat(sLine, 4, sFile, nLine);
		
		pTbl->uAmpCalType = BST_CAL_POWER;
		pTbl->dA = gse_csv_getFloat(sLine, 8, sFile, nLine);
		pTbl->dB = gse_csv_getFloat(sLine, 9, sFile, nLine);
		
		/* Special exception for the magnetic cals which have no antenna len */
		if(pTbl->uFinState & BST_USE_ANT_LEN){
			pTbl->dAntLen = gse_csv_getFloat(sLine, 7, sFile, nLine);
			pTbl->uLoadState |= BST_USE_ANT_LEN;
		}
		else{
			pTbl->dAntLen = 1.0;
		}
		pTbl->dAmpCalHz = (float8) gse_csv_getInt(sLine, 6, sFile, nLine);
		/* Assume 0 setting for dMixer from constructor */
		pTbl->nAmpVer = nCalVer;
		
		pTbl->uLoadState |= BST_AMP_CAL;
		
	}
	
	fclose(fIn);
}

/*****************************************************************************/
/* Load amplitude calibrations for all down-mixed channels */

void load_bstcal_mixed_amp(StrBstCalTblDict* pDict, const char* sPath)
{	
	char sFile[128]; memset(sFile, 0, 128);
	int4        nCalVer      = -1;
	FILE*       fIn          = NULL;
	char        sLine[1024]; memset(sLine, 0, 1024);
	size_t      nLine        = 0;
	int4        nLen         = 0;
	char        sRawField[128]; memset(sRawField, 0, 128);
	char*       sField       = NULL;
	BstCalTbl*  pTbl         = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading amplitude calibrations for down-mixed "
				       "channels", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading amplitude calibrations for down-mixed "
		             "channels", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
	/* Will be reading this file once, and picking which cal tables to 
	   update */
	
	nLine = 0;
		
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;	
			
		nLen = strlen(sLine);
			
		if(sLine[nLen - 2] != '\r' || sLine[nLen - 1] != '\n')
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d has a non PDS "
			             "complient line ending.", sFile, nLine);
			
		if(nLine < 2)      /* 1st line is a header */
			continue;
			
		gse_csv_getField_s(sLine, 0, sRawField, 128);
		sField = gse_strip(sRawField, "\"");
		except_cleanup_push(free, sField);
		
		/* See if we have a cal table that goes with this line */
		if( (pTbl = pDict->getMutable(pDict, sField)) == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Unknown analog "
			             "path name %s", sFile, nLine, sField);
		
		if(pTbl->uLoadState & BST_AMP_CAL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Duplicate analog "
					       "path name %s", sFile, nLine, sField);
		
		except_cleanup_pop(except_call); /* sField */
		
		pTbl->uAmpCalType = BST_CAL_LINEAR;
		pTbl->dA = gse_csv_getFloat(sLine, 4, sFile, nLine);
		pTbl->dB = gse_csv_getFloat(sLine, 5, sFile, nLine);
		pTbl->dPagf = gse_csv_getFloat(sLine, 6, sFile, nLine);
		
		if( (pTbl->uFinState & (BST_USE_ANT_LEN|BST_AMP_CAL)) != 
			 (BST_USE_ANT_LEN|BST_AMP_CAL) )
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Cal tables with channels "
			             "listed in %s are expected to need ampliude calibration "
			             "and effective antenna length calibration.", sFile);
		
		pTbl->dAntLen = gse_csv_getFloat(sLine, 7, sFile, nLine);
		
		
		pTbl->dAmpCalHz = (float8) gse_csv_getInt(sLine, 3, sFile, nLine);
		pTbl->dAmpCalHz *= 1000000.0;
		pTbl->dMixerHz = gse_csv_getFloat(sLine, 2, sFile, nLine);
		pTbl->dMixerHz *= 1000000.0;
		
		pTbl->nAmpVer = nCalVer;
		
		pTbl->uLoadState |= BST_AMP_CAL;
		pTbl->uLoadState |= BST_USE_ANT_LEN;
	}
	
	fclose(fIn);
}

/*****************************************************************************/
/* Load amplitude calibrations for all narrow band spectra channels */

void load_bstcal_nbs_amp(StrBstCalTblDict* pDict, const char* sPath)
{
	char sFile[128]; memset(sFile, 0, 128);
	int4        nCalVer      = -1;
	FILE*       fIn          = NULL;
	char        sLine[1024]; memset(sLine, 0, 1024);
	size_t      nLine        = 0;
	int4        nLen         = 0;
	char        sRawField[128]; memset(sRawField, 0, 128);
	char*       sField       = NULL;
	BstCalTbl*  pTbl         = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading amplitude calibrations for narrow-band "
				       "spectra", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading amplitude calibrations for norraw-band "
		             "spectra", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
	/* Will be reading this file once, and picking which cal tables to 
	   update */
	
	nLine = 0;
		
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;	
			
		nLen = strlen(sLine);
			
		if(sLine[nLen - 2] != '\r' || sLine[nLen - 1] != '\n')
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d has a non PDS "
			             "complient line ending.", sFile, nLine);
			
		if(nLine < 2)      /* 1st line is a header */
			continue;
	
		gse_csv_getField_s(sLine, 0, sRawField, 128);
		sField = gse_strip(sRawField, "\"");
		except_cleanup_push(free, sField);
		
		/* See if we have a cal table that goes with this line */
		if( (pTbl = pDict->getMutable(pDict, sField)) == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Unknown analog "
			             "path name %s", sFile, nLine, sField);
		
		if(pTbl->uLoadState & BST_AMP_CAL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Duplicate analog "
					       "path name %s", sFile, nLine, sField);
		
		except_cleanup_pop(except_call); /* sField */
		
		pTbl->uAmpCalType = BST_CAL_LINEAR;
		pTbl->dA = gse_csv_getFloat(sLine, 8, sFile, nLine);
		pTbl->dB = gse_csv_getFloat(sLine, 9, sFile, nLine);
		pTbl->dPagf = gse_csv_getFloat(sLine, 10, sFile, nLine);
		
		if( (pTbl->uFinState & (BST_USE_ANT_LEN|BST_AMP_CAL)) != 
			 (BST_USE_ANT_LEN|BST_AMP_CAL) )
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Cal tables with channels "
			             "listed in %s are expected to need ampliude calibration "
			             "and effective antenna length calibration.", sFile);
		
		pTbl->dAntLen = gse_csv_getFloat(sLine, 11, sFile, nLine);
		
		pTbl->dAmpCalHz = (float8) gse_csv_getInt(sLine, 2, sFile, nLine);
		pTbl->dAmpCalHz *= 1000000.0;
		pTbl->dMixerHz = gse_csv_getFloat(sLine, 1, sFile, nLine);
		pTbl->dMixerHz *= 1000000.0;
		pTbl->nAmpVer = nCalVer;
		
		pTbl->dLoOutHz = gse_csv_getFloat(sLine, 4, sFile, nLine);
		pTbl->dLoOutHz = (pTbl->dLoOutHz * 1000.0) + pTbl->dMixerHz;
				
		pTbl->dLoInHz = gse_csv_getFloat(sLine, 5, sFile, nLine);
		pTbl->dLoInHz = (pTbl->dLoInHz * 1000.0) + pTbl->dMixerHz;
				
		pTbl->dUpInHz = gse_csv_getFloat(sLine, 6, sFile, nLine);
		pTbl->dUpInHz = (pTbl->dUpInHz * 1000.0) + pTbl->dMixerHz;
				
		pTbl->dUpOutHz = gse_csv_getFloat(sLine, 7, sFile, nLine);
		pTbl->dUpOutHz = (pTbl->dUpOutHz * 1000.0) + pTbl->dMixerHz;
		
		pTbl->uLoadState |= BST_USE_ANT_LEN;
		pTbl->uLoadState |= BST_AMP_CAL;
	}
	
	fclose(fIn);
}

/****************************************************************************/
/* Frequency corrections for NBS products */

void load_bstcal_nbs_freq(StrBstCalTblDict* pDict, const char* sPath)
{
	char        sFile[128]   = {'\0'};
	int4        nCalVer      = -1;
	FILE*       fIn          = NULL;
	int4        nOffsets     = 0;
	char        sLine[1024]  = {'\0'};
	size_t      nLine        = 0;
	int4        nLen         = 0;
	GseStrAry*  pChan        = NULL;
	char        sRawField[128] = {'\0'};
	int4        i            = 0;
	float       dOffset      = 0.0;
	const char* sChannel     = NULL;
	BstCalTbl*  pTbl         = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading frequency calibrations for down-mixed "
				       "channels", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading frequency calibrations for down-mixed "
		             "channels", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
	/* First read once to get the file length */
	nOffsets = -1;
	while( fgets(sLine, 1028, fIn) != NULL )
		nOffsets += 1;
	
	fseek(fIn, 0, SEEK_SET);
	
	/* Will be reading this file once, and picking which cal tables to 
	   update */
	
	nLine = 0;
	pChan = new_GseStrAry(83);  /* 41 * 2 + 1 */
	pChan->pAry[0] = NULL;  /* This is the offset column */
		
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;	
			
		nLen = strlen(sLine);
			
		if(sLine[nLen - 2] != '\r' || sLine[nLen - 1] != '\n')
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d has a non PDS "
			             "complient line ending.", sFile, nLine);
		
		nLen = gse_csv_fields(sLine);
		if(nLen != 83)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Expect 83 columns, "
			             "found %d.", sFile, nLine, nLen);
			
		/* The first line is special, get the channel offsets from it */
		if(nLine == 1){
			
			for(i = 1; i < 83; i++){
				gse_csv_getField_s(sLine, i, sRawField, 128);
				pChan->pAry[i] = gse_strip(sRawField, "\"");	
			}
			
			continue;
		}
	
		/* The rest of the lines start with an offset, then 72 frequency 
		   adjustments, one per NBS channel */
		dOffset = (float8)gse_csv_getInt(sLine, 0, sFile, nLine);
		dOffset *= 1000.0;
		
		for(i = 1; i < 83; i++){
			
			sChannel = pChan->pAry[i];
			
			/* See if we have a cal table that goes with this line */
			if( (pTbl = pDict->getMutable(pDict, sChannel)) == NULL)
				gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Unknown analog "
				             "path name %s", sFile, 1, sChannel);
			
			if(pTbl->pFOffsets == NULL){
				pTbl->nFreqs = nOffsets;
				pTbl->pFOffsets = (float8*)gse_calloc(nOffsets, sizeof(float8));
				pTbl->pFGains = (float8*)gse_calloc(nOffsets, sizeof(float8));
			}
			
			pTbl->pFOffsets[nLine - 2] = dOffset;
			pTbl->pFGains[nLine - 2] = gse_csv_getFloat(sLine, i, sFile, nLine);
			
			/* Hey, I filled out something for this table, so mark it as having
			   frequency information, though we don't know the info to be complete
			*/
			pTbl->nFreqVer = nCalVer;
			pTbl->uLoadState |= BST_FREQ_CAL;
		}
	}
	
	fclose(fIn);
}

/*****************************************************************************/
/* Cal Table to string dictionary */


#define C StrBstCalTblDict
#define K char
#define V BstCalTbl

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcmp(a, b)

#include <rpwgse/ptrDictTplt.c>

#undef KEY_COMPARE
#undef DEL_VALS
#undef FREE_KEYS

#undef V
#undef K
#undef C













