/** @file BstCalTbl.h */
#ifndef _BstCalTbl_H_
#define _BstCalTbl_H_

#include <rpwgse/GseTypes.h>
#include <rpwgse/dict.h>

#include "caldir_util.h"


#ifdef __cplusplus
extern "C" {
#endif

/* For the directly sampled data this is taken from the spreadsheets */
#define BST_DOWN_MIXED_SAMPLE_RATE_HZ 1312500.0

/* Used to define the "I'm Loaded" bitmask */
#define BST_AMP_CAL 0x1
#define BST_ATTN_CAL 0x2
#define BST_FREQ_CAL 0x4
#define BST_USE_ANT_LEN 0x8

#define BST_CAL_LINEAR 0x01
#define BST_CAL_POWER  0x02


/*****************************************************************************
 * Cal Table holder, holds the calibration to convert the data from a SINGLE
 * ANALOG CHANNEL into physical units.  Amplitude calibration is always
 * present.  Frequency calibration is only present for NBS data.
 */
 
/* #define BST_CAL_EQMASK 0x03

#define BST_HAS_FREQ_GAINS 0x4
#define BST_HAS_FREQ_MASK 0x4 */

typedef struct bst_cal_tbl{
	
	/* Overall stuff ******************************************************/
	
	/** The name of the channel */
	char* sChannel;
	
	/** The name of the attenuator channel */
	char* sAttnChan;
	
	/** 0 - 7 depending on which tables have been loaded.
	 *   0x0 - None
	 *   0x1 - Amplitude Cal present
	 *   0x2 - Attn Cal present
	 *   0x4 - Freq cal present
	 * These values are OR'ed together 
	 */
	uint4 uLoadState;
	
	/** The value uLoadState should have when ready to use */	
	uint4 uFinState;
	
	/** The sampling interval, may be used to get the frequency bins
	    for NBS data by dividing by N */
	
	double dSampHz;
	
	
	/* Amplitude Calibration Stuff ****************************************/
	
	/** How to use the calibration factors dA & dB, 
	 * i.e. is this a linear or a power relation 
	 */
	uint4  uAmpCalType;
	
	/** Cal coefficient A, may be slope may be factor */	
	float8 dA;
	
	/** Cal coefficinet B, may be power, may be intercept */
	float8 dB;
	
	/** Pre-amp gain factor, needed when using board level calibrations */
	float8 dPagf;

	/** Effective antenna length for the channel */
	float8 dAntLen;
	
	/** Midband frequency at which the amplitude calibration was preformed,
	 *  in Hertz.
	 */
	float8 dAmpCalHz;     
	
	/** Mixer setting in Hz, may be 0 */
	float8 dMixerHz;      

	/** Version number of the amplitude cal table */
	int4 nAmpVer;

	
	/* Attenuator calibration stuff ***************************************/
		
	/** Number of attenuator settings for this channel, may be 0 */
	int4 nAttns;
	
	/* All possible attenuator factors for this channel */
	attn_cal_t* pAttns;
	
	/* Version number of the attenuator calibration */
	int4 nAttnVer;      
	
   
	/* Frequency correction stuff *****************************************/
	
	/** Lower Sideband Outer Edge of where frequency calibration applies */
	float8 dLoOutHz;
	
	/** Lower Sideband Inner Edge of where frequency calibration applies */
	float8 dLoInHz;
	
	/** Upper Sideband Inner Edge of where frequency calibration applies */
	float8 dUpInHz;
	
	/** Upper Sideband Outer Edge of where frequency calibration applies */
	float8 dUpOutHz;
	
	/** Number of "channel to channel" gain factors loaded*/
	int4 nFreqs;    
		
	/** The X value in terms of an offset from mixer in Hz */ 
	float8* pFOffsets;  
	
	/** The Y value in terms of an ampltude factor */
	float8* pFGains;    
	
	/* Noise bandwidth of each channel in Hertz is determined by
	    the sample size, and is not a hard locked value */
	/* float8 dNoiseBW; */
	
	/** Frequency calibration table version */
	int4 nFreqVer;
	
	
	/** Is the cal table ready? */
	bool (*isLoaded)(const struct bst_cal_tbl* pThis);
	
	/* Getting frequency corrections */
	float8 (*freqFact)(const struct bst_cal_tbl* pThis, float8 dFreq);
		
	/** Destructor */
	void (*del)(void* vpThis);

} BstCalTbl;

/** Make a new channel cal object
 *
 * @public 
 * @memberof BstCalTbl
 */
BstCalTbl* new_BstCalTbl(const char* sChanName, const char* sAttnChName,
                         uint4 uFinState, int4 nNumAttns);

/*****************************************************************************/
/* Cal table dictionaries */

#define S str_bstcaltbl_dict
#define C StrBstCalTblDict
#define K char
#define V BstCalTbl

#include <rpwgse/ptrDictTplt.h>

#undef V
#undef K
#undef C
#undef S


/** Load amplitude calibrations for directly sampled channels 
 *
 * @param pDict a string to BstCalTbl map, where map keys are expected to
 *        be the channel names in the amplitude calibration files.
 *
 * @param sPath the location of the WAV_CAL_BST_DIRECT_AMP_Vxx.CSV file to
 *        read.
 */
void load_bstcal_direct_amp(StrBstCalTblDict* pDict, const char* sPath);

/** Load amplitude calibrations for down-mixed channels
 *
 * @param pDict a string to BstCalTbl map, where map keys are expected to
 *        be the channel names in the amplitude calibration files.
 *
 * @param sPath the location of the WAV_CAL_BST_DIRECT_AMP_Vxx.CSV file to
 *        read.
 */
void load_bstcal_mixed_amp(StrBstCalTblDict* pDict, const char* sPath);


/** Load amplitude calibrations for narrow-band spectra channels
 *
 * @param pDict a string to BstCalTbl map, where map keys are expected to
 *        be the channel names in the amplitude calibration files.
 *
 * @param sPath the location of the WAV_CAL_BST_NBS_AMP_Vxx.CSV file to
 *        read.
 */
void load_bstcal_nbs_amp(StrBstCalTblDict* pDict, const char* sPath);


/** Loading attenuator calibrations into the burst calibration tabels 
 *
 * @param pDict a string to BstCalTbl map.  This function ignores the 
 *        map keys, so use keys that make sense for other functions.
 *
 * @param sPath the location of the WAV_CAL_BST_DIRECT_AMP_Vxx.CSV file to
 *        read.
 */
void load_bstcal_attns(StrBstCalTblDict* pDict, const char* sPath);

/** load frequency corrections for narrow-band spectra channels
 *
 * @param pDict a string to BstCalTbl map, where map keys are expected to
 *        be the channel names in the amplitude calibration files.
 *
 * @param sPath the location of the WAV_CAL_BST_NBS_FREQ_Vxx.CSV file to
 *        read.
 */
void load_bstcal_nbs_freq(StrBstCalTblDict* pDict, const char* sPath);


#ifdef __cplusplus
}
#endif

#endif /* _BstCalTbl_H_ */
