#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <math.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>

#include <jwaves/JWavesSDSblk.h>
#include <jwaves/JWavesHrbhBlk.h>
#include <jwaves/JWavesConv.h>

#include "caldir_util.h"
#include "jno_spice.h"
#include "BstCalib.h"


/****************************************************************************/
/* Compiled in config file directory location.*/
#ifndef CFG
#error Compiled in default configuration directory is missing.
#endif

#define _QDEF(x) #x
#define QDEF(x) _QDEF(x)

/* Need to load the tables:

    WAV_CAL_BST_DIRECT_AMP_Vxx.CSV
	 WAV_CAL_BST_MIXED_AMP_Vxx.CSV
	 WAV_CAL_BST_NBS_AMP_Vxx.CSV
	 WAV_CAL_BST_NBS_FREQ_Vxx.CSV

*/

/*****************************************************************************/
/* Trivial constructor for BstRow objects */

BstRow* new_BstRow(const BstCalTbl* pCal)
{
	size_t uLen = 0;
	BstRow* pThis = (BstRow*)gse_calloc(1, sizeof(BstRow));
	
	memset(pThis->sChannel, '\0', 14);
	/* memset(pThis->sScet, '\0', 24); */
	
	/* Set what you can */
	
	if( (uLen = strlen(pCal->sChannel)) > 14)
		strncpy(pThis->sChannel, pCal->sChannel, 14);
	else
		strncpy(pThis->sChannel, pCal->sChannel, uLen);
	
	pThis->nAmpCalVer = (int1) pCal->nAmpVer;
	pThis->nAttnCalVer = (int1) pCal->nAttnVer;
	pThis->nFreqCalVer = (int1) pCal->nFreqVer;
	
	pThis->del = free;
	
	/* Can't do sampling interval as that would require packet 
	   level knowledge */
	
	return pThis;	
}


/*****************************************************************************/

float8buf_t g_dSideBuf = {NULL, 0};

BstRow* BstCalib_calSec(const BstCalib* pThis, const GsePkt* pPkt, byte uSec)
{
	const U4BstCalTblDict*    pDict     = NULL;
	JWavesSDSblk*             pSds      = NULL;
	byte                      uUseMixer = 0;
	JWavesHrbhBlk*            pBst      = NULL;
	uint4                     uKey      = 0;
	BstRow*                   pRow      = NULL;
	const BstCalTbl*          pTbl      = NULL;
	struct jwaves_float4_conv conv      = {0};
	float8                    dAttnFact = 1.0;
	float4                    dClipHi   = 0.0;
	float4                    dClipLo   = 0.0;
	float8                    dAvg      = 0.0;
	float8                    dFreq     = 0.0;
	int4                      i         = 0;
	int4                      nOffset   = 0;
	int4                      nSamples  = 0;
	int4                      nClipped  = 0;
	int4                      iMixer    = 0;    /* Index of mixing freq */
	float8                    dFactor   = 0.0;  /* Chan to chan gain factor */
	float8                    dTmp      = 0.0;
	
	if((uSec != 0)&&(uSec != 2))
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	pDict = pThis->pTblsById;
	
	pSds = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	
	/* The I/Q data are calibrated using .5 MHz mixer settings, so there
	 * exists keys for:  3.5, 4.5, 5.5 etc.
	 * we are going to collapse keys from 3.0 MHz to 43.75 MHz by knocking
	 * off the fractional part and using the .5 MHz seting for that integer
	 * frequency.
	 */
	uUseMixer = pSds->uMixer;
	if((uUseMixer >= 12) && (uUseMixer <= 175)){
		
		/* If the mixer is at a half-MHz leave it alone, a mixer value of
		   14 corresponds to an oscilator setting of 3.5 MHz (3.5 * 4 = 14) */
		if((uUseMixer - 14) % 4 != 0){			
			uUseMixer =  (( (int)(uUseMixer / 4.0) ) * 4) + 2 ;
			
			gse_info_v("Calibrating a %.2f MHz mixer seting as a %.1f MHz mixer "
			            "setting.", pSds->uMixer/4.0, uUseMixer/4.0);
		}
	}
	   
	uKey = ((pSds->uPSID << 16)&0xFF0000) | ((uSec << 8)&0xFF00) | 
			 (uUseMixer & 0xFF) ;
	
	if( ! pDict->hasKey(pDict, uKey) ){
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Missing calibration table for key "
		             "PSID = %02X, Section = %u, Band = %02X",
		             pSds->uPSID, uSec, pSds->uMixer);
	}
	pTbl = pDict->get(pDict, uKey);
	
	pRow = new_BstRow(pTbl);
	
	/* Internal only vars */
	pRow->uPsid = pSds->uPSID;
	pRow->uSec  = uSec;
	
	/* Let's do the bin header stuff first because it's easy */
	if(pPkt->blkCount(pPkt, JW_HRS_BIN_HDR_ID, 0) != 0){
			
		pBst = (JWavesHrbhBlk*)pPkt->getBlkByTypeIdx(pPkt, JW_HRS_BIN_HDR_ID, 0, 0);
		
		pRow->uQFacSec = pBst->uQFsec;
		pRow->uQFacSub = pBst->uQFsubSec;
		pRow->uQProcSec = pBst->uJunoProcSec;
		pRow->uQProcSub = pBst->uJunoProcSubSec;
		pRow->uQfactor = pBst->uQF;
	}
	else{
		pRow->uQFacSec = 0;
		pRow->uQFacSub = 0;
		pRow->uQProcSec = 0;
		pRow->uQProcSub = 0;
		pRow->uQfactor = 0;
	}
	
	/* Get the sclk */
	pRow->dColSclk = ((float8)pSds->uColSCLK) + (((float8)pSds->uColRTI) / 40.0);
	
	/* Set default NRD flags */
	pRow->bNro = false;
	pRow->bNra = false;
	
	/* Get the preamp and receiver atteunuation factors */
	pRow->uPaAttn = (pSds->uAttn >> 5) & 0x1;
	
	if( jwaves_islfrsci(pSds->uPSID) ){
		pRow->uRecAttn = (pSds->uAttn >> 4) & 0x1;
	}
	else{
		if( jwaves_ishfrsci(pSds->uPSID) )
			pRow->uRecAttn = pSds->uAttn & 0xF;
		else
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Unknown science PSID: %02hhX",
			             pSds->uPSID);
	}

	for(i = 0; i < pTbl->nAttns; i++){
		if((pTbl->pAttns[i].nPaVal == pRow->uPaAttn) && 
			(pTbl->pAttns[i].nRecVal == pRow->uRecAttn))
			dAttnFact = pTbl->pAttns[i].dFactor;
	}
	
	/* Get the SCET, YYYY-DDDTHH:MM:SS.sss format */
	if(! pThis->pAdj->adjust(pThis->pAdj, &(pRow->dColSclk)))
		gse_throw(GSE_XCEPT, GSE_XNOKEY);
	
	/* Set the mixer frequency for any of the HFRH types */
	if((pSds->uPSID==0x88)||(pSds->uPSID==0x89)||(pSds->uPSID==0x8C)||
		(pSds->uPSID==0x8D)||(pSds->uPSID==0xF1)||(pSds->uPSID==0xF2) ){
		if((pSds->uMixer < 0x0C)||(pSds->uMixer> 0xBF))
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Unexpected band value in "
			             "in packet, PSID = %02hhX, BAND = 0x%02hhX", 
			             pSds->uPSID, pSds->uMixer);
			
		pRow->fMixerHz = ((float4)pSds->uMixer) * 250000.0;
	}
		
	
	/* Handle the data conversion */
	
	init_jwaves_float4_conv(&conv);
	jwaves_bytes2float4(pSds->uFmt, pPkt->getData(pPkt), 0, &conv);
	except_cleanup_push(free, conv.ptr);
	
	switch(pSds->uPSID){
	/* NRD Waveforms */
	case 0xA0:
	case 0xA4:
	case 0xA5:
		pRow->fSampInt = 1.0 / pTbl->dSampHz;
		
		if((nSamples % 3) != 0)
			gse_throw_m(GSE_XCEPT, GSE_XDECODE, "In Noise Reduction Diagnostic "
							"packet, number of samples is not a multiple of 3.");
		
		nSamples = conv.len / 3;
		
		/* Calculate the offset based on a non-truncated number of samples */
		nOffset = nSamples * uSec;
		
		/* Last section has 128 bytes (64 samples) cut off */	
		if(uSec == 2){
			nSamples -= 64;
			pRow->bNra = true; /* Section idx 2 is the mitigated data */
		}
			
		pRow->bNro = true;
		
		if(uSec > 2)
			gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Only sections 0-2 are supported"
			            " when calibrating LFR NRD Waveform packets.");
						
		/* Calc the clipped values */
		dClipHi = 65536.0 - 327.68;
		dClipLo = 327.68;
		nClipped = 0;
		for(i = 0; i < nSamples; i++){
			if((conv.ptr[i+nOffset] < dClipLo)||(conv.ptr[i+nOffset] > dClipHi))
				nClipped += 1;
		}
		pRow->fClippedFrac = ((float8)nClipped) / nSamples;
		
		/* Now for the calibration, 1st get the average */
		for(i = 0; i < nSamples; i++) dAvg += conv.ptr[i+nOffset];
		dAvg /= nSamples;
		
		pRow->data.len = nSamples;
		pRow->data.ptr = (float4*)gse_calloc(nSamples, sizeof(float4));
		
		for(i=0; i< nSamples; i++){
			if( pTbl->uAmpCalType == BST_CAL_POWER) {
			
				pRow->data.ptr[i] = (pow( fabs( conv.ptr[i+nOffset] - dAvg), pTbl->dB) *
				                     pTbl->dA * dAttnFact ) / pTbl->dAntLen;
			}
			else{
				pRow->data.ptr[i] = pTbl->dA*fabs(conv.ptr[i+nOffset] - dAvg) + pTbl->dB;
				pRow->data.ptr[i] *= dAttnFact / pTbl->dAntLen;
			}
			
			pRow->data.ptr[i] *= pTbl->dPagf;  /* Pre-amp gain */
			
			if(conv.ptr[i+nOffset] < dAvg) pRow->data.ptr[i] *= -1.0;
		}
		break;
		
		
	/* All Wavefroms */
	case 0x88:
	case 0x89:
	case 0x8C:
	case 0x8D:
	case 0xA1:
	case 0x02:
	case 0xA2:
	case 0xA3:
	case 0x7B:
	case 0x7F:
	case 0x3F:
	case 0x3B:
		pRow->fSampInt = 1.0 / pTbl->dSampHz;
		
		nOffset = 0;
		nSamples = conv.len;
		
		if((pSds->uPSID == 0xA1)||(pSds->uPSID == 0x02)||(pSds->uPSID == 0xA2)||
		   (pSds->uPSID == 0xA3)){
			/* Note, with the overloaded band byte, even single-section LFR 
			   waveforms can be products of the mitigation algorithm */
			if(pSds->uMixer == 0xF1){
				pRow->bNra = true;
				pRow->bNro = true;
			}
		}
				
		/* Calc the clipped values */
		dClipHi = 65536.0 - 327.68;
		dClipLo = 327.68;
		nClipped = 0;
		for(i = 0; i < nSamples; i++){
			if((conv.ptr[i+nOffset] < dClipLo)||(conv.ptr[i+nOffset] > dClipHi))
				nClipped += 1;
		}
		pRow->fClippedFrac = ((float8)nClipped) / nSamples;
		
		/* Now for the calibration, 1st get the average */
		for(i = 0; i < nSamples; i++) dAvg += conv.ptr[i+nOffset];
		dAvg /= nSamples;
		
		pRow->data.len = nSamples;
		pRow->data.ptr = (float4*)gse_calloc(nSamples, sizeof(float4));
		
		for(i=0; i< nSamples; i++){
			if( pTbl->uAmpCalType == BST_CAL_POWER) {
			
				pRow->data.ptr[i] = (pow( fabs( conv.ptr[i+nOffset] - dAvg), pTbl->dB) *
				                     pTbl->dA * dAttnFact ) / pTbl->dAntLen;
			}
			else{
				pRow->data.ptr[i] = pTbl->dA*fabs(conv.ptr[i+nOffset] - dAvg) + pTbl->dB;
				pRow->data.ptr[i] *= dAttnFact / pTbl->dAntLen;
			}
			
			pRow->data.ptr[i] *= pTbl->dPagf;  /* Pre-amp gain */
			
			if(conv.ptr[i+nOffset] < dAvg) pRow->data.ptr[i] *= -1.0;
		}
		break;
			
			
	/* All spectra */	
	case 0xF1:
	case 0xF2:
		
		/* Check for clipping */
		dClipHi = 65536.0 - 327.68;
		dClipLo = 327.68;
		nClipped = 0;
		for(i = 0; i < conv.len; i++){
			if((conv.ptr[i] < dClipLo)||(conv.ptr[i] > dClipHi))
				nClipped += 1;
		}
		pRow->fClippedFrac = ((float8)nClipped) / conv.len;
		
		/* Sampling interval depends on the data size */
		pRow->fSampInt = pTbl->dSampHz / conv.len;
		
		/* Okay, for this data we are going to assume that both points at
		   index (N/2 - 1) and (N/2) represent the mixing frequency.
			So the 1st frequency will be:
			
			Fm - Fs(0.5 - 1/N)
		*/
		pRow->fFirstFreq = pTbl->dMixerHz - pTbl->dSampHz * (0.5 - 1.0/conv.len);
		
		/* Do all the calibrations in a double precision array, alloc if needed */
		if( g_dSideBuf.len < conv.len){
			if(g_dSideBuf.ptr) free(g_dSideBuf.ptr);
			g_dSideBuf.len = conv.len;
			g_dSideBuf.ptr = (float8*)gse_calloc(g_dSideBuf.len, sizeof(float8));
		}		
				
		/* First do a simple coversion to voltage using amp cal */
		for(i = 0; i< conv.len; i++){
			g_dSideBuf.ptr[i] = (pTbl->dA * conv.ptr[i] + pTbl->dB) *
			                    dAttnFact * pTbl->dPagf;
		}
		
		/* Now add in the frequency corrections (this is a PITA) */
		
		/* Lower sideband */
		iMixer = (conv.len/2) - 1;
		for(i = 0; i < conv.len/2; i++){
			
			dFreq = pTbl->dMixerHz - (iMixer - i) * pRow->fSampInt;
			
			if(dFreq < pTbl->dLoOutHz) dFreq = pTbl->dLoOutHz;
			if(dFreq > pTbl->dLoInHz) dFreq = pTbl->dLoInHz;
			
			/* Get the frequency adjust ment */
			dFactor = pTbl->freqFact(pTbl, dFreq);
			
			g_dSideBuf.ptr[i] *= dFactor;
		}
		
		/* Upper sideband */
		iMixer = conv.len/2;
		for(i = conv.len/2; i < conv.len; i++){
			
			dFreq = pTbl->dMixerHz + (i - iMixer) * pRow->fSampInt;
			
			if(dFreq < pTbl->dUpInHz) dFreq = pTbl->dUpInHz;
			if(dFreq > pTbl->dUpOutHz) dFreq = pTbl->dUpOutHz;
			
			/* Get the frequency adjustment */
			dFactor = pTbl->freqFact(pTbl, dFreq);
			
			g_dSideBuf.ptr[i] *= dFactor;
		}

		/* okay, now copy to output as spectral density */
		pRow->data.len = conv.len - 1;
		pRow->data.ptr = (float4*)gse_calloc(pRow->data.len, sizeof(float4));
		
		/* Lower sideband spectral density (including mixer freq) */
		for(i = 0; i < conv.len/2; i++){
			pRow->data.ptr[i] = (g_dSideBuf.ptr[i] * g_dSideBuf.ptr[i]);
			
			/* Where did 1922.6 come from ??? */
			/* pRow->data.ptr[i] /= pTbl->dAntLen * pTbl->dAntLen * 1922.6; */
			pRow->data.ptr[i] /= pTbl->dAntLen * pTbl->dAntLen * pRow->fSampInt;
		}
		
		/* Mixer spectral density */
		iMixer = conv.len/2;
		dTmp = (g_dSideBuf.ptr[iMixer] * g_dSideBuf.ptr[iMixer]);
		dTmp /= pTbl->dAntLen * pTbl->dAntLen * pRow->fSampInt;
		
		pRow->data.ptr[iMixer - 1] += dTmp;  /* 1st part is from above */
		pRow->data.ptr[iMixer - 1] /= 2.0;
		
		for(i = iMixer + 1; i< conv.len; i++){
			pRow->data.ptr[i-1] = (g_dSideBuf.ptr[i] * g_dSideBuf.ptr[i]);
			pRow->data.ptr[i-1] /= pTbl->dAntLen * pTbl->dAntLen * pRow->fSampInt;
		}
		
		break;
	
	default:
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Unknown PSID %02hhX",
		             pSds->uPSID);
		break;
	}
	
	except_cleanup_pop(except_call); /* conv.ptr */
	return pRow;
}



BstRowAry* BstCalib_cal(const BstCalib* pThis, const GsePkt* pPkt)
{
	JWavesSDSblk*           pSds    = NULL;
	BstRowAry*              pRowAry = NULL;
	BstRow*                 pRow    = NULL;
	
	static const except_id_t catch_spec[] = {{GSE_XCEPT, GSE_XDECODE}};
	except_t*               pExcept = NULL;
	char                    sMsg[256] = {'\0'};
	
	if( ! pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0) ){
		gse_error("Non-science data encountered in PDS L2 input, FIX this!");
		return NULL;
	}
	
	pSds = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	/* Catch "can't calibrate" exceptions and just leave that packet
		   out of the output */
	except_try_push(catch_spec, 1, &pExcept);
	if(pExcept == NULL){	
		pRow = BstCalib_calSec(pThis, pPkt, 0);
	}
	else{
		gse_shortprn_except(sMsg, 255, pExcept);
		gse_error(sMsg);
		pRow = NULL;
	}
	except_try_pop();
	
	if(pRow == NULL) return NULL;
	
	pRowAry = new_BstRowAry(0);
	except_cleanup_push(pRowAry->del, pRowAry);
	pRowAry->add(pRowAry, pRow);
	

	/* Handle last sections, for A0, A4 & A5 */
	if((pSds->uPSID == 0xA0)||(pSds->uPSID == 0xA4)||(pSds->uPSID == 0xA5)){	
		
		/* Going to have to put the 1st section on the bump to *UNC list */
		pRow->bBumped = true;
		
		pRow = BstCalib_calSec(pThis, pPkt, 2);
		pRowAry->add(pRowAry, pRow);
	}
	
	except_cleanup_pop(except_no_call);
	return pRowAry;
}

/*****************************************************************************/
void delete_BstCalib(void* vpThis)
{
	BstCalib* pThis = (BstCalib*)vpThis;
	
	if(pThis->pTblsById)
		pThis->pTblsById->del(pThis->pTblsById);
	
	if(pThis->pTbls)
		pThis->pTbls->del(pThis->pTbls);
	
	free(vpThis);
}

/*****************************************************************************/
BstCalib* new_BstCalib(const char* sGroup, HskFineAdj* pAdj)
{
	BstCalib* pThis = NULL;
	
	GseCStrAry*       pPrefix   = NULL;
	GseStrAry*        pCalFiles = NULL;
	StrBstCalTblDict* pD	       = NULL; 
	U4BstCalTblDict*  pI        = NULL;
	BstCalTbl*        pCal      = NULL;
	byte              uBand     = 0;
	int4              i         = 0;
	char*             sChan     = NULL;
	const char*       sKey      = NULL;
	const BstCalTbl*  pConstCal = NULL;
	
	pThis = (BstCalib*) gse_calloc(1, sizeof(BstCalib));
	except_cleanup_push(delete_BstCalib, pThis);
	
	/* Init the CALIB dir */
	pPrefix = new_GseCStrAry(5);
	pPrefix->pAry[0] = "WAV_CAL_BST_DIRECT_AMP_V";
	pPrefix->pAry[1] = "WAV_CAL_BST_MIXED_AMP_V";
	pPrefix->pAry[2] = "WAV_CAL_BST_NBS_AMP_V";
	pPrefix->pAry[3] = "WAV_CAL_BST_NBS_FREQ_V";
	pPrefix->pAry[4] = "WAV_CAL_ATTN_V";
	
	pCalFiles = init_calib_dir(sGroup, pPrefix);
	pPrefix->del(pPrefix);
	
	except_cleanup_push(pCalFiles->del, pCalFiles);
	
	pThis->del = delete_BstCalib;
	pThis->cal = BstCalib_cal;
	
	pThis->pAdj = pAdj;
	
	/* Generate the calibration data holders */
	
	pThis->pTblsById = new_U4BstCalTblDict(NULL);
	pThis->pTblsById->disownAllVals(pThis->pTblsById);  /* Next dict owns the vals */
	
	pThis->pTbls = new_StrBstCalTblDict(NULL);
	pD = pThis->pTbls;
	
	/* Key is a uint composed of the bytes: 00 PSID SECTION BAND,
	   where PSID is the 2nd MSB byte, and BAND is the LSB */
	
	pI = pThis->pTblsById;  
	
	/* Directs */
	pCal = new_BstCalTbl("LFR_LO_B", NULL, BST_AMP_CAL, 0);
	pD->insert(pD, CS("LFR_LO_B"), pCal);
	
	/* Single section packets */
	pI->insert(pI, 0xA10000, pCal);    /* S+N            pre 2011-10 */
	
	pI->insert(pI, 0x0200F0, pCal);    /* S+N */          /* new 2013-12-11 */
	pI->insert(pI, 0x0200F1, pCal);    /* (S+N) - f(N)*/  /* new 2013-12-11 */
	
	pI->insert(pI, 0xA100F0, pCal);    /* S+N */
	pI->insert(pI, 0xA100F1, pCal);    /* (S+N) - f(N)*/
	
	/* Triple section packets */
	pI->insert(pI, 0xA00000, pCal);    /* S+N            pre 2011-10 */
	pI->insert(pI, 0xA00200, pCal);    /* (S+N) - f(N)   pre 2011-10 */
	pI->insert(pI, 0xA000F3, pCal);    /* (S+N)  */
	pI->insert(pI, 0xA002F3, pCal);    /* (S+N) - f(N) */
	
	
	
	pCal = new_BstCalTbl("LFR_LO_E", "LFR_LO_E",
			               BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 4);
	pD->insert(pD, CS("LFR_LO_E"), pCal);
	
   /* Single section packets */
	pI->insert(pI, 0xA20000, pCal);    /* S+N            pre 2011-10 */
	pI->insert(pI, 0xA200F0, pCal);    /* S+N */
	pI->insert(pI, 0xA200F1, pCal);    /* (S+N) - f(N)*/
	
	/* Triple section packets */
	pI->insert(pI, 0xA40000, pCal);    /* S+N            pre 2011-10 */
	pI->insert(pI, 0xA40200, pCal);    /* (S+N) - f(N)   pre 2011-10 */
	pI->insert(pI, 0xA400F3, pCal);    /* (S+N)  */
	pI->insert(pI, 0xA402F3, pCal);    /* (S+N) - f(N) */	
	
	
	pCal = new_BstCalTbl("LFR_HI_E", "LFR_HI_E",
	                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 4);
	pD->insert(pD, CS("LFR_HI_E"), pCal);
	
	/* Single section packets */
	pI->insert(pI, 0xA30000, pCal);    /* S+N            pre 2011-10 */
	pI->insert(pI, 0xA300F0, pCal);    /* S+N */
	pI->insert(pI, 0xA300F1, pCal);    /* (S+N) - f(N)*/
	
	/* Triple section packets */
	pI->insert(pI, 0xA50000, pCal);    /* S+N            pre 2011-10 */
	pI->insert(pI, 0xA50200, pCal);    /* (S+N) - f(N)   pre 2011-10 */
	pI->insert(pI, 0xA500F3, pCal);    /* (S+N)  */
	pI->insert(pI, 0xA502F3, pCal);    /* (S+N) - f(N) */	
	
		
	pCal = new_BstCalTbl("HFR_BASE", "HFR_LO_44",
	                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
	pD->insert(pD, CS("HFR_BASE"), pCal);
	pI->insert(pI, 0x7F0000, pCal);
	pI->insert(pI, 0x3F0000, pCal);
	
	
	pCal = new_BstCalTbl("HFWBR_BASE", "HFR_LO_45",
	                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
	pD->insert(pD, CS("HFWBR_BASE"), pCal);
	pI->insert(pI, 0x7B0000, pCal);
	
	/* Early versions of the flight software used 3B for the dual route 
	   waveform data.  Don't know what ID was used for the dual route spectra
		data, maybe it wasn't marked as special. -cwp */
	pI->insert(pI, 0x3B0000, pCal);
	
	
	/* Make 6 * 41 cal tables ! */
	
	for(i = 3; i<44; i++){
		
		uBand = i*4 + 2;
		
		/* Mixed */
		sChan = NS("HFWBR_I_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_45",
		                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0x880000 | uBand, pCal);
		
		sChan = NS("HFWBR_Q_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_45",
		                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0x890000 | uBand, pCal);
			
		
		sChan = NS("HFR_I_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_44",
		                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0x8C0000 | uBand, pCal);
		
		sChan = NS("HFR_Q_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_44",
		                     BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0x8D0000 | uBand, pCal);
		
		
		/* NBS */
		sChan = NS("HFWBR_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_45",
		               BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN|BST_FREQ_CAL, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0xF10000 | uBand, pCal);
		
		sChan = NS("HFR_%d.5", i);
		pCal = new_BstCalTbl(sChan, "HFR_HI_44",
				         BST_AMP_CAL|BST_ATTN_CAL|BST_USE_ANT_LEN|BST_FREQ_CAL, 32);
		pD->insert(pD, sChan, pCal);
		pI->insert(pI, 0xF20000 | uBand, pCal);
	}
	
	
	/* Okay, now that we have cal data holders and cal files, combine them */
	for(i = 0; i<pCalFiles->num; i++)
	{
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_ATTN_V")){
			load_bstcal_attns(pD, pCalFiles->pAry[i]);
			continue;
		}
	
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_BST_DIRECT_AMP_V")){
			load_bstcal_direct_amp(pD, pCalFiles->pAry[i]);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_BST_MIXED_AMP_V")){
			load_bstcal_mixed_amp(pD, pCalFiles->pAry[i]);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_BST_NBS_AMP_V")){
			load_bstcal_nbs_amp(pD, pCalFiles->pAry[i]);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_BST_NBS_FREQ_V")){
			load_bstcal_nbs_freq(pD, pCalFiles->pAry[i]);
			continue;
		}
		
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "I have no idea what cal file %s is"
		             " for.", pCalFiles->pAry[i]);
	}
	
	/* Don't need the cal files string array anymore */
	except_cleanup_pop(except_call);  /* pCalFiles */
	
	/* Okay now check and see if everything initialized correctly,
	   The keys for the cal tables are the channel names */
	for(sKey = pD->first(pD); sKey != NULL; sKey = pD->next(pD, sKey)){
		pConstCal = pD->get(pD, sKey);
		
		if(! pConstCal->isLoaded(pConstCal) )
			gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "Cal table %s was not fully "
			             "initialized, are you missing a CALIB file?", 
			             pConstCal->sChannel);
	}
	
	except_cleanup_pop(except_no_call); /* pThis */ 
	return pThis;
}

/*****************************************************************************/
/* Arrays of measurements instantiation */

#define C BstRowAry
#define T BstRow
#define CASCADE_DEL

#include <rpwgse/ptrAryTplt.c>

#undef CASCADE_DEL
#undef C
#undef T

/****************************************************************************/
/* Dictionary of uint4 to cal table objects */

#define C U4BstCalTblDict
#define K uint4
#define V BstCalTbl
#define DEL_VALS

#include <rpwgse/dictTplt.c>

#undef DEL_VALS
#undef V
#undef K
#undef C

