/** @file BstCalib.H Turn packets into calibrated data rows and write them out */

#ifndef _BstCalib_H_
#define _BstCalib_H_

#include <stdio.h>

#include "HskFineAdj.h"
#include "BstCalTbl.h"

#ifdef __cplusplus
extern "C" {
#endif


/****************************************************************************/
/* Arrays of burst data */

typedef struct bst_row{

   /* Internal only */
	byte uPsid;
	byte uSec;
	
	/* Bin Header */
	uint4 uQFacSec;
	uint2 uQFacSub;
	uint4 uQProcSec;
	uint2 uQProcSub;
	byte  uQfactor;
	
	/* Data row prefix */
	char sChannel[14]; /* Null terminated internally, needs space padding
	                      when written to disk */
	float8 dColSclk;  /* The spacecraft sclk at the start of the capture */
	/* SCET written to file after sclk */
	
	bool bNro;        /* Is noise reduction on? */
	bool bNra;        /* Was noise reduction applied */
	
	bool bBumped;     /* Needs to be bumped to Alt *UNC file 'cause two 
	                     waveforms are present */
	
	int1 nAmpCalVer;
	int1 nAttnCalVer;
	byte uPaAttn;
	byte uRecAttn;
	int1 nFreqCalVer;
	
	float4 fMixerHz; 
	float4 fClippedFrac;
	
	float4 fSampInt;   /* In seconds for waveforms, Hertz for spectra */
	float4 fFirstFreq; /* 0 (DC) for directly sampled data */
	
	/* Data */	
	float4buf_t data; /* calibrated amplitudes */
		
	void (*del)(void* vpThis);
}BstRow;

/* Do trivial initialization for burst row objects */
BstRow* new_BstRow(const BstCalTbl* pCal);

/****************************************************************************/
/* Arrays of measurements */

#define S bst_row_ary
#define C BstRowAry
#define T BstRow

#include <rpwgse/ptrAryTplt.h>

#undef S
#undef C
#undef T

/****************************************************************************/
/* Dictionary of uint4 to cal table objects */

#define S u4_bstcaltbl_dict
#define C U4BstCalTblDict
#define K uint4
#define V BstCalTbl

#include <rpwgse/dictTplt.h>

#undef V
#undef K
#undef C
#undef S


/****************************************************************************/
/** Packet to calibrated waveform converter.
 *
 * This class doesn't know or care about how the rows will be stored,
 * it just generates them from the input data.
 */
typedef struct bst_calib{
	
	/* Fill in a measurement from data in a packet */
	BstRowAry* (*cal)(const struct bst_calib* pThis, const GsePkt* pPkt);
	
	void (*del)(void* vpThis);
	
	/* The adjustment bureau -talk about an screwed up spelling word */
	HskFineAdj* pAdj;
	
	/* An dictionary of cal tables, holds calibration coffcients needed to
	   convert instrument values to physical units */
	StrBstCalTblDict* pTbls;
	
	/* Same dictionary as above, but (1) this dictionary doesn't own the
	   tables and (2) the keys are made up of the following byte values from
		the packet combinded into a uint4 as follows:  00 PSID SECTION BAND
	*/
	U4BstCalTblDict* pTblsById;
		
	
} BstCalib;


/** Create a new burst data calibrator
 * @public
 * @memberof BstCalib
 */
BstCalib* new_BstCalib(const char* sGroup, HskFineAdj* pAdj);

#ifdef __cplusplus
}
#endif

#endif /* _BstCalib_H_ */
