#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseMD5.h>
#include <rpwgse/GseConv.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>

#include "jno_spice.h"
#include "BstRowWriter.h"

/*****************************************************************************/
/* Sort the rows into the appropriate list */

bool BstRowWriter_add(BstRowWriter* pThis, BstRow* pRow)
{
	int4 iList = -1;
	
	switch(pRow->uPsid){
		
	/* HFR/HFWBR baseband */
	case 0x7B:
	case 0x7F:
	case 0x3F:
	case 0x3B:             /* Flight software change to 3F for this in 2010 */
		iList = BSTROWS_E;
		break;
		
	/* HFR/HFWBR I/Q data */
	case 0x88:
	case 0x89:
	case 0x8C:
	case 0x8D:
		iList = BSTROWS_MIX;
		break;
	
	/* HFR/HFWBR Spectra */
	case 0xF1:
	case 0xF2:
		iList = BSTROWS_NBS;
		break;		
	
	/* LFR lo/hi E */
	case 0xA2:
	case 0xA3:
	case 0xA4:
	case 0xA5:
		if(pRow->bBumped)
			iList = BSTROWS_EUNC;
		else
			iList = BSTROWS_E;
		break;
	
		
	/* LFR B */
	case 0xA0:
	case 0xA1:
	case 0x02:
		if(pRow->bBumped)
			iList = BSTROWS_BUNC;
		else
			iList = BSTROWS_B;
		break;
			
	default:
		gse_throw(GSE_XCEPT, GSE_XWTF);
	}
	
	
	pThis->rows[iList]->append(pThis->rows[iList], pRow);
	if(pThis->fQfAry[iList] == -1.0){
		pThis->fQfAry[iList] = pRow->uQfactor;
	}
	else{
		if(pThis->fQfAry[iList] != pRow->uQfactor){
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "0x%08X: Expected a QFactor of %.0f,"
					       " received QFactor of %u", (uint4) pRow->dColSclk,
					       pThis->fQfAry[iList], pRow->uQfactor);
		}
	}
	
	/* For 1MHz spectrum data we are going to have to have all rows be
	   the same length or else the frequency offsets are screwed up.
		I don't know what to do about this condition should it occur, but at
		least catch it */
	if((iList == BSTROWS_NBS) && (pThis->maxSamples[iList] != 0)){
		if(pRow->data.len != pThis->maxSamples[iList] )
			gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "This just in, 1MHz specta"
			             " length changes in the middle of a session, old num "
					       " samples %zu, new num samples %zu.  In other news "
			             " pigs are flying over Iowa City", 
		                pThis->maxSamples[iList], pRow->data.len);
	}
	
	if(pThis->maxSamples[iList] < pRow->data.len)
		pThis->maxSamples[iList] = pRow->data.len;
	
	/* Keep track of the overall session sclks */
	if( pThis->dSesStartSclk == 0.0){
		pThis->dSesStartSclk = pRow->dColSclk;
	}
	else{
		if(pThis->dSesStartSclk > pRow->dColSclk)
			pThis->dSesStartSclk = pRow->dColSclk;
	}
	
	if( pThis->dSesStopSclk == 0.0){
		pThis->dSesStopSclk = pRow->dColSclk + 1/40.0;
	}
	else{
		if(pThis->dSesStopSclk < (pRow->dColSclk + 1/40.0))
			pThis->dSesStopSclk = pRow->dColSclk + 1/40.0;
	}
	
	return true;
}

/*****************************************************************************/
/* Write the bin header.  Note the QFactor can't change during data collect  */
/* due to the check in BstRowWriter_add so just get the values from the 1st  */
/* packet in the stack */

uint4 _writeScet(float8 dSclk, byte* pWrite)
{
	char sScet[24] = {'\0'};
	size_t u = 0, uLen = 0;
	
	jno_sclk2str(dSclk, sScet, 24);
	
	uLen = strlen(sScet);
	memcpy(pWrite, sScet, uLen);
	pWrite += uLen;
	
	for(u = uLen; u < 23; u++){
		*pWrite = ' ';
		pWrite += 1;
	}
	
	*pWrite = '\0';

	return 24;
}

void BstRowWriter_writeBinHdr(BstRowWriter* pThis, FILE* pOut, BstRow* pRow,
                         GseMD5Hasher* pHasher, int4 nMaxSamples)
{
	
	bytebuf_t buf = {NULL, 0}; /* Get a buffer that's 76 + samples * 4 long */
	byte* pWrite = NULL;
	float8 dSclk  = 0.0;
	size_t uLen = 0;
	int4 i = 0; 
	
	buf.len = 76 + (nMaxSamples * 4);
	
	if(buf.len < 135)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Wow, this file's rows "
		             "are so short that the label template will need to be "
				       "changed, expected at least %d bytes got %d",
		             135, nMaxSamples*4 + 76);

	buf.ptr = (byte*)gse_calloc(buf.len, sizeof(byte));
	except_cleanup_push(free, buf.ptr);
	
	pWrite = buf.ptr; 
	gse_u4ToLSBbytes(pWrite, nMaxSamples*4 + 76);       /* Record length */
	pWrite += 4;
	
	gse_f8ToLSBbytes(pWrite, pThis->dSesStartSclk);     /* Session start SCLK */
	pWrite += 8;
	
	pWrite += _writeScet(pThis->dSesStartSclk, pWrite); /* Scet + 1 NULL */
	
	/* Okay, if this is a record session (QFactor == 0) then we need to
	   be careful and only store spaces for the SCET */
	
	if(pRow->uQfactor != 0){
	
		dSclk = (float8)(pRow->uQFacSec);                /* Q SCLK */
		dSclk += ((float8)pRow->uQFacSub) / 65536.0;
	
		gse_f8ToLSBbytes(pWrite, dSclk);
		pWrite += 8;
	
		pWrite += _writeScet(dSclk, pWrite);              /* Q SCET + 1 NULL */
	}
	else{
		pWrite += 8;                                      /* Q SCLK */
		
		for(i = 0; i<23; i++){
			*pWrite = ' ';  /* Space fill */
			pWrite += 1;
		}
		pWrite += 1; /* Null byte */
	}
	
	dSclk = (float8)(pRow->uQProcSec);
	dSclk += ((float8)pRow->uQProcSub)/ 65536.0;
	
	gse_f8ToLSBbytes(pWrite, dSclk);                    /* Proc SCLK */
	pWrite += 8;
	
	pWrite += _writeScet(dSclk, pWrite);                /* Proc SCET + 1 NULL */
	
	*pWrite = pRow->uQfactor;                     /* one and only qfactor !!! */
	pWrite += 1;
	
	pWrite += 3;                                       /* 3 NULL bytes */
	
	gse_f8ToLSBbytes(pWrite, pThis->dSesStopSclk);     /* Session start SCLK */
	pWrite += 8;
	
	pWrite += _writeScet(pThis->dSesStopSclk, pWrite); /* Scet + 1 NULL */
	
	/* Let them eat null's for the rest of the row... */
	
	/* Write to disk, with appropriate padding */	
	uLen = fwrite(buf.ptr, sizeof(byte), buf.len, pOut);
	
	if(uLen != buf.len)
		gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "Only %zu of %zu row bytes were "
		             "written.", uLen, buf.len);
	
	/* Update the hasher */
	pHasher->update(pHasher, buf.ptr, buf.len);
	
	except_cleanup_pop(except_call); /* done with buf.ptr */
}

/*****************************************************************************/
/* write the frequency offsets, since the A/D system for the HFRH channels   */
/* always runs at the same frequency, this is just dependent on the number   */
/* waveform samples collected */

void BstRowWriter_writeFreqOffsets(BstRowWriter* pThis, FILE* pOut,
                                   GseMD5Hasher* pHasher, int4 nMaxSamples)
{
	
	bytebuf_t buf      = {NULL, 0}; /* Get a buffer that's 76 + samples * 4 long */
	byte*     pWrite   = NULL;
	size_t    uLen     = 0;
	float8    dDeltaF  = 0.0;
	float8    dFreqOff = 0.0;
	int4      i        = 0;
	int4      iMix     = 0;
	
	/* make sure nMaxSamples is odd, the Mixer frequency should be in the 
	   middle with two even sides */
	if( (nMaxSamples  % 2) != 1)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Expected an odd number of "
		             "samples for the 1MHz spectra measurements, recieved %d",
		             nMaxSamples);
	
	buf.len = 76 + (nMaxSamples * 4);
	buf.ptr = (byte*)gse_calloc(buf.len, sizeof(byte));
	
	pWrite = buf.ptr;
	
	*pWrite = 'N'; pWrite += 1;
	*pWrite = 'B'; pWrite += 1;
	*pWrite = 'S'; pWrite += 1;
	for(i = 3; i < 14; i++){
		*pWrite = ' ';
		pWrite += 1;
	}
	
	/* okay, skip 58 nulls so that frequnce offsets line up with the 
	   amplitudes */
	pWrite += 58;
	
	/* Write the number of frequencies */
	gse_u4ToLSBbytes(pWrite, (uint4) nMaxSamples);
	pWrite += 4;
	
	/* Now, write one frequency offset for each data point */
	
	
	iMix = nMaxSamples/2;  /* this truncates, that's good */
	
	dDeltaF = BST_DOWN_MIXED_SAMPLE_RATE_HZ / (nMaxSamples + 1);
	
	for(i = 0; i < iMix; i++){
		dFreqOff = (i - iMix) * dDeltaF;
		gse_f4ToLSBbytes(pWrite, dFreqOff);
		pWrite += 4;
	}
	
	gse_f4ToLSBbytes(pWrite, 0.0); /* The mixer frequency */
	pWrite += 4;
	
	for(i = iMix + 1; i < nMaxSamples; i++){
		dFreqOff = (i - iMix) * dDeltaF;
		gse_f4ToLSBbytes(pWrite, dFreqOff);
		pWrite += 4;
	}
	
	/* Write the data */
	uLen = fwrite(buf.ptr, sizeof(byte), buf.len, pOut);
	
	if(uLen != buf.len)
		gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "Only %zu of %zu row bytes were "
		             "written.", uLen, buf.len);
	
	/* Okay update the hasher */
	pHasher->update(pHasher, buf.ptr, buf.len);
}

/*****************************************************************************/
/* Now down to brass tacks, write the row to the file, 1st write the         */
/* row in memory then send to the disk as a single operation, this will      */
/* hopefully prevent half written rows, in most cases. */

bytebuf_t g_bufRow = {NULL, 0};

void BstRowWriter_writeRow(BstRowWriter* pThis, FILE* pOut, BstRow* pRow,
		                     GseMD5Hasher* pHasher, int4 nMaxSamples)
{
	size_t u      = 0;
	size_t uLen   = 0;
	byte*  pWrite = 0;
	
	/* Make sure each row is at least as long as the header row, or else
	   we will have a problem */
	if(nMaxSamples*4 + 76 < 135)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "Wow, this file's rows "
		             "are so short that the label template will need to be "
				       "changed, expected at least %d bytes got %d",
		             135, nMaxSamples*4 + 76);
	
	if(g_bufRow.len == 0){
		if(g_bufRow.ptr) free(g_bufRow.ptr);
		g_bufRow.len = nMaxSamples*4 + 76;
		g_bufRow.ptr = (byte*)gse_calloc(g_bufRow.len, sizeof(byte));
	}
	
	/* Check that nMaxSamples doesn't change */
	gse_assert(g_bufRow.len == (nMaxSamples*4 + 76));
	
	/* 1st null the row so that we can just skip over null byte regions */
	memset(g_bufRow.ptr, 0, g_bufRow.len);
	
	pWrite = g_bufRow.ptr;
	
	uLen = strlen(pRow->sChannel);             /* Channel */
	memcpy(pWrite, pRow->sChannel, uLen);
	pWrite += uLen;
	for(u = uLen; u < 14; u++){
		*pWrite = ' ';
		pWrite += 1;
	}

	pWrite += 2;                               /* 2 NULLs */
	
	gse_f8ToLSBbytes(pWrite, pRow->dColSclk);  /* SCLK */
	pWrite += 8;
	
	pWrite += _writeScet(pRow->dColSclk, pWrite);  /* Scet + 1 NULL */
		
	if(pRow->bNro) *pWrite = 1;               /* NRD is on */
	pWrite += 1;
	
	if(pRow->bNra) *pWrite = 1;               /* NRD used on this row */
	pWrite += 1;
	
	*pWrite = pRow->nAmpCalVer;               /* Amplitude cal ver no */
	pWrite += 1;
	
	*pWrite = pRow->nAttnCalVer;              /* Attenuation cal ver no */
	pWrite += 1;
	
	*pWrite = pRow->uPaAttn;                  /* PreAmp attn setting */
	pWrite += 1;
	
	*pWrite = pRow->uRecAttn;                 /* Rec attn setting */
	pWrite += 1;
	
	*pWrite = pRow->nFreqCalVer;              /* Freq cal ver no */
	pWrite += 1;
	
	pWrite += 1;                              /* 1 NULL */
	
	gse_f4ToLSBbytes(pWrite, pRow->fMixerHz); /* Mixer freq */
	pWrite += 4;
	
	gse_f4ToLSBbytes(pWrite, pRow->fClippedFrac); /* Clipped fraction */
	pWrite += 4;
	
	gse_f4ToLSBbytes(pWrite, pRow->fSampInt); /* Sample interval (sec or hertz)*/
	pWrite += 4;
	
	gse_f4ToLSBbytes(pWrite, pRow->fFirstFreq); /* first frequency */
	pWrite += 4;
	
	gse_u4ToLSBbytes(pWrite, (uint4)pRow->data.len); /* Num points */
	pWrite += 4;
	
	
	/* Now write the amplitudes */
	for(u = 0; u<pRow->data.len; u++){
		gse_f4ToLSBbytes(pWrite, pRow->data.ptr[u]);
		pWrite += 4;
	}
	
	/* Write to disk, with appropriate padding */	
	uLen = fwrite(g_bufRow.ptr, sizeof(byte), g_bufRow.len, pOut);
	
	if(uLen != g_bufRow.len)
		gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "Only %zu of %zu row bytes were "
		             "written.", uLen, g_bufRow.len);
	
	/* Update the hasher */
	pHasher->update(pHasher, g_bufRow.ptr, g_bufRow.len);
}

/*****************************************************************************/
int4 BstRowWriter_cmp(const BstRow* p1, const BstRow* p2)
{
	/* All else being equaly: store via the following proiority (low to high):	*/
/*	static const byte uOrderB[] = {
		0xA1,0x02,0xA0,       / * LFR B * /
		0
	};
*/
	
	static const byte uOrderE[] = {
		0xA2,0xA4,          /* LFR Lo E */
		0xA3,0xA5,          /* LFR HI E */
		0x7B,0x7F,0x3F,0x3B,/* HFR Baseband */
		0x88,0x89,0x8C,0x8D,/* HFR Mixed Wfrm */
		0
	};
		
/*	static const byte uOrderNBS[] = {
		0xF1, 0xF2,    / * HFR Narrow Band Spectra * /
		0
	}; */
		
	if(p1->dColSclk != p2->dColSclk){
		if(p1->dColSclk > p2->dColSclk)
			return 1;
		else
			return -1;
	}

	int4 i, i1, i2;
	
	for(i = 0; uOrderE[i] != 0; i++){
		if(p1->uPsid == uOrderE[i]) i1 = i;
		if(p2->uPsid == uOrderE[i]) i2 = i;
	}
	
	if(i1 != i2){
		if(i1 > i2) return 1;
		else return -1;
	}
	
	/* Okay, I don't know how to sort these */
	uint4 u4 = (uint4)p1->dColSclk;
	
	gse_throw_fm(GSE_XCEPT, GSE_XASSERT, "%s: Encountered two packets with the "
	             "same SCLK: 0x%08X, id1: 0x%02X, id2: 0x%02X", u4,
	             p1->uPsid, p2->uPsid);
	
	return 0;
}


/*****************************************************************************/
/* Replace and Renumber 
 *
 * Given a temp file and it's hash, find the associated product file and see
 * what action should be taken based on it's hash.
 *
 * Expects temp paths to have .tmp on the end.
 */
char* BstRowWriter_reRe(BstRowWriter* pThis, const char* sTmpPath,
                        const char* sTmpHash)
{
	char* sDir              = NULL;
	GseStrList* pDir        = NULL;
	const char* sPath       = NULL;
	const char* sOldVerPath = NULL;
	char* sOldVerLbl        = NULL;
	char* sNewPath          = NULL;
	int4 nNewVer            = 1;
	GseMD5Hasher* pPrevHash = NULL;
	char sNewVer[16]        = {'\0'};
	int4 i                  = 0;       
	
	sDir = gse_dname(sTmpPath);
	
	/* Now find the previous version of this file, if it exists */
	pDir = gse_listdir(sDir, 1);
		
	for(sPath = pDir->get(pDir, 0); sPath != NULL;sPath = pDir->next(pDir))
		if(strncmp(sPath, sTmpPath, strlen(sTmpPath) - 10) == 0)
			if(strncmp(sPath + strlen(sPath) - 4, ".DAT", 4) == 0)
				sOldVerPath = sPath;
		
	if(sOldVerPath != NULL){
		if(!gse_str2int4(sOldVerPath + strlen(sOldVerPath) - 6, &nNewVer))
			gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "%s: Can't find file's version"
					       " number", sOldVerPath);
		else
			nNewVer += 1;
	
		/* Compare the hash */
		pPrevHash = gse_get_file_md5(sOldVerPath);
	
		if( strcmp(sTmpHash, pPrevHash->finalStr(pPrevHash)) == 0){
			gse_notice_v("%s: Updates unneccesary, temporary output discarded.",
		                strstr(sOldVerPath, "DATA"));
			
			pPrevHash->del(pPrevHash);
			free(sDir);
			gse_remove(sTmpPath);
			return CS(sOldVerPath);
		}
	}
	
	/* Going to be writing something, get it's permanent name */
	sprintf(sNewVer, "%02d", nNewVer);
		
	if(strlen(sNewVer) > 2){
		gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "%s: New version would be %d "
		             "but version numbers > 99 are not allowed by the VOLSIS.",
		             sTmpPath, nNewVer);
	}
	
	free(sDir);
	
	sNewPath = CS(sTmpPath);
	
	except_cleanup_push(free, sNewPath);
	sNewPath[ strlen(sNewPath) - 4] = '\0';
	sNewPath[ strlen(sNewPath) - 6] = sNewVer[0];
	sNewPath[ strlen(sNewPath) - 5] = sNewVer[1];
	
	/* If there was an old version mention that you are updating */
	if(sOldVerPath != NULL)
		gse_notice_v("%s: Updating to ver. %d", sOldVerPath, nNewVer);
	else
		gse_notice_v("%s: Writing", sNewPath);
	
	gse_rename(sTmpPath, sNewPath);
		
	if(sOldVerPath != NULL){
		gse_remove(sOldVerPath);
		i = strlen(sOldVerPath);
		sOldVerLbl = (char*)sOldVerPath;
		sOldVerLbl[i-3] = 'L';
		sOldVerLbl[i-2] = 'B';
		sOldVerLbl[i-1] = 'L';
			
		if(gse_isfile(sOldVerLbl)) gse_remove(sOldVerLbl);
	}
		
	except_cleanup_pop(except_no_call); /* sNewPath */
	
	return sNewPath;
}
	
/*****************************************************************************/
/* GseStrAry* */
void BstRowWriter_writeAll(
	BstRowWriter* pThis, const char* sGroup, JnoPhaser* pPhaser, 
	const gse_caltime_t* pStart)
{
	const mission_period_t* pPer = NULL;
	char* sDir             = NULL;
	int i                  = 0;
	BstRowList* pList      = NULL;
	int4 nMaxSamples       = 0;
	BstRow* pRow           = NULL;
	const char* sType      = NULL;
	char* sTmpPath         = NULL;
	FILE* fOut             = NULL;
	GseMD5Hasher* pHasher  = NULL;
	const char* sNewHash   = NULL;
	char* sFinal           = NULL;
	
	/* GseStrAry* pFinalPaths = NULL; */

	/* check unneeded, throws if can't ret. good phase*/
	pPer = pPhaser->getPhaseFromScet(pPhaser, pStart); 
	
	if(getenv("WAVES_DATA") == NULL)
		gse_throw(GSE_XCEPT, GSE_XMISSINGVAR);
	
	/* There are two versions of the path scheme, orbits which have one
	   directory for the period, and all the rest which use day-long sub dirs */
	if(pPer->bIsOrbit)
		sDir = NS("%s/%s/JNOWAV_1000/DATA/WAVES_BURST/%s", getenv("WAVES_DATA"),
		          sGroup, pPer->sSubDir);
	else
		sDir = NS("%s/%s/JNOWAV_1000/DATA/WAVES_BURST/%s/%04d_%03d",
		          getenv("WAVES_DATA"), sGroup, pPer->sSubDir, pStart->year,
				    pStart->doy);
	
	/* For each list that has data, open a tmp file and write the data */
	for(i = 0; i < BSTROWS_LISTS; i++){
		
		pList = pThis->rows[i];
		nMaxSamples = pThis->maxSamples[i];
		
		if(pList->length(pList) < 1)
			continue;
		
		/* Sort the data for this list */
		pList->sort(pList, BstRowWriter_cmp);
		
		if(pThis->fQfAry[i] == 0) 
			sType = "REC";
		else
			sType = "BST";
		
		sTmpPath = NS("%s/WAV_%04d%03dT%02d%02d%02.0f_%s_%s_Vxx.DAT.tmp", sDir, 
		           pStart->year, pStart->doy, pStart->hour, pStart->min,
		           pStart->sec, pThis->mnemonicAry[i], sType);
		
		/* Make the directories to the filename */
		gse_mkdirsto(sTmpPath, GSE_MODE_775);
				
		/* Write the whole file and update a hash of the contents */
		gse_debug_v("Opening Temp file: %s", sTmpPath);
		fOut = gse_fopen(sTmpPath, "wb");
		pHasher = new_GseMD5Hasher();
		
		/* Write the QFactor header here first */
		BstRowWriter_writeBinHdr(pThis, fOut, pList->get(pList, 0), pHasher,
				                   nMaxSamples);
		
		
		/* If this is the 1MHz spectrum table we are going to need to add in
		   the frequency offsets */
		if(i == BSTROWS_NBS)
			BstRowWriter_writeFreqOffsets(pThis, fOut, pHasher, nMaxSamples);
		
		
		/* clear out the row buffer before starting on a new product */
		if(g_bufRow.ptr) free(g_bufRow.ptr);
		g_bufRow.ptr = NULL;
		g_bufRow.len = 0;
		
		for(pRow = pList->get(pList, 0); pRow != NULL; pRow = pList->next(pList)){
			BstRowWriter_writeRow(pThis, fOut, pRow, pHasher, nMaxSamples);
		}
		
		sNewHash = CS(pHasher->finalStr(pHasher));
		
		pHasher->del(pHasher);
		fclose(fOut);
		
		/* Get the permanent location */
		sFinal = BstRowWriter_reRe(pThis, sTmpPath, sNewHash);
		
		/* pFinalPaths->add(pFinalPaths, sFinal); */
		
	}
	
	
	/* return pFinalPaths; */
}

/*****************************************************************************/
void delete_BstRowWriter(void* vpThis)
{
	int i;
	BstRowWriter* pThis = (BstRowWriter*)vpThis;
	
	for(i = 0; i < BSTROWS_LISTS; i++) 
		pThis->rows[i]->del(pThis->rows[i]);
	
	free(vpThis);
}


/*****************************************************************************/
BstRowWriter* new_BstRowWriter()
{
	BstRowWriter* pThis = NULL;
	int i;
		
	pThis = (BstRowWriter*) gse_calloc(1, sizeof(BstRowWriter));
	except_cleanup_push(delete_BstRowWriter, pThis);
	
	pThis->del      = delete_BstRowWriter;
	pThis->add      = BstRowWriter_add;
	pThis->writeAll = BstRowWriter_writeAll;	
	
	/* Init the storage */
	for(i = 0; i < BSTROWS_LISTS; i++){
		pThis->rows[i] = new_BstRowList();
		pThis->fQfAry[i] = -1.0;
		pThis->maxSamples[i] = 0;
	}
	
	pThis->mnemonicAry[BSTROWS_E] = "E";
	pThis->mnemonicAry[BSTROWS_B] = "B";
	pThis->mnemonicAry[BSTROWS_EUNC] = "EUNC";
	pThis->mnemonicAry[BSTROWS_BUNC] = "BUNC";
	pThis->mnemonicAry[BSTROWS_MIX]  = "MIX";
	pThis->mnemonicAry[BSTROWS_NBS] = "NBS";
	
	except_cleanup_pop(except_no_call); /* pThis */
	return pThis;
}


/*****************************************************************************/
/* Lists of measurements instantiation */

#define C BstRowList
#define I BstRow
#define CASCADE_DEL

#include <rpwgse/listTplt.c>

#undef CASCADE_DEL
#undef I
#undef C
