#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseMultiFileSrc.h>
#include <rpwgse/GseTime.h>
#include <rpwgse/GseMath.h>

#include <jwaves/JWavesConv.h>
#include <jwaves/JWavesChanMgr.h>

#include "HskFineAdj.h"

/*****************************************************************************/
bool HskFineAdj_adjust(HskFineAdj* pThis, float8* pSclk)
{
	/* Optimized for forward iteration 
	   1. Check to see that the current time is less than *pSclk, if not
		   reset pCurr to pFirst.
			
	   2. See if *pSclk >= pCur and < pNext.  If so, iterpolate, if not
		   increment pCur and try again, until you run off the end of the list
	*/
	int4 nCmp = (int4)(*pSclk);
	float8 fFine = 0.0;
	
	if(pThis->_pList->length(pThis->_pList) < 2){
		gse_error("Fine time list contains less that 2 items");
		return false;
	}
	
	if(pThis->_pCur->nSclk > *pSclk){
		pThis->_pCur = pThis->_pList->get(pThis->_pList, 0);
		pThis->_pNext = pThis->_pList->next(pThis->_pList);
	}
	
	/* Find the bounding item */
	bool bBounded = false;
	while(true){
		
		if((pThis->_pCur->nSclk <= nCmp)&&(pThis->_pNext->nSclk > nCmp)){
			bBounded = true;
			break;
		}
		
		pThis->_pCur = pThis->_pNext;
		pThis->_pNext = pThis->_pList->next(pThis->_pList);
		if(pThis->_pNext == NULL){
			/* At the end of the list, reset and return false */
			pThis->_pCur = pThis->_pList->get(pThis->_pList, 0);
			pThis->_pNext = pThis->_pList->next(pThis->_pList);
			break;
		}
	}
	
	if(bBounded){
		fFine = gse_lerp(pThis->_pCur->nSclk, pThis->_pNext->nSclk,
		                 pThis->_pCur->uFine, pThis->_pNext->uFine, *pSclk);
	}
	else{
		/* Okay am I within 10 minutes of the head of list (lowest value) */
		if((*pSclk < pThis->_pList->pHead->pItem->nSclk) &&
			(pThis->_pList->pHead->pItem->nSclk - *pSclk < 600)){
		
			fFine = pThis->_pList->pHead->pItem->uFine;
		}
		else{
			
			/* Okay am I within 10 minutes of the tail of the list (highest value) */
			if((*pSclk >= pThis->_pList->pTail->pItem->nSclk) &&
				(*pSclk - pThis->_pList->pTail->pItem->nSclk  < 600)){
		
				fFine = pThis->_pList->pTail->pItem->uFine;
			}
			else{
				/* You're screwed */
				gse_error_v("SCLK 0x%08X is > 10 minutes outside HSK fine adjust"
				            " list, are you missing HSK data?", *pSclk);
				return false;
			}
		}
	}
	
	fFine /= 65536.0;
		
	*pSclk = *pSclk + fFine;
	return true;
}

/*****************************************************************************/
void HskFineAdj_prnFines(HskFineAdj* pThis, FILE* pOut)
{
	fine_time_t* pFine = NULL;
	float4 dAdj = 0.0;
	int i = 0;
	struct timeval tv;
	char sTime[32];

	for(pFine = pThis->_pList->get(pThis->_pList, 0);
		 pFine != NULL;
		 pFine = pThis->_pList->next(pThis->_pList)){
		
		dAdj = ((float4)pFine->uFine) / 65536.0;
		
		jwaves_sclk2tv(pFine->nSclk, &tv);
		gse_tv2iso_gm_s(&tv,GSE_MILLISECS, sTime, 31);
		
		fprintf(pOut, "%4i  SCET: %s  SCLK: 0x%08X  ADJ: 0x%04X (%.5f)\n", i, 
		        sTime, pFine->nSclk, pFine->uFine, dAdj);
		i++;
	}
}

/*****************************************************************************/
void delete_HskFineAdj(void* vpThis)
{
	HskFineAdj* pThis = (HskFineAdj*)vpThis;
	if(pThis->_pList != NULL) pThis->_pList->del(pThis->_pList);
}

/*****************************************************************************/
HskFineAdj* new_HskFineAdj(const char* sGroup, const char* sBeg,
		                     const char* sEnd)
{
	HskFineAdj* pThis    = NULL;
	GsePktSrc*  pSrc     = NULL;
	GseFileChanMgr* pMgr = NULL;
	GsePkt* pPkt         = NULL;
	fine_time_t ftCur    = {0,0};
	fine_time_t ftLast   = {0,0};
	fine_time_t* pFine   = NULL;
	cbytebuf_t cbuf      = {NULL, 0};
	
	pThis = (HskFineAdj*)gse_calloc(1, sizeof(HskFineAdj));
	except_cleanup_push(delete_HskFineAdj, pThis);
	
	/* Hook up functions and data */
	pThis->adjust   = HskFineAdj_adjust;
	pThis->del      = delete_HskFineAdj;
	pThis->prnFines = HskFineAdj_prnFines;
	pThis->_pList   = new_FinesList();
			
	/* read without header block interperatation, we don't need it for this */
	pMgr = (GseFileChanMgr*) new_JWavesChanMgr(sBeg, sEnd, "hsk.dat", sGroup);
	pSrc = (GsePktSrc*)new_GseMultiFileSrc(pMgr, gse_generic_blocks, at_eof_stop,
                                          10);
	
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
		
		if( (pPkt->pHdr->uContent & 0xF0) != 0x60 ){
			gse_error_v("%s: Non-housekeeping data in file", pSrc->name(pSrc));
			goto CLEAN_PKT;
		}
		
		cbuf = pPkt->getData(pPkt);
			
		if(cbuf.len != 128){
			gse_error_v("%s: HSK packet data was %zu bytes, expected 128",
			            cbuf.len, pSrc->name(pSrc));
			goto CLEAN_PKT;
		}
		
		if((cbuf.ptr[7] & 0x1) != 0){
			gse_error_v("%s: Waves on internal time, skipping Fine time field",
			            pSrc->name(pSrc));
			goto CLEAN_PKT;
		}
			
		ftCur.nSclk = (gse_u4FromMSBbytes(cbuf.ptr + 4) & 0xFFFFFFFE);
		ftCur.uFine = gse_u2FromMSBbytes(cbuf.ptr + 30);
		
		/* Only store this new fine time if different from the previous one
		   or at least 5 minutes away in time */
		if((ftCur.uFine != ftLast.uFine)||(ftCur.nSclk >= ftLast.nSclk + 5*60)){
			
			ftLast = ftCur;
			pFine = (fine_time_t*)gse_calloc(1, sizeof(fine_time_t));
			pFine->nSclk = ftCur.nSclk;
			pFine->uFine = ftCur.uFine;
			pThis->_pList->append(pThis->_pList, pFine);
		}
		
CLEAN_PKT:
		pPkt->del(pPkt);
	}
	
	pSrc->del(pSrc);
	
	/* Setup pointer for the adjust function */
	if(pThis->_pList->length(pThis->_pList) > 1){
		pThis->_pCur = pThis->_pList->get(pThis->_pList, 0);
		pThis->_pNext = pThis->_pList->next(pThis->_pList);
	}
		
	if(pThis->_pList->length(pThis->_pList) < 2)
		gse_error_v("Less than two valid fine time fields in HSK.DAT files from %s to %s",
		            sBeg, sEnd);
		
	except_cleanup_pop(except_no_call); /* pThis */
	return pThis;
}

/*****************************************************************************/
/* Manually instantiate the template (Since we can't use C++, thanks for the */
/* extra work guys.) */

#define C FinesList
#define I fine_time_t
#define CASCADE_FREE

#include <rpwgse/listTplt.c>

#undef CASCADE_FREE
#undef I
#undef C
