#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <ctype.h>
#include <strings.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseTime.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseConv.h>

/* #include <jwaves/JWavesConv.h> */
#include <jwaves/JWavesDefs.h>



#include "jno_spice.h"
#include "JnoPhaser.h"

/****************************************************************************/
/* Compiled in config file directory location.*/
#ifndef CFG
#error Compiled in default configuration directory is missing.
#endif

#define _QDEF(x) #x
#define QDEF(x) _QDEF(x)

/*****************************************************************************/
/* Given a juno SCLK, determine the subdirectory which should contain the 
   data */
const mission_period_t* JnoPhaser_getPhase(const JnoPhaser* pThis, float8 dSclk)
{
	const PeriodListNode* pNode = NULL;
	const mission_period_t* pPer = NULL;
	char sScet[32];
	
	/* 1st See if I'm in an orbit first, skip the convienence functions because
	   they modify things. */
	for(pNode = pThis->_pOrbits->pHead; pNode != NULL; pNode = pNode->pNext){
		pPer = pNode->pItem;
		if((pPer->dSclkBeg <= dSclk) && (dSclk < pPer->dSclkEnd))
			return pPer;
	}
	
	/* Okay try for a general period next */
	for(pNode = pThis->_pPhases->pHead; pNode != NULL; pNode = pNode->pNext){
		pPer = pNode->pItem;
		if((pPer->dSclkBeg <= dSclk) && (dSclk < pPer->dSclkEnd))
			return pPer;
	}
		
	gse_error_v("Time 0x%08X (%s) doesn't fit into any mission phase or orbit"
			      "number, check your config files.", dSclk,
	            jno_sclk2str(dSclk, sScet, 32));
	
	return NULL;  /* Well, no dice, see if config files are up2date */
}

/*****************************************************************************/
/* Given a juno SCET, determine the subdirectory which should contain the
   data */

const mission_period_t* JnoPhaser_getPhaseFromScet(const JnoPhaser* pThis,
		                                             const gse_caltime_t* pCt)
{
	float8 dSclk;
	
	dSclk = jno_ct2sclk(pCt);
	return pThis->getPhase(pThis, dSclk);
}

/*****************************************************************************/
void _prnPeriods(PeriodList* pPeriods, FILE* pOut)
{
	char sBeg[64];
	char sEnd[64];
	
	gse_caltime_t* pCt = NULL;
	
	mission_period_t* pPeriod = NULL;
	for(pPeriod = pPeriods->get(pPeriods, 0);
		 pPeriod != NULL;
		 pPeriod = pPeriods->next(pPeriods)){
		
		pCt = &(pPeriod->ctBeg);
		sprintf(sBeg, "%04d-%03dT%02d:%02d:%06.3f", pCt->year, pCt->doy,
		        pCt->hour, pCt->min, pCt->sec);
		
		pCt = &(pPeriod->ctEnd);
		sprintf(sEnd, "%04d-%03dT%02d:%02d:%06.3f", pCt->year, pCt->doy,
		        pCt->hour, pCt->min, pCt->sec);
		
		fprintf(pOut, "NAME: \"%s\"  BEG: %s  END: %s  SCLKBEG: %09.3f  "
				  "SCLKEND: %09.3f  SUBDIR: %s\n", pPeriod->sName, sBeg, sEnd,
		        pPeriod->dSclkBeg, pPeriod->dSclkEnd, pPeriod->sSubDir);
	}
}


void JnoPhaser_prnPhases(JnoPhaser* pThis, FILE* pOut)
{
	_prnPeriods(pThis->_pPhases, pOut);	
}

void JnoPhaser_prnOrbits(JnoPhaser* pThis, FILE* pOut)
{
	_prnPeriods(pThis->_pOrbits, pOut);	
}

/*****************************************************************************/
/* Destruction */
void delete_JnoPhaser(void* vpThis)
{
	JnoPhaser* pThis = (JnoPhaser*)vpThis;
	
	if(pThis->_pPhases != NULL)
		pThis->_pPhases->del(pThis->_pPhases);
	
	if(pThis->_pOrbits != NULL)
		pThis->_pOrbits->del(pThis->_pOrbits);
}

/*****************************************************************************/
/* Construction */

void _timeFromCol(const char* sCfgFile, size_t nLine, const char* sLine,
                  const char* sColName, int4 nCol, gse_caltime_t* pCt,
                  float8* pSclk)
{
	char* sRawField = NULL;
	char* sField = NULL;
	
	if( (sRawField = gse_csv_getField(sLine,nCol)) != NULL){
		sField = gse_strip(sRawField, "\"");
		free(sRawField);
	}
		
	if((sField == NULL) || 
		(! gse_str2ct_gm(sField, pCt, false) ) )
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Error in %s on line %d",
		             sCfgFile, sColName, nLine);
			
	*pSclk = jno_str2sclk(sField);
}

mission_period_t* _ReadPeriod(const char* sCfgFile, int4 nLine,
		const char* sLine)
{
	char* sRawName = NULL;
	char* sName = NULL;
	int4 nFields = gse_csv_fields(sLine);
	
	mission_period_t* pPeriod = 
			(mission_period_t*)gse_calloc(1, sizeof(mission_period_t));
	
	
	except_cleanup_push(free, pPeriod);
	
	if(nFields != 3)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Syntax error on line %d, "
		             "expected 3 fields not %d", sCfgFile, nLine, nFields);
	
	_timeFromCol(sCfgFile, nLine, sLine, "start time", 0, &(pPeriod->ctBeg),
	             &(pPeriod->dSclkBeg));
	
	_timeFromCol(sCfgFile, nLine, sLine, "stop time", 1, &(pPeriod->ctEnd),
	             &(pPeriod->dSclkEnd));
	
	/* CHECK PERIODS last longer than a day! */
	
	if( (sRawName = gse_csv_getField(sLine,2)) != NULL){
		sName = gse_strip(sRawName, "\"");
		free(sRawName);
	}
	
	if(sName == NULL)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Empty phase/orbit name on"
		             " line %d", sCfgFile, nLine);
	
	if(strlen(sName) > 63)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Phase/orbit name too > 63"
		             " characers on line %d", sCfgFile, nLine);
	
	strcpy(pPeriod->sName, sName);
	free(sName);
	
	except_cleanup_pop(except_no_call);
	
	return pPeriod;
}


JnoPhaser* new_JnoPhaser()
{
	const char* sPhaseFile = QDEF(CFG) "/pds/mission_phase.conf";
	const char* sOrbitFile = QDEF(CFG) "/pds/orbits.conf";
	
	JnoPhaser* pThis     = NULL;
	FILE* fIn            = NULL;
	char sLineBuf[256];  memset(sLineBuf, 0, 256);
	char* sLine          = NULL;
	int4 nLine           = 0;
	size_t i             = 0;
	size_t nLen          = 0;
	int4 nOrbit          = -1;
	mission_period_t* pPeriod = NULL;    
	
	pThis = (JnoPhaser*)gse_calloc(1, sizeof(JnoPhaser));
	except_cleanup_push(delete_JnoPhaser, pThis);
	
	/* Hook up the functions */
	pThis->del              = delete_JnoPhaser;
	pThis->getPhase         = JnoPhaser_getPhase;
	pThis->getPhaseFromScet = JnoPhaser_getPhaseFromScet;
	pThis->prnPhases        = JnoPhaser_prnPhases;
	pThis->prnOrbits        = JnoPhaser_prnOrbits;
	
	pThis->_pPhases = new_PeriodList();
	pThis->_pOrbits = new_PeriodList();
	
	
	/* Load the mission phase list */
	gse_notice_v("Loading mission phase table: %s", sPhaseFile);
	
	if(!gse_isfile(sPhaseFile))
		gse_throw_fm(GSE_XCEPT, GSE_XMISSINGFILE, "%s: Mission phase file missing",
		             sPhaseFile);
	
	fIn = gse_fopen(sPhaseFile, "rb");
	
	while( fgets(sLineBuf, 1023, fIn) != NULL){
		nLine += 1;
		if((sLine = gse_strip_cmnt(sLineBuf, '#')) == NULL)
			continue;
		
		pPeriod = _ReadPeriod(sPhaseFile, nLine, sLine);
		pPeriod->bIsOrbit = false;
		
		snprintf(pPeriod->sSubDir, 127, "%04d%03d_%s", 
		         pPeriod->ctBeg.year, pPeriod->ctBeg.doy, /*
		         pPeriod->ctEnd.year, pPeriod->ctEnd.doy,*/ pPeriod->sName);
		pPeriod->sSubDir[127] = '\0';
		
		nLen = strlen(pPeriod->sSubDir);
		for(i=0; i<nLen; i++){
			pPeriod->sSubDir[i] = toupper(pPeriod->sSubDir[i]);
			if(pPeriod->sSubDir[i] == ' ')
				pPeriod->sSubDir[i] = '_';
		}
		
		pThis->_pPhases->append(pThis->_pPhases, pPeriod);
		free(sLine);
	}
	fclose(fIn);
	pThis->_pPhases->sort(pThis->_pPhases, period_cmp);
	
	
	/* Load the orbit name list */
	gse_notice_v("Loading orbit number table: %s", sOrbitFile);
	
	if(!gse_isfile(sOrbitFile))
		gse_throw_fm(GSE_XCEPT, GSE_XMISSINGFILE, "%s: Orbit number file missing",
		             sOrbitFile);
	
	fIn = gse_fopen(sOrbitFile, "rb");
	nLine = 0;
	while( fgets(sLineBuf, 1023, fIn) != NULL){
		nLine += 1;
		if((sLine = gse_strip_cmnt(sLineBuf, '#')) == NULL)
			continue;
		
		pPeriod = _ReadPeriod(sOrbitFile, nLine, sLine);
		pPeriod->bIsOrbit = true;
		
		if(strcasecmp(pPeriod->sName, "N/A") == 0){
			free(pPeriod);
			continue;
		}
		
		nLen = strlen(pPeriod->sName);
		for(i=0;i<nLen;i++){
			if(!isdigit(pPeriod->sName[i]))
				gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d, orbit number %s "
			             "contains non-digits", sOrbitFile, nLine, pPeriod->sName);
		}
		
		if((!gse_str2int4(pPeriod->sName, &nOrbit))||(nOrbit < 0)){
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Line %d, orbit number %s "
			             "doesn't convert to a positive integer.", sOrbitFile, 
			             nLine, pPeriod->sName);
		}
		
		snprintf(pPeriod->sSubDir, 127, "%04d%03d_ORBIT_%02d",
		         pPeriod->ctBeg.year, pPeriod->ctBeg.doy, /*
				   pPeriod->ctEnd.year, pPeriod->ctEnd.doy, */ nOrbit);
		
		pPeriod->sSubDir[127] = '\0';
				
		pThis->_pOrbits->append(pThis->_pOrbits, pPeriod);
		free(sLine);
	}
	fclose(fIn);
	pThis->_pOrbits->sort(pThis->_pOrbits, period_cmp);
		
	except_cleanup_pop(except_no_call);
	return pThis;
}

/*****************************************************************************/
/* Mission period structure handling */

#define C PeriodList
#define I mission_period_t
#define CASCADE_FREE

#include <rpwgse/listTplt.c>

#undef CASCADE_FREE
#undef I
#undef C

int4 period_cmp(const mission_period_t* pOne, const mission_period_t* pTwo)
{
	return gse_ctcmp(&(pOne->ctBeg), &(pTwo->ctBeg));
}

