#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseMath.h>
#include <rpwgse/GseConv.h>

#include "SrvCalTbl.h"

/*****************************************************************************/
/* WARNING!!!! A GLOBAL, okay since this is app code but not for use in libs */
float8 g_pFtInstVal[512];
bool g_bInitFtInstVal = false;


/* consturctor helper: Init list of posible instrument values */
void _SrvCalTbl_initInstVals()
{
	int4 i = 0;
	float4 dVal = 0;
	int4 exp = 0;
	int4 frac = 0;
	byte* pB0 = (byte*)(&dVal);
	byte* pB1 = (byte*)(&dVal);
	
	
	if(g_bIsLSBfirst){
		pB0 += 3;             /* Points to the last address byte */
		pB1 += 2;             /* Points to the next to last address byte */
	}
	else{
		/* pB0 unchanged points to 1st address byte */
		pB1 += 1;         	 /* Points to next address byte */
	}
	
	for(i = 0; i<512; i++){
		exp = (i >> 3)&0x3F;
		frac = i & 0x7;
		
		exp -= 20;
		exp += 127;
		
		dVal = 0.0;
		
		/* Set the high byte to contain most of the exponent, mask sign bit */
		*pB0 =  (exp >> 1)&0x7F;
		
		/* Set the next byte to contain last of exponent + frac */
		*pB1 =  (((exp & 0x01) << 7) & 0x80) | ((frac << 4)&0x70); 
		
		g_pFtInstVal[i] = dVal;
	}
	g_bInitFtInstVal = true;
}

/*****************************************************************************/

float8 SrvCalTbl_convVal(const SrvCalTbl* pThis, size_t iChan, float8 dInstVal)
{
	
	int4 first, mid, last;
	int4 iRow = -1;
	int4 iPrevRow = -1;
	int4 iNextRow = -1;
	float dRet;
	
	if(!pThis->isLoaded(pThis))
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "Can't convert insturment values, "
		             "Cal table %s is not fully initalized", pThis->sName);
	
	/* Get the index of the matching inst values column, Thanks Fred Swartz !! */
	first = 0;
	last = pThis->uVals;
	while(first <= last){
		mid = (first + last) / 2; /* Compute the mid point */
		
		if(dInstVal > pThis->pInstVal[mid]){
			first = mid + 1;       /* bigger than mid, repeat search in last half */
		}
		else{
			if (dInstVal < pThis->pInstVal[mid]){
				last = mid - 1;
			}
			else{
				iRow = mid;
				break;
			}
		}
	}
	
	/* Found it! */
	if(iRow != -1){
		dRet = pThis->pChanFrms[iChan][iRow];
	}
	else{
		/* If couldn't land on a particular index, interp if between two of them */
		iPrevRow = first - 1;
		iNextRow = first;
		
		if(iPrevRow < 0)
			gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "%s: Instrument value %0.4e "
			             "< the mimimum %0.4e", pThis->sName, dInstVal,
			             pThis->pInstVal[0]);
			
		if(iNextRow >= pThis->uVals)
			gse_throw_fm(GSE_XCEPT, GSE_XARYINDEX, "%s: Instrument value %0.4e "
			             "> the maximum %0.4e", pThis->sName, dInstVal, 
					       pThis->pInstVal[pThis->uVals-1]);
		
		dRet = gse_lerp(pThis->pInstVal[iPrevRow],         /* X0 */
				          pThis->pInstVal[iNextRow],         /* X1 */
		                pThis->pChanFrms[iChan][iPrevRow], /* Y0 */
		                pThis->pChanFrms[iChan][iNextRow], /* Y1 */
		                dInstVal);                         /* Xinterp */
	}
	return dRet;
}
		
/*****************************************************************************/
/* Load a table's channel properties */

void SrvCalTbl_loadChanProp(SrvCalTbl* pThis, const char* sPath)
{
	size_t     nLen         = 0;
	char       sFile[128];  memset(sFile, 0, 128);
	char       sLine[1024]; memset(sLine, 0, 1024);
	FILE*      fIn          = NULL;
	int4       nLine        = 0;
	char       sField[128]; memset(sField, 0, 128);
	/* float8     dAntLen      = 0; Rolled into main table */
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading channel properties", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading channel properties", sPath);
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s: Couldn't get basename from "
		             "path", sPath);
	
	pThis->nPropVer = get_calfile_version(sPath);
	
	if((pThis->uLoadState & 0x2) == 0x2)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Properties already loaded!", sFile);
		
	/* WARNING: In general you CAN'T use fgets to read CSV files because 
	   newline characters are legal, as long as they are in quotes.  I
		can do this here but DON'T TAKE AS A GENERAL EXAMPLE !!! */
	
	fIn = gse_fopen(sPath, "rb");
		
	pThis->iPos = 0;
	nLine = 0;
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;
		nLen = strlen(sLine);
		if(sLine[nLen - 2] != '\r' ||sLine[nLen -1] != '\n')
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Line %d is non PDS "
			          "complient line ending.", sFile, nLine);
		
		if(nLine < 3) continue;
		
		if(pThis->iPos >= pThis->uChan)
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Contains more than "
			             "%d calibration values ", sFile, pThis->uChan);
		
		pThis->pCenters[pThis->iPos] = gse_csv_getFloat(sLine, 1, sFile, nLine);
		pThis->pNbws[pThis->iPos] = gse_csv_getFloat(sLine, 2, sFile, nLine);
		
		/* Effective antenna length can vary as a function of frequency */
		if((pThis->nPropVer == 0)||(!pThis->bUseAntLen))
			pThis->pAntLen[pThis->iPos] = 1.0;
		else
			pThis->pAntLen[pThis->iPos] = gse_csv_getFloat(sLine, 3, sFile, nLine);
					
		pThis->iPos += 1;
	}
		
	if(pThis->iPos < pThis->uChan)
		gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Contains too few channel "
			             "definitions. Read %d, expected %d.", sFile, pThis->iPos,
		                pThis->uChan);
	
	pThis->uLoadState |= 0x2;
	fclose(fIn);
}

/*****************************************************************************/
/* Load the table's conversion array */

void SrvCalTbl_loadConv(SrvCalTbl* pThis, const char* sPath)
{
	char       sFile[128];  memset(sFile, 0, 128);
	char       sField[128]; memset(sField, 0, 128);
	char       sLine[1024]; memset(sLine, 0, 1024);
	size_t     nLen         = 0;
	FILE*      fIn          = NULL;
	int4       nLine        = 0;
	float8     dVal         = 0.0;
	int8       nVal         = 0.0;
	float8     dFrac        = 0.0;
	size_t     iChan        = 0;
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading transfer function", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading transfer function", sPath);
	
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s: Couldn't get basename from "
		             "path", sPath);	
	
	pThis->nPropVer = get_calfile_version(sPath);
	
	if((pThis->uLoadState & 0x4) == 0x4)
		gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: value coversion already "
		             "loaded!", sFile);
	
	fIn = gse_fopen(sPath, "rb");
	nLine = 0;
	pThis->iPos = 0;
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;
		nLen = strlen(sLine);
		if(sLine[nLen - 2] != '\r' ||sLine[nLen -1] != '\n')
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Line %d is non PDS "
			          "complient line ending.", sFile, nLine);
			
		/* TODO: Put a check for the units in here */
		
		if(nLine < 3) continue;
		
		/* Check the instrument value is within 0.1% */		
		if(pThis->uChan == 38 /* HFR SA DATA */){
			nVal = gse_csv_getInt(sLine, 0, sFile, nLine);
			
			if(nVal != pThis->iPos)
				gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Line %d Instrument "
				             "value is expected to be %d", sFile, nLine, pThis->iPos);
			
			pThis->pInstVal[pThis->iPos] = (float8)pThis->iPos;
		}			
		else{
			dVal = gse_csv_getFloat(sLine, 0, sFile, nLine);
			
			dFrac = abs(dVal-g_pFtInstVal[pThis->iPos])/g_pFtInstVal[pThis->iPos];
			
			if(dFrac > 0.001)
				gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Line %d Instrument "
			          "value is off by more than 0.1%", sFile, nLine);
			
			pThis->pInstVal[pThis->iPos] = g_pFtInstVal[pThis->iPos];
		}
	
		/* Okay, load one VRMS conversion for all channels */
		for(iChan = 0; iChan < pThis->uChan; iChan++){
			
			dVal = gse_csv_getFloat(sLine, iChan + 1, sFile, nLine);				
			pThis->pChanFrms[iChan][pThis->iPos] = dVal;
		}			
		
		pThis->iPos += 1;
	}
	
	fclose(fIn);
		
	if(pThis->iPos < pThis->uVals)
		gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Contains too few rows. Read "
			          "%d, expected %d.", sFile, pThis->iPos, pThis->uVals);
	pThis->uLoadState |= 0x4;
}

/*****************************************************************************/
void SrvCalTbl_copyChanProp(SrvCalTbl* pThis, const SrvCalTbl* pOther)
{
	size_t i;
	
	/* This class currently has a fixed-up-front internal allocation size,
	   you could change this if needed, but for now we'll roll with it */
	
	if(pThis->uChan != pOther->uChan)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s: Can't copy in channel "
		             "properties from %s due to a mismatch in the number "
		             "of channels", pThis->sName, pOther->sName);
		
	for(i = 0; i<pThis->uChan; i++){
		pThis->pCenters[i] = pOther->pCenters[i];		
		pThis->pNbws[i] = pOther->pNbws[i];
		if(pThis->bUseAntLen)
			pThis->pAntLen[i] = pOther->pAntLen[i];
	}
	
	pThis->uLoadState |= (pOther->uLoadState & 0x2);
}

/*****************************************************************************/
bool SrvCalTbl_isLoaded(const SrvCalTbl* pThis)
{
	return (pThis->uLoadState == 0x7);
}

/*****************************************************************************/
/* destructor */

void delete_SrvCalTbl(void* vpThis)
{
	size_t i;
	SrvCalTbl* pThis = (SrvCalTbl*)vpThis;
	
	free(pThis->sName);
	free(pThis->pCenters);
	free(pThis->pNbws);
	
	if(pThis->pChanFrms != NULL){
		for(i = 0; i< pThis->uChan; i++)
			free(pThis->pChanFrms[i]);
	}
	
	free(pThis->pAttns);
	free(pThis);
}

/*****************************************************************************/

SrvCalTbl* new_SrvCalTbl(const char* sName, size_t uChannels, size_t uValues, 
                         size_t uAttns, bool bUseAtnLen)
{
	size_t i;
	
	if(uChannels <= 0)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "uChannels = %zu", uChannels);
		
	if(uValues <= 0)
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "uValues = %zu", uValues);
	
	if(sName == NULL)
		gse_throw(GSE_XCEPT, GSE_XNULLPTR);
	
	/* Init the global 9-bit float instrument values array, WARNING: THIS IS
	   NOT THREAD SAFE! */
	if( ! g_bInitFtInstVal ) _SrvCalTbl_initInstVals();
	
	SrvCalTbl* pThis = (SrvCalTbl*)gse_calloc(1, sizeof(SrvCalTbl));
	pThis->del          = delete_SrvCalTbl;
	pThis->loadChanProp = SrvCalTbl_loadChanProp;
	pThis->copyChanProp = SrvCalTbl_copyChanProp;
	pThis->loadConv     = SrvCalTbl_loadConv;
	pThis->isLoaded     = SrvCalTbl_isLoaded;
	pThis->convVal      = SrvCalTbl_convVal;
	
	pThis->sName = CS(sName);
	pThis->uChan = uChannels;
	
	pThis->pCenters = (float8*)gse_calloc(uChannels, sizeof(float8));
	pThis->pNbws = (float8*)gse_calloc(uChannels, sizeof(float8));
	pThis->pAntLen = (float8*)gse_calloc(uChannels, sizeof(float8));
	pThis->bUseAntLen = bUseAtnLen;
	
	pThis->uVals = uValues;
	pThis->pInstVal = (float8*)gse_calloc(uValues, sizeof(float8));
	
	pThis->pChanFrms = (float8**)gse_calloc(uChannels, sizeof(float8*));
	for(i=0;i<pThis->uChan;i++)
		pThis->pChanFrms[i] = (float8*)gse_calloc(uValues, sizeof(float8));
	
	pThis->uAttns = uAttns;
	if(uAttns > 0)
		pThis->pAttns = (attn_cal_t*)gse_calloc(uAttns, sizeof(attn_cal_t));
	
	return pThis;
}

/*****************************************************************************/
void load_named_attns(StrSrvCalTblDict* pDict, const char* sPath)
{
	char sFile[128]; memset(sFile, 0, 128);
	int4        nCalVer      = -1;
	char        sLine[1024]; memset(sLine, 0, 1024);
	size_t      nLine        = 0;
	size_t      nLen         = 0;
	FILE*       fIn          = NULL;
	char        sRawField[128]; memset(sRawField, 0, 128);
	char*       sField       = NULL;
	SrvCalTbl*  pTbl         = NULL;
	attn_cal_t* pAtn         = NULL;
	const char* sKey         = NULL;
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	/* Set the utility counter in each cal table to 0 */
	for(sKey = pDict->first(pDict); sKey != NULL; sKey = pDict->next(pDict, sKey)){
		pTbl = pDict->getMutable(pDict, sKey);
		pTbl->iPos = 0;
	}
	
	if(strstr(sPath, "CALIB"))
		gse_notice_v("%s: Loading attenuator calibrations", strstr(sPath, "CALIB"));
	else
		gse_notice_v("%s: Loading attenuator calibrations", sPath);
		
	nCalVer = get_calfile_version(sPath);
	
	fIn = gse_fopen(sPath, "rb");
	
		
	/* Read once to see how big the list is and to check PDS compliance */
	nLine = 0;
	while( fgets(sLine, 1023, fIn) != NULL){
		nLine += 1;
		
		nLen = strlen(sLine);
		if(sLine[nLen - 2] != '\r' ||sLine[nLen -1] != '\n')
			gse_throw_fm(GSE_XCEPT,GSE_XDECODE,"%s: Line %d is non PDS "
			          "complient line ending.", sFile, nLine);
			
		if(nLine < 3) continue;
		
		gse_csv_getField_s(sLine, 0, sRawField, 128);
		sField = gse_strip(sRawField, "\"");
		
		if( (pTbl = pDict->getMutable(pDict, sField)) == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Unknown receiver "
			             "section name %s", sFile, nLine, sField);
		
		if(pTbl->pAttns != NULL){
			
			pTbl->nAttnVer = nCalVer;
			
			if(pTbl->iPos >= pTbl->uAttns)
				gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Too many attenuator "
				             "settings for reciever section %s", sFile, nLine, sField);
			
			free(sField);
			
			pAtn = pTbl->pAttns + pTbl->iPos;
			
			pAtn->nPaVal = gse_csv_getInt(sLine, 1, sFile, nLine);
			pAtn->nRecVal = gse_csv_getInt(sLine, 2, sFile, nLine);
			pAtn->dDb = gse_csv_getFloat(sLine, 3, sFile, nLine);
			pAtn->dFactor = gse_csv_getFloat(sLine, 4, sFile, nLine);
		
			pTbl->iPos += 1;
		}
		else{
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s, line %d: Waves receiver "
			             "section %s doesn't have switchable attenuators.", sFile,
			              nLine, sField);
		}
	}
	
	fclose(fIn);
	
	/* make sure everyone is loaded */
	for(sKey = pDict->first(pDict); sKey != NULL; sKey = pDict->next(pDict, sKey)){
		pTbl = pDict->getMutable(pDict, sKey);
		
		if(pTbl->iPos != pTbl->uAttns)
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "%s: Not all attenuation "
			             "calibrations were present.", sFile);
	
		pTbl->uLoadState |= 0x1;
	}
}

/*****************************************************************************/
/* Cal Table to string dictionary */


#define C StrSrvCalTblDict
#define K char
#define V SrvCalTbl

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcmp(a, b)

#include <rpwgse/ptrDictTplt.c>

#undef KEY_COMPARE
#undef DEL_VALS
#undef FREE_KEYS

#undef V
#undef K
#undef C
