#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>

#include <jwaves/JWavesSDSblk.h>
#include <jwaves/JWavesConv.h>

#include "caldir_util.h"
#include "SrvCalib.h"


/*****************************************************************************/

bool SrvCalib_cal(const SrvCalib* pThis, const GsePkt* pPkt, uint2 uSec, 
                  srv_row_sec_t* pOut)
{
	JWavesSDSblk*              pSds     = NULL;
	uint4                      uKey     = 0;
	byte                       uPsid    = 0;
	byte                       uBand    = 0;
	const srv_row_sec_t*       pTplt    = NULL;
	size_t                     iAttn    = 0;
	size_t                     iChan    = 0;
	float8                     dFactor  = 1.0;  
	cbytebuf_t                 cbuf     = {0};
	float8                     dFrms    = 0.0;  /* Field rms magnitude */
	struct jwaves_float4_conv  conv     = {0};
	size_t                     uSamples = 0;
	size_t                     uExpect  = 0;
	size_t                     uOffset  = 0;
	float8                     dInstVal = 0.0;
	
	pSds = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	uPsid = pSds->uPSID;
	uBand = pSds->uMixer;
	
	uKey = ((uPsid<<16)&0xFF0000) | ((uSec << 8)&0xFF00) | (uBand&0xFF);
	
	if( (pTplt = pThis->pTplts->get(pThis->pTplts, uKey)) == NULL){
		gse_error_v("Can't locate section template for Pkt Type: 0x%02X, section "
		            "%d and NRD marker: 0x%02X", pSds->uPSID, uSec, uBand);
		return false;
	}
	
	/* Initialize the section... */
	memcpy(pOut, pTplt, sizeof(srv_row_sec_t));
	
	/* Save the collection sclk */
	pOut->dColSclk = ((float8)pSds->uColSCLK) + ((float8)pSds->uColRTI)/40.0;
	
	/* Get the attenuation factor */
	
	if((uSec == 1) || ( jwaves_isBsci(pSds->uPSID) )){
		pOut->nPreAttn = 0;  /* B and Noise paths do not use attenuators, */
		pOut->nRecAttn = 0;  /* ignore the packet contents for these */
	}
	else{
		pOut->nPreAttn = ((pSds->uAttn & 0x20) >> 5) & 0x1;
		if( jwaves_islfrsci(pSds->uPSID))
			pOut->nRecAttn = ((pSds->uAttn & 0x10) >> 4) & 0x1;
		else if( jwaves_ishfrsci(pSds->uPSID))
			pOut->nRecAttn = (pSds->uAttn & 0x0F);
		else
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Packet ID: 0x%02X is not science data!", pSds->uPSID);
		
		for(iAttn = 0; iAttn < pOut->pCalTbl->uAttns; iAttn++){
			if((pOut->pCalTbl->pAttns[iAttn].nPaVal == pOut->nPreAttn)&&
				(pOut->pCalTbl->pAttns[iAttn].nRecVal == pOut->nRecAttn))
				dFactor = pOut->pCalTbl->pAttns[iAttn].dFactor;
		}
	}
	
	
	/* Now get the calibrated data */
	if((pSds->uPSID == 0x46)||(pSds->uPSID == 0x47)){
		
		cbuf = pPkt->getData(pPkt);
		
		if(cbuf.len != pOut->pCalTbl->uChan){
			gse_error_v("Packet type 0x%02X: Expected %d samples of data but "
			            "recieved %d", pSds->uPSID, pOut->pCalTbl->uChan, 
			            pOut->pCalTbl->uChan);
			return false;
		}
		
		for(iChan = 0; iChan < cbuf.len; iChan++){
			dFrms = pOut->pCalTbl->pChanFrms[iChan][cbuf.ptr[iChan]];
			dFrms *= dFactor;
			
			/* Put into spectral density (V^2/m^2)/Hz */
			if(pOut->pCalTbl->bUseAntLen){
				dFrms /= pOut->pCalTbl->pAntLen[iChan];
			}
			
			pOut->dVals[iChan] = (dFrms * dFrms) / pOut->pCalTbl->pNbws[iChan];
			
			pOut->uVals += 1;
		}
	}
	else{
		
		uExpect = pOut->pCalTbl->uChan;
		uOffset = 0;
		if((pSds->uPSID == 0xF0)||(pSds->uPSID == 0xF4)||(pSds->uPSID == 0xF5)){
			uExpect *= 3;
			uOffset =  (pOut->pCalTbl->uChan) * uSec;
		}
		
		uSamples = jwaves_bytes2float4(pSds->uFmt, pPkt->getData(pPkt), 0, &conv);
		
		if(uSamples != uExpect){
			gse_error_v("Packet type 0x%02X: Expected %d samples of "
			            "data but received %d", pSds->uPSID, pTplt->uChan,
			            uSamples);
			free(conv.ptr);
			return false;
		}
		
		if((pSds->uPSID == 0xF0)||(pSds->uPSID == 0xF4)||(pSds->uPSID == 0xF5)){
			if(uSamples % 3 != 0)
				gse_error_v("Packet type 0x%02X: Number of measurements, %d, is "
				            "not evenly divisable by 3", pSds->uPSID, uSamples);
			uSamples = uSamples/3;
		}
				
		for(iChan = 0; iChan<uSamples; iChan++){
			
			dInstVal = conv.ptr[iChan+uOffset];
			
			dFrms = pOut->pCalTbl->convVal(pOut->pCalTbl, iChan, dInstVal);
			
			/* Note, the B channel and the noise lines do NOT have attenuators */
			if((uSec != 1)&&(! jwaves_isBsci(pSds->uPSID) ))
				dFrms *= dFactor;
			
			/* Put into spectral density (V^2/m^2)/Hz or (nT^2)/Hz */
			if(pOut->pCalTbl->bUseAntLen){
				dFrms /= pOut->pCalTbl->pAntLen[iChan];
			}

			pOut->dVals[iChan] = (dFrms * dFrms) / pOut->pCalTbl->pNbws[iChan];
		
			pOut->uVals += 1;
		}
		
		free(conv.ptr);
	}
		
	return true;
}


/*****************************************************************************/
void delete_SrvCalib(void* vpThis)
{
	SrvCalib* pThis = (SrvCalib*)vpThis;
	if(pThis->pTbls != NULL) pThis->pTbls->del(pThis->pTbls);
	if(pThis->pTplts != NULL) pThis->pTplts->del(pThis->pTplts);
	free(vpThis);
}

/*****************************************************************************/
/* Get a row section template based on the packet id, data section and       */
/* an indication on wether Noise reduction is turned on or off (this last    */
/* bit in info has been crammed into the band byte to make Willy's life      */
/* easier, even though it's not very clean                                   */

const srv_row_sec_t* SrvCalib_getTplt(const struct srv_calib* pThis,
                                      byte uPsid, byte uSec, byte uBand)
{
	const srv_row_sec_t* pTplt = NULL;
	uint4 uKey;
	
	uKey = ((uPsid<<16)&0xFF0000) | ((uSec << 8)&0xFF00) | (uBand&0xFF);
			
	if( (pTplt = pThis->pTplts->get(pThis->pTplts, uKey)) == NULL)
		gse_throw_fm(GSE_XCEPT, GSE_XNOKEY, "Can't locate section template for "
		             "Pkt Type: 0x%02X, section %d", uPsid, uSec);
	return pTplt;
}

/*****************************************************************************/
/* Consturctor helper, intialize all the row section templates */

void _SrvCalib_mkSec(SrvCalib* pThis, uint4 uPsid, uint4 uSec, uint4 uBand, 
                     const char* sTbl, int4 nNRDon, int4 nLfrKind,
		               int4 nHFRboard, const char* sBinPrefix)
{
	uint4 uKey = ((uPsid<<16)&0xFF0000) | ((uSec<<8)&0xFF00) | (uBand&0xFF);
	
	const SrvCalTbl* pTbl = pThis->pTbls->get(pThis->pTbls, sTbl);
	
	if(pTbl == NULL)
		gse_throw_fm(GSE_XCEPT, GSE_XNOKEY, "Calibration table %s is not loaded", sTbl);		             
	
	if(!pTbl->isLoaded(pTbl))
		gse_throw_fm(GSE_XCEPT, GSE_XNOTREADY, "Calibration table %s is not "
		             "ready for use", sTbl);
	
	srv_row_sec_t* pSec = (srv_row_sec_t*)gse_calloc(1, sizeof(srv_row_sec_t));
	if(sBinPrefix == NULL)
		pSec->sBinPrefix = sTbl;
	else
		pSec->sBinPrefix = sBinPrefix;
	
	/* Note, the noise and B-Field records do not have attenuators */
	if( (uSec == 2) || jwaves_isBsci(uPsid) )
		snprintf(pSec->sCalVer, 10, "%02d:%02d", pTbl->nPropVer,
		         pTbl->nCalVer);
	else
		snprintf(pSec->sCalVer, 10, "%02d:%02d:%02d", pTbl->nPropVer,
		         pTbl->nCalVer, pTbl->nAttnVer);
	
	pSec->sCalVer[9] = '\0'; /* Just to be safe */
	
	pSec->uChan = pTbl->uChan;
	
	pSec->nNRDon = nNRDon;
	pSec->nLfrKind = nLfrKind;
	pSec->nHFRboard = nHFRboard;
	
	pSec->pCalTbl = pTbl;
	
	pThis->pTplts->insert(pThis->pTplts, uKey, pSec);
}

/*****************************************************************************/
/* Constructor helper, get a caltable that matches a conversion values table */
/* name. */

char* _SrvCalib_calfile2tblkey(const char* sPath)
{
	size_t nLen = 0;
	size_t i = 0;
	size_t iStart = 0;
	size_t iEnd = 0;
	char* sKey = NULL;
	size_t j = 0;
	
	nLen = strlen(sPath);
	
	/* Count backwards until hitting a '/' */
	for(i = nLen - 1; i >= 0; i--){
		if(sPath[i] == '/'){
			iStart = i;
			break;
		}
	}
	
	if(iStart == 0) goto FILE2TBL_ERR;
	if(strstr(sPath, "ATTN"))
		iStart += 9;    /* Attenuator table is not Survey data specific */
	else
		iStart += 13;
	
	/* Count backwards until hitting a '_' */
	for(i = nLen - 1; i >= 0; i--){
		if(sPath[i] == '_'){
			iEnd = i;
			break;
		}
	}
	
	if(iEnd == 0) goto FILE2TBL_ERR;
	
	if(iStart >= iEnd) goto FILE2TBL_ERR;
	
	sKey = (char*)gse_calloc(iEnd - iStart + 1, sizeof(char));
	j = 0;
	for(i = iStart; i<iEnd; i++){
		sKey[j] = sPath[i];
		j++;
	}
	sKey[iEnd] = '\0';
	
	return sKey;
	
FILE2TBL_ERR:
	gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s: Can't determine which calibration "
	             "table corresponts to this file.", sPath);
	return NULL;
}

/*****************************************************************************/
SrvCalib* new_SrvCalib(const char* sGroup)
{
	GseCStrAry*    pPrefix   = NULL;
	GseStrAry*     pCalFiles = NULL;
	
	SrvCalib*      pThis    = NULL;
	StrSrvCalTblDict* pD	 = NULL;   /* Dictionary of caltable objects */
	size_t         i        = 0;
	SrvCalTbl*     pTbl1    = NULL;
	SrvCalTbl*     pTbl2    = NULL;
	char*          sCalName = NULL;
	const char*    sKey     = NULL;  
	
	/* Init the CALIB dir first */
	
	pPrefix = new_GseCStrAry(14);
	pPrefix->pAry[0] = "WAV_CAL_SRV_LFR_LO_PROP_V";
	pPrefix->pAry[1] = "WAV_CAL_SRV_LFR_LO_B_V";
	pPrefix->pAry[2] = "WAV_CAL_SRV_LFR_LO_E_V";
	pPrefix->pAry[3] = "WAV_CAL_SRV_LFR_LO_N_V";

	pPrefix->pAry[4] = "WAV_CAL_SRV_LFR_HI_PROP_V";
	pPrefix->pAry[5] = "WAV_CAL_SRV_LFR_HI_E_V";
	pPrefix->pAry[6] = "WAV_CAL_SRV_LFR_HI_N_V";
	
	pPrefix->pAry[7] = "WAV_CAL_SRV_HFR_LO_PROP_V";
	pPrefix->pAry[8] = "WAV_CAL_SRV_HFR_LO_44_V";
	pPrefix->pAry[9] = "WAV_CAL_SRV_HFR_LO_45_V";
	
	pPrefix->pAry[10] = "WAV_CAL_SRV_HFR_HI_PROP_V";
	pPrefix->pAry[11] = "WAV_CAL_SRV_HFR_HI_44_V";
	pPrefix->pAry[12] = "WAV_CAL_SRV_HFR_HI_45_V";
	pPrefix->pAry[13] = "WAV_CAL_ATTN_V";
	
	pCalFiles = init_calib_dir(sGroup, pPrefix);
	pPrefix->del(pPrefix);
	
	except_cleanup_push(pCalFiles->del, pCalFiles);
	
	pThis = (SrvCalib*)gse_calloc(1, sizeof(SrvCalib));
	except_cleanup_push(delete_SrvCalib, pThis);
	
	pThis->del = delete_SrvCalib;
	pThis->cal = SrvCalib_cal;
	pThis->getTplt = SrvCalib_getTplt;
			
	pThis->pTbls = new_StrSrvCalTblDict(NULL);
	pD = pThis->pTbls;
		
	pD->insert(pD, CS("LFR_LO_E"), new_SrvCalTbl("LFR_LO_E", 43, 512, 4, true));
	pD->insert(pD, CS("LFR_LO_B"), new_SrvCalTbl("LFR_LO_B", 43, 512, 0, false));
	pD->insert(pD, CS("LFR_LO_N"), new_SrvCalTbl("LFR_LO_N", 43, 512, 0, false));
	pD->insert(pD, CS("LFR_HI_E"), new_SrvCalTbl("LFR_HI_E", 18, 512, 4, true));
	pD->insert(pD, CS("LFR_HI_N"), new_SrvCalTbl("LFR_HI_N", 18, 512, 0, false));
	pD->insert(pD, CS("HFR_LO_44"), new_SrvCalTbl("HFR_LO_44", 27, 512, 32, true));
	pD->insert(pD, CS("HFR_LO_45"), new_SrvCalTbl("HFR_LO_45", 27, 512, 32, true));
	pD->insert(pD, CS("HFR_HI_44"), new_SrvCalTbl("HFR_HI_44", 38, 256, 32, true));
	pD->insert(pD, CS("HFR_HI_45"), new_SrvCalTbl("HFR_HI_45", 38, 256, 32, true));
	
	for(i = 0; i<pCalFiles->num; i++){
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_ATTN_V")){
			load_named_attns(pD, pCalFiles->pAry[i]);
			continue;
		}
			
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_SRV_LFR_LO_PROP_V")){
			pTbl1 = pD->getMutable(pD, "LFR_LO_E");
			pTbl1->loadChanProp(pTbl1, pCalFiles->pAry[i]);
			
			pTbl2 = pD->getMutable(pD, "LFR_LO_B");
			pTbl2->copyChanProp(pTbl2, pTbl1);
			
			pTbl2 = pD->getMutable(pD, "LFR_LO_N");
			pTbl2->copyChanProp(pTbl2, pTbl1);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i],"WAV_CAL_SRV_LFR_HI_PROP_V")){
			pTbl1 = pD->getMutable(pD, "LFR_HI_E");
			pTbl1->loadChanProp(pTbl1, pCalFiles->pAry[i]);
			
			pTbl2 = pD->getMutable(pD, "LFR_HI_N");
			pTbl2->copyChanProp(pTbl2, pTbl1);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_SRV_HFR_LO_PROP_V")){
			pTbl1 = pD->getMutable(pD, "HFR_LO_44");
			pTbl1->loadChanProp(pTbl1, pCalFiles->pAry[i]);
			
			pTbl2 = pD->getMutable(pD, "HFR_LO_45");
			pTbl2->copyChanProp(pTbl2, pTbl1);
			continue;
		}
		
		if(strstr(pCalFiles->pAry[i], "WAV_CAL_SRV_HFR_HI_PROP_V")){
			pTbl1 = pD->getMutable(pD, "HFR_HI_44");
			pTbl1->loadChanProp(pTbl1, pCalFiles->pAry[i]);
			
			pTbl2 = pD->getMutable(pD, "HFR_HI_45");
			pTbl2->copyChanProp(pTbl2, pTbl1);
			continue;
		}
		
		sCalName = _SrvCalib_calfile2tblkey(pCalFiles->pAry[i]);
		
		if( (pTbl1 = pD->getMutable(pD, sCalName)) == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XNOKEY, "No calibration table named %s",
			             sCalName);
		else
			pTbl1->loadConv(pTbl1, pCalFiles->pAry[i]);

		free(sCalName);
	}
	
	/* Make sure all calibration tables are fully initialized */
	for(sKey = pD->first(pD); sKey != NULL; sKey = pD->next(pD, sKey)){
		pTbl1 = pD->getMutable(pD, sKey);
		
		if(! pTbl1->isLoaded(pTbl1) ){
			gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Cal table %s is not "
			             "fully initialized", pTbl1->sName);
		}
	}
	
	
	/* Okay, now initialize the row section templates and hook them up to 
	   packet setions, there are 17! of these */
	pThis->pTplts = new_PktSecRowSecDict("(PSID,SEC)-to->RowSection");

	/* Willy's new markers in the band byte for Noise reduction indication */
	_SrvCalib_mkSec(pThis, 0x91, 0, 0xF0, "LFR_LO_B", 0, 1 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x92, 0, 0xF0, "LFR_LO_E", 0, 1 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x93, 0, 0xF0, "LFR_HI_E", 0, 1 /* S+N */, 0, NULL);
	
	_SrvCalib_mkSec(pThis, 0x91, 0, 0xF1, "LFR_LO_B", 1, 3 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x92, 0, 0xF1, "LFR_LO_E", 1, 3 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x93, 0, 0xF1, "LFR_HI_E", 1, 3 /* S   */, 0, NULL);
	
	_SrvCalib_mkSec(pThis, 0xF0, 0, 0xF3, "LFR_LO_B", 1, 1 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF0, 1, 0xF3, "LFR_LO_B", 1, 2 /* N   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF0, 2, 0xF3, "LFR_LO_N", 1, 3 /* S   */, 0, NULL);

	_SrvCalib_mkSec(pThis, 0xF4, 0, 0xF3, "LFR_LO_E", 1, 1 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF4, 1, 0xF3, "LFR_LO_E", 1, 2 /* N   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF4, 2, 0xF3, "LFR_LO_N", 1, 3 /* S   */, 0, NULL);
				
	_SrvCalib_mkSec(pThis, 0xF5, 0, 0xF3, "LFR_HI_E", 1, 1 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF5, 1, 0xF3, "LFR_HI_E", 1, 2 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF5, 2, 0xF3, "LFR_HI_N", 1, 3 /* N   */, 0, NULL);
	
	_SrvCalib_mkSec(pThis, 0x5F, 0, 0xF0, "HFR_LO_44", 0, 0, 44, "HFR_LO");
	_SrvCalib_mkSec(pThis, 0x5B, 0, 0xF0, "HFR_LO_45", 0, 0, 45, "HFR_LO");
	_SrvCalib_mkSec(pThis, 0x3B, 0, 0xF0, "HFR_LO_45", 0, 0, 45, "HFR_LO");
	
	
	
	/* Pre 2011-10 FSW */
	_SrvCalib_mkSec(pThis, 0x91, 0, 0, "LFR_LO_B", 1, 3 /* S */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x92, 0, 0, "LFR_LO_E", 1, 3 /* S */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0x93, 0, 0, "LFR_HI_E", 1, 3 /* S */, 0, NULL);
	
	_SrvCalib_mkSec(pThis, 0xF0, 0, 0, "LFR_LO_B", 1, 1 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF0, 1, 0, "LFR_LO_B", 1, 2 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF0, 2, 0, "LFR_LO_N", 1, 3 /* N   */, 0, NULL);

	_SrvCalib_mkSec(pThis, 0xF4, 0, 0, "LFR_LO_E", 1, 1 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF4, 1, 0, "LFR_LO_E", 1, 2 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF4, 2, 0, "LFR_LO_N", 1, 3 /* N   */, 0, NULL);
				
	_SrvCalib_mkSec(pThis, 0xF5, 0, 0, "LFR_HI_E", 1, 1 /* S   */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF5, 1, 0, "LFR_HI_E", 1, 2 /* S+N */, 0, NULL);
	_SrvCalib_mkSec(pThis, 0xF5, 2, 0, "LFR_HI_N", 1, 3 /* N   */, 0, NULL);
	
	_SrvCalib_mkSec(pThis, 0x5F, 0, 0x00, "HFR_LO_44", 0, 0, 44, "HFR_LO");
	_SrvCalib_mkSec(pThis, 0x5B, 0, 0x00, "HFR_LO_45", 0, 0, 45, "HFR_LO");
	_SrvCalib_mkSec(pThis, 0x3B, 0, 0x00, "HFR_LO_45", 0, 0, 45, "HFR_LO");
	
	_SrvCalib_mkSec(pThis, 0x47, 0, 0x01, "HFR_HI_44", 0, 0, 44, "HFR_HI");
	_SrvCalib_mkSec(pThis, 0x46, 0, 0x01, "HFR_HI_45", 0, 0, 45, "HFR_HI");
	
	except_cleanup_pop(except_no_call); /* pThis */ 
	except_cleanup_pop(except_call);    /* pCalFiles */
	return pThis;
}

/*****************************************************************************/
#define C PktSecRowSecDict
#define K uint4
#define V srv_row_sec_t

#define FREE_VALS

#include <rpwgse/dictTplt.c>

#undef FREE_VALS
#undef C
#undef K
#undef V
