#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <strings.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseMD5.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseFile.h>

#include "jno_spice.h"

#include "SrvRowWriter.h"

/****************************************************************************/

FILE* _SrvRowWriter_openFile(SrvRowWriter* pThis, float8 dSclk,
                            const srv_row_t* pRow)
{
	const mission_period_t* pPer = NULL;
	gse_caltime_t ctRow = {0};
	gse_caltime_t ctFileBeg = {0};
	gse_caltime_t ctFileEnd = {0};
	FILE* fOut = NULL;
	
	
	/* Lock in the max sections and file category variables */ 
	if(pThis->nMaxSec == 0){
		if((pRow->nSections < 1)||(pRow->nSections == 3)||(pRow->nSections > 4))
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Don't know how to handle a file"
		   	          "with %d sections.", pRow->nSections);
		else
			pThis->nMaxSec = pRow->nSections;
	}
	
	if(pThis->sCat == NULL) pThis->sCat = CS(pRow->sRowType);
	
	
	pPer = pThis->pPhaser->getPhase(pThis->pPhaser, dSclk);
		
	/* Get year and doy for sclk */
	jno_sclk2ct(dSclk, &ctRow);
		
	/* If the day containing this dSclk also contians the start of
	   the period, set the filename's time of day component to 
		match the start of the period, else set it to 00:00:00. */
	if((pPer->ctBeg.year == ctRow.year)&&(pPer->ctBeg.doy == ctRow.doy)){
		pThis->sFile = NS("WAV_%04d%03dT%02d%02d%02.0f_%s_Vxx.CSV.tmp",
		                  pPer->ctBeg.year, pPer->ctBeg.doy, pPer->ctBeg.hour,
		                  pPer->ctBeg.min, pPer->ctBeg.sec, pThis->sCat);
		ctFileBeg = pPer->ctBeg;
	}
	else{
		pThis->sFile = NS("WAV_%04d%03dT000000_%s_Vxx.CSV.tmp", ctRow.year,
		                  ctRow.doy, pThis->sCat);
		
		ctFileBeg.year = ctRow.year;
		ctFileBeg.doy = ctRow.doy;
		gse_ct_norm_doy(&ctFileBeg);
	}
		
	/* If the day containing ths dSclk also contains the end of the
	   phase, then set the dFileEnd to the end of the period, else set
		it to the beginning of the next day */
	if((pPer->ctEnd.year == ctRow.year)&&(pPer->ctEnd.doy == ctRow.doy)){
		ctFileEnd = pPer->ctEnd;
	}
	else{
		ctFileEnd.year = ctRow.year;
		ctFileEnd.doy  = ctRow.doy + 1;
		gse_ct_norm_doy(&ctFileEnd);
	}
		
	pThis->dFileBeg = jno_ct2sclk(&ctFileBeg);
	pThis->dFileEnd = jno_ct2sclk(&ctFileEnd);	
	
	/* Okay there are two versions of the path scheme, orbits that 
	   have one directory for the whole period, and all the rest which
		use day-long sub directories (yyyy_ddd) */
	if(pPer->bIsOrbit)
		/* No Day directories */
		pThis->sPath = NS("%s/%s/JNOWAV_1000/DATA/WAVES_SURVEY/%s/%s",
		                  getenv("WAVES_DATA"), pThis->sGroup, pPer->sSubDir,
		                  pThis->sFile);
	else
		/* Daily directories */
		pThis->sPath = NS("%s/%s/JNOWAV_1000/DATA/WAVES_SURVEY/%s/%04d_%03d/%s",
		                  getenv("WAVES_DATA"), pThis->sGroup, pPer->sSubDir,
		                  ctFileBeg.year, ctFileBeg.doy, pThis->sFile);
	
	
	/* Open the file */
	if(gse_log_getlevel() <= GSE_NOTICE){
		char sFileBeg[32];  
		char sFileEnd[32]; 
		sprintf(sFileBeg, "%04d-%03dT%02d:%02d:%06.3f", ctFileBeg.year,
	        ctFileBeg.doy, ctFileBeg.hour, ctFileBeg.min, ctFileBeg.sec);
		
		sprintf(sFileEnd, "%04d-%03dT%02d:%02d:%06.3f", ctFileEnd.year,
	        ctFileEnd.doy, ctFileEnd.hour, ctFileEnd.min, ctFileEnd.sec);
		
		gse_notice_v("Opening: %s for data from %s to %s", pThis->sFile, sFileBeg,
		             sFileEnd);
	}
	
	gse_mkdirsto(pThis->sPath, GSE_MODE_775);
	fOut = gse_fopen(pThis->sPath, "wb");

	/* Okay, write the header */
	size_t i = 0;
	size_t j = 0;
	size_t nCols = 3; /* SCLK, SCET, HAS_BURST */
	size_t nEmpties = 3;

	for(i=0;i<pRow->nSections;i++){
		if(pRow->sec[i].nLfrKind > 0){
			nCols += 7;  /* Offset, ATTN, ATTN, CAL_VER, NRD, LFRKIND, LFRKIND */
			nEmpties += 7;
		}
		else{
			nCols += 5;  /* Offset, ATTN, ATTN, CAL_VER, BOARD */
			nEmpties += 5;
		}
		
		nCols += pRow->sec[i].uChan;  /* Frequencies */
	}
	
	/* Description Rows */
	fprintf(fOut, "\"%s\"", pRow->sDesc);
	for(i=0; i<nCols-1; i++) putc(',', fOut);
	fputs("\r\n", fOut);
	
	fputs("\"The 4th line of this file contains channel frequency centers in "
	      "Hertz\",,,,\"The 5th line of this file contains channel noise "
			"bandwidths in Hertz\"", fOut);
	
	for(i=0; i<nCols-5; i++) putc(',', fOut);
	fputs("\r\n", fOut);
		
	
	/* Bin Names Row */
	for(i=0; i<nEmpties-1; i++) fputs(",", fOut);
	
	for(i=0; i<pRow->nSections; i++){
		for(j=1; j <= pRow->sec[i].uChan; j++)
			fprintf(fOut, ",\"%s_%02zu\"", pRow->sec[i].sBinPrefix, j);
	}
	fputs("\r\n", fOut);
	
	
	/* Header Sections + Frequency Centers (Hz) */
	fputs(",,", fOut);
	
	for(i=0;i<pRow->nSections;i++){
		if(pRow->sec[i].nLfrKind > 0){
			for(j=0; j<7; j++){
				fprintf(fOut, ",\"%s\"", pRow->sec[i].sBinPrefix);
			}
		}
		else{
			for(j=0; j<5; j++){
				fprintf(fOut, ",\"%s\"", pRow->sec[i].sBinPrefix);
			}
		}
	}
	for(i=0; i<pRow->nSections; i++){
		for(j=0; j < pRow->sec[i].uChan; j++)
			fprintf(fOut, ",\"%10.4e\"", pRow->sec[i].pCalTbl->pCenters[j]);
	}
	
	fputs("\r\n", fOut);
	
	
	/* "Header Columns" names + Noise Bandwidths */
	fputs("\"SCLK (sec)\",\"SCET\",\"HAS_BURST\"", fOut);
	
	for(i=0; i<pRow->nSections; i++){		
		fputs(",\"OFFSET (sec)\",\"PA_ATTN\",\"REC_ATTN\",\"CAL_VER\"", fOut);
		if(pRow->sec[i].nLfrKind == 0)
			fputs(",\"BOARD_NUM\"", fOut);
		else
			fputs(",\"NR_ON\",\"LFR_SRC_NUM\",\"LFR_SRC_NAME\"", fOut);
	}
	for(i=0; i<pRow->nSections; i++){
		for(j=0; j < pRow->sec[i].uChan; j++)
			fprintf(fOut, ",\"%10.4e\"", pRow->sec[i].pCalTbl->pNbws[j]);
	}
	fputs("\r\n", fOut);
	
	return fOut;
}


/****************************************************************************/
/* Row contains:                                                            */
/*                                                                          */
/* SCLK,SCET,HAS_BURST                                                      */
/* Offset, ATTN, ATTN, CAL_VER, NRD, LFRKIND, LFRKIND  <= Each LFR Section  */
/* Offset, ATTN, ATTN, CAL_VER                         <= Each HFR Section  */
/* Samples                                                                  */
/* Samples  (Opt)                                                           */
/* Samples  (Opt)                                                           */
/* Samples  (Opt)                                                           */
/*                                                                          */
/****************************************************************************/

const char* g_lfrSpecType[4] = {NULL, "S+f(N)", "f(N)", "S+f(N)-m(N)"};

void _SrvRowWriter_writeRowToFile(SrvRowWriter* pThis, const srv_row_t* pRow)
{
	size_t i = 0;
	size_t j = 0;
	
	/* Check the number of sections and the row type against locked in vals */
	
	if(pRow->nSections != pThis->nMaxSec)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "%s: Row contains %d sections, "
		"expected %d", pRow->sBaseScet, pRow->nSections, pThis->nMaxSec);

	if(strcmp(pRow->sRowType, pThis->sCat) != 0)
		gse_throw_fm(GSE_XCEPT, GSE_XENCODE, "%s: Row is of type '%s', expected"
		             " '%s'", pRow->sRowType, pThis->sCat);
	
	/* Header Section */
	
	/* SCLK,SCET,HAS_BURST */
	fprintf(pThis->fOut, "%13.3f", pRow->dBaseSclk); /* SCLK */
	fprintf(pThis->fOut, ",%s,0", pRow->sBaseScet); /* SCET, has burst */
	
	for(i = 0; i<pRow->nSections; i++){
		
		/* ,Offset ,ATTN ,ATTN ,CAL_VER */
		if(pRow->sec[i].uVals == 0)	
			fputs(",,,,", pThis->fOut);
		else
			fprintf(pThis->fOut, ",%.3f,%d,%d,\"%s\"", pRow->sec[i].dSclkOff,
					  pRow->sec[i].nPreAttn, pRow->sec[i].nRecAttn,
					  pRow->sec[i].sCalVer);
		
		
		if(pRow->sec[i].nLfrKind == 0){
			/* HFR Board Number */
			if(pRow->sec[i].uVals == 0)
				fputs(",", pThis->fOut);
			else
				fprintf(pThis->fOut, ",%d", pRow->sec[i].nHFRboard);
		}
		else{
			/* NRD, LFRKIND, LFRKIND */
			
			if(pRow->sec[i].uVals == 0)
				fputs(",,,", pThis->fOut);
			else
				fprintf(pThis->fOut, ",%d,%d,\"%s\"", pRow->sec[i].nNRDon,
				        pRow->sec[i].nLfrKind, 
				        g_lfrSpecType[pRow->sec[i].nLfrKind]);
		}
	}
	
	
	/* Spectra Densities (Desc from pRow should give units!!) 
	   NOTE: nVals can range from 0 to centers array length */
	for(i=0; i< pRow->nSections; i++){
		
		if(pRow->sec[i].uVals > 0)
			pThis->nSectionsOut += 1;  /* Basically a count of packets out */
		
		for(j=0; j< pRow->sec[i].uChan; j++){
			
			if(j >= pRow->sec[i].uVals )
				fputc(',', pThis->fOut);
			else
				fprintf(pThis->fOut, ",%10.4e", pRow->sec[i].dVals[j]);
		}
	}
	
	/* End of row */
	fputs("\r\n", pThis->fOut);
	
	pThis->nRowsOut += 1;
}

/****************************************************************************/
int4 SrvRowWriter_secsWritten(const SrvRowWriter* pThis){
	return pThis->nSectionsOut;
}

int4 SrvRowWriter_rowsWritten(const SrvRowWriter* pThis){
	return pThis->nRowsOut;
}

/****************************************************************************/
/* Replace and renumber  - all files are written as _Vxx.CSV.tmp, move this */
/* into the premanent position.                                             */
void _SrvRowWriter_reRe(SrvRowWriter* pThis)
{
	GseStrList*   pList       = NULL;
	size_t        i           = 0;
	GseMD5Hasher* pCurHash    = NULL;
	GseMD5Hasher* pPrevHash   = NULL;
	char*         sDir        = NULL;
	const char*   sPath       = NULL;
	const char*   sOldVerPath = NULL;
	int4          nNewVer     = 1;
	bool          bDelOutput  = false;
	char          sNewVer[16]; memset(sNewVer, 0, 16);
	char*         sNewPath    = NULL;
	char*         sOldVerLbl  = NULL;
	
	fclose(pThis->fOut); /* Close the current file */
	pThis->fOut = NULL;
	
	/* get the file who's name matches the pattern sFile[-10] and ends in *.CSV */
	sDir = CS(pThis->sPath);
	except_cleanup_push(free, sDir); /* sDir */
	
	for(i=strlen(sDir) - 1; i >= 0; i--){
		if(sDir[i] == '/'){
			sDir[i] = '\0';
			break;
		}
	}
	
	pList = gse_listdir(sDir, 1);
	
	except_cleanup_push(pList->del, pList);  /* pList */
	
	for(sPath = pList->get(pList, 0); sPath != NULL; sPath = pList->next(pList))
		if(strncmp(sPath, pThis->sPath, strlen(pThis->sPath) - 10) == 0)
			if(strncmp(sPath + strlen(sPath) - 4, ".CSV", 4) == 0)
				sOldVerPath = sPath;
	
	if(sOldVerPath != NULL){
		
		if(!gse_str2int4(sOldVerPath + strlen(sOldVerPath) - 6, &nNewVer))
			gse_throw_fm(GSE_XCEPT, GSE_XIOFAIL, "%s: Can't find file's version"
					       " number", sOldVerPath);
		else
			nNewVer += 1;
		
		/* Compare the hash */
		pCurHash = gse_get_file_md5(pThis->sPath);
		pPrevHash = gse_get_file_md5(sOldVerPath);
		
		if( memcmp(pCurHash->digest, pPrevHash->digest, 16) == 0)
			bDelOutput = true;
		
		pPrevHash->del(pPrevHash); pPrevHash = NULL;
		pCurHash->del(pCurHash);   pCurHash = NULL;
	}
	/* else{
		gse_notice_v("%s: No previous version to update.", 
		             strstr(pThis->sPath, "DATA"));
	}
	*/
	
	if(bDelOutput){
		gse_notice_v("%s: Updates unneccesary, temporary output discarded.",
			           strstr(sOldVerPath, "DATA"));	
		gse_remove(pThis->sPath);
	}
	else{
		/* Going to be writing something, get it's permanent name */
		sprintf(sNewVer, "%02d", nNewVer);
	
		if(strlen(sNewVer) > 2){
			gse_throw_fm(GSE_XCEPT, GSE_XNOTYETIMP, "%s: New version would be %d "
			             "but version numbers > 99 are not allowed by the VOLSIS.",
			             pThis->sFile, nNewVer);
		}
	
		sNewPath = CS(pThis->sPath);
		
		except_cleanup_push(free, sNewPath);
		sNewPath[ strlen(sNewPath) - 4] = '\0';
		sNewPath[ strlen(sNewPath) - 6] = sNewVer[0];
		sNewPath[ strlen(sNewPath) - 5] = sNewVer[1];
		
		/* If there was an old version mention that you are updating */
		if(sOldVerPath != NULL)
			gse_notice_v("%s: Updating file to version %d", sOldVerPath, nNewVer);
		else
			gse_notice_v("%s: Writing file", sNewPath);
	
		gse_rename(pThis->sPath, sNewPath);
		
		if(sOldVerPath != NULL){
			gse_remove(sOldVerPath);
			i = strlen(sOldVerPath);
			sOldVerLbl = (char*)sOldVerPath;
			sOldVerLbl[i-3] = 'L';
			sOldVerLbl[i-2] = 'B';
			sOldVerLbl[i-1] = 'L';
			
			if(gse_isfile(sOldVerLbl)) gse_remove(sOldVerLbl);
		}
		
		except_cleanup_pop(except_call); /* sNewPath */
	}		
	
	free(pThis->sPath); pThis->sPath = NULL;
	free(pThis->sFile); pThis->sFile = NULL;
	except_cleanup_pop(except_call); /* pList */
	except_cleanup_pop(except_call); /* sDir */
}

/****************************************************************************/
/* Can have 1 of three situations when a new row arrives:                   */
/*    1. Something open to the incorrect spot.                              */
/*    2. Nothing open.                                                      */
/*    3. Something open to the correct spot.                                */

void SrvRowWriter_writeRow(SrvRowWriter* pThis, const srv_row_t* pRow)
{
		
	/* 1. Something open to the incorrect spot, close out the file, will move
	 *    us to state 2 if needed. */
	if((pThis->fOut != NULL)&&
		((pRow->dBaseSclk < pThis->dFileBeg)||pThis->dFileEnd <= pRow->dBaseSclk))
		_SrvRowWriter_reRe(pThis);
	
	/* 2. Nothing open. Open up the file, will move us to state 3 if needed. */
	if(pThis->fOut == NULL)		
		pThis->fOut = _SrvRowWriter_openFile(pThis, pRow->dBaseSclk, pRow);
	
	
	/* 3. Something open to the right spot. */
	_SrvRowWriter_writeRowToFile(pThis, pRow);
	
}

/****************************************************************************/
void SrvRowWriter_finish(SrvRowWriter* pThis)
{
	if(pThis->fOut != NULL)
		_SrvRowWriter_reRe(pThis);
}

/****************************************************************************/
void delete_SrvRowWriter(void* vpThis)
{
	SrvRowWriter* pThis = (SrvRowWriter*)vpThis;
	if(pThis->fOut) fclose(pThis->fOut);
	free(pThis->sCat);
	free(pThis->sPath);
	free(pThis->sFile); 	
}

/****************************************************************************/
SrvRowWriter* new_SrvRowWriter(const JnoPhaser* pPhaser, const char* sGroup)
{
	SrvRowWriter* pThis = NULL;
	char* sGroupDir     = NULL;
	char* sVolRoot      = NULL;
	char* sDatRoot      = NULL;
		
	if(pPhaser == NULL)  gse_throw_m(GSE_XCEPT, GSE_XBADARG, "pPhaser == NULL");
	if(sGroup == NULL)   gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sGroup == NULL");
	
	pThis = (SrvRowWriter*)gse_calloc(1, sizeof(SrvRowWriter));
	except_cleanup_push(delete_SrvRowWriter, pThis);
	
	pThis->writeRow = SrvRowWriter_writeRow;
	pThis->secsWritten = SrvRowWriter_secsWritten;
	pThis->rowsWritten = SrvRowWriter_rowsWritten;
	pThis->finish = SrvRowWriter_finish;
	pThis->del = delete_SrvRowWriter;
	
	pThis->pPhaser = pPhaser;
	pThis->sGroup = sGroup;
	pThis->nMaxSec = 0;  /* Lock these in on the */
	pThis->sCat = NULL;  /* first row received   */
	
	/* Calloc gets the rest to appropriate NULL or 0 values */
	
	/* Make sure the DATA directory exists */
	sGroupDir = NS("%s/%s", getenv("WAVES_DATA"), sGroup);
	if(!gse_isdir(sGroupDir)){
		gse_error_v("Data group directory %s doesn't exist.", sGroupDir);
		return NULL;
	}
	
	sVolRoot = NS("%s/%s/JNOWAV_1000", getenv("WAVES_DATA"), sGroup);
	if(!gse_isdir(sVolRoot)){
		gse_mkdirs(sVolRoot, 0775);
		gse_notice_v("Creating Volume Directory: %s", sVolRoot);
	}
	
	sDatRoot = NS("%s/DATA", sVolRoot);
	if(!gse_isdir(sDatRoot)){
		gse_mkdirs(sDatRoot, 0775);
		gse_notice_v("Creating DATA Directory: %s", sDatRoot);
	}
	
	free(sGroupDir);
	free(sVolRoot);
	free(sDatRoot);
	
	except_cleanup_pop(except_no_call);
	return pThis;
}

/*****************************************************************************/
/* instantiate a string to SrvRowWriter dictionary */

#define C StrSrvWriterDict
#define K char
#define V SrvRowWriter

#define FREE_KEYS
#define DEL_VALS

#define KEY_COMPARE(a, b) strcasecmp(a, b)

#include <rpwgse/ptrDictTplt.c>

#undef FREE_KEYS
#undef DEL_VALS
#undef C
#undef K
#undef V

