#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif


#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseFile.h>

#include <jwaves/JWavesSDSblk.h>

#include "jno_spice.h"
#include "SrvRower.h"


/*****************************************************************************/
void SrvRower_setFlush(SrvRower* pThis){
	pThis->bFlush = true;
}

int4 SrvRower_inPktSecs(const SrvRower* pThis){
	return pThis->nSecIn;
}

/*****************************************************************************/
bool SrvRower_addPkt(SrvRower* pThis, const GsePkt* pPkt)
{
	JWavesSDSblk* pSci = NULL;
	float4 dSclk = 0.0;
	uint4 nBaseSclk = 0;
	uint4 nPktSecs = 1;
	uint4 uPktSec = 0;
	size_t i = 0;
	uint4 uRowList = 99;
	uint4 uRowSec = 99;
	
	SrvRowList* pList = NULL;
	srv_row_t*  pRow = NULL;
	const SrvCalib* pCalib = pThis->pCalib;  /* Make the rows shorter */
	
	/* Won't take anymore packets if in flush mode */
	if(pThis->bFlush)
		gse_throw_m(GSE_XCEPT, GSE_XNOTREADY, "Srv Data Rower has been set to "
		            "flush mode, cannot accept new packets");
	
	/* 1st determine which section this packet goes with */
	if(!pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0)){
		gse_debug("Ignoring non science packet");
		return false;
	}
	
	pSci = (JWavesSDSblk*)pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	/* Skip the LRF Q-Factor packets, we don't process them */
	if(pSci->uPSID == 0x21) return false;
	
	/* Alright, convert each section */
	if((pSci->uPSID == 0xF0)||(pSci->uPSID == 0xF4)||(pSci->uPSID == 0xF5))
		nPktSecs = 3;
	
	pThis->nSecIn += nPktSecs;
	
	/* Check timing */
	dSclk = ((float8)pSci->uColSCLK) + ((float8)pSci->uColRTI)/40.0;
	if(pThis->dLastSclk != -1.0){
		if(dSclk < pThis->dLastSclk){
			gse_error_v("New packet time, %.3f, is not >= to last packet time, "
			            "%.3f! Droping the packet. You need to sort the data.",
			            dSclk, pThis->dLastSclk);
			return false;	
		}
	}
	
	nBaseSclk = pSci->uColSCLK;
	if(pSci->uColRTI >= 20)
		nBaseSclk += 1;
	
	for(uPktSec = 0; uPktSec < nPktSecs; uPktSec++){
	
		/* Find which list this goes in */
		uRowList = 99;
		uRowSec = 99;
		for(i = 0; i<SRVROWER_NMAPS; i++){
			if( (pSci->uPSID == pThis->map[i].uPsid) &&
			    (uPktSec == pThis->map[i].uSec) ){
				
				uRowList = pThis->map[i].uListNum;
				uRowSec = pThis->map[i].uRowSec;
			}
		}
		
		if(uRowList == 99){
			gse_error_v("Can't determine row list for packet type 0x%02X, data "
			            "section %u", pSci->uPSID, uPktSec);
			return false;
		}
	
		/* If this second hasn't been seen, create a new row and add in the
		   converted values */
		pList = pThis->pLists[uRowList];
		if((pList->length(pList) < 1) || 
	      ( (pList->get(pList, -1))->nBaseSclk < nBaseSclk) ){
			
			pRow = (srv_row_t*)gse_calloc(1, sizeof(srv_row_t));
			memcpy(pRow, &(pThis->tplts[uRowList]), sizeof(srv_row_t));
			pRow->nBaseSclk = nBaseSclk;
			
			if(! pCalib->cal(pCalib, pPkt, uPktSec, &(pRow->sec[uRowSec]) )){
				free(pRow);
				return false;
			}
			
			pList->append(pList, pRow);		
		}
		else{
		
			pRow = pList->get(pList, -1); 
			
			/* This second has been created, either... */		
			if( pRow->nBaseSclk == nBaseSclk ){
				
				
				/* Add to it, or... */
				if( pRow->sec[uRowSec].uVals == 0){
					
					if(!pCalib->cal(pCalib, pPkt, uPktSec, &(pRow->sec[uRowSec])))
						return false;
				}
				
				/* Make a new row for this same second. */
				/* This assumes the next packet will have a time greater than or */
				/* equal to the current one */
				else{
					
					gse_notice_v("Extra 0x%02X packet, splitting row for second: "
					             "0x%08X", pSci->uPSID, nBaseSclk);
					
					pRow = (srv_row_t*)gse_calloc(1, sizeof(srv_row_t));
					memcpy(pRow, &(pThis->tplts[uRowList]), sizeof(srv_row_t));
					pRow->nBaseSclk = nBaseSclk;
					
					if(! pCalib->cal(pCalib, pPkt, uPktSec, &(pRow->sec[uRowSec]) )){
						free(pRow);
						return false;
					}
					pList->append(pList, pRow);
					
				}
			}
			else{
				gse_throw_m(GSE_XCEPT, GSE_XBADARG, "Input data not sorted by time");
			}
		}
	}
	
	pThis->dLastSclk = dSclk;
	
	return true;
}

/*****************************************************************************/
void _SrvRower_adjSclks(SrvRower* pThis, SrvRowList* pOut)
{
	srv_row_t* pRow = NULL;
	size_t iSec = 0;
	
	for(pRow = pOut->get(pOut, 0); pRow != NULL; pRow = pOut->next(pOut)){
		
		pRow->dBaseSclk = (float8) pRow->nBaseSclk;
		
		if(! pThis->pAdj->adjust(pThis->pAdj, &(pRow->dBaseSclk)))
			gse_throw(GSE_XCEPT, GSE_XNOKEY);
		
		/* Get the base sclk as an scet */
		jno_sclk2str(pRow->dBaseSclk, pRow->sBaseScet, 32);
		
		for(iSec=0; iSec<pRow->nSections; iSec++){
			
			if(pRow->sec[iSec].uVals == 0)
				continue;
			
			if(! pThis->pAdj->adjust(pThis->pAdj, &(pRow->sec[iSec].dColSclk) ))
				gse_throw(GSE_XCEPT, GSE_XNOKEY);
			
			pRow->sec[iSec].dSclkOff = pRow->sec[iSec].dColSclk - pRow->dBaseSclk;
		}
	}
}

/*****************************************************************************/
void _SrvRower_prnRows(SrvRower* pThis, SrvRowList* pOut)
{
	srv_row_t* pRow = NULL;
	size_t iSec = 0;
	
	for(pRow = pOut->get(pOut, 0); pRow != NULL; pRow = pOut->next(pOut)){
		printf("%4s | %s, 0x%08X:", pRow->sRowType, pRow->sBaseScet,
		       pRow->nBaseSclk);
		
		for(iSec = 0; iSec < pRow->nSections; iSec++){
			
			if(pRow->sec[iSec].uVals == 0)
				fputs("   -------- ------", stdout);
			else
				printf("   %7s %6.3f", pRow->sec[iSec].sBinPrefix,
				       pRow->sec[iSec].dSclkOff);
		}
		
		putchar('\n');
	}	
}

/*****************************************************************************/
SrvRowList* _SrvRower_flushRows(SrvRower* pThis){
	
	size_t iList = 0;
	SrvRowList* pList = NULL;
	SrvRowList* pOut = NULL;
	
	for(iList = 0; iList < SRVROWER_NLISTS; iList++){
	
		pList = pThis->pLists[iList];
		
		while(pList->length(pList) > 0){
			
			if(pOut == NULL)
				pOut = new_SrvRowList();
			
			pOut->append(pOut, pList->pop(pList, 0));
		}
	}
	
	if(pOut == NULL) return NULL;
		
	/* Adjust all outbound sclks based on the spice kernels */
	_SrvRower_adjSclks(pThis, pOut);
		
		
	/* As a diagnostic, print the rows going out the door */
	if(pThis->bDiag)
		_SrvRower_prnRows(pThis, pOut);
	
	return pOut;
}

/*****************************************************************************/
SrvRowList* SrvRower_getRows(SrvRower* pThis)
{
	SrvRowList* pOut = NULL;
	
	size_t iList   = 0;
	SrvRowList* pList = NULL;
	int4 nLastSec  = -1;
	
	size_t nAdded   = 0;
	srv_row_t*  pRow = NULL;
	bool bCanMerge   = false;
	srv_row_t*  pNeg3 = NULL;
	srv_row_t*  pNeg2 = NULL;
	srv_row_t*  pNeg1 = NULL;
	
	size_t iSec = 0;
	
	/* The logic for flushing is completly different */
	if(pThis->bFlush)
		return _SrvRower_flushRows(pThis);
	
	pOut = new_SrvRowList();
	
	for(iList = 0; iList < SRVROWER_NLISTS; iList++){
		pList = pThis->pLists[iList];
		
		if(pList->length(pList) < 3)
			continue;
	
		nLastSec = (pList->get(pList, -1))->nBaseSclk;
		
		/* Pop all of the items out of this list except the last two, stop if
		   you hit a sclk that is closer than 2 seconds from the end */
		nAdded = 0;
		while(pList->length(pList) > 2){
			
			pRow = pList->get(pList, 0);
			
			if(pRow->nBaseSclk <= nLastSec - 2){
				pOut->append(pOut, pList->pop(pList, 0));
				nAdded +=1;
			}
			else{
				break;
			}
		}
		
		if(nAdded == 0)
			continue;
		
		
		/* Check for a second from the last pull down.  Requirements: */
		/* Distance from -1 to -2 >= 2 and distance from -2 to -3 is < 2 */
		bCanMerge = false;
		pNeg1 = pList->get(pList, -1);
		pNeg2 = pList->get(pList, -2);
		pNeg3 = pOut->get(pOut, -1);
		
		if((pNeg1->nBaseSclk - pNeg2->nBaseSclk >= 2)&&
			(pNeg2->nBaseSclk - pNeg3->nBaseSclk) < 2){
		
			/* So the merge has an obvious dir closer to the neg3 item now to
			   see if there are holes for the merge: */
			bCanMerge = true;
			for(iSec = 0; iSec < pNeg3->nSections; iSec++){
								
				if((pNeg2->sec[iSec].uVals > 0)&&(pNeg3->sec[iSec].uVals > 0)){
					bCanMerge = false;
					break;
				}
			}
			
			if(bCanMerge){
				/* Pull data in from row -2 which is still in the main list to
				   this out going row */
				for(iSec = 0; iSec < pNeg3->nSections; iSec++){
					
					if( pNeg3->sec[iSec].uVals == 0)
						memcpy(&(pNeg3->sec[iSec]), &(pNeg2->sec[iSec]),
								 sizeof(srv_row_sec_t));						
				}
				
				pNeg2 = pList->pop(pList, -2);
				free(pNeg2);
			}
		}
	}
	
	if(pOut->length(pOut) == 0){
		pOut->del(pOut);
		return NULL;
	}
	
	/* Adjust all outbound sclks based on the spice kernels */
	_SrvRower_adjSclks(pThis, pOut);
		
		
	/* As a diagnostic, print the rows going out the door */
	if(pThis->bDiag)
		_SrvRower_prnRows(pThis, pOut);
	
	return pOut;
}


/*****************************************************************************/
/* Destructor */
void delete_SrvRower(void* vpThis)
{
	SrvRower* pThis = (SrvRower*)vpThis;
	size_t i;
	
	for(i=0;i<SRVROWER_NLISTS;i++)
		if(pThis->pLists[i] != NULL)
			pThis->pLists[i]->del(pThis->pLists[i]);
	
	free(pThis);
}


/*****************************************************************************/
/* Constructor, reads in the cal files (could break off reading into something
   else to allow for internal tables, but this is not needed right now */

void _SrvRower_map(SrvRower* pThis, size_t iMap, byte uPsid, byte uSec,
                   byte uBand, uint4 uListNum, uint4 uRowSec){
	
	pThis->map[iMap].uPsid    = uPsid;
	pThis->map[iMap].uSec     = uSec;
	pThis->map[iMap].uBand    = uBand;
	pThis->map[iMap].uListNum = uListNum;
	pThis->map[iMap].uRowSec  = uRowSec;
	
	/* Also, init this section of the indicated row */
	memcpy( &(pThis->tplts[uListNum].sec[uRowSec]), 
	        pThis->pCalib->getTplt(pThis->pCalib, uPsid, uSec, uBand),
			  sizeof(srv_row_sec_t));
}

SrvRower* new_SrvRower(HskFineAdj* pAdj, const SrvCalib* pCalib)
{
	size_t i;
	SrvRower* pThis = NULL;
	
	pThis = (SrvRower*)gse_calloc(1, sizeof(SrvRower));
	pThis->addPkt   = SrvRower_addPkt;
	pThis->setFlush = SrvRower_setFlush;
	pThis->getRows  = SrvRower_getRows;
	pThis->inPktSecs = SrvRower_inPktSecs;
	pThis->del      = delete_SrvRower;
	except_cleanup_push(delete_SrvRower, pThis);
	
	pThis->pAdj = pAdj;
	pThis->pCalib = pCalib;
	
	pThis->bFlush = false;
	pThis->dLastSclk = -1.0;
	
	for(i = 0; i<SRVROWER_NLISTS; i++)
		pThis->pLists[i] = new_SrvRowList();
	
	/* Init the row templates:  calloc sets all fields to zero, so we only 
	   need to set non-NULL values */
	
	/* Primary Survey mode product, all normal E-fields */
	pThis->tplts[0].sDesc = "Juno Waves Electric Field Spectral Densities in "
	                     "(Volts^2/meter^2)/Hertz";
	pThis->tplts[0].sRowType = "E";
	pThis->tplts[0].nSections = 4;
	
	
	/* Primary Survey mode product, the B-Fields */
	pThis->tplts[1].sDesc = "Juno Waves Magnetic Field Spectral Densities in "
	                     "(nanoTesla/Hertz)";
	pThis->tplts[1].sRowType = "B";
	pThis->tplts[1].nSections = 1;
	
	/* Noise data survey product, all noise E-fields */
	pThis->tplts[2].sDesc = "Juno Waves Power Supply Noise Power Spectral "
	                      "Densities in (Volts^2)/Hertz";
	pThis->tplts[2].sRowType = "EINT";
	pThis->tplts[2].nSections = 2;
	
	/* Noise data survey product, all noise B-Fields */
	pThis->tplts[3].sDesc = "Juno Waves Supply Noise Power Spectral Densities "
	                      "in (Volts^2)/Hertz";
	pThis->tplts[3].sRowType = "BINT";
	pThis->tplts[3].nSections = 1;
	
	/* Alt data E data surve product, usually the S+N data */
	pThis->tplts[4].sDesc = "Juno Waves Un-corrected Electric Spectral "
	                        "Densities in (Volts^2)/Hertz";
	pThis->tplts[4].sRowType = "EUNC";
	pThis->tplts[4].nSections = 2;
	
	/* Primary Survey mode product, the B-Fields */
	pThis->tplts[5].sDesc = "Juno Waves Un-corrected Magnetic Field Spectral "
	                        "Densities in (nanoTesla/Hertz)";
	pThis->tplts[5].sRowType = "BUNC";
	pThis->tplts[5].nSections = 1;
	
	
	/* Now, map packet types to sections with in rows */
	/*                      PSID, SEC,       BAND,  LIST,     ROW_SEC */
	
	
	/* The new lists, after the 2011-10 FSW change to stuff NRD enable into
	   the band byte, not that the 91, 92 and 93 packets split into two
		settings */
	
	_SrvRower_map(pThis,  0, 0x91,  0 /* S+N*/,  0xF0, 1 /* "B"    */,  0);
	_SrvRower_map(pThis,  1, 0x91,  0 /* S  */,  0xF1, 1 /* "B"    */,  0);
	_SrvRower_map(pThis,  2, 0xF0,  0 /* S  */,  0xF3, 1 /* "B"    */,  0);
	_SrvRower_map(pThis,  3, 0xF0,  1 /* S+N*/,  0xF3, 5 /* "BUNC" */,  0);
	_SrvRower_map(pThis,  4, 0xF0,  2 /* N  */,  0xF3, 3 /* "BINT" */,  0);
	
	/* Electric LFR Lo E packet id to row list map */
	_SrvRower_map(pThis,  5, 0x92,  0 /* S+N*/,  0xF0, 0 /* "E"    */,  0);
	_SrvRower_map(pThis,  6, 0x92,  0 /* S  */,  0xF1, 0 /* "E"    */,  0);
	_SrvRower_map(pThis,  7, 0xF4,  0 /* S  */,  0xF3, 0 /* "E"    */,  0);
	_SrvRower_map(pThis,  8, 0xF4,  1 /* S+N*/,  0xF3, 4 /* "EUNC" */,  0);
	_SrvRower_map(pThis,  9, 0xF4,  2 /* N  */,  0xF3, 2 /* "EINT" */,  0);
	
	/* Electric LFR Hi E packet to row list map */
	_SrvRower_map(pThis, 10, 0x93,  0 /* S+N*/,  0xF0, 0 /* "E"    */,  1);
	_SrvRower_map(pThis, 11, 0x93,  0 /* S  */,  0xF1, 0 /* "E"    */,  1);
	_SrvRower_map(pThis, 12, 0xF5,  0 /* S  */,  0xF3, 0 /* "E"    */,  1);
	_SrvRower_map(pThis, 13, 0xF5,  1 /* S+N*/,  0xF3, 4 /* "EUNC" */,  1);
	_SrvRower_map(pThis, 14, 0xF5,  2 /* N  */,  0xF3, 2 /* "EINT" */,  1);
	
	/* Electric HFR Lo packet to row list map */
	_SrvRower_map(pThis, 15, 0x5F,  0 /* S  */,  0xF0, 0 /* "E"    */,  2);
	_SrvRower_map(pThis, 16, 0x5B,  0 /* S  */,  0xF0, 0 /* "E"    */,  2);
	_SrvRower_map(pThis, 17, 0x3B,  0 /* S  */,  0xF0, 0 /* "E"    */,  2);
	
	
	/* The old lists Pre 2011-10 FSW change, band byte always indicated
	   0 and NR was always assumed to be on for packet types 0x90 */
	
	/* Magnetic field packet id to row list map */
	_SrvRower_map(pThis, 18, 0x91,  0 /* S  */,  0, 1 /* "B"    */,  0);
	_SrvRower_map(pThis, 19, 0xF0,  0 /* S  */,  0, 1 /* "B"    */,  0);
	_SrvRower_map(pThis, 20, 0xF0,  1 /* S+N*/,  0, 5 /* "BUNC" */,  0);
	_SrvRower_map(pThis, 21, 0xF0,  2 /* N  */,  0, 3 /* "BINT" */,  0);
	
	/* Electric LFR Lo E packet id to row list map */
	_SrvRower_map(pThis, 22, 0x92,  0 /* S  */,  0, 0 /* "E"    */,  0);
	_SrvRower_map(pThis, 23, 0xF4,  0 /* S  */,  0, 0 /* "E"    */,  0);
	_SrvRower_map(pThis, 24, 0xF4,  1 /* S+N*/,  0, 4 /* "EUNC" */,  0);
	_SrvRower_map(pThis, 25, 0xF4,  2 /* N  */,  0, 2 /* "EINT" */,  0);
	
	/* Electric LFR Hi E packet to row list map */
	_SrvRower_map(pThis, 26, 0x93,  0 /* S  */,  0, 0 /* "E"    */,  1);
	_SrvRower_map(pThis, 27, 0xF5,  0 /* S  */,  0, 0 /* "E"    */,  1);
	_SrvRower_map(pThis, 28, 0xF5,  1 /* S+N*/,  0, 4 /* "EUNC" */,  1);
	_SrvRower_map(pThis, 29, 0xF5,  2 /* N  */,  0, 2 /* "EINT" */,  1);
	
	
	
	/* Electric HFR Lo packet to row list map */
	_SrvRower_map(pThis, 30, 0x5F,  0 /* S  */,  0, 0 /* "E"    */,  2);
	_SrvRower_map(pThis, 31, 0x5B,  0 /* S  */,  0, 0 /* "E"    */,  2);
	_SrvRower_map(pThis, 32, 0x3B,  0 /* S  */,  0, 0 /* "E"    */,  2);
	
	/* Electric HFR Hi packet to row list map */
	_SrvRower_map(pThis, 33, 0x47,  0 /* S  */,  0x01, 0 /* "E"    */,  3);
	_SrvRower_map(pThis, 34, 0x46,  0 /* S  */,  0x01, 0 /* "E"    */,  3);
	
	
	except_cleanup_pop(except_no_call);
	return pThis;
}

/*****************************************************************************/
/* SrvRow structures are free'able (no del member) */

#define C SrvRowList
#define I srv_row_t
#define CASCADE_FREE

#include <rpwgse/listTplt.c>

#undef CASCADE_FREE
#undef I
#undef C

/*****************************************************************************/
/* srv_row_t structures are free'able */

/* #define S SrvRowAry
#define T srv_row_t
#define CASCADE_FREE

#include <rpwgse/ptrAryTplt.c>

#undef CASCADE_FREE
#undef T
#undef S
*/
