/** @file SvrRower.h */
#ifndef _SrvRower_H_
#define _SrvRower_H_

#include <rpwgse/GseTypes.h>

#include "SrvCalib.h"
#include "HskFineAdj.h"

#ifdef __cplusplus
extern "C" {
#endif


/*****************************************************************************/
/* A single Row of SRV data, these are generated by the SrvRower */

typedef struct srv_row{
	
	/* Properties for this row type */
	const char* sDesc;    /* Ex: "Juno Waves Electric Field Survey	Data from
	                             18.9kHz to 150kHz" */
	
	/* Used to keep the output file names straight */
	const char* sRowType; /* "E", "B", "EN", "BN", "EALT", "BALT" */
	
	size_t nSections;     /*  4,   1,   2,    1,    2,      1 */
	
	/* Timing for this particular row */
	int4   nBaseSclk;   /* Acts as the row identifier, used during row build */
	float8 dBaseSclk;   /* The fine time and spice adj. base row time */ 
	char sBaseScet[32];  /* This is in YYYY-DDDTHH:MM:SS.SSS format */
	
	/* Data for this particular row, not all data sections are filled */
	srv_row_sec_t sec[4];

}srv_row_t;

/*****************************************************************************/
/* List of row objects */

#define S srv_row_list
#define C SrvRowList
#define I srv_row_t

#include <rpwgse/listTplt.h>

#undef I
#undef C
#undef S

/*****************************************************************************/
/* An array of row objects, used in the return of getRows */

/* #define S srv_row_array
#define C SrvRowAry
#define T srv_row_t

#include <rpwgse/ptrAryTplt.h>

#undef T
#undef C
#undef S
*/

/*****************************************************************************/
/* Mapping packet types to rows */

typedef struct srv_pkt2list_map_item{
	byte uPsid;        /* the packet type */
	byte uSec;         /* The packet data section number */
	byte uBand;        /* Indicator of the noise reduction being on or off */
	
	uint4 uListNum;    /* Which row list does this packet go in */
	uint4 uRowSec;     /* Which row section does this packet type map to */
}srv_pkt2list_map_item_t;

/*****************************************************************************/
/* Handle Rowing Up LRS measurements */

#define SRVROWER_NLISTS   6
#define SRVROWER_PKTSECS 11
#define SRVROWER_NMAPS   35


/* NOTE:  if you don't intermix calls to getRows with calls to addPkt the
   row collapse doesn't get triggered, this could be remmided if need be */
typedef struct srv_rower{

	/** Add a packet to the rower.
	 * @return If true, the packet was translated in to a row section. */
	bool (*addPkt)(struct srv_rower* pThis, const GsePkt* pPkt);
		
	/** Get 0-N rows of calibrated data. Returns 0 if no new rows are
	 *  currently ready, if the return > 0, the caller is the new
	 *  owner of the rows and must dispose of thier memory as appropriate.
	 *  Due to the nature of waves data, adding a new packet can trigger
	 *  the creation of up to 3 rows.
	 *
	 * @param pThis the rower
	 *
	 * @returns an array of pointers to SrvCalRow objects allocated on the
	 *          heap, or NULL if no rows are ready.  
	 */
	SrvRowList* (*getRows)(struct srv_rower* pThis);
	
	/** Set's the rower to flush mode, which will cause all half full rows 
	 * that may be waiting for new packets to be calibrated and output.  
	 * Any future calls to addPkt will throw an exception. */
	void (*setFlush)(struct srv_rower* pThis);
	
	/** Get the number of science packet sections that have been handed to
	    the rower */
	int4 (*inPktSecs)(const struct srv_rower* pThis);

	/** Destructor */
	void (*del)(void* vpThis);
	
	
	const SrvCalib* pCalib;  /* Packet section to row section maker */
	
	HskFineAdj* pAdj;        /* Sclk Adjuster */
	
	bool bFlush;             /* If on flush incomplete rows */
	
	bool bDiag;              /* Turn on diagnostics mode */
	float8 dLastSclk;        /* Used to make sure input data is in order */
	size_t nSecIn;          /* Keep track of number of input packet sections */
	
	SrvRowList* pLists[SRVROWER_NLISTS]; /* Array of Lists of Undelivered rows */
	srv_row_t tplts[SRVROWER_NLISTS];  /* Array of new row templates */
	
	/* Map to figure out where packets go and how to init new row members */
	srv_pkt2list_map_item_t map[SRVROWER_NMAPS];

} SrvRower;


SrvRower* new_SrvRower(HskFineAdj* pAdj, const SrvCalib* pCalib);


#ifdef __cplusplus
}
#endif

#endif /* _SrvRower_H_ */
