#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseLog.h>
#include <rpwgse/GseFile.h>
#include <rpwgse/GseWrap.h>
#include <rpwgse/GseConv.h>


/****************************************************************************/
/* Compiled in config file directory location.*/
#ifndef CFG
#error Compiled in default configuration directory is missing.
#endif

#define _QDEF(x) #x
#define QDEF(x) _QDEF(x)

/*****************************************************************************/
/* callback to prune non cal files from a list */

bool rm_non_cal(void* pCtx, const char* sPathItem)
{
	char sBasename[128];
	GseCStrAry* pCalPrefix = (GseCStrAry*)pCtx;
	size_t i, nLen;
	bool bKeep = false;
	
	/* If can't get the basename this entry is no good */
	if( ! gse_basename(sPathItem, sBasename, 128) ) return true;
	
	for(i = 0; i< pCalPrefix->num; i++){
	
		nLen = strlen(pCalPrefix->pAry[i]);
		
		if( strncmp(sBasename, pCalPrefix->pAry[i], nLen) == 0){
			
			nLen = strlen(sBasename);
			if( sBasename[nLen - 4] == '.' && sBasename[nLen - 3] == 'C' &&
				 sBasename[nLen - 2] == 'S' && sBasename[nLen - 1] == 'V' )
				bKeep = true;
				break;
		}
	}
	
	if( ! bKeep ) return true;
	
	return false;
}

/*****************************************************************************/
/* General function, get a cal file's version number */

int4 get_calfile_version(const char* sPath){
	
	char sFile[128]; memset(sFile, 0, 128);
	int4 nVer = -1;
	size_t nLen, nNewLen;
	
	if(sPath == NULL || (strlen(sPath) < 23) )
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sPathIn too short");
	
	if(strlen(sPath) > 1023)
		gse_throw_m(GSE_XCEPT, GSE_XBADARG, "sPathIn too long");
	
	if(!gse_basename(sPath, sFile, 128))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of '%s'",
		             sPath);
	
	/* Check for ending in .CSV, then null at decimal point */
	nLen = strlen(sFile);
	if( sFile[nLen - 4] != '.' || sFile[nLen - 3] != 'C' ||
		 sFile[nLen - 2] != 'S' || sFile[nLen - 1] != 'V' ){
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s: Filename doesn't end in .CSV", sFile);
	}
	sFile[nLen - 4] = '\0';
	nNewLen = strlen(sFile);
	
	if((!gse_str2int4(sFile + nNewLen - 2, &nVer))||(nVer < 0))
		gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "%s.CSV: Couldn't get version number", sFile);
	
	sFile[nLen - 4] = '.';
	
	return nVer;
}

/*****************************************************************************/
/* Initialize the volume, if need be, return the names of the calibration */
/* to use on the volume or NULL if there is an issue. */

GseStrAry* init_calib_dir(const char* sGroup, GseCStrAry* pPrefix)
{
	char* sGroupDir = NULL;
	char* sVolRoot = NULL; 
	char* sDatRoot = NULL;
	char* sCalRoot = NULL;
	char* sCalCfg  = NULL;
	char sCfgBase[128];
	char sCalibBase[128];
	
	const char* sPrefix = NULL;
	const char* sCalFile = NULL;
	const char* psHighest[14] = {0}; /* using a c trick to null the array */
	
	GseStrAry* pCalPaths = new_GseStrAry(pPrefix->num);
	
	int4 iCal = 0;
	size_t nLen = 0;
	
	GseStrList* pCfgList = NULL;
	GseStrList* pCalibList = NULL;
	
	/* The group directory should already exist */
	sGroupDir = NS("%s/%s", getenv("WAVES_DATA"), sGroup);
	if(!gse_isdir(sGroupDir))
		gse_throw_fm(GSE_XCEPT, GSE_XMISSINGDIR, "Data group directory %s "
		             "doesn't exist.", sGroupDir);
	
	sVolRoot = NS("%s/%s/JNOWAV_1000", getenv("WAVES_DATA"), sGroup);
	if(!gse_isdir(sVolRoot)){
		gse_mkdirs(sVolRoot, 0775);
		gse_notice_v("Creating Volume Directory: %s", sVolRoot);
	}
	
	sCalRoot = NS("%s/CALIB", sVolRoot);
	if(!gse_isdir(sCalRoot)){
		gse_mkdirs(sCalRoot, 0775);
		gse_notice_v("Creating CALIB Directory: %s", sCalRoot);
	}
	
	/* Now copy in the calibration files if need be */
	sCalCfg = CS( QDEF(CFG) "/pds" );
	if(!gse_isdir(sCalCfg))
		gse_throw_fm(GSE_XCEPT, GSE_XMISSINGDIR, "Config directory %s doesn't "
		             " exist", sCalCfg);
	
	/* Get the contents of the CFG dir */
	if( (pCfgList = gse_listdir( sCalCfg, 1)) == NULL)
		gse_throw_fm(GSE_XCEPT, GSE_XMISSINGFILE, "Config directory %s is empty!",
		             sCalCfg);
	
	/* Prune off items that don't start with one of the preficies above or
	   doesn't end in .CSV */
	pCfgList->rmWhere(pCfgList, rm_non_cal, pPrefix);
	
	/* Okay, there had better be at least num config files left */
	if(pCfgList->length(pCfgList) < pCalPaths->num){
		gse_error_v("Config directory %s is missing 1 or more calibration "
		            "tables, the following cal files were present:",sCalCfg);
		if(pCfgList->length(pCfgList) == 0)
			fprintf(stderr, "        (NONE)\n");
		else
			for(iCal=0; iCal< pCfgList->length(pCfgList); iCal++)
				fprintf(stderr, "        %s\n", pCfgList->get(pCfgList, iCal));
		
		/* Going to exit out don't worry about the destructor */
		exit(113);
		/* gse_throw(GSE_XCEPT, GSE_XMISSINGFILE); */
	}
	
	/* Find the highest version of each calibration file in the config
	   directory */
	for(sCalFile = pCfgList->get(pCfgList,0); sCalFile != NULL; 
		 sCalFile = pCfgList->next(pCfgList) ){
	
		if(! gse_basename(sCalFile, sCfgBase, 128) ) continue;
		
		for(iCal = 0; iCal < pCalPaths->num; iCal++){
			sPrefix = pPrefix->pAry[iCal];
			if(strncmp(sCfgBase, sPrefix, strlen(sPrefix)) == 0){
		
				if(psHighest[iCal] == NULL){
					psHighest[iCal] = sCalFile;
				}
				else{
					if(strcmp(sCalFile, psHighest[iCal]) > 0)
						psHighest[iCal] = sCalFile;
				}
				
				break;
			}
		}
	}
	
	/* Check again to see if I have one of each config table */
	for(iCal = 0; iCal < pCalPaths->num; iCal++){
		if(psHighest[iCal] == NULL)
			gse_throw_fm(GSE_XCEPT, GSE_XMISSINGFILE, "%s: Needs a calibration "
			             "file with the pattern %s??.CSV", sCalCfg,
			             pPrefix->pAry[iCal]);
	}
	
	/* Okay now compare the highest config files to the ones in the
	   PDS dir, for each one that doesn't exist or is a lower version
		number, replace it */
	
	if( (pCalibList = gse_listdir(sCalRoot, 1)) != NULL)
		pCalibList->rmWhere(pCalibList, rm_non_cal, pPrefix);
	
	
	/* For each calib file from the config, see if we need to add, update
	   or do nothing to the config file in the volume */
	for(iCal = 0; iCal < pCalPaths->num; iCal++){
		sPrefix = pPrefix->pAry[iCal];
		
		char* sFoundOne = NULL;
		if(pCalibList != NULL){
			for( sCalFile = pCalibList->get(pCalibList,0); sCalFile != NULL;
		     sCalFile = pCalibList->next(pCalibList)){
			
				if( ! gse_basename(sCalFile, sCalibBase, 128)) continue;
				
				nLen = strlen(sCalibBase);
				
				if((strncmp(sCalibBase, sPrefix, strlen(sPrefix)) == 0)&&
					(strcmp(".CSV", sCalibBase + nLen - 4) == 0) ){
					
					if(sFoundOne != NULL){
						gse_throw_fm(GSE_XCEPT, GSE_XNOTUNIQUE, "More than version"
						             "of the same calibration file found in %s",
						             sCalRoot);
					}
					
					sFoundOne = (char*)sCalFile;
					break;
				}
			}
		}
		
		if(sFoundOne == NULL){
			gse_notice_v("Copying %s to %s", psHighest[iCal], sCalRoot);
			gse_copyfile_p(psHighest[iCal], sCalRoot, GSE_MODE_664);
			
			/* Save off the new calibration file name */
			if( ! gse_basename(psHighest[iCal], sCalibBase, 128))
				gse_throw(GSE_XCEPT,GSE_XWTF);
			pCalPaths->pAry[iCal] = NS("%s/%s", sCalRoot, sCalibBase);
		}
		else{
			
			if( ! gse_basename(sFoundOne, sCalibBase, 128) )
				gse_throw(GSE_XCEPT, GSE_XWTF);
			if( ! gse_basename(psHighest[iCal], sCfgBase, 128) )
				gse_throw(GSE_XCEPT, GSE_XWTF);
			
			if(strcmp(sCfgBase, sCalibBase) > 0){
				gse_notice_v("Removing %s", sFoundOne);
				gse_remove(sFoundOne);
				nLen = strlen(sFoundOne);
				
				sFoundOne[nLen - 4] = '.';
				sFoundOne[nLen - 3] = 'L';
				sFoundOne[nLen - 2] = 'B';
				sFoundOne[nLen - 1] = 'L';
				gse_notice_v("Removing %s", sFoundOne);
				gse_remove(sFoundOne);
				
				gse_notice_v("Copying %s to %s", psHighest[iCal], sCalRoot);
				gse_copyfile_p(psHighest[iCal], sCalRoot, GSE_MODE_664);
				
				/* Save off the new calibration file name */
				if( ! gse_basename(psHighest[iCal], sCalibBase, 128))
					gse_throw(GSE_XCEPT,GSE_XWTF);			
				pCalPaths->pAry[iCal] = NS("%s/%s", sCalRoot, sCalibBase);
				
			}
			else{
				gse_debug_v("%s: Retained", sFoundOne);
				
				/* Save off the existing calibration file name */
				if( ! gse_basename(sFoundOne, sCalibBase, 128))
					gse_throw(GSE_XCEPT,GSE_XWTF);			
				pCalPaths->pAry[iCal] = NS("%s/%s", sCalRoot, sCalibBase);
			}
		}
	}
	
	if(pCalibList != NULL) pCalibList->del(pCalibList);
	if(pCfgList != NULL) pCfgList->del(pCfgList);
	
	free(sGroupDir);
	free(sVolRoot);
	free(sDatRoot);
	free(sCalRoot);
	free(sCalCfg);
	
	return pCalPaths;
}
