#define _POSIX_C_SOURCE 200112

#ifdef GSE_THREAD_SAFE
#include <pthread.h>
#endif

#include <math.h>
#include <string.h>
#include <SpiceUsr.h>

#include <rpwgse/GseExcept.h>
#include <rpwgse/GseConv.h>

#include <jwaves/JWavesDefs.h>

#include "jno_spice.h"

/*****************************************************************************/
/* Excerpt from SpiceZdf.h on types:                                         */
/*                                                                           */
/* The SPICE typedefs map in an arguably natural way to ANSI C types:        */
/*                                                                           */
/*    SpiceBoolean -> int                                                    */
/*    SpiceChar    -> char                                                   */
/*    SpiceDouble  -> double                                                 */
/*    SpiceInt     -> int or long                                            */
/*    ConstX       -> const X  (X = any of the above types)                  */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Bounce spice errors to an exception, if failed_c indicates a problem */

void jno_spice_check(){
	int iNext;
	char sBuff[1841];
	
	if( failed_c() ) {
		getmsg_c("SHORT", 41, sBuff);
		iNext = strlen(sBuff);
		sBuff[iNext] = ' ';
		iNext++;
		getmsg_c("LONG", 1841-iNext , sBuff+iNext);
		
		gse_throw_m(SPICE_XCEPT, SPICE_XERR, sBuff);
	}
}

/*****************************************************************************/
/* Initialize Spice, or throw an exception */

void jno_spice_init(const char* sFurnish){

	errprt_c("SET", 0, "ALL");     /* Want all error text */
	erract_c("SET", 0, "RETURN");  /* Just return on error, we'll check for it */
	errdev_c("SET", 0, "NULL");    /* Don't go printing to the screen on error */

	
	furnsh_c(sFurnish);
	jno_spice_check();
}

/*****************************************************************************/
char* jno_sclk2str(float8 dSclk, char* sScet, size_t nLen){
	float8 et;
	int4 nSclk, nRem;
	char sSclk[64];
	
	sSclk[63] = '\0'; /* Insure null termination */
			
	nSclk = (int4) trunc(dSclk);
	nRem = (dSclk - nSclk) * 65536;
	snprintf(sSclk, 63, "%08d:%05d", nSclk, nRem);
	
	scs2e_c(-61999, sSclk, &et);
	
	jno_spice_check();
	timout_c(et, "YYYY-DOYTHR:MN:SC.###", nLen, sScet);
	jno_spice_check();
	
	return sScet;
}

/*****************************************************************************/
void jno_sclk2ct(float8 dSclk, gse_caltime_t* pCt)
{
	char sTime[32];
	
	jno_sclk2str(dSclk, sTime, 32);
	sTime[4] = '\0';
	sTime[8] = '\0';
	sTime[11] = '\0';
	sTime[14] = '\0';
			
	if(! gse_str2int4(sTime, &pCt->year)) goto SCLK2CT_ERR;
	if(! gse_str2int4(sTime+5, &pCt->doy)) goto SCLK2CT_ERR;
	if(! gse_str2int4(sTime+9, &pCt->hour)) goto SCLK2CT_ERR;
	if(! gse_str2int4(sTime+12, &pCt->min)) goto SCLK2CT_ERR;
	if(! gse_str2float8(sTime+15, &pCt->sec)) goto SCLK2CT_ERR;

	gse_ct_norm_doy(pCt);
	return;
	
SCLK2CT_ERR:
	sTime[4] = '-';
	sTime[8] = 'T';
	sTime[11] = ':';
	sTime[14] = ':';
	gse_throw_fm(GSE_XCEPT, GSE_XDECODE, "Couldn't convert %s to a "
	             "gse_caltime_t", sTime);
}


/*****************************************************************************/
float8 jno_ct2sclk(const gse_caltime_t* pCt){
	
	float8 dSclk;
	int4 nRem;
	float8 et;
	char sTime[32];
	
	sprintf(sTime, "%04d-%03dT%02d:%02d:%06.3f", pCt->year,
	        pCt->doy, pCt->hour, pCt->min, pCt->sec);

	utc2et_c(sTime, &et);
	jno_spice_check();
	
	/* sce2c_c(JUNO_SPICE_ID, et, &dSclk); */
	sce2s_c(-61999, et, 31, sTime);  /* Now using sTime as an output */
	jno_spice_check();
	
	if((sTime[12] != '.')||(sTime[1] != '/'))
		gse_throw_fm(SPICE_XCEPT, SPICE_XERR, "Unexpected SCLK format in string "
		             "%s", sTime);
	
	sTime[1] = '\0'; sTime[12] = '\0';
	
	if( gse_str2float8(sTime+2, &dSclk) ){
		if( gse_str2int4(sTime+13, &nRem)){
			if((dSclk > 0)&&(nRem >= 0)){
				dSclk += ((double)nRem)/65536.0;
				return dSclk;
			}
		}
	}
	
	gse_throw_fm(SPICE_XCEPT, SPICE_XERR, "Can't convert %s to a valid floating"
	             " point SCLK.", sTime);
	return 0;
}

/*****************************************************************************/
float8 jno_str2sclk(const char* sScet){
	
	float8 et;
	float8 dSclk;
	
	utc2et_c(sScet, &et);
	jno_spice_check();
	
	sce2c_c(JUNO_SPICE_ID, et, &dSclk);
	dSclk = dSclk / 256.0;
	jno_spice_check();
	return dSclk;
}
