#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <strings.h>

#include <das.h>
#include <das_cli.h>
#include <rpwgse/Gse.h>
#include <jwaves/JWaves.h>

#include "jno_spice.h"
#include "HskFineAdj.h"

#define JNO_FURNISH "/opt/project/juno/spice/JunoSCLKKernels.list"
const char* g_sKernels = JNO_FURNISH;

/*****************************************************************************/
const char* g_sDesc = 
"Reads Juno Level 2 datasets and outputs specta of raw data numbers as\n"
"   a das2 stream.  The datasets 'survey' and 'burst_wbr_hi' are are already\n"
"   in spectral form all others are converted via a hann window and a fourier\n"
"   transform.  Resulting values are *not* divided by the frequency bandwidth.\n";

const char* g_sFooter = 
"See Also:\n"
"   waves_pds_l2wfrmrdr\n"
"\n"
"Maintainer:\n"
"   Chris Piker <chris-piker@uiowa.edu>\n\n";

/*****************************************************************************/
void setLogLevel(const char* sLevel){
	if(strlen(sLevel) < 1)
		gse_throw(GSE_XCEPT, GSE_XBADARG);
	
	switch(sLevel[0]){
		case 'c': gse_log_setlevel(GSE_CRITICAL); return;
		case 'e': gse_log_setlevel(GSE_ERROR); return;
		case 'w': gse_log_setlevel(GSE_WARNING); return;
		case 'd': gse_log_setlevel(GSE_DEBUG); return;
	}
	gse_log_setlevel(GSE_NOTICE);
}

/*****************************************************************************/
/* Structure to keep track of what channels go with which packet types */

typedef struct chan_2_das_map {

	int nDasPktId;
	const char* sDataset;
	bool bSentHdr;
	size_t uYitems;
	const char** pChannels;
	
	float rLastStartFreq;  /* Only used by HFR Hi */
}chan_2_das_map_t;


/*****************************************************************************/
/* Function to see if a particular channel should be ignored */

chan_2_das_map_t* getChanMap(const char* sDataSet, chan_2_das_map_t* pMaps,
		                       GseXYdata* pData)
{

	/* Look over all maps looking to see if the channel ID in the returned 
	   dataset is one of the ones I should output */
	chan_2_das_map_t* pMap = pMaps;
	const char* pChannel = NULL;
	while(pMap->nDasPktId != 0){
		
		/* Map not used in this dataset */
		if(strcmp(pMap->sDataset, sDataSet) == 0){
		
			for(int i = 0; pMap->pChannels[i] != NULL; i++){
				pChannel = pMap->pChannels[i];
				if( strcasecmp(pChannel, pData->channel(pData)) == 0)
					return pMap;
			}
		}	
		pMap++;
	}

	return NULL;
}

/*****************************************************************************/
/* (Re)send a packet header */

void sendHeader(chan_2_das_map_t* pMap, GseXYdata* pData, GseFileStr* pBuf,
                bool bAscii)
{
	pMap->uYitems = pData->length(pData);
	
	const char* sValueFmt; 
	if(bAscii)
		sValueFmt = "ascii13";
	else
		sValueFmt = "little_endian_real4";
	
	pBuf->clear(pBuf);
	pBuf->putstr(pBuf, "<packet>\r\n");
	pBuf->putstr(pBuf, 
"  <properties title=\"Juno Waves Raw Survey\" xLabel=\"SCET %{RANGE}\""
" yLabel=\"Frequncy (kHz)\" zLabel=\"Raw Value\" />\r\n");
	pBuf->putstr(pBuf, "  <x type=\"time25\" units=\"t2000\"></x>\r\n");
	pBuf->printf(pBuf,
"  <yscan type=\"%s\" nitems=\"%zu\" yUnits=\"%s\"\r\n",
                sValueFmt, pData->length(pData), pData->Xunit(pData));
	pBuf->putstr(pBuf, "    yTags=\"");
	
	const float4* pX = pData->Xdata(pData);
	size_t uLen = pData->length(pData);
	for(size_t u = 0; u < uLen; u++){
		if(u != 0) pBuf->putstr(pBuf, ", ");
		pBuf->printf(pBuf, "%.4e", pX[u]);
	}
	pBuf->putstr(pBuf, "\">\r\n  </yscan>\r\n");
	pBuf->putstr(pBuf, "</packet>\r\n");
	printf("[%02d]%06zd%s", pMap->nDasPktId, pBuf->strlen(pBuf), pBuf->getstr(pBuf));
			
	pMap->bSentHdr = true;
}

/*****************************************************************************/
/* Send a data packet */
bool sendData(chan_2_das_map_t* pMap, float8 rSclk, GseXYdata* pData, 
		        GseFileStr* pBuf, bool bAscii)
{
	
	/* Convert the sclk to a scet */
	char sScet[32] = {'\0'};
	jno_sclk2str(rSclk, sScet, 31);
	
	printf(":%02d:%23s  ", pMap->nDasPktId, sScet);
	
	const float* pY = pData->Ydata(pData);
	size_t uLen = pData->length(pData);
	byte u1Out[4] = {0};
	
	/* Make sure we write little endian on either type of hardware */
	for(size_t u = 0; u < uLen; u++){
		
		if(bAscii){
			if(u < (uLen - 1))
				printf("%11.4e  ", pY[u]);
			else
				printf("%11.4e\r\n", pY[u]);
		}
		else{
			gse_f4ToLSBbytes(u1Out, pY[u]);
			fwrite(u1Out, sizeof(byte), 4, stdout);
		}
	}
	
	return true;
}
		
/*****************************************************************************/
int main(int nArgs, char** sArgs){
	
	/* Make the map of channels to packet ID's */
	chan_2_das_map_t maps[] = {
	
	{1, "survey_b", false, 0,  (const char*[]){"LFR B spec", "LFR B spec NRD S", NULL}, 0.0f},
			
	{2, "survey_e", false, 0,  (const char*[]){"LFR Lo E spec", "LFR Lo E spec NRD S", NULL}, 0.0f},
	{3, "survey_e", false, 0,  (const char*[]){"LFR Hi E spec", "LFR Hi E spec NRD S", NULL}, 0.0f},
	{4, "survey_e", false, 0,  (const char*[]){"HFR base", "HFR base fallback", NULL}, 0.0f},
	{5, "survey_e", false, 0,  (const char*[]){"HFR", "HFR fallback", NULL}, 0.0f},
	
	{6, "burst_b",  false, 0, (const char*[]){"LFR B", "LFR B NRD S", NULL}, 0.0f},

	{7, "burst_e", false, 0, (const char*[]){"LFR Lo E", "LFR Lo E NRD S", NULL}, 0.0f},

	{8, "burst_e", false, 0, (const char*[]){"LFR Hi E", "LFR Hi E NRD S", NULL}, 0.0f},
	
	{9, "burst_e", false, 0, (const char*[]){"HFWBR base", "HFWBR base fallback", 
	                             "HFWBR base dual", "HFWBR base dual fallback", NULL}, 0.0f},
	{10,"burst_e", false, 0, (const char*[]){"HFWBR band spec","HFR band spec",NULL}, 0.0f},
	
	{0}
	};
		
	DasSelector sels[] = {
		
		{"scet", timept_t, OPTIONAL, (const char* []){OP_GE, OP_LT, NULL},
	    "Spacecraft Event Time"},
		
		{"group", string_t, OPTIONAL, (const char* []){OP_EQ, NULL},
		 "The data group from which to pull records, defaults to 'flight'"},
				
		{"dataset", string_t, ENUM|OPTIONAL, (const char*[]){"survey_b",
		 "survey_e","burst_b", "burst_e", NULL}, 
		 "Select the type of L2 data to output.  The 'survey_b' and 'survey_e'\n"
		 "      datasets as well as the upper bands of the 'burst_e' start as spectra\n"
		 "      and are not converted to spectra by the reader.  The 'burst_b' and \n"
		 "      the lower 3 bands of the 'burst_e' datasets have been transmitted to\n"
		 "      the ground as waveforms and must be converted internally. The default\n"
		 "      is dataset is 'survey_e'."
		 },
		 
		{"ascii", bool_t, OPTIONAL, NULL, 
		 "If set to 'true', the reader produces an ascii stream.  By default\n"
		 "a binary stream is output."
		},
		
		{"level", string_t, ENUM|OPTIONAL, 
		(const char*[]){"error","warning","info","debug",NULL},
	   "The logging level on the server, this output is not directly available"
		" to remote clients"
		},
		
		{"ucal", bool_t, OPTIONAL, NULL,
		 "If set to 'true', the reader gathers input packets from *.ucal files,\n"
		 "which have the same record format as the PDS L2 data, but are not\n"
		 "guaranteed to be in time order.  Time revisons in the input files\n"
		 "will cause gaps in the output display."
		},
		
		{NULL}
	};
	
	DasOutput outs[] = {
		{"time", "UTC", 0, NULL, 
		 "Time point at which the capture that generated in a set of frequency\n"
		 "      measurements was initiated.  The times are generated by running the\n"
		 "      the SCLK values through spice, and HSK fine times are used to adjust\n"
		 "      the SCLK prior to conversion" },
				
		{"frequency", "Hz", 0, NULL, "The frequency bin center."},
				
		{"amplitude", "counts", 0, (const char*[]){"frequency","time",NULL},
		 "The instrument uncompressed instrument value in each frequency channel"},
		
		{NULL}
	};
	
	gse_thread_init(GSE_INFO, NULL);
	
	das_parsecmdline(nArgs, sArgs, sels, outs, g_sDesc, g_sFooter);
	
	setLogLevel(das_get_selenum(sels, "level", "info"));
			
	/* Init spice, will need this to convert start/stop times to SCLK and
	   packet SCLKs back to SCET.  All time comparisons are done using SCLK */
	jno_spice_init(JNO_FURNISH);
	
	gse_caltime_t ctBeg = {2012, 9, 1, 245, 0, 0, 0.0};
	das_get_seltime(sels, "scet", OP_GE, &(ctBeg.year), &(ctBeg.month),
			          &(ctBeg.dom), &(ctBeg.doy), &(ctBeg.hour), &(ctBeg.min),
			          &(ctBeg.sec));
	char sSclkBeg[40] = {'\0'};
	snprintf(sSclkBeg, 39, "%.3f", jno_ct2sclk(&ctBeg));
		
	gse_caltime_t ctEnd = {2030, 1, 1, 1, 0, 0, 0.0};
	das_get_seltime(sels, "scet", OP_LT, &(ctEnd.year), &(ctEnd.month),
			          &(ctEnd.dom), &(ctEnd.doy), &(ctEnd.hour), &(ctEnd.min),
			          &(ctEnd.sec));
	char sSclkEnd[40] = {'\0'};
	snprintf(sSclkEnd, 39, "%.3f", jno_ct2sclk(&ctEnd));	
	
	
	const char* sGroup = das_get_selstr(sels, "group", OP_EQ, "flight");
	const char* sDataSet = das_get_selenum(sels, "dataset", "survey_e");
	
	bool bAscii = das_get_selbool(sels, "ascii", false);
	
	
	/* Send a stream header, then setup to catch any exceptions and transmit an
	   exception header */
	GseFileStr* pOut = new_GseFileStr(16384);
	pOut->printf(pOut, "<stream><properties Datum:xTagWidth=\"31.0 s\" /></stream>\r\n");
	printf("[00]%06zd%s", pOut->strlen(pOut), pOut->getstr(pOut));
	pOut->clear(pOut);
	
	/* About to use a bunch of GSE functions, if any of these throw an exception
	   catch it and send it to the client program */
	int nRet           = 0;
	char sExcept[1024] = {'\0'};
	int nPktsSent      = 0;
	chan_2_das_map_t* pMap = 0;
	except_t* pExcept  = NULL;
	static const except_id_t catch_spec[] = { {XCEPT_GROUP_ANY,XCEPT_CODE_ANY}};
	
	
	except_try_push(catch_spec, 1, &pExcept);
	if(pExcept == NULL){
		
		/* Figure out the file type */
		const char* sTypes = "srv.pkt";
		if(das_get_selbool(sels, "ucal", false))
			sTypes = "lrs.ucal";
		
		if( strncmp(sDataSet, "survey_", 7) != 0){
			sTypes = "rec.pkt,bst.pkt";
			if(das_get_selbool(sels, "ucal", false))
				sTypes = "hrs_rec.ucal,hrs_bin.ucal";
		}
	
		GseFileChanMgr* pMgr = 	
			(GseFileChanMgr*)new_JWavesChanMgr(sSclkBeg, sSclkEnd, sTypes, sGroup);
	
		GsePktSrc* pSrc = 
			(GsePktSrc*)new_GseMultiFileSrc(pMgr, jwaves_blocks, at_eof_stop, 10);
	
		HskFineAdj* pAdj = new_HskFineAdj(sGroup, sSclkBeg, sSclkEnd);
	
	
		GsePkt* pPkt = NULL;
		GseXYdataAry* pDataAry = NULL;
		GseXYdata* pData = NULL;
		float8 rSclk = 0L;
		while( (pPkt= pSrc->readNext(pSrc)) != NULL){
			
			pDataAry = fact_JWavesXYdataAry(pPkt, JW_DATA_RAW_FLOATS, 
			                                pSrc->pktIdx(pSrc), pSrc->offset(pSrc));
			
			/* This can happen when we hit MRO's or other non-science data */
			if(pDataAry == NULL){
				pPkt->del(pPkt);
				continue;
			}
			
			for(int iData = 0; iData<pDataAry->num; iData++){	
				pData = pDataAry->pAry[iData];
				
				if( (pMap = getChanMap(sDataSet, maps, pData)) == NULL) 
					continue;
				
				/* Have I sent the headers yet for this channel type?  Note the HFWBR Hi
				   data is odd in that it's frequency table can change */
				if( (! pMap->bSentHdr ) ||
					 ((pMap->nDasPktId == 10)&&(pMap->rLastStartFreq != pData->Ydata(pData)[0])) || 
					 (pMap->uYitems != pData->length(pData))
				  )
					sendHeader(pMap, pData, pOut, bAscii);
				
				rSclk = jwaves_pkt2sclk(pPkt);
				pAdj->adjust(pAdj, &rSclk);
				
				if(sendData(pMap, rSclk, pData, pOut, bAscii))
					nPktsSent += 1;
				
				if( pData->Xtype(pData) == FREQUENCY)
					pMap->rLastStartFreq = pData->Xdata(pData)[0];
			}
		
			pDataAry->del(pDataAry);
			pPkt->del(pPkt);
		}
		
		pSrc->del(pSrc);
		pAdj->del(pAdj);
		
	}
	else{
		gse_sprint_except(sExcept, 1023, pExcept);
		int nLen = strlen(sExcept);
		
		for(int i = 0; i < nLen; i++){ /* Swap double quotes for single quotes */
			if(sExcept[i] == '"'){
				sExcept[i] = '\x27';
			}
		}
		
		pOut->clear(pOut);
		pOut->printf(pOut, "<exception type=\"ReaderError\" message=\"%s\" />\r\n", sExcept);
		printf("[xx]%06zd%s", pOut->strlen(pOut), pOut->getstr(pOut));
		/* Also send the exception to standard error */
		fputs(sExcept, stderr);
		nRet = (pExcept->except_id.except_group * 10) + pExcept->except_id.except_code;
	}
	except_try_pop();
	
	if(nRet != 0){
		pOut->del(pOut);
		return nRet;
	}
	
	if(nPktsSent == 0){
		pOut->clear(pOut);
		pOut->printf(pOut, "<exception type=\"NoDataInInterval\" message=\"No %s data "
		             "found in the interval %s to %s for data group %s.\"/>\r\n",
		             sDataSet, gse_ct2iso_gm(&ctBeg, GSE_MILLISECS),
				       gse_ct2iso_gm(&ctEnd, GSE_MILLISECS), sGroup);
		printf("[xx]%06zd%s", pOut->strlen(pOut), pOut->getstr(pOut));	
	}
	
	pOut->del(pOut);
	return nRet;	
}









