#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <stdio.h>

#ifdef __MINGW32__
#include <argp/argp.h>
#else
#include <argp.h>
#endif

#include <rpwgse/GseInit.h>
#include <rpwgse/GseString.h>
#include <rpwgse/GseFile.h>

#include <jwaves/JWavesChanMgr.h>
#include <jwaves/JWavesPkt.h>
#include <jwaves/JWavesSDSblk.h>
#include <jwaves/JWavesHrbhBlk.h>

#include "jno_spice.h"
#include "JnoPhaser.h"
#include "HskFineAdj.h"
#include "BstCalib.h"
#include "BstRowWriter.h"


/****************************************************************************/
/* Compiled in config file directory location.*/
#ifndef CFG
#error Compiled in default configuration directory is missing.
#endif

#define _QDEF(x) #x
#define QDEF(x) _QDEF(x)

#define JNO_FURNISH "/opt/project/juno/spice/JunoSCLKKernels.list"
const char* g_sKernels = JNO_FURNISH;

/*****************************************************************************/
/* GNG Argp based, command Line argument handling */

const char* argp_program_version = "waves_pds_mkbst $Rev: 11362 $ "
  "$Date: 2013-03-14 13:47:01 -0500 (Thu, 14 Mar 2013) $";
const char* argp_program_bug_address = "<chris-piker@uiowa.edu>";

static char g_args_doc[] = "GROUP";

static char g_opts_doc[] = 
"\n"
"waves_pds_mkbst - Generate Level 3 WAVES_BURST data from L2 data.\n"
"\n"
"Options:\v"
"Environment:\n"
"  WAVES_DATA is taken to be the root directory for the Juno Waves data storage\n"
"  tree.\n"
"\n"
"Files:\n"
"   waves_pds_mkbst expects to find calibration tables in the config directory:\n"
"\n"
"   " QDEF(CFG) "/pds\n"
"\n"
"   Before generating and L3 data, BURST calibration tables in the target\n"
"   volume's CALIB directory are compared against calibration tables in the\n"
"   config directory.  Lower version number tables on the volume are updated\n"
"   if higher version number tables are available from the config location.\n"
"\n"
"   In addition the mission phase and orbit number tables:\n"
"\n"
"   " QDEF(CFG) "/pds/mission_phase.conf\n"
"   " QDEF(CFG) "/pds/orbits.conf\n"
"\n"
"are required to determine output subdirectories under DATA.\n"
"\n"
"Examples:\n"
"   Processing the antena deployment data:\n"
"\n"
"       waves_pds_mkbst -T 2011-08-09,2011-08-11 flight\n"
"\n"
"   Processing an entire SVT run:\n"
"\n"
"      waves_pds_mkbst naked6/svt/2011_08_29_lvco_d5\n";

static struct argp_option g_options[] = {

	{"GROUP", 1, NULL, OPTION_DOC|OPTION_NO_USAGE, 
	 "This program reads L2 product files under: "
	 "$WAVES_DATA/GROUP/JNOWAV_0000.  Here GROUP is the Waves data group. "
	 "For data delivered to SwRI you'll want to use the value 'flight'.\n", 0},

	{"input", 'i', "L2_FILE", 0, "Specify a single input L2_FILE to process, "
    "this can be done because high rate bins are processed as atomic units."
	 "There is no equivalent for low rate data, which is handled on a time "
	 "span basis.\n", 1},
			
	{"kernel-list", 'k', "FILE", 0,
	 "Specify an alternate SPICE kernel list file, the default is\n"
	 JNO_FURNISH "\n", 1},
	 
	{"no-check-l2", 'x', NULL, 0,
	 "Don't check the names of the L2 files against the SCET of the 1st "
	 "packet given by spice.  This may be needed when processing pre-launch "
	 "data using post-launch spice kernels.\n", 1},
							
	{"level", 'l', "LEVEL", 0,
	 "Set the log output verbosity.  Here LEVEL is one of 'critical', 'error',"
	 "'warning', 'notice', 'debug', 'trace' in order if increasing log output "
	 "volume.\n", 1},
			
	{"time", 'T', "TIME1,TIME2", 0,
	 "Instead of file based reading and writing, handle data over a defined "
	 "time period.  TIME1 and TIME2 may be specified in YYYY-DDD or YYYY-MM-DD "
	 "format. These values will be truncated to the nearest day.\n", 1},
	 	 
	{0}
};

typedef struct opts{
	const char* sBeg;
	const char* sEnd;
	int nLevel;
	char* sGroup;
	const char* sKernels;
	const char* sInput;
	bool bCheckL2Times;
}Opts;

static error_t parse_opt(int nKey, char* sArg, struct argp_state* pState)
{
	Opts* pOpts = pState->input;
	gse_caltime_t ctBeg;
	gse_caltime_t ctEnd;
	char* sBeg;
	char* sEnd;
	
	switch(nKey){
		
		/* Pull out the begin an end times */
		case 'T':			  
			sBeg = CS(sArg);
			sEnd = strchr(sBeg, ',');

			if(sEnd != NULL){
				*sEnd = '\0';
				sEnd += 1;
			}
			if(!sEnd || (sEnd == '\0') || (sBeg == '\0') ){
				fprintf(stderr, "ERROR: Bad time range for -T argument: '%s'\n", sArg);
				return EINVAL;
			}
			
			if(!gse_str2ct_gm(sBeg, &ctBeg, false) ||
				!gse_str2ct_gm(sEnd, &ctEnd, false))  {
				fprintf(stderr, "ERROR: Bad time range for -T argument: '%s'\n", sArg);
				return EINVAL;
			}
			
			free(sBeg);
			
			if(ctBeg.hour != 0 || ctBeg.min != 0 || ctBeg.sec != 0.0){
				fprintf(stderr, "WARNING: Sub-day time resolution dropped from "
				        "beginning time.\n");
			}
			pOpts->sBeg = NS("%04d-%02d-%02d", ctBeg.year, ctBeg.month, ctBeg.dom);
			
			if(ctEnd.hour != 0 || ctEnd.min != 0 || ctEnd.sec != 0.0){
				fprintf(stderr, "WARNING: Sub-day time resolution dropped from "
				        "ending time.\n");
			}
			pOpts->sEnd = NS("%04d-%02d-%02d", ctEnd.year, ctEnd.month, ctEnd.dom);
			
			break;
			
		case 'l':
			switch(sArg[0]){
				case 'c': pOpts->nLevel = GSE_CRITICAL; break;
				case 'e': pOpts->nLevel = GSE_ERROR;    break;
				case 'w': pOpts->nLevel = GSE_WARNING;  break;
				case 'n': case 'i': pOpts->nLevel = GSE_NOTICE; break;
				case 'd': pOpts->nLevel = GSE_DEBUG;    break;
				case 't': pOpts->nLevel = GSE_TRACE;    break;
				default:
					return EINVAL;
			}
			break;
			
		case 'k': pOpts->sKernels = sArg; break;
		
		case 'x': pOpts->bCheckL2Times = false; break;
		
		case 'i': pOpts->sInput = sArg; break;
			
		case ARGP_KEY_ARG:
			pOpts->sGroup = sArg;
			
			if(pState->next < pState->argc)
				return ARGP_ERR_UNKNOWN;
			
			break;
		
		default:
			return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static struct argp g_parser = {g_options, parse_opt, g_args_doc, g_opts_doc};

/*****************************************************************************/
/* Use the JWaves Channel Manger to scoop these up */

GseStrList* getHrsByTime(const char* sBeg, const char* sEnd, const char* sGroup)
{
	GseStrList* pList = NULL;
	GseFileChanMgr* pMgr = NULL;
	const char* sPath = NULL;
	int4 iChan = -1, nChan = 0;
	
	pMgr = (GseFileChanMgr*) new_JWavesChanMgr(sBeg, sEnd, "rec.pkt,bst.pkt",
	                                           sGroup);
	pList = new_GseStrList();
	
	nChan = pMgr->channels(pMgr);
	for(iChan = 0; iChan < nChan; iChan++){
		sPath = NULL;
		while( (sPath = pMgr->nextFile(pMgr, iChan, sPath)) )
			pList->append(pList, CS(sPath));
	}
		
	pMgr->del(pMgr); 
	pMgr = NULL;
	
	if(pList->length(pList) > 0){
		pList->sort(pList, strcmp);
	}
	else{
		pList->del(pList);
		pList = NULL;
	}
	
	return pList;
}

/*****************************************************************************/
/* Utility Function: Get an SCET from the filename */

bool ct_from_l2_name(const char* sPath, gse_caltime_t* pCt, bool bThrow)
{
	char sBase[128] = {'\0'};
	char sTmp[10] = {'\0'};
	int4 nTmp, i;
	
	if((sBase == NULL)||(pCt == NULL)) gse_throw(GSE_XCEPT, GSE_XNULLPTR);
	
	if(! gse_basename(sPath, sBase, 127) ){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't get the basename of %s", sPath);
		}
		else{
			gse_error_v("Couldn't get the basename of %s", sPath);
			return false;
		}
	}
	
	if(strlen(sBase) < 18){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Basename %s is too short to contain"
			             " the WAV_ prefix plus the SCET to the second", sBase);
		}
		else{
			gse_error_v("Basename %s is too short to contain"
			            " the WAV_ prefix plus the SCET to the second", sBase);
			return false;
		}
	}
	
	/* Year */
	for(i=0;i<4;i++) sTmp[i] = sBase[i+4];
	sTmp[4] = '\0';
	if(!gse_str2int4(sTmp, &(pCt->year))){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert %s in %s to a "
			            "year value", sTmp, sBase);
		}
		else{
			gse_error_v("Couldn't convert %s in %s to a year value", sTmp, sBase);
			return false;
		}
	}
	
	/* Day of year */
	sTmp[0] = sBase[8]; sTmp[1] = sBase[9]; sTmp[2] = sBase[10];  sTmp[3] = '\0';
	if(!gse_str2int4(sTmp, &(pCt->doy))){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert %s in %s to a "
			             "day of year value", sTmp, sBase);
		}
		else{
			gse_error_v("Couldn't convert %s in %s to a day of year value", sTmp, sBase);
			return false;
		}
	}
				
	/* Hour of day */
	sTmp[0] = sBase[12]; sTmp[1] = sBase[13]; sTmp[2] = '\0';
	if(!gse_str2int4(sTmp, &(pCt->hour))){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert %s in %s to an "
			             "hour of day value", sTmp, sBase);
		}
		else{
			gse_error_v("Couldn't convert %s in %s to an hour of day value", sTmp, sBase);
			return false;
		}
	}

	/* minute of year */
	sTmp[0] = sBase[14]; sTmp[1] = sBase[15]; sTmp[3] = '\0';
	if(!gse_str2int4(sTmp, &(pCt->min))){
		if(bThrow){
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert %s in %s to a "
			             "minute of hour value", sTmp, sBase);
		}
		else{
			gse_error_v("Couldn't convert %s in %s to a minute of hour value", sTmp, sBase);
			return false;
		}
	}
	
	/* Second of minute */
	sTmp[0] = sBase[16]; sTmp[1] = sBase[17]; sTmp[3] = '\0';
	if(!gse_str2int4(sTmp, &nTmp)){
		if(bThrow){			
			gse_throw_fm(GSE_XCEPT, GSE_XBADARG, "Couldn't convert %s in %s to a "
			             "second of minute value", sTmp, sBase);
		}
		else{
			gse_error_v("Couldn't convert %s in %s to a second of minute value", sTmp, sBase);
			return false;
		}
	}
			
	pCt->sec = (float8)nTmp;
	
	gse_ct_norm_doy(pCt);
	
	return true;
}

/*****************************************************************************/
/* Checking input files before processing */

bool checkL2BurstFile(HskFineAdj* pAdj, const char* sPath,
		                gse_caltime_t* pCtFile, bool bCheckTimes)
{	
	char sBase[128] = {'\0'};
	GsePktSrc* pSrc          = NULL;
	bool bRet                = false;
	GsePkt* pPkt             = NULL;
	JWavesSDSblk* pSci = NULL;
	JWavesHrbhBlk* pBlk = NULL;
	gse_caltime_t ctData     = {0};
	float8 dSclk             = -1.0;
	float8 dFullRes          = 0.0;
			
	if(!ct_from_l2_name(sPath, pCtFile, false) )
		goto checkl2_done;
	
	/* read the 1st packet */
	pSrc = (GsePktSrc*)new_GseFilePktSrc(sPath, jwaves_blocks, at_eof_stop);
	
	/* Get a short name for reporting, if ctFromL2Filename stops using
	   gse_basename the put extra checks here. */
	gse_basename(sPath, sBase, 127);
	
	pPkt = pSrc->readNext(pSrc);
	if(!pPkt){
		gse_error_v("%s: Can't read the 1st packet", sBase);
		goto checkl2_done;
	}
	
	if( (pPkt->pHdr->uContent & 0xF0) != 0x80){
		gse_error_v("%s: First packet is not science data", sBase);
		goto checkl2_done;
	}
	
	if( (pPkt->pHdr->uContent & 0x0F) < 0x04){
		gse_error_v("%s: First packet contains raw data", sBase);
		goto checkl2_done;
	}
		
	if( ! pPkt->hasUniqueBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0) ){
		gse_error_v("%s: First packet doesn't have a science data status block.", 
				      sBase);
		goto checkl2_done;
	}
	
	pSci = (JWavesSDSblk*) pPkt->getBlkByType(pPkt, JW_SDS_BLK_TYPE_ID, 0);
	
	dSclk = pSci->uColSCLK;
	dSclk += ((float8) pSci->uColRTI) / 40.0;
	
	/* Try to adjust the sclk using the fine time array */
	if(! pAdj->adjust(pAdj, &dSclk) ){
		gse_error_v("%s: Can't get fine time adjustment.", sBase);
		goto checkl2_done;
	}
	
	jno_sclk2ct(dSclk, &ctData);
	
	/* Compare truncated seconds */
	dFullRes = ctData.sec;
	ctData.sec = (int4)ctData.sec;
	
	if(bCheckTimes && (gse_ctcmp(pCtFile, &ctData) != 0)){
		gse_error_v("%s: SCET mismatch, from filename: %04d-%03dT%02d:%02d:%02.0f"
		            " from 1st pkt: %04d-%03dT%02d:%02d:%06.3f.", sBase,
		            pCtFile->year, pCtFile->doy, pCtFile->hour, pCtFile->min, pCtFile->sec,
		            ctData.year, ctData.doy, ctData.hour, ctData.min, dFullRes);
		goto checkl2_done;
	}
	
	/* Check to seed if qFactor existance matches the filename */
	if( ! pPkt->hasUniqueBlkByType(pPkt, JW_HRS_BIN_HDR_ID, 0) ){
		
		if(strstr(sBase, "REC") == NULL) {	
			gse_error_v("%s: First packet of non-record file doesn't have a "
		            "high-rate bin header block.",  sBase);
			goto checkl2_done;
		}
	}
	else{
	
		pBlk = (JWavesHrbhBlk*) pPkt->getBlkByType(pPkt, JW_HRS_BIN_HDR_ID, 0);
	
		if((pBlk->uQF == 0) && (strstr(sBase, "REC") == NULL)){
			gse_error_v("%s: First packet has QFactor = 0, but 'REC' is not part "
			            " of the filename.", sBase);
			goto checkl2_done;
		}
	
		if((pBlk->uQF != 0) && (strstr(sBase, "BST") == NULL)){
			gse_error_v("%s: First packet has QFactor > 0, but 'BST' is not part "
			            " of the filename.", sBase);
			goto checkl2_done;
		}
	}
		
	/* all checks passed */
	bRet = true;
	
checkl2_done:
	if(pPkt != NULL) pPkt->del(pPkt);
	if(pSrc != NULL) pSrc->del(pSrc);
	return bRet;
}

/*****************************************************************************/

int main(int argc, char** argv)
{
	int4 nRet             = 0;
	GseStrList* pFiles    = NULL;
	const char* sPath     = NULL;
	char* sVolRoot        = NULL;
	JnoPhaser* pPhaser    = NULL;
	HskFineAdj* pAdj      = NULL;
	BstCalib* pCalib      = NULL;
	gse_caltime_t ctFile  = {0};
	BstRowWriter* pWriter = NULL;
	GsePktSrc* pSrc       = NULL;
	GsePkt* pPkt          = NULL;
	BstRowAry* pRowAry    = NULL;
	int4 i                = 0;
	
	Opts opts = {"2000-01-02", "2068-01-19", GSE_NOTICE, NULL, g_sKernels, 
	             NULL, true};
	
	gse_thread_init(GSE_ERROR, NULL);
	argp_parse(&g_parser, argc, argv, ARGP_IN_ORDER, 0, &opts);
	
	gse_log_setlevel(opts.nLevel);
	
	/* Init spice */
	jno_spice_init(opts.sKernels);
	
	/* Check for the data group */
	if(!opts.sGroup){
		gse_error("GROUP parameter is missing.  Use --help for more info.");
		return 1;
	}
	
	/* Check for the WAVES_DATA env */
	if(getenv("WAVES_DATA") == NULL){
		gse_error("Environment variable WAVES_DATA is not defined.");
		return 3;
	}
	
	/* Set up the umask so that PDS data can be read on the web */
	gse_notice("Setting process umask to 0002");
	umask( S_IWOTH );
	
	sVolRoot = NS("%s/%s/JNOWAV_1000", getenv("WAVES_DATA"), opts.sGroup);
	gse_notice_v("Output volume root: %s", sVolRoot);
	
	
	/* Set up a phaser object to handle output directory names */
	pPhaser = new_JnoPhaser();
	
	/* Set up a fine time adjuster to get accurate packet times */
	pAdj = new_HskFineAdj(opts.sGroup, opts.sBeg, opts.sEnd);
	
	
	/* Set up the calibrator to write files */
	pCalib = new_BstCalib(opts.sGroup, pAdj);
	
	/* Get the list of input files */
	if(opts.sInput){
		pFiles = new_GseStrList();
		pFiles->append(pFiles, gse_abspath(opts.sInput));
	}
	else{
		pFiles = getHrsByTime(opts.sBeg, opts.sEnd, opts.sGroup);
	}
	
	if(pFiles == NULL){
		gse_notice_v("No high rate for %s to %s in the %s group\n", 
				       opts.sBeg, opts.sEnd, opts.sGroup);
		return nRet;        
	}
	
	/* Main Loop */
	if( (sPath = pFiles->get(pFiles, 0) ) != NULL){
	
		do{ 
			gse_notice_v("%s: Processing...", sPath);
			
			if(!checkL2BurstFile(pAdj, sPath, &ctFile, opts.bCheckL2Times)){
				nRet |= 0x2;
			}		
			else{
								
				/* Make an input source */
				pSrc = (GsePktSrc*)new_GseFilePktSrc(sPath, jwaves_blocks,
						                               at_eof_stop);
				
				/* Make a new file writer */
				pWriter = new_BstRowWriter();
				
				/* loop reading packets */
				while( (pPkt = pSrc->readNext(pSrc)) != NULL){
					
					if( (pRowAry = pCalib->cal(pCalib, pPkt)) != NULL){ ;
					
						for(i = 0; i<pRowAry->num; i++){
						
						/* If writer keeps the data, null the ptr so it's not deleted */
							if(pWriter->add(pWriter, pRowAry->pAry[i]))
								pRowAry->pAry[i] = NULL;
						}
						
						pRowAry->del(pRowAry);
					}
					pPkt->del(pPkt);
				}
				
				pSrc->del(pSrc);         /* Free the source */
				
				/* Write, update or retain the high rate files, 
				   Note: The writer is going to have to buffer everything
					      as the session start-stop time goes into the 
							header section of each file. */
				
				pWriter->writeAll(pWriter, opts.sGroup, pPhaser, &ctFile);
				pWriter->del(pWriter);   /* Free the writer */
			}
		}
		while( (sPath = pFiles->next(pFiles) ) != NULL);
		
	}
	
	pFiles->del(pFiles);
	pCalib->del(pCalib);
	pAdj->del(pAdj);
	pPhaser->del(pPhaser);
	free(sVolRoot);
	
	return nRet;
}
