#define _POSIX_C_SOURCE 200112

#ifdef GSE_POSIX_THREADS
#include <pthread.h>
#endif

#include <stdio.h>

#ifdef __MINGW32__
#include <argp/argp.h>
#else
#include <argp.h>
#endif

#include <strings.h>

#include <rpwgse/Gse.h>
#include <jwaves/JWaves.h>

#include "jno_spice.h"

#include "HskFineAdj.h"      /* Calibrating data */
#include "SrvCalib.h"
#include "SrvRower.h"

#include "JnoPhaser.h"       /* Writing data */
#include "SrvRowWriter.h"

/****************************************************************************/
/* Compiled in config file directory location.*/
#ifndef CFG
#error Compiled in default configuration directory is missing.
#endif

#define _QDEF(x) #x
#define QDEF(x) _QDEF(x)

#define JNO_FURNISH "/opt/project/juno/spice/JunoSCLKKernels.list"
const char* g_sKernels = JNO_FURNISH;

/*****************************************************************************/
/* GNG Argp based, command Line argument handling */

const char* argp_program_version = "waves_pds_mksrv $Rev: 11289 $ "
  "$Date: 2012-02-14 10:16:07 -0600 (Tue, 14 Feb 2012) $";
const char* argp_program_bug_address = "<chris-piker@uiowa.edu>";

static char g_args_doc[] = "GROUP";

static char g_opts_doc[] = 
"\n"
"waves_pds_mksrv - Generate Level 3 WAVES_SURVEY data from L2 data.\n"
"\n"
"Options:\v"
"Environment:\n"
"  WAVES_DATA is taken to be the root directory for the Juno Waves data storage\n"
"  tree.\n"
"\n"
"Files:\n"
"   waves_pds_mksrv expects to find calibration tables in the config directory:\n"
"\n"
"   " QDEF(CFG) "/pds\n"
"\n"
"   Before generating any L3 data, SURVEY mode calibration tables in the target\n"
"   volume's CALIB directory are compared against calibration tables in the \n"
"   config directory.  Lower version number tables on the volume are updated if\n"
"   higher version number tables are available from the config location.\n"
"\n"
"   In addition the mission phase and orbit number tables:\n"
"\n"
"   " QDEF(CFG) "/pds/mission_phase.conf\n"
"   " QDEF(CFG) "/pds/orbits.conf\n"
"\n"
"are required to determine output subdirectories under DATA.\n"
"\n"
"Examples:\n"
"   Processing the antena deployment data:\n"
"\n"
"      waves_pds_mksrv -T 2011-08-09,2011-08-11 flight\n"
"\n"
"   Processing an entire SVT run:\n"
"\n"
"      waves_pds_mksrv naked6/svt/2011_08_29_lvco_d5\n";

static struct argp_option g_options[] = {

	{"GROUP", 1, NULL, OPTION_DOC|OPTION_NO_USAGE, 
	 "This program reads L2 product files under: "
	 "$WAVES_DATA/GROUP/JNOWAV_0000.  Here GROUP is the Waves data group. "
	 "For data delivered to SwRI you'll want to use the value 'flight'.\n", 0},

	{"kernel-list", 'k', "FILE", 0,
	 "Specify an alternate SPICE kernel list file, the default is\n"
	 JNO_FURNISH "\n", 1},
							
	{"level", 'l', "LEVEL", 0,
	 "Set the log output verbosity.  Here LEVEL is one of 'critical', 'error',"
	 "'warning', 'notice', 'debug', 'trace' in order if increasing log output "
	 "volume.\n", 1},
			
	{"time", 'T', "TIME1,TIME2", 0,
	 "Instead of file based reading and writing, handle data over a defined "
	 "time period.  TIME1 and TIME2 may be specified in YYYY-DDD or YYYY-MM-DD "
	 "format.  These values will be trunkated to the nearest day.\n", 1},
	 
	{0}
};

typedef struct opts{
	const char* sBeg;
	const char* sEnd;
	int nLevel;
	char* sGroup;
	const char* sKernels;
}Opts;

static error_t parse_opt(int nKey, char* sArg, struct argp_state* pState)
{
	Opts* pOpts = pState->input;
	gse_caltime_t ctBeg;
	gse_caltime_t ctEnd;
	char* sBeg;
	char* sEnd;
	
	switch(nKey){
		
		/* Pull out the begin an end times */
		case 'T':			  
			sBeg = CS(sArg);
			sEnd = strchr(sBeg, ',');

			if(sEnd != NULL){
				*sEnd = '\0';
				sEnd += 1;
			}
			if(!sEnd || (sEnd == '\0') || (sBeg == '\0') ){
				fprintf(stderr, "ERROR: Bad time range for -T argument: '%s'\n", sArg);
				return EINVAL;
			}
			
			if(!gse_str2ct_gm(sBeg, &ctBeg, false) ||
				!gse_str2ct_gm(sEnd, &ctEnd, false))  {
				fprintf(stderr, "ERROR: Bad time range for -T argument: '%s'\n", sArg);
				return EINVAL;
			}
			
			free(sBeg);
			
			if(ctBeg.hour != 0 || ctBeg.min != 0 || ctBeg.sec != 0.0){
				fprintf(stderr, "WARNING: Sub-day time resolution dropped from "
				        "beginning time.\n");
			}
			pOpts->sBeg = NS("%04d-%02d-%02d", ctBeg.year, ctBeg.month, ctBeg.dom);
			
			if(ctEnd.hour != 0 || ctEnd.min != 0 || ctEnd.sec != 0.0){
				fprintf(stderr, "WARNING: Sub-day time resolution dropped from "
				        "ending time.\n");
			}
			pOpts->sEnd = NS("%04d-%02d-%02d", ctEnd.year, ctEnd.month, ctEnd.dom);
			
			break;
			
		case 'l':
			switch(sArg[0]){
				case 'c': pOpts->nLevel = GSE_CRITICAL; break;
				case 'e': pOpts->nLevel = GSE_ERROR;    break;
				case 'w': pOpts->nLevel = GSE_WARNING;  break;
				case 'n': case 'i': pOpts->nLevel = GSE_NOTICE; break;
				case 'd': pOpts->nLevel = GSE_DEBUG;    break;
				case 't': pOpts->nLevel = GSE_TRACE;    break;
				default:
					return EINVAL;
			}
			break;
			
		case 'k': pOpts->sKernels = sArg; break;
			
		case ARGP_KEY_ARG:
			pOpts->sGroup = sArg;
			
			if(pState->next < pState->argc)
				return ARGP_ERR_UNKNOWN;
			
			break;
		
		default:
			return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static struct argp g_parser = {g_options, parse_opt, g_args_doc, g_opts_doc};

/*****************************************************************************/

int main(int argc, char** argv)
{
	SrvCalib*         pCalib = NULL;
	GseFileChanMgr*   pMgr = NULL;
	GsePktSrc*        pSrc = NULL;
	GsePkt*           pPkt = NULL;
	JnoPhaser*        pPhaser = NULL;
	HskFineAdj*       pAdj = NULL;
	SrvRower*         pRower = NULL;
	srv_row_t*        pRow = NULL;
	SrvRowList*       pRows = NULL;
	StrSrvWriterDict* pWriters = NULL;
	SrvRowWriter*     pWriter = NULL;
	const char*       sFileType = NULL;
	int4              nSecIn = 0;
	int4              nSecOut = 0;
	
	gse_thread_init(GSE_ERROR, NULL);
	
	Opts opts = {"2000-01-02", "2068-01-19", GSE_NOTICE, NULL, g_sKernels};
	argp_parse(&g_parser, argc, argv, ARGP_IN_ORDER, 0, &opts);
	
	gse_log_setlevel(opts.nLevel);
	
	/* Init spice */
	jno_spice_init(opts.sKernels);
	
	/* Check for the data group */
	if(!opts.sGroup){
		gse_error("GROUP parameter is missing.  Use --help for more info.");
		return 1;
	}
	
	/* Check for the WAVES_DATA env */
	if(getenv("WAVES_DATA") == NULL){
		gse_error("Environment variable WAVES_DATA is not defined.");
		return 3;
	}
	
	/* Set up the umask so that PDS data can be read on the web */
	gse_notice("Setting process umask to 0002");
	umask( S_IWOTH );
	
	gse_notice_v("Output volume root: %s/%s/JNOWAV_1000", getenv("WAVES_DATA"),
	             opts.sGroup);
	
	/* Make a new calibrator */
	pCalib = new_SrvCalib(opts.sGroup);
	
	gse_notice_v("Input volume root: %s/%s/JNOWAV_0000", getenv("WAVES_DATA"),
	             opts.sGroup);
	
	/* Make a packet source */
	pMgr = (GseFileChanMgr*)
			  new_JWavesChanMgr(opts.sBeg, opts.sEnd, "srv.pkt", opts.sGroup);
	pSrc = (GsePktSrc*)new_GseMultiFileSrc(pMgr, jwaves_blocks, at_eof_stop, 10);
	
	/* Load the fine time alteration table */
	pAdj = new_HskFineAdj(opts.sGroup, opts.sBeg, opts.sEnd);
	
	if(gse_log_getlevel() <= GSE_DEBUG)
		pAdj->prnFines(pAdj, stdout);
	
	/* Make an Survey data row creator */
	pRower = new_SrvRower(pAdj, pCalib);
	
	/* pRower->bDiag = true; */
	
	/* Make an object to handle mission phases and orbit numbers */
	pPhaser = new_JnoPhaser();
	
	/* Print diagnostics if at debug level */
	if(gse_log_getlevel() <= GSE_DEBUG){
		pPhaser->prnPhases(pPhaser, stdout);
		pPhaser->prnOrbits(pPhaser, stdout);
	}
	
	/* A dictionary for looking up row writers by filetype */
	pWriters = new_StrSrvWriterDict(NULL);
		
	/* Main reading/writing loop */
	while( (pPkt = pSrc->readNext(pSrc)) != NULL){
	
		if(pRower->addPkt(pRower, pPkt)){
		
			if( (pRows = pRower->getRows(pRower)) != NULL){
		
				for(pRow = pRows->get(pRows, 0); pRow != NULL; pRow = pRows->next(pRows)){
								
					/* See if this row-type has a writer, if not make one */
					if( !pWriters->hasKey(pWriters, pRow->sRowType) ){
						pWriter = new_SrvRowWriter(pPhaser, opts.sGroup);
						pWriters->insert(pWriters, CS(pRow->sRowType), pWriter);
					}
					
					pWriter = pWriters->getMutable(pWriters, pRow->sRowType);
					pWriter->writeRow(pWriter, pRow);
				}
				pRows->del(pRows);
			}
		}
		pPkt->del(pPkt);
	}
	
	/* No new packets for the time period, set the rower to flush mode,
	   and read out rows in progress. */
	pRower->setFlush(pRower);
	while( (pRows = pRower->getRows(pRower)) != NULL){
		
		for(pRow = pRows->get(pRows, 0); pRow != NULL; pRow = pRows->next(pRows)){
			
			if( !pWriters->hasKey(pWriters, pRow->sRowType) ){
				pWriter = new_SrvRowWriter(pPhaser, opts.sGroup);
				pWriters->insert(pWriters, CS(pRow->sRowType), pWriter);
			}
			pWriter = pWriters->getMutable(pWriters, pRow->sRowType);
			pWriter->writeRow(pWriter, pRow);
		}
		pRows->del(pRows);
	}
	
	nSecIn = pRower->inPktSecs(pRower);
	
	/* Tell the writers that no new rows are on the way and to finish up the
	   files they are working on */
	for(sFileType = pWriters->first(pWriters); sFileType != NULL; 
		 sFileType = pWriters->next(pWriters, sFileType)){
		
		pWriter = pWriters->getMutable(pWriters, sFileType);
		pWriter->finish(pWriter);
		nSecOut += pWriter->secsWritten(pWriter);
	}
	
	/* Clean memory so that mtrace can find leaks in areas that matter */
	pWriters->del(pWriters);
	
	pRower->del(pRower);
	pAdj->del(pAdj);
	pPhaser->del(pPhaser);
	pSrc->del(pSrc);
	pCalib->del(pCalib);
	
	int nRet = 0;
	/* printf("%d packets read, %d packets written", nSecIn, nSecOut); */
	gse_notice_v("%d sweeps read, %d sweeps written", nSecIn, nSecOut);
	
	if(nSecIn != nSecOut){
		nRet = 7;
		gse_error("Input sweeps != Output sweeps");
	}
	
	return nRet;
}


